/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.10 $
 * $Date: 2007/11/21 20:18:23 $
 * $Author: sgop $
 */

#include <string.h>

#include "action_item.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  char* section;
  action_item_t* action;
} user_action_t;

GList* Actions = NULL;
static GList** UserActions = NULL;
static int Size = 0;

/** Generic function to create new actions.
 * @param name if parent is NULL then this is the name for later access, otherwise it is the label.
 * @param parent the parent actions or NULL for a toplevel node.
 * @param flags the flags (see the action flags above)
 * @param action the activation function
 * @param check the function which selects the process-able objects from the real selection
 */
action_item_t*
action_item_new(const char* name, action_item_t* parent, int flags,
                ActionItemFunc action, ActionItemCheck check) {
  action_item_t* ai;

  g_assert(parent || name);
  
  ai = g_malloc0(sizeof(*ai));
  ai->name = name?g_strdup(name):NULL;
  ai->flags = flags;
  ai->action = action;
  ai->check = check;
  ai->children = NULL;
  ai->user_data = NULL;
  ai->link = NULL;

  if (parent)
    parent->children = g_list_append(parent->children, ai);
  else
    Actions = g_list_append(Actions, ai);
  return ai;
}

/** Separator. Wrapper for action_item_new().
 * @param parent the parent action.
 * @return the new action item
 * @note these actions should not have any children
 */
action_item_t* action_item_sep(action_item_t* parent) {
  return action_item_new(NULL, parent, 0, NULL, NULL);
}

/** A root action. Wrapper for action_item_new()
 * @param name the name for later access
 * @return the new root action item
 */
action_item_t* action_item_root(const char* name) {
  return action_item_new(name, NULL, 0, NULL, NULL);
}

/** A simple action. Wrapper for action_item_new().
 * @param name the label for the action.
 * @param parent the parent.
 * @param flags the flags
 * @param action the activation function
 * @return the new action item
 */
action_item_t* action_item_simple(const char* name,
                                  action_item_t* parent,
                                  int flags, ActionItemFunc action) {
  return action_item_new(name, parent, flags, action, NULL);
}

/**
 * Link to other actions. Wrapper for action_item_new()
 * @param name the name of the actions
 * @param parent the parent action
 * @param link the linker function
 * @param embed whether or not to embed the linked actions instead of
 * creating a submenu (not functional atm)
 * @return the new action item
 */
action_item_t* action_item_sub(const char* name, action_item_t* parent,
                               ActionItemLink link, gboolean embed)
{
    action_item_t* item;
    item = action_item_new(name, parent, 
                           embed?AI_LINK_EMBED:0, NULL, NULL);
    item->link = link;
    return item;
}

static action_item_t* on_get_sub_actions(GList* selection, const char* data)
{
    (void)selection;
    return action_item_search(data);
}

/**
 * Link to other actions as submenu. Wrapper for action_item_new()
 * @param name the name of the actions
 * @param parent the parent action
 * @param subname the name of the sub actions, must be a static string
 * @return the new action item
 */
action_item_t* action_item_link(const char* name, action_item_t* parent,
                                const char* subname)
{
    action_item_t* item;
    item = action_item_new(name, parent, AI_LOCAL_DATA, NULL, NULL);
    item->user_data = (void*)subname;
    item->link = (ActionItemLink)on_get_sub_actions;
    return item;
}

/**
 * An internal node with will have some subactions. Wrapper for
 * action_item_new().
 * @param name the name of the actions
 * @param parent the parent action
 * @return the new action item
 */
action_item_t* action_item_node(const char* name, action_item_t* parent) {
  return action_item_new(name, parent, 0, NULL, NULL);
}

/** Destroy actions. Destroys the action and all its children.
 * @param ai the action item
 */
void action_item_destroy(action_item_t* ai) {
  GList* dlist;

  for (dlist = ai->children; dlist; dlist = dlist->next) {
    action_item_t* item = dlist->data;
    action_item_destroy(item);
  }
  g_list_free(ai->children);
  g_free(ai);
  //FIXME: handle the case that this action has a parent
  Actions = g_list_remove(Actions, ai);
}

/** Search for actions.
 * @param name the name of the actions
 * @return the actions if found, or NULL
 */
action_item_t* action_item_search(const char* name)
{
    GList* dlist;
    
    for (dlist = Actions; dlist; dlist = dlist->next)
    {
        action_item_t* item = dlist->data;
        if (!strcmp(item->name, name)) return item;
    }
    g_log("CORE", G_LOG_LEVEL_WARNING,
          "action item '%s' not found", name);
    g_assert(FALSE);
    return NULL;
}

static action_item_t* action_user_find(int ID, const char* section) {
  GList* dlist;
  if (ID < 0 || ID >= Size) return NULL;
  if (!UserActions || !UserActions[ID]) return NULL;
  for (dlist = UserActions[ID]; dlist; dlist = dlist->next) {
    user_action_t* action = dlist->data;
    if (!strcmp(action->section, section))
      return action->action;
  }
  return NULL;
}

static void action_user_register(int ID, const char* section,
                                 action_item_t* item) {
  action_item_t* action;
  user_action_t* ua;

  if (ID >= Size) {
    int i1;
    UserActions =
      g_realloc(UserActions, sizeof(*UserActions)*(ID+1));
    for (i1 = Size; i1 < ID+1; i1++) UserActions[i1] = NULL;
    Size = ID+1;
  }

  action = action_user_find(ID, section);
  if (action) {
    g_log("CORE", G_LOG_LEVEL_WARNING,
          "user actions for plugin '%d' and section '%s' already defined",
          ID, section);
    return;
  }
  ua = g_malloc(sizeof(*ua));
  ua->section = g_strdup(section);
  ua->action = item;
  UserActions[ID] = g_list_append(UserActions[ID], ua);
}

action_item_t* user_action_upload_find(int ID) {
  return action_user_find(ID, "Upload");
}
action_item_t* user_action_result_find(int ID) {
  return action_user_find(ID, "Result");
}
action_item_t* user_action_source_find(int ID) {
  return action_user_find(ID, "Source");
}
action_item_t* user_action_browse_find(int ID) {
  return action_user_find(ID, "Browse");
}

/** Register actions which handle users from uploads
 * @param ID the plugin which provides the actions
 * @param item the actions
 */
void user_action_upload_register(int ID, action_item_t* item) {
  action_user_register(ID, "Upload", item);
}

/** Register actions which handle users from search results
 * @param ID the plugin which provides the actions
 * @param item the actions
 */
void user_action_result_register(int ID, action_item_t* item) {
  action_user_register(ID, "Result", item);
}

/** Register actions which handle users from download sources
 * @param ID the plugin which provides the actions
 * @param item the actions
 */
void user_action_source_register(int ID, action_item_t* item) {
  action_user_register(ID, "Source", item);
}

/** Register actions which handle users from browse results
 * @param ID the plugin which provides the actions
 * @param item the actions
 */
void user_action_browse_register(int ID, action_item_t* item) {
  action_user_register(ID, "Browse", item);
}


#ifdef WIN32
# define _ACTION_ITEM_C_
# include "lcorealiasdef.c"
#endif
