/* Copyright (C) 2005 sgop@users.sourceforge.net
   This is free software distributed under the terms of the
   GNU Public License.  See the file COPYING for details. */

#ifndef _ACTION_ITEM_H_
#define _ACTION_ITEM_H_

#include <glib.h>

G_BEGIN_DECLS

/** @defgroup ActionModule Actions to build up menus
 * 
 * Action items are good for defining context sensitive menus. The
 * GUI can implement a framework to convert the actions to popup
 * menus. The actions are not restricted to one single object, they
 * can operate on multiple selections. This way it is very easy to
 * define a menu for the GUI.
 *
 * <b>Example:</b>
 * Defining actions:
 * <div style='border:1px solid #aaa;padding:10px;background-color:#eee'><pre>
 * static action_item_t* nap_actions_user_menu(const char* name) {
 *   action_item_t* item = action_item_root(name);
 *   action_item_simple("Open chat...", item, 0, (ActionItemFunc)on_nap_chat);
 *   action_item_simple("Send message...", item, AI_RESTRICT_1, NULL);
 *   action_item_sep(item);
 *   action_item_simple("Whois user", item, 0, (ActionItemFunc)on_nap_whois);
 *   action_item_new("Add to buddies", item, 0,
 *                   (ActionItemFunc)on_nap_buddy,
 *                   (ActionItemCheck)on_nap_check_not_buddy);
 *   action_item_simple("Browse user", item, 0, (ActionItemFunc)on_nap_browse);
 * }
 * </pre></div>
 *
 * Using the conversion function to extract the user from different objects:
 * <div style='border:1px solid #aaa;padding:10px;background-color:#eee'><pre>
 *   action_item_t* item;
 *   item = nap_actions_user_menu("Upload nap users actions");
 *   item->convert = (ActionObjectConvert)nap_upload_to_nu;
 *   user_action_upload_register(NAP_ID, item);
 *   item = nap_actions_user_menu("Source nap users actions");
 *   item->convert = (ActionObjectConvert)nap_source_to_nu;
 *   user_action_source_register(NAP_ID, item);
 * </pre></div>
 *
 * @ingroup BasicFramework
 * @{
 */

/** @name Action item flags */
/*@{*/
#define AI_GLOBAL     0x01 ///< The action doesn't need an object to operate on. It can be performed any time.
#define AI_RESTRICT_1 0x02 ///< The action only operates on one single object.
#define AI_HIDE_0     0x04 ///< Don't display the action if there is no object selected that can be processed.
#define AI_LINK_EMBED 0x08 ///< Embed this action in the parent instead of building a submenu (not functional atm)
#define AI_CHECK      0x10 ///< The action represents a check button
#define AI_RADIO      0x20 ///< The action represents a radio button
#define AI_LOCAL_DATA 0x40 ///< The action has its local data, it does not inherit its parents data.
#define AI_TOGGLE     (AI_CHECK|AI_RADIO)
/*@}*/

#define ACTION_PROCESS_LIST(func, type) static void       \
    on_ ## func (GList* list) {                           \
        GList* dlist;																			\
        for (dlist = list; dlist; dlist = dlist->next) {	\
            type* item = dlist->data;                     \
            func(item);                                   \
        }																									\
    }

typedef struct _action_item_t action_item_t;

/**
 * The function which is called when an action is activated. 
 * @param objects the selection to perform that action on. Note that
 * this selection is already narrowed down by the ActionItemCheck
 * function which means that you only get object here that actually
 * can be processed.
 * @param data the userdata passed down from the toplevel action, or
 * some local data (see AI_LOCAL_DATA)
 */
typedef void (*ActionItemFunc)(GList* objects, void* data);
/**
 * Actions can dynamically link to another action item. This is for
 * example useful if the (sub)menu depends on the selection.
 * @param items the selection
 * @param data the userdata of the toplevel node or some local data.
 */
typedef action_item_t* (*ActionItemLink)(GList* items, void* data);
/**
 * Sometimes you need the same actions in different places to operate
 * on different objects. You could create several actions with
 * different handlers (ActionItemFunc) - or you create several actions
 * with the same handlers and an additional conversion functions which
 * convert the object in such way, that they can be passed to that one
 * implementation of the ActionItemFunc callbacks.
 * @note This conversion function is inherited by children, you need
 * to set it for the toplevel action only.
 */
typedef void* (*ActionObjectConvert)(void* object, void* data);
/**
 * Check whether the object can be processed by that action.
 */
typedef gboolean (*ActionItemCheck)(void* object, void* data);


/** An action. You rarely access the members directly. */
struct _action_item_t {
  char* name;                  ///< the name (label) of the action
  ActionItemFunc action;       ///< the function to be called if action is activated.
  ActionItemCheck check;       ///< the function that is called to detected whether an object is passed to the activation function
  ActionItemLink link;         ///< this action links to other actions instead of defining an activation function
  ActionObjectConvert convert; ///< function to convert the objects before they are passed to the check/activation functions
  int flags;                   ///< the flags of the action
  GList* children;             ///< child actions, don't access directly.
  void* user_data;             ///< the userdata which can be forced to inherit the parents userdata if AI_LOCAL_DATA is set.
};

action_item_t*
action_item_new(const char* name, action_item_t* parent, int flags,
                ActionItemFunc action, ActionItemCheck check);
action_item_t* action_item_root(const char* name);
action_item_t* action_item_sep(action_item_t* parent);
action_item_t* action_item_simple(const char* name, action_item_t* parent,
                                  int flags, ActionItemFunc action);
action_item_t* action_item_sub(const char* name, action_item_t* parent,
                               ActionItemLink link, gboolean embed);
action_item_t* action_item_link(const char* name, action_item_t* parent,
                                const char* subname);

action_item_t* action_item_node(const char* name, action_item_t* parent);


void action_item_destroy(action_item_t* ia);

action_item_t* action_item_search(const char* name);


/** @name User actions */
/*@{*/
action_item_t* user_action_upload_find(int ID);
action_item_t* user_action_result_find(int ID);
action_item_t* user_action_source_find(int ID);
action_item_t* user_action_browse_find(int ID);

void user_action_upload_register(int ID, action_item_t* item);
void user_action_result_register(int ID, action_item_t* item);
void user_action_source_register(int ID, action_item_t* item);
void user_action_browse_register(int ID, action_item_t* item);
/*@}*/

/*@}*/

G_END_DECLS
#endif
