/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.12 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include <string.h>

#include "attribute.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

struct _attr_init_t {
  char* name;
  char* domain;
  AttrPrint do_print;
  UIPrintFunc  ui_print;
  XMLPrintFunc xml_print;
  CompareFunc  compare;
  FreeFunc     free;
  CopyFunc     copy;
  StoreFunc    store;
  StoreStrFunc store_str;
  int references;
  int used;
};

static GList* AttrInit = NULL;
static gboolean AllowUnregistered = FALSE;

/** Setup the loader. You can supress and discard unregistered
 * attributes while loading.
 * @param allow TRUE to keep unregistered attribtues as string, FALSE to discard
 */
void attr_allow_unregistered(gboolean allow) {
  AllowUnregistered = allow;
}

/** Clear an attribute container.
 * @param ac the container
 */
void attr_container_empty(attr_container_t* ac) {
  GList* dlist;

  if (ac->attr) {
    for (dlist = ac->attr; dlist; dlist = dlist->next) {
      attribute_t* attr = dlist->data;
      attr_destroy(attr);
    }
    g_list_free(ac->attr);
    ac->attr = NULL;
    object_set_dirty(OBJ(ac));
  }
}

/** Destroy a container. Use this function only from the destroy
 * function of a derives class.
 * @param ac the container to destroy
 */
void attr_container_destroy(attr_container_t* ac) {
  g_assert(ac);
  attr_container_empty(ac);
  object_destroy(OBJ(ac));
}

/** Initializes an attributes container. Call from the init function
 * of your derived class.
 * @param ac the attribute container.
 * @param name the name of the container, passed to object_init()
 * @note the object is initialized with sync to TRUE
 */
void attr_container_init(attr_container_t* ac, const char* name) {
  g_assert(ac);
  object_init(OBJ(ac), name, TRUE);
  object_add_type(OBJ(ac), "AttributeContainer");
  OBJ(ac)->destroy = (ObjectFunc)attr_container_destroy;
  ac->attr = NULL;
}

/** Create a new attribute container.
 * @return a new attribtue container
 */
attr_container_t* attr_container_new() {
  attr_container_t* result = g_malloc0(sizeof(*result));
  attr_container_init(result, "AttributeContainer");
  return result;
}

/** Copy a attribute container and some of its attributes.
 * @param dest destination container
 * @param source source container
 * @param domain the attribute domain which shoould be copied, or NULL
 * to copy all attributes.
 */
void attr_container_copy(attr_container_t* dest, attr_container_t* source,
                         const char* domain) {
  GList* dlist;

  g_assert(dest && source);
  
  object_copy(OBJ(dest), OBJ(source));
  attr_container_empty(dest);
  if (source->attr) {
    for (dlist = source->attr; dlist; dlist = dlist->next) {
      attribute_t* attr = dlist->data;
      if (!domain || !strcmp(attr->_ai->domain, domain))
        dest->attr = g_list_append(dest->attr, attr_clone(attr));
    }
    object_set_dirty(OBJ(dest));
  }
}

/** Copy some attributes from one container to another. This is
 * similar to attr_container_copy() with the difference that already
 * existing attributes are not overwritten.
 * @param dest destination container
 * @param src source container
 * @param domain the attribute domain which shoould be copied, or NULL
 * to copy all attributes.
 */
void attr_container_adapt(attr_container_t* dest, 
                          attr_container_t* src,
                          const char* domain) {
  GList* dlist;

  g_assert(dest && src);

  for (dlist = src->attr; dlist; dlist = dlist->next) {
    attribute_t* attr = dlist->data;
    attribute_t* attr2;
    
    if (domain && strcmp(domain, attr->_ai->domain)) continue;
    attr2 = attr_find(dest, attr->_ai->domain, attr->_ai->name);
    if (!attr2) {
      attr2 = attr_clone(attr);
      dest->attr = g_list_append(dest->attr, attr2);
    } else if (attr->_ai->compare(attr->value, attr2->value) != 0) {
      g_log("ATTR", G_LOG_LEVEL_WARNING,
            "attribute '%s@%s' differs"
            " | dest: %s [%s] %s"
            " | src:  %s [%s] %s",
            attr->_ai->name, attr->_ai->domain,

            OBJ(dest)->name, object_get_type(OBJ(dest)),
            attr2->_ai->xml_print(attr2->value),

            OBJ(src)->name, object_get_type(OBJ(src)),
            attr->_ai->xml_print(attr->value));
    }
  }
  object_set_dirty(OBJ(dest));
}


/** Clone a container. Creates a new container and copies the object
 * using attr_container_copy().
 * @param ac the source container
 * @return the cloned container
 */
attr_container_t* attr_container_clone(attr_container_t* ac) {
  attr_container_t* result;

  if (!ac) return NULL;
  result = attr_container_new();
  attr_container_copy(result, ac, NULL);
  return result;
}

/** Compare two containers.
 * @param ac1 the first container
 * @param ac2 the second container
 * @return -1 if ac1 < ac2, 1 if ac1 > ac2 or 0 if containers are equal.
 */
int attr_container_compare(attr_container_t* ac1,
                           attr_container_t* ac2) {
  int l1, l2;
  GList* dlist;
  attribute_t* attr2;

  if (!ac2) return !ac1;
  if (!ac1) return -1;
  
  l1 = g_list_length(ac1->attr);
  l2 = g_list_length(ac2->attr);
  if (l1 < l2) return -1;
  if (l1 > l2) return 1;
  for (dlist = ac1->attr; dlist; dlist = dlist->next) {
    attribute_t* attr = dlist->data;
    attr2 = attr_find(ac2, attr->_ai->domain, attr->_ai->name);
    if (!attr2) return 1;
    if (attr->_ai->compare(attr->value, attr2->value) != 0)
      return 1;
  }
  return 0;
}

/** Apply a function to some attributes in the container.
 * @param ac the container
 * @param domain the domain of attributes to inspect, or NULL to apply on all attributes.
 * @param func the function to apply
 * @param data user data passed to the function
 */
void attr_container_foreach(attr_container_t* ac, const char* domain,
                            AttrFunc func, void* data) {
  GList* dlist;
  for (dlist = ac->attr; dlist; dlist = dlist->next) {
    attribute_t* attr = dlist->data;
    if (domain && strcmp(attr->_ai->domain, domain)) continue;
    func(attr, data);
  }
}

/** Remove an attribute from a container. Only removes instances that
 * actually are in the container, clones are not removed.
 * @param ac the container
 * @param attr the attributes to remove
 */
void attr_remove(attr_container_t* ac, attribute_t* attr) {
  g_assert(ac && attr);
  ac->attr = g_list_remove(ac->attr, attr);
  attr_destroy(attr);
  object_set_dirty(OBJ(ac));
}

/** Destroy an attribute.
 * @param attr the attributes
 */
void attr_destroy(attribute_t* attr) {
  attr_init_t* ai;

  if (!attr) return;
  ai = attr->_ai;
  if (attr->value && ai->free) ai->free(attr->value);
  ai->references--;
  if (!ai->references) {
    AttrInit = g_list_remove(AttrInit, ai);
    if (ai->name) g_free(ai->name);
    if (ai->domain) g_free(ai->domain);
    g_free(ai);
  }
  g_free(attr);
}

/** Copy an attributes.
 * @param[in] dest the destination
 * @param[out] source the source.
 */
void attr_copy(attribute_t* dest, attribute_t* source) {
  if (!dest || !source) return;

  if (dest->value && dest->_ai->free) {
    dest->_ai->free(dest->value);
    dest->value = NULL;
  }
  dest->value = source->_ai->copy(NULL, source->value);
  dest->refreshed = source->refreshed;
  dest->do_print = source->do_print;

  dest->_ai = source->_ai;
  dest->_ai->references++;
}

/** Clone an attribute.
 * @param[in] attr the attribute to clone
 * @return the cloned attribute
 */
attribute_t* attr_clone(attribute_t* attr) {
  attribute_t* result;

  if (!attr) return NULL;
  result = g_malloc0(sizeof(*attr));
  attr_copy(result, attr);
  return result;
}

/** Compare two attributes.
 * @param[in] attr1 the first attribute
 * @param[in] attr2 the second attribute
 * @return 0 if equal, -1 if attr1 < attr2, 1 if attr1 > attr2
 */
int attr_compare(attribute_t* attr1, attribute_t* attr2) {
  if (!attr1) return attr2?1:0;
  if (!attr2) return -1;
  return attr1->_ai->compare(attr1->value, attr2->value);
}

static attr_init_t*
attr_init_find(const char* domain, const char* name) {
  GList* dlist;
  
  for (dlist = AttrInit; dlist; dlist = dlist->next) {
    attr_init_t* ai = dlist->data;
    if (!strcmp(ai->domain, domain) && !strcmp(ai->name, name))
      return ai;
  }
  return NULL;
}

GList* attr_get_registered(const char* domain) {
  GList* result = NULL;
  GList* dlist;
  for (dlist = AttrInit; dlist; dlist = dlist->next) {
    attr_init_t* init = dlist->data;
    if (!strcmp(domain, init->domain))
      result = g_list_append(result, init->name);
  }
  return result;
}

static attr_init_t*
attr_register_internal(const char* domain, const char* name, AttrPrint do_print,
                       UIPrintFunc ui_print,  XMLPrintFunc xml_print,
                       CompareFunc compare, CopyFunc copy, FreeFunc free,
                       StoreStrFunc store_str, StoreFunc store) {
  attr_init_t* init;
  
  init = g_malloc0(sizeof(*init));
  init->domain = g_strdup(domain);
  init->name = g_strdup(name);
  init->do_print = do_print;
  init->ui_print = ui_print;
  init->xml_print = xml_print;
  init->compare = compare;
  init->copy = copy;
  init->free = free?free:(FreeFunc)g_free;
  init->store = store;
  init->store_str = store_str;
  init->references = 1;
  init->used = 0;
  AttrInit = g_list_append(AttrInit, init);
  return init;
}

/** Register a new attribute which then can be used in the attr_container_t
 * @param domain the attribute domain
 * @param name the attribute name
 * @param do_print print hint for the UI, don't print when 0, always
 * print when 1, maybe print when 2
 * @param ui_print function used to print in the UI
 * @param xml_print function to export attribute to a xml file
 * @param compare function to compare two attributes
 * @param copy function to copy an attribute
 * @param free function to free an attribute
 * @param store_str function to create an attribute from a string
 * @param store function to create an attribute from specific arguments
 */
void attr_register(const char* domain, const char* name, AttrPrint do_print,
                   UIPrintFunc ui_print,
                   XMLPrintFunc xml_print,
                   CompareFunc compare,
                   CopyFunc copy,
                   FreeFunc free,
                   StoreStrFunc store_str,
                   StoreFunc store) {
  attr_init_t* init;

  init = attr_init_find(domain, name);
  if (init) {
    g_log("ATTR", G_LOG_LEVEL_WARNING,
          "attribute '%s' already registered for domain '%s'",
	  name, domain);
    return;
  }
  init =
    attr_register_internal(domain, name, do_print, ui_print, xml_print,
                           compare, copy, free, store_str, store);

  g_log("ATTR", G_LOG_LEVEL_DEBUG,
        "registered attribute '%s' for domain '%s'", name, domain);
}

/** Wrapper for attr_register() for hash attibutes
 * @param domain the attribute domain
 * @param name the attribute name
 * @param do_print printing hint, see attr_register()
 * @param print function used to print in the UI, or NULL to use default.
 */
void attr_register_hash(const char* domain, const char* name,
                        AttrPrint do_print, UIPrintFunc print) {
  attr_register(domain, name, do_print,
                print?print:(UIPrintFunc)print_hash,
                (XMLPrintFunc)print_hash,
                (CompareFunc)compare_hash,
                (CopyFunc)copy_hash,
                (FreeFunc)free_hash,
                (StoreStrFunc)store_str_hash,
                (StoreFunc)store_hash);
}

/** Wrapper for attr_register() for string attibutes
 * @param domain the attribute domain
 * @param name the attribute name
 * @param do_print printing hint, see attr_register()
 * @param print function used to print in the UI, or NULL to use default.
 */
void attr_register_str(const char* domain, const char* name,
                       AttrPrint do_print, UIPrintFunc print) {
  attr_register(domain, name, do_print,
                print?print:(UIPrintFunc)print_string,
                (XMLPrintFunc)print_string,
                (CompareFunc)compare_string,
                (CopyFunc)copy_string,
                (FreeFunc)NULL,
                (StoreStrFunc)store_str_string,
                (StoreFunc)store_string);
}

/** Wrapper for attr_register() for int attibutes
 * @param domain the attribute domain
 * @param name the attribute name
 * @param do_print printing hint, see attr_register()
 * @param print function used to print in the UI, or NULL to use default.
 */
void attr_register_int(const char* domain, const char* name,
                       AttrPrint do_print, UIPrintFunc print) {
  attr_register(domain, name, do_print,
                print?print:(UIPrintFunc)print_int,
                (XMLPrintFunc)print_int,
                (CompareFunc)compare_int,
                (CopyFunc)copy_int,
                (FreeFunc)NULL,
                (StoreStrFunc)store_str_int,
                (StoreFunc)store_int);
}

/** Wrapper for attr_register() for boolean attibutes
 * @param domain the attribute domain
 * @param name the attribute name
 * @param do_print printing hint, see attr_register()
 * @param print function used to print in the UI, or NULL to use default.
 */
void attr_register_bool(const char* domain, const char* name,
                        AttrPrint do_print, UIPrintFunc print) {
  attr_register(domain, name, do_print,
                print?print:(UIPrintFunc)print_bool,
                (XMLPrintFunc)print_bool,
                (CompareFunc)compare_bool,
                (CopyFunc)copy_bool,
                (FreeFunc)NULL,
                (StoreStrFunc)store_str_bool,
                (StoreFunc)store_bool);
}

/** Find an attribute in a container.
 * @param[in] ac the container
 * @param[in] domain domain of the attribute to look for 
 * @param[in] name name of the attribute to look for
 * @return the attribute, or NULL if not found
 */
attribute_t* attr_find(attr_container_t* ac, const char* domain, const char* name) {
  GList* dlist;
  
  if (!ac) return NULL;
  for (dlist = ac->attr; dlist; dlist = dlist->next) {
    attribute_t* attr = dlist->data;
    if (!strcmp(attr->_ai->name, name) &&
	!strcmp(attr->_ai->domain, domain)) return attr;
  }
  return NULL;
}

/** Change the value of an attribute.
 * @param[in] ac the container
 * @param[in] domain the domain of the attribute to change.
 * @param[in] name the name of the attribute to change.
 * @param[in] ... arguments which are stored
 */
attribute_t* attr_store(attr_container_t* ac, const char* domain,
                        const char* name, ...) {
  attr_init_t* ai;
  attribute_t* attr;
  va_list ap;

  attr = attr_find(ac, domain, name);
  if (!attr) {
    ai = attr_init_find(domain, name);
    if (!ai) {
      g_log("ATTR", G_LOG_LEVEL_WARNING,
            "attribute '%s' not registered for domain '%s'", 
	    name, domain);
      g_assert(FALSE);
      return NULL;
    }
    attr = g_malloc0(sizeof(*attr));
    attr->do_print = ai->do_print;
    attr->_ai = ai;
    ai->used++;
    ai->references++;
    ac->attr = g_list_append(ac->attr, attr);

    va_start(ap, name);
    attr->value = attr->_ai->store(attr->value, ap);
    va_end(ap);
    //    attr->refreshed = CurrentTime;
    g_assert(attr->value);
    attr->refreshed = time(NULL);
    object_set_dirty(OBJ(ac));
  } else {
    void* res;
    va_start(ap, name);
    res = attr->_ai->store(attr->value, ap);
    va_end(ap);
    //    attr->refreshed = CurrentTime;
    if (res) {
      attr->value = res;
      attr->refreshed = time(NULL);
      object_set_dirty(OBJ(ac));
    }
  }
  return attr;
}

/** Change the value of an attribute using the content of a string.
 * @param[in] ac the container
 * @param[in] domain the domain of the attribute to change.
 * @param[in] name the name of the attribute to change.
 * @param[in] str the string which is a serialized attribute
 */
attribute_t* 
attr_store_str(attr_container_t* ac, const char* domain,
               const char* name, const char* str) {
  attr_init_t* ai;
  attribute_t* attr;
  attr_init_t* dai = NULL;
  
  attr = attr_find(ac, domain, name);
  if (!attr) {
    ai = attr_init_find(domain, name);
    if (!ai) {
      if (AllowUnregistered) {
        // this should be only during loading attributes from disc
/*         g_log("ATTR", G_LOG_LEVEL_DEBUG, */
/*               "attribute '%s@%s' stripped", name, domain); */
/*         return NULL; */
        dai = attr_register_internal(domain, name, ATTR_SKIP,
                                     (UIPrintFunc)print_string,
                                     (XMLPrintFunc)print_string,
                                     (CompareFunc)compare_string,
                                     (CopyFunc)copy_string,
                                     (FreeFunc)NULL,
                                     (StoreStrFunc)store_str_string,
                                     (StoreFunc)store_string);
        ai = dai;
        g_log("ATTR", G_LOG_LEVEL_DEBUG,
              "found unregistered '%s' : treating as string", name);
      } else {
        g_log("ATTR", G_LOG_LEVEL_DEBUG,
              "** attribute '%s' not registered for domain '%s'", 
              name, domain);
        g_assert(FALSE);
        return NULL;
      }
    }
    ai->used++;
    ai->references++;

    attr = g_malloc0(sizeof(*attr));
    attr->do_print = ai->do_print;
    attr->_ai = ai;
    ac->attr = g_list_append(ac->attr, attr);

    attr->value = attr->_ai->store_str(attr->value, str);
/*     g_assert(attr->value); */
    //    attr->refreshed = CurrentTime;
    attr->refreshed = time(NULL);
    object_set_dirty(OBJ(ac));
  } else {
    void* res;
    res = attr->_ai->store_str(attr->value, str);
    //    attr->refreshed = CurrentTime;
    if (res) {
      attr->value = res;
      attr->refreshed = time(NULL);
      object_set_dirty(OBJ(ac));
    }
  }
  return attr;
}


/** Print attributes into a string. Only attributes with print mode
 * equal to 1 are printed (see attr_store() and attr_strore_str())
 * @param[in] ac the container which holds the attributes.
 * @return the printed attributes.
 * @note don't free the result (fixme: make secure)
 */
char* attr_print_all(attr_container_t* ac) {
  GList* dlist;
  int cnt = 0;
  int added = 0;
  char* result = get_text_buffer();
  const char* text;
  unsigned int size = text_buffer_size();

  for (dlist = ac->attr; dlist; dlist = dlist->next) {
    attribute_t* attr = dlist->data;
    attr_init_t* ai = attr->_ai;
    if (attr->do_print == ATTR_IMPORTANT && ai->ui_print &&
        (text = ai->ui_print(attr->value))) {
      if (strlen(text)+3+added >= size) break;
      if (cnt) {
        strcpy(result+added, ", ");
        added += 2;
      }
      strcpy(result+added, text);
      added += strlen(text);
      cnt++;
    }
  }
  result[added] = 0;
  return result;
}

GList* attr_get_list(attr_container_t* ac) {
  g_assert(ac);
  return ac->attr;
}

void attr_print_stats() {
  GList* dlist;

  g_log("ATTR", G_LOG_LEVEL_DEBUG,
        "- statistics ------------------------------------");
  for (dlist = AttrInit; dlist; dlist = dlist->next) {
    attr_init_t* ai = dlist->data;
    g_log("ATTR", G_LOG_LEVEL_DEBUG,
          " %15s @ %-10s : used %5d times, still %d refs",
          ai->name, ai->domain, ai->used, ai->references);
  }
  g_log("ATTR", G_LOG_LEVEL_DEBUG,
        "-------------------------------------------------");
}

/** Print the attribute using the ui print function
 * @param attr the attribute to print.
 * @return the printed attribute
 * @note You don't need to free the result.
 */
const char* attr_fancy_print(attribute_t* attr) {
  if (attr && attr->_ai && attr->_ai->ui_print)
    return attr->_ai->ui_print(attr->value);
  else
    return NULL;
}

/** Print the attribute to be store to an xml file.
 * @param attr the attribute to print.
 * @return the printed attribute
 * @note You don't need to free the result.
 */
const char* attr_xml_print(attribute_t* attr) {
  if (attr && attr->_ai && attr->_ai->xml_print)
    return attr->_ai->xml_print(attr->value);
  else
    return NULL;
}

/** Get the name of an attribute
 * @param attr the attribute.
 * @return the name of the attribute
 */
const char* attr_get_name(attribute_t* attr) {
  if (attr && attr->_ai) return attr->_ai->name;
  else return NULL;
}

#ifdef WIN32
# define _ATTRIBUTES_C_
# include "lcorealiasdef.c"
#endif
