/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.6 $
 * $Date: 2006/07/22 17:11:45 $
 * $Author: sid4u $
 */

#ifndef _ATTR_H_
#define _ATTR_H_

#include "l_time.h"
#include <glib.h>

#include "utils.h"
#include "object.h"

G_BEGIN_DECLS

/** \defgroup ClassAttribute The attribute container
 * @ingroup BasicFramework
 */
/*@{*/

/** @name Print priority of attributes */
typedef enum {
  ATTR_SKIP,       ///< never print this attribute
  ATTR_IMPORTANT,  ///< always print attribute
  ATTR_USEFUL,     ///< useful to print attribute
  ATTR_UNIMPORTANT ///< unimportant attribute
} AttrPrint;

typedef struct _attribute_t attribute_t;

typedef void (*AttrFunc)(attribute_t* attr, void* data);

/** Class which can hold attributes */
typedef struct {
  object_t _object;   ///< the base class
  GList* attr;        ///< the list of attribute_t. Never access directly.
} attr_container_t;
#define ATTRC(arg) ((attr_container_t*)arg)

typedef struct _attr_init_t attr_init_t;

/** The attribute class. */
struct _attribute_t {
  attr_init_t* _ai;   ///< private
  void* value;        ///< the value, rarely used directly.
  time_t refreshed;   ///< timestamp of last attribute change
  int do_print;       ///< print mode of the attribute
};


/** @name General handling */
/*@{*/
void attr_allow_unregistered(gboolean allow);
void attr_register(const char* domain, const char* name, AttrPrint do_print,
                   UIPrintFunc ui_print, XMLPrintFunc xml_print,
                   CompareFunc compare, CopyFunc copy, FreeFunc free,
                   StoreStrFunc store_str, StoreFunc store);
void attr_register_hash(const char* domain, const char* name, AttrPrint do_print, UIPrintFunc print);
void attr_register_str (const char* domain, const char* name, AttrPrint do_print, UIPrintFunc print);
void attr_register_int (const char* domain, const char* name, AttrPrint do_print, UIPrintFunc print);
void attr_register_bool(const char* domain, const char* name, AttrPrint do_print, UIPrintFunc print);

void attr_print_stats(void);
/*@}*/


/** @name Attribute handling */
/*@{*/
attribute_t* attr_clone(attribute_t* attr);
void attr_copy(attribute_t* dest, attribute_t* source);
void attr_destroy(attribute_t* attr);
int attr_compare(attribute_t* attr1, attribute_t* attr2);

const char* attr_fancy_print(attribute_t* attr);
const char* attr_xml_print(attribute_t* attr);
const char* attr_get_name(attribute_t* attr);

/** Gets the pointer to the attribute value
 * @param attr the attribute
 * @param type the type to cast to
 */
#define attr_get_ptr(attr, type)  ((type*)((attr)->value))
/** Gets the attribute value
 * @param attr the attribute
 * @param type the type to cast to
 */
#define attr_get_ref(attr, type)  (*(type*)((attr)->value))
/** Gets one item from a multivalue attribute
 * @param attr the attribute
 * @param type the type to cast to
 * @param i the index within the multivalue
 */
#define attr_get_refi(attr, type, i)  ((type*)((attr)->value))[i]
/*@}*/


/** @name Attribute container handling */
/*@{*/

/** Store all attributes as xml.
 * @param ac the attribute container
 * @param elem the xmlNodePtr to save to
 */
#define ATTR_XML_PRINT_ALL(ac, elem) \
{\
  GList* _dlist;\
  for (_dlist = ac->attr; _dlist; _dlist = _dlist->next) {\
    attribute_t* attr = _dlist->data;\
    const char* value = attr_xml_print(attr);\
    if (value) {\
      xmlNodePtr _xsub = xmlNewTextChild(elem, NULL, "Attr", value);\
      xmlNewProp(_xsub, "Name", attr_get_name(attr));\
    }\
  }\
}

void attr_remove(attr_container_t* ac, attribute_t* attr);

void attr_container_init(attr_container_t* ac, const char* name);
attr_container_t* attr_container_new(void);
void attr_container_copy(attr_container_t* dest, 
                         attr_container_t* source,
                         const char* domain);
attr_container_t* attr_container_clone(attr_container_t* ac);
void attr_container_destroy(attr_container_t* ac);
int attr_container_compare(attr_container_t* ac1,
                           attr_container_t* ac2);
void attr_container_adapt(attr_container_t* dest, 
                          attr_container_t* source,
                          const char* domain);
void attr_container_empty(attr_container_t* ac);
void attr_container_foreach(attr_container_t* ac, const char* domain,
                            AttrFunc func, void* data);

attribute_t* attr_find(attr_container_t* ac, const char* domain, const char* name);
attribute_t* attr_store(attr_container_t* ac, const char* domain, const char* name, ...);
attribute_t* attr_store_str(attr_container_t* ac, const char* domain, const char* name, const char* str);
char* attr_print_all(attr_container_t* ac);
GList* attr_get_list(attr_container_t* ac);

/*@}*/

/*@}*/

G_END_DECLS

#endif
