/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.13 $
 * $Date: 2007/11/22 08:42:39 $
 * $Author: sid4u $
 */

#include <string.h>
#include <stdio.h>
#ifndef WIN32
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#endif

#include "b_socket.h"
#include "utils.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

struct _b_socket_priv_t {
  buffer_t* in_buffer;
  buffer_t* out_buffer;
  int close_when_sent;
  gboolean consuming;
  unsigned int max_insize;

  BReadFunc read;
  BWriteFunc write;

  BConsumeInputFunc consume_input;
  BPrepareFunc prepare_out;
};

typedef struct {
  module_t _module;
} bsocket_module_t;

static bsocket_module_t* BSocketModule = NULL;

static const char* b_socket_consume_input(b_socket_t* socket) {
  buffer_t* buffer = socket->priv->in_buffer;
  print_raw("consume message", buffer->data + buffer->consumed,
            buffer->size - buffer->consumed);
  buffer_clear(buffer);
  return NULL;
}

/** Ends a buffered socket. Call from ending function of derived class only. 
 * @param s the socket
 * @param message reason of ending.
 * @return TRUE if socket should be destroyed, FALSE otherwise
 */
gboolean b_socket_end(b_socket_t* s, const char* message) {
  g_assert(BSocketModule);
  /*
  if (buffer_avail(s->priv->out_buffer) > 0) {
    g_log("BSOCK", G_LOG_LEVEL_WARNING,
	  "closing socket with pending out buffer %u",
	  buffer_avail(s->priv->out_buffer));
  }
  */
  if (s->priv->in_buffer) buffer_destroy(s->priv->in_buffer);
  s->priv->in_buffer = NULL;
  if (s->priv->out_buffer) buffer_destroy(s->priv->out_buffer);
  s->priv->out_buffer = NULL;

  socket_end(SOCK(s), message);

  return TRUE;
}

/** Destroys a buffered socket. Call from destroy function of derived
 * class only.
 * @param s the socket
 */
void b_socket_destroy(b_socket_t* s) {
  g_assert(BSocketModule);
  if (s->priv->in_buffer) buffer_destroy(s->priv->in_buffer);
  if (s->priv->out_buffer) buffer_destroy(s->priv->out_buffer);
  socket_destroy(SOCK(s));
}

static void b_socket_consume(b_socket_t* s) {
  const char* message;

  if (s->priv->consuming) return;  // don't recurse ever

  s->priv->consuming = TRUE;

  while ((message = s->priv->consume_input(s))) {
    if (strcmp(message, "AGAIN")) {
      buffer_t* b = s->priv->out_buffer;
      // if out buffer has data, keep sending it.
      if (buffer_avail(b)) {
        g_log("SOCK", G_LOG_LEVEL_DEBUG,
              "delayed socket close: '%s'", message);
        b_socket_close_when_sent(s, ST_CONSUME_ERROR);
      } else {
        socket_done(SOCK(s), ST_CONSUME_ERROR, message);
      }
      break;
    }
  }

  s->priv->consuming = FALSE;
}

static void b_socket_receive(b_socket_t* s)
{
    buffer_t* buffer = s->priv->in_buffer;
    
    if (buffer->size >= s->priv->max_insize)
    {
        socket_done(SOCK(s), ST_INPUT_OVERFLOW,
                    l_printf("input buffer overflow (%d)", s->priv->max_insize));
        return;
    }
    if (s->priv->read(s) > 0) b_socket_consume(s);
}

static void b_socket_send(b_socket_t* s) {
  if (!s->priv->prepare_out(s)) return;
  
  if (buffer_avail(s->priv->out_buffer) > 0) {
    if (s->priv->write(BSOCK(s)) <= 0) return;
  } else {
    socket_suspend(SOCK(s), DIR_OUT);
    if (s->priv->close_when_sent >= 0)
      socket_done(SOCK(s), s->priv->close_when_sent, "everything sent");
  }
}

static gboolean b_socket_prepare_out(b_socket_t* s) {
  if (s->priv->out_buffer->consumed)
    buffer_consume(s->priv->out_buffer, 0);
  return TRUE;
}

/** Init a buffered socket. Call from init function of derived
 * class only.
 * @param s the socket
 * @param name the name of the socket.
 * @param timeout idle time after socket is closed, or 0 for no timeout.
 */
void b_socket_init(b_socket_t* s, const char* name, int timeout) {
  g_assert(BSocketModule);
  socket_init(SOCK(s), name, timeout);
  object_add_type(OBJ(s), "BSocket");
  OBJ(s)->destroy = (ObjectFunc)b_socket_destroy;
  socket_set_end(SOCK(s), (SockEndFunc)b_socket_end);
  s->priv = g_malloc0(sizeof(*s->priv));
  s->priv->in_buffer = NULL;
  s->priv->out_buffer = NULL;
  s->priv->max_insize = 2048;
  s->priv->close_when_sent = -1;
  s->priv->consume_input = b_socket_consume_input;
  s->priv->prepare_out = b_socket_prepare_out;
  s->priv->consuming = FALSE;

  socket_set_handler(SOCK(s), (LIOFunc)b_socket_receive, DIR_IN);
  socket_set_handler(SOCK(s), (LIOFunc)b_socket_send, DIR_OUT);
}

void b_socket_move(b_socket_t* dest, b_socket_t* src, gboolean copy_buffer) {
  socket_move(SOCK(dest), SOCK(src));
  if (copy_buffer) {
    buffer_t* buffer;
    buffer = dest->priv->in_buffer;
    dest->priv->in_buffer = src->priv->in_buffer;
    src->priv->in_buffer = buffer;
    buffer = dest->priv->out_buffer;
    dest->priv->out_buffer = src->priv->out_buffer;
    src->priv->out_buffer = buffer;
  }
}

void b_socket_set_limit(b_socket_t* s, unsigned int max_size) {
  if (max_size) s->priv->max_insize = max_size;
}

/** Starts receiving and consuming data.
 * The socket might be destroyed when this function returns.
 * @param s the socket.
 * @param max_size size the input buffer overflows.
 */
void b_socket_start_receiving(b_socket_t* s, unsigned int max_size) {
  g_assert(BSocketModule);
  if (max_size) s->priv->max_insize = max_size;
  if (!s->priv->in_buffer) s->priv->in_buffer = buffer_new(1024);
  socket_resume(SOCK(s), DIR_IN);

  // check for consume, since this might be a copied socket
  // with a copied buffer
  //FIXME: this might be dangerous..
  if (buffer_avail(s->priv->in_buffer)) b_socket_consume(s);
}

/** Starts sending the output buffer.
 * @param s the socket
 */
void b_socket_start_sending(b_socket_t* s) {
  g_assert(BSocketModule);
  socket_resume(SOCK(s), DIR_OUT);
}

/** Close the socket when everything is sent.
 * @param s the socket.
 * @param status the ending status
 */
void b_socket_close_when_sent(b_socket_t* s, int status) {
  g_assert(BSocketModule);
  // stop receiving
  socket_suspend(SOCK(s), DIR_IN);
  s->priv->close_when_sent = status;
}

/** FIXME */
buffer_t* b_socket_get_out_buffer(b_socket_t* socket, unsigned space) {
  g_assert(BSocketModule);
  if (space) {
    if (socket->priv->out_buffer)
      buffer_need_space(socket->priv->out_buffer, space);
    else
      socket->priv->out_buffer = buffer_new(space);
  }
  return socket->priv->out_buffer;
}

/** FIXME */
buffer_t* b_socket_get_in_buffer(b_socket_t* s) {
  g_assert(BSocketModule);
  return s->priv->in_buffer;
}

/** FIXME */
buffer_t* b_socket_set_in_buffer(b_socket_t* s, buffer_t* buffer) {
  buffer_t* temp = s->priv->in_buffer;
  s->priv->in_buffer = buffer;
  return temp;
}

/** FIXME */
buffer_t* b_socket_set_out_buffer(b_socket_t* s, buffer_t* buffer) {
  buffer_t* temp = s->priv->out_buffer;
  s->priv->out_buffer = buffer;
  return temp;
}

/** FIXME */
void b_socket_set_read_write(b_socket_t* s,
                             BReadFunc read, BWriteFunc write) {
  if (read) s->priv->read = read;
  if (write) s->priv->write = write;
}

/** Set the function that is called when there is incoming data to
 * consume.
 * @param s the socket.
 * @param func the function to call
 */
void b_socket_set_consume(b_socket_t* s, BConsumeInputFunc func) {
  s->priv->consume_input = func;
}

/** Set a special function that is called before data is sent on that
 * socket.  Uploads for example fill the buffer by reading data from a
 * file.
 * @param s the socket.
 * @param func the prepare function.
 */
void b_socket_set_prepare_out(b_socket_t* s, BPrepareFunc func) {
  s->priv->prepare_out = func;
}

static void bsockets_init() {
}
static void bsockets_run() {
}
static void bsockets_exit() {
}

static void bsocket_module_init(bsocket_module_t* module) {
  module_init(MODULE(module), "BSocket Module",
	      (ModuleFunc)bsockets_init,
	      (ModuleFunc)bsockets_run,
	      (ModuleFunc)bsockets_exit);

  module_require(lsocket_module_instance);
  
}


/** Require this function if you directly use this module.
 * @return the module instance
 * @see module_require()
 */
module_t* bsocket_module_instance() {
  if (!BSocketModule) {
    BSocketModule = g_malloc0(sizeof(*BSocketModule));
    bsocket_module_init(BSocketModule);
  }
  return MODULE(BSocketModule);
}

#ifdef WIN32
#define _B_SOCKET_C_
#include "lcorealiasdef.c"
#endif
