/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.12 $
 * $Date: 2007/11/21 20:18:23 $
 * $Author: sgop $
 */

#include <glib.h>

#include "browse.h"
#include "file_tree.h"
#include "action_item.h"
#include "ui.h"
#include "l_i18n.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
  GList* browses;
} browse_module_t;

static browse_module_t* BrowseModule = NULL;

browse_t* browse_find(user_t* user)
{
    GList* dlist;
    g_assert(BrowseModule);
    for (dlist = BrowseModule->browses; dlist; dlist = dlist->next)
    {
        browse_t* browse = dlist->data;
        if (browse->user == user) return browse;
    }
    return NULL;
}

static void browse_node_init(browse_node_t* node, const char* name)
{
    file_node_init(FILE_NODE(node), name);
    object_add_type(OBJ(node), "BrowseNode");
}

static file_node_t* browse_node_new(const char* name)
{
    browse_node_t* node = g_malloc0(sizeof(*node));
    browse_node_init(node, name);
    return FILE_NODE(node);
}

void browse_destroy(browse_t* browse)
{
    g_assert(BrowseModule);
    
    if (OBJ(browse)->destroy == (ObjectFunc)browse_destroy)
    {
        g_log("BROWSE", G_LOG_LEVEL_WARNING,
              "browse destroy function should be overridden");
    }
    BrowseModule->browses = g_list_remove(BrowseModule->browses, browse);
    
    user_unref(browse->user);
    
    file_tree_destroy(FILE_TREE(browse));
}

void browse_init(browse_t* browse, user_t* user)
{
    const char* name = user->print(user);
    
    g_assert(BrowseModule);
    file_tree_init(FILE_TREE(browse), name, name);
    object_add_type(OBJ(browse), "Browse");
    OBJ(browse)->destroy = (ObjectFunc)browse_destroy;
    FILE_TREE(browse)->node_new = browse_node_new;
    //  FILE_TREE(browse)->node_recent = browse_node_recent;
    
    browse->user = user_ref(user);
    
    BrowseModule->browses = g_list_append(BrowseModule->browses, browse);
    object_make_public(OBJ(browse), NULL);
}

void browse_set_status(browse_t* browse, const char* status)
{
    g_assert(BrowseModule);
    UI_file_tree_update_label(FILE_TREE(browse), status);
}

void browse_set_progress(browse_t* browse, double prog, const char* status)
{
    g_assert(BrowseModule);
    UI_file_tree_set_progress(FILE_TREE(browse), prog);
    UI_file_tree_set_progress_name(FILE_TREE(browse), status);
}

void browse_insert_file(browse_t* browse, file_t* file)
{
    file_node_t* node;
    
    g_assert(BrowseModule);
    node = file_tree_insert_file(FILE_TREE(browse), file, 
                                 FN_FLAG_SHARED|FN_FLAG_RECENT);
    //  if (node) BROWSE_NODE(node)->last_seen = 0;
    //  browse->status++;
}


static void on_download_from_result(GList* nodes)
{
    GList* dlist;
    for (dlist = nodes; dlist; dlist = dlist->next)
    {
        file_node_t* node = dlist->data;
        if (node->file) download_from_file(node->file);
    }
}

static void on_download_from_source(GList* nodes, browse_t* browse)
{
    GList* dlist;
    for (dlist = nodes; dlist; dlist = dlist->next)
    {
        file_node_t* node = dlist->data;
        download_t* download;
        
        if (!node || !node->file) continue;
        // FIXME
        download = download_from_file(node->file);
        if (!download) continue;
        download_add_source(download, node->file, browse->user);
    }
}

static gboolean on_check_source(file_node_t* node, browse_t* browse)
{
    return (node && browse && node->file &&
            browse->user && source_factory_get(browse->user->ID));
}

static action_item_t* on_get_protocol_user_menu(GList* selection, browse_t* browse)
{
    (void)selection;
    return user_action_browse_find(browse->user->ID);
}

static void browse_actions_init(void)
{
    action_item_t* item;
    
    item = action_item_root("Browse file actions");
    action_item_simple(N_("Download"), item, 0,
                       (ActionItemFunc)on_download_from_result);
    action_item_simple(N_("Download as..."), item, 0,
                       (ActionItemFunc)NULL/*on_download_folder*/);
    action_item_sep(item);
    action_item_new(N_("Download from source"), item, 0,
                    (ActionItemFunc)on_download_from_source,
                    (ActionItemCheck)on_check_source);
    
    action_item_sep(item);
    action_item_sub(N_("User menu"), item, 
                    (ActionItemLink)on_get_protocol_user_menu,
                    FALSE);
}

static void browses_init()
{
    browse_actions_init();
}

static void browses_run()
{
}

static void browses_exit()
{
}

static void browse_module_init(browse_module_t* module)
{
    module_init(MODULE(module), "Browse Module",
                (ModuleFunc)browses_init,
                (ModuleFunc)browses_run,
                (ModuleFunc)browses_exit);
    
    module->browses = NULL;

    module_require(user_module_instance);
}


module_t* browse_module_instance()
{
    if (!BrowseModule)
    {
        BrowseModule = g_malloc0(sizeof(*BrowseModule));
        browse_module_init(BrowseModule);
    }
    return MODULE(BrowseModule);
}

#ifdef WIN32
#define _BROWSE_C_
#include "lcorealiasdef.c"
#endif
