/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.10 $
 * $Date: 2006/07/22 17:11:45 $
 * $Author: sid4u $
 */

#include <string.h>

#include <libxml/parser.h>

#include "chat.h"
#include "buddy.h"
#include "plugin.h"
#include "utils.h"
#include "preferences.h"
#include "xml_handler.h"
#include "ui.h"
#include "l_i18n.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
  GList* buddies;
  GList* updater;
} buddy_module_t;

static buddy_module_t* BuddyModule = NULL;

/** Register callback to monitor changes of the buddy itself.
 * Call this function if you like to react to changes in the buddy name, etc.
 * @param func the buddy function
 */
void buddy_register_update(BuddyFunc func)
{
    g_assert(BuddyModule);
    if (func) BuddyModule->updater = g_list_append(BuddyModule->updater, func);
}

void buddy_propagate_update(buddy_t* buddy)
{
    GList* dlist;
    g_assert(BuddyModule);
    for (dlist = BuddyModule->updater; dlist; dlist = dlist->next)
    {
        BuddyFunc func = dlist->data;
        func(buddy);
    }
}

static void buddy_load(void)
{
    attr_allow_unregistered(TRUE);
    lopster_xml_load("core:buddies");
    attr_allow_unregistered(FALSE);
}

static void buddy_save(void)
{
    if (module_running(MODULE(BuddyModule)))
        lopster_xml_save("core:buddies", 1000);
}

static void buddy_recalc_online(buddy_t* buddy) {
  GList* dlist;
  int cnt = 0;
  for (dlist = buddy->users; dlist; dlist = dlist->next) {
    user_t* user = dlist->data;
    if (user->online) cnt++;
  }
  if (cnt != buddy->online) {
    if (buddy->online == 0) {
      chat_print(NULL, "+\0client",
                 "+\1c_text", OBJ(buddy)->name,
                 "+\1c_client", _("is online"),
                 "+\0nl", NULL);
    } else if (cnt == 0) {
      chat_print(NULL, "+\0client",
                 "+\1c_text", OBJ(buddy)->name,
                 "+\1c_client", _("is offline"),
                 "+\0nl", NULL);
    }
    buddy->online = cnt;
    object_set_dirty(OBJ(buddy));
  }
}
static void on_user_changed(user_t* user, buddy_t* buddy) {
  (void)user;
  buddy_recalc_online(buddy);
}

static void buddy_destroy(buddy_t* buddy) {
  GList* dlist;

  for (dlist = buddy->users; dlist; dlist = dlist->next) {
    user_t* user = dlist->data;
    // this basically is a buddy_remove_user() with all the
    // unnecessary stuff skipped
    user_stop_monitor(user);
    user_unref(user);
  }
  g_list_free(buddy->users);
  buddy->users = NULL;

  attr_container_empty(ATTRC(buddy));
  g_free(buddy->description);

  BuddyModule->buddies = g_list_remove(BuddyModule->buddies, buddy);
  attr_container_destroy(ATTRC(buddy));
  buddy_save();
}

static void buddy_init(buddy_t* buddy, const char* name)
{
    g_assert(buddy && name);
    
    attr_container_init(ATTRC(buddy), name);
    object_add_type(OBJ(buddy), "Buddy");
    OBJ(buddy)->destroy = (ObjectFunc)buddy_destroy;
    object_set_sync(OBJ(buddy), FALSE);
    buddy->online = 0;
    buddy->users = NULL;
}

/** Create a new buddy.
 * @param nick the name of the buddy
 * @return a new buddy
 */
buddy_t* buddy_new(const char* nick)
{
    buddy_t* buddy;
    
    g_assert(BuddyModule);
    
    buddy = g_malloc0(sizeof(*buddy));
    buddy_init(buddy, nick);
    BuddyModule->buddies = g_list_append(BuddyModule->buddies, buddy);

    object_make_public(OBJ(buddy), NULL);
    
    return buddy;
}

/** Change the description of the buddy.
 * @param buddy the buddy
 * @param des the new description
 */
void buddy_set_description(buddy_t* buddy, const char* des)
{
    g_assert(buddy);
    g_free(buddy->description);
    buddy->description = des?g_strdup(des):NULL;
    // FIXME: inform gui?
}

/** Add a user that belongs to that buddy.
 * @param buddy the buddy
 * @param user the user to add
 */
void buddy_add_user(buddy_t* buddy, user_t* user) {
  if (g_list_find(buddy->users, user)) {
    g_log("BUDDY", G_LOG_LEVEL_WARNING,
          "user '%s' already part of buddy '%s'",
          user->print(user), OBJ(buddy)->name);
    return;
  }
  user_ref(user);
  buddy->users = g_list_append(buddy->users, user);
  object_monitor_d(OBJ(user), (ObjectFunc)on_user_changed, buddy);
  user_start_monitor(user);

  on_user_changed(NULL, buddy);
  buddy_save();
}

/** Remove a user from the buddy.
 * @param buddy the buddy
 * @param user the user to remove
 */
void buddy_remove_user(buddy_t* buddy, user_t* user) {
  if (!g_list_find(buddy->users, user)) {
    g_log("BUDDY", G_LOG_LEVEL_WARNING,
          "user '%s' not part of buddy '%s'",
          user->print(user), OBJ(buddy)->name);
    return;
  }
  buddy->users = g_list_remove(buddy->users, user);
  object_unmonitor_d(OBJ(user), (ObjectFunc)on_user_changed, buddy);
  user_stop_monitor(user);
  user_unref(user);

  on_user_changed(NULL, buddy);
  buddy_save();
}

/** Call a function for every user that belongs to the buddy and a
 * certain network.
 * @param buddy the buddy
 * @param ID the ID of the network (plugin)
 * @param func the function to call.
 * @param data user data to pass to the function
 */
void buddy_foreach_user(buddy_t* buddy, int ID,
                        UserForeach func, void* data) {
  GList* dlist;
  for (dlist = buddy->users; dlist; dlist = dlist->next) {
    user_t* user = dlist->data;
    if (user->ID != ID) continue;
    if (func(user, data)) return;
  }
}

static gboolean buddy_find_user(buddy_t* buddy, user_t* user) {
  GList* dlist;
  for (dlist = buddy->users; dlist; dlist = dlist->next) {
    user_t* u = dlist->data;
    if (u == user) return TRUE;
  }
  return FALSE;
}

/** Find the buddy a certain user belongs to.
 * @param user the user looking for
 * @return the buddy or NULL if none found.
 */
buddy_t* buddy_find_from_user(user_t* user) {
  GList* dlist;

  g_assert(BuddyModule);
  for (dlist = BuddyModule->buddies; dlist; dlist = dlist->next) {
    buddy_t* buddy = dlist->data;
    if (buddy_find_user(buddy, user)) return buddy;
  }
  return NULL;
}

/** Check whether the buddy contains user from a certain network.
 * @param buddy the buddy to look at.
 * @param ID the id of the network
 * @return TRUE if user from that network exists in that buddy, FALSE otherwise
 */
gboolean buddy_has_user(buddy_t* buddy, int ID) {
  GList* dlist;

  g_assert(BuddyModule);
  for (dlist = buddy->users; dlist; dlist = dlist->next) {
    user_t* user = dlist->data;
    if (user->ID == ID) return TRUE;
  }
  return FALSE;
}

static void xml_save_buddy(buddy_t* buddy, xmlNodePtr xroot)
{
    xmlNodePtr xnode = xmlNewChild(xroot, NULL, "Buddy", NULL);
    xmlNewProp(xnode, "Name", OBJ(buddy)->name);
    if (buddy->description)
        xmlNewTextChild(xnode, NULL, "Description", buddy->description);
    ATTR_XML_PRINT_ALL(ATTRC(buddy), xnode);
}

static void xml_save_buddies(xmlDocPtr xfile, xmlNodePtr xroot)
{
    (void)xfile;
    g_list_foreach(BuddyModule->buddies, (GFunc)xml_save_buddy, xroot);
}

static buddy_t* xml_get_buddy(xmlDocPtr xdoc, xmlNodePtr xnode)
{
    buddy_t* buddy;
    xmlChar* attr;
    xmlChar* attr2;
    xmlNodePtr xsub;
    
    attr = xmlGetProp(xnode, "Name");
    if (!attr) return NULL;
    
    buddy = buddy_new(attr);
    xmlFree(attr);
    
    xsub = xnode->xmlChildrenNode;
    while (xsub != NULL)
    {
        if (!strcmp(xsub->name, "Attr"))
        {
            attr = xmlGetProp(xsub, "Name");
            if (attr)
            {
                attr2 = xmlNodeListGetString(xdoc, xsub->xmlChildrenNode, 1);
                attr_store_str(ATTRC(buddy), "buddy", attr, attr2);
                xmlFree(attr);
                xmlFree(attr2);
            }
        }
        else if (!strcmp(xsub->name, "Description"))
        {
            attr = xmlNodeListGetString(xdoc, xsub->xmlChildrenNode, 1);
            if (attr)
            {
                buddy_set_description(buddy, attr);
                xmlFree(attr);
            }
        }
        xsub = xsub->next;
    }
    return buddy;
}

static void xml_get_buddies(xmlDocPtr xfile, xmlNodePtr xroot)
{
    xmlNodePtr xsub = xroot->xmlChildrenNode;
    
    while (xsub != NULL)
    {
        if (!strcmp(xsub->name, "Buddy"))
        {
            xml_get_buddy(xfile, xsub);
        }
        xsub = xsub->next;
    }
}

static void buddies_init()
{
    lopster_xml_register_home("core:buddies", "buddy.xml",
                              "Buddies", xml_get_buddies, xml_save_buddies);
}

static void buddies_run()
{
    buddy_load();
    g_list_foreach(BuddyModule->buddies, (GFunc)buddy_propagate_update, NULL);
}

static void buddies_exit() {
  g_list_free(BuddyModule->updater);
  BuddyModule = NULL;
}

static void buddy_module_init(buddy_module_t* module) {
  module_init(MODULE(module), "Buddy Module",
              (ModuleFunc)buddies_init,
              (ModuleFunc)buddies_run,
              (ModuleFunc)buddies_exit);

  module->buddies = NULL;
  module->updater = NULL;

  module_require(pref_module_instance);
  module_require(chat_module_instance);
  module_require(user_module_instance);
  module_require(xml_module_instance);
}

/** If you support buddies, require this module
 * @see module_require()
 */
module_t* buddy_module_instance() {
  if (!BuddyModule) {
    BuddyModule = g_malloc0(sizeof(*BuddyModule));
    buddy_module_init(BuddyModule);
  }
  return MODULE(BuddyModule);
}

#ifdef WIN32
#define _BUDDY_C_
#include "lcorealiasdef.c"
#endif
