/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.7 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include <string.h>

#include <glib.h>

#include "buffer.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

static void buffer_init(buffer_t* buffer, unsigned size) {
  buffer->max = size;
  buffer->size = 0;
  buffer->consumed = 0;
  buffer->data = g_malloc(sizeof(char)*size);
}

/** Create a new buffer.
 * @param size the size of the buffer.
 * @return the new buffer.
 */
buffer_t* buffer_new(unsigned int size) {
  buffer_t* buffer;
  
  buffer = g_malloc(sizeof(*buffer));
  buffer_init(buffer, size);
  return buffer;
}

/** Destroy a buffer.
 * @param buffer the buffer.
 */
void buffer_destroy(buffer_t* buffer) {
  if (!buffer) return;
  g_free(buffer->data);
  g_free(buffer);
}

/** Check how many bytes there are not consumed yet in the buffer.
 * @param buffer the buffer
 * @return the number of unconsumed bytes in the buffer.
 */
unsigned int buffer_avail(buffer_t* buffer) {
  if (!buffer) return 0;
  return buffer->size-buffer->consumed;
}

/** Check how many space there is still available in the buffer.
 * @param buffer the buffer
 * @return the free space
 */
unsigned int buffer_space(buffer_t* buffer) {
  if (!buffer) return 0;
  return buffer->max-buffer->size;
}

/** Make sure that there is enough space for writing.
 * @param buffer the buffer
 * @param space the space needed for writing.
 * @return the pointer to the buffer data.
 * @note The buffer data might be reallocated, so old buffer pointers
 * get invalid.
 */
char* buffer_need_space(buffer_t* buffer, unsigned int space) {
  while (buffer->max < space + buffer->size)
    buffer_resize(buffer, buffer->max*2);
  return buffer->data + buffer->size;  // buffer_write_pos();
}

/** Get the reading pointer of the buffer. This is the start of the
 * unconsumed data. 
 * @param buffer the buffer
 * @return the pointer to the first unconsumed bytes
 */
char* buffer_read_pos(buffer_t* buffer) {
  return buffer->data + buffer->consumed;
}

/** Get the writing pointer of the buffer. This is the start of the
 * free space. 
 * @param buffer the buffer
 * @return the pointer to the first free bytes
 */
char* buffer_write_pos(buffer_t* buffer) {
  return buffer->data + buffer->size;
}

/** Resize the buffer.
 * @param buffer
 * @param new_size the new size.
 */
void buffer_resize(buffer_t* buffer, unsigned int new_size) {
  if (buffer->max == new_size) return;
  buffer->data = g_realloc(buffer->data, sizeof(char)*new_size);
  buffer->max = new_size;
  if (buffer->size > buffer->max)
    buffer->size = buffer->max;
  if (buffer->consumed > buffer->size)
    buffer->consumed = buffer->size;
}

/** Clear the buffer. 
 * @param buffer the buffer.
 * @note the buffer data does not get deleted. To free the data, use
 * buffer_resize with argument 0.
 */
void buffer_clear(buffer_t* buffer) {
  if (buffer) {
    buffer->consumed = 0;
    buffer->size = 0;
  }
}

/** Consumes some bytes from the buffer. The remaining data is shifted down.
 * @param buffer the buffer
 * @param how_many number of bytes to consume.
 */
void buffer_consume(buffer_t* buffer, unsigned int how_many) {
  how_many += buffer->consumed;
  if (!how_many) return;

  if (how_many >= buffer->size) {
    buffer->size = 0;
  } else {
    buffer->size -= how_many;
    memmove(buffer->data, buffer->data+how_many, buffer->size);
  }
  buffer->consumed = 0;
  if (buffer->size < buffer->max) buffer->data[buffer->size] = 0;
}

/** Consumes some bytes from the buffer. The remaining data is not
 * shifted down which means that the amount of free space does not increase.
 * @param buffer the buffer
 * @param how_many number of bytes to consume.
 * @note several calls of this function plus one calls to
 * buffer_consume() with argument 0 is more efficient that always
 * using buffer_consume()
 */
void buffer_consume_virt(buffer_t* buffer, unsigned int how_many) {
  buffer->consumed += how_many;
  if (buffer->consumed > buffer->size)
    buffer->consumed = buffer->size;
}

/** Store some data in the buffer.
 * @param buffer the buffer
 * @param data the data to store.
 * @param len the len of the data
 */
void buffer_add_data(buffer_t* buffer, const void* data, unsigned len) {
  char* pos;
  g_assert(buffer && data);
  pos = buffer_need_space(buffer, len);
  memcpy(pos, data, len);
  buffer->size += len;
}

/** Store a string in the buffer (without traling 0).
 * @param buffer the buffer
 * @param str the string to store.
 */
void buffer_add_string(buffer_t* buffer, const char* str) {
  unsigned len;
  g_assert(buffer && str);
  
  len = strlen(str);
  buffer_need_space(buffer, len+1);
  memcpy(buffer_write_pos(buffer), str, len+1);
  buffer->size += len;
}

void buffer_add_converted_string(buffer_t* buffer, const char* str,
                                 StringConvertFunc func) {
  unsigned avail = strlen(str);
  while (avail > 0) {
    unsigned res_len;
    const char* res = func(&str, &avail, &res_len);
    if (res && res_len) {
      buffer_need_space(buffer, res_len);
      memcpy(buffer_write_pos(buffer), res, res_len);
      buffer->size += res_len;
    }
  }
}
				 
#ifdef WIN32
# define _BUFFER_C_
# include "lcorealiasdef.c"
#endif
