/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.13 $
 * $Date: 2008/12/23 13:38:14 $
 * $Author: sgop $
 */

#include <string.h>

#include "preferences.h"
#include "chat.h"
#include "string_list.h"
#include "ui.h"
#include "l_i18n.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
  chat_page_t* current;
  GList* pages;
  int buffer;
} chat_module_t;

static chat_module_t* ChatModule = NULL;

static chat_page_t*
chat_page_new_simple(const char* name, const char* label, const char* title);

/** Set the active chat buffer of a chat page. By default, buffer 0 is
 * active (which should always be the main screen). Change the buffer
 * if you like to print to a differnt buffer of the page. You always
 * should revert to buffer 0 when you are done.
 * @param num the number of the buffer.
 * @see chat_print()
 */
void chat_set_buffer(int num) {
  ChatModule->buffer = num;
}

/** The the current chat page.
 * @return the current page, or NULL
 */
chat_page_t* chat_page_get_current() {
  return ChatModule->current;
}

static chat_page_t* chat_page_search(const char* name) {
  GList* dlist;

  for (dlist = ChatModule->pages; dlist; dlist = dlist->next) {
    chat_page_t* page = dlist->data;
    if (!strcmp(OBJ(page)->name, name)) return page;
  }
  return NULL;
}

static chat_page_t* chat_page_get_printable(chat_page_t* page) {
  if (!page) page = ChatModule->current;
  if (!UI_chat_page_printable(page)) {
    page = chat_page_search("__Dummy__");
    if (!page)
      page = chat_page_new_simple("__Dummy__", _("Temporary"), _("Temporary page"));
  }
  return page;
}      

/** Notfiy that the current page has changed.
 * @param page the new current page.
 */
void chat_page_set_current(chat_page_t* page) {
  ChatModule->current = page;
}

/** Close a chat page. Always use this function and never use
 * object_kill() or your own destructor to close a page page. The
 * close() method is called and the depending on the return value the
 * page is destroyed.
 * @param page the page to close.
 */
void chat_page_close(chat_page_t* page)
{
    if (page->close(page)) object_kill(OBJ(page));
}

static gboolean chat_page_closing(chat_page_t* page)
{
    if (page->ID != -1 && page->close == chat_page_closing)
    {
        g_log("CHAT", G_LOG_LEVEL_WARNING,
              "page close func must be overloaded");
    }
    return TRUE;
}

/** Free a chat page. Call only from destroy method of derived class.
 * @param page the chat page to destroy.
 */
void chat_page_destroy(chat_page_t* page)
{
    if (page->ID != -1 && OBJ(page)->destroy == (ObjectFunc)chat_page_destroy) {
        g_log("CHAT", G_LOG_LEVEL_WARNING,
              "page destroy func must be overloaded");
    }
    ChatModule->pages = g_list_remove(ChatModule->pages, page);
    g_free(page->label);
    g_free(page->title);
    object_destroy(OBJ(page));
}

/** Change the label of the chat page.
 * @param page the chat page
 * @param label the new label for the page
 */
void chat_page_set_label(chat_page_t* page, const char* label)
{
    g_free(page->label);
    page->label = g_strdup(label);
    object_set_dirty(OBJ(page));
}

/** Change the title of the chat page.
 * @param page the chat page
 * @param tite the new title for the page
 */
void chat_page_set_title(chat_page_t* page, const char* title)
{
    g_free(page->title);
    page->title = g_strdup(title);
    object_set_dirty(OBJ(page));
}

/** Initialize a chat page. Call only from init function of derived
 * class.
 * @param page the chat page.
 * @param name the name of the page.
 * @param label the tab label of the page.
 * @param title the title of the page.
 * @param ID the ID of the plugin that creates the page.
 */
void chat_page_init(chat_page_t* page, const char* name, const char* label,
                    const char* title, int ID)
{
    object_init(OBJ(page), name, TRUE);
    object_add_type(OBJ(page), "ChatPage");
    OBJ(page)->destroy = (ObjectFunc)chat_page_destroy;
    page->label = g_strdup(label);
    page->title = g_strdup(title);
    page->active = TRUE;
    page->ID = ID;
    page->close = chat_page_closing;
    //  page->expand_command = NULL;
}

/** Create a simple new chat page. This page is not shown, you should
 * use chat_page_new_simple() if you like to use your own UI creation
 * function.
 * @param name the internal name of the chat page.
 * @param label the label of the page (this label is used for the UI tab)
 * @param title the title of the page (this title is used for the UI tab)
 * @return a new chat page.
 */
chat_page_t* chat_page_new(const char* name, const char* label, const char* title)
{
    chat_page_t* page = g_malloc0(sizeof(*page));
    chat_page_init(page, name, label, title, -1);
    ChatModule->pages = g_list_append(ChatModule->pages, page);
    return page;
}

static chat_page_t*
chat_page_new_simple(const char* name, const char* label, const char* title)
{
    chat_page_t* page;
    
    page = chat_page_new(name, label, title);
    UI_chat_create_simple_page(page);
    return page;
}

/** Print a message to the current buffer (see chat_set_buffer()). The
 * variable arguments (NULL terminated) specify the message to
 * print. Every message consists of a style name followed by the
 * according attributes. The styles must be registered in the UI.
 *
 * @param page the page to print, or NULL to use the current, or if
 * there is not current use a temp page.
 * @param ... paramters that specify the message.
 * 
 * Example:
 * <div style='border:1px solid #aaa;padding:10px;background-color:#eee'><pre>
 *    chat_print(page,
 *               "+\2nick",       user, "<\0",
 *               "+\4nick_text",  user, message,
 *               "+\0c_bracket",  ">",
 *               "+\0nl",
 *               NULL);
 * </pre></div>
 *
 * The style names (e.g. '+\\2nick') have two characters prependend,
 * the first character must always be '+', the second specifies the
 * print flags (@see defines PAD_LEFT, PAD_RIGHT; PAD_HILIGHT).
 * The rest of the message ('nick') is the actual style to use
 * (registered in the UI). The style 'nick' for instance requires two
 * attributes (arguments), 'user' and '<\\0'.
 * 
 * The above example prints an emote message: \<user message\>\\n
 */
void chat_print(chat_page_t* page, ...) {
  va_list ap;
  
  va_start(ap, page);
  page = chat_page_get_printable(page);
  UI_chat_print(page, ChatModule->buffer, ap);
  va_end(ap);
}

/** Set a page active or inactive. Inactive means that the page gets
 * inactive status in the UI.
 * @param page the chat page
 * @param active TRUE to set active, FALSE to set inactive
 */
void chat_page_set_active(chat_page_t* page, gboolean active)
{
    if (!page) return;
    if (page->active == active) return;  //TEST
    page->active = active;
    object_set_dirty(OBJ(page));
}

static chat_page_t* chat_page_global(void)
{
    chat_page_t* page = chat_page_search("__Global__");
    if (!page)
        page = chat_page_new_simple("__Global__", _("Global"), _("Global messages"));
    return page;
}

static void source_message(chat_page_t* page, const char* sender, const char* type) {
  if (sender || type) {
    chat_print(page, "+\0c_bracket", "[", NULL);
    if (sender)
      chat_print(page, "+\0c_text", sender, NULL);
    if (sender && type)
      chat_print(page, "+\0c_bracket", ":", NULL);
    if (type)
      chat_print(page, "+\0c_text", type, NULL);
    chat_print(page, "+\2c_bracket", "]", NULL);
  }
}

/** Print a normal message. This is similar to user_message() with the
 * difference that there is no user.
 * @param page the chat page, or NULL to use current one.
 * @param sender the the server/net/protocol the message comes from.
 * @param type the message type
 * @param message the message
 */
void normal_message(chat_page_t* page, const char* sender,
                    const char* type, const char* message)
{
    char* pos1 = NULL;
    
    if (!message) message = "";
    do
    {
        pos1 = strchr(message, '\n');
        if (pos1) *pos1 = 0;
        chat_print(page, "+\2server", NULL);
        source_message(page, sender, type);
        chat_print(page,
                   "+\4c_text", message,
                   "+\0nl", NULL);
        if (!pos1) break;
        message = pos1 + 1;
    } while (*message != 0);
}

/** Print a server message.
 * @param sender the sender which usually is the server/net/protocol name
 * @param message the message
 */
void server_message(const char* sender, const char* message)
{
    chat_page_t* page;
    
    if ((pref_get_ref("/core/chat/hide", int) & HIDE_SERVER)) return;
    
    page = chat_page_search("__Server__");
    if (!page) page = chat_page_new_simple("__Server__", _("Server"), _("Server messages"));
    normal_message(page, sender, NULL, message);
}

/** Print a message of the day line.
 * @param sender the sender which usually is the server/net/protocol name
 * @param message the message, may be NULL
 */
void motd_message(const char* sender, const char* message) {
  chat_page_t* page;

  if ((pref_get_ref("/core/chat/hide", int) & HIDE_MOTD)) return;
  
  page = chat_page_search("__MOTD__");
  if (!page) page = chat_page_new_simple("__MOTD__", _("MOTD"), _("Message Of The Day"));
  normal_message(page, sender, NULL, message);
}


/** Print a user from a user.
 * @param page the page to print to, or NULL to use current
 * @param sender the server/net/protocol this message comes from, or NULL
 * @param type the type of the message (private,global,etc), or NULL
 * @param user the user this message comes from
 * @param message the message, may be NULL
 * @note sender and type are used to prefix the message on a generic
 * page, you normally pass NULL here if you print a message to a page
 * where the source of the message is obvious (e.g. in napster public page)
 */
void user_message(chat_page_t* page, const char* sender,
                  const char* type, const char* user, const char* message)
{
    char* pos1;
    
    if (!message) message = "";
    
    do
    {
        pos1 = strchr(message, '\n');
        if (pos1) *pos1 = 0;
        chat_print(page, "+\2stamp", NULL);
        source_message(page, sender, type);
        chat_print(page,
                   "+\2nick", user, "<>",
                   "+\4c_text", message,
                   "+\0nl", NULL);
        if (!pos1) break;
        message = pos1 + 1;
    }
    while (*message != 0);
}

/** Print a global message.
 * @param sender the sender which usually is the server/net/protocol name
 * @param user the user the message comes from, can be NULL
 * @param message the message, may be NULL
 */
void global_message(const char* sender, const char* user, const char* message) {
  chat_page_t* page;

  if (pref_get_ref("/core/chat/hide", int) & HIDE_GLOBAL) return;

  if (pref_get_ref("/core/chat/pipe", int) & PIPE_GLOBAL)
    page = chat_page_global();
  else
    page = NULL;
  
  if (user) {
    user_message(page, sender, page?NULL:_("global"), user, message);
  } else {
    normal_message(page, sender, NULL, message);
  }
}

/** Print a emote message. This is quite similar to user_message()
 * @param page the chat page, or NULL to use current one.
 * @param sender the the server/net/protocol the message comes from.
 * @param type the message type
 * @param user the user this message comes from
 * @param message the message, may be NULL
 */
void emote_message(chat_page_t* page, const char* sender,
                   const char* type, const char* user, 
                   const char* message)
{
    char* pos1;
    
    if (!message) message = "";
    
    do
    {
        pos1 = strchr(message, '\n');
        if (pos1) *pos1 = 0;
        chat_print(page, "+\2stamp", NULL);
        source_message(page, sender, type);
        chat_print(page,
                   "+\2nick", user, "<\0",
                   "+\4nick_text", user, message,
                   "+\0c_bracket", ">",
                   "+\0nl", NULL);
        if (!pos1) break;
        message = pos1 + 1;
    }
    while (*message != 0);
}

/** Print a private message. This is quite similar to
 * user_message(). You should use user_message() if you print to a
 * private page of that user as this function uses special colors
 * @param page the chat page, or NULL to use current one.
 * @param sender the the server/net/protocol the message comes from.
 * @param type the message type
 * @param user the user this message comes from
 * @param message the message, may be NULL
 */
void private_message(chat_page_t* page, const char* sender,
                     const char* type, const char* user, 
                     const char* message) {
  char* pos1;

  if (!message) message = "";
  if (!sender) sender = "??";
  
  do {
    pos1 = strchr(message, '\n');
    if (pos1) *pos1 = 0;
    chat_print(page, "+\2stamp", NULL);
    source_message(page, sender, type);
    chat_print(page,
               "+\2nick", user, "<>",
               "+\4c_priv_msg", message,
               "+\0nl", NULL);
    if (!pos1) break;
    message = pos1 + 1;
  } while (*message != 0);
}

/** @defgroup ChatPrefValues Chat settings
 * @ingroup PrefModuleValues
 * 
 * <table border='0' style='border-spacing:1px;background-color:#888'>
 *  <tr style='background-color:#fff;'>
 *    <td>/core/chat/hide</td>        <td>int</td>       <td>hide flags for chat message</td>
 *  </tr><tr style='background-color:#efe;'>
 *    <td>/core/chat/pipe</td>        <td>int</td>       <td>pipe flags for chat message</td>
 *  </tr><tr style='background-color:#fff;'>
 *    <td>/core/chat/timestamp</td>   <td>int</td>       <td>Format of the timestamp</td>
 *  </tr><tr style='background-color:#efe;'>
 *    <td>/core/chat/auto_rejoin</td> <td>boolean</td>   <td>whether or not to auto rejoin channels when kicked/con lost, etc</td>
 *  </tr><tr style='background-color:#fff;'>
 *    <td>/core/chat/colored_nicks</td><td>int</td>      <td>how to color the nicks in the chat page</td>
 *  </tr><tr style='background-color:#efe'>
 *    <td>/core/chat/prefix/server</td><td>string</td>   <td>prefix for server message</td>
 *  </tr><tr style='background-color:#fff;'>
 *    <td>/core/chat/prefix/client</td><td>string</td>   <td>prefix for client message</td>
 *  </tr><tr style='background-color:#efe'>
 *    <td>/core/chat/prefix/join</td><td>string</td>   <td>prefix for join message</td>
 *  </tr><tr style='background-color:#fff;'>
 *    <td>/core/chat/prefix/part</td><td>string</td>   <td>prefix for part message</td>
 *  </tr><tr style='background-color:#efe'>
 *    <td>/core/chat/prefix/text</td><td>string</td>   <td>prefix for other message</td>
 *  </tr>
 * </table>
 *
 * @{
 * @}
 */
static void chats_init() {
  pref_register_int   ("/core/chat/hide", TRUE, HIDE_NONE);
  pref_add_constraint_minmax("/core/chat/hide", 0, HIDE_MASK);
  pref_register_int   ("/core/chat/pipe", TRUE, PIPE_PRIVATE|PIPE_GLOBAL);
  pref_add_constraint_minmax("/core/chat/pipe", 0, PIPE_MASK);
  pref_register_int   ("/core/chat/timestamp", TRUE, TIME_HOUR_MIN_SEC);
  pref_add_constraint_minmax("/core/chat/timestamp", TIME_NONE, TIME_HOUR_MIN_SEC);
  pref_register_bool  ("/core/chat/auto_rejoin", TRUE, TRUE);
  pref_register_int   ("/core/chat/colored_nicks", TRUE, 2);
  pref_add_constraint_minmax("/core/chat/colored_nicks", 0, 3);

  pref_register_string("/core/chat/prefix/server", TRUE, "*");
  pref_register_string("/core/chat/prefix/client", TRUE, "#");
  pref_register_string("/core/chat/prefix/join", TRUE, ">");
  pref_register_string("/core/chat/prefix/part", TRUE, "<");
  pref_register_string("/core/chat/prefix/text", TRUE, " ");

  string_list_register("core:chat:hilight", N_("Chat highlightning"), TRUE);
}

static void chat_module_init(chat_module_t* module) {
  module_init(MODULE(module), "Chat Module",
	      (ModuleFunc)chats_init, NULL, NULL);

  module->current = NULL;
  module->pages = NULL;
  module->buffer = 0;

  module_require(pref_module_instance);
  module_require(string_list_module_instance);
  module_require(ui_module_instance);
  
}


/** Require this module if you like to use chat.
 * @see module_require()
 * @return the chat module
 */
module_t* chat_module_instance() {
  if (!ChatModule) {
    ChatModule = g_malloc0(sizeof(*ChatModule));
    chat_module_init(ChatModule);
  }
  return MODULE(ChatModule);
}

#ifdef WIN32
#define _CHAT_C_
#include "lcorealiasdef.c"
#endif
