/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.7 $
 * $Date: 2006/07/22 17:11:46 $
 * $Author: sid4u $
 */

#include <string.h>
#include <stdio.h>

#include <glib.h>

#include "l_i18n.h"
#include "chat.h"
#include "utils.h"
#include "plugin.h"
#include "commands.h"
#include "completion.h"
#ifdef WIN32
#include "win32_defs.h"
#endif
#ifdef WIN32
#include "lcorealias.h"
#endif

#define MAX_HISTORY   1000


typedef struct {
  module_t _module;
  GList* commands;
  completion_t* completion;
  char* history[MAX_HISTORY];
  char* historyM[MAX_HISTORY];
  int hist_pos;
  int hist_end;
  GPtrArray* handler;
} command_module_t;

static command_module_t* CommandModule = NULL;

static void command_history_apply(const char* str);

/** Install a command handler.
 * @param ID the id of the network (plugin)
 * @param handler the command handler
 */
void command_handler_install(int ID, CommandHandler handler) {
  
  g_assert(CommandModule);
  if (ID < -1 || ID >= 1000) {
    g_log("COMMAND", G_LOG_LEVEL_WARNING,
          "tried to set command handler at invalid index %d", ID);
  } else {
    if (CommandModule->handler->len < (unsigned)ID+2) 
      g_ptr_array_set_size(CommandModule->handler, ID+2);
    if (g_ptr_array_index(CommandModule->handler, ID+1)) {
      g_log("COMMAND", G_LOG_LEVEL_WARNING,
            "override command handler at index %d", ID);
    }
    g_ptr_array_index(CommandModule->handler, ID+1) = handler;
  }
}

static CommandHandler command_handler_find(int ID) {
  if (ID < -1 || CommandModule->handler->len < (unsigned)ID+2) return NULL;
  return g_ptr_array_index(CommandModule->handler, ID+1);
}

/** Print an error.
 * @param page the chat page to print to, or NULL to use current.
 * @param message the error message.
 */
void command_error(chat_page_t* page, const char* message) {
  g_assert(CommandModule);
  chat_print(page, "+\0client",
             "+\1c_client", message,
             "+\1nl", NULL);
}

static void command_destroy(command_t* command) {
  g_free(command->name);
  g_free(command->args);
  g_free(command->description);
  g_free(command);
}

/** Private function, not used by plugins 
 * @param iname command name
 * @param id the network id.
 * @return the command, or NULL if not found.
 */
command_t* command_search(const char *iname, int id) {
  command_t* comm;
  GList* dlist;
  
  g_assert(CommandModule);
  if (!iname) return NULL;
  if (*iname == '/') iname++;
  if (*iname == 0) return NULL;

  for (dlist = CommandModule->commands; dlist; dlist = dlist->next) {
    comm = dlist->data;
    if ((id == -2 || comm->id == id) &&
        !strcasecmp(iname, comm->name)) return comm;
  }
  return NULL;
}

/** Initialize the command. Call from the init method of your derived
 * class.
 * @param command the command.
 * @param id the id of the network which defines the command.
 * @param name the name of the command.
 * @param syntax the syntax of the command arguments.
 * @param des the description of the command
 */
void command_init(command_t* command, int id, const char* name,
                  const char* syntax, const char* des) {
  
  g_assert(CommandModule);
  g_assert(command);
  g_assert(!command_search(command->name, id));
  command->id = id;
  command->name = g_strdup(name);
  command->args = g_strdup(syntax);
  command->description = g_strdup(des);
  g_log("COMMAND", G_LOG_LEVEL_DEBUG,
        "registered new command [%s] '%s'",
        plugin_get_name(plugin_get(id)), name);
  CommandModule->commands = g_list_append(CommandModule->commands, command);

}

static command_t*
command_search_for_plugin(GList* commands, int id, unsigned int* num) {
  GList* dlist;
  command_t* result = NULL;
  
  if (num) *num = 0;
  
  for (dlist = commands; dlist; dlist = dlist->next) {
    command_t* c = dlist->data;
    if (c->id == id) {
      if (num) (*num)++;
      if (!result) result = c;
    }
  }
  return result;
}

static void
commands_print(chat_page_t* page, GList* results, int id) {
  int len = 0;
  GList* dlist;

  for (dlist = results; dlist; dlist = dlist->next) {
    command_t* com = dlist->data;
    if (id != -2 && com->id != id) continue;
    len++;
  }

  if (id == -2) {
    chat_print(page, "+\0client",
               "+\1c_client", l_printf(_("%d matching commands"), len),
               "+\0nl", NULL);
  } else {
    chat_print(page, "+\0client",
               "+\1c_client", l_printf(_("%d matching commands"), len),
               "+\1c_text", "(",
               "+\0c_text", plugin_get_name(plugin_get(id)),
               "+\0c_text", ")",
               "+\0nl", NULL);
  }
  for (dlist = results; dlist; dlist = dlist->next) {
    command_t* com = dlist->data;
    if (id != -2 && com->id != id) continue;
    chat_print(page, "+\0client",
               "+\1c_text", "  ",
               "+\1c_bracket", "[",
               "+\0c_client", l_printf("%-12s", com->name),
               "+\0c_bracket", "]",
               "+\1c_text", "(",
               "+\0c_text", plugin_get_name(plugin_get(com->id)),
               "+\0c_text", ")",
               "+\0nl", NULL);
  }
}

void command_handle(chat_page_t* page, char *data) {
  char* command;
  command_t* comm;
  GList* results;
  int id;
  CommandHandler handler;

  g_assert(CommandModule);
  if (!data) return;
  
  command_history_apply(data);

  if (*data == '/') {
    command = arg(data+1, 0);
    data = arg(NULL, 1);

    // get all possible commands
    results = completion_complete(CommandModule->completion, command, NULL);
  } else {
    results = NULL;
    command = NULL;
  }

  COMMAND_ERROR(!results && command, page,
                _("Unknown command! Type /help for a list of commands"));

  // check whether a specific plugin was specified
  if (command && data) {
    id = plugin_get_id(data);
    if (id > -2) {
      data = arg(data, 0);
      data = arg(NULL, 1);
    }
  } else {
    id = -2;
  }

  if (results) {
    unsigned int num;
    if (id > -2) {
      comm = command_search_for_plugin(results, id, &num);
    } else if (g_list_length(results) == 1) {
      comm = results->data;
      num = 1;
    } else {
      id = (page?page->ID:-1);
      comm = command_search_for_plugin(results, id, &num);
    }
    // note: id != -2 now
    if (!comm) {
      chat_print(page, "+\0client",
                 "+\1proto", plugin_get_name(plugin_get(id)),
                 "+\1c_client", _("Command /"),
                 "+\0c_text", command,
                 "+\1c_client", _("is not supported"),
                 "+\0nl", NULL);
      return;
    } else if (num > 1) {
      commands_print(page, results, id);
      return;
    } else if (strcasecmp(command, comm->name)) {
      // FIXME: dangerous? the current page might change with chat_print()..
      chat_print(page, "+\0client",
                 "+\1c_client", _("Using command /"),
                 "+\0c_text", comm->name,
                 "+\1c_text", "(",
                 "+\0proto", plugin_get_name(plugin_get(comm->id)),
                 "+\0c_text", ")",
                 "+\0nl", NULL);
    }
    id = comm->id;
  } else {
    comm = NULL;
    id = (page?page->ID:-1);
  }
  
  handler = command_handler_find(id);
  if (!handler)
    COMMAND_ERROR(!handler, page, 
                  l_printf(_("%s has no command handler installed"), 
                           plugin_get_name(plugin_get(id))));
  handler(page, comm, data);
}


/** Get the name of the command.
 * @param command the command.
 * @return the name of the command.
 */
static char* command_name(command_t* command) {
  g_assert(CommandModule);
  return command->name;
}

static char* expand_command(chat_page_t* page, const char* command) {
  GList* results;
  char* result;
  int len;

  results = completion_complete(CommandModule->completion, command, &result);
  len = g_list_length(results);
  if (len > 1) {
    commands_print(page, results, -2);
  } else if (len == 0) {
    chat_print(page, "+\0client",
               "+\1c_client", _("No matching command found"),
               "+\0nl", NULL);
  }
  if (result) {
    char* temp = result;
    result = g_strdup_printf("/%s%s", temp, (len==1)?" ":"");
    g_free(temp);
  }
  return result;
}

char* command_expand(chat_page_t* page, const char* command) {
  char* pos;
  
  g_assert(CommandModule);
  g_assert(command);

  if (*command == 0 || *command != '/') {
    if (page && page->expand) return page->expand(page, command);
  } else if (*command == '/') {
    pos = strchr(command, ' ');
    if (!pos) return expand_command(page, command+1);
    else if (page && page->expand) return page->expand(page, command);
  }
  chat_print(page, "+\0client",
             "+\1c_client", _("No expand function found"),
             "+\0nl", NULL);
  return NULL;
}

static void command_history_update_internal(int pos, const char* str) {
  if (!str) return;
  g_free(CommandModule->historyM[pos]);
  CommandModule->historyM[pos] = g_strdup(str);
}

static void command_history_apply(const char* str) {
  // reset modification
  command_history_update_internal
    (CommandModule->hist_pos, CommandModule->history[CommandModule->hist_pos]);

  // add new item
  g_free(CommandModule->history[CommandModule->hist_end]);
  CommandModule->history[CommandModule->hist_end] = g_strdup(str);

  // init modification
  command_history_update_internal(CommandModule->hist_end, 
                                  CommandModule->history[CommandModule->hist_end]);

  // step to next item (new end of history)
  if (++CommandModule->hist_end >= MAX_HISTORY) CommandModule->hist_end = 0;
  CommandModule->hist_pos = CommandModule->hist_end;

  // clear it
  g_free(CommandModule->history[CommandModule->hist_pos]);
  CommandModule->history[CommandModule->hist_pos] = g_strdup("");

  // and init modification
  command_history_update_internal(CommandModule->hist_pos, 
                                  CommandModule->history[CommandModule->hist_pos]);

}

void command_history_update(const char* str) {
  g_assert(CommandModule);
  command_history_update_internal(CommandModule->hist_pos, str);
}

const char* command_history_up() {
  int pos;

  g_assert(CommandModule);

  pos = CommandModule->hist_pos-1;
  if (pos < 0) pos = MAX_HISTORY-1;

  if (pos == CommandModule->hist_end || !CommandModule->historyM[pos]) {
    return NULL;
  } else {
    CommandModule->hist_pos = pos;
    return CommandModule->historyM[pos];
  }
}

const char* command_history_down() {
  g_assert(CommandModule);
  if (CommandModule->hist_pos == CommandModule->hist_end) return NULL;
  if (++CommandModule->hist_pos >= MAX_HISTORY) CommandModule->hist_pos = 0;
  return CommandModule->historyM[CommandModule->hist_pos];
}

void command_foreach(CommandForeachFunc func, int id, void* data) {
  GList* dlist;
  g_assert(CommandModule);
  for (dlist = CommandModule->commands; dlist; dlist = dlist->next) {
    command_t* command = dlist->data;
    if (id < -1 || command->id == id) func(command, data);
  }
}

static void commands_init() {
  
  CommandModule->completion = 
    completion_new((CompletionFunc)command_name);
  completion_set_compare
    (CommandModule->completion, (GCompletionStrncmpFunc)strncasecmp);
  completion_set_items(CommandModule->completion, &CommandModule->commands);

  CommandModule->handler = g_ptr_array_new();
  
}

static void commands_run() {
}
static void commands_exit() {
  int i1;

  g_list_foreach(CommandModule->commands, (GFunc)command_destroy, NULL);
  g_list_free(CommandModule->commands);
  CommandModule->commands = NULL;
  completion_free(CommandModule->completion);
  for (i1 = 0; i1 < MAX_HISTORY; i1++) {
    g_free(CommandModule->history[i1]);
    g_free(CommandModule->historyM[i1]);
  }

  g_ptr_array_free(CommandModule->handler, TRUE);
}

static void command_module_init(command_module_t* module) {
  module_init(MODULE(module), "Command Module",
	      (ModuleFunc)commands_init,
	      (ModuleFunc)commands_run,
	      (ModuleFunc)commands_exit);
  
  module->commands = NULL;
  module->completion = NULL;
  module->hist_pos = 0;
  module->hist_end = 0;
  module->handler = NULL;

  memset(module->history, 0, sizeof(char*)*MAX_HISTORY);
  memset(module->historyM, 0, sizeof(char*)*MAX_HISTORY);

}


module_t* command_module_instance() {
  if (!CommandModule) {
    CommandModule = g_malloc0(sizeof(*CommandModule));
    command_module_init(CommandModule);
  }
  return MODULE(CommandModule);
}

#ifdef WIN32
#define _COMMAND_C_
#include "lcorealiasdef.c"
#endif
