/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.7 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include <string.h>

#include <glib.h>

#include "completion.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

struct _completion_t {
  GList** items;
  GList* cache;

  CompletionFunc func;
  gchar* prefix;
  CompletionStrncmpFunc strncmp_func;
};

static char* id_func(char* str) {
  return str;
}

/** Create a new completion object.
 * @param func function which converts an object to a string, or NULL
 * to take the objects itself as strings.
 * @return a new completion object
 */
completion_t* completion_new(CompletionFunc func) {
  completion_t* gcomp;
  
  gcomp = g_new(completion_t, 1);
  gcomp->items = NULL;
  gcomp->cache = NULL;
  gcomp->prefix = NULL;
  gcomp->func = func?func:(CompletionFunc)id_func;
  gcomp->strncmp_func = strncmp;

  return gcomp;
}

/** Set the items which are used to complete with.
 * @param cmp the completion object
 * @param items a pointer to the items
 * @note The GList the item pointer points to must stay valid as long
 * as this completion exists.
 */
void completion_set_items(completion_t* cmp, GList** items) {
  
  g_return_if_fail(cmp != NULL);
  
  if (cmp->prefix) {
    g_free(cmp->prefix);
    cmp->prefix = NULL;
  }
  g_list_free(cmp->cache);
  cmp->cache = NULL;

  cmp->items = items;
}

static void completion_check_cache(completion_t* cmp, gchar** new_prefix) {
  register GList* list;
  register gsize len;  
  register gsize i;
  register gsize plen;
  gchar* postfix;
  gchar* s;
  
  if (!cmp->cache) {
    if (new_prefix) *new_prefix = NULL;
    return;
  }
  
  len = strlen(cmp->prefix);
  list = cmp->cache;
  s = cmp->func(list->data);
  postfix = s + len;
  plen = strlen(postfix);
  list = list->next;

  strncpy(cmp->prefix, s, len);
  
  while (list && plen) {
    s = cmp->func(list->data);
    s += len;
    for (i = 0; i < plen; ++i) {
      if (postfix[i] != s[i])
	break;
    }
    plen = i;
    list = list->next;
  }
  
  if (new_prefix) {
    *new_prefix = g_new0(gchar, len + plen + 1);
    strncpy(*new_prefix, cmp->prefix, len);
    strncpy(*new_prefix + len, postfix, plen);
  }
}

/** Complete a prefix and return a list of possible matches.
 * @param cmp the completion object.
 * @param prefix a string to match.
 * @param[out] new_prefix the most common prefix of the matches.
 * @return a list of matches, see completion_set_items()
 */
GList* completion_complete(completion_t* cmp, const gchar* prefix, gchar** new_prefix) {
  gsize len;
  GList* list;
  
  g_return_val_if_fail(cmp != NULL, NULL);
  g_return_val_if_fail(cmp->items != NULL, NULL);
  g_return_val_if_fail(prefix != NULL, NULL);

  g_assert(cmp->func);
  
  len = strlen(prefix);
  
  g_list_free(cmp->cache);
  cmp->cache = NULL;

  list = *(cmp->items);
  while (*prefix && list) {
    if (!cmp->strncmp_func(prefix, cmp->func(list->data), len))
      cmp->cache = g_list_append(cmp->cache, list->data);
    list = list->next;
  }

  g_free(cmp->prefix);

  if (cmp->cache) cmp->prefix = g_strdup(prefix);
  else cmp->prefix = NULL;

  completion_check_cache(cmp, new_prefix);
  
  return *prefix?cmp->cache:*(cmp->items);
}

/** Destroy a completion object. 
 * @param cmp the completion
 */
void completion_free(completion_t* cmp) {
  g_return_if_fail(cmp != NULL);
  
  g_list_free(cmp->cache);
  g_free(cmp->prefix);
  g_free(cmp);
}

/** Set the completion compare func. Use this function to use another 
 * compare function than strncmp().
 * @param cmp the completion
 * @param strncmp_func the new compare function
 */
void completion_set_compare(completion_t* cmp, CompletionStrncmpFunc strncmp_func) {
  cmp->strncmp_func = strncmp_func;
}

#ifdef WIN32
# define _COMPLETION_C_
# include "lcorealiasdef.c"
#endif

