/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.28 $
 * $Date: 2007/11/21 20:18:23 $
 * $Author: sgop $
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifndef WIN32
#include <unistd.h>
#endif

#include <libxml/parser.h>

#include "download.h"
#include "l_base.h"
#include "l_timer.h"
#include "ui.h"
#include "utils.h"
#include "l_limit.h"
#include "preferences.h"
#include "action_item.h"
#include "open_actions.h"
#include "plugin.h"
#include "xml_handler.h"
#include "l_i18n.h"
#include "l_io.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
    module_t _module;
    GList* downloads;
    GList* schedule;
    GList* started;
    gboolean queue_check;
    gboolean ignore_save;
    rate_t* rate;
    FileToSourceFunc* source_funcs;
    DownloadMatchFunc* match_funcs;
} download_module_t;

static download_module_t* DownloadModule = NULL;

#if !defined(O_BINARY)
# define O_BINARY  0
#endif

#define SEG_OVERLAP   100

static void download_deactivate(download_t* download);
static void download_insert_done_segment_with_error(download_t* download, file_size_t start, file_size_t stop, unsigned err);
static void download_insert_done_segment(download_t* download, file_size_t start, file_size_t stop);
static void download_insert_free_segment(download_t* download, file_size_t start, file_size_t stop);
static void source_remove_segment(source_t* source, segment_t* segment);
static void download_release_segment(download_t* download, source_t* source, segment_t* segment);
static gboolean download_check_finish(download_t* download, gboolean move);
static void source_set_status(source_t* source, int status);



void download_register_match(int ID, DownloadMatchFunc match)
{
    g_assert(DownloadModule);
    if (ID < 0 || ID >= plugin_get_number()) return;
    DownloadModule->match_funcs[ID] = match;
}

DownloadMatchFunc download_match_get(int ID)
{
    g_assert(DownloadModule);
    if (ID < 0 || ID >= plugin_get_number()) return NULL;
    return DownloadModule->match_funcs[ID];
}

gboolean download_matches_source(download_t* download, file_t* file, user_t* user)
{
    DownloadMatchFunc match = user?download_match_get(user->ID):NULL;
    if (!match) return download_check_file(download, file);  //fallback.
    else return match(download, file, user);
}


static void downloads_save(void) {
    lopster_xml_save("core:downloads", 20000);
}

static void downloads_load(void) {
    attr_allow_unregistered(TRUE);
    DownloadModule->ignore_save = TRUE;
    
    lopster_xml_load("core:downloads");
    
    DownloadModule->ignore_save = FALSE;
    attr_allow_unregistered(FALSE);
}


static void download_create_search_keys(download_t* download) 
{
    char* suf;
    GSList* keys = NULL;
    unsigned cnt = 3;

    if (!FILE(download)->filename) return;
    
    if (download->search_keys)
    {
        g_slist_foreach(download->search_keys, (GFunc)g_free, NULL);
        g_slist_free(download->search_keys);
        download->search_keys = NULL;
    }

    suf = strrchr(FILE(download)->filename, '.');
    if (suf) *suf = 0;
    make_list_from_string(&keys, FILE(download)->filename, " ^_-.,:;\t(){}[]<>\"!#*+'");
    
    while (cnt && keys)
    {
        GSList* res = NULL;
        GSList* dlist;
        unsigned len = 0;
        for (dlist = keys; dlist; dlist = dlist->next)
        {
            char* key = dlist->data;
            unsigned l = strlen(key);
            if (l > len) 
            {
                len = l;
                res = dlist;
            }
        }
        // don't add keys with less than 3 chars if we already
        // got other keys
        if (len < 3 && download->search_keys) break;
        download->search_keys = g_slist_append(download->search_keys, res->data);
        keys = g_slist_delete_link(keys, res);
        cnt--;
    }
    g_slist_foreach(keys, (GFunc)g_free, NULL);
    g_slist_free(keys);
    if (suf) *suf = '.';
}

static void download_limit_changed(void)
{
    DownloadModule->queue_check = TRUE;
}

static void download_add_limit_real(download_t* download, limit_t* limit)
{
    download->limits = g_list_append(download->limits, limit);
    if (!object_monitor_find(OBJ(limit), (ObjectFunc)download_limit_changed, NULL))
        object_monitor_d(OBJ(limit), (ObjectFunc)download_limit_changed, NULL);
}

static void download_add_limit(download_t* download, const char* name)
{
    limit_t* limit = limit_find(name);
    if (!limit) {
        g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
              "limit '%s' not found", name);
        return;
    }
    download_add_limit_real(download, limit);
}

int download_get_rate(download_t* download)
{
    return rate_get_rate(download->rate);
}

static void download_destroy(download_t* download)
{
    GList* dlist;
    
    DownloadModule->downloads = g_list_remove(DownloadModule->downloads, download);
    
    download_deactivate(download);
    
    if (download->rate) rate_destroy(download->rate);
    
    g_slist_foreach(download->search_keys, (GFunc)g_free, NULL);
    g_slist_free(download->search_keys);
    download->search_keys = NULL;
    
    g_list_free(download->limits);
    download->limits = NULL;
    
    for (dlist = download->segments_used; dlist; dlist = dlist->next)
    {
        dl_segment_t* seg = dlist->data;
        segment_destroy(seg->segment);
        g_free(seg);
    }
    g_list_free(download->segments_used);
    download->segments_used = NULL;
    
    segments_destroy(download->segments_done);
    download->segments_done = NULL;
    
    segments_destroy(download->segments_free);
    download->segments_free = NULL;
    
    file_destroy(FILE(download));
    
    downloads_save();
}

static void download_update_filename(download_t* download) 
{
    static gboolean warned = FALSE;
    const char* path = pref_get_ptr("/core/path/incomplete", char);
    char* filename;
    
    // download to current directory if no incomplete path set.
    if (path && *path)
    {
        filename = g_strdup_printf(SDIR("%s") "%s", path, FILE(download)->filename);
    } 
    else 
    {
        if (!warned) 
        {
            lopster_notify
                (LOPSTER_MESSAGE_WARNING, _("Incomplete settings"),
                 _("You should setup a folder for your incomplete files! "
                   "Otherwise the current working directory is used for downloading."));
            warned = TRUE;
        }
        filename = g_strdup(FILE(download)->filename);
    }
    file_set_locname(FILE(download), filename);
}

static void download_init(download_t* download, file_t* file)
{
    file_init(FILE(download));
    if (file) 
    {
        file_copy(FILE(download), file, "file");
        download_create_search_keys(download);
        download_insert_free_segment(download, 0, file->size);
        download_update_filename(download);
    }
    object_add_type(OBJ(download), "Download");
    OBJ(download)->destroy = (ObjectFunc)download_destroy;
    
    download->sources = NULL;
    download->flags = 0;
    download->no_active = 0;
    download->segments_done = NULL;
    download->rate = rate_new(0);
  
    // add limit per file.
    download_add_limit(download, "Download File Limit");
}

static void on_download_destroy(download_t* download)
{
    if (download && !download_check_finish(download, FALSE))
    {
        char* temp = locale_from_utf8(FILE(download)->localname);
        // FIXME: only delete if existent
        // FIXME: win32 unlink safe ?
        if (temp)
        {
            unlink(temp);
            g_free(temp);
        }
    }
}

static download_t* download_new(file_t* file)
{
    download_t* download = g_malloc0(sizeof(*download));
    download_init(download, file);
    
    DownloadModule->downloads = g_list_append(DownloadModule->downloads, download);
    object_notify_d(OBJ(download), (ObjectFunc)on_download_destroy, NULL);
    
    return download;
}

download_t* download_from_file(file_t* file)
{
    download_t* download = download_new(file);

/*   printf("----FILE %p\n", attr_find(ATTRC(file), "file", "urn:tree:tiger")); */
/*   printf("----DOWN %p\n", attr_find(ATTRC(download), "file", "urn:tree:tiger")); */
    
    g_assert(DownloadModule);
    search_new_from_download(download);
    
    object_make_public(OBJ(download), NULL);

    downloads_save();
    return download;
}

static void xml_save_segment(segment_t* segment, xmlNodePtr xsub)
{
    xmlNodePtr xsub2 = xmlNewChild(xsub, NULL, "Segment", NULL);
    xmlNewProp(xsub2, "Start", l_printf(FF, segment->start));
    xmlNewProp(xsub2, "Stop", l_printf(FF, segment->stop));
    if (segment->error)
        xmlNewProp(xsub2, "Error", l_printf("%u", segment->error));
}

static void xml_save_download(download_t* download, xmlNodePtr xnode)
{
    file_t* file = FILE(download);
    GList* dlist;
    xmlNodePtr xsub;
    GSList* slist;
    
    xsub = xmlNewChild(xnode, NULL, "Download", NULL);
    xmlNewProp(xsub, "Size", l_printf(FF, file->size));
    xmlNewTextChild(xsub, NULL, "Name", file->localname);
    for (slist = download->search_keys; slist; slist = slist->next)
    {
        char* key = slist->data;
        xmlNewTextChild(xsub, NULL, "SearchKey", key);
    }
    
    ATTR_XML_PRINT_ALL(ATTRC(file), xsub);
    
    segments_foreach(download->segments_done,
                     (SegmentsForeach)xml_save_segment, xsub);
    
    for (dlist = download->segments_used; dlist; dlist = dlist->next)
    {
        dl_segment_t* seg = dlist->data;
        segment_t* segment = seg->segment;
        xmlNodePtr xsub2;
        
        if (segment->size <= 100) continue;
        xsub2 = xmlNewChild(xsub, NULL, "Segment", NULL);
        xmlNewProp(xsub2, "Start", l_printf(FF, segment->start));
        xmlNewProp(xsub2, "Stop", l_printf(FF, segment->start+segment->size));
        if (segment->error)
            xmlNewProp(xsub2, "Error", l_printf("%u", segment->error));
    }
}

static void xml_save_downloads(xmlDocPtr xfile, xmlNodePtr xroot)
{
    (void)xfile;
    g_list_foreach(DownloadModule->downloads, (GFunc)xml_save_download, xroot);
}
 
static void _download_extract_free_segments(download_t* download)
{
    g_assert(!download->segments_used);
    g_assert(!download->segments_free);
    
    download->segments_free =
        segments_create_complement(download->segments_done, FILE(download)->size);
}

static void _download_created(download_t* download)
{
    if (!download->search_keys) download_create_search_keys(download);
    _download_extract_free_segments(download);
    object_make_public(OBJ(download), NULL);
    download_check_finish(download, FALSE);
}

static void xml_get_download(xmlDocPtr xfile, xmlNodePtr xnode) 
{
    xmlNodePtr xsub;
    xmlChar* attr;
    xmlChar* attr2;
    file_t* file;
    download_t* download;
    
    download = download_new(NULL);
    file = FILE(download);
    
    attr = xmlGetProp(xnode, "Size");
    file->size = strtoll(attr, NULL, 10);
    xmlFree(attr);
    
    xsub = xnode->xmlChildrenNode;
    while (xsub) 
    {
        if (!strcmp(xsub->name, "Name")) 
        {
            attr = xmlNodeListGetString(xfile, xsub->xmlChildrenNode, 1);
            if (attr) 
            {
                file_set_locname(file, g_strdup(attr));
                file_set_origname(file, g_strdup(attr));
                xmlFree(attr);
            }
        } 
        else if (!strcmp(xsub->name, "Attr"))
        {
            attr = xmlGetProp(xsub, "Name");
            if (attr) {
                attr2 = xmlNodeListGetString(xfile, xsub->xmlChildrenNode, 1);
                attr_store_str(ATTRC(file), "file", attr, attr2);
                xmlFree(attr);
                xmlFree(attr2);
            }
        } 
        else if (!strcmp(xsub->name, "SearchKey")) 
        {
            attr = xmlNodeListGetString(xfile, xsub->xmlChildrenNode, 1);
            if (attr)
            {
                download->search_keys =
                    g_slist_append(download->search_keys, g_strdup(attr));
                xmlFree(attr);
            }
        } 
        else if (!strcmp(xsub->name, "Segment"))
        {
            file_size_t start, stop;
            xmlChar* attr3 = xmlGetProp(xsub, "Error");
            
            attr = xmlGetProp(xsub, "Start");
            attr2 = xmlGetProp(xsub, "Stop");
            if (attr && attr2) 
            {
                start = strtoll(attr, NULL, 10);
                stop = strtoll(attr2, NULL, 10);
                if (attr3)
                    download_insert_done_segment_with_error(download, start, stop, atoi(attr3));
                else
                    download_insert_done_segment(download, start, stop);
            }
            if (attr) xmlFree(attr);
            if (attr2) xmlFree(attr2);
            if (attr3) xmlFree(attr3);
        }
        xsub = xsub->next;
    }
    
    _download_created(download);
}

static void xml_get_downloads(xmlDocPtr xfile, xmlNodePtr xroot)
{
    xmlNodePtr xsub = xroot->xmlChildrenNode;
    xsub = xroot->xmlChildrenNode;
    while (xsub != NULL)
    {
        if (!strcmp(xsub->name, "Download")) 
        {
            xml_get_download(xfile, xsub);
        }
        xsub = xsub->next;
    }
}

/** Import incomplete downloads from lopster1.
 * @param file the open file
 */
void download_import_from_lopster1(FILE* file) 
{
    char line[2048];
    download_t* download = NULL;

    while (mfgets(line, sizeof(line), file)) 
    {
        if (line[0] == '-') 
        {
            char* temp1 = arg(line+1, 0);
            char* temp2 = arg(NULL, 0);
            char* temp3 = arg(NULL, 0);  // ignored
            char* temp4 = arg(NULL, 0);  // ignored
            file_size_t start;
            file_size_t length;
          
            if (!temp2 || !download) continue;
#ifdef WIN32
            start = g_ascii_strtoull(temp1, NULL, 10);
            length = g_ascii_strtoull(temp2, NULL, 10);
#else
            // g_ascii_strtoull() is glib >= 2.12
            start = strtoull(temp1, NULL, 10);
            length = strtoull(temp2, NULL, 10);
#endif          
            if (length < 1000) continue;
          
            download_insert_done_segment(download, start, start+length);
        } 
        else 
        {
            char* filename    = arg(line, 2);
            char* size        = arg(NULL, 2);
            char* dirname     = arg(NULL, 2);  // ignored
            char* user        = arg(NULL, 2);  // <= v1.0.1
            char* org_file    = arg(NULL, 2);  // <= v1.0.1
            char* search      = arg(NULL, 2);
            char* last_access = arg(NULL, 2);  // ingored
            char* status      = arg(NULL, 2);  // ignored
          
            if (!search) continue;
          
            if (!status) 
            {
                // backward compatible to 1.0.1
                // user and org_file tag removed
                status = search;
                last_access = org_file;
                search = user;
            }
      
            if (download) _download_created(download);
            download = download_new(NULL);
          
            // set filename
            file_set_locname(FILE(download), g_strdup(filename));
            file_set_origname(FILE(download), g_strdup(filename));
            // set filesize
            FILE(download)->size = strtoll(size, NULL, 10);
            // set search keys
            download->search_keys = NULL;
            make_list_from_string(&download->search_keys, search, " ");
        }
    }
    if (download) _download_created(download);
}



static void download_delete(download_t* download)
{
    object_kill(OBJ(download));
}

static void action_source_remove(GList* sources)
{
    GList* dlist;
    for (dlist = sources; dlist; dlist = dlist->next)
    {
        source_t* source = dlist->data;
        socket_done(SOCK(source), ST_DELETE, "removed by user");
    }
}

static action_item_t* on_get_protocol_user_menu(GList* selection)
{
    GList* dlist;
    action_item_t* result = NULL;
    
    for (dlist = selection; dlist; dlist = dlist->next)
    {
        source_t* source = dlist->data;
        action_item_t* item = user_action_source_find(source->user->ID);
        if (!item) return NULL;
        if (!result) result = item;
        else if (result != item) return NULL;
    }
    return result;
}

void download_find_sources(download_t* download)
{
    g_assert(DownloadModule);
    if (!download_check_finish(download, FALSE))
        search_new_from_download(download);
}

static void on_download_show_details(GList* downloads)
{
    GList* dlist;
    for (dlist = downloads; dlist; dlist = dlist->next) 
    {
        download_t* download = dlist->data;
        UI_download_show_details(download);
    }
}

#define DOWNLOAD_LIST_ACTION(func) static void                \
    on_ ## func (GList* downloads) {                          \
        GList* dlist;                                         \
        for (dlist = downloads; dlist; dlist = dlist->next) { \
            download_t* download = dlist->data;               \
            func(download);                                   \
        }                                                     \
    }

DOWNLOAD_LIST_ACTION(download_delete);
DOWNLOAD_LIST_ACTION(download_find_sources);

static gboolean on_check_find(download_t* download)
{
    return (download && !download_check_finish(download, FALSE));
}

static void download_actions_init(void)
{
    action_item_t* item;
    
    item = action_item_root("Download actions");
    action_item_open_file(N_("Open file"), item, NULL);
    action_item_sep(item);
    action_item_simple(N_("Delete download"), item, 0,
                       (ActionItemFunc)on_download_delete);
    action_item_sep(item);
    action_item_new(N_("Find sources"), item, 0,
                    (ActionItemFunc)on_download_find_sources,
                    (ActionItemCheck)on_check_find);
    action_item_sep(item);
    action_item_simple(N_("Show details"), item, 0,
                       (ActionItemFunc)on_download_show_details);
  
    item = action_item_root("Source actions");
    action_item_simple(N_("Remove source"), item, 0,
                       (ActionItemFunc)action_source_remove);
    action_item_sep(item);
    action_item_sub(N_("User menu"), item,
                    (ActionItemLink)on_get_protocol_user_menu,
                    FALSE);
}

static void source_add_limit_real(source_t* source, limit_t* limit, int group)
{
    source->limits[group] = g_list_append(source->limits[group], limit);
    if (!object_monitor_find(OBJ(limit), (ObjectFunc)download_limit_changed, NULL))
        object_monitor_d(OBJ(limit), (ObjectFunc)download_limit_changed, NULL);
}

/** Add a limit for a source.
 * @param source the source
 * @param name the name of the limit
 * @param group when the limit should be decreased/increased. If the
 * group for example is SOURCE_REMOTE, then source is counted as soon
 * as it has at lease status SOURCE_REMOTE
 * @see limit_new()
 */
void source_add_limit(source_t* source, const char* name, int group)
{
    limit_t* limit = limit_find(name);
    if (!limit) 
    {
        g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
              "limit '%s' not found", name);
        return;
    }
    source_add_limit_real(source, limit, group);
}


static void add_progress(segment_t* segment, file_size_t* progress)
{
    *progress += segment->size;
}

file_size_t download_get_progress(download_t* download)
{
    GList* dlist;
    file_size_t done = 0;

    g_assert(DownloadModule);
    segments_foreach(download->segments_done, (SegmentsForeach)add_progress, &done);

    for (dlist = download->segments_used; dlist; dlist = dlist->next)
    {
        dl_segment_t* seg = dlist->data;
        segment_t* segment = seg->segment;
        if (segment->size > 0) done += segment->size;
    }
    return done;
}

static void source_limits_adjust(source_t* source, int status, int new_status)
{
    if (new_status > status) 
    {
        // new_status > status  ==>  increase limits
        while (status < new_status) 
        {
            status++;
            if (source->limits[status]) limits_increase(source->limits[status]);
        }
    } 
    else 
    {
        // new_status < status  ==>  decrease limits
        while (status > new_status) 
        {
            if (source->limits[status]) limits_decrease(source->limits[status]);
            status--;
        }
        DownloadModule->queue_check = TRUE;
    }
}

static void source_unschedule(source_t* source) 
{
    if (source->scheduled) 
    {
        DownloadModule->schedule = g_list_remove(DownloadModule->schedule, source);
        source->scheduled = FALSE;
    }
}

static void source_schedule(source_t* source) 
{
    if (!source->scheduled) 
    {
        DownloadModule->queue_check = TRUE;
        DownloadModule->schedule = g_list_append(DownloadModule->schedule, source);
        source->scheduled = TRUE;
    }
}


static void source_set_status(source_t* source, int status) 
{
    if (source->status == status) return;
    
    source_limits_adjust(source, source->status, status);
    
    // sort out
    if (source->status == SOURCE_REMOTE) 
    {
        if (source->timeout) lopster_timeout_remove(source->timeout);
        source->timeout = 0;
    } 
    else if (source->status == SOURCE_STARTED || source->status == SOURCE_LOADING) 
    {
        DownloadModule->started = g_list_remove(DownloadModule->started, source);
        socket_set_timeout(SOCK(source), 0);
    }
    
    g_assert(source->timeout == 0);
    
    // set status
    source->status = status;
    
  // sort in
    if (source->status == SOURCE_STARTED || source->status == SOURCE_LOADING) 
    {
        DownloadModule->started = g_list_append(DownloadModule->started, source);
        socket_set_timeout(SOCK(source), 90);
    }
    
    object_set_dirty(OBJ(source));
}

int source_get_progress(source_t* source) 
{
    segment_t* seg;
    
    g_assert(DownloadModule);
    if (!source || !source->cur_segment) return 0;
    seg = source->cur_segment;
    if (seg->start >= seg->stop) return 0;
    return seg->size*100/(seg->stop-seg->start);
}

/** Destroy the source. Called by derived destroy method.
 * @param source the source to destroy
 */
void source_destroy(source_t* source) 
{
    int i1;
    g_assert(DownloadModule);
    
    source->download->sources = g_list_remove(source->download->sources, source);
    source->download->no_source--;
    object_set_dirty(OBJ(source->download));
    
    source_set_status(source, SOURCE_UNKNOWN);
    source_unschedule(source);
    
    user_unref(source->user);
    
    object_kill(OBJ(source->file));
    for (i1 = 0; i1 < SOURCE_SIZE; i1++) g_list_free(source->limits[i1]);
    
    tcp_socket_destroy(TSOCK(source));
}

static gboolean source_retry(source_t* source) 
{
    if (source->timeout) lopster_timeout_remove(source->timeout);
    source->timeout = 0;
    source_schedule(source);
    return FALSE;
}

/** Call this function if download is not possible due to a full
 * remote queue.
 * @param source the source
 * @param t time in seconds when to restart the source.
 * @param pos remote queue position 
 */
void source_set_remote(source_t* source, unsigned t, unsigned pos) 
{
    g_assert(t > 0);
    g_assert(DownloadModule);
    
    source->remote = pos;
    
    // make sure the source is not scheduled
    source_unschedule(source);
    source_set_status(source, SOURCE_REMOTE);
    
    // set timer to re-schedule.
    if (!source->timeout && t)
        source->timeout = lopster_timeout_add(t*1000, (GSourceFunc)source_retry, source);
    
}

static void source_close_file(source_t* source) 
{
    if (source->fd) 
    {
        fclose(source->fd);
        source->fd = NULL;
    }
}

typedef struct 
{
    thread_t _thread;
    file_size_t size;
    char* src;
    char* dest;
    FILE* file1;
    FILE* file2;
} move_t;

static void move_destroy(move_t* move) 
{
    int status = thread_get_status(THREAD(move));
    char* temp;
    
    fclose(move->file1);
    fclose(move->file2);
    
    if (status == THREAD_COMPLETED) 
    {
        temp = locale_from_utf8(move->src);
    } 
    else 
    {
        temp = locale_from_utf8(move->dest);
    }
    unlink(temp);
    g_free(temp);
    
    g_free(move->dest);
    g_free(move->src);
    
    thread_destroy(THREAD(move));
}

static unsigned move_main(move_t* move) 
{
#define BLOCK 1024*64
    static char buffer[BLOCK];
    gboolean finished;
    file_size_t done = 0;
    
    // FIXME: make 64bit compatible
    finished = FALSE;
    while (!finished) 
    {
        unsigned ret1;
        
        if ((ret1 = fread(buffer, 1, BLOCK, move->file1)) != BLOCK) 
        {
            if (ferror(move->file1))
                return THREAD_FAILED;
            if (feof(move->file1)) finished = TRUE;
        }
        if (fwrite(buffer, 1, ret1, move->file2) != ret1) 
            return THREAD_FAILED;
        
        if (thread_interrupted(THREAD(move))) return THREAD_INTERRUPTED;
        
        done += ret1;
        thread_set_progress(THREAD(move), done*100/move->size);
        //    sleep(1);
    }
    
  return THREAD_COMPLETED;
}

static move_t* move_new(download_t* download, const char* destination) 
{
    move_t* move;
    FILE* file1;
    FILE* file2;
    char* src = locale_from_utf8(FILE(download)->localname);
    char* dest = locale_from_utf8(destination);
    
    if ((file1 = fopen(src, "rb")) == NULL)
    {
        g_warning("Unable to open file '%s' (%s) !\n", 
                  src, g_strerror(errno));
        g_free(src);
        g_free(dest);
        return NULL;
    }

    if ((file2 = fopen(dest, "wb")) == NULL)
    {
        fclose(file1);
        g_warning("Unable to create file '%s' (%s) !\n",
                  dest, g_strerror(errno));
        g_free(src);
        g_free(dest);
        return NULL;
    }
    
    file_set_locname(FILE(download), g_strdup(destination));
    file_set_origname(FILE(download), g_strdup(destination));
    
    move = g_malloc0(sizeof(*move));
    thread_init(THREAD(move), _("Moving finished download"), 200);
    OBJ(move)->destroy = (ObjectFunc)move_destroy;
    move->file1 = file1;
    move->file2 = file2;
    move->dest = g_strdup(destination);
    move->src = g_strdup(FILE(download)->localname);
    move->size = FILE(download)->size;
    
    thread_set_description
        (THREAD(move), l_printf(_("from %s\nto %s"), move->src, move->dest));
    
    return move;
}



static gboolean download_move_file(download_t* download) 
{
    const char* folder = mtype_get_folder(FILE(download)->media);
    char* dest = l_printf(SDIR("%s") "%s", folder, FILE(download)->filename);
    char* src = locale_from_utf8(FILE(download)->localname);
    char* temp;
    int ret;
    
    /* First try and link it to the new location */
    temp = locale_from_utf8(dest);
    ret = rename(src, temp);
    g_free(temp);
    g_free(src);
    if (ret == -1 && (errno == EXDEV || errno == EPERM)) 
    {
        move_t* move = move_new(download, dest);
        if (move) thread_start(THREAD(move), (ThreadMainFunc)move_main);
    } 
    else 
    {
        file_set_locname(FILE(download), g_strdup(dest));
        file_set_origname(FILE(download), g_strdup(dest));
    }
    return FALSE;
}

static void download_deactivate(download_t* download) 
{
    GList* dlist;
    dlist = download->sources;
    while (dlist) 
    {
        source_t* source = dlist->data;
        dlist = dlist->next;
        socket_done(SOCK(source), ST_DELETE, "finished");
    }
    g_assert(!download->sources);
    if (download->search) 
    {
        object_kill(OBJ(download->search));
        download->search = NULL;
    }
}


static gboolean download_rs(download_t* download) 
{
    download_deactivate(download);
    return FALSE;
}

static gboolean download_check_finish(download_t* download, gboolean move) 
{
    segment_t* segment;
    
    if (download->flags & DOWNLOAD_FINISHED) return TRUE;
    
    if ((segment = segments_get_first(download->segments_done)) &&
        segment->start == 0 && segment->stop >= FILE(download)->size) 
    {
        download->flags |= DOWNLOAD_FINISHED;
        if (move) 
        {
            lopster_idle_add((GSourceFunc)download_move_file, download);
            lopster_idle_add((GSourceFunc)download_rs, download);
        }
        return TRUE;
    }
    return FALSE;
}

/* static gboolean download_check_finish(download_t* download, gboolean move) { */

/*   if (download->flags & DOWNLOAD_FINISHED) return TRUE; */
/*   //FIXME: segments_cover() doesn't check for resume errors, right? */
/*   if (segments_cover(download->segments_done, 0, FILE(download)->size) == 1) { */
/*     download->flags |= DOWNLOAD_FINISHED; */
/*     if (move) { */
/*       lopster_idle_add((GSourceFunc)download_move_file, download); */
/*       lopster_idle_add((GSourceFunc)download_rs, download); */
/*     } */
/*     return TRUE; */
/*   } */
/*   return FALSE; */
/* } */

/** End the source. Called by derived end method.
 * @param source the source.
 * @param message the ending message.
 * @return TRUE
 */
gboolean source_end(source_t* source, const char* message) 
{
    gboolean check;
    
    g_assert(DownloadModule);
    //  source_set_remote(source, 10, 0);
    
    check = (source->used_segments != NULL);
    while (source->used_segments)
        source_remove_segment(source, source->used_segments->data);
    
    if (check) download_check_finish(source->download, TRUE);
    
    object_set_dirty(OBJ(source));
    object_set_dirty(OBJ(source->download));
    
    tcp_socket_end(TSOCK(source), message);
    
    return TRUE;
}

static void on_source_destroyed(source_t* source, user_t* user) 
{
    user->sources = g_list_remove(user->sources, source);
}

static void source_set_user(source_t* source, user_t* user) 
{
    const char* name = l_printf
        ("d:%s:%s:dlimit", plugin_get_short_name(plugin_get(user->ID)), user->print(user));
    limit_t* limit = limit_find(name);
    
    if (!limit) limit = limit_new(name, "/core/download/limit_per_user");
    source_add_limit_real(source, limit, SOURCE_REMOTE);
    
    source->user = user_ref(user);
    
    user->sources = g_list_append(user->sources, source);
    object_notify_o(OBJ(source), (ObjectFunc)on_source_destroyed, OBJ(user));
}

/** Initialize a source. Called by the init function of derived sources.
 * @param source the source
 * @param file the file of the source
 * @param user the user offering the file
 * @param overlap TRUE if segments should be overlapped for integrity, FALSE otherwise
 */
void source_init(source_t* source, file_t* file, user_t* user, gboolean overlap) 
{
    g_assert(DownloadModule);
    
    tcp_socket_init(TSOCK(source), user?OBJ(user)->name:"Source", 0);
    object_add_type(OBJ(source), "Source");
    OBJ(source)->destroy = (ObjectFunc)source_destroy;
    socket_set_end(SOCK(source), (SockEndFunc)source_end);
    
    source->file = file_clone(file, "file");
    source->download = NULL;
    source->timeout = 0;
    source->remote = 0;
    source->cur_segment = NULL;
    source->fd = NULL;
    source->overlap = overlap;
    source->status = SOURCE_UNKNOWN;
    source->scheduled = FALSE;
    source->limits[SOURCE_UNKNOWN] = NULL;
    source->limits[SOURCE_REMOTE]  = NULL;
    source->limits[SOURCE_STARTED] = NULL;
    source->limits[SOURCE_LOADING] = NULL;
    
    source_set_user(source, user);
    
    // add global limit
    source_add_limit(source, "Download Limit", SOURCE_LOADING);
    // add global bandwidth limit
    socket_add_volume(SOCK(source), "Download Bandwidth Limit");
    // register bandwidth load history
    socket_add_load(SOCK(source), "Transfer");
    // monitor the rate
    socket_add_default_rate(SOCK(source), DIR_IN);
    socket_add_rate(SOCK(source), DownloadModule->rate);
    
    source_schedule(source);
}

/*
static source_t* source_new(file_t* file) {
  source_t* source;

  source = g_malloc(sizeof(*source));
  source_init(source, file, -1, NULL);
  
  return source;
}
*/

static source_t*
download_find_source(download_t* download, const char* fname, user_t* user) 
{
    GList* dlist;
    
    for (dlist = download->sources; dlist; dlist = dlist->next) 
    {
        source_t* source = dlist->data;
        if (user->ip && source->user->ip) 
        {
            if (source->user->ip != user->ip) continue;
        }
        else 
        {
            if (source->user != user) continue;
        }
        if (fname && strcmp(source->file->filename, fname)) continue;
        return source;
    }
    return NULL;
}

static void download_add_source_real(download_t* download, source_t* source) 
{
    g_assert(DownloadModule);
    
    attr_container_adapt(ATTRC(download), ATTRC(source->file), "file");
    download->sources = g_list_append(download->sources, source);
    download->no_source++;
    source->download = download;
    socket_add_rate(SOCK(source), download->rate);
    
    object_make_public(OBJ(source), NULL);
    
    object_set_dirty(OBJ(download));
}

static void
download_insert_free_segment(download_t* download,
                             file_size_t start, file_size_t stop) 
{
    download->segments_free =
        segments_insert(download->segments_free, start, stop, 0, SEG_ERR_NONE);
}

static void
download_insert_done_segment_with_error(download_t* download, file_size_t start,
                                        file_size_t stop, unsigned err)
{
    download->segments_done =
        segments_insert(download->segments_done, start, stop, stop-start, err);
}

static void
download_insert_done_segment(download_t* download, file_size_t start, file_size_t stop)
{
    download_insert_done_segment_with_error(download, start, stop, SEG_ERR_NONE);
}

// NOTE: this function only allows to free the END of a used segment
/*
static void
download_cut_segment(download_t* download,
		     file_size_t start, file_size_t stop) {
  GList* dlist;

  for (dlist = download->segments_used; dlist; dlist = dlist->next) {
    dl_segment_t* dseg = dlist->data;
    segment_t* seg = dseg->segment;
    if (start >= seg->start && stop <= seg->stop) {
      g_assert(start > seg->start + seg->size);
      g_assert(stop == seg->stop);
      seg->stop = start;
      break;
    }
  }
  g_assert(dlist);
  download_insert_free_segment(download, start, stop);
}
*/
				  
static void _source_recalc_stop(source_t* source)
{
    segment_t* segment = source->cur_segment;
    
    if (!segment)
    {
        source->real_stop = 0;
    } 
    else if (source->overlap)
    {
        file_size_t rint;
        
        if (source->file->size - segment->stop < SEG_OVERLAP)
            rint = source->file->size - segment->stop;
        else
            rint = SEG_OVERLAP;
        //FIXME: get correct real segment stop.
        if (download_segment_is_done
            (source->download, segment->stop, segment->stop + rint))
        {
            source->real_stop = segment->stop + rint;
        } 
        else
        {
            source->real_stop = segment->stop;
        }
    } 
    else
    {
        source->real_stop = segment->stop;
    }
}

static void _segment_reset_stop(dl_segment_t* segment, file_size_t stop)
{
    segment_t* seg = segment->segment;
    source_t* source = segment->owner;
    
    g_log("DOWNLOAD", G_LOG_LEVEL_DEBUG,
          "changing segment stop to ("FF"-"FF") "FF,
          seg->start, seg->stop, stop);
    seg->stop = stop;
    
    if (source)
    {
        g_assert(source->cur_segment == seg);
        _source_recalc_stop(source);
        // next line should be done in seg_changed() func
        //FIXME:    socket_rate_set_todo(SOCK(source), DIR_IN, seg->stop-seg->start);
        if (source->seg_changed) source->seg_changed(source);
    }
    
}

static void _segment_reset_start(source_t* source, segment_t* segment)
{
    if (source->overlap)
    {
        file_size_t rint;
        
        // begin overlap
        if (segment->start < SEG_OVERLAP)
            rint = segment->start;
        else
            rint = SEG_OVERLAP;
        if (download_segment_is_done
            (source->download, segment->start - rint, segment->start))
        {
            segment->size = -rint;
        } 
        else
        {
            segment->size = 0;
        }
        
    } 
    else
    {
        segment->size = 0;
    }
}

static void source_remove_segment(source_t* source, segment_t* segment)
{
    if (!segment) return;
    if (source->cur_segment == segment)
    {
        source_close_file(source);
        source->cur_segment = NULL;
    }
    source->used_segments =
        g_list_remove(source->used_segments, segment);
    download_release_segment(source->download, source, segment);
    if (!source->used_segments) source->download->no_active--;
}

/** Use an attached segment as next segment for download.  The current
 * segment (if any) will be de-attached silently and can be used for
 * other sources again (if not downloaded already).
 * @param source the source
 * @return the net available segment, or NULL if non found
 * @see source_use_segment()
 */
segment_t* source_next_segment(source_t* source)
{
    g_assert(DownloadModule);
    
    source_remove_segment(source, source->cur_segment);
    g_assert(source->cur_segment == NULL);
    if (source->used_segments)
    {
        source->cur_segment = source->used_segments->data;
        /* FIXME:
           socket_rate_set_todo
           (SOCK(source), DIR_IN, 
           source->cur_segment->stop - source->cur_segment->start);
        */
        _segment_reset_start(source, source->cur_segment);
        _source_recalc_stop(source);
        
        return source->cur_segment;
    } 
    else
    {
        return NULL;
    }
}

static void source_add_segment(source_t* source, segment_t* segment)
{
    if (!source->used_segments) source->download->no_active++;
    source->used_segments =
        g_list_append(source->used_segments, segment);
    
    //  if (!source->cur_segment) source_next_segment(source);
}

/*
static void _download_segments_print(download_t* download) {
  GList* dlist;

  printf("DONE\n");
  segments_foreach(download->segments_done, (SegmentsForeach)segment_print, NULL);
  
  printf("USED\n");
  for (dlist = download->segments_used; dlist; dlist = dlist->next) {
    dl_segment_t* segment = dlist->data;
    segment_print(segment->segment);
  }

  printf("FREE\n");
  segments_foreach(download->segments_free, (SegmentsForeach)segment_print, NULL);
}
*/

/** Attach a segment for this source. 
 * @param source the source
 * @param start the start of the segment
 * @param stop the end of the segment
 * @param split_allowed whether this segment can be split to get a
 * segment for other sources.
 */
void source_use_segment(source_t* source, file_size_t start,
                        file_size_t stop, gboolean split_allowed)
{
    GList* dlist;
    download_t* download = source->download;
    segment_t* segment;
    dl_segment_t* dseg;
    
    g_assert(DownloadModule);
    
    g_log("DOWNLOAD", G_LOG_LEVEL_DEBUG,
          "trying to use segment "FF"-"FF, start, stop);
    //  _download_segments_print(download);
    
    // safety check whether the segment intersects a finished segment
    if (segments_intersect(download->segments_done, start, stop))
    {
        g_assert(FALSE);
        return;
    }

    // check whether the segment starts in a used segment
    for (dlist = download->segments_used; dlist; dlist = dlist->next)
    {
        dl_segment_t* segment = dlist->data;
        segment_t* seg = segment->segment;
        
        if (start >= seg->start && start < seg->stop)
        {
            // error if segment cant be splitted
            if (!segment->split_allowed) return;
            // error if start inside already downloaded block
            if (start < seg->start+seg->size) return;
            download_insert_free_segment(download, start, seg->stop);
            _segment_reset_stop(segment, start);
            source_use_segment(source, start, stop, split_allowed);
            return;
        }
    }

    segment = segments_cut(download->segments_free, start, stop);
    if (!segment) return;
    
    dseg = g_malloc0(sizeof(*dseg));
    dseg->segment = segment;
    dseg->owner = source;
    dseg->split_allowed = split_allowed;
    download->segments_used =
        g_list_append(download->segments_used, dseg);
    g_log("DOWNLOAD", G_LOG_LEVEL_DEBUG,
          "using segment "FF"-"FF, segment->start, segment->stop);
    source_add_segment(source, segment);
    
}

/** Find a segment and attach it to the source.  This is basically a
 * wrapper for source_request_segment(), source_use_segment() and
 * source_next_segment()
 * @param source the source
 * @return the next segment, or NULL if non available
 */
segment_t* source_attach_segment(source_t* source)
{
    segment_t* segment = source_request_segment(source, 0);
    if (!segment) return NULL;
    source_use_segment(source, segment->start, segment->stop, TRUE);
    return source_next_segment(source);
}

static void find_free_segment(segment_t* seg, segment_t** result)
{
    if (*result == NULL || (*result)->stop-(*result)->start > seg->stop-seg->start)
        *result = seg;
}

/** Request a segment from the file which can be used for download.
 * The result is a static variable. The segment will not be reserved
 * in any way at this point. Its only to get a potential segment.  If
 * you like to use it for download, use the function
 * source_use_segment() afterwards.
 * @param source the download
 * @param min the minimum size of the requested segment
 * @return pointer to a static segment
 */
segment_t* source_request_segment(source_t* source, file_size_t min)
{
    static segment_t result;
    GList* dlist;
    segment_t* temp;
    download_t* download = source->download;
    
    g_assert(DownloadModule);
    // FIXME: implement better algorithm
    
    // use free segment if there is any.
    temp = NULL;
    segments_foreach(download->segments_free, (SegmentsForeach)find_free_segment, &temp);
    if (temp)
    {
        memcpy(&result, temp, sizeof(result));
        _segment_reset_start(source, &result);
        return &result;
    }
    
    temp = NULL;
    
    if (min < 128*1024) min = 128*1024;  // we cut at least 128k
    min = min*3/2;   // we cut only 2/3 from the rest of the segment
    
    // try to cut from used segment; use the biggest
    for (dlist = download->segments_used; dlist; dlist = dlist->next)
    {
        dl_segment_t* segment = dlist->data;
        segment_t* seg = segment->segment;
        if (!segment->split_allowed) continue;
        if (seg->stop - seg->start - seg->size <= min) continue;
        min = seg->stop - seg->start - seg->size;
        temp = seg;
    }
    
    if (!temp) return NULL;

    // we cut 2/3 of the remaining segment
    result.start = (temp->start*2 + temp->size*2 + temp->stop)/3;
    result.stop = temp->stop;
    result.size = 0;
    
    _segment_reset_start(source, &result);
    return &result;
}

static void find_free_segment_at(segment_t* seg, segment_t** result)
{
    if ((*result)->size == 0)
    {
        if ((*result)->start >= seg->start && (*result)->start < seg->stop)
            *result = seg;
    }
}

/** Request a segment from the file at a certain position
 * The result is a static variable. The segment will not be reserved
 * in any way at this point. Its only to get a potential segment.  If
 * you like to use it for download, use the function
 * source_use_segment() afterwards.
 * @param source the download
 * @param offset the offset where the segment shall start
 * @return pointer to a static segment
 */
segment_t* source_request_segment_at(source_t* source, file_size_t offset)
{
    static segment_t result;
    GList* dlist;
    segment_t* temp;
    download_t* download = source->download;
    
    g_assert(DownloadModule);
    
    // check if it's still free
    temp = &result;
    result.start = result.stop = offset;
    result.size = 0;
    segments_foreach(download->segments_free, (SegmentsForeach)find_free_segment_at, &temp);
    if (temp != &result)
    {
        memcpy(&result, temp, sizeof(result));
        _segment_reset_start(source, &result);
        return &result;
    }
    
    temp = NULL;
    
    // try to cut from used segment
    for (dlist = download->segments_used; dlist; dlist = dlist->next)
    {
        dl_segment_t* segment = dlist->data;
        segment_t* seg = segment->segment;

        if (offset >= seg->start && offset < seg->stop)
        {
            if (!segment->split_allowed) return NULL;
            if (seg->start+seg->size + 1024 >= offset) return NULL;

            result.start = offset;
            result.stop = seg->stop;
            result.size = 0;
            _segment_reset_start(source, &result);
            return &result;
        }
    }
    
    return NULL;
}

static void
download_release_segment(download_t* download, source_t* source, segment_t* segment)
{
    GList* dlist;
    
    for (dlist = download->segments_used; dlist; dlist = dlist->next)
    {
        dl_segment_t* dseg = dlist->data;
        segment_t* seg = dseg->segment;
        if (seg == segment)
        {
            g_assert(source == dseg->owner);
            
            if (seg->size < 0) 
                seg->size = 0;
            else if (seg->start + seg->size > seg->stop)
                seg->size = seg->stop - seg->start;
            
            download_insert_done_segment(download, seg->start, seg->start+seg->size);
            download_insert_free_segment(download, seg->start+seg->size, seg->stop);
            g_free(dseg);
            segment_destroy(seg);
            download->segments_used = g_list_delete_link(download->segments_used, dlist);
            downloads_save();
            return;
        }
    }
    g_assert(FALSE);
    segment_destroy(segment);
}

static gboolean source_allowed(source_t* source)
{
    int i1;

/*   g_log("DEBUG", G_LOG_LEVEL_DEBUG, */
/*         "source allowed? %d %s", source->status, OBJ(source)->name); */

    // check download limits FIXME: not functional, cause limits get not adjusted
    if (limits_reached(source->download->limits))
    {
        return FALSE;
    }

    for (i1 = 0; i1 < SOURCE_SIZE; i1++)
    {
        if (!source->limits[i1]) continue;
        // only limits > source->status get increased
/*     g_log("DEBUG", G_LOG_LEVEL_DEBUG, "-check limits %d", i1); */
        if (i1 > source->status)
        {
            // check whether limit is hit
            if (limits_reached(source->limits[i1])) return FALSE;
        }
        else
        {
            GList* dlist;
            // check whether limit is even *exceeded*, hit is not enough
            for (dlist = source->limits[i1]; dlist; dlist = dlist->next)
            {
                limit_t* limit = dlist->data;
                if (limit_get_used(limit) > limit_get_limit(limit))
                {
/*           g_log("LIMIT", G_LOG_LEVEL_DEBUG, "limit '%s' exceeded %d/%d", */
/*                 OBJ(limit)->name, limit->used, limit->limit); */
                    return FALSE;
/*         } else { */
/*           g_log("LIMIT", G_LOG_LEVEL_DEBUG, "limit '%s' not exceeded %d/%d", */
/*                 OBJ(limit)->name, limit->used, limit->limit); */
                }
            }
        }
    }
    return TRUE;
}

static gboolean source_start(source_t* source)
{
    
    source_unschedule(source);
    source_set_status(source, SOURCE_STARTED);
    
    //    UI.source_show(source);
    
    if (!source->start(source))
    {
        socket_done(SOCK(source), ST_ERROR, "start error");
        return FALSE;
    }
    else
    {
        return TRUE;
    }
}

static gboolean downloads_check()
{
    static unsigned calls = 0;
    GList* dlist;
    
    g_assert(DownloadModule);
    
    calls++;
    if ((calls % 60) == 0) downloads_save();
    
    if (!DownloadModule->queue_check || !DownloadModule->schedule) return TRUE;
    
    for (dlist = DownloadModule->schedule; dlist; )
    {
        source_t* source = dlist->data;
        dlist = dlist->next;
        if (source_allowed(source) && source_start(source)) return TRUE;
    }
    // we didnt start an download, disable check.
    DownloadModule->queue_check = FALSE;
    return TRUE;
}

gboolean
download_segment_is_done(download_t* download, file_size_t start, file_size_t stop)
{
    GList* dlist;
    
    if (stop <= start) return TRUE;
    
    // FIXME: calculate exactly
    if (segments_cover(download->segments_done, start, stop) != -1) 
        return TRUE;
    
    for (dlist = download->segments_used; dlist; dlist = dlist->next)
    {
        dl_segment_t* seg = dlist->data;
        segment_t* segment = seg->segment;
        file_size_t seg_stop;
        
        // TEST: assume some bytes download if this is a freshly used segment
        if (segment->size == 0) seg_stop = segment->start + SEG_OVERLAP;
        else seg_stop = segment->start+segment->size;
        
        if (start >= segment->start && stop <= seg_stop) return TRUE;
    }
    return FALSE;
}

static gboolean source_open_file(source_t* source)
{
    char* filename;
    segment_t* segment;
    
    g_assert(!source->fd);
    
    segment = source->cur_segment;
    if (!segment)
    {
        g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
              "could not open file (no segment)");
        g_assert(FALSE);
        return FALSE;
    }
    
    filename = auto_text(locale_from_utf8(FILE(source->download)->localname));
    g_log("DOWNLOAD", G_LOG_LEVEL_DEBUG, "opening file '%s'", filename);

    source->fd = fopen(filename, "r+b");
    if (!source->fd) {
        source->fd = fopen(filename, "w+b");
    }

    /* FIXME: do more checks to avoid side effects.
       if (source->fd == -1 && source->download->no_active == 1) {
       // incomplete folder changed? fallback..
       download_update_filename(source->download);
       filename = text_from_utf8(FILE(source->download)->filename);
       source->fd = open(filename, O_RDWR|O_CREAT|O_BINARY,
       S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
       }
    */

    if (!source->fd)
    {
        g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
              "cout not open '%s' for download", FILE(source->download)->localname);
        return FALSE;
    }
    else if (FSEEK(source->fd, segment->start+segment->size, SEEK_SET) != 0)
    {
        g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
              "error when seeking to "FF, segment->start+segment->size);
        source_close_file(source);
        return FALSE;
    }
    return TRUE;
}

static const char* source_get_input(source_t* source)
{
    buffer_t* buffer = b_socket_get_in_buffer(BSOCK(source));
    segment_t* segment = source->cur_segment;
    int avail;
    int avail2;
    
    if (!source->fd && !source_open_file(source))
        return "could not open file for download";
    avail = buffer_avail(buffer);
    if (!avail) return NULL;
    
    // make sure we dont read to much from the buffer
    if (avail > source->real_stop - segment->start - segment->size)
        avail = source->real_stop - segment->start - segment->size;
    
    // THIS IS RESUME START CHECK
    if (segment->size < 0)
    {
        char* buf;
        int res;
        
        if (-segment->size > avail) avail2 = avail;
        else avail2 = -segment->size;
        g_log("DOWNLOAD", G_LOG_LEVEL_DEBUG,
              "begin resume check: checking %u bytes", avail2);
        // load data for resume check.
        buf = g_malloc(avail2);
        
        if ((res = fread(buf, 1, avail2, source->fd)) != avail2)
        {
            g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
                  "could not read end of segment for resume check (%d/%d bytes)", res, avail2);
            g_free(buf);
            //FIXME: we should mark the previous segment.
            return "download file read error";
        }
        
        // now do the resume check of n2 bytes data
        if (memcmp(buf, buffer_read_pos(buffer), avail2) != 0)
        {
            g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
                  "resume check of %d bytes failed", avail2);
            print_raw("FILE", buf, avail2);
            print_raw("NET", buffer_read_pos(buffer), avail2);
            // FIXME: mark previous segment
            g_free(buf);
            return "begin resume check failed";
        }
        g_free(buf);
        
        // ok, resume check passed, now decrease the number of bytes
        // that are left for writing..(Note: avail2 is always <= avail)
        avail -= avail2;
        segment->size += avail2;
        buffer_consume_virt(buffer, avail2);
        
        // return if no data left...
        if (avail == 0) return NULL;
    }

    // ok, avail bytes left to write from buffer position avail2, but
    // at most to position segment->stop (we might have received
    // the data to real_stop)
    if (segment->start+segment->size+avail > segment->stop)
    {
        avail2 = segment->stop - segment->start - segment->size;
        // if we already wrote the whole segment we dont write any data
        if (avail2 < 0) avail2 = 0;
    }
    else
    {
        // write all the data.
        avail2 = avail;
    }
  
    // now writing avail2 bytes...
    if (avail2 > 0)
    {
#ifdef WIN32
        // Win32 has a bug ? with files opened with r+ or w+, after doing some read or write
        // before writing again , position on the file should be set again (shame)
        if ( FSEEK(source->fd, ftell( source->fd ), SEEK_SET) != 0 )
        {
            g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
                  "could not set position to the file '%s'", source->file->filename);
            return "download file position to write error";
        }
#endif
        
        if ( fwrite(buffer_read_pos(buffer), 1, avail2, source->fd) != (size_t)avail2)
        {
            g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
                  "could not write %d bytes to file '%s'", avail2,
                  source->file->filename);
            return "download file write error";
        }
        segment->size += avail2;
        avail -= avail2;
        buffer_consume(buffer, avail2);
    }
    
    // 'avail' bytes left for resume check at segment end
    if (avail > 0)
    {
        char* buf;
        // avail should now
        g_log("DOWNLOAD", G_LOG_LEVEL_DEBUG,
              "end resume check: checking %d bytes", avail);
        // load data for resume check.
        buf = g_malloc(avail);
        if (fread(buf, 1, avail, source->fd) != (size_t)avail)
        {
            g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
                  "could not read start of segment (%d bytes)", avail);
            g_free(buf);
            segment->error |= SEG_ERR_NEXT;
            return "download file read error";
        }
        
        // now do the resume check of n bytes data
        if (download_segment_is_done(source->download, segment->stop, source->real_stop))
        {
            if (memcmp(buf, buffer_read_pos(buffer), avail) != 0)
            {
                g_log("DOWNLOAD", G_LOG_LEVEL_WARNING,
                      "resume end check of %d bytes failed", avail);
                g_free(buf);
                segment->error |= SEG_ERR_NEXT;
                return "end resume check failed";
            }
        }
        else
        {
            g_log("DOWNLOAD", G_LOG_LEVEL_DEBUG,
                  "resume end check of %d bytes skipped", avail);
        }
        
        g_free(buf);
        segment->size += avail;
        buffer_consume(buffer, avail);
    }

    // finish the segment
    //  if (segment->start+segment->size >= segment->stop) {
    if (segment->start+segment->size >= source->real_stop)
    {
        socket_set_status(SOCK(source), ST_DL_SEG_FIN);
        return "segment finished";
    }
    return NULL;
}


/** Start the download. There must be a current segment and the source
 * must be connected to the peer.  To set the current segment, use
 * source_use_segment() to add segments to the source, and
 * source_next_segment() to set one of the added segments as current
 * segment.
 * @param source the source
 * @return NULL on success, a message on error
 */
const char* source_start_downloading(source_t* source)
{
    if (!source_open_file(source))
        return "could not open file for download";
    
    g_assert(DownloadModule);
    
    // make sure this source is not scheduled anymore.
    source_unschedule(source);
    source_set_status(source, SOURCE_LOADING);
    
    b_socket_set_consume(BSOCK(source),(BConsumeInputFunc)source_get_input);
    b_socket_start_receiving(BSOCK(source), 102400);
    
    return NULL;
}

/** Find a source of a user.
 * @param user the user of the source
 * @param winname the windows filename of the file to look for.
 * @return the source of found, or NULL
 */
source_t* source_find(user_t* user, const char* winname)
{
    GList* dlist;
    for (dlist = user->sources; dlist; dlist = dlist->next)
    {
        source_t* source = dlist->data;
        g_assert(source->user == user);
        if (!strcmp(source->file->netname, winname)) return source;
    }
    return NULL;
}

/** Register a function that creates a source from a given user and file.
 * @param ID the ID of the plugin
 * @param create the function that creates the source.
 */
void source_factory_register(int ID, FileToSourceFunc create)
{
    g_assert(DownloadModule);
    if (ID < 0 || ID >= plugin_get_number()) return;
    DownloadModule->source_funcs[ID] = create;
}

FileToSourceFunc source_factory_get(int ID)
{
    g_assert(DownloadModule);
    if (ID < 0 || ID >= plugin_get_number()) return NULL;
    return DownloadModule->source_funcs[ID];
}

void download_clear_finished()
{
    GList* dlist;
    
    g_assert(DownloadModule);
    dlist = DownloadModule->downloads;
    while (dlist)
    {
        download_t* download = dlist->data;
        dlist = dlist->next;
        if (download->flags & DOWNLOAD_FINISHED)
            object_kill(OBJ(download));
    }
}

/** Add a new source for a download. The source is generated from
 * the given file and user.
 * @param download the download
 * @param file the file the user offers
 * @param user the user offering the file
 * @return a new source, or an duplicate existing source, or NULL on error.
 */
source_t* download_add_source(download_t* download, file_t* file, user_t* user)
{
    source_t* source;
    FileToSourceFunc func;
    
    g_assert(DownloadModule);
    func = source_factory_get(user->ID);
    if (!func) return NULL;
    
    source = download_find_source(download, file->filename, user);
    if (!source)
    {
        source = func(download, file, user);
        if (source) download_add_source_real(download, source);
        DownloadModule->queue_check = TRUE;
/*   } else { */
/*     char* ip1 = ntoa(user->ip); */
/*     char* ip2 = ntoa(source->user->ip); */
/*     g_message("  source already found for %s (%s): %s (%s)",  */
/*               OBJ(user)->name, ip1, OBJ(source->user)->name, ip2); */
    }
    return source;
}

/**
 * Check whether a file matches a download. Currently the size and
 * the search_keys of the download are compared.
 * @param download the download.
 * @param file the file to check
 * @return TRUE if file matches, FALSE otherwise
 */
gboolean download_check_file(download_t* download, file_t* file)
{
    GSList* slist;
    
    if (FILE(download)->size != file->size) return FALSE;
    
    for (slist = download->search_keys; slist; slist = slist->next)
    {
        char* token = slist->data;
        if (!strcasestr(file->localname, token)) return FALSE;
    }
    return TRUE;
}

/** Find a download.
 * @param localname the local filename of the download.
 * @param size the size of the file.
 * @return the download, or NULL
 */
download_t* download_find(const char* localname, file_size_t size)
{
    GList* dlist;
    
    for (dlist = DownloadModule->downloads; dlist; dlist = dlist->next)
    {
        download_t* download = dlist->data;
        file_t* file = dlist->data;
        if (file->size == size && !strcmp(file->localname, localname))
            return download;
    }
    return NULL;
}

void downloads_foreach(ObjectFunc func, void* data)
{
    g_assert(func);
    g_assert(DownloadModule);
    g_list_foreach(DownloadModule->downloads, (GFunc)func, data);
}

rate_t* downloads_get_rate()
{
    g_assert(DownloadModule);
    return DownloadModule->rate;
}

static void downloads_init()
{
    lopster_xml_register_home("core:downloads", "downloads.xml",
                              "Incomplete", xml_get_downloads, xml_save_downloads);
    
    pref_register_int   ("/core/download/limit", TRUE, 10);
    pref_add_constraint_min("/core/download/limit", 0);
    pref_register_int   ("/core/download/limit_per_user", TRUE, 1);
    pref_add_constraint_min("/core/download/limit", 0);
    pref_register_int   ("/core/download/limit_per_file", TRUE, 10);
    pref_add_constraint_min("/core/download/limit_per_file", 0);
    pref_register_int   ("/core/download/bandlimit", TRUE, 1024*200);
    pref_add_constraint_minmax("/core/download/bandlimit", 0, 1<<30);
    pref_register_string("/core/path/incomplete", TRUE, NULL);
    
    DownloadModule->rate = rate_new(0);
    DownloadModule->source_funcs =
        g_malloc0(sizeof(FileToSourceFunc) * plugin_get_number());
    DownloadModule->match_funcs =
        g_malloc0(sizeof(DownloadMatchFunc) * plugin_get_number());
    
    download_actions_init();
}

static void downloads_run()
{
    // register limits
    limit_new("Download Limit", "/core/download/limit");
    limit_new("Download File Limit", "/core/download/limit_per_file");
    
    // register bandwidth limits
    volume_new("Download Bandwidth Limit", "/core/download/bandlimit");
    
    downloads_load();
    
    g_list_foreach(DownloadModule->downloads, (GFunc)download_find_sources, NULL);
    
    lopster_call_periodic(1000, (CallFunc)downloads_check, NULL);
}

static void downloads_exit()
{
    g_free(DownloadModule->source_funcs);
    g_free(DownloadModule->match_funcs);
    DownloadModule = NULL;
}

static void download_module_init(download_module_t* module)
{
    module_init(MODULE(module), "Download Module",
                (ModuleFunc)downloads_init,
                (ModuleFunc)downloads_run,
                (ModuleFunc)downloads_exit);
    
    module->downloads = NULL;
    module->schedule = NULL;
    module->started = NULL;
    module->queue_check = TRUE;
    module->ignore_save = FALSE;
    module->rate = NULL;
    module->source_funcs = NULL;
    module->match_funcs = NULL;
    
    module_require(pref_module_instance);
    module_require(search_module_instance);
    module_require(lsocket_module_instance);
    module_require(tsocket_module_instance);
    module_require(limit_module_instance);
    module_require(base_module_instance);
}

/** Require this module if you support downloads.
 * @return the download module
 */
module_t* download_module_instance()
{
    if (!DownloadModule)
    {
        DownloadModule = g_malloc0(sizeof(*DownloadModule));
        download_module_init(DownloadModule);
    }
    return MODULE(DownloadModule);
}

#ifdef WIN32
#define _DOWNLOAD_C_
#include "lcorealiasdef.c"
#endif
