/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.13 $
 * $Date: 2006/07/31 19:13:22 $
 * $Author: sgop $
 */

#ifndef _DOWNLOAD_H_
#define _DOWNLOAD_H_

/** \defgroup DownloadModule Downloads in lopster
 * @ingroup FileSharingFramework
 */
/*@{*/

typedef struct _source_t source_t;
typedef struct _download_t download_t;

#include "files.h"
#include "tcp_socket.h"
#include "file_segment.h"
#include "action_item.h"
#include "user.h"
#include "module.h"

#include "search.h"

G_BEGIN_DECLS

#define RESUME_ERR_END   0x01
#define RESUME_ERR_BEG   0x02
#define RESUME_ERR_HASH  0x04

#define ST_DL_SEG_FIN     300
#define ST_SOURCE_REMOTE  301

/** Function to start a source.
 * @param source the source.
 * @return FALSE on error, TRUE otherwise
 */
typedef gboolean (*SourceStartFunc)(source_t* source);

/** Function that gets called if current segment has changed.
 * @param source the source.
 */
typedef void (*SourceSegChangedFunc)(source_t* source);

/** Create source from user and file */
typedef source_t* (*FileToSourceFunc)(const download_t* download, file_t* file, user_t*);

/** Function which checks whether a source matches a download.
 * @param download the download
 * @param file the source file
 * @param user the user sharing the file
 * @return TRUE if file matches, FALSE otherwise
 */
typedef gboolean (*DownloadMatchFunc)(download_t* download, file_t* file, user_t* user);


typedef struct {
    segment_t* segment;
    source_t* owner;
    gboolean split_allowed;
} dl_segment_t;

#define DOWNLOAD_FINISHED   0x01

struct _download_t {
    file_t _file;
    GSList* search_keys;
    search_t* search;
    GList* limits;
    int flags;
    int no_active;
    int no_source;

    GList* sources;
    GList* segments_used;  // of type dl_segment_t
    segments_t* segments_done;  // of type segments_t
    segments_t* segments_free;  // of type segments_t
    rate_t* rate;
};


#define SOURCE_UNKNOWN  0
#define SOURCE_REMOTE   1   // remotely queued
#define SOURCE_STARTED  2   // download requested
#define SOURCE_LOADING  3   // downloading
#define SOURCE_SIZE     4   // number of defines


#define SOURCE(arg) ((source_t*)(arg))
/** A source. */
struct _source_t {
    tcp_socket_t _tsock;     ///< the base class
    file_t* file;            ///< the file of that user
    user_t* user;            ///< the user
    download_t* download;    ///< the download this source belongs to
    GList* limits[SOURCE_SIZE];
    int status;
    gboolean scheduled;
    int timeout;
    int remote;
    GList* used_segments;
    segment_t* cur_segment;
    file_size_t real_stop;
    gboolean overlap;        ///< set by derived init(), default: TRUE
  
    FILE* fd;
    SourceStartFunc start;
    SourceSegChangedFunc seg_changed;
};


/** @name Downloads */
/*@{*/
module_t* download_module_instance();
source_t* download_add_source(download_t* download, file_t* file, user_t* user);
gboolean download_matches_source(download_t* download, file_t* file, user_t* user);
gboolean download_check_file(download_t* download, file_t* file);
void download_register_match(int ID, DownloadMatchFunc match);
DownloadMatchFunc download_match_get(int ID);
/*@}*/


/** @name Source functions */
/*@{*/
void source_init(source_t* source, file_t* file, user_t* user, gboolean overlap);
gboolean source_end(source_t* source, const char* message);
void source_destroy(source_t* source);
segment_t* source_request_segment(source_t* source, file_size_t min);
segment_t* source_request_segment_at(source_t* source, file_size_t offset);
void source_use_segment(source_t* source, file_size_t start, file_size_t stop, gboolean split_allowed);
segment_t* source_next_segment(source_t* source);
segment_t* source_attach_segment(source_t* source);
const char* source_start_downloading(source_t* source);
void source_set_remote(source_t* source, unsigned t, unsigned pos);
source_t* source_find(user_t* user, const char* winname);
void source_add_limit(source_t* source, const char* name, int group);
void source_factory_register(int ID, FileToSourceFunc create);
/*@}*/


/*@}*/

FileToSourceFunc source_factory_get(int ID);
void download_import_from_lopster1(FILE* file);
download_t* download_from_file(file_t* file);
void download_find_sources(download_t* download);
download_t* download_find(const char* localname, file_size_t size);
file_size_t download_get_progress(download_t* download);
int download_get_rate(download_t* download);
rate_t* downloads_get_rate();
gboolean download_segment_is_done(download_t* download, file_size_t start, file_size_t stop);
void download_clear_finished(void);
void downloads_foreach(ObjectFunc func, void* data);
int source_get_progress(source_t* source);

G_END_DECLS
#endif
