/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.8 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include <errno.h>
#ifndef WIN32
#include <unistd.h>
#endif
#ifdef WIN32
#include <io.h> 
#endif
#include <string.h>
#include <fcntl.h>

#include "exec.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

#define MAX_PROCESS       5

static GList* Exec = NULL;
static GList* ExecRunning = NULL;
static GList* ExecQueue = NULL;
static unsigned NumberRunning = 0;

static void exec_fire_up(exec_t* exec);

static int exec_find_free_id() {
  static int ID = 0;
  return ID++;
}

static void exec_destroy(exec_t* exec) {
  Exec = g_list_remove(Exec, exec);
  if (exec->argv) g_strfreev(exec->argv);
  if (exec->buffer) buffer_destroy(exec->buffer);
  exec->buffer = NULL;
  g_free(exec);
}

void exec_set_output_func(exec_t* exec, ExecReadFunc func, void* data) {
  exec->read = func;
  exec->read_data = data;
  if (func) {
    if (!exec->buffer) exec->buffer = buffer_new(1024*10);
  } else {
    if (exec->buffer) buffer_destroy(exec->buffer);
    exec->buffer = NULL;
  }
}

static void exec_try_dequeue() {
  exec_t* exec;

  if (!ExecQueue) return;
  if (NumberRunning >= MAX_PROCESS) return;

  exec = ExecQueue->data;
  ExecQueue = g_list_delete_link(ExecQueue, ExecQueue);
  exec->status = EXEC_INACTIVE;  // NOTE: check with exec_set_status()
  exec_fire_up(exec);
}

static void exec_set_status(exec_t* exec, unsigned status) {
  if (exec->status == status) return;

  if (exec->status == EXEC_RUNNING) {
    ExecRunning = g_list_remove(ExecRunning, exec);
    NumberRunning--;
    //    exec_try_dequeue();
  } else if (exec->status == EXEC_QUEUED) {
    ExecQueue = g_list_remove(ExecQueue, exec);
  }

  exec->status = status;

  if (exec->status == EXEC_RUNNING) {
    ExecRunning = g_list_append(ExecRunning, exec);
    NumberRunning++;
  } else if (exec->status == EXEC_QUEUED) {
    ExecQueue = g_list_append(ExecQueue, exec);
  }
}

static void exec_end(exec_t * exec, int status) {
  exec_set_status(exec, status);

  if (exec->pid) g_spawn_close_pid(exec->pid);
  if (exec->read_input) g_source_remove(exec->read_input);
  if (exec->end) exec->end(exec, exec->end_data);

  exec_destroy(exec);

  exec_try_dequeue();
}

static gboolean exec_read(exec_t* exec) {
  char* pos = buffer_read_pos(exec->buffer);
  char* nl = strchr(pos, '\n');

  while (nl) {
    *nl = 0;
    g_log("EXEC", G_LOG_LEVEL_DEBUG,
          "%03d: %s", exec->internal_id, pos);
    buffer_consume_virt(exec->buffer, nl-pos+1);
    pos = nl+1;
    nl = strchr(pos, '\n');
  }
  buffer_consume(exec->buffer, 0);
  return TRUE;
}
static void exec_init(exec_t* exec, char** argv) {
  exec->status = EXEC_INACTIVE;
  exec->argv = argv;
  exec->pid = 0;
  exec->read = NULL;
  exec->read_data = NULL;
  exec->end = NULL;
  exec->end_data = NULL;
  exec->internal_id = exec_find_free_id();
  exec->buffer = NULL;
  exec_set_output_func(exec, (ExecReadFunc)exec_read, NULL);
}

exec_t* exec_new(char** argv) {
  exec_t* exec = g_malloc0(sizeof(*exec));
  exec_init(exec, argv);
  Exec = g_list_append(Exec, exec);
  return exec;
}
  
static void exec_queue(exec_t* exec) {
  exec_set_status(exec, EXEC_QUEUED);
  g_log("EXEC", G_LOG_LEVEL_DEBUG,
        "queued process %d", exec->internal_id);
}

static gboolean exec_output_real(exec_t* exec) {
  buffer_t* buffer = exec->buffer;
  int res = read(exec->read_fd, buffer_write_pos(buffer), buffer_space(buffer)-1);

  switch (res) {
  case -1:
    g_log("EXEC", G_LOG_LEVEL_WARNING,
          "could not read from process %d %d:(%s)",
          exec->internal_id, errno, g_strerror(errno));
    return FALSE;
  case 0:
    g_source_remove(exec->read_input);
    exec->read_input = 0;
    return FALSE;
  }
  
  buffer->size += res;
  buffer->data[buffer->size] = 0;
  if (!exec->read(exec, exec->read_data)) return FALSE;

  return TRUE;
}

static void on_exec_exit(GPid pid, int status, exec_t* exec) {
  g_assert(exec && exec->pid == pid);
  g_log("EXEC", G_LOG_LEVEL_DEBUG,
        "process %d exited with status %d", exec->internal_id, status);

  exec_end(exec, status==0?EXEC_SUCCESS:EXEC_ABORTED);
}

static gboolean
on_exec_output(GIOChannel* source, GIOCondition condition, exec_t* exec) {
  (void)source;
  if (condition & G_IO_HUP) return FALSE;
  if (condition & G_IO_ERR) return FALSE;
  
  return exec_output_real(exec);
}

static void on_exec_func_destroy(exec_t* exec) {

  // flush input to output
  if (exec->read_input != 0) {
/*     fcntl(exec->read_fd, F_SETFL, O_NONBLOCK); */
    exec_output_real(exec);
    exec->read_input = 0;
  }

}

static void exec_fire_up(exec_t* exec) {
  GError* error = NULL;

  g_assert(exec && exec->argv && exec->argv[0]);
  g_assert(exec->status == EXEC_INACTIVE);
  
  if (g_spawn_async_with_pipes
      (NULL, exec->argv, NULL,
       G_SPAWN_SEARCH_PATH|G_SPAWN_DO_NOT_REAP_CHILD|G_SPAWN_STDERR_TO_DEV_NULL,
       NULL, NULL, &exec->pid,
       NULL, exec->read?(&exec->read_fd):NULL, NULL, &error)) {

    exec_set_status(exec, EXEC_RUNNING);
    g_log("EXEC", G_LOG_LEVEL_DEBUG, "started process %d", exec->internal_id);

    g_child_watch_add(exec->pid, (GChildWatchFunc)on_exec_exit, exec);
    if (exec->read) {
      GIOChannel* channel = g_io_channel_unix_new(exec->read_fd);
      exec->read_input = 
        g_io_add_watch_full(channel, G_PRIORITY_DEFAULT,
                            (G_IO_IN|G_IO_HUP|G_IO_ERR),
                            (GIOFunc)on_exec_output, exec,
                            (GDestroyNotify)on_exec_func_destroy);
      g_io_channel_unref(channel);
    }
  } else {
    g_log("EXEC", G_LOG_LEVEL_WARNING, "an error occurred: %s",
          error?error->message:"failed to spawn process");
    exec_end(exec, EXEC_FAILED);
  }

  if (error) g_error_free(error);
}

void exec_start(exec_t* exec, ExecEndFunc end, void* data) {
  if (exec->status != EXEC_INACTIVE) return;

  g_log("EXEC", G_LOG_LEVEL_DEBUG,
        "trying to start process %d", exec->internal_id);

  exec->end = end;
  exec->end_data = data;

  if (NumberRunning >= MAX_PROCESS) {
    exec_queue(exec);
  } else {
    exec_fire_up(exec);
  }
}

#ifdef WIN32
# define _EXEC_C_
# include "lcorealiasdef.c"
#endif
