/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.6 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include "file_segment.h"
#include "utils.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

struct _segments_t {
  GList* segments;
  gboolean auto_mode;
};

static void
segment_init(segment_t* segment, file_size_t start,
             file_size_t stop, file_size_t size, unsigned error) {
  segment->start = start;
  segment->stop = stop;
  segment->size = size;
  if (size > 0) segment->error = error;
}

segment_t* segment_new(file_size_t start, file_size_t stop,
                       file_size_t size, unsigned error) {
  segment_t* segment;
  
  segment = g_malloc0(sizeof(*segment));
  segment_init(segment, start, stop, size, error);
  return segment;
}

void segment_destroy(segment_t* segment) {
  if (segment) g_free(segment);
}

void segment_print(segment_t* segment) {
  if (segment)
    g_log("CORE", G_LOG_LEVEL_DEBUG, "  "FF" + "FF" = "FF,
	  segment->start, segment->size, segment->stop);
}

static void segments_init(segments_t* segments) {
  segments->segments = NULL;
  segments->auto_mode = TRUE;
}
segments_t* segments_new() {
  segments_t* segments = g_malloc0(sizeof(*segments));
  segments_init(segments);
  return segments;
}
void segments_destroy(segments_t* segments) {
  if (!segments) return;
  g_list_foreach(segments->segments, (GFunc)segment_destroy, NULL);
  g_list_free(segments->segments);
  g_free(segments);
}
void segments_foreach(segments_t* segments, SegmentsForeach func,
		      void* data) {
  GList* dlist;
  if (!segments) return;
  for (dlist = segments->segments; dlist; dlist = dlist->next) {
    segment_t* segment = dlist->data;
    func(segment, data);
  }
}

void segments_prepend(segments_t* segments, segment_t* segment) {
  if (!segment) return;
  g_assert(segments);
  segments->segments = g_list_prepend(segments->segments, segment);
  segments->auto_mode = FALSE;
}

void segments_append(segments_t* segments, segment_t* segment) {
  if (!segment) return;
  g_assert(segments);
  segments->segments = g_list_append(segments->segments, segment);
  segments->auto_mode = FALSE;
}

segments_t* segments_insert(segments_t* segments, file_size_t start,
                            file_size_t stop, file_size_t size, unsigned error) {
  GList* ndlist = NULL;
  GList* pdlist = NULL;
  unsigned eprev = (error & SEG_ERR_PREV);
  unsigned enext = (error & SEG_ERR_NEXT);
  gboolean ehash = (error & SEG_ERR_HASH);
  gboolean ehash2;
  segment_t* nseg = NULL;
  segment_t* pseg = NULL;
  gboolean mprev = FALSE;
  gboolean mnext = FALSE;
  
  //FIXME: do security check for segment intersection

  if (!segments) segments = segments_new();
  if (start >= stop) return segments;
  g_assert(size >= 0);
  
  if (!segments->auto_mode) {
    segments_append(segments, segment_new(start, stop, size, error));
    return segments;
  }

  for (ndlist = segments->segments; ndlist; ndlist = ndlist->next) {
    segment_t* seg = ndlist->data;
    if (seg->start >= stop) break;
    pdlist = ndlist;
  }
  
  if (pdlist) {
    pseg = pdlist->data;
    ehash2 = (pseg->error & SEG_ERR_HASH);
    mprev = TRUE;
    if (pseg->stop != start) mprev = FALSE;
    else if (eprev) mprev = FALSE;
    else if (ehash2 && !ehash) mprev = FALSE;
    else if (!ehash2 && ehash) mprev = FALSE;
    else if (size > 0 && pseg->size != pseg->stop-pseg->start) mprev = FALSE;
  }
  if (ndlist) {
    nseg = ndlist->data;
    ehash2 = (nseg->error & SEG_ERR_HASH);
    mnext = TRUE;
    if (nseg->start != stop) mnext = FALSE;
    else if (enext) mnext = FALSE;
    else if (ehash2 && !ehash) mnext = FALSE;
    else if (!ehash2 && ehash) mnext = FALSE;
    else if (nseg->size > 0 && size != stop-start) mnext = FALSE;
  }

  if (!mprev && !mnext) {
    // dont merge
    segments->segments =
      g_list_insert_before(segments->segments, ndlist,
                           segment_new(start, stop, size, error));
  } else if (!mprev) {
    // merge with next
    nseg->start = start;
    nseg->size += size;
    nseg->error &= ~SEG_ERR_PREV;
    nseg->error |= eprev;
  } else if (!mnext) {
    // merge with prev
    pseg->stop = stop;
    pseg->size += size;
    pseg->error &= ~SEG_ERR_NEXT;
    pseg->error |= enext;
  } else {
    // merge with both
    pseg->stop = nseg->stop;
    pseg->size += nseg->size;
    pseg->size += size;
    pseg->error &= ~SEG_ERR_NEXT;
    pseg->error |= (nseg->error & SEG_ERR_NEXT);
    segments->segments = g_list_delete_link(segments->segments, ndlist);
    segment_destroy(nseg);
  }    
  return segments;
}

segments_t* segments_create_complement(segments_t* segments, file_size_t size) {
  file_size_t last = 0;
  GList* dlist;
  segments_t* result = segments_new();

  for (dlist = segments?segments->segments:NULL; dlist; dlist = dlist->next) {
    segment_t* segment = dlist->data;
    if (segment->start > last)
      segments_insert(result, last, segment->start, 0, SEG_ERR_NONE);
    last = segment->stop;
  }
  if (last < size) 
    segments_insert(result, last, size, 0, SEG_ERR_NONE);
  return result;
}

int segments_cover(segments_t* segments, file_size_t start, file_size_t stop) {
  GList* dlist;
  unsigned result = 0;
  
  if (start >= stop) return 0;  // covered by 0
  if (!segments) return -1;     // not covered

  // noone needs a general implementation on unsorted segments,
  // so do a the fast specialized implementation only
  if (!segments->auto_mode) {
    g_log("CORE", G_LOG_LEVEL_WARNING,
          "cover() not supported without auto_mode");
    return -1;
  }

  for (dlist = segments->segments; dlist; dlist = dlist->next) {
    segment_t* segment = dlist->data;
    if (segment->stop <= start) continue;  // no intersection
    if (segment->start > start) return -1; // start is not covered.
    if (segment->start+segment->size <= start) return -1; // stop is not covered
    result++;
    if (segment->start+segment->size >= stop) return result; // fully convered
    if (segment->start+segment->size != segment->stop) return -1; // stop cannot be not covered by next segment
    start = segment->stop;  // adjust start
  }
  return -1;
}

gboolean segments_intersect(segments_t* segments, file_size_t start, file_size_t stop) {
  GList* dlist;

  if (!segments) return FALSE;

  for (dlist = segments->segments; dlist; dlist = dlist->next) {
    segment_t* seg = dlist->data;
    if (start < seg->start && stop > seg->stop) return TRUE;
    if (start >= seg->start && start < seg->stop) return TRUE;
    if (stop > seg->start && stop <= seg->stop) return TRUE;
  }
  return FALSE;
}

unsigned segments_count(segments_t* segments) {
  if (!segments) return 0;
  return g_list_length(segments->segments);
}

static segment_t*
segments_cut_from(segments_t* segments, GList* dlist,
                  file_size_t start, file_size_t stop) {
  segment_t* seg = dlist->data;
  file_size_t pos = seg->start+seg->size;
  file_size_t size;
  segment_t* nseg = NULL;
  
  if (start > seg->start && stop < seg->stop) {
    size = (pos>stop)?(pos-stop):0;
    segments->segments =
      g_list_insert_before
      (segments->segments, dlist->next,
       segment_new(stop, seg->stop, size, 
		   seg->error&(SEG_ERR_HASH|SEG_ERR_NEXT)));
    
    // cut the original segment
    seg->stop = start;
    seg->error &= ~SEG_ERR_NEXT;
    if (pos > seg->stop) seg->size = seg->stop-seg->start;
    
    // and create new one
    size =
      (pos>stop)?(stop-start):
      (pos>start)?(pos-start):0;
    nseg = segment_new(start, stop, size, seg->error&SEG_ERR_HASH);
  } else if (start == seg->start && stop == seg->stop) {
    // we simply use the original segment
    segments->segments =
      g_list_delete_link(segments->segments, dlist);
    nseg = seg;
  } else if (start == seg->start) {
    // cut from segment head
    size = (pos>stop)?(stop-start):seg->size;
    nseg = segment_new(start, stop, size,
		       seg->error&(SEG_ERR_HASH|SEG_ERR_PREV));
    seg->start = stop;
    seg->error &= ~SEG_ERR_PREV;
    seg->size -= size;
    if (seg->size == 0) seg->error &= ~SEG_ERR_HASH;
  } else {
    // cut from segment tail
    size = (pos>start)?(pos-start):0;
    nseg = segment_new(start, stop, size,
		       seg->error&(SEG_ERR_HASH|SEG_ERR_NEXT));
    seg->stop = start;
    seg->error &= ~SEG_ERR_NEXT;
    seg->size -= size;
  }
  return nseg;
}
		  
segment_t* segments_cut(segments_t* segments, file_size_t start, file_size_t stop) {
  GList* dlist;
  segment_t* seg = NULL;

  if (!segments) return NULL;
  if (stop <= start) return NULL;

  if (segments->auto_mode) {
    for (dlist = segments->segments; dlist; dlist = dlist->next) {
      seg = dlist->data;
      if (seg->stop <= start) continue;
      if (start < seg->start || stop > seg->stop) return NULL;
      break;
    }
  } else {
    for (dlist = segments->segments; dlist; dlist = dlist->next) {
      seg = dlist->data;
      if (start >= seg->start && stop <= seg->stop) break;
    }
  }
  if (dlist)
    return (segments_cut_from(segments, dlist, start, stop));
  else
    return NULL;
}

segment_t* segments_get_first(segments_t* segments) {
  if (!segments || !segments->segments) return NULL;
  return segments->segments->data;
}

GList* segments_get_list(segments_t* segments) {
  return (segments?segments->segments:NULL);
}

#ifdef WIN32
# define _FILE_SEGMENT_C_
# include "lcorealiasdef.c"
#endif
