/* Copyright (C) 2005-2006 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.13 $
 * $Date: 2008/12/23 13:38:14 $
 * $Author: sgop $
 */

#include <string.h>
#include <stdio.h>

#include <glib.h>

#include "files.h"
#include "file_tree.h"
#include "mtypes.h"
#include "ui.h"
#include "l_timer.h"

#ifdef WIN32
#include "lcorealias.h"
#endif



/** Destroy a node. Call from derived destroy method only.
 * @param node the node to destroy.
 */
void file_node_destroy(file_node_t* node) {
  file_node_t* child;
  file_node_t* next;
  
  g_assert(node);
  
  if (node->file) object_kill(OBJ(node->file));

  child = node->child;
  while (child) {
    next = child->next;
    object_kill(OBJ(child));
    child = next;
  }
  node->child = NULL;
  
  object_destroy(OBJ(node));
}

/** Initializes a file node. Call from derived init method only.
 * @param node the node
 * @param name the name of the node, usually path or filename
 */
void file_node_init(file_node_t* node, const char* name) {
  int i1, i2;

  g_assert(node);
  object_init(OBJ(node), name, TRUE);
  object_add_type(OBJ(node), "FileNode");
  OBJ(node)->destroy = (ObjectFunc)file_node_destroy;
  node->parent = NULL;
  node->child = NULL;
  node->next = NULL;
  node->prev = NULL;
  node->file = NULL;

  for (i1 = 0; i1 <= MEDIA_ANY; i1++) {
    for (i2 = 0; i2 <= 3; i2++) {
      node->fci[i1][i2] = 0;
      node->fsi[i1][i2] = 0;
    }
  }
  node->flags = 0;
}

static file_node_t* file_node_new(const char* name) {
  file_node_t* node = g_malloc0(sizeof(*node));
  file_node_init(node, name);
  return node;
}

/** Get the complete path of the node.
 * @param node the file node
 * @param[out] size the resulting size of the path plus the initial value
 * @return a dynamically path, free with g_free()
 * @note the path always has a dir separator appended.
 */
char* file_node_get_folder(file_node_t* node, int* size) {
  char* result;
  int len;
  int msize = size?*size:0;

  g_assert(node);
  // add space
  len = strlen(OBJ(node)->name);
  msize += len;
  if (len > 0 && OBJ(node)->name[len-1] != DIR_SEP)
    msize++;

  if (!node->parent) {
    // no more parent, we allocate the string here
    // one byte for the \0
    result = g_malloc(sizeof(char)*(msize+1));
    result[msize] = 0;
    // thats the return point, now size is the index
    msize = 0;
  } else {
    result = file_node_get_folder(node->parent, &msize);
  }
  
  strcpy(result+msize, OBJ(node)->name);
  msize += len;
  if (len > 0 && OBJ(node)->name[len-1] != DIR_SEP)
    result[msize++] = DIR_SEP;
  if (size) *size = msize;
  return result;
}

static file_node_t* file_node_search(file_node_t* node, const char* name) {
  while (node) {
    if (!strcmp(OBJ(node)->name, name)) return node;
    node = node->next;
  }
  return 0;
}

static void file_node_created(file_node_t* node) {

  g_assert(node);

  if (node->file) {
    file_t* file = node->file;
    int mt = file->media;
    unsigned group = (node->flags & FN_FLAG_SHARED)?FN_GROUP_SHARED:FN_GROUP_UNSHARED;

    // now set counter
    node->fci[MEDIA_ANY][FN_GROUP_ALL]++;
    node->fsi[MEDIA_ANY][FN_GROUP_ALL] += file->size;
    node->fci[mt][FN_GROUP_ALL]++;
    node->fsi[mt][FN_GROUP_ALL] += file->size;
    
    node->fci[MEDIA_ANY][group]++;
    node->fsi[MEDIA_ANY][group] += file->size;
    node->fci[mt][group]++;
    node->fsi[mt][group] += file->size;

    if (node->flags & FN_FLAG_RECENT) {
      node->fci[MEDIA_ANY][FN_GROUP_RECENT]++;
      node->fsi[MEDIA_ANY][FN_GROUP_RECENT] += file->size;
      node->fci[mt][FN_GROUP_RECENT]++;
      node->fsi[mt][FN_GROUP_RECENT] += file->size;
    }
  } else {
    node->fci[MEDIA_FOLDER][FN_GROUP_ALL]++;
  }
}






static void file_tree_clear_files(file_tree_t *tree) {
  int i1, i2;
  file_node_t* node;
  file_node_t* next;
  
  if (!tree) return;

  // updating user stats
  for (i1 = 0; i1 <= MEDIA_ANY; i1++) {
    for (i2 = 0; i2 < FN_GROUP_SIZE; i2++) {
      if (i1 != MEDIA_ANY && i2 != FN_GROUP_ALL) {
        tree->fci[i1][i2] = 0;
        tree->fsi[i1][i2] = 0;
      }
    }
  }

  node = tree->files;
  while (node) {
    next = node->next;
    object_kill(OBJ(node));
    node = next;
  }
  tree->files = NULL;
}

/** Destroy a file tree. Call from derived destroy function only.
 * @param tree the file tree
 */
void file_tree_destroy(file_tree_t* tree)
{
    if (!tree) return;
    g_free(tree->description);
    if (tree->mark_timeout) lopster_timeout_remove(tree->mark_timeout);
    
    file_tree_clear_files(tree);
    
    object_destroy(OBJ(tree));
}


static gboolean file_tree_node_recent(file_tree_t* tree, file_node_t* node) {
  (void)tree;
  (void)node;
  return FALSE;
}

/** Initialize a file tree. Call from derived init method only.
 * @param tree the file tree
 * @param name the name of the tree
 * @param des the description of the tree
 */
void file_tree_init(file_tree_t* tree, const char* name, const char* des) {
  int i1, i2;

  object_init(OBJ(tree), name, FALSE);
  object_add_type(OBJ(tree), "FileTree");
  OBJ(tree)->destroy = (ObjectFunc)file_tree_destroy;

  tree->description = des?g_strdup(des):NULL;
  tree->last_refreshed = 0;

  for (i1 = 0; i1 <= MEDIA_ANY; i1++) {
    for (i2 = 0; i2 <= 3; i2++) {
      tree->fci[i1][i2] = 0;
      tree->fsi[i1][i2] = 0;
    }
  }

  tree->files = NULL;
  tree->show_media = MEDIA_ANY;
  tree->show_time = FN_GROUP_ALL;
  tree->mark_timeout = 0;

  tree->node_new = file_node_new;
  tree->node_recent = file_tree_node_recent;
}

/** Notify that the filenode has changed.
 * @param tree the tree
 * @param node the node that has changed.
 */
void file_node_update(file_tree_t* tree, file_node_t* node) {
  g_assert(tree && node);
  node->flags |= FN_FLAG_UPDATE;
  object_set_dirty(OBJ(tree));
}

static void file_tree_node_inserted(file_tree_t* tree, file_node_t* node) {
  file_node_t* parent;
  int i1, i2;

  g_assert(tree && node);

  parent = node->parent;
  while (parent) {
    file_node_update(tree, parent);
    for (i1 = 0; i1 <= MEDIA_ANY; i1++) {
      for (i2 = 0; i2 < FN_GROUP_SIZE; i2++) {
        parent->fci[i1][i2] += node->fci[i1][i2];
        parent->fsi[i1][i2] += node->fsi[i1][i2];
      }
    }
    parent = parent->parent;
  }
  file_node_update(tree, node);
  for (i1 = 0; i1 <= MEDIA_ANY; i1++) {
    for (i2 = 0; i2 < FN_GROUP_SIZE; i2++) {
      tree->fci[i1][i2] += node->fci[i1][i2];
      tree->fsi[i1][i2] += node->fsi[i1][i2];
    }
  }
}

static void file_tree_insert_node(file_tree_t* tree, file_node_t* parent, file_node_t* node)
{
    g_assert(node);
  
    // insert the node
    if (!parent)
    {
        node->next = tree->files;
        if (node->next) node->next->prev = node;
        node->prev = NULL;
        node->parent = NULL;
        tree->files = node;
    }
    else
    {
        node->next = parent->child;
        node->prev = NULL;
        if (node->next) node->next->prev = node;
        parent->child = node;
        node->parent = parent;
    }
    
    file_tree_node_inserted(tree, node);
}

/** Insert a file directly in a parent.  Normally
 * file_tree_insert_file() is used to insert files.  But for
 * efficiency reasons this function is used directly when loading
 * file_trees from disc
 * @param tree the file tree
 * @param parent the place where to insert
 * @param file the file to insert.
 * @param flags the flags of the new file node
 * @param check_dup TRUE to check for duplucates, FALSE otherwise
 * @return a new filenode, or the duplicate
 */
file_node_t* file_tree_insert_file_in(file_tree_t* tree, file_node_t* parent,
                                      file_t* file, int flags, gboolean check_dup)
{
    file_node_t* child;
    
    if (check_dup)
    {
        child = file_node_search(parent?parent->child:tree->files, file->filename);
        if (child)
        {
            object_kill(OBJ(file));
            return child;
        }
    }
    
    child = tree->node_new(file->filename);
    child->file = file;
    child->flags = flags;
    
    file_node_created(child);
    file_tree_insert_node(tree, parent, child);
    
    return child;
}

/** Call a function on every node of a (sub)tree. The traversal is
 * aborted if the callback returns not NULL.
 * @param tree the file tree.
 * @param node the root of the subtree, or NULL for the whole tree
 * @param action the callback
 * @param data the user data for the callback
 * @return the result of the callback (if trasversal was aborted), or NULL
 */
file_node_t* file_tree_action(file_tree_t* tree, file_node_t* node, 
                              FileNodeTraverse action, void* data) {
  file_node_t* temp;
  file_node_t* temp2;
  file_node_t* result;
  
  if (!tree) return NULL;
  if (node) result = action(tree, node, data);
  else result = NULL;
  if (result) return result;

  if (node) temp = node->child;
  else temp = tree->files;

  while (temp) {
    temp2 = temp->next;
    result = file_tree_action(tree, temp, action, data);
    if (result) return result;
    temp = temp2;
  }
  return NULL;
}

/** Call a function for every file node. The traversal is aborted if a
 * callback returns not NULL.
 * @param tree the file tree
 * @param func the callback
 * @param data the user data for the callback
 * @note this call is equivalent to file_tree_action(tree, NULL, func, data);
 */
void file_tree_foreach(file_tree_t* tree, FileNodeTraverse func, void* data) {
  file_tree_action(tree, NULL, func, data);
}

/** Notify the tree that a file was shared.
 * @param tree the file tree
 * @param node the node that was shared.
 * @note make sure you only share files that are not shared already
 */
static void file_tree_node_shared(file_tree_t* tree, file_node_t* node) {
  file_node_t* parent;
  int mt;

  g_assert(tree && node && node->file);

  mt = node->file->media;

  parent = node;
  while (parent) {
    file_node_update(tree, parent);
    parent->fci[mt][FN_GROUP_UNSHARED]--;
    parent->fci[mt][FN_GROUP_SHARED]++;
    parent->fsi[mt][FN_GROUP_UNSHARED] -= node->file->size;
    parent->fsi[mt][FN_GROUP_SHARED] += node->file->size;
    parent->fci[MEDIA_ANY][FN_GROUP_UNSHARED]--;
    parent->fci[MEDIA_ANY][FN_GROUP_SHARED]++;
    parent->fsi[MEDIA_ANY][FN_GROUP_UNSHARED] -= node->file->size;
    parent->fsi[MEDIA_ANY][FN_GROUP_SHARED] += node->file->size;
    parent = parent->parent;
  }
  tree->fci[mt][FN_GROUP_UNSHARED]--;
  tree->fci[mt][FN_GROUP_SHARED]++;
  tree->fsi[mt][FN_GROUP_UNSHARED] -= node->file->size;
  tree->fsi[mt][FN_GROUP_SHARED] += node->file->size;
  tree->fci[MEDIA_ANY][FN_GROUP_UNSHARED]--;
  tree->fci[MEDIA_ANY][FN_GROUP_SHARED]++;
  tree->fsi[MEDIA_ANY][FN_GROUP_UNSHARED] -= node->file->size;
  tree->fsi[MEDIA_ANY][FN_GROUP_SHARED] += node->file->size;

}

/** Notify the tree that a file was unshared.
 * @param tree the file tree
 * @param node the node that was shared.
 * @note make sure you only unshare files that are shared
 */
static void file_tree_node_unshared(file_tree_t* tree, file_node_t* node) {
  file_node_t* parent;
  int mt;

  if (!tree || !node || !node->file) return;

  mt = node->file->media;

  parent = node;
  while (parent) {
    file_node_update(tree, parent);
    parent->fci[mt][FN_GROUP_UNSHARED]++;
    parent->fci[mt][FN_GROUP_SHARED]--;
    parent->fsi[mt][FN_GROUP_UNSHARED] += node->file->size;
    parent->fsi[mt][FN_GROUP_SHARED] -= node->file->size;
    parent->fci[MEDIA_ANY][FN_GROUP_UNSHARED]++;
    parent->fci[MEDIA_ANY][FN_GROUP_SHARED]--;
    parent->fsi[MEDIA_ANY][FN_GROUP_UNSHARED] += node->file->size;
    parent->fsi[MEDIA_ANY][FN_GROUP_SHARED] -= node->file->size;
    parent = parent->parent;
  }
  tree->fci[mt][FN_GROUP_UNSHARED]++;
  tree->fci[mt][FN_GROUP_SHARED]--;
  tree->fsi[mt][FN_GROUP_UNSHARED] += node->file->size;
  tree->fsi[mt][FN_GROUP_SHARED] -= node->file->size;
  tree->fci[MEDIA_ANY][FN_GROUP_UNSHARED]++;
  tree->fci[MEDIA_ANY][FN_GROUP_SHARED]--;
  tree->fsi[MEDIA_ANY][FN_GROUP_UNSHARED] += node->file->size;
  tree->fsi[MEDIA_ANY][FN_GROUP_SHARED] -= node->file->size;
}

static void file_tree_node_marked(file_tree_t* tree, file_node_t* node) {
  file_node_t* parent;
  file_t* file;
  int mt;

  g_assert(tree && node && node->file);
  
  file = node->file;
  mt = file->media;

  parent = node;
  while (parent) {
    file_node_update(tree, parent);
    parent->fci[MEDIA_ANY][FN_GROUP_RECENT]++;
    parent->fsi[MEDIA_ANY][FN_GROUP_RECENT] += file->size;
    parent->fci[mt][FN_GROUP_RECENT]++;
    parent->fsi[mt][FN_GROUP_RECENT] += file->size;
    parent = parent->parent;
  }

  tree->fci[MEDIA_ANY][FN_GROUP_RECENT]++;
  tree->fsi[MEDIA_ANY][FN_GROUP_RECENT] += file->size;
  tree->fci[mt][FN_GROUP_RECENT]++;
  tree->fsi[mt][FN_GROUP_RECENT] += file->size;
}

static void file_tree_node_unmarked(file_tree_t* tree, file_node_t* node) {
  file_node_t* parent;
  file_t* file;
  int mt;

  g_assert(tree && node && node->file);
  
  file = node->file;
  mt = file->media;

  parent = node;
  while (parent) {
    file_node_update(tree, parent);
    parent->fci[MEDIA_ANY][FN_GROUP_RECENT]--;
    parent->fsi[MEDIA_ANY][FN_GROUP_RECENT] -= file->size;
    parent->fci[mt][FN_GROUP_RECENT]--;
    parent->fsi[mt][FN_GROUP_RECENT] -= file->size;
    parent = parent->parent;
  }

  tree->fci[MEDIA_ANY][FN_GROUP_RECENT]--;
  tree->fsi[MEDIA_ANY][FN_GROUP_RECENT] -= file->size;
  tree->fci[mt][FN_GROUP_RECENT]--;
  tree->fsi[mt][FN_GROUP_RECENT] -= file->size;
}

static void
file_tree_mark_files_rec(file_tree_t* tree, file_node_t* node, int* timeout) {
  int temp;

  if (node && node->file) {
    temp = tree->node_recent(tree, node);
    if (temp > 0 &&  (*timeout == 0 || temp < *timeout))
      *timeout = temp;
    if (temp && (node->flags & FN_FLAG_RECENT) == 0) {
      node->flags |= FN_FLAG_RECENT;
      file_tree_node_marked(tree, node);
    } else if (!temp && (node->flags & FN_FLAG_RECENT)) {
      node->flags &= ~FN_FLAG_RECENT;
      file_tree_node_unmarked(tree, node);
    }
  } else {
    if (node) node = node->child;
    else node = tree->files;
    while (node) {
      file_tree_mark_files_rec(tree, node, timeout);
      node = node->next;
    }
  }
}

static int file_tree_mark_files(gpointer data) {
  file_tree_t* tree = data;
  int new_timeout = 0;

  if (!tree || !tree->files) return 0;
  
  g_log("FILETREE", G_LOG_LEVEL_DEBUG, "(%s) marking now", OBJ(tree)->name);
  file_tree_mark_files_rec(tree, NULL, &new_timeout);
  
  if (new_timeout > 0) file_tree_mark(tree, new_timeout); 

  return 0;
}

/** Mark recent nodes in the tree.
 * @param tree the tree.
 * @param when seconds when to really mark
 */
void file_tree_mark(file_tree_t* tree, int when) {
  if (!tree) return;

  if (tree->mark_timeout) {
    lopster_timeout_remove(tree->mark_timeout);
    tree->mark_timeout = 0;
  }

  if (when < 0) when = 0;
  if (when > 3600) when = 3600;
  if (when > 0) {
    tree->mark_timeout = lopster_timeout_add(when*1000, file_tree_mark_files, tree);
  } else {
    file_tree_mark_files(tree);
  }
}

/** Insert a folder in the tree.
 * @param tree the filetree
 * @param parent where to insert.
 * @param folder the folder to insert
 * @return the new node
 */
file_node_t* file_tree_insert_folder(file_tree_t* tree, file_node_t* parent, char* folder) {
  file_node_t* node;
  char* pos;

  g_assert(folder && *folder);

  pos = strchr(folder, DIR_SEP);

  if (pos) *pos = 0;
  node = file_node_search(parent?parent->child:tree->files, (*folder)?folder:"/");
  if (!node) {
    node = tree->node_new((*folder)?folder:"/");
    file_node_created(node);
    file_tree_insert_node(tree, parent, node);
  }
  if (pos) folder = pos+1;
  else folder = NULL;

  if (folder && *folder != 0)
    return file_tree_insert_folder(tree, node, folder);
  else return node;
}

static file_node_t*
file_tree_insert_file_real(file_tree_t* tree, file_t* file, int flags, int inh) {
  file_node_t* node;
  char save;
  char* temp;

  if (file->localname >= file->filename) {
    g_log("FILETREE", G_LOG_LEVEL_WARNING,
          "tree_insert_file(): should never happen");
    return NULL;
  }

  // now insert the file;
  // cut the filename (file->filename is linked to file->localname!)
  save = *(file->filename);
  *(file->filename) = 0;
  temp = g_strdup(file->localname);
  *(file->filename) = save;
  node = file_tree_insert_folder(tree, NULL, temp);
  g_free(temp);

  // append the filename again.
  if (inh) {
    if (node->fci[MEDIA_ANY][FN_GROUP_SHARED] > 0 || 
        node->fci[MEDIA_ANY][FN_GROUP_ALL] == 0)
      flags |= FN_FLAG_SHARED;
    else
      flags &= ~FN_FLAG_SHARED;
  }
  node = file_tree_insert_file_in(tree, node, file, flags, TRUE);
  
  return node;
}

/** Insert a file in the tree.
 * @param tree the tree
 * @param file the file to insert
 * @param flags the flags of the resulting node
 * @return the new filenode
 */
file_node_t* file_tree_insert_file(file_tree_t* tree, file_t* file, int flags) {
  return file_tree_insert_file_real(tree, file, flags, 0);
}

/** Insert a file in the tree. The difference to
 * file_tree_insert_file() is that the share status of the parent node
 * is inherited (the flag FN_SHARED)
 * @param tree the tree
 * @param file the file to insert
 * @param flags the flags of the resulting node
 * @return the new filenode
 */
file_node_t* file_tree_insert_file_inherited(file_tree_t* tree, file_t* file, int flags) {
  return file_tree_insert_file_real(tree, file, flags, 1);
}

static file_node_t* on_share(file_tree_t* tree, file_node_t* node, int* data) {
  if (node->file && !(node->flags & FN_FLAG_SHARED) && node->fci[data[0]][data[1]]) {
    node->flags |= FN_FLAG_SHARED;
    file_tree_node_shared(tree, node);
    if (tree->node_shared) tree->node_shared(node);
  }
  return NULL;
}

/** Share a subtree. 
 * @param tree the file tree
 * @param node the subtree to share, or NULL to share all
 * @param media which media to share
 * @param group which group to share
 */
void file_tree_share_node(file_tree_t* tree, file_node_t* node,
                          unsigned media, unsigned group) {
  int data[2] = {media, group};
  file_tree_action(tree, node, (FileNodeTraverse)on_share, data);
}

static file_node_t* on_unshare(file_tree_t* tree, file_node_t* node, int* data)
{
    if (node->file && (node->flags & FN_FLAG_SHARED) && node->fci[data[0]][data[1]])
    {
        node->flags &= ~FN_FLAG_SHARED;
        file_tree_node_unshared(tree, node);
        if (tree->node_unshared) tree->node_unshared(node);
    }
    return NULL;
}

/** Unshare a subtree. 
 * @param tree the file tree
 * @param node the subtree to unshare, or NULL to unshare all
 * @param media which media to share
 * @param group which group to share
 */
void file_tree_unshare_node(file_tree_t* tree, file_node_t* node,
                            unsigned media, unsigned group)
{
    int data[2] = {media, group};
    file_tree_action(tree, node, (FileNodeTraverse)on_unshare, data);
}

static void file_tree_node_removed(file_tree_t* tree, file_node_t* node)
{
    file_node_t* parent;
    int i1, i2;
    
    g_assert(tree && node);
    
    parent = node->parent;
    while (parent)
    {
        file_node_update(tree, parent);
        for (i1 = 0; i1 <= MEDIA_ANY; i1++)
        {
            for (i2 = 0; i2 <= 3; i2++)
            {
                parent->fci[i1][i2] -= node->fci[i1][i2];
                parent->fsi[i1][i2] -= node->fsi[i1][i2];
            }
        }
        parent = parent->parent;
    }
    for (i1 = 0; i1 <= MEDIA_ANY; i1++)
    {
        for (i2 = 0; i2 <= 3; i2++)
        {
            tree->fci[i1][i2] -= node->fci[i1][i2];
            tree->fsi[i1][i2] -= node->fsi[i1][i2];
        }
    }
}

/** Remove a node from the tree.
 * @param tree the file tree
 * @param node the node to remove
 */
void file_tree_remove_node(file_tree_t* tree, file_node_t* node)
{
    file_node_t* parent;
    
    g_assert(tree && node);
    
/*   g_log("DEBUG", G_LOG_LEVEL_DEBUG, "remove node %s", OBJ(node)->name); */
    
    if (node->fci[MEDIA_ANY][FN_GROUP_SHARED])
        file_tree_unshare_node(tree, node, MEDIA_ANY, FN_GROUP_ALL);
    
    file_tree_node_removed(tree, node);
    
    // now remove the node
    if (node->next) node->next->prev = node->prev;
    if (node->prev) node->prev->next = node->next;
    
    if (node->parent)
    {
        if (node->parent->child == node)
            node->parent->child = node->next;
        parent = node->parent;
    }
    else
    {
        tree->files = node->next;
        parent = NULL;
    }
    node->next = node->prev = node->parent = NULL;
    
    object_kill(OBJ(node));
    
    if (parent && !parent->child) file_tree_remove_node(tree, parent);
}


file_node_t* file_tree_search_filename(file_tree_t* tree, char* filename) {
  int len;
  int offset = 0;
  file_node_t* node = tree->files;

  // rec top down tree search
  while (node) {
    len = strlen(OBJ(node)->name);
    if (len && OBJ(node)->name[len-1] == DIR_SEP) len--; //special root case
    //    printf("compare %d bytes: '%s' '%s'\n",
    //	   len, node->name, filename+offset);
    if (len == 0 || 
        (!strncmp(OBJ(node)->name, filename+offset, len) &&
         (filename[offset+len]==DIR_SEP || filename[offset+len]==0))) {
      offset += len;
      if (filename[offset] == DIR_SEP) offset++; //special root case
      if (filename[offset] == 0) {      // filename end
        //	printf("succ\n");
        if (node->file) return node;
        else return NULL;
      } else if (!node->child) {        // not end, but no child
        //	printf("no child\n");
        return NULL;
      } else {                          // continue
        node = node->child;
      }
    } else {
      node = node->next;
    }
  }
  //  printf("no succ\n");
  return NULL;
}

#ifdef WIN32
#define _FILE_TREE_C_
#include "lcorealiasdef.c"
#endif
