/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.6 $
 * $Date: 2006/07/22 17:11:46 $
 * $Author: sid4u $
 */

#ifndef _FILE_TREE_H_
#define _FILE_TREE_H_

#include "mtypes.h"
#include "files.h"
#include "object.h"

G_BEGIN_DECLS

/** \defgroup FileTreeModule Building trees that contain files.
 *
 * Filetrees are trees that contain file nodes. File nodes either have
 * children or they own a file (see @ref FileModule). FileTrees may
 * contain nodes that are derived from ordinary filenodes.
 * 
 * Plugin developers normally don't need to access one of these
 * functions directly.
 *
 * @ingroup BasicFramework
 */
/*@{*/

/** @name FileNode flags. */
/*@{*/
#define FN_FLAG_SHARED 0x01  ///< the file is shared, only makes sense for nodes containing a file
#define FN_FLAG_UPDATE 0x02  ///< the node needs to be updates
#define FN_FLAG_RECENT 0x04  ///< the node contains a recent file
/*@}*/

/** @name FileNode groups. */
/*@{*/
#define FN_GROUP_UNSHARED 0  ///< unshared files
#define FN_GROUP_SHARED   1  ///< shared files
#define FN_GROUP_RECENT   2  ///< recent files.
#define FN_GROUP_ALL      3  ///< all files
#define FN_GROUP_SIZE     4  ///< the number of groups
/*@}*/

typedef struct _file_node_t file_node_t;
typedef struct _file_tree_t file_tree_t;

typedef file_node_t* (*FileNodeNew)(const char* name);
typedef gboolean (*FileNodeRecent)(file_tree_t* tree, file_node_t* node);
typedef file_node_t* (*FileNodeTraverse)(file_tree_t* tree, file_node_t* node, void* data);
typedef void (*FileNodeFunc)(file_node_t* node);

#define FILE_TREE(obj) ((file_tree_t*)obj)
struct _file_tree_t {
  object_t _object;
  char* description;
  time_t last_refreshed;
  file_size_t fsi[MEDIA_ANY+1][FN_GROUP_SIZE];
  int fci[MEDIA_ANY+1][FN_GROUP_SIZE];

  file_node_t* files;
  int show_media;
  int show_time;
  int mark_timeout;
  
  // functions
  FileNodeNew node_new;
  FileNodeRecent node_recent;
  FileNodeFunc node_shared;
  FileNodeFunc node_unshared;
};

#define FILE_NODE(obj) ((file_node_t*)obj)
struct _file_node_t {
  object_t _object;
  file_node_t* parent;
  file_node_t* child;
  file_node_t* next;
  file_node_t* prev;
  file_t* file;
  file_size_t fsi[MEDIA_ANY+1][FN_GROUP_SIZE];
  int fci[MEDIA_ANY+1][FN_GROUP_SIZE];
  int flags;
};

/** @name Tree functions */
/*@{*/
void file_tree_init(file_tree_t* tree, const char* name, const char* des);
void file_tree_destroy(file_tree_t* tree);
file_node_t* file_tree_insert_file_in(file_tree_t* tree, file_node_t* node, 
                                      file_t* file, int flags, gboolean check_dup);
void file_tree_foreach(file_tree_t* tree, FileNodeTraverse func, void* data);
file_node_t* file_tree_action(file_tree_t* tree, file_node_t* node, 
                              FileNodeTraverse action, void* data);
void file_tree_mark(file_tree_t* tree, int when);
file_node_t* file_tree_insert_folder(file_tree_t* tree, file_node_t* parent, char* folder);
file_node_t* file_tree_insert_file(file_tree_t* tree, file_t* file, int flags);
file_node_t* file_tree_insert_file_inherited(file_tree_t* tree, file_t* file, int flags);
void file_tree_share_node(file_tree_t* tree, file_node_t* node, unsigned media, unsigned group);
void file_tree_unshare_node(file_tree_t* tree, file_node_t* node, unsigned media, unsigned group);
void file_tree_remove_node(file_tree_t* tree, file_node_t* node);
/*@}*/


/** @name Node functions */
/*@{*/
void file_node_init(file_node_t* node, const char* name);
void file_node_destroy(file_node_t* node);
char* file_node_get_folder(file_node_t* node, int* size);
void file_node_update(file_tree_t* tree, file_node_t* node);
/*@}*/


/*@}*/

file_node_t* file_tree_search_filename(file_tree_t* tree, char* filename);

G_END_DECLS
#endif
