/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.11 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include <string.h>
#include <ctype.h>

#include "object.h"
#include "files.h"
#include "attribute.h"
#include "mtypes.h"
#include "l_i18n.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

/** Convert a windows filename to the local filename 
 * @param name the filename
 * @return the same filename with converted characters.
 */
char* win2local(char *name) {
  char *ptr = name;

  if (DIR_SEP == '\\') return name;

  if (!ptr || !*ptr) return NULL;

  while (*ptr) {
    if (*ptr == '\\') *ptr = DIR_SEP;
    ptr++;
  }

  return name;
}

/** Convert a local filename to the windows filename 
 * @param name the filename
 * @return the same filename with converted characters.
 */
char* local2win(char* name) {
  char *ptr = name;

  if (DIR_SEP == '\\') return name;
  
  if (!ptr || !*ptr) return name;

  while (*ptr) {
    if (*ptr == DIR_SEP) *ptr = '\\';
    ptr++;
  }

  return name;
}

/** Convert a unix filename to the windows filename 
 * @param name the filename
 * @return the same filename with converted characters.
 */
char* unix2win(char* name) {
  char *ptr = name;
  if (!ptr || !*ptr) return name;

  while (*ptr) {
    if (*ptr == '/') *ptr = '\\';
    ptr++;
  }

  return name;
}

/** Destroy a file. This function may be called from the destroy
 * function of a derived class only.
 * @param file the file.
 */
void file_destroy(file_t* file) {
  if (!file) return;
  g_free(file->localname);
  g_free(file->netname);
  g_free(file->mimetype);
  attr_container_destroy(ATTRC(file));
}

/** Initialize file. This function may be called from the init
 * function of a derived class only.
 * @param file the file.
 */
void file_init(file_t* file) {
  attr_container_init(ATTRC(file), "File");
  object_add_type(OBJ(file), "File");
  OBJ(file)->destroy = (ObjectFunc)file_destroy;
}

/** Create a new file.
 * @return teh new file
 */
file_t* file_new() {
  file_t* file = g_malloc0(sizeof(*file));
  file_init(file);
  return file;
}

/*
void file_set_unixname(file_t* file, const char* name) {
  char* pos;

  if (file->longname) g_free(file->longname);
  file->longname = g_strdup(name);
  pos = strrchr(file->longname, '/');
  if (pos) file->filename = pos+1;
  else file->filename = file->longname;
  file->media = mtype_find_from_suffix(file->filename);
}
*/

/** Set the windows name of the file. The struct is updated accordingly.
 * @param file the file.
 * @param name the windows name of the file.
 */
void file_set_origname(file_t* file, char* name) {
  g_free(file->netname);
  file->netname = local2win(name);
}

/** Set the local name of the file. The struct is updated accordingly.
 * @param file the file.
 * @param name the local name of the file.
 */
void file_set_locname(file_t* file, char* name) {
  char* pos;

  g_free(file->localname);
  file->localname = win2local(name);

  pos = strrchr(file->localname, DIR_SEP);
  if (pos) file->filename = pos+1;
  else file->filename = file->localname;
  g_free(file->mimetype);
  file->mimetype = mtype_get_mime_type(file->localname, &file->media);
}

/** Copy the file.
 * @param dest the destination file.
 * @param source the source file.
 * @param domain the attribute domain which should be copied or NULL to copy all attributes.
 * @see attr_container_copy()
 */
void file_copy(file_t* dest, file_t* source, const char* domain) {
  attr_container_copy(ATTRC(dest), ATTRC(source), domain);
  object_add_type(OBJ(dest), "File");
  g_free(dest->localname);
  dest->localname = g_strdup(source->localname);
  g_free(dest->netname);
  dest->netname = g_strdup(source->netname);
  dest->filename = dest->localname + (source->filename-source->localname);
  dest->media = source->media;
  dest->size = source->size;
  g_free(dest->mimetype);
  dest->mimetype = g_strdup(source->mimetype);
}

/** Clone a file.
 * @param source the file to clone.
 * @param attr the attribute domain to copy, see file_copy()
 * @return the cloned file.
 */
file_t* file_clone(file_t* source, const char* attr) {
  file_t* result;
  if (!source) return NULL;
  result = file_new();
  file_copy(result, source, attr);
  return result;
}

#ifdef WIN32
# define _FILES_C_
# include "lcorealiasdef.c"
#endif
