/* Copyright (C) 2005 sgop@users.sourceforge.net
   This is free software distributed under the terms of the
   GNU Public License.  See the file COPYING for details. */

#include <string.h>
#include <stdio.h>

#include "g_commands.h"
#include "plugin.h"
#include "utils.h"
#include "exec.h"
#include "preferences.h"
#include "l_i18n.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef void (*GCommandFunc)(chat_page_t*, char*);
typedef void (*CommandHandleFunc)(chat_page_t* page, command_t*, char*);
#define G_COMM(f) static void f (chat_page_t* _page, char *_data)

#define G_COMMAND(c) ((g_command_t*)(c))
typedef struct {
  command_t _command;
  GCommandFunc handle;
} g_command_t;

typedef struct {
  module_t _module;
  int cnt;
} g_command_module_t;

static g_command_module_t* GCommandModule = NULL;

static void g_command_new(const char* name, const char* args,
                          const char* des, GCommandFunc func) {
  g_command_t *comm;

  g_assert(GCommandModule);

  g_assert(name);

  comm = g_malloc0(sizeof(g_command_t));
  command_init(COMMAND(comm), -1, name, args, des);
  comm->handle = func;
}

void g_command_handle(chat_page_t* page, g_command_t* command, char* data) {
  if (!command) return;
  command->handle(page, data);
}


G_COMM(gcom_plugin) {
  int i1;

  (void)_data;

  chat_print(_page, "+\0client",
             "+\1c_client", _("List of plugins:"),
             "+\0nl", NULL);
  for (i1 = 0; i1 < plugin_get_number(); i1++) {
    plugin_t* plugin = plugin_get(i1);
    const char* version = plugin_get_version_string(plugin);

    chat_print(_page, "+\0client",
               "+\1c_text", "  ",
               "+\1c_bracket", "[",
               "+\0c_client", l_printf("%-4s", plugin_get_short_name(plugin)),
               "+\0c_bracket", "]",
               "+\1c_text", plugin_get_name(plugin),
               "+\1c_text", l_printf(_("(version %s)"), version),
               "+\0nl", NULL);
  }

  i1 = plugin_get_number();
  chat_print(_page, "+\0client",
             "+\1c_client", l_printf(ngettext("%d plugin loaded!", "%d plugins loaded!", i1), i1),
             "+\0nl", NULL);
}

static void on_print_command(command_t* command, chat_page_t* page) {
  if (GCommandModule->cnt == 0) chat_print(page, "+\2client", NULL);
  chat_print(page,
             "+\1c_bracket", "[",
             "+\0c_client", l_printf("%-12s", command->name),
             "+\0c_bracket", "]", NULL);
  GCommandModule->cnt++;
  if ((GCommandModule->cnt % 5) == 0) {
    chat_print(page, "+\0nl", NULL);
    GCommandModule->cnt = 0;
  }
}

static int
commands_print_help(chat_page_t* page, const char* name, int proto) {
  command_t* command;
  
  g_assert(GCommandModule);

  if (name) {
    command = command_search(name, proto);
    if (!command) return 0;
    chat_print(page, "+\0client",
               "+\1c_client", _("Help for command"),
               "+\1c_server", command->name, NULL);
    if (proto == -1) {
      chat_print(page, "+\1proto", "CORE", "+\0nl", NULL);
    } else {
      chat_print(page, "+\1proto", plugin_get_name(plugin_get(proto)),
                 "+\0nl", NULL);
    }
    chat_print(page, "+\0client",
               "+\1c_client", _(" Syntax     :"),
               "+\1c_text", "/",
               "+\0c_text", command->name,
               "+\1c_text", command->args,
               "+\0nl", NULL);
    chat_print(page, "+\0client",
               "+\1c_client", _(" Description:"),
               "+\1c_text", command->description,
               "+\0nl", NULL);
  } else {
    if (proto == -1) {
      chat_print(page, "+\0client",
                 "+\1c_client", _("List of commands:"),
                 "+\1c_text", "CORE",
                 "+\0nl", NULL);
    } else {
      chat_print(page, "+\0client",
                 "+\1c_client", _("List of commands:"),
                 "+\1c_text", plugin_get_name(plugin_get(proto)),
                 "+\0nl", NULL);
    }
    
    GCommandModule->cnt = 0;
    command_foreach((CommandForeachFunc)on_print_command, proto, page);
    if (GCommandModule->cnt % 5) {
      chat_print(page, "+\0nl", NULL);
      GCommandModule->cnt = 0;
    }
  }
  return 1;
}

G_COMM(gcom_help) {
  int i1;
  char* command = arg(_data, 0);
  int cnt;

  cnt = commands_print_help(_page, command, -1);
  for (i1 = 0; i1 < plugin_get_number(); i1++) {
    cnt += commands_print_help(_page, command, i1);
  }
  COMMAND_ERROR(command && !cnt, _page, 
                _("Unknown command! Type /help for a list of commands"));
}

static void
print_pref_value(pref_node_t* node, chat_page_t* page, char* prefix) {
  GList* dlist;
  char str[1024] = "";
  const char* value;
  char* pos;

  if (!node) return;
  
  if (!prefix) pos = str;
  else pos = str+sprintf(str, "%s", prefix);

  if (node->xml_print) {
    value = node->xml_print(node->value);
    chat_print(page, "+\0client",
               "+\1c_text", l_printf("%-35s", str),
               "+\1c_client", ":",
               "+\1c_text", value?value:"<unknown>",
               "+\0nl", NULL);
  }
  for (dlist = node->children; dlist; dlist = dlist->next) {
    pref_node_t* child = dlist->data;
    sprintf(pos, "/%s", OBJ(child)->name);
    print_pref_value(child, page, str);
  }
}

G_COMM(gcom_pref) {
  char* command;
  char* field;
  const char* value;
  pref_node_t* node;

  command = arg(_data, 0);
  COMMAND_ERROR(!command, _page, _("Argument required"));

  field = arg(NULL, 0);

  if (!strcasecmp(command, "set")) {
    COMMAND_ERROR(!field, _page, _("More arguments required"));
    value = arg(NULL, 1);
    COMMAND_ERROR(!value, _page, _("Value required"));
    pref_store_str(field, value);
  } else if (!strcasecmp(command, "show")) {
    node = pref_search(field?field:"", FALSE);
    COMMAND_ERROR(!node, _page, _("No such node in preferences"));

    print_pref_value(node, _page, field);
  } else {
    COMMAND_ERROR(TRUE, _page, _("Unknown action"));
  }
}

static gboolean exec_to_page(exec_t* exec, chat_page_t* page) {
  char* pos = buffer_read_pos(exec->buffer);
  char* nl = strchr(pos, '\n');

  while (nl) {
    *nl = 0;
    chat_print(page, "+\2client",
               "+\0c_bracket", "[",
               "+\0c_text", "exec",
               "+\0c_bracket", ":",
               "+\0c_text", l_printf("%03d", exec->internal_id),
               "+\0c_bracket", "]",
               "+\1c_text", pos,
               "+\0nl", NULL);
    buffer_consume_virt(exec->buffer, nl-pos+1);
    pos = nl+1;
    nl = strchr(pos, '\n');
  }
  buffer_consume(exec->buffer, 0);
  return TRUE;

}
static void on_page_destroy(chat_page_t* page, exec_t* exec) {
  (void)page;
  exec_set_output_func(exec, NULL, NULL);
}
static void on_exec_destroy(exec_t* exec, chat_page_t* page)
{
    if (page)
    {
        object_unnotify_d(OBJ(page), (ObjectFunc)on_page_destroy, exec);
    }
}
G_COMM(gcom_exec)
{
    char* command;
    char** argv;
    exec_t* exec;
    
    command = arg(_data, 1);
    COMMAND_ERROR(!command, _page, _("Argument required"));
    COMMAND_ERROR(!g_shell_parse_argv(command, NULL, &argv, NULL),
                  _page, _("Could not parse command"));
    exec = exec_new(argv);
    
    if (_page)
    {
        object_notify_d(OBJ(_page), (ObjectFunc)on_page_destroy, exec);
    }
    exec_set_output_func(exec, (ExecReadFunc)exec_to_page, _page);
    exec_start(exec, (ExecEndFunc)on_exec_destroy, _page);
}

static void g_commands_init() {
  command_handler_install(-1, (CommandHandler)g_command_handle);

  g_command_new("plugin", NULL,
                _("List all loaded plugins"),
                gcom_plugin);
  g_command_new("help", "[command]",
                _("Help for <command>"),
                gcom_help);
  g_command_new("pref", "<'set' <field> <value>|'show' <field>>",
                _("Set or show preferences value"),
                gcom_pref);
  g_command_new("exec", "<command>",
                _("Execute <command>"),
                gcom_exec);
}

static void g_commands_run() {
}
static void g_commands_exit() {
  //FIXME: cleanup
}

static void g_command_module_init(g_command_module_t* module) {
  module_init(MODULE(module), "Global Command Module",
              (ModuleFunc)g_commands_init,
              (ModuleFunc)g_commands_run,
              (ModuleFunc)g_commands_exit);
  
  module_require(command_module_instance);
  module_require(plugin_module_instance);
  module_require(pref_module_instance);
  module_require(chat_module_instance);
}


module_t* g_command_module_instance() {
  if (!GCommandModule) {
    GCommandModule = g_malloc0(sizeof(*GCommandModule));
    g_command_module_init(GCommandModule);
  }
  return MODULE(GCommandModule);
}

#undef G_COMMAND
#undef G_COMM

#ifdef WIN32
#define _G_COMMANDS_C_
#include "lcorealiasdef.c"
#endif
