/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.6 $
 * $Date: 2007/04/29 09:21:47 $
 * $Author: sgop $
 */

#include "lcore/utils.h"
#include "ui_dialog.h"
#include "ui_support.h"
#include "ui_main.h"

#ifdef WIN32
#include "lgtk2corealias.h"
#endif

struct _ui_page_t {
  char* title;
  PageCreateFunc create;
  PageChangeFunc enter;
  PageChangeFunc leave;
  ui_dialog_t* dialog;

  GtkWidget* content;
  void* data;
};

struct _ui_dialog_t {
  GtkWidget* win;
  GtkLabel* header;
  GtkNotebook* notebook;
  GtkLabel* warning;
  GtkButton* prev;
  GtkButton* next;
  
  GList* pages;
  ui_page_t* current;
  gboolean allow_cancel;
  char* title;
};

static void ui_dialog_destroy(ui_dialog_t* dialog) {
  GList* dlist;

  g_free(dialog->title);
  for (dlist = dialog->pages; dlist; dlist = dlist->next) {
    ui_page_t* page = dlist->data;
    g_free(page->title);
    g_free(page->data);
    g_free(page);
  }
  g_list_free(dialog->pages);

  gtk_widget_destroy(dialog->win);
  g_free(dialog);
}

static gboolean on_dialog_delete(ui_dialog_t* dialog) {
  if (dialog->allow_cancel) ui_dialog_destroy(dialog);
  return TRUE;
}

static gboolean ui_dialog_leave_page(ui_dialog_t* dialog, gboolean next) {
  const char* res = NULL;
  ui_page_t* page = dialog->current;
  
  if (page->leave) res = page->leave(page, next);
  if (res) {
    gtk_label_set_markup
      (dialog->warning, l_printf("<span foreground='red'>%s</span>", res));
    gtk_widget_show(GTK_WIDGET(dialog->warning));
    return FALSE;
  } else {
    gtk_label_set_markup(dialog->warning, "");
    gtk_widget_hide(GTK_WIDGET(dialog->warning));
    return TRUE;
  }
}

static void ui_dialog_enter_page(ui_dialog_t* dialog, gboolean next) {
  const char* res = NULL;
  ui_page_t* page = dialog->current;
  
  ui_set_header_label(dialog->header, page->title);

  if (!page->content) {
    page->content = page->create(page);
    gtk_notebook_append_page(dialog->notebook, page->content, NULL);
  }

  if (page->enter) res = page->enter(page, next);
  if (res) {
    gtk_label_set_markup
      (dialog->warning, l_printf("<span foreground='red'>%s</span>", res));
    gtk_widget_show(GTK_WIDGET(dialog->warning));
  } else {
    gtk_label_set_markup(dialog->warning, "");
    gtk_widget_hide(GTK_WIDGET(dialog->warning));
  }

  gtk_notebook_set_current_page
    (dialog->notebook,
     gtk_notebook_page_num(dialog->notebook, page->content));
}

static void ui_dialog_next(ui_dialog_t* dialog) {
  GList* dlist;

  if (!dialog->current) {
    dlist = dialog->pages;
  } else if (ui_dialog_leave_page(dialog, TRUE)) {
    dlist = g_list_find(dialog->pages, dialog->current);
    g_assert(dlist);
    dlist = dlist->next;
  } else {
    return;
  }
  if (dlist) {
    dialog->current = dlist->data;
    ui_dialog_enter_page(dialog, TRUE);
    if (!dlist->next) gtk_button_set_label(dialog->next, GTK_STOCK_CLOSE);
    gtk_widget_set_sensitive
      (GTK_WIDGET(dialog->prev), dlist->prev!=NULL);
  } else {
    ui_dialog_destroy(dialog);
  }
}

static void ui_dialog_prev(ui_dialog_t* dialog) {
  GList* dlist;

  if (!dialog->current) return;

  dlist = g_list_find(dialog->pages, dialog->current);
  g_assert(dlist);
  if (!dlist->prev) return;

  gtk_button_set_label(dialog->next, GTK_STOCK_GO_FORWARD);

  if (!ui_dialog_leave_page(dialog, FALSE)) return;
  dlist = dlist->prev;
  dialog->current = dlist->data;
  ui_dialog_enter_page(dialog, FALSE);

  gtk_widget_set_sensitive
    (GTK_WIDGET(dialog->prev), dlist->prev!=NULL);
}

static void ui_dialog_create_win(ui_dialog_t* dialog) {
  GtkWidget* win;
  GtkWidget* vbox;
  GtkWidget* notebook;
  GtkWidget* item;
  GtkWidget* hbox;
  
  dialog->win = win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW (win), dialog->title);
  gtk_window_set_default_size(GTK_WINDOW(win), 500, 300);
  gtk_window_set_type_hint(GTK_WINDOW(win), GDK_WINDOW_TYPE_HINT_DIALOG);
  gtk_window_set_position(GTK_WINDOW(win), GTK_WIN_POS_CENTER_ON_PARENT);
  if (ui_main_get_win())
  {
      gtk_window_set_transient_for(GTK_WINDOW(win), ui_main_get_win());
      gtk_window_set_destroy_with_parent(GTK_WINDOW(win), TRUE);
  }
  g_signal_connect_swapped(G_OBJECT(win), "delete_event",
                           G_CALLBACK(on_dialog_delete), dialog);
  
  vbox = gtk_vbox_new(FALSE, 10);
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(win), vbox);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
  
  // the header
  item = ui_create_header_label("");
  dialog->header = GTK_LABEL(item);
  gtk_misc_set_alignment(GTK_MISC(item), 0, 0.5);
  gtk_widget_show(item);
  gtk_box_pack_start(GTK_BOX(vbox), item, FALSE, FALSE, 0);
  
  // the pages
  notebook = gtk_notebook_new();
  dialog->notebook = GTK_NOTEBOOK(notebook);
  gtk_notebook_set_show_tabs(dialog->notebook, FALSE);
  gtk_notebook_set_show_border(dialog->notebook, FALSE);
  gtk_widget_show(notebook);
  gtk_box_pack_start(GTK_BOX(vbox), notebook, TRUE, TRUE, 0);

  // the warning
  item = gtk_label_new("");
  dialog->warning = GTK_LABEL(item);
  gtk_label_set_use_markup(dialog->warning, TRUE);
  gtk_box_pack_start(GTK_BOX(vbox), item, FALSE, FALSE, 0);

  // the buttons
  hbox = ui_create_bbox(10, GTK_BUTTONBOX_END);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  if (dialog->allow_cancel) {
    item = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
    gtk_container_add(GTK_CONTAINER(hbox), item);
    gtk_widget_show(item);
    g_signal_connect_swapped(G_OBJECT(item), "clicked",
			     G_CALLBACK(ui_dialog_destroy), dialog);
  }

  item = gtk_button_new_from_stock(GTK_STOCK_GO_BACK);
  dialog->prev = GTK_BUTTON(item);
  gtk_container_add(GTK_CONTAINER(hbox), item);
  gtk_widget_show(item);
  g_signal_connect_swapped(G_OBJECT(item), "clicked",
                           G_CALLBACK(ui_dialog_prev), dialog);

  item = gtk_button_new_from_stock(GTK_STOCK_GO_FORWARD);
  dialog->next = GTK_BUTTON(item);
  gtk_container_add(GTK_CONTAINER(hbox), item);
  gtk_widget_show(item);
  g_signal_connect_swapped(G_OBJECT(item), "clicked",
                           G_CALLBACK(ui_dialog_next), dialog);

}

static void
ui_dialog_init(ui_dialog_t* dialog, const char* title, gboolean allow_cancel) {
  dialog->title = g_strdup(title);
  dialog->allow_cancel = allow_cancel;
  ui_dialog_create_win(dialog);
}
  
ui_dialog_t* ui_dialog_new(const char* title, gboolean allow_cancel) {
  ui_dialog_t* dialog = g_malloc0(sizeof(*dialog));
  ui_dialog_init(dialog, title, allow_cancel);
  return dialog;
}

void ui_dialog_hide(ui_dialog_t* dialog) {
  gtk_widget_hide(dialog->win);
}

void ui_dialog_add_page(ui_dialog_t* dialog, ui_page_t* page) {
  dialog->pages = g_list_append(dialog->pages, page);
  page->dialog = dialog;
  if (dialog->current) gtk_button_set_label(dialog->next, GTK_STOCK_GO_FORWARD);
}

void ui_dialog_set_modal(ui_dialog_t* dialog, gboolean set) {
  gtk_window_set_modal(GTK_WINDOW(dialog->win), set);
}

void ui_dialog_start(ui_dialog_t* dialog) {
  if (dialog->current) return; // already started.
  ui_dialog_next(dialog);
  gtk_widget_show(dialog->win);
}

static void
ui_page_init(ui_page_t* page, const char* title, PageCreateFunc create,
             PageChangeFunc enter, PageChangeFunc leave) {
  page->title = g_strdup(title);
  page->create = create;
  page->enter = enter;
  page->leave = leave;
}
ui_page_t* ui_page_new(const char* title, PageCreateFunc create,
		       PageChangeFunc enter, PageChangeFunc leave) {
  ui_page_t* page = g_malloc0(sizeof(*page));
  ui_page_init(page, title, create, enter, leave);
  
  return page;
}

ui_dialog_t* ui_page_get_dialog(ui_page_t* page) {
  return page->dialog;
}

void ui_page_set_data(ui_page_t* page, void* data) {
  g_free(page->data);
  page->data = data;
}
void* ui_page_get_data(ui_page_t* page) {
  return page->data;
}

#ifdef WIN32
#define _UI_DIALOG_C_
#include "lgtk2corealiasdef.c"
#endif
