/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.7 $
 * $Date: 2006/07/22 17:11:58 $
 * $Author: sid4u $
 */

#include <string.h>

#include <libxml/parser.h>
#include <glib.h>

#include "lcore/utils.h"
#include "lcore/xml_handler.h"
#include "ui_layout.h"

#ifdef WIN32
#include "lcore/win32_defs.h"
#endif
#ifdef WIN32
#include "lgtk2corealias.h"
#endif

typedef struct {
  char* name;
  char* value;
  GList* objects;
  LayoutGetFunc get;
  LayoutSetFunc set;
} ui_layout_func_t;

static GList* LayoutFuncs = NULL;

static ui_layout_func_t* ui_layout_func_find(const char* name) {
  GList* dlist;

  g_assert(name);
  for (dlist = LayoutFuncs; dlist; dlist = dlist->next) {
    ui_layout_func_t* func = dlist->data;
    if (!strcmp(func->name, name)) return func;
  }
  return NULL;
}

/** Register a generic layout. If possible use one of the predefined
 * wrappers().
 * @param name the name of the layout.
 * @param get the layout get function
 * @param set the layout set function
 * @note make sure you use ui_layout_monitor() only on object that can
 * be handled by these get and set functions.
 */
void ui_layout_register(const char* name, LayoutGetFunc get, LayoutSetFunc set) {
  ui_layout_func_t* func;

  g_assert(name);

  func = ui_layout_func_find(name);
  if (!func) {
    func = g_malloc(sizeof(*func));
    func->name = g_strdup(name);
    func->value = NULL;
    func->objects = NULL;
    func->set = NULL;
    func->get = NULL;
    LayoutFuncs = g_list_append(LayoutFuncs, func);
  }
  if (get) func->get = get;
  if (set) func->set = set;
}

static void xml_load_layout(xmlDocPtr xfile, xmlNodePtr xroot) {
  xmlNodePtr xsub = xroot->xmlChildrenNode;
  while (xsub != NULL) {
    if (!strcmp((char*)xsub->name, "Item")) {
      ui_layout_func_t* func;
      xmlChar* attr;

      attr = xmlGetProp(xsub, "Name");
      if (!attr) {
        xsub = xsub->next;
        continue;
      }
      
      ui_layout_register(attr, NULL, NULL);
      func = ui_layout_func_find(attr);
      xmlFree(attr);

      attr = xmlNodeListGetString(xfile, xsub->xmlChildrenNode, 1);
      if (!attr) {
        xsub = xsub->next;
        continue;
      }
      
      func->value = g_strdup(attr);
      if (func->set && func->objects) {
        g_list_foreach(func->objects, (GFunc)func->set, func->value);
      }
      xmlFree(attr);
    }
    xsub = xsub->next;
  }
}

/** Start monitoring an object for a layout. This function has two
 * important jobs.
 *
 * (1) it immediately applies the saved layout to the object. This
 * means that when creating widgets, *first* setup the default style,
 * *then* call this function.
 *
 * (2) it saves the layout of the object on exit (or when you call
 * ui_layout_unmonitor())
 *
 * You can add as many objects to a registered layout as you
 * like. This for example is useful for chat pages of one type.
 *
 * @param name the name of the layout
 * @param object the object of to monitor
 */
void ui_layout_monitor(const char* name, void* object) {
  ui_layout_func_t* func;

  func = ui_layout_func_find(name);
  if (!func) {
    g_log("GTK_LAYOUT", G_LOG_LEVEL_WARNING, 
          "layout functions '%s' not found", name);
    return;
  }
  func->objects = g_list_append(func->objects, object);
  if (!func->set) {
    g_log("GTK_LAYOUT", G_LOG_LEVEL_WARNING, 
          "no layout set function for '%s'", name);
    return;
  }
  func->set(object, func->value);
}

/** Stop monitoring an object. The layout of the object is saved.
 * @param name the name of the layout
 * @param object the object
 */
void ui_layout_unmonitor(const char* name, void* object) {
  ui_layout_func_t* func;

  func = ui_layout_func_find(name);
  if (!func) {
    g_log("GTK_LAYOUT", G_LOG_LEVEL_WARNING, 
          "layout functions '%s' not found", name);
    return;
  }
  func->objects = g_list_remove(func->objects, object);
  if (!func->get) {
    g_log("GTK_LAYOUT", G_LOG_LEVEL_WARNING, 
          "no layout get function for '%s'", name);
    return;
  }
  g_free(func->value);
  func->value = func->get(object);
}

  
static void xml_save_layout(ui_layout_func_t* func, xmlNodePtr xroot) {
  xmlNodePtr xsub;
  if (func->get && func->objects) {
    g_free(func->value);
    func->value = func->get(func->objects->data);
  }
  xsub = xmlNewTextChild(xroot, NULL, "Item", func->value);
  xmlNewProp(xsub, "Name", func->name);
}

static void xml_save_layouts(xmlDocPtr xfile, xmlNodePtr xroot) {
  (void)xfile;
  g_list_foreach(LayoutFuncs, (GFunc)xml_save_layout, xroot);
}

void ui_layout_save() {
  lopster_xml_save("ui:core:layout", 1000);
}

void ui_layout_init() {
  lopster_xml_register_home("ui:core:layout", SDIR("gtk2") "layout.xml",
                            "Layout", xml_load_layout, xml_save_layouts);
  lopster_xml_load("ui:core:layout");
}


// predefined funcs
static char* ui_treeview_get_col_size(GtkTreeViewColumn* col) {
  int real_requested_width;

  g_assert(col);
  if (col->use_resized_width) {
    real_requested_width = col->resized_width;
  } else if (gtk_tree_view_column_get_sizing(col) == GTK_TREE_VIEW_COLUMN_FIXED) {
    real_requested_width = col->fixed_width;
  } else {
    real_requested_width = col->requested_width;
    if (real_requested_width < 0)
      real_requested_width = 0;
  }

  if (col->min_width != -1)
    real_requested_width = MAX(real_requested_width, col->min_width);
  if (col->max_width != -1)
    real_requested_width = MIN(real_requested_width, col->max_width);

  return g_strdup_printf("%d", real_requested_width);
}
static void ui_treeview_set_col_size(GtkTreeViewColumn* col, const char* val) {
  int width = 50;
  g_assert(col);
  
  if (val) width = atoi(val);
  if (gtk_tree_view_column_get_sizing(col) == GTK_TREE_VIEW_COLUMN_FIXED)
    gtk_tree_view_column_set_fixed_width(col, width>0?width:100);
}

static char* ui_treeview_get_col_vis(GtkTreeViewColumn* col) {
  g_assert(col);
  return g_strdup_printf("%s", gtk_tree_view_column_get_visible(col)?"yes":"no");
}
static void ui_treeview_set_col_vis(GtkTreeViewColumn* col, const char* val) {
  g_assert(col);
  if (!val)
    gtk_tree_view_column_set_visible(col, TRUE);
  else if (!strcasecmp(val, "TRUE") ||
           !strcasecmp(val, "yes") || !strcmp(val, "1"))
    gtk_tree_view_column_set_visible(col, TRUE);
  else
    gtk_tree_view_column_set_visible(col, FALSE);
}

static char* ui_paned_get_pos(GtkPaned* paned) {
  g_assert(paned);
  return g_strdup_printf("%d", gtk_paned_get_position(paned));
}
static void ui_paned_set_pos(GtkPaned* paned, const char* val) {
  g_assert(paned);
  if (val)
    gtk_paned_set_position(paned, atoi(val));
}

static char* ui_widget_get_vis(GtkWidget* widget) {
  g_assert(widget);
  return g_strdup_printf("%s", GTK_WIDGET_VISIBLE(widget)?"yes":"no");
}
static void ui_widget_set_vis(GtkWidget* widget, const char* val) {
  g_assert(widget);
  if (!val) return;
  else if (!strcasecmp(val, "true") ||
           !strcasecmp(val, "yes") || !strcmp(val, "1"))
    gtk_widget_show(widget);
  else
    gtk_widget_hide(widget);
}

static char* ui_widget_get_exp(GtkExpander* widget) {
  g_assert(widget);
  return g_strdup_printf("%s", gtk_expander_get_expanded(widget)?"yes":"no");
}
static void ui_widget_set_exp(GtkExpander* widget, const char* val) {
  g_assert(widget);
  if (!val) return;
  else if (!strcasecmp(val, "true") ||
           !strcasecmp(val, "yes") || !strcmp(val, "1"))
    gtk_expander_set_expanded(widget, TRUE);
  else
    gtk_expander_set_expanded(widget, FALSE);
}

static char* ui_win_get_size(GtkWindow* win) {
  int x, y;
  g_assert(win);
  gtk_window_get_size(win, &x, &y);
  return g_strdup_printf("%d %d", x, y);
}
static void ui_win_set_size(GtkWindow* win, const char* val) {
  int x=0, y=0;
  g_assert(win);
  if (!val) return;
  if (sscanf(val, "%d %d", &x, &y) == 2) {
    if (x > 2048) x=2048;
    if (y > 2048) y=2048;
    gtk_window_resize(win, x, y);
  }
}

static char* ui_win_get_pos(GtkWindow* win) {
  int x, y;
  g_assert(win);
  gtk_window_get_position(win, &x, &y);
  return g_strdup_printf("%d %d", x, y);
}
static void ui_win_set_pos(GtkWindow* win, const char* val) {
  int x=0, y=0;
  g_assert(win);
  if (!val) return;
  if (sscanf(val, "%d %d", &x, &y) == 2)
    gtk_window_move(win, x, y);
}

static char* ui_spin_get_val(GtkSpinButton* spin) {
  g_assert(spin);
  return g_strdup_printf("%d", gtk_spin_button_get_value_as_int(spin));
}
static void ui_spin_set_val(GtkSpinButton* spin, const char* val) {
  g_assert(spin);
  if (val) gtk_spin_button_set_value(spin, atoi(val));
}

static char* ui_get_int(int* ptr) {
  g_assert(ptr);
  return g_strdup_printf("%d", *ptr);
}
static void ui_set_int(int* ptr, const char* val) {
  g_assert(ptr);
  if (val) *ptr = atoi(val);
}

static char* ui_get_bool(gboolean* ptr) {
  g_assert(ptr);
  return g_strdup_printf("%s", (*ptr)?"yes":"no");
}
static void ui_set_bool(gboolean* ptr, const char* val) {
  g_assert(ptr);
  if (!val || *val == 0) return;
  else if (!strcasecmp(val, "true") ||
           !strcasecmp(val, "yes") || !strcmp(val, "1"))
    *ptr = TRUE;
  else
    *ptr = FALSE;
}



/** Save position of a GtkPaned.
 * @param name the name of the layout
 */
void ui_layout_register_paned(const char* name) {
  ui_layout_register(name,
                     (LayoutGetFunc)ui_paned_get_pos,
                     (LayoutSetFunc)ui_paned_set_pos);
}
static void ui_layout_register_col_size(const char* name) {
  ui_layout_register(name,
                     (LayoutGetFunc)ui_treeview_get_col_size,
                     (LayoutSetFunc)ui_treeview_set_col_size);
}
static void ui_layout_register_col_vis(const char* name) {
  ui_layout_register(name,
                     (LayoutGetFunc)ui_treeview_get_col_vis,
                     (LayoutSetFunc)ui_treeview_set_col_vis);
}
/** Save the size and visibility of a GtkTreeViewColumn.
 * @param name the name of the layout
 */
void ui_layout_register_col(const char* name) {
  ui_layout_register_col_size(l_printf("%s:size", name));
  ui_layout_register_col_vis(l_printf("%s:vis", name));
}
/** Save visibility of a GtkWidget.
 * @param name the name of the layout
 */
void ui_layout_register_vis(const char* name) {
  ui_layout_register(name,
                     (LayoutGetFunc)ui_widget_get_vis,
                     (LayoutSetFunc)ui_widget_set_vis);
}
/** Save the state of a GtkExpander.
 * @param name name of the layout
 */
void ui_layout_register_exp(const char* name) {
  ui_layout_register(name,
                     (LayoutGetFunc)ui_widget_get_exp,
                     (LayoutSetFunc)ui_widget_set_exp);
}
static void ui_layout_register_win_size(const char* name) {
  ui_layout_register(name,
                     (LayoutGetFunc)ui_win_get_size,
                     (LayoutSetFunc)ui_win_set_size);
}
static void ui_layout_register_win_pos(const char* name) {
  ui_layout_register(name,
                     (LayoutGetFunc)ui_win_get_pos,
                     (LayoutSetFunc)ui_win_set_pos);
}
/** Save the position and size of a GtkWindow.
 * @param name name of the layout
 */
void ui_layout_register_win(const char* name) {
  ui_layout_register_win_pos(l_printf("%s:pos", name));
  ui_layout_register_win_size(l_printf("%s:size", name));
}
/** Save the value of a GtkSpinButton.
 * @param name name of the layout
 */
void ui_layout_register_spin(const char* name) {
  ui_layout_register(name,
                     (LayoutGetFunc)ui_spin_get_val,
                     (LayoutSetFunc)ui_spin_set_val);
}
/** Save the value of an int*.
 * @param name name of the layout
 */
void ui_layout_register_int(const char* name) {
  ui_layout_register(name,
                     (LayoutGetFunc)ui_get_int,
                     (LayoutSetFunc)ui_set_int);
}
/** Save the value of an gboolean*.
 * @param name name of the layout
 */
void ui_layout_register_bool(const char* name) {
  ui_layout_register(name,
                     (LayoutGetFunc)ui_get_bool,
                     (LayoutSetFunc)ui_set_bool);
}

#ifdef WIN32
#define _UI_LAYOUT_C_
#include "lgtk2corealiasdef.c"
#endif
