/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.22 $
 * $Date: 2007/11/21 20:15:25 $
 * $Author: sgop $
 */

#include <string.h>

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gtk/gtk.h>

#include "lcore/preferences.h"
#include "lcore/mtypes.h"
#include "lcore/library.h"
#include "lcore/chat.h"
#include "lcore/plugin.h"
#include "lcore/string_list.h"
#include "lcore/l_base.h"
#include "lcore/l_i18n.h"
#include "lcore/ui.h"
#include "ui_preferences.h"
#include "ui_support.h"
#include "ui_main.h"
#include "ui_about.h"
#include "ui_style.h"

#ifdef WIN32
#include "lgtk2corealias.h"
#endif

typedef GtkContainer* (*CreatePageFunc)(ui_pref_page_t*);

struct _ui_pref_page_t {
  char* label;
  char* des;
  gboolean scrolled;
  GtkContainer* page;
  GList* children;
  GList* sections;
  CreatePageFunc create;
  ui_pref_page_t* parent;
  plugin_t* plugin;
};

typedef struct {
  char* name;
  UIPrefFunc apply;
  UIPrefFunc cancel;
  gpointer data;
  gboolean dirty;
} ui_pref_t;

typedef struct {
  GtkWidget* top;
  GtkEntry* dir_entry;
} ui_folder_t;

typedef struct {
  char* text;
  unsigned flag;
  GList* children;
} ui_tree_t;



typedef struct {
  mtype_t* type;
  GtkEntry* download_entry;
  GList* appl;
  GtkBox* appl_box;
} gtype_t;

typedef struct {
  gtype_t* parent;
  GtkWidget* top;
  GtkEntry* app_entry;
  GtkEntry* des_entry;
  GtkToggleButton* flag_check;
} gappl_t;

static gtype_t* Types[MEDIA_ANY] = {
  NULL, NULL, NULL, NULL, 
  NULL, NULL, NULL
};

static GtkWidget* PrefWin = NULL;
static GtkBox* SharedFolderBox = NULL;
static GtkNotebook* NoteBook = NULL;
static GList* LibFolder = NULL;

static ui_pref_page_t* Preferences = NULL;
static ui_pref_page_t* PluginPage = NULL;

static GHashTable* PrefFuncs = NULL;


#define UI_BROWSER_NUM    9

static const char* Browser[UI_BROWSER_NUM] = {
  "",
  "gnome-open",         // Gnome Default
  "epiphany",
  "galeon",
  "konqueror",
  "mozilla",
  "firefox",
  "netscape",
  "opera"
};

#define UI_EMAIL_NUM     5

static const char* EMail[UI_EMAIL_NUM] = {
  "",
  "evolution",
  "mozilla -mail",
  "opera",
  "mutt"
};




static void ui_pref_destroy(ui_pref_t* f) {
  g_free(f->name);
  g_free(f);
}

static ui_pref_t* ui_pref_find(const char* name) {
  return g_hash_table_lookup(PrefFuncs, name);
}

/** Set the apply/cancel functions.
 * @param name the name of the setting
 * @param apply the apply function
 * @param cancel the cancel function
 * @param data the userdata for the cancel/apply function
 */
void ui_pref_set_funcs(const char* name, UIPrefFunc apply,
                       UIPrefFunc cancel, gpointer data) {
  ui_pref_t* f;

  if (g_hash_table_lookup(PrefFuncs, name)) {
    g_log("UI_PREF", G_LOG_LEVEL_WARNING,
          "functions for '%s' already set", name);
    return;
  }
  f = g_malloc(sizeof(*f));
  f->name = g_strdup(name);
  f->apply = apply;
  f->cancel = cancel;
  f->data = data;
  f->dirty = FALSE;
  
  g_hash_table_insert(PrefFuncs, f->name, f);
}

/** Set the GUI dirty. Apply/Cancel only is used when the GUI is
 * marked dirty.
 * @param name the name of the setting
 */
void ui_pref_set_dirty(const char* name) {
  ui_pref_t* pref = ui_pref_find(name);
  pref->dirty = TRUE;
}

static void ui_pref_apply_real(const char* name, ui_pref_t* f) {
  if (!f) f = ui_pref_find(name);
  if (f->dirty && f->apply) {
    f->apply(f->name, f->data);
    f->dirty = FALSE;
  }
}

static void ui_pref_apply_all(void) {
  g_hash_table_foreach(PrefFuncs, (GHFunc)ui_pref_apply_real, NULL);
}

static void ui_pref_cancel_real(const char* name, ui_pref_t* f) {
  if (!f) f = ui_pref_find(name);
  if (f->dirty && f->cancel) {
    f->cancel(f->name, f->data);
    f->dirty = FALSE;
  }
}

/** Reset the GUI to the current setting. Does nothing if GUI is dirty.
 * @param name the name of the setting.
 */
void ui_pref_cancel(const char* name) {
  ui_pref_t* f = ui_pref_find(name);
  if (!f->dirty) {
    f->dirty = TRUE;
    ui_pref_cancel_real(name, f);
  }
}

static void ui_pref_cancel_all(void) {
  g_hash_table_foreach(PrefFuncs, (GHFunc)ui_pref_cancel_real, NULL);
}

 
static void ui_pref_activate_node(ui_pref_page_t* node) {
  int num;
  num = gtk_notebook_page_num(NoteBook, GTK_WIDGET(node->page));
  if (num >= 0) gtk_notebook_set_current_page(NoteBook, num);
}

static gboolean on_add_page(ui_pref_page_t* node) {
  GtkContainer* page;
  ui_pref_page_t* node2;
  GList* dlist;

  g_assert(node->create);
  page = node->create(node);

  if (page) {
    GtkWidget* vbox = gtk_vbox_new(FALSE, 0);
    GtkWidget* hbox = gtk_hbox_new(FALSE, 5);
    GtkWidget* label = ui_create_header_label(node->label);
    GtkWidget* sep = gtk_hseparator_new();
    
    gtk_widget_show(vbox);
    gtk_widget_show(hbox);
    gtk_widget_show(label);
    gtk_widget_show(sep);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 0);
    gtk_container_set_border_width(page, 10);
    if (node->scrolled) {
      GtkWidget* sw = gtk_scrolled_window_new(NULL, NULL);
      gtk_widget_show(sw);
      gtk_scrolled_window_set_shadow_type
        (GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_NONE);
      gtk_scrolled_window_set_policy
        (GTK_SCROLLED_WINDOW(sw), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);

      gtk_scrolled_window_add_with_viewport
        (GTK_SCROLLED_WINDOW(sw), GTK_WIDGET(page));
      gtk_viewport_set_shadow_type
        (GTK_VIEWPORT(gtk_bin_get_child(GTK_BIN(sw))), GTK_SHADOW_NONE);
      
      gtk_box_pack_start(GTK_BOX(vbox), sw, TRUE, TRUE, 0);
    } else {
      gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(page), TRUE, TRUE, 0);
    }

    gtk_misc_set_alignment(GTK_MISC(label), 0, 0);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    node2 = node->parent;
    while (node2) {
      label = ui_create_event_label
        (node2->label, G_CALLBACK(ui_pref_activate_node), node2, TRUE);
      gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
      gtk_box_reorder_child(GTK_BOX(hbox), label, 0);
      node2 = node2->parent;
    }

    sep = gtk_hseparator_new();
    gtk_widget_show(sep);
    gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 0);

    gtk_notebook_append_page(NoteBook, vbox, NULL);
    node->page = GTK_CONTAINER(vbox);
    gtk_widget_show(GTK_WIDGET(page));
  }
  for (dlist = node->children; dlist; dlist = dlist->next) {
    node2 = dlist->data;
    on_add_page(node2);
  }
  return FALSE;
}


static void on_pref_cancel(GtkWidget* win)
{
    ui_pref_cancel_all();
    gtk_widget_hide(win);
}

static gboolean on_pref_delete(GtkWidget* win)
{
    on_pref_cancel(win);
    return TRUE;
}

static void on_pref_ok(GtkWidget* win)
{
    ui_pref_apply_all();
    gtk_widget_hide(win);
}

/*
static void on_pref_reset() {
  ui_pref_cancel();
}
*/

static void on_pref_apply(void) {
  ui_pref_apply_all();
}

GtkWidget* ui_create_preferences()
{
    GtkWidget* window;
    GtkWidget* vbox0;
    GtkWidget* hbox;
    GtkWidget* button;
    GtkWidget* notebook;
    
    if (PrefWin)
    {
        gtk_window_present(GTK_WINDOW(PrefWin));
        return NULL;
    }
    
    PrefWin = window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_default_size(GTK_WINDOW(window), 600, 500);

    gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_DIALOG);
    gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER_ON_PARENT);

    //  gtk_widget_set_events (window, GDK_EXPOSURE_MASK | GDK_BUTTON_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);
    gtk_window_set_title(GTK_WINDOW(window), _("Preferences"));
    gtk_window_set_role(GTK_WINDOW(window), "Preferences");
    
    g_signal_connect(G_OBJECT(window), "delete_event",
                     G_CALLBACK(on_pref_delete), NULL);
    
    vbox0 = gtk_vbox_new(FALSE, 10);
    gtk_container_set_border_width(GTK_CONTAINER(vbox0), 10);
    gtk_container_add(GTK_CONTAINER(window), vbox0);
    gtk_widget_show(vbox0);
    
    notebook = gtk_notebook_new();
    gtk_box_pack_start(GTK_BOX(vbox0), notebook, TRUE, TRUE, 0);
    gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
    gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);
    NoteBook = GTK_NOTEBOOK(notebook);
    gtk_widget_show(notebook);
    
    
    hbox = ui_create_bbox(10, GTK_BUTTONBOX_END);
    gtk_box_pack_end(GTK_BOX(vbox0), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    button = gtk_button_new_from_stock(GTK_STOCK_APPLY);
    gtk_container_add(GTK_CONTAINER(hbox), button);
    g_signal_connect_swapped(G_OBJECT(button), "clicked",
                             G_CALLBACK(on_pref_apply), window);
    gtk_widget_show(button);

    button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
    gtk_container_add(GTK_CONTAINER(hbox), button);
    g_signal_connect_swapped(G_OBJECT(button), "clicked",
                             G_CALLBACK(on_pref_cancel), window);
    gtk_widget_show(button);

    button = gtk_button_new_from_stock(GTK_STOCK_OK);
    gtk_container_add(GTK_CONTAINER(hbox), button);
    g_signal_connect_swapped(G_OBJECT(button), "clicked",
                             G_CALLBACK(on_pref_ok), window);
    gtk_widget_show(button);

    /*
      button = gtk_button_new_from_stock(GTK_STOCK_UNDO);
      gtk_container_add(GTK_CONTAINER(hbox), button);
      g_signal_connect_swapped(G_OBJECT(button), "clicked",
      G_CALLBACK(on_pref_reset), window);
    */

    on_add_page(Preferences);
  
    if (ui_main_get_win())
    {
        gtk_window_set_transient_for(GTK_WINDOW(window), ui_main_get_win());
        gtk_window_set_destroy_with_parent(GTK_WINDOW(window), TRUE);
    }

    gtk_widget_show(window);

    //  gtk_widget_hide(sw);  //TEST

    return window;
}

/******************************************************************/
/* MAP: GtkDrawingArea+Spin <-> int pref node */
/******************************************************************/
typedef struct {
  char* name;
  guint32 range;
  guint32 value;
  int x_drag_start;
  int drag_range;
  GtkWidget* da;
  //    GtkSpinButton* spin;
} ui_speed_t;

#define MAXV   (1<<30)

static int calc_x(guint32 val, guint32 max, int width) {
  if (val >= max) return width;
  return (guint64)val * (guint64)width  / (guint64)max;
}

static gboolean
on_bandwidth_expose(GtkWidget* widget, GdkEventExpose* event, ui_speed_t* data) {
  int width = widget->allocation.width;
  int height = widget->allocation.height;
  unsigned x1, texth, posx, dx;
  GdkGC* gc;
  PangoLayout* layout;
  const char* str;

  (void)event;

  layout = ui_get_text_layout(widget);
  ui_layout_get_ext(layout, "128.0 KiB/s", NULL, &texth);
  texth += 4;

  gdk_draw_rectangle
    (widget->window, ui_style_get_gc("bandwidth:color", "Base"),
     TRUE, 0, 0, width, height);
  gdk_draw_rectangle
    (widget->window, ui_style_get_gc("bandwidth:color", "Border"),
     FALSE, 0, 0, width-1, height-1);
  
  posx = calc_x(data->value, data->range, width-4);
  gdk_draw_rectangle
    (widget->window, ui_style_get_gc("bandwidth:color", "Bar"),
     TRUE, 2, 2, posx, height-4);

  gc = ui_style_get_gc("bandwidth:color", "Text");
  str = print_speed(&data->value);
  pango_layout_set_text(layout, str, -1);
  gdk_draw_layout(widget->window, gc, 6, (height-texth)/2, layout);


  dx = 1;
  while (calc_x(dx, data->range, width-4) < 20) dx *=2;
  
  x1 = 0;
  while (x1 <= data->range) {
    posx = calc_x(x1, data->range, width-4);
    gdk_draw_line(widget->window, ui_style_get_gc("bandwidth:color", "Border"),
                  posx+2, 0, posx+2, 5);
    x1 += dx;
  }

  return TRUE;
}

static void _bandwidth_force_update(GtkWidget* da) {
  if (da->window) gdk_window_invalidate_rect(da->window, NULL, FALSE);
}

static gboolean
on_bandwidth_pressed(GtkWidget* widget, GdkEventButton* event, ui_speed_t* data) {
  int width = widget->allocation.width - 4;
  int x = (int)event->x - 2;

  if (x < 0) x = 0;
  if (x > width) x = width;

  if (event->button == 1) {
    if (x == 0) data->value = 0;
    else data->value = ((guint64)data->range * (guint64)x -1) / (guint64)width +1;
    _bandwidth_force_update(data->da);
    ui_pref_set_dirty(data->name);
  } else {
    data->x_drag_start = x>0?x:1;
    data->drag_range = data->range;
  }

  return TRUE;
}

static gboolean
on_bandwidth_motion(GtkWidget* widget, GdkEventMotion* event, ui_speed_t* data) {
  int width = widget->allocation.width - 4;
  int x = (int)event->x - 2;

  if (x < 0) x = 0;
  if (x > width) x = width;

  if (event->state & GDK_BUTTON1_MASK) {
    if (x == 0) data->value = 0;
    else data->value = ((guint64)data->range * (guint64)x -1) / (guint64)width +1;
    _bandwidth_force_update(data->da);
    ui_pref_set_dirty(data->name);
  } else if (x > 0) {
    guint64 res = (guint64)data->drag_range * (guint64)data->x_drag_start / (guint64)x;
    if (res < 16) res = 16;
    else if (res > MAXV) res = MAXV;
    data->range = res;
    _bandwidth_force_update(data->da);
  }

  return TRUE;
}

static void pref_to_speed(pref_node_t* node, ui_speed_t* speed) {
  (void)node;
  ui_pref_cancel(speed->name);
}
static void on_pref_speed_destroy(ui_speed_t* speed) {
  pref_unobserve(speed->name, (ObjectFunc)pref_to_speed, NULL);
  g_free(speed->name);
  g_free(speed);
}
static void on_int_speed_apply(const char* name, ui_speed_t* speed) {
  pref_store(name, speed->value);
}
static void on_int_speed_cancel(const char* name, ui_speed_t* data) {
  data->range = 16;
  data->value = pref_get_ref(name, int);
  while (data->range < data->value) data->range *= 2;
  _bandwidth_force_update(data->da);
}

static GtkWidget* ui_pref_int_speed(const char* name) {
  const int val = pref_get_ref(name, int);
  GtkWidget* hbox;
  GtkWidget* da;
  //  GtkWidget* spin;
  ui_speed_t* data;

  data = g_malloc0(sizeof(*data));
  data->name = g_strdup(name);
  data->range = 16;
  data->value = val;
  while (data->range < val) data->range *= 2;

  hbox = gtk_hbox_new(FALSE, 5);
  
  data->da = da = gtk_drawing_area_new();
  gtk_widget_set_events(da, GDK_BUTTON1_MOTION_MASK|GDK_BUTTON3_MOTION_MASK|GDK_BUTTON_PRESS_MASK);
  gtk_widget_show(da);
  gtk_box_pack_start(GTK_BOX(hbox), da, TRUE, TRUE, 0);

  g_signal_connect(G_OBJECT(da), "expose-event",
                   G_CALLBACK(on_bandwidth_expose), data);
  g_signal_connect(G_OBJECT(da), "button-press-event",
                   G_CALLBACK(on_bandwidth_pressed), data);
  g_signal_connect(G_OBJECT(da), "motion-notify-event",
                   G_CALLBACK(on_bandwidth_motion), data);

  pref_observe(name, (ObjectFunc)pref_to_speed, data);
  g_signal_connect_swapped(G_OBJECT(da), "destroy",
                           G_CALLBACK(on_pref_speed_destroy), data);
  ui_pref_set_funcs(name, (UIPrefFunc)on_int_speed_apply,
                    (UIPrefFunc)on_int_speed_cancel, data);
  return hbox;
}

/******************************************************************/
/* MAP: GtkEntry <-> any pref node */
/******************************************************************/
static void pref_to_entry(pref_node_t* node, const char* name) {
  (void)node;
  ui_pref_cancel(name);
}

static void on_pref_entry_destroy(char* name) {
  pref_unobserve(name, (ObjectFunc)pref_to_entry, NULL);
  g_free(name);
}

static void on_string_entry_apply(const char* name, GtkEntry* entry)
{
    pref_store_str(name, ui_entry_get_text(entry));
}

static void on_string_entry_cancel(const char* name, GtkEntry* entry)
{
    pref_node_t* node = pref_search(name, TRUE);
    ui_entry_set_text(entry, node->xml_print(node->value));
}

/**
 * Map a string setting (see pref_register_string()) to a GtkEntry.
 * @param name the name of the setting
 * @param info info text which will be displayed if field is empty
 * @return the GtkEntry representing the string setting
 */
GtkWidget* ui_pref_string_entry(const char* name, const char* info)
{
    GtkWidget* entry;
    pref_node_t* node = pref_search(name, TRUE);
    const char* text = node->xml_print?node->xml_print(node->value):NULL;
    char* dup = g_strdup(name);
    
    entry = ui_entry_new_default(text, info);
    gtk_widget_set_size_request(entry, 20, -1);
    pref_observe(name, (ObjectFunc)pref_to_entry, dup);
    g_signal_connect_swapped(G_OBJECT(entry), "destroy",
                             G_CALLBACK(on_pref_entry_destroy), dup);
    ui_pref_set_funcs(name, (UIPrefFunc)on_string_entry_apply,
                      (UIPrefFunc)on_string_entry_cancel, entry);
    g_signal_connect_swapped(G_OBJECT(entry), "changed",
                             G_CALLBACK(ui_pref_set_dirty), dup);
    return entry;
}

GtkWidget* ui_pref_string_entry_inv(const char* name, const char* info)
{
    GtkWidget* entry = ui_pref_string_entry(name, NULL);
    gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
    if (info) ui_entry_set_default(GTK_ENTRY(entry), info);
    return entry;
}

/******************************************************************/
/* MAP: GtkTextView <-> string pref node */
/******************************************************************/
static void pref_to_view(pref_node_t* node, const char* name) {
  (void)node;
  ui_pref_cancel(name);
}

static void on_pref_view_destroy(char* name) {
  pref_unobserve(name, (ObjectFunc)pref_to_view, NULL);
  g_free(name);
}

static void on_string_view_apply(const char* name, GtkTextView* view) {
  GtkTextBuffer* buffer = gtk_text_view_get_buffer(view);
  GtkTextIter start, end;
  char* text;
  
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  text = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
  pref_store_str(name, text);
  g_free(text);
}

static void on_string_view_cancel(const char* name, GtkTextView* view) {
  GtkTextBuffer* buffer = gtk_text_view_get_buffer(view);
  const char* text = pref_get_ptr(name, char);

  gtk_text_buffer_set_text(buffer, text?text:"", -1);
}

/** Map a string setting (see pref_register_string()) to a GtkEntry.
 * @param name the name of the setting
 * @return the GtkEntry representing the string setting
 */
GtkWidget* ui_pref_string_view(const char* name) {
  pref_node_t* node = pref_search(name, TRUE);
  const char* text = node->xml_print?node->xml_print(node->value):NULL;
  char* dup = g_strdup(name);
  GtkTextBuffer* buffer;
  GtkWidget* view;
  GtkWidget* sw;  

  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy
    (GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_IN);

  view = gtk_text_view_new();
  gtk_widget_show(view);
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  if (text) gtk_text_buffer_set_text(buffer, text, -1);

  gtk_container_add(GTK_CONTAINER(sw), view);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(view), GTK_WRAP_WORD);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(view), TRUE);
  gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(view), TRUE);


  pref_observe(name, (ObjectFunc)pref_to_view, dup);
  g_signal_connect_swapped(G_OBJECT(view), "destroy",
                           G_CALLBACK(on_pref_view_destroy), dup);
  ui_pref_set_funcs(name, (UIPrefFunc)on_string_view_apply,
                    (UIPrefFunc)on_string_view_cancel, view);
  g_signal_connect_swapped(G_OBJECT(buffer), "changed",
                           G_CALLBACK(ui_pref_set_dirty), dup);
  return sw;  
}

/******************************************************************/
/* MAP: GtkComboBox <-> int pref node */
/******************************************************************/
static void pref_to_combo(pref_node_t* node, const char* name) {
  (void)node;
  ui_pref_cancel(name);
}
static void  on_pref_combo_destroy(char* name) {
  pref_unobserve(name, (ObjectFunc)pref_to_combo, NULL);
  g_free(name);
}
static void on_int_combo_apply(const char* name, GtkComboBox* combo) {
  pref_store(name, gtk_combo_box_get_active(combo));
}
static void on_int_combo_cancel(const char* name, GtkComboBox* combo) {
  gtk_combo_box_set_active(combo, pref_get_ref(name, int));
}

/** Map an int setting (see pref_register_int()) to a GtkComboBox.
 * @param name the name of the setting
 * @param ... list of strings mapping to [0...]
 * @return the GtkComboBoxText representing the int setting
 */
GtkWidget* ui_pref_int_combo(const char* name, ...)
{
    const int val = pref_get_ref(name, int);
    GtkWidget* combo;
    va_list ap;
    char* text;
    
    combo = gtk_combo_box_new_text();
    va_start(ap, name);
    while ((text = va_arg(ap, char*)) != NULL)
        gtk_combo_box_append_text(GTK_COMBO_BOX(combo), text);
    va_end(ap);
    gtk_combo_box_set_active(GTK_COMBO_BOX(combo), val);
    
    if (name)
    {
        char* dup = g_strdup(name);
        pref_observe(name, (ObjectFunc)pref_to_combo, dup);
        g_signal_connect_swapped(G_OBJECT(combo), "destroy",
                                 G_CALLBACK(on_pref_combo_destroy), dup);
        ui_pref_set_funcs(name, (UIPrefFunc)on_int_combo_apply,
                          (UIPrefFunc)on_int_combo_cancel, combo);
        g_signal_connect_swapped(G_OBJECT(combo), "changed",
                                 G_CALLBACK(ui_pref_set_dirty), dup);
    }
    else
    {
        gtk_widget_set_sensitive(combo, FALSE);
    }
    return combo;
}


/******************************************************************/
/* MAP: GtkComboBox <-> character set */
/******************************************************************/
static void pref_to_charset(pref_node_t* node, const char* name) {
  (void)node;
  ui_pref_cancel(name);
}

static void on_pref_charset_destroy(char* name) {
  pref_unobserve(name, (ObjectFunc)pref_to_charset, NULL);
  g_free(name);
}

static void on_charset_combo_apply(const char* name, GtkWidget* combo) {
  pref_store_str(name, ui_charset_control_get_charset(combo));
}

static void on_charset_combo_cancel(const char* name, GtkWidget* combo) {
  ui_charset_control_set_charset(combo, pref_get_ptr(name, char));
}

/** Map a string charset setting (see pref_register_string()) to a
 * GtkComboBox.
 * @param name the name of the setting
 * @return the GtkComboBox representing the character set encoding
 */
GtkWidget* ui_pref_charset_combo(const char* name) {
  GtkWidget* combo;
  const char* text = pref_get_ptr(name, char);
  char* dup = g_strdup(name);

  combo = ui_create_charset_control(text, NULL, NULL);
/*   gtk_widget_set_size_request(combo, 20, -1); */
  pref_observe(name, (ObjectFunc)pref_to_charset, dup);
  g_signal_connect_swapped(G_OBJECT(combo), "destroy",
                           G_CALLBACK(on_pref_charset_destroy), dup);
  ui_pref_set_funcs(name, (UIPrefFunc)on_charset_combo_apply,
                    (UIPrefFunc)on_charset_combo_cancel, combo);
  g_signal_connect_swapped(G_OBJECT(combo), "changed",
                           G_CALLBACK(ui_pref_set_dirty), dup);
  return combo;
}


/******************************************************************/
/* MAP: DropDown <-> int pref node */
/******************************************************************/
static void pref_to_flags(pref_node_t* node, const char* name) {
  (void)node;
  ui_pref_cancel(name);
}
static void on_pref_tree_destroy(ui_tree_t* tree) {
  GList* dlist;
  for (dlist = tree->children; dlist; dlist = dlist->next) {
    ui_tree_t* sub = dlist->data;
    on_pref_tree_destroy(sub);
  }
  g_free(tree->text);
  g_free(tree);
}
static void on_pref_flags_destroy(GtkWidget* item, char* n) {
  ui_tree_t* tree = g_object_get_data(G_OBJECT(item), "tree");
  char* name = g_object_get_data(G_OBJECT(item), "name");

  g_free(n);
  if (name) {
    pref_unobserve(name, (ObjectFunc)pref_to_flags, NULL);
    g_free(name);
  }
  on_pref_tree_destroy(tree);
}
static void
on_int_flags_apply(const char* name, GtkWidget* button) {
  int val = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(button), "value"));
  pref_store(name, val);
}
static void
on_int_flags_cancel(const char* name, GObject* button) {
  int val = pref_get_ref(name, int);
  g_object_set_data(button, "value", GINT_TO_POINTER(val));
}

static void on_flag_activate(GObject* item, GObject* button) {
  ui_tree_t* tree = g_object_get_data(item, "tree");
  char* name = g_object_get_data(button, "name");
  int val = GPOINTER_TO_INT(g_object_get_data(button, "value"));
  gboolean sync = GPOINTER_TO_INT(g_object_get_data(button, "sync"));
  val ^= tree->flag;
  if (sync && name) {
    pref_store(name, val);
  } else {
    g_object_set_data(button, "value", GINT_TO_POINTER(val));
    if (name) ui_pref_set_dirty(name);
  }
}
static GtkWidget*
on_select_flag_rec(GObject* button, ui_tree_t* tree, int val) {
  GtkWidget* popup;
  GList* dlist;

  popup = gtk_menu_new();
  for (dlist = tree->children; dlist; dlist = dlist->next) {
    ui_tree_t* sub = dlist->data;
    GtkWidget* item = NULL;
    GtkWidget* sub_menu;

    if (sub->children) {
      item = gtk_menu_item_new_with_label(sub->text);
      sub_menu = on_select_flag_rec(button, sub, val);
      gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), sub_menu);
    } else {
      if (sub->text) {
        item = gtk_check_menu_item_new_with_label(sub->text);
        if (val & sub->flag)
          gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item), TRUE);
        if (sub->flag) {
          g_signal_connect(G_OBJECT(item), "activate",
                           G_CALLBACK(on_flag_activate), button);
        } else {
          gtk_widget_set_sensitive(item, FALSE);
        }
      } else {
        item = gtk_menu_item_new();
      }
      g_object_set_data(G_OBJECT(item), "tree", sub);
    }
    gtk_widget_show(item);
    gtk_container_add(GTK_CONTAINER(popup), item);
  }
  return popup;
}

static GtkWidget* on_select_flag(GObject* button) {
  ui_tree_t* tree = g_object_get_data(button, "tree");
  int val = GPOINTER_TO_INT(g_object_get_data(button, "value"));

  return on_select_flag_rec(button, tree, val);
}
static ui_tree_t* ui_pref_get_tree(va_list* ap) {
  ui_tree_t* tree;
  char* text;
  unsigned number;
  unsigned flag;
  unsigned i1;

  number = va_arg(*ap, unsigned);
  text = va_arg(*ap, char*);
  flag = va_arg(*ap, unsigned);
  
  tree = g_malloc(sizeof(*tree));
  tree->text = text?g_strdup(text):NULL;
  tree->flag = flag;
  tree->children = NULL;
  
  for (i1 = 0; i1 < number; i1++) {
    ui_tree_t* t = ui_pref_get_tree(ap);
    tree->children = g_list_append(tree->children, t);
  }

  return tree;
}

/** Map an int setting (see pref_register_int()) to a popup button.
 * Every three arguments represent one menu item:
 *
 * (1) unsigned - number of children
 * (2) char* - item label
 * (3) unsigned - flag represented by item

 * @param name the name of the setting
 * @param sync TRUE to immediately update the setting when the GUI changes.
 * @param pos the popup position (see ui_create_popup_button())
 * @param ... arguments representing the menu structure
 * @return the popup button representing the flags
 */
GtkWidget*  ui_pref_int_flags(const char* name, gboolean sync, int pos, ...) {
  const int val = pref_get_ref(name, int);
  GtkWidget* button;
  va_list ap;
  ui_tree_t* tree;

  va_start(ap, pos);
  tree = ui_pref_get_tree(&ap);
  va_end(ap);

  button =
    ui_create_popup_button(gtk_label_new(tree->text), pos,
                           (PopupFunction)on_select_flag, NULL);
  g_object_set_data(G_OBJECT(button), "sync", GINT_TO_POINTER(sync));
  g_object_set_data(G_OBJECT(button), "value", GINT_TO_POINTER(val));
  g_object_set_data(G_OBJECT(button), "tree", tree);
  g_object_set_data(G_OBJECT(button), "name", name?g_strdup(name):NULL);

  if (name) {
    char* dup = g_strdup(name);
    pref_observe(name, (ObjectFunc)pref_to_flags, dup);
    ui_pref_set_funcs(name, (UIPrefFunc)on_int_flags_apply,
                      (UIPrefFunc)on_int_flags_cancel, button);
    g_signal_connect(G_OBJECT(button), "destroy",
                     G_CALLBACK(on_pref_flags_destroy), dup);
  } else {
    gtk_widget_set_sensitive(button, FALSE);
    g_signal_connect(G_OBJECT(button), "destroy",
		     G_CALLBACK(on_pref_flags_destroy), NULL);
  }
  return button;
}


/******************************************************************/
/* MAP: GtkSpinButton <-> int pref node */
/******************************************************************/
static void pref_to_spin(pref_node_t* node, const char* name) {
  (void)node;
  ui_pref_cancel(name);
}

static void on_pref_spin_destroy(char* name) {
  pref_unobserve(name, (ObjectFunc)pref_to_spin, NULL);
  g_free(name);
}

static void
on_int_spin_apply(const char* name, GtkSpinButton* spin) {
  pref_store(name, gtk_spin_button_get_value_as_int(spin));
}

static void
on_int_spin_cancel(const char* name, GtkSpinButton* spin) {
  gtk_spin_button_set_value(spin, pref_get_ref(name, int));
}

/** Map an int setting (see pref_register_int()) to a GtkSpinButton.
 * @param name the name of the setting
 * @param min the min value allowed
 * @param max the max value allowed
 * @return the GtkSpinButton representing the int setting
 */
GtkWidget* ui_pref_int_spin(const char* name, int min, int max) {
  const int val = pref_get_ref(name, int);
  GtkWidget* spin;

  spin = gtk_spin_button_new_with_range(min, max, 1);
  gtk_spin_button_set_digits(GTK_SPIN_BUTTON(spin), 0);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(spin), val);

  if (name) {
    char* dup = g_strdup(name);
    pref_observe(name, (ObjectFunc)pref_to_spin, dup);
    g_signal_connect_swapped(G_OBJECT(spin), "destroy",
			     G_CALLBACK(on_pref_spin_destroy), dup);
    ui_pref_set_funcs(name, (UIPrefFunc)on_int_spin_apply,
		      (UIPrefFunc)on_int_spin_cancel, spin);
    g_signal_connect_swapped(G_OBJECT(spin), "changed",
			     G_CALLBACK(ui_pref_set_dirty), dup);
  } else {
    gtk_widget_set_sensitive(spin, FALSE);
  }
  
  return spin;
}

/******************************************************************/
/* MAP: EventLabel <-> bool pref node */
/******************************************************************/
static void pref_to_event(pref_node_t* node, const char* name) {
  (void)node;
  ui_pref_cancel(name);
}

static void on_pref_event_destroy(GtkWidget* item, char* name) {
  char* tlabel = g_object_get_data(G_OBJECT(item), "tlabel");
  char* flabel = g_object_get_data(G_OBJECT(item), "flabel");

  pref_unobserve(name, (ObjectFunc)pref_to_event, NULL);
  g_free(name);
  g_free(tlabel);
  g_free(flabel);
}

static void on_en_dis(GtkWidget* ebox, GdkEventButton* event,
		      const char* name) {
  GtkWidget* widget = g_object_get_data(G_OBJECT(ebox), "widget");
  int swap = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(ebox), "swap"));
  gboolean active = (g_object_get_data(G_OBJECT(ebox), "active") == NULL);
  const char* tlabel = g_object_get_data(G_OBJECT(ebox), "tlabel");
  const char* flabel = g_object_get_data(G_OBJECT(ebox), "flabel");

  (void)event;
  ui_event_label_set_text(ebox, active?tlabel:flabel);
  g_object_set_data(G_OBJECT(ebox), "active", active?ebox:NULL);
  if (widget) gtk_widget_set_sensitive(widget, swap?!active:active);
  ui_pref_set_dirty(name);
}

static void
on_bool_label_apply(const char* name, GtkWidget* ebox) {
  gboolean active = (g_object_get_data(G_OBJECT(ebox), "active") != NULL);
  pref_store(name, active);
}

static void
on_bool_label_cancel(const char* name, GtkWidget* ebox) {
  GtkWidget* widget = g_object_get_data(G_OBJECT(ebox), "widget");
  const char* tlabel = g_object_get_data(G_OBJECT(ebox), "tlabel");
  const char* flabel = g_object_get_data(G_OBJECT(ebox), "flabel");
  int swap = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(ebox), "swap"));
  gboolean active = pref_get_ref(name, gboolean);

  ui_event_label_set_text(ebox, active?tlabel:flabel);
  g_object_set_data(G_OBJECT(ebox), "active", active?ebox:NULL);
  if (widget) gtk_widget_set_sensitive(widget, swap?!active:active);
}

/** Map a boolean setting (see pref_register_bool()) to an event label.
 * @param name the name of the setting
 * @param tlabel label to display when setting is TRUE
 * @param flabel label to display when setting is FALSE
 * @param ref widget to disable/enable when setting is FALSE/TRUE 
 * @param swap inverse the behaviour of the widget
 * @return the event label representing the boolean setting
 */
GtkWidget* ui_pref_bool_label(const char* name, const char* tlabel, 
                              const char* flabel, GtkWidget* ref,
                              gboolean swap) {
  GtkWidget* label;
  gboolean val = pref_get_ref(name, gboolean);
  char* dup = g_strdup(name);

  label = ui_create_event_label
    (val?tlabel:flabel, G_CALLBACK(on_en_dis), dup, FALSE);
  
  g_object_set_data(G_OBJECT(label), "widget", ref);
  if (val) g_object_set_data(G_OBJECT(label), "active", label);
  g_object_set_data_full(G_OBJECT(label), "tlabel", g_strdup(tlabel), g_free);
  g_object_set_data_full(G_OBJECT(label), "flabel", g_strdup(flabel), g_free);
  g_object_set_data(G_OBJECT(label), "swap", GINT_TO_POINTER(swap));
  
  pref_observe(name, (ObjectFunc)pref_to_event, dup);
  g_signal_connect(G_OBJECT(label), "destroy",
                   G_CALLBACK(on_pref_event_destroy), dup);
  ui_pref_set_funcs(name, (UIPrefFunc)on_bool_label_apply,
                    (UIPrefFunc)on_bool_label_cancel, label);
  
  if (ref) gtk_widget_set_sensitive(ref, swap?!val:val);

  return label;
}

/******************************************************************/
/* MAP: GtkCheckButton <-> bool pref node */
/******************************************************************/
static void pref_to_check(pref_node_t* node, const char* name) {
  (void)node;
  ui_pref_cancel(name);
}


static void on_en_dis2(GtkToggleButton* check, GtkWidget* widget) {
  int swap = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(check), "swap"));
  gboolean active = (gtk_toggle_button_get_active(check)==TRUE);

  if (widget) gtk_widget_set_sensitive(widget, swap?!active:active);
}

static void on_pref_check_destroy(char* name) {
  pref_unobserve(name, (ObjectFunc)pref_to_check, NULL);
  g_free(name);
}

static void
on_bool_check_apply(const char* name, GtkToggleButton* check) {
  pref_store(name, gtk_toggle_button_get_active(check));
}

static void
on_bool_check_cancel(const char* name, GtkToggleButton* check) {
  gtk_toggle_button_set_active(check, pref_get_ref(name, gboolean));
}

GtkWidget*
ui_pref_bool_check(const char* name, const char* tlabel,
                   GtkWidget* ref, gboolean swap) {
  GtkWidget* check;
  gboolean val = pref_get_ref(name, gboolean);
  char* temp = g_strdup(name);

  check = gtk_check_button_new_with_label(tlabel);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), val);
  g_signal_connect(G_OBJECT(check), "toggled",
                   G_CALLBACK(on_en_dis2), ref);
  g_object_set_data(G_OBJECT(check), "swap", GINT_TO_POINTER(swap));
  
  pref_observe(name, (ObjectFunc)pref_to_check, temp);
  g_signal_connect_swapped(G_OBJECT(check), "destroy",
                           G_CALLBACK(on_pref_check_destroy), temp);
  ui_pref_set_funcs(name, (UIPrefFunc)on_bool_check_apply,
                    (UIPrefFunc)on_bool_check_cancel, check);
  g_signal_connect_swapped(G_OBJECT(check), "toggled",
                           G_CALLBACK(ui_pref_set_dirty), temp);
  
  if (ref) gtk_widget_set_sensitive(ref, val);

  return check;
}


/******************************************************************/
/* MAP: GtkButton <-> pref node representing a image file */
/******************************************************************/

static void pref_to_image(pref_node_t* node, const char* name) {
  (void)node;
  ui_pref_cancel(name);
}

static void on_pref_image_destroy(GObject* button, char* name) {
  GtkWidget* image = gtk_bin_get_child(GTK_BIN(button));
  char* file = g_object_get_data(G_OBJECT(image), "file");
  g_free(file);
  pref_unobserve(name, (ObjectFunc)pref_to_image, NULL);
  g_free(name);
}

static void on_image_button_apply(const char* name, GtkWidget* image) {
  char* file = g_object_get_data(G_OBJECT(image), "file");
  pref_store_str(name, file);
}

#define MAX_IMAGE_WIDTH   200
static void on_image_button_set(GtkWidget* image) {
  char* file = g_object_get_data(G_OBJECT(image), "file");
  int height = image->allocation.height;
  GError* error = NULL;
  GdkPixbuf* buf = NULL;
  
/*   g_message(" -- on_image_button_set %d", height); */
  if (file && *file)
    buf = gdk_pixbuf_new_from_file_at_scale
      (file, MAX_IMAGE_WIDTH/*req.height-BORDER*/, height, TRUE, &error);
  
  if (!buf) {
    if (error) {
      lopster_notify(LOPSTER_MESSAGE_INFO, _("Failed to set image"), error->message);
      g_error_free(error);
      g_free(file);
    }
    g_object_set_data(G_OBJECT(image), "file", NULL);
    // set standard icon
    buf = create_pixbuf("user_icon.png");
    gtk_image_set_from_pixbuf(GTK_IMAGE(image), buf);
    gdk_pixbuf_unref(buf);
  } else {
    gtk_image_set_from_pixbuf(GTK_IMAGE(image), buf);
    gdk_pixbuf_unref(buf);
  }
}

static void on_image_button_size(GtkWidget* image, GtkAllocation* allocation) {
  const GdkPixbuf* buf = gtk_image_get_pixbuf(GTK_IMAGE(image));
  int width, height;
  char* file = g_object_get_data(G_OBJECT(image), "file");

  if (!file || !buf) return;

  width = gdk_pixbuf_get_width(buf);
  height = gdk_pixbuf_get_height(buf);

/*   g_message(" -- %d %d / %d %d", width, height, */
/*             allocation->width, allocation->height); */

  if (width < MAX_IMAGE_WIDTH && height < allocation->height)
    on_image_button_set(image);
  else if (height > allocation->height)
    on_image_button_set(image);
}


static void on_image_button_cancel(const char* name, GtkWidget* image) {
  char* file = pref_get_ptr(name, char);
  char* old_file = g_object_get_data(G_OBJECT(image), "file");

  g_free(old_file);
  g_object_set_data(G_OBJECT(image), "file", file);

  on_image_button_set(image);
}

static void on_browse_image(GtkButton* button, const char* name)
{
    GtkWidget* dialog;
    GtkWidget* image = gtk_bin_get_child(GTK_BIN(button));
    char* file = g_object_get_data(G_OBJECT(image), "file");
    int response;
    
    dialog = gtk_file_chooser_dialog_new
        (_("Choose picture"), GTK_WINDOW(PrefWin),
         GTK_FILE_CHOOSER_ACTION_OPEN,
         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
         _("No image"), 1,
         GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
         NULL);
    gtk_file_chooser_set_local_only(GTK_FILE_CHOOSER(dialog), TRUE);
    gtk_file_chooser_set_select_multiple(GTK_FILE_CHOOSER(dialog), FALSE);
    gtk_file_chooser_set_show_hidden(GTK_FILE_CHOOSER(dialog), FALSE);
    
    if (file && *file)
        gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(dialog), file);
    
    response = gtk_dialog_run(GTK_DIALOG(dialog));
    if (response == GTK_RESPONSE_ACCEPT || response == 1)
    {
        char* new_file;
        if (response == 1) new_file = NULL;
        else new_file = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
        
        g_free(file);
        g_object_set_data(G_OBJECT(image), "file", new_file);
        
        on_image_button_set(image);
        ui_pref_set_dirty(name);
    }
    
    gtk_widget_destroy(dialog);
}

static GtkWidget* ui_pref_image_button(const char* name) {
  GtkWidget* button;  
  GtkWidget* item;
  const char* file = pref_get_ptr(name, char);
  char* dup = g_strdup(name);

  button = gtk_button_new();
  pref_observe(name, (ObjectFunc)pref_to_image, dup);
  g_signal_connect(G_OBJECT(button), "destroy",
                   G_CALLBACK(on_pref_image_destroy), dup);
  g_signal_connect(G_OBJECT(button), "clicked",
                   G_CALLBACK(on_browse_image), dup);

  item = create_pixmap("user_icon.png");
  gtk_widget_set_size_request(item, -1, 20);
  gtk_container_add(GTK_CONTAINER(button), item);
  
  ui_pref_set_funcs(name, (UIPrefFunc)on_image_button_apply,
                    (UIPrefFunc)on_image_button_cancel, item);

  g_object_set_data(G_OBJECT(item), "file", g_strdup(file));
  g_signal_connect(G_OBJECT(item), "size-allocate",
                   G_CALLBACK(on_image_button_size), NULL);
  
  on_image_button_set(item);
  
  return button;
}











/** Create a deactivatable section. Wrapper for ui_create_section()
 * and ui_pref_bool_label().
 * @param header passed to ui_create_section()
 * @param help passed to ui_create_section()
 * @param hbox passed to ui_create_section()
 * @param pref passed to ui_pref_bool_label()
 * @param active passed to ui_pref_bool_label()
 * @param inactive passed to ui_pref_bool_label()
 * @param swap passed to ui_pref_bool_label()
 * @return the new section
 */
GtkWidget*
ui_pref_bool_section(const char* header, const char* help, GtkWidget** hbox,
                     const char* pref, const char* active,
                     const char* inactive, gboolean swap) {
  GtkWidget* vbox;
  GtkWidget* hhbox;
  GtkWidget* label;

  vbox = ui_create_section(header, help, hbox, &hhbox);

  label = ui_pref_bool_label(pref, active, inactive, *hbox, swap);
  gtk_box_pack_end(GTK_BOX(hhbox), label, FALSE, FALSE, 0);
  return vbox;
}



/*********************************************************/


static void on_string_list_apply(const char* name) {
  char* name2 = l_printf("temp:%s", name);
  string_list_copy_items(name, name2);
}
static void on_string_list_cancel(const char* name) {
  char* name2 = l_printf("temp:%s", name);
  string_list_copy_items(name2, name);
}

static void on_add_item(GtkEntry* entry)
{
    char* name = g_object_get_data(G_OBJECT(entry), "string_list");
    const char* text = ui_entry_get_text(entry);
    if (!text || *text == 0) return;
    string_list_add(name, text);
    //  gtk_entry_set_text(entry, "");
    ui_pref_set_dirty(name+5); //skip the string 'temp:'
}

static gboolean on_tree_button(GtkTreeView* tv, GdkEventButton* event,
			       GtkEntry* entry) {
  GtkTreePath* path;
  GtkTreeIter iter;
  GtkTreeModel* model;
  GtkTreeSelection* select;
  char* item;

  if (event->type != GDK_BUTTON_PRESS) return FALSE;
  if (event->button != 1) return FALSE;
  // avoid un/selects if we have mods pressed.
  if (event->state != 0) return FALSE;

  model = gtk_tree_view_get_model(tv);

  if (!gtk_tree_view_get_path_at_pos(tv, event->x, event->y,
				     &path, NULL, NULL, NULL)) return FALSE;
  
  select = gtk_tree_view_get_selection(tv);
  gtk_tree_model_get_iter(model, &iter, path);
  gtk_tree_path_free(path);
  gtk_tree_model_get(model, &iter, 0, &item, -1);
  /*
  if (!gtk_tree_selection_iter_is_selected(select, &iter)) {
    gtk_tree_selection_select_iter(select, &iter);
  }
  */
  ui_entry_set_text(entry, item);

  return FALSE;
}

static GtkWidget*
ui_pref_string_list(const char* name, const char* pref,
                    const char* label1, const char* label2)
{
    GtkWidget* vbox;
    GtkWidget* item;
    GtkWidget* hbox;
    GtkWidget* sw;
    GtkWidget* tv;
    GtkWidget* entry;
    char* name2 = g_strdup_printf("temp:%s", name);
    GtkRequisition req;

    vbox = gtk_vbox_new(FALSE, label2?5:0);
    
    hbox = gtk_hbox_new(FALSE, 0);

    if (!pref) entry = ui_entry_new(NULL);
    else entry = ui_pref_string_entry(pref, NULL);
    gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);
    gtk_widget_size_request(entry, &req);

    g_object_set_data_full(G_OBJECT(entry), "string_list",
                           name2, (GDestroyNotify)g_free);
    g_signal_connect(G_OBJECT(entry), "activate",
                     G_CALLBACK(on_add_item), NULL);

    item = gtk_button_new_from_stock(GTK_STOCK_ADD);
    gtk_box_pack_start(GTK_BOX(hbox), item, FALSE, FALSE, 0);
    gtk_widget_set_size_request(item, -1, req.height);
    g_signal_connect_swapped(G_OBJECT(item), "clicked",
                             G_CALLBACK(on_add_item), entry);

    
    item = ui_create_headed_item(label1?label1:name, hbox);
    gtk_box_pack_start(GTK_BOX(vbox), item, FALSE, FALSE, 0);
    
    sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy
        (GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_scrolled_window_set_shadow_type
        (GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_IN);
    if (label2)
        item = ui_create_headed_item(label2, sw);
    else
        item = sw;
    gtk_box_pack_start(GTK_BOX(vbox), item, TRUE, TRUE, 0);
    
    string_list_copy_items(name2, name);
    tv = ui_string_list_view(name2, "String list actions");
    gtk_widget_show(tv);
    gtk_container_add(GTK_CONTAINER(sw), tv);
    
    if (pref)
    {
        g_signal_connect(G_OBJECT(tv), "button-press-event",
                         G_CALLBACK(on_tree_button), entry);
    }
    
    ui_pref_set_funcs(name,
                      (UIPrefFunc)on_string_list_apply,
                      (UIPrefFunc)on_string_list_cancel, NULL);
    
    gtk_widget_show_all(vbox);
    return vbox;
}

static void on_browse_folder(GtkEntry* entry)
{
    const char* c = ui_entry_get_text(entry);
    char* name = 
        ui_select_folder(_("Choose folder"), (c && *c)?c:NULL, GTK_WINDOW(PrefWin));

    if (name)
    {
        ui_entry_set_text(entry, name);
        g_free(name);
    }
}







static GtkWidget* ui_general_user(gboolean* expand)
{
    GtkWidget* vbox;
    GtkWidget* hbox;
    GtkWidget* table;
    GtkWidget* item;
    
    vbox = ui_create_section
        (_("User settings"), 
         _("Setup your default nickname and password. "
           "These settings normally will be used by plugins "
           "if you don't specify them at plugin level."),
         &hbox, NULL);

    table = gtk_table_new(1, 3, FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);


    item = ui_create_headed_item
        (_("default username"),
         ui_pref_string_entry("/core/user/nick", _("Enter username")));
    gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL,0,0);
    
    item = ui_create_headed_item
         (_("default password"),
          ui_pref_string_entry_inv("/core/user/pass", _("Enter password")));
    gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL,0,0);
    
    item = ui_create_headed_item
        (_("description"), ui_pref_string_view("/core/user/description"));
    gtk_table_attach(GTK_TABLE(table), item, 0, 2, 1, 2,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    
    item = ui_create_headed_item
        (_("picture"), ui_pref_image_button("/core/user/picture"));
    gtk_table_attach(GTK_TABLE(table), item, 2, 3, 0, 2,
                     0,GTK_FILL|GTK_EXPAND,0,0);
    
    
    gtk_widget_show_all(vbox);
    
    if (expand) *expand = TRUE;
    
    return vbox;
}

static GtkWidget* ui_general_ports(gboolean* expand)
{
    GtkWidget* vbox;
    GtkWidget* hbox;
    GtkWidget* item;
    GtkWidget* table;
    
    vbox = ui_pref_bool_section
        (_("Allowed data ports"), 
         _("You should deactivate this box if you cannot accept any connections. "
           "This might highly decrease the number of search result and successful "
           "downloads on some networks, but it usually is still better than giving "
           "false settings here."),
         &hbox, "/core/listen/firewalled",
         _("activate (accept incoming connections)"),
         _("deactivate (don't allow incoming / firewalled)"), TRUE);
    
    table = gtk_table_new(1, 1, FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
    
    item = ui_create_headed_item
        (_("port range (e.g. 5000-5500,6699)"),
         ui_pref_string_entry("/core/listen/range", NULL));
    gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    gtk_widget_show_all(vbox);
    if (expand) *expand = FALSE;
    return vbox;
}

static GtkWidget* ui_general_ip(gboolean* expand)
{
    GtkWidget* vbox;
    GtkWidget* hbox;
    GtkWidget* item;
    GtkWidget* table;
    
    vbox = ui_pref_bool_section
        (_("Force static local ip"), 
         _("If you have a static ip you can specify it here which can "
           "solve problems on some networks. Currently you can only enter "
           "ips in the dot notation."),
         &hbox, "/core/local/force_ip",
         _("disable"), _("enable"), FALSE);
    
    table = gtk_table_new(1, 1, FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
    
    item = ui_create_headed_item
        (_("local ip"), ui_pref_string_entry("/core/local/ip", "xxx.xxx.xxx.xxx"));
    gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    gtk_widget_show_all(vbox);
    
    if (expand) *expand = FALSE;
    return vbox;
}

static void on_ui_combo_apply(const char* name, GtkComboBox* combo)
{
    const char* ui = gtk_combo_box_get_active_text(combo);
    if (!ui || gtk_combo_box_get_active(combo) == 0)
    {
        ui_set_name(NULL, TRUE);
    }
    else
    {
        ui_set_name(ui, TRUE);
    }
    (void)name;
}

static void on_ui_combo_cancel(const char* name, GtkComboBox* combo)
{
    GList* dlist;
    ui_entry_t* current = ui_get_list(&dlist);
    unsigned cnt = 0;
    unsigned index = cnt;

    while (dlist)
    {
        ui_entry_t* entry = dlist->data;
        cnt++;
        if (entry == current) index = cnt;
        dlist = dlist->next;
    }
    gtk_combo_box_set_active(combo, index);
    (void)name;
}

static GtkWidget* _ui_pref_ui_combo()
{
    GList* dlist;
    ui_entry_t* current = ui_get_list(&dlist);
    GtkWidget* combo;
    unsigned cnt = 0;
    unsigned index = cnt;
    char* dup = g_strdup("core:ui");

    combo = gtk_combo_box_new_text();
    gtk_combo_box_append_text
        (GTK_COMBO_BOX(combo), _("Load without any user interface"));
    while (dlist)
    {
        ui_entry_t* entry = dlist->data;
        gtk_combo_box_append_text
            (GTK_COMBO_BOX(combo), _(entry->name));
        dlist = dlist->next;
        cnt++;
        if (entry == current) index = cnt;
    }
    gtk_combo_box_set_active(GTK_COMBO_BOX(combo), index);
    
    g_signal_connect_swapped(G_OBJECT(combo), "destroy",
                             G_CALLBACK(g_free), dup);

    ui_pref_set_funcs(dup, (UIPrefFunc)on_ui_combo_apply,
                      (UIPrefFunc)on_ui_combo_cancel, combo);
    g_signal_connect_swapped(G_OBJECT(combo), "changed",
                             G_CALLBACK(ui_pref_set_dirty), dup);
    
    return combo;
}

static GtkWidget* ui_general_ui(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* table;
  GtkWidget* item;
  
  vbox = ui_create_section
      (_("Graphical user interface"), NULL, &hbox, NULL);
  
  table = gtk_table_new(1, 1, FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
  gtk_table_set_col_spacings(GTK_TABLE(table), 5);
  
  item = ui_create_headed_item
      (_("interface to load"), _ui_pref_ui_combo());
  gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                   GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);

  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox;
}

static GtkWidget* ui_transfer_speed(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* table;
  GtkWidget* item;
  
  vbox = ui_create_section
    (_("Internet connection"),
     _("These settings do not influence your downloads or uploads, they "
       "have informative purpose only, so there's no reason not to be honest here."
       "\n\nLeft-click the bar to set the speed, right-click and drag to scale the bar."
       "\n\nTo limit your transfer speed, check the settings page for transfers."),
     &hbox, NULL);

  table = gtk_table_new(1, 2, FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
  gtk_table_set_col_spacings(GTK_TABLE(table), 5);
  
  item = ui_create_headed_item
    (_("line type"),
     ui_pref_int_combo("/core/line/type", 
                       _("Not sure"),
                       _("Analog modem"),
                       _("ISDN"),
                       _("DSL"),
                       _("Cable"),
                       _("Satellite"),
                       _("Power line"),
                       _("T1"),
                       _("T3+"),
                       NULL));
  gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                   GTK_FILL,GTK_FILL,0,0);

  item = ui_create_headed_item
    (_("maximum upload speed possible"), ui_pref_int_speed("/core/line/speed"));
  gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                   GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);

  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox;
}

static void on_plugin_check_apply(const char* name, GtkToggleButton* check) {
  plugin_t* plugin = g_object_get_data(G_OBJECT(check), "plugin");
  (void)name;
  plugin_known_active(plugin->filename, gtk_toggle_button_get_active(check));
}

static void on_plugin_check_cancel(const char* name, GtkToggleButton* check) {
  plugin_t* plugin = g_object_get_data(G_OBJECT(check), "plugin");
  (void)name;
  if (!plugin_known_as_active(plugin->filename))
    gtk_toggle_button_set_active(check, FALSE);
  else
    gtk_toggle_button_set_active(check, TRUE);
}

static gboolean on_plugin_about(plugin_t* plugin)
{
    const char** authors = NULL;
    unsigned len = g_list_length(plugin->info->authors);
    
    if (len > 0) {
        GList* dlist;
        int i1;
        
        authors = g_malloc0(sizeof(*authors)*(len+1));
        for (i1 = 0, dlist = plugin->info->authors; dlist; dlist = dlist->next, i1++)
            authors[i1] = dlist->data;
    }
    ui_about_show
        (GTK_WINDOW(PrefWin), plugin->info->name, plugin_get_version_string(plugin),
         plugin->info->copyright, plugin->info->license,
         plugin->info->website, authors, _(plugin->info->comment), plugin->info->icon);
    return TRUE;
    
    g_free(authors);
}

static GtkWidget* ui_transfer_download(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* table;
  GtkWidget* item;
  
  vbox = ui_create_section(_("Download limits"), NULL, &hbox, NULL);

  table = gtk_table_new(1, 3, FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
  gtk_table_set_col_spacings(GTK_TABLE(table), 5);
  gtk_table_set_row_spacings(GTK_TABLE(table), 3);
  
  item = ui_create_headed_item
    (_("overall"), ui_pref_int_spin("/core/download/limit", 0, 10000));
  gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                   GTK_FILL,GTK_FILL,0,0);

  item = ui_create_headed_item
    (_("per user"), ui_pref_int_spin("/core/download/limit_per_user", 0, 10000));
  gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                   GTK_FILL,GTK_FILL,0,0);

  item = ui_create_headed_item
    (_("bandwidth"), ui_pref_int_speed("/core/download/bandlimit"));
  gtk_table_attach(GTK_TABLE(table), item, 2, 3, 0, 1,
                   GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);

  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox;
}

static GtkWidget* ui_transfer_upload(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* table;
  GtkWidget* item;
  
  vbox = ui_create_section(_("Upload limits"), NULL, &hbox, NULL);

  table = gtk_table_new(1, 3, FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
  gtk_table_set_col_spacings(GTK_TABLE(table), 5);

  item = ui_create_headed_item
    (_("overall"), ui_pref_int_spin("/core/upload/limit", 0, 10000));
  gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                   GTK_FILL,GTK_FILL,0,0);

  item = ui_create_headed_item
    (_("per user"), ui_pref_int_spin("/core/upload/limit_per_user", 0, 10000));
  gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                   GTK_FILL,GTK_FILL,0,0);

  item = ui_create_headed_item
    (_("bandwidth"), ui_pref_int_speed("/core/upload/bandlimit"));
  gtk_table_attach(GTK_TABLE(table), item, 2, 3, 0, 1,
                   GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);

  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox;
}

static GtkWidget* ui_transfer_incomplete(gboolean* expand)
{
    GtkWidget* vbox;
    GtkWidget* hbox;
    GtkWidget* table;
    GtkWidget* item;
    
    vbox = ui_create_section
        (_("Temporary files"),
         _("The incomplete folder is used for files while they are downloading. "
           "When a download is finished it will be moved to the folder "
           "as specified for the according media type."),
         &hbox, NULL);
    
    table = gtk_table_new(1, 1, FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
    gtk_table_set_row_spacings(GTK_TABLE(table), 3);
    
    item = ui_create_functional_headed_item
        (_("path for incomplete files"),
         ui_pref_string_entry("/core/path/incomplete", _("Location of files while downloading")),
         _("change"), G_CALLBACK(on_browse_folder));
    gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    gtk_widget_show_all(vbox);
    
    if (expand) *expand = FALSE;
    return vbox;
}

static void ui_other_node(pref_node_t* node, GtkBox* box)
{
    if (node->xml_print && node->store_str && node->save)
    {
        const char* name = pref_node_fullname(node);
        GtkWidget* item;
        if (ui_pref_find(name)) return;
        item = ui_create_headed_item(name, ui_pref_string_entry(name, NULL));
        gtk_box_pack_start(box, item, FALSE, FALSE, 0);
    }
}

static GtkWidget* ui_other_list(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* vbox2;

  vbox = ui_create_section
    (_("Unhandled settings"),
     _("These are settings which are not covered by any other "
       "section in this dialog. A separate user interface either "
       "isn't implemented yet or the plugin that would handle them "
       "is currently not loaded. It also can be that some of these "
       "items are outdated and not used by any plugin or the core "
       "application anymore. If you start lopster will all plugins "
       "being active, the outdated items will "
       "be automatically discarded and not saved to disk anymore."),
     &hbox, NULL);

  vbox2 = gtk_vbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), vbox2, TRUE, TRUE, 0);
  
  pref_foreach((ObjectFunc)ui_other_node, vbox2);

  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox;
}

static void on_add_sl(const char* name, GtkBox* vbox) {
  GtkWidget* item;

  if (ui_pref_find(name)) return;
  item = ui_pref_string_list
    (name, NULL, _(string_list_get_label(name)), NULL);
  gtk_box_pack_start(vbox, item, FALSE, FALSE, 0);
}

static GtkWidget* ui_other_string_list(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* vbox2;

  vbox = ui_create_section(_("Unhandled string lists"), NULL, &hbox, NULL);

  vbox2 = gtk_vbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), vbox2, TRUE, TRUE, 0);
  
  string_lists_foreach((StringListFunc)on_add_sl, vbox2);

  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox;
}

static void ui_media_set_dirty()
{
    ui_pref_set_dirty("MediaTypes");
}

static GtkWidget* ui_media_path(gtype_t* type)
{
    GtkWidget* vbox;
    GtkWidget* hbox;
    GtkWidget* item;
    GtkWidget* entry;
    
    vbox = ui_create_section
        (_("Download folder"),
         _("The directory specifies the location where the "
           "downloads are moved to after they are finished."),
         &hbox, NULL);
    
    entry = ui_entry_new(type->type->download?type->type->download:NULL);
    type->download_entry = GTK_ENTRY(entry);
    
    g_signal_connect(G_OBJECT(entry), "changed",
                     G_CALLBACK(ui_media_set_dirty), NULL);
    
    item = ui_create_functional_headed_item
        (_("download path"), entry, _("change"), G_CALLBACK(on_browse_folder));
    
    gtk_box_pack_start(GTK_BOX(hbox), item, TRUE, TRUE, 0);
    
    return vbox;
}

static void ui_appl_remove(gappl_t* gappl) {
  gtk_widget_destroy(gappl->top);
  gappl->parent->appl = g_list_remove(gappl->parent->appl, gappl);
  g_free(gappl);

  ui_media_set_dirty();
}

static void ui_appl_add(gtype_t* parent, appl_t* app)
{
    GtkWidget* table;
    GtkWidget* label;
    GtkWidget* entry;
    GtkWidget* hbox;
    GtkWidget* item;
    gappl_t* gappl;
    
    gappl = g_malloc0(sizeof(gappl_t));
    parent->appl = g_list_append(parent->appl, gappl);
    gappl->parent = parent;
    
    table = gtk_table_new(1, 3, FALSE);
    gtk_box_pack_start(parent->appl_box, table, TRUE, TRUE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
    
    gappl->top = table;
    
    entry = ui_entry_new(NULL);
    if (app && app_get_command(app))
        ui_entry_set_text(GTK_ENTRY(entry), app_get_command(app));
    gtk_widget_set_size_request(entry, 50, -1);
    gappl->app_entry = GTK_ENTRY(entry);
    g_signal_connect(G_OBJECT(entry), "changed",
                     G_CALLBACK(ui_media_set_dirty), NULL);

    item = ui_create_advanced_headed_item(_("command"), entry, &hbox);
    label = ui_create_event_label
        (_("remove"), G_CALLBACK(ui_appl_remove), gappl, TRUE);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

    gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);


    entry = ui_entry_new(NULL);
    if (app && app_get_name(app))
        ui_entry_set_text(GTK_ENTRY(entry), app_get_name(app));
    gtk_widget_set_size_request(entry, 50, -1);
    gappl->des_entry = GTK_ENTRY(entry);
    g_signal_connect(G_OBJECT(entry), "changed",
                     G_CALLBACK(ui_media_set_dirty), NULL);
  
    item = ui_create_headed_item(_("description"), entry);
    gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
  
    item = gtk_check_button_new_with_label(_("multiple items"));
    gappl->flag_check = GTK_TOGGLE_BUTTON(item);
    //  gtk_toggle_button_set_mode(gappl->flag_check, FALSE);
    if (app && app_get_multiple(app))
        gtk_toggle_button_set_active(gappl->flag_check, TRUE);
    g_signal_connect(G_OBJECT(item), "toggled",
                     G_CALLBACK(ui_media_set_dirty), NULL);
  
    item = ui_create_headed_item(_("can open"), item);
    gtk_table_attach(GTK_TABLE(table), item, 2, 3, 0, 1,
                     GTK_FILL,GTK_FILL,0,0);
  
    gtk_widget_show_all(table);
}

static void ui_appl_new(gtype_t* parent)
{
    GList* dlist;
    for (dlist = parent->appl; dlist; dlist = dlist->next)
    {
        gappl_t* gappl = dlist->data;
        const char* text = ui_entry_get_text(gappl->app_entry);
        if (!text || *text == 0) return;
    }
    ui_appl_add(parent, NULL);
}

static GtkWidget* ui_media_appl(gtype_t* gtype, const char* header) {
  GtkWidget* vbox;
  GtkWidget* vbox2;
  GtkWidget* hbox;
  GtkWidget* hhbox;
  GtkWidget* label;
  GList* dlist;

  vbox = ui_create_section
    (header?header:_("Applications"),
     _("The command is parsed like the shell would do it "
       "and every argument '%f' is replaced with the target name. "
       "If there is no '%f', the target is automatically appended "
       "to the command.\n\n"
       "The description is used as label in the context menu.\n\n"
       "The 'can handle' option allows you to control whether you "
       "will be able to select more than one target for open (since "
       "that makes no sense with several commands)."),
     &hbox, &hhbox);

  vbox2 = gtk_vbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), vbox2, TRUE, TRUE, 0);

  gtype->appl_box = GTK_BOX(vbox2);

  label = ui_create_event_label
    (_("add new application"), G_CALLBACK(ui_appl_new), gtype, TRUE);
  gtk_box_pack_end(GTK_BOX(hhbox), label, FALSE, FALSE, 0);

  for (dlist = gtype->type->app; dlist; dlist = dlist->next) {
    appl_t* app = dlist->data;
    ui_appl_add(gtype, app);
  }
  return vbox;
}

static void g_type_destroy(gtype_t* type) {
  g_list_foreach(type->appl, (GFunc)g_free, NULL);
  g_list_free(type->appl);
  g_free(type);
}

static GtkWidget* ui_media_general(int no, const char* label) {
  GtkWidget* vbox0;
  GtkWidget* vbox;
  mtype_t* mtype = mtype_find(no);

  vbox0 = gtk_vbox_new(FALSE, 10);

  if (Types[no]) g_type_destroy(Types[no]);
  Types[no] = g_malloc0(sizeof(gtype_t));
  Types[no]->type = mtype;

  if (no != MEDIA_FOLDER) {
    vbox = ui_media_path(Types[no]);
    gtk_box_pack_start(GTK_BOX(vbox0), vbox, FALSE, FALSE, 0);
  }

  vbox = ui_media_appl(Types[no], label);
  gtk_box_pack_start(GTK_BOX(vbox0), vbox, FALSE, FALSE, 0);

  gtk_widget_show_all(vbox0);

  return vbox0;
}

static GtkWidget* ui_media_unknown(gboolean* expand) {
  if (expand) *expand = FALSE;
  return ui_media_general(MEDIA_UNKNOWN, NULL);
}
static GtkWidget* ui_media_audio(gboolean* expand) {
  if (expand) *expand = FALSE;
  return ui_media_general(MEDIA_AUDIO, NULL);
}
static GtkWidget* ui_media_video(gboolean* expand) {
  if (expand) *expand = FALSE;
  return ui_media_general(MEDIA_VIDEO, NULL);
}
static GtkWidget* ui_media_application(gboolean* expand) {
  if (expand) *expand = FALSE;
  return ui_media_general(MEDIA_APPLICATION, NULL);
}
static GtkWidget* ui_media_image(gboolean* expand) {
  if (expand) *expand = FALSE;
  return ui_media_general(MEDIA_IMAGE, NULL);
}
static GtkWidget* ui_media_document(gboolean* expand) {
  if (expand) *expand = FALSE;
  return ui_media_general(MEDIA_DOCUMENT, NULL);
}



static void ui_browser_changed(GtkComboBox* combo)
{
    int active = gtk_combo_box_get_active(combo);
    GtkEntry* entry = g_object_get_data(G_OBJECT(combo), "entry");
    
    if (active >= 0 && active < UI_BROWSER_NUM)
        ui_entry_set_text(entry, Browser[active]);
}

static GtkWidget* ui_media_browser() 
{
    GtkWidget* vbox;
    GtkWidget* item;
    GtkWidget* hbox;
    GtkWidget* table;
    GtkWidget* combo;
    GtkWidget* entry;
    GtkRequisition req;
    
    vbox = ui_create_section(_("Web browser"), NULL, &hbox, NULL);
    
    table = gtk_table_new(1, 3, FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
    
    entry = ui_pref_string_entry("/core/desktop/browser", NULL);
    gtk_widget_size_request(entry, &req);
    item = ui_create_headed_item(_("browser command"), entry);
    gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    combo = gtk_combo_box_new_text();
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("custom"));
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("Gnome Default"));
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), "Epiphany");
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), "Galeon");
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), "Konqueror");
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), "Mozilla");
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), "Firefox");
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), "Netscape");
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), "Opera");
    gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);
    gtk_widget_set_size_request(combo, 150, req.height);
    g_object_set_data(G_OBJECT(combo), "entry", entry);
    item = ui_create_headed_item(_("predefined"), combo);
    gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                     GTK_FILL,GTK_FILL,0,0);
    
    g_signal_connect_swapped(G_OBJECT(combo), "changed",
                             G_CALLBACK(ui_browser_changed), combo);
    
    gtk_widget_show_all(vbox);
    
    return vbox;
}

static void ui_email_changed(GtkComboBox* combo)
{
    int active = gtk_combo_box_get_active(combo);
    GtkEntry* entry = g_object_get_data(G_OBJECT(combo), "entry");
    
    if (active >= 0 && active < UI_EMAIL_NUM)
        ui_entry_set_text(entry, EMail[active]);
}

static GtkWidget* ui_media_email()
{
    GtkWidget* vbox;
    GtkWidget* item;
    GtkWidget* hbox;
    GtkWidget* table;
    GtkWidget* combo;
    GtkWidget* entry;
    GtkRequisition req;
    
    vbox = ui_create_section(_("E-Mail client"), NULL, &hbox, NULL);
    
    table = gtk_table_new(1, 3, FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
    
    entry = ui_pref_string_entry("/core/desktop/email", NULL);
    gtk_widget_size_request(entry, &req);
    item = ui_create_headed_item(_("email command"), entry);
    gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    combo = gtk_combo_box_new_text();
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("custom"));
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("Evolution Mail Reader"));
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("Mozilla Mail"));
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("Opera Mail"));
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo), _("Mutt"));
    gtk_combo_box_set_active(GTK_COMBO_BOX(combo), 0);
    gtk_widget_set_size_request(combo, 150, req.height);
    g_object_set_data(G_OBJECT(combo), "entry", entry);
    item = ui_create_headed_item(_("predefined"), combo);
    gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                     GTK_FILL,GTK_FILL,0,0);
    
    g_signal_connect_swapped(G_OBJECT(combo), "changed",
                             G_CALLBACK(ui_email_changed), combo);
    
    gtk_widget_show_all(vbox);
    
    return vbox;
}

static GtkWidget* ui_media_rest(gboolean* expand) {
  GtkWidget* vbox0 = ui_media_general(MEDIA_FOLDER, _("Folder applications"));
  GtkWidget* vbox;

  vbox = ui_media_browser();
  gtk_box_pack_start(GTK_BOX(vbox0), vbox, FALSE, FALSE, 0);
  gtk_box_reorder_child(GTK_BOX(vbox0), vbox, 0);

  vbox = ui_media_email();
  gtk_box_pack_start(GTK_BOX(vbox0), vbox, FALSE, FALSE, 0);
  gtk_box_reorder_child(GTK_BOX(vbox0), vbox, 1);

  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox0;
}

static void ui_lib_set_dirty()
{
    ui_pref_set_dirty("/core/lib/current");
}

static void ui_lib_remove_folder(ui_folder_t* folder)
{
    gtk_widget_destroy(folder->top);
    LibFolder = g_list_remove(LibFolder, folder);
    g_free(folder);
    ui_lib_set_dirty();
}

static void ui_lib_change_folder(ui_folder_t* folder)
{
    const char* c = ui_entry_get_text(folder->dir_entry);
    char* name =
        ui_select_folder(_("Change folder"), (c && *c)?c:NULL, GTK_WINDOW(PrefWin));

    if (name)
    {
        ui_entry_set_text(folder->dir_entry, name);
        g_free(name);
    }
}

static void ui_lib_add_folder(GtkBox* box, const char* dir)
{
    GtkWidget* label;
    GtkWidget* hbox;
    GtkWidget* hhbox;
    GtkWidget* item;
    ui_folder_t* f;
    GtkRequisition req;

    f = g_malloc(sizeof(*f));
    LibFolder = g_list_append(LibFolder, f);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox);

    item = ui_entry_new(dir);
    f->dir_entry = GTK_ENTRY(item);
    gtk_box_pack_start(GTK_BOX(hbox), item, TRUE, TRUE, 0);
    gtk_widget_size_request(item, &req);

    item = gtk_button_new_from_stock(GTK_STOCK_REMOVE);
    gtk_box_pack_start(GTK_BOX(hbox), item, FALSE, FALSE, 0);
    gtk_widget_set_size_request(item, -1, req.height);
    g_signal_connect_swapped(G_OBJECT(item), "clicked",
                             G_CALLBACK(ui_lib_remove_folder), f);

    item = ui_create_advanced_headed_item(_("folder"), hbox, &hhbox);
    g_signal_connect(G_OBJECT(f->dir_entry), "changed",
                     G_CALLBACK(ui_lib_set_dirty), NULL);
    gtk_box_pack_start(box, item, FALSE, FALSE, 0);
    f->top = item;

    label = ui_create_event_label
        (_("change"), G_CALLBACK(ui_lib_change_folder), f, TRUE);
    gtk_box_pack_start(GTK_BOX(hhbox), label, FALSE, FALSE, 0);

    gtk_widget_show_all(item);
}

static void ui_lib_new_folder(GtkBox* box)
{
    char* name = ui_select_folder(_("Choose folder"), NULL, GTK_WINDOW(PrefWin));

    if (name)
    {
        GList* dlist;
        
        for (dlist = LibFolder; dlist; dlist = dlist->next)
        {
            ui_folder_t* folder = dlist->data;
            const char* text = ui_entry_get_text(folder->dir_entry);
            if (!text || *text == 0)
            {
                ui_entry_set_text(folder->dir_entry, name);
                break;
            }
        }
        if (!dlist)
        {
            ui_lib_add_folder(box, name);
            ui_lib_set_dirty();
        }
        g_free(name);
    }
}

static void lib_to_folder(pref_node_t* node)
{
    const char* name = pref_node_fullname(node);
    ui_pref_cancel(name);
}

static GtkWidget* ui_library_paths(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* vbox2;
  GtkWidget* hbox;
  GtkWidget* hhbox;
  GtkWidget* label;
  library_t* lib = library_get_current();
  GList* dlist;

  vbox = ui_create_section(_("Shared folders"), NULL, &hbox, &hhbox);

  vbox2 = gtk_vbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox), vbox2, TRUE, TRUE, 0);

  SharedFolderBox = GTK_BOX(vbox2);
  
  label = ui_create_event_label
    (_("add new folder"), G_CALLBACK(ui_lib_new_folder), vbox2, TRUE);
  gtk_box_pack_end(GTK_BOX(hhbox), label, FALSE, FALSE, 0);

  if (lib) {
    for (dlist = lib->shared_folder; dlist; dlist = dlist->next) {
      char* folder = dlist->data;
      ui_lib_add_folder(GTK_BOX(vbox2), folder);
    }
  }

  pref_observe("/core/lib/current", (ObjectFunc)lib_to_folder, NULL);

  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox;
}

static GtkWidget* ui_chat_functional_general(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* table;
  GtkWidget* item;
  
  vbox = ui_create_section(_("Automatically rejoin channels"), NULL, &hbox, NULL);
  
  table = gtk_table_new(1, 1, FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
  gtk_table_set_col_spacings(GTK_TABLE(table), 5);
  gtk_table_set_row_spacings(GTK_TABLE(table), 3);
  
  item = ui_pref_int_combo
    ("/core/chat/auto_rejoin", 
     _("don't rejoin, but leave page open and deactivated"),
     _("if part was not requested (kicked, connection lost, etc.)"),
     NULL);
  gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                   GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
  
  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox;
}

static GtkWidget* ui_chat_pipe_hide(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* table;
  GtkWidget* item;
  
  vbox = ui_create_section(_("Message filter and redirection"),
                           NULL, &hbox, NULL);

  table = gtk_table_new(1, 2, FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
  gtk_table_set_col_spacings(GTK_TABLE(table), 5);
  gtk_table_set_row_spacings(GTK_TABLE(table), 3);
  
  item = 
    ui_pref_int_flags("/core/chat/hide", FALSE, POS_LEFT_DOWN,
                      5, _("Do not show any ..."), 0,
                      0, _("server messages"), HIDE_SERVER,
                      0, _("global messages"), HIDE_GLOBAL,
                      0, _("channel parts"), HIDE_PARTS,
                      0, _("channel joins"), HIDE_JOINS,
                      0, _("messages of the day"), HIDE_MOTD);
  gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                   GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);

  item =
    ui_pref_int_flags("/core/chat/pipe", FALSE, POS_LEFT_DOWN,
                      4, _("Separate window for ..."), 0,
                      0, _("private messages"), PIPE_PRIVATE,
                      0, _("global messages"), PIPE_GLOBAL,
                      0, _("operator messages"), PIPE_OPERATOR,
                      0, _("channel operator messages"), PIPE_CHAN_OP);
  gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                   GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);

  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox;
}

static GtkWidget* ui_user_away(gboolean* expand)
{
    GtkWidget* vbox;
    GtkWidget* item;
    GtkWidget* hbox;
    
    vbox = ui_pref_bool_section
        (_("Away status"), NULL, &hbox, "/core/user/away",
         _("deactivate away mode"), _("activate away mode"), FALSE);
    
    item = ui_pref_string_list
        ("core:user:away", "/core/user/away_message",
         _("away message"), _(string_list_get_label("core:user:away")));
    gtk_box_pack_start(GTK_BOX(hbox), item, TRUE, TRUE, 0);
    
    gtk_widget_show_all(vbox);
    
    if (expand) *expand = TRUE;
    return vbox;
}

static GtkWidget* ui_http_proxy(gboolean* expand)
{
    GtkWidget* vbox;
    GtkWidget* hbox;
    GtkWidget* item;
    GtkWidget* table;
    
    vbox = ui_pref_bool_section
        (_("Http proxy"), NULL, &hbox, "/core/http/proxy/enabled",
         _("disable"), _("enable"), FALSE);
    
    table = gtk_table_new(1, 2, FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
    
    item = ui_create_headed_item
        (_("host address"), ui_pref_string_entry("/core/http/proxy/host", NULL));
    gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    item = ui_create_headed_item
        (_("port"), ui_pref_int_spin("/core/http/proxy/port", 1, 65535));
    gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                     GTK_FILL,GTK_FILL|GTK_EXPAND,0,0);
    
    gtk_widget_show_all(vbox);
    
    if (expand) *expand = FALSE;
    return vbox;
}

static GtkWidget* ui_http_control(gboolean* expand)
{
    GtkWidget* vbox;
    GtkWidget* hbox;
    GtkWidget* item;
    GtkWidget* table;
    
    vbox = ui_pref_bool_section
        (_("Remote control via web browser"), 
         _("Leave the login or password field blank to use the "
           "default settings from the core. The resource files "
           "are used to build the web sites, they usually include "
           "a style sheets and javascript file."),
         &hbox, "/core/web/active", _("disable"), _("enable"), FALSE);
    
    table = gtk_table_new(1, 2, FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
    gtk_table_set_row_spacings(GTK_TABLE(table), 3);
    
    item = ui_create_headed_item
        (_("login"),
         ui_pref_string_entry("/core/web/username", _("Using default username")));
    gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    item = ui_create_headed_item
        (_("password"),
         ui_pref_string_entry_inv("/core/web/password", _("Using default password")));
    gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    item = ui_create_headed_item
        (_("port"), ui_pref_int_spin("/core/web/port", 1, 65535));
    gtk_table_attach(GTK_TABLE(table), item, 2, 3, 0, 1,
                     GTK_FILL,GTK_FILL|GTK_EXPAND,0,0);
    
    
    item = ui_create_functional_headed_item
        (_("location of resource files"),
         ui_pref_string_entry("/core/web/resource_path", NULL),
         _("change"), G_CALLBACK(on_browse_folder));
    gtk_table_attach(GTK_TABLE(table), item, 0, 3, 1, 2,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    gtk_widget_show_all(vbox);
    
    if (expand) *expand = FALSE;
    return vbox;
}

static GtkWidget* ui_chat_appearance_general(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* item;
  GtkWidget* table;
  
  vbox = ui_create_section(_("General"), NULL, &hbox, NULL);

  table = gtk_table_new(1, 2, FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
  gtk_table_set_col_spacings(GTK_TABLE(table), 5);
  gtk_table_set_row_spacings(GTK_TABLE(table), 3);
  
  item = ui_create_headed_item
    (_("format of the timestamps"),
     ui_pref_int_combo("/core/chat/timestamp", 
                       _("no timestamps"),
                       _("hour:minute"),
                       _("hour:minute:second"),
                       NULL));
  gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                   GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
  
  item = ui_create_headed_item
    (_("colored nicks"),
     ui_pref_int_combo("/core/chat/colored_nicks", 
                       _("default color"),
                       _("dark colors"),
                       _("medium colors"),
                       _("light colors"), NULL));
  gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                   GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);

  gtk_widget_show_all(vbox);

  if (expand) *expand = FALSE;
  return vbox;
}

static GtkWidget* ui_chat_prefix(gboolean* expand)
{
    GtkWidget* vbox;
    GtkWidget* hbox;
    GtkWidget* item;
    GtkWidget* table;
    
    vbox = ui_create_section
        (_("Prefixes for all kinds of messages"),
         _("These are prefixes for all kinds of message that can "
           "appear in a chat page. The style of these prefixes can "
           "be set up in the styles dialog."),
         &hbox, NULL);
    
    table = gtk_table_new(1, 5, FALSE);
    gtk_box_pack_start(GTK_BOX(hbox), table, TRUE, TRUE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
    gtk_table_set_row_spacings(GTK_TABLE(table), 3);
    
    item = ui_create_headed_item
        (_("server"), ui_pref_string_entry("/core/chat/prefix/server", NULL));
    gtk_table_attach(GTK_TABLE(table), item, 0, 1, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    item = ui_create_headed_item
        (_("client"), ui_pref_string_entry("/core/chat/prefix/client", NULL));
    gtk_table_attach(GTK_TABLE(table), item, 1, 2, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    item = ui_create_headed_item
        (_("join"), ui_pref_string_entry("/core/chat/prefix/join", NULL));
    gtk_table_attach(GTK_TABLE(table), item, 2, 3, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    item = ui_create_headed_item
        (_("part"), ui_pref_string_entry("/core/chat/prefix/part", NULL));
    gtk_table_attach(GTK_TABLE(table), item, 3, 4, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    item = ui_create_headed_item
        (_("other"), ui_pref_string_entry("/core/chat/prefix/text", NULL));
    gtk_table_attach(GTK_TABLE(table), item, 4, 5, 0, 1,
                     GTK_FILL|GTK_EXPAND,GTK_FILL|GTK_EXPAND,0,0);
    
    gtk_widget_show_all(vbox);
    
    if (expand) *expand = FALSE;
    return vbox;
}

static GtkWidget* ui_chat_hilight(gboolean* expand) {
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* item;
  
  vbox = ui_create_section(_("Chat highlightning"), NULL, &hbox, NULL);

  item = ui_pref_string_list
    ("core:chat:hilight", NULL,
     _("highlight lines that contain one of these terms"), NULL);
  gtk_box_pack_start(GTK_BOX(hbox), item, TRUE, TRUE, 0);
  
  gtk_widget_show_all(vbox);

  if (expand) *expand = TRUE;
  return vbox;
}


static ui_pref_page_t* 
ui_pref_page_new(const char* label, gboolean scrolled, const char* des) {
  ui_pref_page_t* node;

  g_assert(!PrefWin);
  
  node = g_malloc0(sizeof(*node));
  node->label = g_strdup(label);
  node->des = g_strdup(des);
  node->scrolled = scrolled;
  node->children = NULL;
  node->sections = NULL;
  node->page = NULL;
  node->create = NULL;
  return node;
}

static GtkContainer* ui_pref_create_leaf(ui_pref_page_t* node) {
  GList* dlist;
  GtkWidget* vbox0;

  vbox0 = gtk_vbox_new(FALSE, 15);

  for (dlist = node->sections; dlist; dlist = dlist->next) {
    CreateSectionFunc func = dlist->data;
    gboolean expand = FALSE;
    GtkWidget* item = func(&expand);

    gtk_box_pack_start(GTK_BOX(vbox0), item, expand, expand, 0);
    gtk_widget_show(item);
  }

  gtk_widget_show(vbox0);

  return GTK_CONTAINER(vbox0);
}

/** Create a leaf page. Use ui_pref_register_page() if the page should
 * have subpages.
 * @param label a short description
 * @param scrolled whether the page should be able to scroll vertically
 * @param des a longer description of the page.
 * @param first_func the first section.
 * @param ... a NULL terminated list of additional section fucns
 * @return a new preferences page
 */
ui_pref_page_t* ui_pref_register_leaf(const char* label, gboolean scrolled, const char* des,
                                      CreateSectionFunc first_func, ...) {
  ui_pref_page_t* node;
  CreateSectionFunc create = first_func;
  va_list ap;

  node = ui_pref_page_new(label, scrolled, des);
  node->create = ui_pref_create_leaf;

  va_start(ap, first_func);
  while (create) {
    node->sections = g_list_append(node->sections, create);
    create = va_arg(ap, CreateSectionFunc);
  }
  va_end(ap);

  return node;
}

static GtkContainer* ui_pref_create_page(ui_pref_page_t* node) {
  GList* dlist;
  GtkWidget* vbox0;
  GtkWidget* vbox;
  GtkWidget* hbox;
  GtkWidget* hhbox;
  GtkWidget* label;

  vbox0 = gtk_vbox_new(FALSE, 10);

  for (dlist = node->children; dlist; dlist = dlist->next) {
    ui_pref_page_t* child = dlist->data;

    vbox = ui_create_section(child->label, NULL, &hbox, &hhbox);
    gtk_widget_show(vbox);
    label = ui_create_event_label
      (_("enter page"), G_CALLBACK(ui_pref_activate_node), child, TRUE);
    gtk_box_pack_end(GTK_BOX(hhbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);

    if (child->des) {
      label = gtk_label_new(child->des);
      gtk_widget_show(label);
      gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
      gtk_label_set_selectable(GTK_LABEL(label), TRUE);

      gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);
      gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
    } else {
      gtk_widget_destroy(hbox);
    }

    gtk_box_pack_start(GTK_BOX(vbox0), vbox, FALSE, FALSE, 0);
  }

  gtk_widget_show(vbox0);

  return GTK_CONTAINER(vbox0);
}

/** Create a preferences page. Use this function to create pages that
 * only contains subpages.
 * @param label the short description fo the page.
 * @param scrolled whether that page should have a vertical scrollbar
 * @param des a longer description of the page, or NULL
 * @param first_child the first subpage
 * @param ... a list of additional subpages terminated by NULL
 * @return a new preferences page
 * @see ui_pref_register_leaf()
 */
ui_pref_page_t* ui_pref_register_page(const char* label, gboolean scrolled, const char* des,
                                      ui_pref_page_t* first_child, ...) {
  ui_pref_page_t* node;
  ui_pref_page_t* sub = first_child;
  va_list ap;

  node = ui_pref_page_new(label, scrolled, des);
  node->create = ui_pref_create_page;

  va_start(ap, first_child);
  while (sub) {
    node->children = g_list_append(node->children, sub);
    sub->parent = node;
    sub = va_arg(ap, ui_pref_page_t*);
  }
  va_end(ap);

  return node;
}

static ui_pref_page_t* ui_pref_find_plugin(ui_pref_page_t* page, plugin_t* plugin) {
  GList* dlist;
  for (dlist = page->children; dlist; dlist = dlist->next) {
    ui_pref_page_t* node = dlist->data;
    if (node->plugin == plugin) return node;
  }
  return NULL;
}

static GtkContainer* ui_pref_create_plugin(ui_pref_page_t* page) {
  GtkWidget* vbox0;
  GList* dlist;
  GtkWidget* item;

  vbox0 = gtk_vbox_new(FALSE, 10);

  item = gtk_label_new
    (_("<b>Note:</b> De/activating plugins requires a restart of the program."));
  gtk_label_set_use_markup(GTK_LABEL(item), TRUE);
  gtk_misc_set_alignment(GTK_MISC(item), 0, 0.5);
  gtk_box_pack_start(GTK_BOX(vbox0), item, FALSE, FALSE, 0);

  item = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(vbox0), item, FALSE, FALSE, 0);

  for (dlist = plugin_get_list(); dlist; dlist = dlist->next) {
    plugin_t* plugin = dlist->data;
    GtkWidget* vbox;
    GtkWidget* hbox;
    GtkWidget* label;
    GtkWidget* check;
    ui_pref_page_t* child;

    vbox = gtk_vbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(vbox0), vbox, FALSE, FALSE, 0);

    // the header
    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

    item = gtk_image_new_from_stock
      (ui_main_get_plugin_stock_from_plugin(plugin), GTK_ICON_SIZE_MENU);
    gtk_box_pack_start(GTK_BOX(hbox), item, FALSE, FALSE, 0);

    check = gtk_check_button_new();
    if (plugin_known_as_active(plugin->filename))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), TRUE);
    gtk_box_pack_start(GTK_BOX(hbox), check, TRUE, TRUE, 0);
    
    
    label = gtk_label_new(l_printf("<b>%s</b> (%s)", plugin->info->name,
                                   plugin_status_string(plugin)));
    gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
    gtk_container_add(GTK_CONTAINER(check), label);
    if (plugin->status != PLUGIN_ACTIVE && plugin->status != PLUGIN_INACTIVE)
      gtk_widget_set_sensitive(check, FALSE);

    child = ui_pref_find_plugin(page, plugin);
    label = ui_create_event_label
      (_("configure"), G_CALLBACK(ui_pref_activate_node), child, TRUE);
    gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    if (!child) ui_event_label_set_sensitive(label, FALSE);

    label = ui_create_event_label
      (_("about"), G_CALLBACK(on_plugin_about), plugin, TRUE);
    gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);

    // the filename
    label = gtk_label_new(l_printf("%s", plugin->filename));
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
    gtk_misc_set_padding(GTK_MISC(label), 20, 0);
    gtk_label_set_selectable(GTK_LABEL(label), TRUE);
    gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
    
    /*
    if (plugin->status != PLUGIN_INACTIVE &&
	plugin->status != PLUGIN_DUPLICATE)
      gtk_widget_set_sensitive(check, FALSE);
    */
    g_object_set_data(G_OBJECT(check), "plugin", plugin);
    ui_pref_set_funcs(plugin->info->name, (UIPrefFunc)on_plugin_check_apply,
                      (UIPrefFunc)on_plugin_check_cancel, check);
    g_signal_connect_swapped(G_OBJECT(check), "toggled",
                             G_CALLBACK(ui_pref_set_dirty), plugin->info->name);
  }

  gtk_widget_show_all(vbox0);

  return GTK_CONTAINER(vbox0);
}

static ui_pref_page_t*
ui_register_plugin_page(const char* label, gboolean scrolled, const char* des) {
  ui_pref_page_t* node;

  PluginPage = node = ui_pref_page_new(label, scrolled, des);
  node->create = ui_pref_create_plugin;

  return node;
}

/** Add the preferences page of a plugin to the dialog.
 * @param ID the ID of the plugin
 * @param page the preferences page
 * @see ui_pref_register_page(), ui_pref_register_leaf()
 */
void ui_pref_hook_plugin(int ID, ui_pref_page_t* page) {
  plugin_t* plugin = plugin_get(ID);
  page->plugin = plugin;
  PluginPage->children = g_list_append(PluginPage->children, page);
  page->parent = PluginPage;
}


void ui_pref_init() {
  PrefFuncs = g_hash_table_new_full
    ((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
     (GDestroyNotify)NULL, (GDestroyNotify)ui_pref_destroy);

  Preferences =
      ui_pref_register_page
      (    _("Preferences"), FALSE, NULL,
           ui_pref_register_page
           (    _("Filesharing"), FALSE, _("Basic settings to make lopster ready for filesharing."),
                ui_pref_register_leaf
                (    _("My identity"), FALSE, _("Give some information about you and your connection"),
                     ui_general_user,
                     ui_transfer_speed,
                     ui_general_ip,
                     NULL),
                ui_pref_register_leaf
                (    _("Shared files"), TRUE, _("Add/remove folders you like to share"),
                     ui_library_paths,
                     NULL),
                ui_pref_register_leaf
                (    _("Transfers"), FALSE, _("Settings for both the downloads and the uploads"),
                     ui_general_ports,
                     ui_transfer_download,
                     ui_transfer_upload,
                     ui_transfer_incomplete,
                     NULL),
                NULL),
           ui_pref_register_page
           (    _("Client"), FALSE, _("Modify the settings to get comfortable with this program"),
                ui_pref_register_leaf
                (    _("Basic client setup"), TRUE, _("Basic settings for lopster"),
                     ui_general_ui,
                     ui_http_control,
                     ui_http_proxy,
                     NULL),
                ui_pref_register_page
                (    _("Paths/Applications"), FALSE, _("Settings for download paths and applications to use"),
                     ui_pref_register_leaf
                     (    _("Unknown files"), TRUE, _("Files that don't belong to another media type"),
                          ui_media_unknown,
                          NULL),
                     ui_pref_register_leaf
                     (    _("Audio files"), TRUE, NULL, ui_media_audio, NULL),
                     ui_pref_register_leaf
                     (    _("Video files"), TRUE, NULL, ui_media_video, NULL),
                     ui_pref_register_leaf
                     (    _("Application files"), TRUE, NULL, ui_media_application, NULL),
                     ui_pref_register_leaf
                     (    _("Image files"), TRUE, NULL, ui_media_image, NULL),
                     ui_pref_register_leaf
                     (    _("Text files"), TRUE, NULL, ui_media_document, NULL),
                     ui_pref_register_leaf
                     (    _("Browser/Email/Folders"), TRUE, NULL, ui_media_rest, NULL),
                     NULL),
                ui_pref_register_leaf
                (    _("Chat functionality"), FALSE, _("Change functional chat settings here"),
                     ui_chat_functional_general,
                     ui_chat_pipe_hide,
                     ui_user_away,
                     NULL),
                ui_pref_register_leaf
                (    _("Chat appearance"), FALSE, _("Change visual chat settings here"),
                     ui_chat_appearance_general,
                     ui_chat_prefix,
                     ui_chat_hilight,
                     NULL),
                NULL),
           ui_register_plugin_page
           (    _("Plugins"), TRUE, _("Settings for the plugins")),
           ui_pref_register_leaf
           (    _("Other"), TRUE, _("Settings which have no interface"),
                ui_other_list,
                ui_other_string_list,
                NULL),
           NULL);

}

static void on_media_apply()
{
    int i1;
    const char* text;
    const char* text2;
    appl_t* app;
    GList* dlist;
    
    // FIXME: speedup
    mtype_clear_all();
    
    for (i1 = 0; i1 < MEDIA_ANY; i1++)
    {
        gtype_t* type = Types[i1];
        mtype_t* mtype = type->type;
        
        if (type->download_entry)
        {
            text = ui_entry_get_text(type->download_entry);
            if (text && *text) mtype->download = g_strdup(text);
        }
    
        for (dlist = type->appl; dlist; dlist = dlist->next)
        {
            gappl_t* gapp = dlist->data;
            
            text = ui_entry_get_text(gapp->app_entry);
            if (!text || *text == 0) continue;
            text2 = ui_entry_get_text(gapp->des_entry);
            
            app = app_new(text, text2);
            app_set_multiple(app, gtk_toggle_button_get_active(gapp->flag_check));
            mtype->app = g_list_append(mtype->app, app);
        }
    }
    mtype_save();
}

static void on_media_cancel()
{
    unsigned i1;
    
    // FIXME: speedup
    for (i1 = 0; i1 < MEDIA_ANY; i1++)
    {
        gtype_t* gtype = Types[i1];
        mtype_t* type = gtype->type;
        GList* dlist;
        
        // restore download path
        if (gtype->download_entry)
            ui_entry_set_text(gtype->download_entry,
                              type->download?type->download:"");
    // restore applications
        while (gtype->appl) ui_appl_remove(gtype->appl->data);
        for (dlist = type->app; dlist; dlist = dlist->next)
        {
            appl_t* app = dlist->data;
            ui_appl_add(gtype, app);
        }
    }
}

static void on_library_apply()
{
    library_t* lib = library_get_current();
    GList* dlist;
    
    // FIXME: speedup
    if (!lib) return;
    library_clear_shared(lib);
    
    for (dlist = LibFolder; dlist; dlist = dlist->next)
    {
        ui_folder_t* folder = dlist->data;
        const char* text = ui_entry_get_text(folder->dir_entry);
        
        if (!text || *text == 0) continue;
        
        if (!directory_exists(text))
        {
            lopster_notify(LOPSTER_MESSAGE_WARNING, _("Directory does not exist"),
                           l_printf(_("The directory <b>%s</b> does not exist and "
                                      "therefore cannot be used for sharing."), text));
            continue;
        }
        
        library_add_shared(lib, text);
    }
    
    library_refresh(TRUE);
}

static void on_library_cancel() {
  library_t* lib = library_get_current();
  GList* dlist;

  if (!SharedFolderBox) return;
  // FIXME: speedup
  while (LibFolder) ui_lib_remove_folder(LibFolder->data);

  if (lib) {
    for (dlist = lib->shared_folder; dlist; dlist = dlist->next) {
      char* folder = dlist->data;
      ui_lib_add_folder(SharedFolderBox, folder);
    }
  }
}

static void on_register_temp_lists(const char* name) {
  char* name2 = l_printf("temp:%s", name);
  const char* label = string_list_get_label(name);
  string_list_register(name2, label?label:name, string_list_get_allow_ws(name));
  string_list_set_save(name2, FALSE);
}

void ui_pref_post_init() {
  int i1;

  for (i1 = -1; i1 < plugin_get_number(); i1++) {
    plugin_t* plugin = plugin_get(i1);
    char* pref = l_printf("/version/%s", plugin_get_short_name(plugin));
    ui_pref_set_funcs(pref, NULL, NULL, NULL);
  }

  ui_pref_set_funcs("core:known:plugins", NULL, NULL, NULL);
  ui_pref_set_funcs("/core/lib/mark_val", NULL, NULL, NULL);

  string_lists_foreach((StringListFunc)on_register_temp_lists, NULL);
  
  // some special apply function which set a complex structure
  // instead of simple pref items
  ui_pref_set_funcs("MediaTypes", (UIPrefFunc)on_media_apply,
                    (UIPrefFunc)on_media_cancel, NULL);
  
  ui_pref_set_funcs("/core/lib/current", (UIPrefFunc)on_library_apply,
                    (UIPrefFunc)on_library_cancel, NULL);
}

#ifdef WIN32
#define _UI_PREFERENCES_C_
#include "lgtk2corealiasdef.c"
#endif
