/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.24 $
 * $Date: 2009/02/01 13:52:41 $
 * $Author: sgop $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <gtk/gtk.h>

#include "ui_main.h"
#include "ui_search.h"
#include "ui_support.h"
#include "ui_layout.h"
#include "ui_treeview.h"
#include "ui_tooltips.h"
#include "lcore/attribute.h"
#include "lcore/plugin.h"
#include "lcore/utils.h"
#include "lcore/search.h"
#include "lcore/files.h"
#include "lcore/mtypes.h"
#include "lcore/search.h"
#include "lcore/download.h"
#include "lcore/l_base.h"
#include "lcore/l_i18n.h"

#ifdef WIN32
#include "lgtk2corealias.h"
#endif

typedef struct {
  GtkWidget* main;
  GtkTreeStore* store;
  GtkTreeView* view;
  GtkTreeViewColumn* col[4];
  GtkWidget* result_label;
  GtkWidget** source_label;
  int updater;
} search_gui_t;

typedef struct {
  char* name;
} search_filter_t;

static GtkBox* LeftContent = NULL;
static GtkEntry* IncludeEntry = NULL;
static GtkEntry* ExcludeEntry = NULL;
static GtkEntry* MinSizeEntry = NULL;
static GtkEntry* MaxSizeEntry = NULL;
static GtkEntry* MinBitrateEntry = NULL;
static GtkEntry* MaxBitrateEntry = NULL;
static GtkEntry* MinFrequencyEntry = NULL;
static GtkEntry* MaxFrequencyEntry = NULL;
static GtkEntry* MinDurationEntry = NULL;
static GtkEntry* MaxDurationEntry = NULL;
static GtkListStore* SearchModel = NULL;
static GtkNotebook* NoteBook = NULL;
static GtkLabel* PatternLabel = NULL;
static GtkWidget* PatternButton = NULL;
static search_pattern_t* CurPattern = NULL;
static GtkTable* SumTable = NULL;
static GtkBox* SumBox = NULL;
static GtkImage* MediaImage = NULL;

static GList* OptionFuncs = NULL;

static GList* Searches = NULL;
static GList* Patterns = NULL;

static int Updater = 0;

typedef struct {
  char* name;
  char* text;
  GetSetOptionFunc get;
  GetSetOptionFunc set;
  GtkWidget* expander;
  GtkWidget* label;
} search_option_t;


static GtkListStore* ui_search_store_get(void) {
  if (!SearchModel) {
    SearchModel = gtk_list_store_new
      (3,
       G_TYPE_POINTER,  // data pointer
       G_TYPE_STRING,
       G_TYPE_INT);
  }
  return SearchModel;
}

static search_option_t* ui_search_find_option(const char* name) {
  GList* dlist;
  search_option_t* option;

  for (dlist = OptionFuncs; dlist; dlist = dlist->next) {
    option = dlist->data;
    if (!strcmp(option->name, name)) return option;
  }
  return NULL;
}

static GtkWidget*
add_search_option(const char* name, GtkWidget* widget, search_option_t* option) {
  GtkWidget* label;
  GtkWidget* exp;

  exp = gtk_expander_new(name);
  gtk_widget_show(exp);
  if (option) option->expander = exp;

  gtk_widget_show(widget);
  gtk_container_add(GTK_CONTAINER(exp), widget);

  label = gtk_label_new(name);
  gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
  gtk_widget_show(label);
  gtk_expander_set_label_widget(GTK_EXPANDER(exp), label);

  gtk_box_pack_start(LeftContent, exp, FALSE, TRUE, 0);

  ui_layout_register_exp
    (l_printf("core:search:option:%s:exp", option?option->name:name));
  ui_layout_monitor
    (l_printf("core:search:option:%s:exp", option?option->name:name),
     exp);
		      
  return label;
}
		      
/** Function to add search fields to the side bar in the search page.
 */
void ui_search_register_option(const char* name, const char* label,
                               CreateOptionUIFunc create,
                               GetSetOptionFunc get,
                               GetSetOptionFunc set) {
  search_option_t* option;

  option = ui_search_find_option(name);
  if (option) {
    g_log("GTK_SEARCH", G_LOG_LEVEL_WARNING,
          "search option '%s' already registered", name);
    return;
  }

  g_log("GTK_SEARCH", G_LOG_LEVEL_DEBUG,
        "registering search option '%s'", name);
  option = g_malloc0(sizeof(*option));
  option->name = g_strdup(name);
  option->text = g_strdup(label);
  option->get = get;
  option->set = set;
  option->expander = NULL;

  OptionFuncs = g_list_append(OptionFuncs, option);

  if (create) {
    option->label = add_search_option(label, create(), option);
  } else {
    option->label = NULL;
  }
}
			       
static void ui_search_set_current_pattern(search_pattern_t* pattern, int media) {
  if (media == MEDIA_UNKNOWN) media = MEDIA_ANY;
  if (pattern) {
    gtk_label_set_text(PatternLabel, OBJ(pattern)->name);
  } else {
    gtk_label_set_text
      (PatternLabel, l_printf(_("Search for '%s'"), _(int2media(media, 0))));
  }
  CurPattern = pattern;
  gtk_widget_set_sensitive(PatternButton, (pattern!=NULL));
}

static search_pattern_t* ui_search_get_temp_pattern(void) {
  static search_pattern_t* TempPattern = NULL;

  if (!TempPattern) {
    TempPattern = search_pattern_new();
    object_set_name(OBJ(TempPattern), "TempPattern");
  }
  g_slist_foreach(TempPattern->include, (GFunc)g_free, NULL);
  g_slist_free(TempPattern->include);
  TempPattern->include = NULL;
  g_slist_foreach(TempPattern->exclude, (GFunc)g_free, NULL);
  g_slist_free(TempPattern->exclude);
  TempPattern->exclude = NULL;
  attr_container_empty(ATTRC(TempPattern));
  
  return TempPattern;
}

static search_pattern_t* ui_search_pattern_get(void) {
  GList* dlist;
  search_pattern_t* pattern = NULL;
  search_option_t* option;
  
  pattern = search_pattern_new();
  
  for (dlist = OptionFuncs; dlist; dlist = dlist->next) {
    option = dlist->data;
    option->get(pattern);
  }
  if (pattern->include) {
    char* str = make_string_from_list(pattern->include, " ");
    object_set_name(OBJ(pattern), str);
    g_free(str);
  } else {
    object_set_name(OBJ(pattern), "-");
  }
  return pattern;
}

static void ui_pattern_update_summary() {
  search_pattern_t* search = ui_search_pattern_get();
  GtkWidget* table;
  int line = 0;

  if (SumTable) {
    gtk_widget_destroy(GTK_WIDGET(SumTable));
    SumTable = NULL;
  }

  if (search) {
    table = gtk_table_new(1, 2, FALSE);
    SumTable = GTK_TABLE(table);
    gtk_box_pack_start(SumBox, table, FALSE, FALSE, 0);
    gtk_table_set_col_spacings(GTK_TABLE(table), 10);

    if (search->exclude) {
      ui_table_add_line(GTK_TABLE(table), &line, _("Exclude"),
                        make_string_from_list(search->exclude, " "));
    }

    ui_table_add_attrc(GTK_TABLE(table), &line, ATTRC(search), "pattern");
    if (line > 0) gtk_widget_show(table);

    object_kill(OBJ(search));
  }

}

static gboolean ui_patterns_check(void) {
  GList* dlist;
  search_option_t* option;
  search_pattern_t* orig;
  int result;
  char* text;
  search_pattern_t* pattern = ui_search_get_temp_pattern();
  
  for (dlist = OptionFuncs; dlist; dlist = dlist->next) {
    option = dlist->data;
    result = option->get(pattern);
    if (!option->label) continue;
    if (result == 1) {
      text = g_strdup_printf("<b>%s</b>", _(option->text));
      gtk_label_set_markup(GTK_LABEL(option->label), text);
      g_free(text);
    } else if (result == 0) {
      gtk_label_set_markup(GTK_LABEL(option->label), _(option->text));
    }
  }
  orig = search_pattern_find(pattern);
  ui_search_set_current_pattern(orig, pattern->media);
  
  ui_pattern_update_summary();
  
  Updater = 0;
  return FALSE;
}

static gboolean on_input_expose(void) {
  if (!Updater)
    Updater = g_timeout_add(500, (GSourceFunc)ui_patterns_check, NULL);
  return FALSE;
}

void ui_search_input_monitor(GtkWidget* widget) {
  g_signal_connect(G_OBJECT(widget), "expose-event",
                   G_CALLBACK(on_input_expose), NULL);
}

static int ui_search_option_set_term(search_pattern_t* pattern) {
  (void)pattern;
  //  if (pattern)
  //    gtk_entry_set_text(IncludeEntry, pattern->term?pattern->term:"");
  return 1;
}
static int ui_search_option_get_term(search_pattern_t* pattern) {
  const char* text;

  if (!pattern) return 1; 
  
  g_slist_foreach(pattern->include, (GFunc)g_free, NULL);
  g_slist_free(pattern->include);
  pattern->include = NULL;
  g_slist_foreach(pattern->exclude, (GFunc)g_free, NULL);
  g_slist_free(pattern->exclude);
  pattern->exclude = NULL;

  text = gtk_entry_get_text(IncludeEntry);
  if (text && *text) {
    search_string_tokenize(l_strdup(text), &pattern->include, &pattern->exclude);
    return 1;
  } else {
    return 0;
  }
}

static void ui_search_set_current_media(int media)
{
    int current = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(MediaImage), "media"));
    if (media == current) return;
    g_object_set_data(G_OBJECT(MediaImage), "media", GINT_TO_POINTER(media));
    gtk_image_set_from_stock(MediaImage, MediaStock[media],
                             GTK_ICON_SIZE_SMALL_TOOLBAR); 
    on_input_expose();
}

static int ui_search_option_set_realm(search_pattern_t* pattern) {
  if (!pattern) return 1;
  ui_search_set_current_media(pattern->media);
  return 1;
}

static int ui_search_option_get_realm(search_pattern_t* pattern) {
  if (!pattern) return 1;
  pattern->media = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(MediaImage), "media"));
  return 1;
}

static GtkWidget* ui_search_option_create_exclude(void) {
  GtkWidget* entry;

  entry = gtk_entry_new();
  ExcludeEntry = GTK_ENTRY(entry);
  ui_search_input_monitor(entry);
  
  //  gtk_widget_show(entry);

  return entry;
}
static int
ui_search_option_set_exclude(search_pattern_t* pattern) {
  if (!pattern) return 1;
  if (pattern->exclude) {
    char* str = make_string_from_list(pattern->exclude, " ");
    gtk_entry_set_text(ExcludeEntry, str);
    g_free(str);
    return 1;
  } else {
    gtk_entry_set_text(ExcludeEntry, "");
    return 0;
  }
}
static int
ui_search_option_get_exclude(search_pattern_t* pattern) {
  const char* text;

  if (!pattern) return 1; 
  text = gtk_entry_get_text(ExcludeEntry);
  if (text && *text) {
    search_string_tokenize(l_strdup(text), &pattern->exclude, NULL);
    return 1;
  } else {
    return 0;
  }
}

static GtkWidget* ui_search_option_create_size(void) {
  GtkWidget* hbox;
  GtkWidget* entry;
  GtkWidget* label;

  hbox = gtk_hbox_new (FALSE, 5);
  //  gtk_widget_show (hbox);
  entry = gtk_entry_new();
  MinSizeEntry = GTK_ENTRY(entry);
  gtk_widget_show(entry);
  ui_search_input_monitor(entry);
  gtk_widget_set_size_request(entry, 0, -1);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  label = gtk_label_new(_("to"));
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);

  entry = gtk_entry_new();
  MaxSizeEntry = GTK_ENTRY(entry);
  gtk_widget_show(entry);
  ui_search_input_monitor(entry);
  gtk_widget_set_size_request(entry, 0, -1);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  return hbox;
}

static int
ui_search_option_set_size(search_pattern_t* pattern) {
  attr_container_t* ac;
  attribute_t* attr;

  if (!pattern) return 1; 
  ac = ATTRC(pattern);
  attr = attr_find(ac, "pattern", "Size");
  if (attr) {
    gtk_entry_set_text(MinSizeEntry,
		       attr_get_refi(attr, file_size_t, 0)?
		       print_bytes_exact(attr_get_ptr(attr, file_size_t)):"");
    gtk_entry_set_text(MaxSizeEntry,
		       attr_get_refi(attr, file_size_t, 1)?
		       print_bytes_exact(attr_get_ptr(attr, file_size_t)+1):"");
    return 1;
  } else {
    gtk_entry_set_text(MinSizeEntry, "");
    gtk_entry_set_text(MaxSizeEntry, "");
    return 0;
  }
}
static int ui_search_option_get_size(search_pattern_t* pattern) {
  file_size_t min, max;
  
  if (!pattern) return 1;
  min = extract_bytes(gtk_entry_get_text(MinSizeEntry));
  max = extract_bytes(gtk_entry_get_text(MaxSizeEntry));

  if (min || max) {
    attr_store(ATTRC(pattern), "pattern", "Size", min, max);
    return 1;
  } else {
    return 0;
  }
}

static GtkWidget* ui_search_option_create_duration(void) {
  GtkWidget* hbox;
  GtkWidget* entry;
  GtkWidget* label;

  hbox = gtk_hbox_new (FALSE, 5);
  //  gtk_widget_show (hbox);
  entry = gtk_entry_new();
  MinDurationEntry = GTK_ENTRY(entry);
  gtk_widget_show(entry);
  ui_search_input_monitor(entry);
  gtk_widget_set_size_request(entry, 0, -1);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  label = gtk_label_new(_("to"));
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);

  entry = gtk_entry_new();
  MaxDurationEntry = GTK_ENTRY(entry);
  gtk_widget_show(entry);
  ui_search_input_monitor(entry);
  gtk_widget_set_size_request(entry, 0, -1);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  return hbox;
}
static int
ui_search_option_set_duration(search_pattern_t* pattern) {
  attr_container_t* ac;
  attribute_t* attr;

  if (!pattern) return 1;
  ac = ATTRC(pattern);
  attr = attr_find(ac, "pattern", "Duration");
  if (attr) {
    gtk_entry_set_text(MinDurationEntry,
		       attr_get_refi(attr, int, 0)?
		       print_seconds_exact(attr_get_ptr(attr, int)):"");
    gtk_entry_set_text(MaxDurationEntry,
		       attr_get_refi(attr, int, 1)?
		       print_seconds_exact(attr_get_ptr(attr, int)+1):"");
    return 1;
  } else {
    gtk_entry_set_text(MinDurationEntry, "");
    gtk_entry_set_text(MaxDurationEntry, "");
    return 0;
  }
}
static int
ui_search_option_get_duration(search_pattern_t* pattern) {
  int min, max;
  
  if (!pattern) return 1; 
  min = extract_duration(gtk_entry_get_text(MinDurationEntry));
  max = extract_duration(gtk_entry_get_text(MaxDurationEntry));

  if (min || max) {
    attr_store(ATTRC(pattern), "pattern", "Duration", min, max);
    return 1;
  } else {
    return 0;
  }
}

static GtkWidget* ui_search_option_create_bitrate(void) {
  GtkWidget* hbox;
  GtkWidget* entry;
  GtkWidget* label;

  hbox = gtk_hbox_new (FALSE, 5);
  //  gtk_widget_show (hbox);
  entry = gtk_entry_new();
  MinBitrateEntry = GTK_ENTRY(entry);
  gtk_widget_show(entry);
  ui_search_input_monitor(entry);
  gtk_widget_set_size_request(entry, 0, -1);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  label = gtk_label_new(_("to"));
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);

  entry = gtk_entry_new();
  MaxBitrateEntry = GTK_ENTRY(entry);
  gtk_widget_show(entry);
  ui_search_input_monitor(entry);
  gtk_widget_set_size_request(entry, 0, -1);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  return hbox;
}
static int
ui_search_option_set_bitrate(search_pattern_t* pattern) {
  attr_container_t* ac;
  attribute_t* attr;
  char text[100];

  if (!pattern) return 1;
  ac = ATTRC(pattern);
  attr = attr_find(ac, "pattern", "Bitrate");
  if (attr) {
    attr_get_refi(attr, int, 0)?
      sprintf(text, "%d", ((int*)(attr->value))[0]):
      (text[0] = 0);
    gtk_entry_set_text(MinBitrateEntry, text);
    attr_get_refi(attr, int, 1)?
      sprintf(text, "%d", ((int*)(attr->value))[1]):
      (text[0] = 0);
    gtk_entry_set_text(MaxBitrateEntry, text);
    return 1;
  } else {
    gtk_entry_set_text(MinBitrateEntry, "");
    gtk_entry_set_text(MaxBitrateEntry, "");
    return 0;
  }
}
static int
ui_search_option_get_bitrate(search_pattern_t* pattern) {
  int min, max;
  
  if (!pattern) return 1;
  min = atoi(gtk_entry_get_text(MinBitrateEntry));
  max = atoi(gtk_entry_get_text(MaxBitrateEntry));

  if (min || max) {
    attr_store(ATTRC(pattern), "pattern", "Bitrate", min, max);
    return 1;
  } else {
    return 0;
  }
}

static GtkWidget* ui_search_option_create_frequency(void) {
  GtkWidget* hbox;
  GtkWidget* entry;
  GtkWidget* label;

  hbox = gtk_hbox_new (FALSE, 5);
  //  gtk_widget_show (hbox);
  entry = gtk_entry_new();
  MinFrequencyEntry = GTK_ENTRY(entry);
  gtk_widget_show(entry);
  ui_search_input_monitor(entry);
  gtk_widget_set_size_request(entry, 0, -1);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  label = gtk_label_new(_("to"));
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 0);

  entry = gtk_entry_new();
  MaxFrequencyEntry = GTK_ENTRY(entry);
  gtk_widget_show(entry);
  ui_search_input_monitor(entry);
  gtk_widget_set_size_request(entry, 0, -1);
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  return hbox;
}
static int
ui_search_option_set_frequency(search_pattern_t* pattern) {
  attr_container_t* ac;
  attribute_t* attr;
  char text[100];

  if (!pattern) return 1;
  ac = ATTRC(pattern);
  attr = attr_find(ac, "pattern", "Frequency");
  if (attr) {
    ((int*)(attr->value))[0]?
      sprintf(text, "%d", ((int*)(attr->value))[0]):
      (text[0] = 0);
    gtk_entry_set_text(MinFrequencyEntry, text);
    ((int*)(attr->value))[1]?
      sprintf(text, "%d", ((int*)(attr->value))[1]):
      (text[0] = 0);
    gtk_entry_set_text(MaxFrequencyEntry, text);
    return 1;
  } else {
    gtk_entry_set_text(MinFrequencyEntry, "");
    gtk_entry_set_text(MaxFrequencyEntry, "");
    return 0;
  }
}
static int ui_search_option_get_frequency(search_pattern_t* pattern) {
  int min, max;
  
  if (!pattern) return 1;
  min = atoi(gtk_entry_get_text(MinFrequencyEntry));
  max = atoi(gtk_entry_get_text(MaxFrequencyEntry));

  if (min || max) {
    attr_store(ATTRC(pattern), "pattern", "Frequency", min, max);
    return 1;
  } else {
    return 0;
  }
}

static void ui_search_pattern_apply(search_pattern_t* pattern) {
  GList* dlist;
  search_option_t* option;
  int result;

  if (!pattern) return;
  
  // well, this is very easy to do.
  for (dlist = OptionFuncs; dlist; dlist = dlist->next) {
    option = dlist->data;
    result = option->set(pattern);
    if (!option->expander) continue;
    gtk_expander_set_expanded(GTK_EXPANDER(option->expander), (result==1));
  }
  on_input_expose();
}

static void on_search(void)
{
    search_pattern_t* pattern = ui_search_pattern_get();
    if (pattern)
    {
        search_new_from_pattern(pattern, -1);
        object_kill(OBJ(pattern));
    }
    gtk_entry_set_text(IncludeEntry, "");
}

static void ui_search_init(void) {
  static int calls = 0;

  if (calls) return;
  calls++;

  // register pattern fields, used to get/set the pattern
  // attributes from/to the UI
  ui_search_register_option("Search", NULL, NULL,
                            ui_search_option_get_term,
                            ui_search_option_set_term);
  ui_search_register_option("Realm", NULL, NULL,
                            ui_search_option_get_realm,
                            ui_search_option_set_realm);
  ui_search_register_option("Exclude", N_("Exclude"),
                            ui_search_option_create_exclude,
                            ui_search_option_get_exclude,
                            ui_search_option_set_exclude);
  ui_search_register_option("Size", N_("Size"),
                            ui_search_option_create_size,
                            ui_search_option_get_size,
                            ui_search_option_set_size);
  ui_search_register_option("Bitrate", N_("Bitrate"),
                            ui_search_option_create_bitrate,
                            ui_search_option_get_bitrate,
                            ui_search_option_set_bitrate);
  ui_search_register_option("Frequency", N_("Frequency"),
                            ui_search_option_create_frequency,
                            ui_search_option_get_frequency,
                            ui_search_option_set_frequency);
  ui_search_register_option("Duration", N_("Duration"),
                            ui_search_option_create_duration,
                            ui_search_option_get_duration,
                            ui_search_option_set_duration);

  on_input_expose();
}

static gboolean
on_toggle_advanced(GtkWidget* elabel, GdkEventButton *event,
                   GtkNotebook* notebook) {
  int pagenum =
    GPOINTER_TO_INT(g_object_get_data(G_OBJECT(elabel), "pagenum"));

  (void)event;
  if (pagenum == 0) {
    gtk_notebook_next_page(notebook);
    g_object_set_data(G_OBJECT(elabel), "pagenum", GINT_TO_POINTER(1));
    ui_event_label_set_text(elabel, _("search list"));
  } else {
    gtk_notebook_prev_page(notebook);
    g_object_set_data(G_OBJECT(elabel), "pagenum", GINT_TO_POINTER(0));
    ui_event_label_set_text(elabel, _("advanced"));
  }
  return TRUE;
}

static void on_search_changed(GtkTreeSelection *selection)
{
    GtkTreeIter iter;
    GtkTreeModel* model;
    search_gui_t* data;
    search_t* search;
    int index;
    
    if (!gtk_tree_selection_get_selected(selection, &model, &iter))
        return;
    
    gtk_tree_model_get(model, &iter, 0, &search, -1);
    if (!search) return;
    
    data = object_get_data(OBJ(search), "SearchPage");
    if (!data || !data->main) return;
    
    index = gtk_notebook_page_num(NoteBook, data->main);
    if (index < 0) return;
    gtk_notebook_set_current_page(NoteBook, index);
    
    ui_search_pattern_apply(SEARCHP(search));
}

static void
on_search_switch_page(GtkNotebook* notebook, GtkNotebookPage* page,
                      guint page_num, GtkTreeSelection* select)
{
    GtkWidget* widget;
    GList* dlist;
    
    (void)page;
    widget = gtk_notebook_get_nth_page(notebook, page_num);
    for (dlist = Searches; dlist; dlist = dlist->next)
    {
        search_t* search = dlist->data;
        GtkTreeIter* iter = object_get_data(OBJ(search), "SearchEntry");
        search_gui_t* data = object_get_data(OBJ(search), "SearchPage");

        if (iter && data->main == widget)
        {
            if (!gtk_tree_selection_iter_is_selected(select, iter))
                gtk_tree_selection_select_iter(select, iter);
            break;
        }
    }
}

  
static void on_pattern_select(search_pattern_t* pattern) {
  ui_search_pattern_apply(pattern);
  ui_search_set_current_pattern(pattern, pattern->media);
}

static GtkWidget* on_select_pattern(void) {
  GtkWidget* pop;
  GtkWidget* item;
  GList* dlist;

  pop = gtk_menu_new();
  for (dlist = Patterns; dlist; dlist = dlist->next) {
    search_pattern_t* pattern = dlist->data;

    item = gtk_image_menu_item_new_with_label(OBJ(pattern)->name);
    gtk_image_menu_item_set_image
      (GTK_IMAGE_MENU_ITEM(item),
       gtk_image_new_from_stock(MediaStock[pattern->media], GTK_ICON_SIZE_SMALL_TOOLBAR));
    gtk_widget_show(item);
    gtk_container_add(GTK_CONTAINER(pop), item);
    g_signal_connect_swapped(G_OBJECT(item), "activate",
                             G_CALLBACK(on_pattern_select), pattern);
  }
  if (!Patterns) {
    //FIXME: add default patterns
    item = gtk_menu_item_new_with_label(_("No patterns defined"));
    gtk_widget_show(item);
    gtk_container_add(GTK_CONTAINER(pop), item);
  }
  return pop;
}

static gboolean on_add_pattern(GtkEntry* entry) {
  const char* text = gtk_entry_get_text(entry);
  search_pattern_t* pattern;

  if (!text || *text == 0) {
    lopster_notify(LOPSTER_MESSAGE_INFO, _("Missing pattern name"),
                   _("Please enter a name for the search pattern if you wish to save it."));
    return TRUE;
  }
  pattern = ui_search_pattern_get();
  object_set_name(OBJ(pattern), text);
  search_pattern_add(pattern);
  
  gtk_entry_set_text(entry, "");

  on_input_expose();

  return TRUE;
}

static void on_remove_pattern(void) {
  if (!CurPattern) return;
  search_pattern_remove(CurPattern);
}

static GtkWidget* on_result_tip(search_result_t* result, unsigned depth) {
  GtkWidget* table;
  int line = 0;

  (void)depth;

  table = gtk_table_new(1, 2, FALSE);
  gtk_table_set_col_spacings(GTK_TABLE(table), 10);

  ui_table_add_file(GTK_TABLE(table), &line, FILE(result));
  ui_table_add_sep(GTK_TABLE(table), &line);
  ui_table_add_user(GTK_TABLE(table), &line, result->user);

  gtk_widget_show(table);
  return table;
}

static GtkWidget* on_search_tip(search_pattern_t* search, unsigned depth) {
  GtkWidget* table;
  int line = 0;

  (void)depth;

  table = gtk_table_new(1, 2, FALSE);
  gtk_table_set_col_spacings(GTK_TABLE(table), 10);

  if (search->include) {
    ui_table_add_line(GTK_TABLE(table), &line, _("Search for"),
                      make_string_from_list(search->include, " "));
  }

  if (search->exclude) {
    ui_table_add_line(GTK_TABLE(table), &line, _("Exclude"),
                      make_string_from_list(search->exclude, " "));
  }

  ui_table_add_line(GTK_TABLE(table), &line, _("Mediatype"), 
                    _(int2media(search->media, 0)));

  ui_table_add_attrc(GTK_TABLE(table), &line, ATTRC(search), "pattern");

  gtk_widget_show(table);
  return table;
}

static void on_media_select(void* data) {
  int media = GPOINTER_TO_INT(data);
  ui_search_set_current_media(media);
}

static GtkWidget* on_select_media(void) {
  GtkWidget* pop;
  GtkWidget* item;
  unsigned i1;

  pop = gtk_menu_new();

  for (i1 = 0; i1 <= MEDIA_ANY; i1++) {
    if (i1 == MEDIA_UNKNOWN || i1 == MEDIA_FOLDER) continue;
    item = gtk_image_menu_item_new_with_label(_(int2media(i1, 0)));
    gtk_image_menu_item_set_image
      (GTK_IMAGE_MENU_ITEM(item),
       gtk_image_new_from_stock(MediaStock[i1], GTK_ICON_SIZE_SMALL_TOOLBAR));
    gtk_widget_show(item);
    gtk_container_add(GTK_CONTAINER(pop), item);
    g_signal_connect_swapped(G_OBJECT(item), "activate",
                             G_CALLBACK(on_media_select), GINT_TO_POINTER(i1));
  }

  return pop;
}

GtkWidget* ui_create_search_content() {
  GtkWidget* vbox;
  GtkWidget* vbox2;
  GtkWidget* vbox3;
  GtkWidget* hpaned;
  GtkWidget* toplevel;
  GtkWidget* label;
  GtkWidget* hbox;
  GtkWidget* entry;
  GtkWidget* button;
  GtkWidget* image;
  GtkWidget* sw;
  GtkListStore* store;
  GtkWidget* tv;
  GtkWidget* notebook;
  GtkTreeSelection* select;
  GtkTreeViewColumn *column;
  
  toplevel = hpaned = gtk_hpaned_new();
  gtk_widget_show(hpaned);
  ui_layout_monitor("core:search:paned:pos", hpaned);
  //  gtk_box_pack_start(GTK_BOX(vbox0), hpaned, TRUE, TRUE, 0);


  // LEFT
  vbox = gtk_vbox_new(FALSE, 5);
  gtk_widget_show(vbox);
  gtk_paned_pack1(GTK_PANED(hpaned), vbox, FALSE, TRUE);
  
  vbox2 = gtk_vbox_new(FALSE, 2);
  gtk_widget_show(vbox2);
  gtk_box_pack_start(GTK_BOX(vbox), vbox2, FALSE, FALSE, 0);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
  
  PatternButton = button = ui_create_close_button(_("delete this pattern"));
  gtk_widget_show(button);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
  g_signal_connect_swapped(G_OBJECT(button), "clicked", 
                           G_CALLBACK(on_remove_pattern), NULL);
  
  label = gtk_label_new("--");
  PatternLabel = GTK_LABEL(label);
  button = ui_create_popup_button
    (label, POS_LEFT_DOWN, (PopupFunction)on_select_pattern, NULL);
  gtk_widget_show(button);
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  
  image = gtk_image_new_from_stock
    (STOCK_LOPSTER_MTYPE_ALL, GTK_ICON_SIZE_SMALL_TOOLBAR);
  gtk_widget_show(image);
/*   gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 2); */
  MediaImage = GTK_IMAGE(image);
  ui_search_set_current_media(MEDIA_ANY);

  button = ui_create_popup_button
    (image, POS_RIGHT_DOWN, (PopupFunction)on_select_media, NULL);
  gtk_widget_show(button);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);


  entry = gtk_entry_new();
  IncludeEntry = GTK_ENTRY(entry);
  gtk_widget_show(entry);
  gtk_box_pack_start(GTK_BOX(vbox2), entry, FALSE, TRUE, 0);
  g_signal_connect(G_OBJECT(entry), "activate",
                   G_CALLBACK(on_search), NULL);

  

  hbox = gtk_hbox_new(FALSE, 5);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  
  //FIXME: add stock icon
  button = gtk_button_new_with_label(_("start search"));
  gtk_widget_show(button);
  gtk_box_pack_end(GTK_BOX(hbox), button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(on_search), NULL);
  
/*   label = ui_create_event_label */
/*     (_("start search"), G_CALLBACK(on_search), NULL, TRUE); */
/*   gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0); */

  notebook = gtk_notebook_new();
  gtk_widget_show(notebook);
  gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
  gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);
  gtk_box_pack_start(GTK_BOX(vbox), notebook, TRUE, TRUE, 0);
  
  label = ui_create_event_label
    (_("advanced"), G_CALLBACK(on_toggle_advanced), notebook, FALSE);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);

  vbox2 = gtk_vbox_new(FALSE, 5);
  gtk_widget_show(vbox2);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox2, NULL);
  
  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_show(sw);
  gtk_scrolled_window_set_policy
    (GTK_SCROLLED_WINDOW(sw), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_IN);
  gtk_box_pack_start(GTK_BOX(vbox2), sw, TRUE, TRUE, 0);

  store = ui_search_store_get();
  tv = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
  gtk_container_add(GTK_CONTAINER(sw), tv);
  gtk_widget_show(tv);
  gtk_tree_view_set_headers_clickable(GTK_TREE_VIEW(tv), TRUE);
  g_object_set(G_OBJECT(tv), "fixed-height-mode", TRUE, NULL);
  gtk_fancytips_new(GTK_TREE_VIEW(tv), (FancyTipsFunc)on_search_tip, 
                    GTK_FANCY_POS_RIGHT);
  
  select = gtk_tree_view_get_selection(GTK_TREE_VIEW(tv));
  gtk_tree_selection_set_mode(select, GTK_SELECTION_BROWSE);
  g_signal_connect(G_OBJECT(select), "changed",
                   G_CALLBACK(on_search_changed), NULL);
  
  column = ui_tv_column_new
    (_("Searches"), gtk_cell_renderer_text_new(), "text", 1, TRUE, NULL);
  ui_tv_config_default_col(column, 1, "core:search:col:searches", FALSE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tv), column);
  column = ui_tv_column_new
    (_("Results"), gtk_cell_renderer_text_new(), "text", 2, TRUE, NULL);
  ui_tv_config_default_col(column, 2, "core:search:col:results", TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tv), column);

  SumBox = GTK_BOX(vbox2);

  vbox2 = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(vbox2);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox2, NULL);
  g_object_set_data(G_OBJECT(label), "pagenum", GINT_TO_POINTER(0));

  hbox = gtk_hseparator_new();
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 3);

  vbox3 = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(vbox3);
  LeftContent = GTK_BOX(vbox3);
  
  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_show(sw);
  gtk_scrolled_window_set_policy
    (GTK_SCROLLED_WINDOW(sw), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_add_with_viewport
    (GTK_SCROLLED_WINDOW(sw), vbox3);
  gtk_viewport_set_shadow_type
    (GTK_VIEWPORT(gtk_bin_get_child(GTK_BIN(sw))), GTK_SHADOW_NONE);
  gtk_box_pack_start(GTK_BOX(vbox2), sw, TRUE, TRUE, 0);

  hbox = gtk_hbox_new(FALSE, 5);
  gtk_widget_show(hbox);
  gtk_box_pack_end(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);

  entry = gtk_entry_new();
  gtk_widget_show(entry);
  
  g_signal_connect(G_OBJECT(entry), "activate",
		   G_CALLBACK(on_add_pattern), NULL);

  label = ui_create_event_label
    (_("save settings as"), G_CALLBACK(on_add_pattern), entry, TRUE);
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
  {
    GtkRequisition req;
    gtk_widget_size_request(label, &req);
    gtk_widget_set_size_request(entry, 10, req.height+4);
  }
  gtk_box_pack_start(GTK_BOX(hbox), entry, TRUE, TRUE, 0);

  hbox = gtk_hseparator_new();
  gtk_widget_show(hbox);
  gtk_box_pack_end(GTK_BOX(vbox2), hbox, FALSE, FALSE, 3);

  
  // RIGHT
  notebook = gtk_notebook_new();
  gtk_widget_show(notebook);
  gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
  gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);
  g_signal_connect(G_OBJECT(notebook), "switch_page",
                   G_CALLBACK(on_search_switch_page), select);
  NoteBook = GTK_NOTEBOOK(notebook);

  gtk_paned_pack2(GTK_PANED(hpaned), notebook, TRUE, TRUE);

  ui_search_init();
  
  return toplevel;
}

static void ui_search_pattern_update(search_pattern_t* pattern) {
  if (CurPattern == pattern) on_input_expose();
}

static void ui_search_pattern_hide(search_pattern_t* pattern) {
  Patterns = g_list_remove(Patterns, pattern);
  object_unmonitor_d(OBJ(pattern), (ObjectFunc)ui_search_pattern_update, NULL);
  // update global var now, since pattern updater is invoked
  // with some delay
  if (CurPattern == pattern) CurPattern = NULL;
  on_input_expose();
}

void ui_search_pattern_show(search_pattern_t* pattern)
{
    if (g_list_find(Patterns, pattern)) return;
    Patterns = g_list_append(Patterns, pattern);
    //  ui_search_pattern_update(pattern);
    object_notify_d(OBJ(pattern), (ObjectFunc)ui_search_pattern_hide, NULL);
    object_monitor_d(OBJ(pattern), (ObjectFunc)ui_search_pattern_update, NULL);
}

static void ui_search_result_free_gui(search_result_t* result)
{
    object_free_data_common(OBJ(result), "SearchResult");
/*     g_list_foreach(result->children, (GFunc)object_free_data_common, NULL); */
}

static void ui_search_remove(search_t* search)
{
    search_gui_t* data = object_get_data(OBJ(search), "SearchPage");
    GtkTreeIter* iter = object_get_data(OBJ(search), "SearchEntry");
    int index;

    Searches = g_list_remove(Searches, search);
  
    if (data) 
    {
        gtk_tree_store_clear(data->store);

        search_foreach_result(search, (StorageTraverseFunc)ui_search_result_free_gui, NULL);

        ui_layout_unmonitor("core:search:col:filename:size", data->col[0]);
        ui_layout_unmonitor("core:search:col:filename:vis", data->col[0]);
        ui_layout_unmonitor("core:search:col:filesize:size", data->col[1]);
        ui_layout_unmonitor("core:search:col:filesize:vis", data->col[1]);
        ui_layout_unmonitor("core:search:col:sources:size", data->col[2]);
        ui_layout_unmonitor("core:search:col:sources:vis", data->col[2]);
        ui_layout_unmonitor("core:search:col:meta:size", data->col[3]);
        ui_layout_unmonitor("core:search:col:meta:vis", data->col[3]);
        
        index = gtk_notebook_page_num(NoteBook, data->main);
        if (index != -1)
            gtk_notebook_remove_page(NoteBook, index);
    
        if (data->updater) g_source_remove(data->updater);
        data->updater = 0;
    
        g_free(data->source_label);
        object_free_data_common(OBJ(search), "SearchPage");
    }
    if (iter)
    {
        gtk_list_store_remove(ui_search_store_get(), iter);
        object_free_data_common(OBJ(search), "SearchEntry");
    }
}

static gint sort_results_by_source(GtkTreeModel* model,
                                   GtkTreeIter* a, GtkTreeIter* b) {
  search_result_t* r1;
  search_result_t* r2;
  
  gtk_tree_model_get(model, a, 0, &r1, -1);
  gtk_tree_model_get(model, b, 0, &r2, -1);
  
  if (r1->user->ID < r2->user->ID) return -1;
  else if (r1->user->ID > r2->user->ID) return 1;
  else if (r1->weight > r2->weight) return -1;
  else if (r1->weight < r2->weight) return 1;
  else return g_utf8_collate(OBJ(r1->user)->name, OBJ(r2->user)->name); //FIXME: use g_utf8_collate_key()
}

void ui_search_add_page(search_t* search) {
  GtkWidget* sw;
  GtkWidget* hbox0;
  GtkWidget* vbox0;
  GtkWidget* button;
  GtkWidget* label;
  search_gui_t* data;
  GtkTreeIter* iter;
  GtkTreeStore* store;
  GtkWidget* tv;
  GtkTreeViewColumn *column;
  GtkTreeSelection *select;
  int num, i1;

  if (!search) return;

  if (object_get_data(OBJ(search), "SearchPage")) return;

  Searches = g_list_append(Searches, search);
  
  object_set_data(OBJ(search), "SearchPage", data = g_malloc0(sizeof(*data)));
  object_set_data(OBJ(search), "SearchEntry", iter = g_malloc0(sizeof(*iter)));
  
  vbox0 = gtk_vbox_new(FALSE, 2);
  gtk_widget_show(vbox0);
  data->main = vbox0;

  hbox0 = gtk_hbox_new(FALSE, 5);
  gtk_widget_show(hbox0);
  gtk_box_pack_start(GTK_BOX(vbox0), hbox0, FALSE, FALSE, 0);
  
  button = ui_create_close_button(_("close this search page"));
  gtk_widget_show(button);
  gtk_box_pack_start(GTK_BOX(hbox0), button, FALSE, FALSE, 0);
  g_signal_connect_swapped(G_OBJECT(button), "clicked",
                           G_CALLBACK(object_kill), search);
  
  label = gtk_label_new(l_printf("<b>%s</b> [%s]", OBJ(search)->name,
                                 _(int2media(SEARCHP(search)->media, 0))));
  gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
  gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(hbox0), label, TRUE, TRUE, 0);

  button = ui_create_tool_button(GTK_STOCK_REFRESH, _("search again"));
  gtk_widget_show(button);
  gtk_box_pack_end(GTK_BOX(hbox0), button, FALSE, FALSE, 0);
  g_signal_connect_swapped(G_OBJECT(button), "clicked", 
                           G_CALLBACK(search_refresh), search);

  label = gtk_label_new("");
  //  gtk_misc_set_alignment(GTK_MISC(label), 1, 0.5);
  gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
  gtk_widget_show(label);
  gtk_box_pack_end(GTK_BOX(hbox0), label, FALSE, FALSE, 5);
  data->result_label = label;

  num = plugin_get_number();
  data->source_label = g_malloc0(sizeof(*data->source_label)*(num+2));
  for (i1 = num+1; i1 >= 0; i1--) {
    GtkWidget* hbox;
    const char* stock;
    GtkWidget* item;

    hbox = gtk_hbox_new(FALSE, 2);

    if (i1 > num) stock = ui_main_get_plugin_stock(-1);
    else if (i1 == 0) stock = ui_main_get_plugin_stock(-2);
    else stock = ui_main_get_plugin_stock(i1-1);

    if (stock) item = gtk_image_new_from_stock(stock, GTK_ICON_SIZE_MENU);
    else item = gtk_label_new(plugin_get_short_name(plugin_get(i1-1)));
    gtk_widget_show(item);
    gtk_box_pack_start(GTK_BOX(hbox), item, FALSE, FALSE, 0);
    
    item = gtk_label_new(l_printf("%d", search->sources[i1]));
    gtk_widget_show(item);
    gtk_box_pack_start(GTK_BOX(hbox), item, FALSE, FALSE, 0);
    data->source_label[i1] = item;

    if (search->sources[i1] > 0) gtk_widget_show(hbox);
    gtk_box_pack_end(GTK_BOX(hbox0), hbox, FALSE, FALSE, 0);
  }
    
  
  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_box_pack_start(GTK_BOX(vbox0), sw, TRUE, TRUE, 0);
  gtk_widget_show(sw);
  gtk_scrolled_window_set_shadow_type
    (GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_IN);
  gtk_scrolled_window_set_policy
    (GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  data->store = store =
    gtk_tree_store_new(7,
                       G_TYPE_POINTER,  // 0 data pointer
                       G_TYPE_STRING,   // 1 filename
                       G_TYPE_STRING,   // 2 size
                       G_TYPE_STRING,   // 3 source
                       G_TYPE_STRING,   // 4 meta
                       G_TYPE_STRING,   // 5 pix
                       G_TYPE_BOOLEAN   // 6 mark
                       );
  gtk_tree_sortable_set_sort_func
    (GTK_TREE_SORTABLE(store),
     3, (GtkTreeIterCompareFunc)sort_results_by_source,
     NULL, NULL);
  gtk_tree_sortable_set_sort_func
    (GTK_TREE_SORTABLE(store),
     2, (GtkTreeIterCompareFunc)sort_file_by_size,
     store, NULL);
  
  tv = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
  //  gtk_tree_view_columns_autosize(GTK_TREE_VIEW(tv));
  gtk_container_add(GTK_CONTAINER(sw), tv);
  gtk_widget_show(tv);
  gtk_tree_view_set_headers_clickable (GTK_TREE_VIEW (tv), TRUE);
  //  gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (tv), TRUE);
  data->view = GTK_TREE_VIEW(tv);
  g_object_set(G_OBJECT(tv), "fixed-height-mode", TRUE, NULL);
  gtk_fancytips_new(GTK_TREE_VIEW(tv), (FancyTipsFunc)on_result_tip,
                    GTK_FANCY_POS_MOUSE);

  {
    action_item_t* actions[1];
    actions[0] = action_item_search("Search result actions");
    
    ui_tv_handle_button(GTK_TREE_VIEW(tv), 1, actions, search);
  }

  select = gtk_tree_view_get_selection (GTK_TREE_VIEW (tv));
  gtk_tree_selection_set_mode(select, GTK_SELECTION_MULTIPLE);

  
  column = ui_tv_column_new
    (_("Filename"), gtk_cell_renderer_text_new(), "text", 1, TRUE, NULL);
  ui_tv_config_default_col(column, 1, "core:search:col:filename", FALSE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tv), column);
  gtk_tree_view_set_expander_column(GTK_TREE_VIEW (tv), column);
  data->col[0] = column;
  
  column = ui_tv_column_new
    (_("Filesize"), gtk_cell_renderer_text_new(), "text", 2, TRUE, NULL);
  ui_tv_config_default_col(column, 2, "core:search:col:filesize", FALSE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tv), column);
  data->col[1] = column;
  
  column = ui_tv_column_new
    (_("Sources"),
     gtk_cell_renderer_pixbuf_new(), "stock-id", 5, FALSE,
     gtk_cell_renderer_text_new(), "text", 3, TRUE,
     NULL);
  ui_tv_config_default_col(column, 3, "core:search:col:sources", FALSE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tv), column);
  data->col[2] = column;

  column = ui_tv_column_new
    (_("Meta"), gtk_cell_renderer_text_new(), "text", 4, TRUE, NULL);
  ui_tv_config_default_col(column, 4, "core:search:col:meta", TRUE);
  gtk_tree_view_append_column(GTK_TREE_VIEW(tv), column);
  data->col[3] = column;

  ui_tv_config_line_marker(GTK_TREE_VIEW(tv), "search:available", 6);

  g_object_set_data(G_OBJECT(data->main), "search", search);

  // *first* add to store
  gtk_list_store_append(ui_search_store_get(), iter);
  gtk_list_store_set(ui_search_store_get(), iter,
                     0, search,
                     1, OBJ(search)->name,
                     2, search_get_number_results(search),
                     -1);

  // *then* add page, as adding the page triggers callbacks
  gtk_notebook_set_current_page
    (NoteBook, gtk_notebook_append_page(NoteBook, data->main, NULL));

  object_notify_d(OBJ(search), (ObjectFunc)ui_search_remove, NULL);
}

static gboolean on_search_update(search_t* search)
{
    search_gui_t* data = object_get_data(OBJ(search), "SearchPage");
    GtkTreeIter* iter = object_get_data(OBJ(search), "SearchEntry");
    int cnt = search_get_number_results(search);

    // update counter in list.
    if (iter)
    {
        gtk_list_store_set(ui_search_store_get(), iter, 2, cnt, -1);
    }

    // update search page header
    if (data)
    {
        char* l = l_printf(ngettext("%d file", "%d files", cnt), cnt);
        int i1;

        gtk_label_set_markup(GTK_LABEL(data->result_label), l_printf("<b>%s</b>", l));
        
        cnt = 0;
        for (i1 = plugin_get_number()+1; i1 >= 0; i1--)
        {
            if (search->sources[i1] <= 0) continue;
            if (i1 == 0 && cnt < 2) continue;
            gtk_label_set_text(GTK_LABEL(data->source_label[i1]),
                               l_printf("%d", search->sources[i1]));
            if (!GTK_WIDGET_VISIBLE(gtk_widget_get_parent(data->source_label[i1])))
                gtk_widget_show(gtk_widget_get_parent(data->source_label[i1]));
            cnt++;
        }
        if (data->updater) g_source_remove(data->updater);
        data->updater = 0;
    }

    return FALSE;
        
}


static void ui_search_update(search_t* search)
{
    search_gui_t* data = object_get_data(OBJ(search), "SearchPage");
    if (!data || data->updater) return;
    data->updater = g_timeout_add(1000, (GSourceFunc)on_search_update, search);
}

static void ui_search_result_hide(search_result_t* result, search_t* search)
{
    GtkTreeIter* iter;
    search_gui_t* data;

    iter = object_get_data(OBJ(result), "SearchResult");
    if (!iter) return;

    data = object_get_data(OBJ(search), "SearchPage");
    if (data) gtk_tree_store_remove(data->store, iter);

    ui_search_result_free_gui(result);
}

void ui_search_result_show(search_result_t* result, search_t* search)
{
    GtkTreeIter* iter;
    search_gui_t* data;
    
    g_assert(search && result);
  
    data = object_get_data(OBJ(search), "SearchPage");
    if (!data) return;

    iter = object_get_data(OBJ(result), "SearchResult");

    if (!iter)
    {
        GtkTreeIter* parent = NULL;
        if (result->master)
        {
            parent = object_get_data(OBJ(result->master), "SearchResult");
            g_assert(parent);  // really assert?
        }
        object_set_data(OBJ(result), "SearchResult", iter = g_malloc(sizeof(*iter)));
        gtk_tree_store_append(data->store, iter, parent);
    
        gtk_tree_store_set(data->store, iter,
                           0, result,
                           1, FILE(result)->filename,
                           2, print_filesize(&(FILE(result)->size)),
                           -1);
        
        object_notify_d(OBJ(result), (ObjectFunc)ui_search_result_hide, search);
        
        ui_search_update(search);
    }
    
    // fill the entry.
    gtk_tree_store_set(data->store, iter,
                       3, result->user->print(result->user),
                       4, attr_print_all(ATTRC(result)),
                       5, ui_main_get_plugin_stock(result->user->ID),
                       6, result->available,
                       -1);
}

#ifdef WIN32
#define _UI_SEARCH_C_
#include "lgtk2corealiasdef.c"
#endif
