/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.14 $
 * $Date: 2007/11/23 13:11:00 $
 * $Author: sid4u $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "string.h"

#include <libxml/parser.h>
#include <gtk/gtk.h>

#include "lcore/xml_handler.h"
#include "lcore/utils.h"
#include "lcore/object.h"
#include "lcore/l_i18n.h"
#include "ui_style.h"
#include "ui_preferences.h"
#include "ui_support.h"
#include "ui_main.h"
#ifdef WIN32
#include "lcore/win32_defs.h"
#endif

static GHashTable* Styles = NULL;
static GList* StylesOrdered = NULL;
static GList* StyleGroups = NULL;

static GtkWidget* Window = NULL;
static gboolean Loaded = FALSE;
static GtkWidget* StyleWin = NULL;
static GList* Items = NULL;
static GList* StyleObs = NULL;
static char* EditGroup = NULL;
static char* RcFile = NULL;
static GtkLabel* StyleLabel = NULL;
static GtkLabel* RcLabel = NULL;

typedef struct {
  char* name;
  gboolean active;
  char type;
  union {
    GdkColor color;
    char* font;
    gboolean underline;
  };
} ui_style_item_t;

typedef struct {
  object_t _object;
  char* group;
  char* des;
  GdkGC* gc;
  GList* items;
} ui_style_t;

typedef struct {
  ui_style_t* style;
  StyleFunc func;
  gpointer data;
} ui_st_t;


typedef struct {
  ui_style_t* style;
  ui_style_item_t* item;
  GtkToggleButton* check;
  GtkWidget* widget;
} ui_si_t;

static void ui_set_rc_file(const char* filename);
static void ui_styles_set();


static void ui_styles_save(void) {
  if (!Loaded) return;
  lopster_xml_save("ui:core:styles", 5000);
}

void ui_styles_load(void) {
  Loaded = FALSE;
  lopster_xml_load("ui:core:styles");
  Loaded = TRUE;
}

static ui_style_item_t* ui_style_find_item(ui_style_t* style, const char* name) {
  if (style && name) {
    GList* dlist;
    for (dlist = style->items; dlist; dlist = dlist->next) {
      ui_style_item_t* item = dlist->data;
      if (!strcmp(item->name, name)) return item;
    }
  }
  return NULL;
}

static void ui_style_destroy(ui_style_t* style) {
  g_free(style->des);
  if (style->gc) g_object_unref(style->gc);
  object_destroy(OBJ(style));
}

static void ui_style_set_group(ui_style_t* style, const char* group) {
  GList* dlist;
  
  if (!group) return;
  if (style->group) {
    StylesOrdered = g_list_remove(StylesOrdered, style);
    g_free(style->group);
  }
  style->group = g_strdup(group);
  StylesOrdered = g_list_append(StylesOrdered, style);

  for (dlist = StyleGroups; dlist; dlist = dlist->next) {
    char* g = dlist->data;
    if (!strcmp(g, group)) return;
  }
  StyleGroups = g_list_append(StyleGroups, g_strdup(group));
}

static void
ui_style_register_valist(ui_style_t* style, const char* name, va_list ap) {
  while (name) {
    int type = va_arg(ap, int);
    gboolean set = va_arg(ap, gboolean);
    ui_style_item_t* item = NULL;

    if (ui_style_find_item(style, name)) {
      g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
            "style '%s' already has '%s'", OBJ(style)->name, name);
    } else if (type == STYLE_COLOR) {
      item = g_malloc0(sizeof(*item));
      item->name = g_strdup(name);
      item->active = set;
      item->type = type;
      item->color.red = va_arg(ap, int);
      item->color.green = va_arg(ap, int);
      item->color.blue = va_arg(ap, int);
    } else if (type == STYLE_FONT) {
      item = g_malloc0(sizeof(*item));
      item->name = g_strdup(name);
      item->type = type;
      item->font = g_strdup(va_arg(ap, char*));
      item->active = (item->font && *item->font)?set:FALSE;
    } else if (type == STYLE_UNDERLINE) {
      item = g_malloc0(sizeof(*item));
      item->name = g_strdup(name);
      item->active = set;
      item->type = type;
      item->underline = va_arg(ap, gboolean);
    } else {
      g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
            "unknown item '%s' in style '%s'", name, OBJ(style)->name);
    }
    if (item) style->items = g_list_append(style->items, item);
    name = va_arg(ap, char*);
  }
}

/** Register a new style.
 * @param group the group of the style
 * @param name the name of the style
 * @param des the description
 * @param first_name the first style property
 * @param ... more style properties.
 * @note a style property consists of four parts: the style name, the
 * style type, a flag whether it is initially active, and the style
 * args itself depending on the style type. Example:
 * <div style='border:1px solid #aaa;padding:10px;background-color:#eee'><pre>
 *  ui_style_register("General style settings", "filetree:recent",
 *                   "Recent files in the library", 
 *                   "Text",       STYLE_COLOR, TRUE,  0xffff, 0x44e2, 0x0000,
 *                   "Background", STYLE_COLOR, FALSE, 0xcccc, 0xcccc, 0xcccc,
 *                   "Font",       STYLE_FONT,  FALSE, NULL,
 *                   NULL);
 * </pre></div>
 */
void ui_style_register(const char* group, const char* name, const char* des,
                       const char* first_name, ...) {
  ui_style_t* style;
  va_list ap;

  if (!Styles) {
    Styles = g_hash_table_new_full
      ((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
       (GDestroyNotify)NULL, (GDestroyNotify)ui_style_destroy);
  }

  style = g_hash_table_lookup(Styles, name);
  if (style) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "style '%s' already existent", name);
  } else {
    style = g_malloc0(sizeof(*style));
    object_init(OBJ(style), name, FALSE);
    object_add_type(OBJ(style), "Style");
    style->des = g_strdup(des);
    g_hash_table_insert(Styles, OBJ(style)->name, style);
    
    g_log("GTK_STYLE", G_LOG_LEVEL_DEBUG,
          "registerd style '%s'", name);
    va_start(ap, first_name);
    ui_style_register_valist(style, first_name, ap);
    va_end(ap);
    //    object_set_dirty(OBJ(style));
  }
  ui_style_set_group(style, group);
}

static void on_style_changed(ui_style_t* style, ui_st_t* data) {
  g_assert(data->style == style);
  data->func(OBJ(style)->name, data->data);
}

static void on_style_die(ui_style_t* style, ui_st_t* data) {
  g_assert(data->style == style);
  StyleObs = g_list_remove(StyleObs, data);
  g_free(data);
}

/** Observe a style for changes.
 * @param name the name of the style
 * @param func the function to chall on changes.
 * @param data user data for the observer function
 * @see ui_style_register()
 */
void ui_style_observe(const char* name, StyleFunc func, gpointer data) {
  ui_style_t* style = g_hash_table_lookup(Styles, name);
  ui_st_t* st;
  
  if (!style) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "cannot observe style '%s' : not found", name);
    return;
  }
  st = g_malloc(sizeof(*st));
  st->func = func;
  st->data = data;
  st->style = style;

  StyleObs = g_list_prepend(StyleObs, st);

  object_monitor_d(OBJ(style), (ObjectFunc)on_style_changed, st);
  object_notify_d(OBJ(style), (ObjectFunc)on_style_die, st);
}

/** Stop observing a style.
 * @param name the name of the style.
 * @param func the observer function
 * @param data the user data for the observer.
 * @see ui_style_observe()
 */
void ui_style_unobserve(const char* name, StyleFunc func, gpointer data)
{
    ui_style_t* style = g_hash_table_lookup(Styles, name);
    GList* dlist;
    
    if (!style)
    {
        g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
              "cannot unobserve style '%s' : not found", name);
        return;
    }
    for (dlist = StyleObs; dlist; dlist = dlist->next)
    {
        ui_st_t* st = dlist->data;
        if (st->style == style && st->func == func && st->data == data)
        {
            object_unmonitor_d(OBJ(style), (ObjectFunc)on_style_changed, st);
            object_unnotify_d(OBJ(style), (ObjectFunc)on_style_die, st);
            StyleObs = g_list_remove(StyleObs, st);
            g_free(st);
            return;
        }
    }
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "cannot find observer for style '%s'", name);
}

static void
ui_style_set_valist(ui_style_t* style, const char* name, va_list ap) {
  gboolean changed = FALSE;
  while (name) {
    ui_style_item_t* item = ui_style_find_item(style, name);
    if (!item) {
      g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
            "style '%s' has no property '%s' registerd", OBJ(style)->name, name);
    } else if (item->type == STYLE_COLOR) {
      int r = va_arg(ap, int);
      int g = va_arg(ap, int);
      int b = va_arg(ap, int);
      if (!item->active || item->color.red != r ||
          item->color.green != g || item->color.blue != b) {
        item->color.red = r;
        item->color.green = g;
        item->color.blue = b;
        item->active = TRUE;
        changed = TRUE;
      }
    } else if (item->type == STYLE_FONT) {
      char* font = va_arg(ap, char*);
      if (!item->active || strcmp(font, item->font)) {
        g_free(item->font);
        item->font = (font&&*font)?g_strdup(font):NULL;
        item->active = (item->font != NULL);
        changed = TRUE;
      }
    } else if (item->type == STYLE_UNDERLINE) {
      gboolean underline = va_arg(ap, gboolean);
      if (!item->active || item->underline != underline) {
        item->underline = underline;
        item->active = TRUE;
        changed = TRUE;
      }
    } else {
      // there should never be an unknown item registered.
      g_assert(FALSE);
    }
    name = va_arg(ap, char*);
  }
  if (changed) object_set_dirty(OBJ(style));
}

/** Set the properties of a style.
 * @param name the name of the style.
 * @param first_item the name of the first property
 * @param ... the property arguments and more properties
 */
void ui_style_set(const char* name, const char* first_item, ...) {
  ui_style_t* style = g_hash_table_lookup(Styles, name);
  va_list ap;

  if (!style) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "cannot set style '%s' : not found", name);
    return;
  }
  va_start(ap, first_item);
  ui_style_set_valist(style, first_item, ap);
  va_end(ap);
  ui_styles_save();
}

/** Unset a style property.
 * @param name the name of the style.
 * @param item the name of the property to unset.
 */
void ui_style_unset(const char* name, const char* item) {
  ui_style_t* style = g_hash_table_lookup(Styles, name);
  ui_style_item_t* i;

  if (!style) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "cannot unset style '%s' : not found", name);
    return;
  }
  i = ui_style_find_item(style, item);
  if (!i) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "style '%s' has no property '%s'", name, item);
    return;
  }
  if (i->active) {
    i->active = FALSE;
    object_set_dirty(OBJ(style));
    ui_styles_save();
  }
}

/** Check a style for a active property. To check whether the style
 * has an a property property registed use ui_style_has()
 * @param name the name of the style.
 * @param item the name of the property
 * @return TRUE if property is active, FALSE otherwise
 */
gboolean ui_style_sets(const char* name, const char* item) {
  ui_style_t* style = g_hash_table_lookup(Styles, name);
  ui_style_item_t* i;

  if (!style) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "cannot find style '%s'", name);
    return FALSE;
  }
  i = ui_style_find_item(style, item);
  return (i && i->active);
}


/** Get a gc with forground and/or background set.
 * @param name the name of the style.
 * @param color the foreground color, or NULL
 * @return the gc or NULL if style was not found.
 */
GdkGC* ui_style_get_gc(const char* name, const char* color) {
  ui_style_t* style = g_hash_table_lookup(Styles, name);

  if (!style) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "cannot get gc: style '%s' not found", name);
    return NULL;
  }
  if (!style->gc) style->gc = gdk_gc_new(Window->window);
  if (color) {
    ui_style_item_t* item = ui_style_find_item(style, color);
    if (!item) {
      g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
            "style '%s' does not define foreground '%s'", name, color);
    } else if (item->type != STYLE_COLOR) {
      g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
            "property '%s' of style '%s' is no color", color, name);
      /*
    } else if (!item->active) {
      g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
            "property '%s' of style '%s' is not active", color, name);
      */
    } else {
      gdk_gc_set_rgb_fg_color(style->gc, &item->color);
    }
  }
  return style->gc;
}

/** Get a color from the style.
 * @param name the name of the style.
 * @param color the color item of the style.
 * @return the color, or NULL if not found or inactive
 */
GdkColor* ui_style_get_color(const char* name, const char* color) {
  ui_style_t* style = g_hash_table_lookup(Styles, name);
  ui_style_item_t* item;

  if (!style) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "style '%s' not found", name);
  } else if (!(item = ui_style_find_item(style, color))) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "property '%s' in style '%s' is not defined", color, name);
  } else if (item->type != STYLE_COLOR) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "property '%s' in style '%s' is not a color", color, name);
  } else if (!item->active) {
    /*
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "property '%s' in style '%s' is not active", color, name);
    */
  } else {
    return &item->color;
  }
  return NULL;
}

/** Get a font from the style.
 * @param name the name of the style.
 * @param font the property name 
 * @return the fontname or NULL if not found or unset.
 */
const char* ui_style_get_font(const char* name, const char* font) {
  ui_style_t* style = g_hash_table_lookup(Styles, name);
  ui_style_item_t* item;
  
  if (!style) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "style '%s' not found", name);
  } else if (!(item = ui_style_find_item(style, font))) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "property '%s' in style '%s' is not defined", font, name);
  } else if (item->type != STYLE_FONT) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "property '%s' in style '%s' is not a font", font, name);
  } else if (!item->active) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "property '%s' in style '%s' is not active", font, name);
  } else {
    return (item->font && *item->font)?item->font:NULL;
  }
  return NULL;
}

/** Get the underlined property from the style.
 * @param name the name of the style.
 * @param underline the property name 
 * @return the TRUE if underlined, FALSE otherwise
 */
gboolean ui_style_get_underline(const char* name, const char* underline) {
  ui_style_t* style = g_hash_table_lookup(Styles, name);
  ui_style_item_t* item;
  
  if (!style) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "style '%s' not found", name);
  } else if (!(item = ui_style_find_item(style, underline))) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "property '%s' in style '%s' is not defined", underline, name);
  } else if (item->type != STYLE_UNDERLINE) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "property '%s' in style '%s' is not a underline", underline, name);
  } else if (!item->active) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "property '%s' in style '%s' is not active", underline, name);
  } else {
    return item->underline;
  }
  return FALSE;
}


void ui_styles_set_main(GtkWidget* widget) {
  g_assert(widget);
  
  if (!GTK_WIDGET_REALIZED(widget)) gtk_widget_realize(widget);
  g_assert(widget->window);
  Window = widget;
  
  //  ui_styles_load();
}



static void ui_styles_apply(void) {
  GList* dlist;

  for (dlist = Items; dlist; dlist = dlist->next) {
    ui_si_t* si = dlist->data;
    ui_style_t* style = si->style;
    ui_style_item_t* item = si->item;
    
    if (gtk_toggle_button_get_active(si->check)) {
      if (item->type == STYLE_COLOR) {
        GdkColor color;
        gtk_color_button_get_color(GTK_COLOR_BUTTON(si->widget), &color);
        ui_style_set(OBJ(style)->name, item->name, 
                     color.red, color.green, color.blue, NULL);
      } else if (item->type == STYLE_FONT) {
        const char* font =
          gtk_font_button_get_font_name(GTK_FONT_BUTTON(si->widget));
        ui_style_set(OBJ(style)->name, item->name, font, NULL);
      } else if (item->type == STYLE_UNDERLINE) {
        gboolean active =
          gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(si->widget));
        ui_style_set(OBJ(style)->name, item->name, active, NULL);
      } else {
        g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
              "item '%s' from style '%s' has invalid type %d",
              OBJ(style)->name, item->name, item->type);
        g_assert(FALSE);
      }
    } else {
      ui_style_unset(OBJ(style)->name, item->name);
    }
  }
  {
    const char* text;
    text = gtk_label_get_text(StyleLabel);
    if (text && *text == 0) text = NULL;
    if (lopster_xml_change("ui:core:styles", text, TRUE)) ui_styles_set();
    text = gtk_label_get_text(RcLabel);
    if (RcFile != text && (!RcFile || !text || strcmp(RcFile, text)))
      ui_set_rc_file(text);
  }
}

static void on_styles_cancel(GtkWidget* win) {
  gtk_widget_hide(win);
}

static void on_styles_apply(GtkWidget* win) {
  (void)win;
  ui_styles_apply();
}

static void on_styles_ok(GtkWidget* win) {
  ui_styles_apply();
  on_styles_cancel(win);
}

static void on_item_activate(GtkToggleButton* button, GtkWidget* item) {
  gtk_widget_set_sensitive(item, gtk_toggle_button_get_active(button));
}

static GtkToggleButton*
ui_create_style_item(GtkWidget* box, ui_style_item_t* item, GtkWidget* widget) {
  GtkWidget* hbox;
  GtkWidget* check;

  hbox = gtk_vbox_new(FALSE, 0);
  check = gtk_check_button_new_with_label(_(item->name));
  gtk_box_pack_start(GTK_BOX(hbox), check, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), widget, TRUE, TRUE, 0);
  if (item->active)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(check), TRUE);
  else
    gtk_widget_set_sensitive(widget, FALSE);
  g_signal_connect(G_OBJECT(check), "toggled",
                   G_CALLBACK(on_item_activate), widget);

  gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 0);

  return GTK_TOGGLE_BUTTON(check);
}

static void on_style_changed2(ui_style_t* style, ui_si_t* item) {
  ui_style_item_t* i = item->item;

  g_assert(style == item->style);

  if (i->type == STYLE_COLOR) {
    gtk_color_button_set_color(GTK_COLOR_BUTTON(item->widget), &i->color);
  } else if (i->type == STYLE_FONT) {
    if (i->font)
      gtk_font_button_set_font_name(GTK_FONT_BUTTON(item->widget), i->font);
  } else if (i->type == STYLE_UNDERLINE) {
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(item->widget), i->underline);
  }
  gtk_toggle_button_set_active(item->check, i->active);
}

static void ui_create_style_edit(ui_style_t* style, GtkBox* box)
{
    GtkWidget* hbox;
    GtkWidget* exp;
    GtkWidget* label;
    GList* dlist;
    
    if (!style->des || !style->group) return;
    if (strcmp(style->group, EditGroup)) return;
    exp = gtk_expander_new(l_printf("<b>%s</b>", _(style->des)));
    
    hbox = gtk_hbox_new(FALSE, 5);
    gtk_container_add(GTK_CONTAINER(exp), hbox);
    gtk_expander_set_expanded(GTK_EXPANDER(exp), FALSE);
    gtk_expander_set_use_markup(GTK_EXPANDER(exp), TRUE);
    
    label = gtk_label_new("");
    gtk_widget_set_size_request(label, 20, -1);
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    
    //////////////////////
    
    for (dlist = style->items; dlist; dlist = dlist->next)
    {
        ui_style_item_t* i = dlist->data;
        ui_si_t* item = g_malloc0(sizeof(*item));
        
        item->style = style;
        item->item = i;
        
        if (i->type == STYLE_COLOR)
        {
            item->widget = gtk_color_button_new_with_color(&i->color);
        }
        else if (i->type == STYLE_FONT)
        {
            if (!i->font || *i->font == 0) item->widget = gtk_font_button_new();
            else item->widget = gtk_font_button_new_with_font(i->font);
        }
        else if (i->type == STYLE_UNDERLINE)
        {
            item->widget = gtk_check_button_new_with_label(_("yes"));
            if (i->underline)
                gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(item->widget), TRUE);
        }
        else
        {
            g_assert(FALSE);
            continue;
        }
        Items = g_list_append(Items, item);
        item->check = ui_create_style_item(hbox, i, item->widget);
        object_monitor_d(OBJ(style), (ObjectFunc)on_style_changed2, item);
    }
    
    gtk_box_pack_start(box, exp, FALSE, FALSE, 0);
}

static void on_change_file(GtkLabel* label)
{
    const char* c = gtk_label_get_text(label);
    char* filename =
        ui_select_file(_("Choose file"), (c && *c)?c:NULL, GTK_WINDOW(StyleWin));

    if (filename) 
    {
        gtk_label_set_text(label, filename);
        g_free(filename);
    }
}

static void ui_style_settings_save() {
  lopster_xml_save("ui:core:settings", 0);
}

static void ui_styles_set() {
  ui_styles_load();
  ui_style_settings_save();
}
static void ui_set_rc_file(const char* filename) {
  g_free(RcFile);
  RcFile = g_strdup(filename);
  ui_style_settings_save();
  gtk_rc_reparse_all();
}
static void on_load_rc(GtkLabel* label) {
  const char* text = gtk_label_get_text(label);
  if (text && *text && strcmp(text, _("none loaded")))
    ui_set_rc_file(text);
}
static void on_load_styles(GtkLabel* label) {
  const char* text = gtk_label_get_text(label);
  if (text && *text) {
    lopster_xml_change("ui:core:styles", text, TRUE);
    ui_styles_set();
  }
}

void ui_create_style_editor()
{
    GtkWidget* window;
    GtkWidget* button;
    GtkWidget* vbox0;
    GtkWidget* vbox;
    GtkWidget* vbox1;
    GtkWidget* hbox;
    GtkWidget* hhbox;
    GtkWidget* label;
    GtkWidget* temp;
    GList* dlist;
  
    if (StyleWin)
    {
        gtk_window_present(GTK_WINDOW(StyleWin));
        return;
    }
  
    StyleWin = window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), _("Style editor"));
    gtk_window_set_role(GTK_WINDOW(window), "Styles");
    gtk_window_set_default_size(GTK_WINDOW(window), 600, 500);
    gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_DIALOG);
    gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER_ON_PARENT);
    if (ui_main_get_win())
    {
        gtk_window_set_transient_for(GTK_WINDOW(window), ui_main_get_win());
        gtk_window_set_destroy_with_parent(GTK_WINDOW(window), TRUE);
    }

    g_signal_connect(G_OBJECT(window), "delete_event",
                     G_CALLBACK(on_styles_cancel), NULL);
  
    vbox0 = gtk_vbox_new(FALSE, 10);
    gtk_container_add(GTK_CONTAINER(window), vbox0);
    gtk_container_set_border_width(GTK_CONTAINER(vbox0), 10);
  

    vbox = gtk_vbox_new(FALSE, 10);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
    temp = ui_create_scrolled_area(vbox, FALSE);
    gtk_box_pack_start(GTK_BOX(vbox0), temp, TRUE, TRUE, 0);


    // first item
    label = ui_create_header_box(_("Style files for lopster"));
    gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

    vbox1 = ui_create_section(_("Currently loaded gtk rc-file"), NULL, &hbox, &hhbox);
    gtk_box_pack_start(GTK_BOX(vbox), vbox1, FALSE, FALSE, 0);

    label = gtk_label_new(RcFile?RcFile:_("none loaded"));
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    RcLabel = GTK_LABEL(label);
    gtk_label_set_selectable(RcLabel, TRUE);

    label = ui_create_event_label
        (_("reload now"), G_CALLBACK(on_load_rc), RcLabel, TRUE);
    gtk_box_pack_end(GTK_BOX(hhbox), label, FALSE, FALSE, 0);
    label = ui_create_event_label
        (_("change (needs restart)"), G_CALLBACK(on_change_file), RcLabel, TRUE);
    gtk_box_pack_end(GTK_BOX(hhbox), label, FALSE, FALSE, 0);

    gtk_widget_show_all(vbox1);

    vbox1 = ui_create_section(_("Currently loaded style-file (see sections below)"), NULL, &hbox, &hhbox);
    gtk_box_pack_start(GTK_BOX(vbox), vbox1, FALSE, FALSE, 0);

    if (lopster_xml_find("ui:core:styles"))
        label = gtk_label_new(lopster_xml_find("ui:core:styles"));
    else
        label = gtk_label_new("");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    StyleLabel = GTK_LABEL(label);
    gtk_label_set_selectable(StyleLabel, TRUE);

    label = ui_create_event_label
        (_("reload now"), G_CALLBACK(on_load_styles), StyleLabel, TRUE);
    gtk_box_pack_end(GTK_BOX(hhbox), label, FALSE, FALSE, 0);
    label = ui_create_event_label
        (_("change"), G_CALLBACK(on_change_file), StyleLabel, TRUE);
    gtk_box_pack_end(GTK_BOX(hhbox), label, FALSE, FALSE, 0);

    gtk_widget_show_all(vbox1);

    for (dlist = StyleGroups; dlist; dlist = dlist->next) {
        EditGroup = dlist->data;

        vbox1 = gtk_vbox_new(FALSE, 0);
        gtk_box_pack_start(GTK_BOX(vbox), vbox1, FALSE, FALSE, 0);

        label = ui_create_header_box(_(EditGroup));
        gtk_box_pack_start(GTK_BOX(vbox1), label, FALSE, FALSE, 0);
    
        //    g_hash_table_foreach(Styles, (GHFunc)ui_create_style_edit, vbox1);
        g_list_foreach(StylesOrdered, (GFunc)ui_create_style_edit, vbox1);
    }

    // separator
    temp = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox0), temp, FALSE, FALSE, 0);

    // buttons
    hbox = ui_create_bbox(10, GTK_BUTTONBOX_END);
    gtk_box_pack_end(GTK_BOX(vbox0), hbox, FALSE, FALSE, 0);

    button = gtk_button_new_from_stock(GTK_STOCK_APPLY);
    gtk_container_add(GTK_CONTAINER(hbox), button);
    g_signal_connect_swapped(G_OBJECT(button), "clicked",
                             G_CALLBACK(on_styles_apply), window);

    button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
    gtk_container_add(GTK_CONTAINER(hbox), button);
    g_signal_connect_swapped(G_OBJECT(button), "clicked",
                             G_CALLBACK(on_styles_cancel), window);

    button = gtk_button_new_from_stock(GTK_STOCK_OK);
    gtk_container_add(GTK_CONTAINER(hbox), button);
    g_signal_connect_swapped(G_OBJECT(button), "clicked",
                             G_CALLBACK(on_styles_ok), window);

    gtk_widget_show_all(window);
}


static void xml_get_style(xmlNodePtr xnode) {
  xmlNodePtr xsub;
  xmlChar* name;
  ui_style_t* style;

  name = xmlGetProp(xnode, "name");
  if (!name) return;
  style = g_hash_table_lookup(Styles, name);
  //  ui_style_register(NULL, name, NULL, STYLE_NONE, STYLE_NONE);
  if (!style) {
    g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
          "style '%s' not found", name);
    return;
  }
  xsub = xnode->xmlChildrenNode;

  while (xsub != NULL) {
    int r=0,g=0,b=0;
    ui_style_item_t* item;

    if (!strcmp(xsub->name, "Color")) {
      xmlChar* attr = xmlGetProp(xsub, "name");
      xmlChar* attr2 = xmlGetProp(xsub, "active");
      item = ui_style_find_item(style, attr);
      if (attr) xmlFree(attr);
      if (item) {
        attr = xmlGetProp(xsub, "red");
        if (attr && !strncmp(attr, "0x", 2)) {
          r = strtoul(attr+2, NULL, 16);
        } else {
          r = atoi(attr);
        }
        if (attr) xmlFree(attr);
        attr = xmlGetProp(xsub, "green");
        if (attr && !strncmp(attr, "0x", 2)) {
          g = strtoul(attr+2, NULL, 16);
        } else {
          g = atoi(attr);
        }
        if (attr) xmlFree(attr);
        attr = xmlGetProp(xsub, "blue");
        if (attr && !strncmp(attr, "0x", 2)) {
          b = strtoul(attr+2, NULL, 16);
        } else {
          b = atoi(attr);
        }
        if (attr) xmlFree(attr);

        item->color.red = r;
        item->color.green = g;
        item->color.blue = b;
        item->active = (attr2 && (!strcasecmp(attr2, "yes") || !strcmp(attr2, "1")));
      }
      if (attr2) xmlFree(attr2);
    } else if (!strcmp(xsub->name, "Font")) {
      xmlChar* attr = xmlGetProp(xsub, "name");
      xmlChar* attr2 = xmlGetProp(xsub, "active");
      item = ui_style_find_item(style, attr);
      if (attr) xmlFree(attr);
      if (item) {
        attr = xmlGetProp(xsub, "font");
        if (attr) {
          item->font = g_strdup(attr);
          xmlFree(attr);
        }
        item->active = (attr2 && (!strcasecmp(attr2, "yes") || !strcmp(attr2, "1")));
      }
      if (attr2) xmlFree(attr2);
    } else if (!strcmp(xsub->name, "Underline")) {
      xmlChar* attr = xmlGetProp(xsub, "name");
      xmlChar* attr2 = xmlGetProp(xsub, "active");
      item = ui_style_find_item(style, attr);
      if (item) {
        if (attr) xmlFree(attr);
        attr = xmlGetProp(xsub, "underline");
        item->underline = (attr && (!strcasecmp(attr, "yes") || !strcmp(attr, "1")));
        item->active = (attr2 && (!strcasecmp(attr2, "yes") || !strcmp(attr2, "1")));
      }
      if (attr) xmlFree(attr);
      if (attr2) xmlFree(attr2);
    } else if (strcmp(xsub->name, "text")) {
      g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
            "unknown node in xml file: '%s'", xsub->name);
    }
    xsub = xsub->next;
  }
  xmlFree(name);

  object_set_dirty(OBJ(style));
}

static void xml_load_styles(xmlDocPtr xfile, xmlNodePtr xroot) {
  xmlNodePtr xsub = xroot->xmlChildrenNode;
  (void)xfile;
  while (xsub != NULL) {
    if (!strcmp(xsub->name, "Style")) {
      xml_get_style(xsub);
    } else if (strcmp(xsub->name, "text")) {
      g_log("GTK_STYLE", G_LOG_LEVEL_WARNING,
            "unknown node in xml file: '%s'", xsub->name);
    }
    xsub = xsub->next;
  }
}


static void xml_save_style_item(ui_style_item_t* item, xmlNodePtr xnode) {
  xmlNodePtr xsub;
  
  if (item->type == STYLE_COLOR) {
    xsub = xmlNewChild(xnode, NULL, "Color", NULL);
    xmlNewProp(xsub, "name", item->name);
    xmlNewProp(xsub, "active", item->active?"yes":"no");
    xmlNewProp(xsub, "red", l_printf("0x%04x", item->color.red));
    xmlNewProp(xsub, "green", l_printf("0x%04x", item->color.green));
    xmlNewProp(xsub, "blue", l_printf("0x%04x", item->color.blue));
  } else if (item->type == STYLE_FONT) {
    xsub = xmlNewChild(xnode, NULL, "Font", NULL);
    xmlNewProp(xsub, "name", item->name);
    xmlNewProp(xsub, "active", item->active?"yes":"no");
    if (item->font) xmlNewProp(xsub, "font", item->font);
  } else if (item->type == STYLE_UNDERLINE) {
    xsub = xmlNewChild(xnode, NULL, "Underline", NULL);
    xmlNewProp(xsub, "name", item->name);
    xmlNewProp(xsub, "active", item->active?"yes":"no");
    xmlNewProp(xsub, "underline", item->underline?"yes":"no");
  }
}

static void xml_save_style(const char* key, ui_style_t* style, xmlNodePtr xnode) {
  xmlNodePtr xsub = xmlNewChild(xnode, NULL, "Style", NULL);
  xmlNewProp(xsub, "name", OBJ(style)->name);
  g_list_foreach(style->items, (GFunc)xml_save_style_item, xsub);

  (void)key;
}

static void xml_save_styles(xmlDocPtr xfile, xmlNodePtr xroot) {
  (void)xfile;
  g_hash_table_foreach(Styles, (GHFunc)xml_save_style, xroot);
}



static void xml_load_settings(xmlDocPtr xfile, xmlNodePtr xroot) {
  xmlNodePtr xsub = xroot->xmlChildrenNode;
  xmlChar* attr;

  while (xsub != NULL) {
    if (!strcmp(xsub->name, "StyleFile")) {
      attr = xmlNodeListGetString(xfile, xsub->xmlChildrenNode, 1);
      if (attr && *attr) lopster_xml_change("ui:core:styles", attr, TRUE);
      if (attr) xmlFree(attr);
    } else if (!strcmp(xsub->name, "RcFile")) {
      attr = xmlNodeListGetString(xfile, xsub->xmlChildrenNode, 1);
      if (attr && *attr) RcFile = g_strdup(attr);
      if (attr) xmlFree(attr);
    }
    xsub = xsub->next;
  }
}

static void xml_save_settings(xmlDocPtr xfile, xmlNodePtr xroot) {
  const char* filename = lopster_xml_find("ui:core:styles");
  (void)xfile;
  if (filename)
    xmlNewTextChild(xroot, NULL, "StyleFile", filename);
  if (RcFile && *RcFile)
    xmlNewTextChild(xroot, NULL, "RcFile", RcFile);
}

void ui_styles_init() {
  lopster_xml_register_home("ui:core:settings", SDIR("gtk2") "settings.xml",
                            "Settings", xml_load_settings, xml_save_settings);
  lopster_xml_register_home("ui:core:styles", SDIR("gtk2") "styles.xml",
                            "Styles", xml_load_styles, xml_save_styles);
  
  Styles = g_hash_table_new_full
    ((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
     (GDestroyNotify)g_free, (GDestroyNotify)NULL);

  if (!lopster_xml_load("ui:core:settings")) {
#ifdef PACKAGE_SOURCE_DIR
    RcFile =
      g_strdup_printf(SDIR(PACKAGE_SOURCE_DIR) SDIR("lcore")
		      SDIR("gtk2_ui") "gtkrc-2.0-lopster");
#else
#ifndef WIN32
    RcFile =
      g_strdup_printf(SDIR(PACKAGE_DATA_DIR) SDIR("gtk2")
                      "gtkrc-2.0-lopster");
#else
    RcFile =
      g_strdup_printf(SDIR(PACKAGE_DATA_DIR) SDIR("Themes") SDIR("gtk2")
                      "gtkrc-2.0-lopster");
#endif
#endif
  }
  if (RcFile) gtk_rc_parse(RcFile);
}

