/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.19 $
 * $Date: 2008/12/23 13:38:14 $
 * $Author: sgop $
 */

#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <string.h>
#include "lcore/l_time.h"
#include "lcore/l_i18n.h"

#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

#include "lcore/utils.h"
#include "lcore/l_base.h"
#include "ui_layout.h"
#include "ui_support.h"
#include "ui_tooltips.h"
#include "ui_main.h"

#ifdef WIN32
#include "lgtk2corealias.h"
#endif

/* #define EVENT_FMT  "-%s" */
/* #define EVENT_HFMT "-<u>%s</u>" */
#define EVENT_FMT  "[%s]"
#define EVENT_HFMT "[%s]"

static PangoLayout* TextLayout = NULL;

/** Load a pixmap from file. The file is search in the
 * pixmap folders registered by lopster_pixmap_dir_add()
 * @param filename the name of the file.
 * @return the GtkImage, or an empty image if not found
 */
GtkWidget* create_pixmap(const gchar* filename) {
  const char* pathname = NULL;
  GtkWidget* pixmap;
  
  if (!filename || !filename[0]) {
    return gtk_image_new();
  }

  pathname = lopster_pixmap_find(filename);

  if (!pathname) {
    g_log("GTK", G_LOG_LEVEL_WARNING, "Couldn't find pixmap file: %s", filename);
    return gtk_image_new();
  }

  pixmap = gtk_image_new_from_file(pathname);
  return pixmap;
}

/** Load a pixbuf from file. The file is search in the
 * pixmap folders registered by lopster_pixmap_dir_add()
 * @param filename the name of the file.
 * @return the GdkPixbuf, or NULL if not found
 */
GdkPixbuf* create_pixbuf(const gchar *filename) {
  const char* pathname = NULL;
  GdkPixbuf* pix;
  
  if (!filename || !filename[0]) return NULL;

  pathname = lopster_pixmap_find(filename);

  if (!pathname) {
    g_log("GTK", G_LOG_LEVEL_WARNING,
          "Couldn't find pixmap file: %s", filename);
    return NULL;
  }

  pix = gdk_pixbuf_new_from_file(pathname, NULL);
  return pix;
}


/** Change the text of the header label previously created with
 * ui_create_header_label()
 * @param label the label
 * @param text the new text.
 */
void ui_set_header_label(GtkLabel* label, const char* text) {
/*   gtk_label_set_markup */
/*     (label, l_printf("<span weight=\"bold\" size=\"larger\">%s</span>", text)); */
  gtk_label_set_markup(label, l_printf("<span size=\"larger\">%s</span>", text));
}

/** Create a header label. The name of the label is 'HeaderLabel' so its
 * style can be modified by the gtk2 rc file.
 * @param text the text of the label
 * @return a GtkLabel
 */
GtkWidget* ui_create_header_label(const char* text) {
  GtkWidget* label;

/*   label = gtk_label_new(l_printf("<span weight=\"bold\" size=\"larger\">%s</span>", text)); */
  label = gtk_label_new(l_printf("<span size=\"larger\">%s</span>", text));
  gtk_widget_set_name(label, "HeaderLabel");
  gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
  gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
  return label;
}

/** Create a header label. The name of the label is 'HeaderBox' so its
 * style can be modified by the gtk2 rc file. This is similar to
 * ui_create_header_label() but its packed into a event box so the
 * background color can be set.
 * @param text the text of the label
 * @return a GtkLabel
 */
GtkWidget* ui_create_header_box_old(const char* text) {
  GtkWidget* label;
  GtkWidget* ebox;

  ebox = gtk_event_box_new();
  gtk_widget_set_name(ebox, "HeaderBox");

  label = gtk_label_new(l_printf("<span weight=\"bold\" size=\"larger\">%s</span>", text));
  gtk_widget_show(label);
  gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
  gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
  gtk_misc_set_padding(GTK_MISC(label), 5, 1);

  gtk_container_add(GTK_CONTAINER(ebox), label);

  return ebox;
}

static gboolean
on_ebox_enter(GtkWidget* box, GdkEventCrossing* event, GtkLabel* label) {
  static GdkCursor* Cursor = NULL;
  const char* text;

  (void)event;
  if (GTK_WIDGET_SENSITIVE(label)) {
    text = g_object_get_data(G_OBJECT(label), "label");
    gtk_label_set_markup(label, l_printf(EVENT_HFMT, text));
    gtk_widget_set_state(GTK_WIDGET(label), GTK_STATE_PRELIGHT);

    if (!Cursor) Cursor = gdk_cursor_new(GDK_HAND2);
    gdk_window_set_cursor(box->window, Cursor);
  }
  return TRUE;
}
static gboolean
on_ebox_leave(GtkWidget* box, GdkEventCrossing* event, GtkLabel* label) {
  const char* text;

  (void)event;
  //  if (GTK_WIDGET_HAS_FOCUS(gtk_widget_get_parent(GTK_WIDGET(label)))) return TRUE;
/*   if (GTK_WIDGET_SENSITIVE(label)) { */
    text = g_object_get_data(G_OBJECT(label), "label");
    gtk_label_set_markup(label, l_printf(EVENT_FMT, text));
    gtk_widget_set_state(GTK_WIDGET(label), GTK_STATE_NORMAL);
    gdk_window_set_cursor(box->window, NULL);
/*   } */
  return TRUE;
}

static void on_clear_label(GtkWidget* widget) {
  char* label = g_object_get_data(G_OBJECT(widget), "label");
  g_free(label);
}

/** Create a new event label. An event label is similar to a html link
 * and can be clicked to do some action.
 * @param text the text of the label
 * @param callback the function to be called when clicking the label
 * @param user_data the data passed to the callback function
 * @param swapped TRUE to call the GCallback with swapped arguments
 * @return the label
 */
GtkWidget* ui_create_event_label(const char* text, GCallback callback,
                                 gpointer user_data, gboolean swapped) {
  GtkWidget* ebox;
  GtkWidget* label;
  char* temp;

  if (!callback) text = l_printf(_("%s (NI)"), text);

  ebox = gtk_event_box_new();
  gtk_widget_set_events(ebox,
                        GDK_BUTTON_PRESS_MASK |
                        GDK_KEY_PRESS_MASK |
                        // GDK_FOCUS_CHANGE_MASK |
                        GDK_ENTER_NOTIFY_MASK |
                        GDK_LEAVE_NOTIFY_MASK);
  GTK_WIDGET_SET_FLAGS(ebox, GTK_CAN_FOCUS);

  gtk_widget_show(ebox);

  label = gtk_label_new(l_printf(EVENT_FMT, text));
  gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
  gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
  temp = g_strdup(text);
  g_object_set_data(G_OBJECT(label), "label", temp);
  g_signal_connect(G_OBJECT(label), "destroy",
                   G_CALLBACK(on_clear_label), NULL);

  g_signal_connect(G_OBJECT(ebox), "enter-notify-event",
                   G_CALLBACK(on_ebox_enter), label);
  g_signal_connect(G_OBJECT(ebox), "leave-notify-event",
                   G_CALLBACK(on_ebox_leave), label);
  
  gtk_widget_set_name(label, "EventLabel");
  gtk_widget_show(label);
  gtk_container_add(GTK_CONTAINER(ebox), label);

  if (callback) {
    g_object_set_data(G_OBJECT(ebox), "callback", callback);
    if (swapped) {
      g_signal_connect_swapped(G_OBJECT(ebox), "button-press-event",
			       callback, user_data);
    } else {
      g_signal_connect(G_OBJECT(ebox), "button-press-event",
		       callback, user_data);
    }
  } else {
    ui_event_label_set_sensitive(ebox, FALSE);
  }
  return ebox;
}

void mtype_browse(const char* url);
static void on_url_clicked(const char* url) {
  mtype_browse(url);
}

/** Create a label that represnts an URL.
 * @param text the label text
 * @param url the actual url
 * @return the event widget
 */
GtkWidget* ui_create_url_label(const char* text, const char* url) {
  GtkWidget* label;
  char* dup = g_strdup(url);

  label = ui_create_event_label(text, G_CALLBACK(on_url_clicked), dup, TRUE);
  g_signal_connect_swapped(G_OBJECT(label), "destroy", G_CALLBACK(g_free), dup);
  
  return label;
}

/** Set or unset a event label sensitive.
 * @param ebox the event label
 * @param set TRUE to set sensitive, FALSE to disable
 */
void ui_event_label_set_sensitive(GtkWidget* ebox, gboolean set) {
  gpointer callback = g_object_get_data(G_OBJECT(ebox), "callback");
  GtkWidget* child = GTK_BIN(ebox)->child;
  if (GTK_WIDGET_SENSITIVE(child) == set) return;

  if (set && !callback) return;
  gtk_widget_set_sensitive(child, set);

  if (!set) {
    if (!callback) return;
    g_signal_handlers_block_matched
      (ebox, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, callback, NULL);
  } else {
    g_signal_handlers_unblock_matched
      (ebox, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, callback, NULL);
  }
}

/** Change the text of the event label
 * @param ebox the event label
 * @param text the new text of the label
 */
void ui_event_label_set_text(GtkWidget* ebox, const char* text) {
  GtkWidget* label;
  char* temp;
  
  label = GTK_BIN(ebox)->child;
  temp = g_object_get_data(G_OBJECT(label), "label");
  g_free(temp);
  g_object_set_data(G_OBJECT(label), "label", g_strdup(text));
  gtk_label_set_text(GTK_LABEL(label), l_printf(EVENT_FMT, text));
}

/** Create a button box used in dialog boxes.
 * @param spacing the space between the buttons.
 * @param layout the gtk layout for the buttons, normally GTK_BUTTONBOX_END
 * @return the GtkButtonBox, which basically is a GtkHBox
 */
GtkWidget* ui_create_bbox(gint spacing, gint layout) {
  GtkWidget *bbox;

  bbox = gtk_hbutton_box_new();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), layout);
  gtk_box_set_spacing(GTK_BOX(bbox), spacing);
  //  gtk_button_box_set_child_size(GTK_BUTTON_BOX(bbox), child_w, child_h);
  //  gtk_widget_show(bbox);

  return bbox;
}

static void on_show_hide(GtkWidget* ebox, GdkEventButton* event, GtkWidget* widget) {
  char* new_text;

  (void)event;
  if (GTK_WIDGET_VISIBLE(widget)) {
    new_text = g_object_get_data(G_OBJECT(widget), "show_text");
    gtk_widget_hide(widget);
  } else {
    new_text = g_object_get_data(G_OBJECT(widget), "hide_text");
    gtk_widget_show(widget);
  }
  ui_event_label_set_text(ebox, new_text);
}

/** Create a special event label for showing/hiding a widget. This is
 * a wrapper for ui_create_event_label().
 * @param show_text label when the widget is hidden.
 * @param hide_text label when the widget is shown.
 * @param widget widget which visibility should be toggled.
 * @return the event label
 */
GtkWidget* ui_create_show_hide_label(const char* show_text, const char* hide_text,
                                     GtkWidget* widget) {
  char* t1;
  char* t2;
  GtkWidget* label;

  t1 = g_strdup(show_text);
  t2 = g_strdup(hide_text);
  g_object_set_data(G_OBJECT(widget), "show_text", t1);
  g_object_set_data(G_OBJECT(widget), "hide_text", t2);
  g_signal_connect_swapped(G_OBJECT(widget), "destroy",
			   G_CALLBACK(g_free), t1);
  g_signal_connect_swapped(G_OBJECT(widget), "destroy",
			   G_CALLBACK(g_free), t2);

  label = ui_create_event_label
    (GTK_WIDGET_VISIBLE(widget)?t2:t1, G_CALLBACK(on_show_hide), widget, FALSE);
  return label;
}

/** Create a too button with an icon and without label.
 * @param stock the stock icon
 * @param tip the tooltip for the button
 * @return a GtkButton with an icon
 * @note the button has name 'ToolButton' so you can change the style
 * of the button in lopster's gtkrc file.
 */
GtkWidget* ui_create_tool_button(const char* stock, const char* tip) {
  GtkWidget* pixmap;
  GtkRcStyle* rcstyle;
  GtkWidget* button;

  /*
  pixmap =
    gtk_image_new_from_stock(STOCK_LOPSTER_CLOSE,
			     GTK_ICON_SIZE_SMALL_TOOLBAR);
  */
  pixmap = gtk_image_new_from_stock(stock, GTK_ICON_SIZE_MENU);
  gtk_widget_show(pixmap);

  rcstyle = gtk_rc_style_new();
  rcstyle->xthickness = rcstyle->ythickness = 1;
  
  button = gtk_button_new();
  gtk_widget_modify_style(button, rcstyle);
  g_object_unref(rcstyle);

  gtk_widget_set_name(button, "ToolButton");
  g_object_set(G_OBJECT(button), "can-focus", FALSE, "can-default", FALSE, NULL);


  gtk_container_add(GTK_CONTAINER(button), pixmap);
  gtk_button_set_relief(GTK_BUTTON(button), GTK_RELIEF_NONE);

  if (tip) {
      gtk_widget_set_tooltip_text(button, tip);
  }

  return button;
}

/** Create a close button. Wrapper for ui_create_tool_button() using
 * the STOCK_LOPSTER_CLOSE as stock icon.
 * @param tip the tooltip for the button
 * @return the tool button
 */
GtkWidget* ui_create_close_button(const char* tip) {
  return ui_create_tool_button(STOCK_LOPSTER_CLOSE, tip);
  //  return ui_create_tool_button(GTK_STOCK_STOP);
  //  return ui_create_tool_button(GTK_STOCK_CLOSE);
}

static void on_set_date(GtkWindow* window) {
  GtkCalendar* cal = g_object_get_data(G_OBJECT(window), "calendar");
  GtkButton* button = g_object_get_data(G_OBJECT(window), "button");
  struct tm* tv = g_object_get_data(G_OBJECT(button), "date");
  guint d, m, y;

  gtk_calendar_get_date(cal, &y, &m, &d);
  tv->tm_mday = d;
  tv->tm_mon = m;
  tv->tm_year = y;

  gtk_button_set_label
    (button, l_printf("%02d.%02d.%d", tv->tm_mday, tv->tm_mon, tv->tm_year));

  gtk_widget_destroy(GTK_WIDGET(window));
}

static void on_date_set(GtkButton* button) {
  GtkWidget* window;
  GtkWidget* cal;
  GtkWidget* frame;
  struct tm* tv;
  GtkWidget* vbox;
  GtkWidget* but;
  int x, y;
  gint screen_width;
  gint screen_height;
  GdkScreen *screen;
  GtkRequisition req;

  tv = g_object_get_data(G_OBJECT(button), "date");
  
  window = gtk_window_new(GTK_WINDOW_POPUP);

  gtk_window_set_resizable(GTK_WINDOW(window), FALSE);
  gtk_window_set_modal(GTK_WINDOW(window), TRUE);
  gtk_window_set_type_hint(GTK_WINDOW(window), GDK_WINDOW_TYPE_HINT_DIALOG);
  gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER_ON_PARENT);
  if (ui_main_get_win())
  {
      gtk_window_set_transient_for(GTK_WINDOW(window), ui_main_get_win());
      gtk_window_set_destroy_with_parent(GTK_WINDOW(window), TRUE);
  }
  g_object_set_data(G_OBJECT(window), "button", button);

  frame = gtk_frame_new(NULL);
  gtk_container_add(GTK_CONTAINER(window), frame);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_OUT);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(frame), vbox);
  //  gtk_container_set_border_width(GTK_CONTAINER(vbox), 2);

  cal = gtk_calendar_new();
  gtk_box_pack_start(GTK_BOX(vbox), cal, TRUE, TRUE, 0);
  g_object_set(G_OBJECT(cal), "day", tv->tm_mday, NULL);
  g_object_set(G_OBJECT(cal), "month", tv->tm_mon, NULL);
  g_object_set(G_OBJECT(cal), "year", tv->tm_year, NULL);
  g_object_set_data(G_OBJECT(window), "calendar", cal);
  //  gtk_calendar_mark_day(GTK_CALENDAR(cal), tv->tm_mday);

  g_signal_connect_swapped(G_OBJECT(cal), "day-selected-double-click",
                           G_CALLBACK(on_set_date), window);
  
  frame = gtk_frame_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);

  but = gtk_button_new_from_stock(GTK_STOCK_CLOSE);
  gtk_container_add(GTK_CONTAINER(frame), but);
  gtk_widget_set_size_request(but, -1, 20);
  g_signal_connect_swapped(G_OBJECT(but), "clicked",
                           G_CALLBACK(on_set_date), window);
  


  // maybe replace with gtk function
  gdk_window_get_origin(GTK_WIDGET(button)->window, &x, &y);

  x += GTK_WIDGET(button)->allocation.x;
  y += GTK_WIDGET(button)->allocation.y + GTK_WIDGET(button)->allocation.height;
  gtk_widget_size_request(GTK_WIDGET(button), &req);
  screen = gtk_widget_get_screen(GTK_WIDGET(button));
  screen_width = gdk_screen_get_width(screen);
  screen_height = gdk_screen_get_height(screen);

  x = CLAMP(x, 0, MAX (0, screen_width - req.width));
  y = CLAMP(y, 0, MAX (0, screen_height - req.height));

  gtk_window_move(GTK_WINDOW(window), x, y);

  gtk_widget_show_all(window);

}

/** Create a date button for selecting a date. The button has the
 * named data 'date' set which is a pointer to 'struct tm'
 * @return the button
 */
GtkWidget* ui_create_date_button() {
  GtkWidget* button;
  struct tm* tv = g_malloc(sizeof(*tv));
  struct tm* tv2;
  time_t t;

  t = time(NULL);
  tv2 = localtime(&t);
  memcpy(tv, tv2, sizeof(*tv2));
  tv->tm_year += 1900;

  button = gtk_button_new_with_label
    (l_printf("%02d.%02d.%d", tv->tm_mday, tv->tm_mon, tv->tm_year));
  g_object_set_data(G_OBJECT(button), "date", tv);
  g_signal_connect_swapped(G_OBJECT(button), "destroy",
                           G_CALLBACK(g_free), tv);
  g_signal_connect(G_OBJECT(button), "clicked",
                   G_CALLBACK(on_date_set), NULL);

  return button;
}

/** Add a header to a widget. There will be no space between the
 * header and the content. The header will be left aligned.
 * @param header text of the header.
 * @param item the widget
 * @return the GtkVBox containing both the header and the item
 */
GtkWidget* ui_create_headed_item(const char* header, GtkWidget* item) {
  GtkWidget* vbox;
  GtkWidget* label;

  vbox = gtk_vbox_new(FALSE, 0);
  /*
  if (strchr(header, '_')) {
    label = gtk_label_new("");
    gtk_label_set_text_with_mnemonic(GTK_LABEL(label), header);
    gtk_label_set_mnemonic_widget(GTK_LABEL(label), item);
  } else
  */
  {
    label = gtk_label_new(header);
  }
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
  gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);

  gtk_box_pack_start(GTK_BOX(vbox), item, TRUE, TRUE, 0);
  gtk_widget_show_all(vbox);
  
  return vbox;
}

/** Same as a normal headed item. Additionally the header is added to
 * a hbox to which you can add more things.
 * @param header the header text
 * @param item the content
 * @param[out] hbox which will hold hbox of the header, or NULL
 * @return the GtkVBox containing the header box and the item
 */
GtkWidget*
ui_create_advanced_headed_item(const char* header, GtkWidget* item, GtkWidget** hbox) {
  GtkWidget* vbox;
  GtkWidget* hbox0;
  GtkWidget* label;

  vbox = gtk_vbox_new(FALSE, 0);

  hbox0 = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(vbox), hbox0, FALSE, FALSE, 0);
  
  if (strchr(header, '_')) {
    label = gtk_label_new("");
    gtk_label_set_text_with_mnemonic(GTK_LABEL(label), header);
    gtk_label_set_mnemonic_widget(GTK_LABEL(label), item);
  } else {
    label = gtk_label_new(header);
  }
  gtk_box_pack_start(GTK_BOX(hbox0), label, FALSE, FALSE, 0);

  if (hbox) *hbox = hbox0;

  gtk_box_pack_start(GTK_BOX(vbox), item, FALSE, FALSE, 0);
  gtk_widget_show_all(vbox);
  
  return vbox;
}

/** Similar to ui_create_headed_item(). Additionally a event label is
 * to the right side of the header. This function is a wrapper for
 * ui_create_advanced_headed_item().
 * @param header the header text, will be left aligned
 * @param item content
 * @param action the action text, will be right aligned in the header.
 * @param func the function to call when the action is triggered. The
 * first argument of the function will be the content widget.
 * @return the GtkVBox containing the header and the item.
 */
GtkWidget*
ui_create_functional_headed_item(const char* header, GtkWidget* item,
                                 const char* action, GCallback func) {
  GtkWidget* result;
  GtkWidget* hbox;
  GtkWidget* label;

  result = ui_create_advanced_headed_item(header, item, &hbox);
  label = ui_create_event_label(action, func, item, TRUE);
  gtk_widget_show(label);
  gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);

  return result;
}

/** Create a scrolled area around a widget.
 * @param content the content of the scrollable area.
 * @param s TRUE to add a top and bottom separator line outside the
 * scrolled area
 * @return a GtkVBox or GtkScrolledWindow, depending on s
 */
GtkWidget* ui_create_scrolled_area(GtkWidget* content, gboolean s) {
  GtkWidget* vbox;
  GtkWidget* sep;
  GtkWidget* sw;

  if (s) {
    vbox = gtk_vbox_new(FALSE, 0);
  
    sep = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 0);
    
    sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_box_pack_start(GTK_BOX(vbox), sw, TRUE, TRUE, 0);
    gtk_scrolled_window_set_shadow_type
      (GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_NONE);
    gtk_scrolled_window_set_policy
      (GTK_SCROLLED_WINDOW(sw), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
    
    gtk_scrolled_window_add_with_viewport
      (GTK_SCROLLED_WINDOW(sw), content);
    gtk_viewport_set_shadow_type
      (GTK_VIEWPORT(gtk_bin_get_child(GTK_BIN(sw))), GTK_SHADOW_NONE);
    
    sep = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 0);
    
    gtk_widget_show_all(vbox);
    return vbox;
  } else {
    sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_shadow_type
      (GTK_SCROLLED_WINDOW(sw), GTK_SHADOW_NONE);
    gtk_scrolled_window_set_policy
      (GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    
    gtk_scrolled_window_add_with_viewport
      (GTK_SCROLLED_WINDOW(sw), content);
    gtk_viewport_set_shadow_type
      (GTK_VIEWPORT(gtk_bin_get_child(GTK_BIN(sw))), GTK_SHADOW_NONE);
    
    gtk_widget_show_all(sw);
    gtk_widget_set_size_request(sw, 10, 10);
    return sw;
  }
}

/** Creates a PangoLayout for a widget.
 * @param widget a widget
 * @return the layout object
 */
PangoLayout* ui_get_text_layout(GtkWidget* widget) {
  g_assert(GTK_IS_WIDGET(widget));
  if (!TextLayout) {
    PangoContext* context =
      gtk_widget_get_pango_context(widget);
    TextLayout = pango_layout_new(context);
  }

  return TextLayout;
}

/** Get the width and height of a text for a layout.
 * @param layout the layout to be used.
 * @param text the text which should be used.
 * @param[out] width the resulting width, or NULL
 * @param[out] height the resulting height, or NULL
 * @note The text will be set in the layout, so you can use the layout
 * directly afterwards to draw the text. No need to call
 * pango_layout_set_text().
 * @see ui_get_text_layout()
 */
void ui_layout_get_ext(PangoLayout* layout, const char* text,
                       int* width, int* height) {
  PangoRectangle rect;
  pango_layout_set_text(layout, text, -1);
  pango_layout_get_pixel_extents(layout, &rect, NULL);
  if (width) *width = rect.width;
  if (height) *height = rect.height;
}

/**
 * Show file chooser dialog.
 * @return newly allocated filename which was selected, or NULL if none was selected.
 */
char* ui_select_file(const char* title, const char* current, GtkWindow* parent)
{
    GtkWidget *dialog;
    char* result = NULL;
    
    dialog = gtk_file_chooser_dialog_new
        (title, parent,
         GTK_FILE_CHOOSER_ACTION_OPEN,
         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
         GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
         NULL);
    
    if (current)
        gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(dialog), current);
    
    if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
    {
        result = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
    }
    gtk_widget_destroy(dialog);
    return result;
}

/**
 * Show file chooser dialog.
 * @return newly allocated filename which was selected, or NULL if none was selected.
 */
char* ui_select_folder(const char* title, const char* current, GtkWindow* parent)
{
    GtkWidget *dialog;
    char* result = NULL;

    dialog = gtk_file_chooser_dialog_new
        (title, parent,
         GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
         GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
         NULL);
    
    if (current)
/*         gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), current); */
        gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(dialog), current);
    
    if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
    {
        result = gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(dialog));
    }
    gtk_widget_destroy(dialog);
    return result;
}


#define MARGIN_INDENT 20
/** Create a section with a bold header and a content box, which is
 * indented by 20 pixels. This function can be used by callbacks that
 * are passed to ui_pref_register_leaf()
 * @param header the header text
 * @param help some help text, or NULL. If there is a help text, a
 * small help icon will appear next to the header. Moving the mouse
 * over it will create a tooltip containing the help text.
 * @param[out] hbox the resulting hbox for the content under the
 * header. This box already contains an invisible item of width 20.
 * @param[out] header_hbox the resulting hbox of the header which can be
 * used to add additional stuff there, or NULL
 * @return the GtkVBox containing all the things
 */
GtkWidget* ui_create_section(const char* header, const char* help,
                             GtkWidget** hbox, GtkWidget** header_hbox) {
  GtkWidget* box;
  GtkWidget* label;
  GtkWidget* hbox2;

  box = gtk_vbox_new(FALSE, 5);
  
  hbox2 = gtk_hbox_new(FALSE, 5);
  gtk_widget_show(hbox2);
  gtk_box_pack_start(GTK_BOX(box), hbox2, FALSE, FALSE, 0);
  
  label = gtk_label_new(l_printf("<b>%s</b>", header));
  gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE, 0);
  if (help) {
    GtkWidget* ebox = gtk_event_box_new();
    GtkWidget* pixmap =
      gtk_image_new_from_stock
      (GTK_STOCK_DIALOG_INFO/*GTK_STOCK_HELP*/, GTK_ICON_SIZE_MENU);
    gtk_widget_set_events(ebox,
			  GDK_ENTER_NOTIFY_MASK |
			  GDK_LEAVE_NOTIFY_MASK);
    gtk_widget_show(ebox);
    gtk_widget_show(pixmap);
    gtk_container_add(GTK_CONTAINER(ebox), pixmap);
    gtk_box_pack_start(GTK_BOX(hbox2), ebox, FALSE, FALSE, 0);
    gtk_widget_set_tooltip_text(ebox, help);
  }
  if (header_hbox) *header_hbox = hbox2;

  *hbox = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(*hbox);
  gtk_box_pack_start(GTK_BOX(box), *hbox, TRUE, TRUE, 0);
  label = gtk_label_new("");
  gtk_widget_show(label);
  gtk_widget_set_size_request(label, MARGIN_INDENT, -1);
  gtk_box_pack_start(GTK_BOX(*hbox), label, FALSE, FALSE, 0);

  return box;
}

GtkWidget* ui_create_expander_section(const char* header, const char* help, GtkWidget** hbox) {
  GtkWidget* exp;
  GtkWidget* label;
  GtkWidget* hbox2;

  exp = gtk_expander_new(header);
  
  hbox2 = gtk_hbox_new(FALSE, 5);
  gtk_widget_show(hbox2);
  
  label = gtk_label_new(l_printf("<b>%s</b>", header));
  gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
  gtk_widget_show(label);
  gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE, 0);
  
  if (help) {
    GtkWidget* ebox = gtk_event_box_new();
    GtkWidget* pixmap =
      gtk_image_new_from_stock
      (GTK_STOCK_DIALOG_INFO/*GTK_STOCK_HELP*/, GTK_ICON_SIZE_MENU);
    gtk_widget_set_events(ebox,
                          GDK_ENTER_NOTIFY_MASK |
                          GDK_LEAVE_NOTIFY_MASK);
    gtk_widget_show(ebox);
    gtk_widget_show(pixmap);
    gtk_container_add(GTK_CONTAINER(ebox), pixmap);
    gtk_box_pack_start(GTK_BOX(hbox2), ebox, FALSE, FALSE, 0);
    gtk_widget_set_tooltip_text(ebox, help);
  }
  gtk_expander_set_label_widget(GTK_EXPANDER(exp), hbox2);

  *hbox = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(*hbox);
  label = gtk_label_new("");
  gtk_widget_show(label);
  gtk_widget_set_size_request(label, MARGIN_INDENT, -1);
  gtk_box_pack_start(GTK_BOX(*hbox), label, FALSE, FALSE, 0);

  gtk_container_add(GTK_CONTAINER(exp), *hbox);

  return exp;
}

/** Add widgets to a table with two columns.
 * @param table the table
 * @param line the line number in the table, it will be increased automatically
 * @param item1 the left item
 * @param item2 the right item
 */
void ui_table_add_item(GtkTable* table, int* line, GtkWidget* item1, GtkWidget* item2) {

  gtk_widget_show(item1);
  gtk_table_attach(GTK_TABLE(table), item1, 0, 1, *line, *line+1,
                   GTK_FILL, GTK_FILL, 0, 0);

  gtk_widget_show(item2);
  gtk_table_attach(GTK_TABLE(table), item2, 1, 2, *line, *line+1,
                   GTK_FILL|GTK_EXPAND, GTK_FILL, 0, 0);

  (*line)++;
}

/** Add a new line to the table with two columns.
 * @param table the table
 * @param line the line number in the table, it will be increased automatically
 * @param l the first column
 * @param text the second column
 */
void ui_table_add_line(GtkTable* table, int* line, const char* l, const char* text)
{
    GtkWidget* label1 = gtk_label_new(NULL);
    GtkWidget* label2 = gtk_label_new(text);

    gtk_label_set_markup(GTK_LABEL(label1), l_printf("<b>%s:</b>", l));
    gtk_misc_set_alignment(GTK_MISC(label1), 1, 0.5);
    gtk_misc_set_alignment(GTK_MISC(label2), 0, 0.5);
    
    ui_table_add_item(table, line, label1, label2);
}

/** Add an attribute to a table with two columns.
 * @param table the table
 * @param attr the attribute to print
 * @param line the line number in the table, it will be increased automatically
 * if the attribute actually was printed.
 */
void ui_table_add_attr(GtkTable* table, int* line, attribute_t* attr) {
  const char* val = attr_fancy_print(attr);
  if (val) ui_table_add_line(table, line, _(attr_get_name(attr)), val);
}

/** Add a separator to a table with two columns.
 * @param table the table
 * @param line the line number in the table, it will be increased automatically
 */
void ui_table_add_sep(GtkTable* table, int* line) {
  GtkWidget* item = gtk_hseparator_new();

  gtk_widget_show(item);
  gtk_table_attach(GTK_TABLE(table), item, 0, 2, *line, *line+1,
                   GTK_FILL, GTK_FILL, 0, 2);

  (*line)++;
}

static int* Line = NULL;
static int Added = 0;
static int Group = 0;
static void _on_attr_print(attribute_t* attr, GtkTable* table) {
  if (attr->do_print == Group) {
    if (!Added && *Line > 0) ui_table_add_sep(table, Line);
    Added++;
    ui_table_add_attr(table, Line, attr);
  }
}

/** Add an attribute container to a table with two columns.
 * @param table the table
 * @param cont the attribute container to print
 * @param line the line number in the table, it will be increased automatically
 * @param domain the domain of attributes to add, or NULL to add all
 */
void ui_table_add_attrc(GtkTable* table, int* line, attr_container_t* cont,
                        const char* domain) {
  Line = line;
  
  Added = 0;
  Group = ATTR_IMPORTANT;
  attr_container_foreach(cont, domain, (AttrFunc)_on_attr_print, table);

  Added = 0;
  Group = ATTR_USEFUL;
  attr_container_foreach(cont, domain, (AttrFunc)_on_attr_print, table);

  Added = 0;
  Group = ATTR_UNIMPORTANT;
  attr_container_foreach(cont, domain, (AttrFunc)_on_attr_print, table);

}

/** Add a file to a table with two columns.
 * @param table the table
 * @param file the file to print
 * @param line the line number in the table, it will be increased automatically
 */
void ui_table_add_file(GtkTable* table, int* line, file_t* file)
{
    ui_table_add_line(table, line, _("File"), file->filename);
    
    if (file->mimetype)
        ui_table_add_line(table, line, _("Mimetype"), file->mimetype);
    
    ui_table_add_line(table, line, _("Size"),
                      l_printf("%s ("FF")", print_filesize(&file->size), file->size));
    
    ui_table_add_attrc(table, line, ATTRC(file), "file");
    
}

/** Add a user to a table with two columns.
 * @param table the table
 * @param user the user to print
 * @param line the line number in the table, it will be increased automatically
 */
void ui_table_add_user(GtkTable* table, int* line, user_t* user) {
  const char* nick;
  if (user_ip(user))
    nick = l_printf("%s (%s)", user->print(user), ntoa(user_ip(user)));
  else
    nick = user->print(user);
  ui_table_add_line(GTK_TABLE(table), line, _("Source"), nick);
}

static gboolean on_entry_focus_in(GtkEntry* entry)
{
    gint active = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(entry), "active"));
    gint vis = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(entry), "visibility"));

    gtk_entry_set_visibility(GTK_ENTRY(entry), (gboolean)vis);
    if (active == 1) gtk_entry_set_text(entry, "");
    return FALSE;
}

static gboolean on_entry_focus_out(GtkEntry* entry)
{
    gint active = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(entry), "active"));
    gint vis = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(entry), "visibility"));
    const char* text = gtk_entry_get_text(entry);
    
    if (active == 0) return FALSE;

    if (text && *text)
    {
        g_object_set_data(G_OBJECT(entry), "active", GINT_TO_POINTER(2));
        gtk_entry_set_visibility(GTK_ENTRY(entry), vis);
    }
    else
    {
        const char* default_text = g_object_get_data(G_OBJECT(entry), "default");
        gtk_entry_set_text(entry, default_text);
        g_object_set_data(G_OBJECT(entry), "active", GINT_TO_POINTER(1));
        gtk_entry_set_visibility(GTK_ENTRY(entry), TRUE);
    }
    return FALSE;
}

/**
 * Create new entry field.
 * @param text the text of the field.
 * @return the new entry field
 */
GtkWidget* ui_entry_new(const char* text)
{
    GtkWidget* entry = gtk_entry_new();
    if (text) gtk_entry_set_text(GTK_ENTRY(entry), text);
    return entry;
}

/**
 * Create new entry field with default text.
 * @param text the text of the field.
 * @param def text which is display if field is empty
 * @return the new entry field
 */
GtkWidget* ui_entry_new_default(const char* text, const char* def)
{
    GtkWidget* entry = ui_entry_new(text);
    if (def) ui_entry_set_default(GTK_ENTRY(entry), def);
    return entry;
}

/**
 * Create new entry field with default text for passwords.
 * @param text the text of the field.
 * @param def text which is display if field is empty
 * @return the new entry field
 */
GtkWidget* ui_entry_new_default_pass(const char* text, const char* def)
{
    GtkWidget* entry = ui_entry_new(text);
    gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
    if (def) ui_entry_set_default(GTK_ENTRY(entry), def);
    return entry;
}

/**
 * Setup an entry field that shows a default text when no text was
 * entered and the entry is not focused.
 * @param entry the entry field
 * @param default_text the default text
 */
void ui_entry_set_default(GtkEntry* entry, const char* default_text)
{
    gint active = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(entry), "active"));
    gboolean vis = gtk_entry_get_visibility(GTK_ENTRY(entry));

    g_object_set_data(G_OBJECT(entry), "visibility", GINT_TO_POINTER(vis));

    if (!default_text)
    {
        if (active == 1) gtk_entry_set_text(entry, "");
        if (active > 0)
        {
            //FIXME: disconnect signals
            g_log("GTK", G_LOG_LEVEL_WARNING, "ui_entry_set_default(): missing implementation");
        }
        g_object_set_data(G_OBJECT(entry), "active", GINT_TO_POINTER(0));
        return;
    }
    else
    {
        char* dup = g_strdup_printf("<%s>", default_text);
        const char* text = gtk_entry_get_text(entry);
        
        if (active != 1 && text && *text)
        {
            g_object_set_data(G_OBJECT(entry), "active", GINT_TO_POINTER(2));
        }
        else
        {
            gtk_entry_set_text(GTK_ENTRY(entry), dup);
            g_object_set_data(G_OBJECT(entry), "active", GINT_TO_POINTER(1));
            gtk_entry_set_visibility(GTK_ENTRY(entry), TRUE);
        }
        g_object_set_data_full(G_OBJECT(entry), "default", dup, (GDestroyNotify)g_free);

        if (active == 0)
        {
            g_signal_connect(G_OBJECT(entry), "focus-in-event",
                             G_CALLBACK(on_entry_focus_in), NULL);
            g_signal_connect(G_OBJECT(entry), "focus-out-event",
                             G_CALLBACK(on_entry_focus_out), NULL);
        }
    }
}

/**
 * Get the text of the entry. Never call on widgets that have focus.
 * @param entry the entry.
 * @return the text of the entry.
 */
const char* ui_entry_get_text(GtkEntry* entry)
{
    gint active = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(entry), "active"));
    if (active == 1) return "";
    else return gtk_entry_get_text(entry);
}

void ui_entry_set_text(GtkEntry* entry, const char* text)
{
    gtk_entry_set_text(entry, text);
    on_entry_focus_out(entry);
}

static void on_banner_size_request(
        GtkDrawingArea* da, GtkRequisition* requisition, GdkPixbuf** pix)
{
    if (pix[0] && pix[1] && pix[2])
    {
        requisition->width =
            gdk_pixbuf_get_width(pix[0]) + gdk_pixbuf_get_width(pix[2]);
        requisition->height =
            gdk_pixbuf_get_height(pix[0]);
    }
    else
    {
        requisition->width = 10;
        requisition->height = 10;
    }
}

static gboolean on_banner_expose(
        GtkWidget* widget, GdkEventExpose* event, GdkPixbuf** pix)
{
    int x = widget->allocation.x;
    int y = widget->allocation.y;
    int width  = widget->allocation.width;
    int height = widget->allocation.height;
    int w[3] = {0,0,0};
    int h[3] = {0,0,0};
    GdkWindow* window = widget->window;
    GdkGC* gc;
    const char* text = g_object_get_data(G_OBJECT(widget), "label");

    x = y = 0;
    if (pix[0]) w[0] = gdk_pixbuf_get_width(pix[0]);
    if (pix[1]) w[1] = gdk_pixbuf_get_width(pix[1]);
    if (pix[2]) w[2] = gdk_pixbuf_get_width(pix[2]);
    if (pix[0]) h[0] = gdk_pixbuf_get_height(pix[0]);
    if (pix[1]) h[1] = gdk_pixbuf_get_height(pix[1]);
    if (pix[2]) h[2] = gdk_pixbuf_get_height(pix[2]);
    
    gc = gdk_gc_new(window);

    if (pix[0])
        gdk_draw_pixbuf(window, gc, pix[0], 0, 0,
                        x, y, w[0], h[0],
                        GDK_RGB_DITHER_NONE, 0, 0);
    width -= (w[0] + w[2]);
    x += w[0];

    if (width > 0 && pix[1])
    {
        while (width >= w[1])
        {
            gdk_draw_pixbuf(window, gc, pix[1], 0, 0,
                            x, y, w[1], h[1],
                            GDK_RGB_DITHER_NONE, 0, 0);
            x += w[1];
            width -= w[1];
        }
        if (width > 0)
        {
            gdk_draw_pixbuf(window, gc, pix[1], 0, 0,
                            x, y, width, h[1],
                            GDK_RGB_DITHER_NONE, 0, 0);
            x += width;
            width = 0;
        }
    }

    if (pix[2])
        gdk_draw_pixbuf(window, gc, pix[2], 0, 0,
                        x, y, w[2], h[2],
                        GDK_RGB_DITHER_NONE, 0, 0);

    if (text)
    {
        int dx, dy;
        PangoLayout* layout = ui_get_text_layout(widget);
        ui_layout_get_ext(layout, text, &dx, &dy);
        gdk_draw_layout(window, widget->style->white_gc, 10, (height-dy)/2, layout);
    }

    return FALSE;
}

GtkWidget* ui_create_header_box(const char* text)
{
    GtkWidget* da;
    static GdkPixbuf* pix[3] = { NULL, NULL, NULL };
    
    if (!pix[0])
        pix[0] = create_pixbuf("header-logo-left.png");
    if (!pix[1])
        pix[1] = create_pixbuf("header-logo-slice.png");
    if (!pix[2])
        pix[2] = create_pixbuf("header-logo-right.png");

    da = gtk_drawing_area_new();
    g_object_set_data_full(G_OBJECT(da), "label", g_strdup(text), g_free);

    g_signal_connect(G_OBJECT(da), "expose-event",
                     G_CALLBACK(on_banner_expose), pix);
    g_signal_connect(G_OBJECT(da), "size-request",
                     G_CALLBACK(on_banner_size_request), pix);

    return da;
}

GtkWidget* ui_create_window(const char* title, unsigned width, unsigned height)
{
    GtkWidget* win;

    win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(win), title);
    if (width > 0 && height > 0)
        gtk_window_set_default_size(GTK_WINDOW(win), width, height);

    gtk_window_set_type_hint(GTK_WINDOW(win), GDK_WINDOW_TYPE_HINT_DIALOG);
    gtk_window_set_position(GTK_WINDOW(win), GTK_WIN_POS_CENTER_ON_PARENT);
    if (ui_main_get_win())
    {
        gtk_window_set_transient_for(GTK_WINDOW(win), ui_main_get_win());
        gtk_window_set_destroy_with_parent(GTK_WINDOW(win), TRUE);
    }
    return win;
}



#ifdef WIN32
#define _UI_SUPPORT_C_
#include "lgtk2corealiasdef.c"
#endif
