/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.8 $
 * $Date: 2007/11/21 20:15:25 $
 * $Author: sgop $
 */

#include "lcore/utils.h"
#include "lcore/action_item.h"
#include "lcore/l_i18n.h"
#include "ui_layout.h"
#include "ui_style.h"

#ifdef WIN32
#include "lgtk2corealias.h"
#endif

typedef struct {
  GtkTreeView* tv;
  char* layout;
  int col;
} ui_tv_t;

static int SelectionDepth = -1;
static gboolean SelectionInverted = FALSE;

void ui_tv_config_default_col(GtkTreeViewColumn* col, int c,
                              const char* layout, gboolean last) {
  (void)last;
  gtk_tree_view_column_set_resizable(col, TRUE);
  if (c >= 0)
    gtk_tree_view_column_set_sort_column_id(col, c);

  gtk_tree_view_column_set_sizing(col, GTK_TREE_VIEW_COLUMN_FIXED);
  if (layout) {
    ui_layout_monitor(l_printf("%s:size", layout), col);
    ui_layout_monitor(l_printf("%s:vis", layout), col);
  }
}
			   
static void setup_base(GtkTreeViewColumn* col, GtkCellRenderer* rend,
                       const char* name, int model_col) {
  int data_col =
    GPOINTER_TO_INT(g_object_get_data(G_OBJECT(rend), "data_col"));
  char* attr = g_object_get_data(G_OBJECT(rend), "attr");
  gtk_tree_view_column_set_attributes(col, rend, attr, data_col, NULL);

  if (ui_style_sets(name, "Background")) {
    g_object_set(G_OBJECT(rend), "cell-background-gdk",
                 ui_style_get_color(name, "Background"), NULL);
    gtk_tree_view_column_add_attribute
      (col, rend, "cell-background-set", model_col);
  } else {
    g_object_set(G_OBJECT(rend), "cell-background-set", FALSE, NULL);
  }
}
		      
static void setup_text(GtkTreeViewColumn* col, GtkCellRenderer* rend,
                       const char* name, int model_col) {

  if (ui_style_sets(name, "Text")) {
    g_object_set(G_OBJECT(rend), "foreground-gdk",
                 ui_style_get_color(name, "Text"), NULL);
    gtk_tree_view_column_add_attribute
      (col, rend, "foreground-set", model_col);
  } else {
    g_object_set(G_OBJECT(rend), "foreground-set", FALSE, NULL);
  }
  if (ui_style_sets(name, "Font")) {
    g_object_set(G_OBJECT(rend), "font", ui_style_get_font(name, "Font"), NULL);
    gtk_tree_view_column_add_attribute(col, rend, "family-set", model_col);
    gtk_tree_view_column_add_attribute(col, rend, "size-set", model_col);
    gtk_tree_view_column_add_attribute(col, rend, "variant-set", model_col);
    gtk_tree_view_column_add_attribute(col, rend, "weight-set", model_col);
    gtk_tree_view_column_add_attribute(col, rend, "style-set", model_col);
  } else {
    g_object_set(G_OBJECT(rend),
                 "family-set", FALSE, "size-set", FALSE,
                 "variant-set", FALSE, "style-set", FALSE,
                 "weight-set", FALSE, NULL);
  }

}

static void ui_marker_changed(ui_tv_t* data)
{
    GList* cols = gtk_tree_view_get_columns(data->tv);
    GList* dlist;
    
    for (dlist = cols; dlist; dlist = dlist->next)
    {
        GtkTreeViewColumn* col = dlist->data;
        GList* rends = gtk_tree_view_column_get_cell_renderers(col);
        GList* dlist2;
        
        for (dlist2 = rends; dlist2; dlist2 = dlist2->next)
        {
            GtkCellRenderer* rend = dlist2->data;
            setup_base(col, rend, data->layout, data->col);
            if (GTK_IS_CELL_RENDERER_TEXT(rend)) 
                setup_text(col, rend, data->layout, data->col);
            else if (GTK_IS_CELL_RENDERER_PIXBUF(rend))
                ;
            else if (GTK_IS_CELL_RENDERER_TOGGLE(rend))
                ;
            else
                g_log("UI", G_LOG_LEVEL_WARNING, "un-handled cell renderer");
        }
        g_list_free(rends);
    }
    g_list_free(cols);
}

static void ui_marker_changed_cb(const char* name, ui_tv_t* data) {
  (void)name;
  ui_marker_changed(data);
}

static void on_marker_destroy(ui_tv_t* data) {
  ui_style_unobserve(data->layout, (StyleFunc)ui_marker_changed_cb, data);
  g_free(data->layout);
  g_free(data);
}

/** 
 * Apply layout to lines in treeview if a certain boolean column in
 * the model has value TRUE. This only has any effect for visible
 * columns which where created with ui_tv_column_new()
 */
void ui_tv_config_line_marker(GtkTreeView* tv, const char* layout,
                              int model_col)
{
    ui_tv_t* data = g_malloc(sizeof(*data));
    
    data->tv = tv;
    data->layout = g_strdup(layout);
    data->col = model_col;
    
    ui_style_observe(layout, (StyleFunc)ui_marker_changed_cb, data);
    
    g_signal_connect_swapped(G_OBJECT(tv), "destroy",
                             G_CALLBACK(on_marker_destroy), data);
    
    ui_marker_changed(data);
}

static void
ui_tv_column_set_data(GtkTreeViewColumn* col, GtkCellRenderer* rend,
                      const char* attr, int data_col)
{
    gtk_tree_view_column_set_attributes(col, rend, attr, data_col, NULL);
    g_object_set_data(G_OBJECT(rend), "data_col", GINT_TO_POINTER(data_col));
    g_object_set_data(G_OBJECT(rend), "attr", (char*)attr);
}

GtkTreeViewColumn* ui_tv_column_new(const char* title, ...)
{
    GtkCellRenderer* rend;
    GtkTreeViewColumn* col;
    va_list ap;
    
    col = gtk_tree_view_column_new();
    gtk_tree_view_column_set_title(col, title);
    
    va_start(ap, title);
    while ((rend = va_arg(ap, GtkCellRenderer*)) != NULL)
    {
        const char* attr = va_arg(ap, char*);
        int data_col = va_arg(ap, int);
        gboolean expand = va_arg(ap, gboolean);
        
        gtk_tree_view_column_pack_start(col, rend, expand);
        ui_tv_column_set_data(col, rend, attr, data_col);
    }
    va_end(ap);
    
    return col;
}

static GList* _select_items(GList* sel, action_item_t* action) {
  static GList* result = NULL;
  GList* dlist;
  
  if (!action->check) return sel;
  if (result) {
    g_list_free(result);
    result = NULL;
  }
  for (dlist = sel; dlist; dlist = dlist->next) {
    void* real_obj = action->convert?
      action->convert(dlist->data, action->user_data):dlist->data;
    if (action->check(real_obj, action->user_data))
      result = g_list_append(result, dlist->data);
  }
  if ((action->flags & AI_TOGGLE) && result == NULL) {
    SelectionInverted = TRUE;
    return sel;
  } else {
    SelectionInverted = FALSE;
    return result;
  }
}

static void _inherit_properties(action_item_t* dest, action_item_t* src) {
  if (!dest || !src || dest == src) return;
  if (!(dest->flags & AI_LOCAL_DATA)) dest->user_data = src->user_data;
  if (!dest->convert) dest->convert = src->convert;
}

static GList* _real_objects(action_item_t* item, GList* selection) {
  static GList* result = NULL;

  if (!item->convert) return selection;
  if (result) g_list_free(result);
  result = NULL;
  for (; selection; selection = selection->next) {
    void* object = selection->data;
    void* real_obj = item->convert(object, item->user_data);
    result = g_list_append(result, real_obj);
  }
  return result;
}

static action_item_t* _real_actions(action_item_t* actions, GList* selection) {
  action_item_t* new_actions;
  GList* s;

  // return self if this is no link
  if (!actions || !actions->link) return actions;

  // get the list of real object
  s = _real_objects(actions, selection);
  // get the linked actions
  new_actions = actions->link(s, actions->user_data);
  // inherit the user data + convert function
  _inherit_properties(new_actions, actions);

  return new_actions;
}

static void count_func(GtkTreeModel* model, GtkTreePath* path,
		       GtkTreeIter* iter, GList** list) {
  void* object;

  if (SelectionDepth == 0 || SelectionDepth == gtk_tree_path_get_depth(path)) {
    gtk_tree_model_get(model, iter, 0, &object, -1);
    (*list) = g_list_append(*list, object);
  }
}

static GList* ui_get_depth_selection(GtkTreeSelection* select, int depth) {
  GList* result = NULL;

  SelectionDepth = depth;
  gtk_tree_selection_selected_foreach
    (select, (GtkTreeSelectionForeachFunc)count_func, &result);
  return result;
}

static void on_item_action(GObject* item, action_item_t* action) {

  g_assert(action);
  g_assert(action->action);

  if (action->flags & AI_GLOBAL) {
    action->action(NULL, action->user_data);
  } else {
    GList* sel = g_object_get_data(item, "Selection");
    GList* s = _select_items(sel, action);
    GList* objects = _real_objects(action, s);
    action->action(objects, action->user_data);
  }
}

static GtkWidget*
ui_create_view_popup_rec(action_item_t* actions,
                         GList* selection, unsigned num) {
  GtkWidget* menu;
  GtkWidget* submenu;
  GtkWidget* item;
  unsigned rnum;
  GList* dlist;
  gboolean last_item_was_separator = TRUE;  // dont start with seps

  if (!actions) return NULL;

  menu = gtk_menu_new();
  for (dlist = actions->children; dlist; dlist = dlist->next) {
    action_item_t* child = dlist->data;
    action_item_t* real_child;

    // inherit the user data + convert function
    _inherit_properties(child, actions);
    // get real child in case of action link
    real_child = _real_actions(child, selection);
    
    if (real_child) {
      // get the number of valid items
      GList* s = _select_items(selection, real_child);
      rnum = g_list_length(s);
      // if no valid items, continue if HIDE_0 flag is set
      if (rnum == 0 && (real_child->flags & AI_HIDE_0)) continue;
    } else {
      rnum = 0;   // not used, avoid compiler warnings
    }

    // CREATE THE ITEM
    if (!real_child) {
      // link that cannot expand, set original name and deactivate
      g_assert(child->name); // a separator should never link
      item = gtk_menu_item_new_with_label(_(child->name));
      last_item_was_separator = FALSE;
    } else if (!real_child->name) {
      // separator
      if (last_item_was_separator) continue;
      item = gtk_menu_item_new();
      last_item_was_separator = TRUE;
    } else if (real_child->children) {
      // parent node, use the name of the original child
      item = gtk_menu_item_new_with_label(_(child->name));
      last_item_was_separator = FALSE;
    } else {
      char* l;

      if (num > 1 && !(real_child->flags & AI_GLOBAL))
        l = l_printf("[%d] %s%s", rnum, _(child->name),
                     real_child->action?"":_(" (NI)"));
      else
        l = l_printf("%s%s", _(child->name),
                     real_child->action?"":_(" (NI)"));

      if (real_child->flags & AI_TOGGLE) {
        item = gtk_check_menu_item_new_with_label(l);
        if (real_child->flags & AI_RADIO)
          gtk_check_menu_item_set_draw_as_radio
            (GTK_CHECK_MENU_ITEM(item), TRUE);
        if (SelectionInverted || rnum < num)
          gtk_check_menu_item_set_active
            (GTK_CHECK_MENU_ITEM(item), TRUE);
        if (rnum < num)
          gtk_check_menu_item_set_inconsistent
            (GTK_CHECK_MENU_ITEM(item), TRUE);
      } else {
        item = gtk_menu_item_new_with_label(_(l));
      }
      last_item_was_separator = FALSE;
    }
    gtk_widget_show(item);
    gtk_container_add(GTK_CONTAINER(menu), item);
    
    // SETUP THE ITEM
    if (!real_child || !child->name || 
        (rnum == 0 && !(real_child->flags & AI_GLOBAL))) {
      gtk_widget_set_sensitive(item, FALSE);
    } else if (rnum > 1 && (real_child->flags & AI_RESTRICT_1)) {
      gtk_widget_set_sensitive(item, FALSE);
    } else if (real_child->action) {
      g_object_set_data(G_OBJECT(item), "Selection", selection);
      g_signal_connect(G_OBJECT(item), "activate",
                       G_CALLBACK(on_item_action), real_child);
    } else if (real_child->children) {
      submenu = ui_create_view_popup_rec(real_child, selection, num);
      if (submenu)
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), submenu);
      else
        gtk_widget_set_sensitive(item, FALSE);
    } else {
      gtk_widget_set_sensitive(item, FALSE);
    }
  }

  // FIXME:
  // if last item was separator, remove it...
  return menu;
}

static GtkWidget*
ui_create_view_popup(GtkTreeView* tv, action_item_t** actions) {
  GtkTreeSelection* select;
  GtkWidget* popup;
  GtkWidget** spopup;
  unsigned num_actions, num_popup;
  int i1;
  GList* selection = NULL;
  void* user_data = g_object_get_data(G_OBJECT(tv), "action_data");

  num_actions =
    GPOINTER_TO_INT(g_object_get_data(G_OBJECT(tv), "num_actions"));
  
  if (num_actions == 0) return NULL;

  spopup = g_malloc(sizeof(*spopup)*num_actions);
  select = gtk_tree_view_get_selection(tv);

  num_popup = 0;
  for (i1 = num_actions-1; i1 >= 0; i1--) {
    unsigned cnt;
    action_item_t* a;

    if (actions[i1] && !(actions[i1]->flags & AI_LOCAL_DATA))
      actions[i1]->user_data = user_data;

    // select all nodes of a particular depth (or all if num_actions==1)
    selection = ui_get_depth_selection(select, num_actions==1?0:i1+1);
    cnt = g_list_length(selection);
    
    // at least show one popup, even if no item selected
    if (cnt > 0 || (i1 == 0 && num_popup == 0)) {
      a = _real_actions(actions[i1], selection);
      spopup[i1] = ui_create_view_popup_rec(a, selection, cnt);
    } else {
      spopup[i1] = 0;
    }
    if (spopup[i1]) {
      num_popup++;
      g_signal_connect_swapped(G_OBJECT(spopup[i1]), "destroy",
			       G_CALLBACK(g_list_free), selection);
    } else {
      g_list_free(selection);
    }
  }

  if (num_popup == 0) {
    popup = NULL;
  } else if (num_popup == 1) {
    popup = NULL;
    for (i1 = 0; i1 < (int)num_actions; i1++) {
      if (spopup[i1]) {
        popup = spopup[i1];
        break;
      }
    }
  } else {
    popup = gtk_menu_new();
    for (i1 = 0; i1 < (int)num_actions; i1++) {
      if (spopup[i1]) {
        GtkWidget* item = gtk_menu_item_new_with_label(_(actions[i1]->name));
        gtk_widget_show(item);
        gtk_container_add(GTK_CONTAINER(popup), item);
        gtk_widget_show(spopup[i1]);
        gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), spopup[i1]);
      }
    }
  }
  g_free(spopup);
  return popup;
}

static gboolean
on_view_button(GtkTreeView* tv, GdkEventButton *event,
	       action_item_t** actions) {
  GtkWidget *pop;
  GtkTreeIter iter;
  GtkTreeModel* model;
  GtkTreePath* path;
  GtkTreeSelection* select;
  void* ptr;
  gboolean selected = TRUE;
  
  if (event->window != gtk_tree_view_get_bin_window(tv)) 
    return FALSE;
  model = gtk_tree_view_get_model(tv);
  
  select = gtk_tree_view_get_selection(tv);
  if (gtk_tree_view_get_path_at_pos(tv, event->x, event->y,
				    &path, NULL, NULL, NULL)) {
    gtk_tree_model_get_iter(model, &iter, path);
    selected = gtk_tree_selection_iter_is_selected(select, &iter);
    gtk_tree_model_get(model, &iter, 0, &ptr, -1);
    gtk_tree_path_free(path);
  } else {
    ptr = NULL;
  }

  if (event->button == 1) {
    return FALSE;
  } else if (event->button == 3) {
    if (!selected) {
      gtk_tree_selection_unselect_all(select);
      gtk_tree_selection_select_iter(select, &iter);
    }
    pop = ui_create_view_popup(tv, actions);
    if (pop) gtk_menu_popup(GTK_MENU(pop), NULL, NULL, NULL, NULL,
			    event->button, 
			    gdk_event_get_time((GdkEvent*)event));
    return TRUE;
  }

  return FALSE;
}

static gboolean
on_view_f10(GtkTreeView* tv, action_item_t** actions) {
  GtkWidget* pop;
  
  pop = ui_create_view_popup(tv, actions);
  if (pop) gtk_menu_popup(GTK_MENU(pop), NULL, NULL, NULL, NULL,
			  0, gdk_event_get_time((GdkEvent*)0));
  return TRUE;
}

void ui_tv_handle_button(GtkTreeView* tv, unsigned num_actions,
			 action_item_t** actions, void* user_data) {
  unsigned i1;
  action_item_t** a = NULL;

  g_assert(num_actions == 0 || actions);
  if (num_actions > 0) {
    a = g_malloc(sizeof(*a) * num_actions);
    for (i1 = 0; i1 < num_actions; i1++) {
      a[i1] = actions[i1];
    }
    g_signal_connect_swapped
      (G_OBJECT(tv), "destroy", G_CALLBACK(g_free), a);
  }

  g_object_set_data(G_OBJECT(tv), "num_actions", 
		    GINT_TO_POINTER(num_actions));
  g_object_set_data(G_OBJECT(tv), "action_data", user_data);

  g_signal_connect(G_OBJECT(tv), "popup-menu",
		   G_CALLBACK(on_view_f10), a);
  
  g_signal_connect(G_OBJECT(tv), "button-press-event",
		   G_CALLBACK(on_view_button), a);

}

#ifdef WIN32
#define _UI_TREEVIEW_C_
#include "lgtk2corealiasdef.c"
#endif
