/* Copyright (C) 2005-2006 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.15 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#ifdef WIN32
#include <ctype.h>
#endif

#include "utils.h"
#include "l_base.h"
#include "l_i18n.h"
#include "string_list.h"
#include "attribute.h"
#include "preferences.h"
#ifdef WIN32
#include "lcorealias.h"
#endif


typedef struct {
    CallFunc func;
    void* data;
    int id;
} lopster_call_t;

typedef struct {
    module_t _module;
//    GHashTable* periodic_funcs;       // periodically called functions
    GList* pixmap_dir;
    GList* notify;           // list of user notification functions
    int pending_tasks;       // number of tasks that need time before exiting
    GList* exit_funcs;       // called when exit is requested
    GList* shutdown_funcs;   // called shortly before shutdown
    int mode;
} base_module_t;

static time_t BTime = 0;
static char* LopsterHome = NULL;
static int* ARGC = NULL;
static char*** ARGV = NULL;

/*#ifdef WIN32
time_t wCurrentTime = 0;
#else
time_t CurrentTime = 0;
#endif
*/
static base_module_t* BaseModule = NULL;


/** Add a pixmap directory. This usually is not needed for a plugin,
 * if it installs it pixmaps in the default directory
 * @param dir the directory to add.
 */
void lopster_pixmap_dir_add(const char* dir)
{
    g_assert(BaseModule);
    if (BaseModule)
    {
        BaseModule->pixmap_dir =
            g_list_append(BaseModule->pixmap_dir, g_strdup(dir));
    }
}

/** Find a pixmap. The pixmap will be searched in the directories
 * added with lopster_pxmap_dir_add().
 * @param name name of the file.
 * @return the full path, or NULL if not found
 */
const char* lopster_pixmap_find(const char* name) {
    g_assert(BaseModule);
    if (BaseModule) {
        GList* dlist;
        for (dlist = BaseModule->pixmap_dir; dlist; dlist = dlist->next) {
            char* dir = dlist->data;
            char* pathname = l_printf(SDIR("%s") "%s", dir, name);
            if (g_file_test(pathname, G_FILE_TEST_EXISTS))
                return pathname;
        }
    }
    return NULL;
}

/** Install an exit handler. The function is called shortly before lopster
 * starts to shutdown all modules..
 * @param func function to be called at exit
 */
void lopster_call_at_exit(void (*func)()) {
    g_assert(BaseModule);
    if (BaseModule) {
        BaseModule->shutdown_funcs =
            g_list_append(BaseModule->shutdown_funcs, func);
    }
}


void lopster_set_home(const char* home) {
    g_free(LopsterHome);
    LopsterHome = g_strdup(home);
}

/** Get the home dir of lopster. If not overridden with a command line
 * argument, the home dir of lopster currently is $HOME/.2lopster
 * @return the home dir of lospter
 */
const char* lopster_get_home() {
    return LopsterHome;
}

time_t BootTime() {
    return BTime;
}

void lopster_set_argcv(int* argc, char*** argv) {
    ARGC = argc;
    ARGV = argv;
}
int* lopster_get_argc() {
    return ARGC;
}
char*** lopster_get_argv() {
    return ARGV;
}

/** Present a message to the user.
 * @param type see the LOPSTER_MESSAGE_ defines
 * @param header short version of the message.
 * @param text the message
 */
void lopster_notify(int type, const char* header, const char* text)
{
    g_assert(BaseModule);
    if (BaseModule)
    {
        GList* dlist;
        for (dlist = BaseModule->notify; dlist; dlist = dlist->next)
        {
            NotifyFunc func = dlist->data;
            func(type, header, text);
        }
    }
}

/** Register a handler for the user messages.
 * The gui normally uses this function to catch the user messages
 * an presents them in a dialog window.
 * @param func the function to be called on a notification
 */
void lopster_notify_register(NotifyFunc func)
{
    g_assert(BaseModule);
    if (BaseModule)
    {
        BaseModule->notify = g_list_append(BaseModule->notify, func);
    }
}

/** Register an exist task. Use this function and you will be 
 * asked on exits whether you need some more time (threads for example
 * try to finish their work). Once you finished your work you must notify
 * the core with lopster_exit_task_terminated().
 * @note you always must finish your work immediately in mode LOPSTER_EXIT_NOW
 */
void lopster_exit_task_register(ExitFunc func) {
    g_assert(BaseModule);
    BaseModule->exit_funcs = 
        g_list_append(BaseModule->exit_funcs, func);
}

static void lopster_terminate(void) {
    g_assert(BaseModule);
    if (BaseModule) {
        GList* dlist;

        g_log("CORE", G_LOG_LEVEL_DEBUG, "shutting down lopster NOW");
        BaseModule->mode = LOPSTER_EXIT_NOW;
  
        for (dlist = BaseModule->shutdown_funcs; dlist; dlist = dlist->next) {
            CallFunc func = dlist->data;
            func(NULL);
        }
    }
}

/** Request lopster to terminate. This is the usual way to exit lopster.
 * @param mode see the LOPSTER_EXIT_* defines
 */
void lopster_termination_request(int mode) {
    g_assert(BaseModule);
    if (BaseModule) {
        GList* dlist;
        g_log("CORE", G_LOG_LEVEL_DEBUG,
              "requested termination %s",
              mode==LOPSTER_EXIT_NOW?"NOW":
              mode==LOPSTER_EXIT_SOON?"SOON":
              mode==LOPSTER_EXIT_SAFE?"SAFE":"???");
        if (BaseModule->mode != LOPSTER_RUNNING) return;
    
        g_assert(mode == LOPSTER_EXIT_SAFE ||
                 mode == LOPSTER_EXIT_SOON ||
                 mode == LOPSTER_EXIT_NOW);
    
        BaseModule->mode = mode;
        BaseModule->pending_tasks = 0;  //FIXME: really reset?
    
        g_log("CORE", G_LOG_LEVEL_DEBUG,
              "asking modules for termination");
    
        for (dlist = BaseModule->exit_funcs; dlist; dlist = dlist->next) {
            ExitFunc func = dlist->data;
            if (!func(mode)) BaseModule->pending_tasks++;
        }
    
        if (BaseModule->pending_tasks > 0) {
            g_log("CORE", G_LOG_LEVEL_DEBUG,
                  "%d modules need more time", BaseModule->pending_tasks);
      
            if (mode == LOPSTER_EXIT_NOW) {
                g_log("CORE", G_LOG_LEVEL_WARNING,
                      "...but we don't have time, forcing exit");
                BaseModule->pending_tasks = 0;
            }
        }
    
        if (BaseModule->pending_tasks == 0) lopster_terminate();
    }
}

/** Current mode of lopster: LOPSTER_STARTING, etc.
 * @return the mode
 */
int lopster_get_mode() {
    g_assert(BaseModule);
    return BaseModule->mode;
}

/** Notify the core that you don't need more time any longer.
 * @see lopster_exit_task_register() 
 */
void lopster_exit_task_terminated() {
    g_assert(BaseModule);
    g_log("CORE", G_LOG_LEVEL_DEBUG,
          "exit task terminated %d", BaseModule->pending_tasks);
    BaseModule->pending_tasks--;
    if (BaseModule->pending_tasks > 0) return;
    if (BaseModule->mode == LOPSTER_EXIT_NOW ||
        BaseModule->mode == LOPSTER_EXIT_SOON ||
        BaseModule->mode == LOPSTER_EXIT_SAFE)
        lopster_terminate();
}

#ifndef HAVE_STRCASESTR
const char* strcasestr(const char* str, const char* sub) {
    const char* p;
    const char* startn = NULL;
    const char* np = NULL;

    for (p = str; *p; p++) {
        if (np) {
            if (toupper(*p) == toupper(*np)) {
                if (!*++np) return startn;
            } else np = NULL;
        } else if (toupper(*p) == toupper(*sub)) {
            np = sub + 1;
            startn = p;
        }
    }
    return NULL;
}
#endif



/** @defgroup BasePrefValues Base settings
 * @ingroup PrefModuleValues
 * 
 * <table border='0' style='border-spacing:1px;background-color:#888'>
 *  <tr style='background-color:#fff;'>
 *    <td>/core/user/nick</td>        <td>string</td>    <td>default username</td>
 *  </tr><tr style='background-color:#efe;'>
 *    <td>/core/user/pass</td>        <td>string</td>    <td>default password</td>
 *  </tr><tr style='background-color:#fff;'>
 *    <td>/core/user/away</td>        <td>boolean</td>   <td>AFK mode yes/no</td>
 *  </tr><tr style='background-color:#efe;'>
 *    <td>/core/user/away_message</td><td>string</td>    <td>away message</td>
 *  </tr><tr style='background-color:#fff;'>
 *    <td>/core/line/type</td>        <td>int</td>       <td>line type</td>
 *  </tr><tr style='background-color:#efe'>
 *    <td>/core/line/speed</td>       <td>int</td>       <td>line speed in B/s</td>
 *  </tr><tr style='background-color:#fff;'>
 *    <td>/core/desktop/browser</td>  <td>string</td>    <td>web browser</td>
 *  </tr><tr style='background-color:#efe;'>
 *    <td>/core/desktop/email</td>    <td>string</td>    <td>email program</td>
 *  </tr>
 * </table>
 *
 * @{
 * @}
 */
static void base_init() {
/*    BaseModule->periodic_funcs =
        g_hash_table_new_full(g_direct_hash, g_direct_equal,
                              NULL, (GDestroyNotify)lopster_call_destroy);
*/
    pref_register_string("/core/user/nick", TRUE, "TestUser");
    pref_register_string("/core/user/pass", TRUE, "testpass");
    pref_register_string("/core/user/description", TRUE, NULL);
    pref_register_string("/core/user/picture", TRUE, NULL);
    pref_register_bool  ("/core/user/away", FALSE, FALSE);
    pref_register_string("/core/user/away_message", TRUE, NULL);
    pref_register_int   ("/core/line/type", TRUE, LINE_OTHER);
    pref_add_constraint_minmax("/core/line/type", LINE_OTHER, LINE_T3);
    pref_register_int   ("/core/line/speed", TRUE, 0);
    pref_add_constraint_minmax("/core/line/speed", 0, 1<<30);
    pref_register_string("/core/desktop/browser", TRUE, "gnome-open");
    pref_register_string("/core/desktop/email", TRUE, "evolution");

    string_list_register("core:user:away", N_("Predefined away messages"), TRUE);

#ifdef PACKAGE_SOURCE_DIR
    lopster_pixmap_dir_add(SDIR(PACKAGE_SOURCE_DIR) "pixmaps");
#else
    lopster_pixmap_dir_add(SDIR(PACKAGE_DATA_DIR) "pixmaps");
#endif

}

static void base_run() {
    BaseModule->mode = LOPSTER_RUNNING;
   // lopster_call_periodic(1000, (CallFunc)update_time, NULL);
}

static void base_exit() {
    g_list_foreach(BaseModule->pixmap_dir, (GFunc)g_free, NULL);
    g_list_free(BaseModule->pixmap_dir);
    BaseModule->pixmap_dir = NULL;

    // cleaning delayed calls
 //   g_hash_table_destroy(BaseModule->periodic_funcs);
 //   BaseModule->periodic_funcs = NULL;
  
    g_list_free(BaseModule->shutdown_funcs);
    BaseModule->exit_funcs = NULL;

    g_list_free(BaseModule->exit_funcs);
    BaseModule->exit_funcs = NULL;

    g_list_free(BaseModule->notify);
    BaseModule->notify = NULL;

    attr_print_stats();  //FIXME: remove

    BaseModule = NULL;
}

static void base_module_init(base_module_t* module) {
    module_init(MODULE(module), "Base Module",
                (ModuleFunc)base_init, (ModuleFunc)base_run, (ModuleFunc)base_exit);
 //   module->periodic_funcs = NULL;
    module->pixmap_dir = NULL;
    module->notify = NULL;
    module->exit_funcs = NULL;
    module->shutdown_funcs = NULL;
    module->pending_tasks = 0;
    module->mode = LOPSTER_STARTING;
#ifdef WIN32
    /*wCurrentTime =*/ BTime = time(NULL);
#else
   /* CurrentTime =*/ BTime = time(NULL);
#endif
    module_require(pref_module_instance);
    module_require(string_list_module_instance);
    // module_require(timer_module_instance);
}

/** Require this module almost always.
 * @return the base module
 */
module_t* base_module_instance(void) {
    if (!BaseModule) {
        BaseModule = g_malloc0(sizeof(*BaseModule));
        base_module_init(BaseModule);
    }
    return MODULE(BaseModule);
}


#ifdef WIN32
# define _L_BASE_C_
# include "lcorealiasdef.c"
#endif
