/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.12 $
 * $Date: 2007/11/22 08:42:39 $
 * $Author: sid4u $
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#endif
#include <string.h>

#include <glib.h>

#include "buffer.h"
#include "l_http.h"
#include "l_resolve.h"
#include "buffer.h"
#include "zbuffer.h"
#include "utils.h"
#include "preferences.h"
#include "url.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

#define CRLF   "\r\n"
#define DCRLF  "\r\n\r\n"

typedef struct {
  module_t _module;
  GList* https;
} http_module_t;

static http_module_t* HttpModule = NULL;

static void http_reset(http_t* http);

void http_destroy(http_t* http)
{
    g_assert(HttpModule);
    HttpModule->https = g_list_remove(HttpModule->https, http);
    
    if (http->url) object_kill(OBJ(http->url));
    if (http->proxy)  g_free(http->proxy);
    if (http->header) g_free(http->header);
    http_tags_destroy(http->header_info);
    tcp_socket_destroy(TSOCK(http));
}

gboolean http_end(http_t* http, const char* message)
{
    switch (socket_get_status(SOCK(http)))
    {
    case ST_INPUT_COMPLETE:
        socket_set_status(SOCK(http), ST_UNKNOWN);
        http_reset(http);
        if (!http->keep_alive)
        {
            if (buffer_avail(b_socket_get_out_buffer(BSOCK(http), 0)))
                b_socket_close_when_sent(BSOCK(http), ST_DONE);
            else
                break;
        }
        return FALSE;
    }
    tcp_socket_end(TSOCK(http), message);
    return TRUE;
}

gboolean http_decompress_bz(http_t* http)
{
    buffer_t* decomp = buffer_new(1024);
    buffer_t* buffer = b_socket_get_in_buffer(BSOCK(http));
    compress_t* inflate =
        compress_new(COMPRESS_BZIP2, FALSE, buffer, decomp);
    
    //    printf("we have %d bytes\n", buffer->size);
    if (!inflate)
    {
        buffer_destroy(decomp);
        g_log("HTTP", G_LOG_LEVEL_WARNING, "could not init decomp");
        return FALSE;
    }
    if (!compress_action(inflate, TRUE))
    {
        compress_destroy(inflate);
        buffer_destroy(decomp);
        g_log("HTTP", G_LOG_LEVEL_WARNING, "could not inflate");
        return FALSE;
    }
    compress_destroy(inflate);
    
    // copy remaining data
    if (buffer_avail(buffer) > 0)
    {
        //      printf("we have %d bytes left\n", buffer->size);
        buffer_need_space(decomp, buffer_avail(buffer));
        memcpy(buffer_write_pos(decomp), buffer_read_pos(buffer),
               buffer_avail(buffer));
    }
    buffer_destroy(b_socket_set_in_buffer(BSOCK(http), decomp));
    http->content_length = decomp->size;
    return TRUE;
}

static const char* _http_load_finished(http_t* http)
{
    //  printf("load finished %d\n", http->content_type);
    if (http->content_type == CONTENT_BZIP2)
    {
        http_decompress_bz(http);
    }
    socket_set_status(SOCK(http), ST_INPUT_COMPLETE);
    return "got everything";
}
static const char* http_consume_body_chunked(http_t* http) {
  char* chunk;
  char* write_pos;
  int chunk_size;
  char* temp;
  char* lf;
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(http));
  
  //  printf("consumed chunked\n");
  g_assert(buffer);
  write_pos = chunk = buffer_read_pos(buffer);

  while ((lf = strchr(chunk, '\r')) != NULL) {
    if (lf[1] != '\n') break;
    // cut extension
    temp = strchr(chunk, ';');
    if (temp) *temp = 0;
    
    chunk_size = strtol(chunk, NULL, 16);
    if (!chunk_size) {
      // all chunks read, now pass to consumer
      buffer->size = write_pos - buffer->data;
      buffer->consumed = 0;
      /*
      printf("length compare %d / %d\n",
	     http->content_length, buffer->size);
      */
      http->content_length = buffer->size;
      return _http_load_finished(http);
    }
    if (lf+4 + chunk_size > buffer->data + buffer->size) break;

    memmove(write_pos, lf+2, chunk_size);
    write_pos += chunk_size;
    chunk = lf+4+chunk_size;
  }

  // not enough data
  if (write_pos != chunk) {
    chunk_size = (buffer_write_pos(buffer)-chunk);
    memmove(write_pos, chunk, chunk_size);
    buffer->size -= (chunk-write_pos);
    buffer->consumed = (write_pos-buffer->data);
  }
  return NULL;
  
}

static const char* http_process_header(http_t* http)
{
    char* temp;
    int res;
    
    if (!http->header) return "no header";
    
    {
        GList* dlist;
        g_message("- %s", http->header);
        for (dlist = http->header_info->tags; dlist; dlist = dlist->next) {
            http_tag_t* h = dlist->data;
            g_message("  - %20s : %s", h->name, h->value);
        }
    }
    
    if (strncmp(http->header, "HTTP/", 5))
        return "'HTTP' expected";
    
    arg(http->header, 0);
    temp = arg(NULL, 0);
    if (!temp) return "unparsable header";
    res = atoi(temp);
    temp = arg(NULL, 1);
    
    if (res != 200)
    {
        socket_set_status(SOCK(http), ST_HEADER_ERROR);
        return temp?l_strdup(temp):"http get error";
    }
    
    return NULL;
}

static int _http_get_content_type(const char* type) {
  if (!type) return CONTENT_UNKNOWN;
  if (!strcmp(type, "text/html")) return CONTENT_RAW;
  if (!strcmp(type, "text/css")) return CONTENT_RAW;
  if (!strcmp(type, "text/plain")) return CONTENT_RAW;
  if (!strcmp(type, "x-bzip2")) return CONTENT_BZIP2;
  g_log("HTTP", G_LOG_LEVEL_WARNING,
        "unknown Content-Type '%s' found", type);
  return CONTENT_UNKNOWN;
}

static const char* http_consume_header(http_t* http) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(http));
  char* header;
  char* line;
  char* lf;
  char* lf2;
  char* pos2;

  // NOTE: incoming buffers always are 0-terminated!
  // so str-functions can be used.

  header = buffer_read_pos(buffer);
  //  print_raw("header", buffer->data, buffer->size);

  line = header;
  while (1) {
    int len;

    lf = strchr(line, '\r');
    lf2 = strchr(line, '\n');
    if (lf) {
      if (lf2 < lf) {
	lf = lf2;  //FIXME: really correct?
	lf2 = lf+1;
      } else if (lf[1] == '\n') {
	lf2 = lf+2;
      } else {
	lf2 = lf+1;
      }
    } else if (lf2) {
      lf = lf2;
      lf2 = lf+1;
    } else break;

    *lf = 0;
    len = lf - line;
    
    buffer_consume_virt(buffer, lf2-line);

    // end of header?
    if (len == 0) {
      const char* res;
      buffer_consume(buffer, 0);
      b_socket_set_consume(BSOCK(http), http->process_body);
      res = http->process_header(BSOCK(http));
      if (res) return res;  // error
      //      printf("got header end %d\n", buffer->size);
      if (buffer->size || http->content_length == 0) {
	// doc complete or more input data to process
	return "AGAIN";
      } else {
	// wait for more input
	return NULL;  
      }
    } else if (!http->header) {
      http->header = g_strdup(line);
    } else if ((pos2 = strchr(line, ':')) != NULL) {
      *pos2++ = 0;
      while (*pos2 == ' ') pos2++;
      http_tags_add(http->header_info, line, pos2);
      
      b_socket_set_limit(BSOCK(http), 1024000);

      if (!strcasecmp(line, "Transfer-Encoding")) {
	//	printf("* transfer encoding is '%s'\n", pos2);
	if (!strcasecmp(pos2, "chunked")) 
	  http->process_body =
	    (BConsumeInputFunc)http_consume_body_chunked;
      } else if (!strcasecmp(line, "Content-length")) {
	//	printf("* content lenght is '%s'\n", pos2);
	http->content_length = atoi(pos2);
      } else if (!strcasecmp(line, "Content-Type")) {
	//	printf("* content type is '%s'\n", pos2);
	http->content_type = _http_get_content_type(pos2);
      } else if (!strcasecmp(line, "Connection")) {
	//	printf("* keep alive is '%s'\n", pos2);
	if (!strncasecmp(pos2, "Keep-Alive", 10)) 
	  http->keep_alive = TRUE;
	else
	  http->keep_alive = FALSE;
      }
    } else {
      g_log("HTTP", G_LOG_LEVEL_WARNING,
	    "unparsable line: '%s'", line);
    }
    line = lf2;
  }
  // free some buffer space if neccessary
  if (buffer->size == buffer->max && buffer->consumed)
    buffer_consume(buffer, 0);

  return NULL;
}

static void http_reset(http_t* http) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(http));

  buffer_consume(buffer, http->content_length);
  http->content_length = 0;
  http->keep_alive = FALSE;

  g_free(http->header);
  http->header = NULL;

  http_tags_clear(http->header_info);
  
  b_socket_set_consume
    (BSOCK(http), (BConsumeInputFunc)http_consume_header);
}

static const char* http_process_body(http_t* http)
{
    buffer_t* buffer = b_socket_get_in_buffer(BSOCK(http));
    
    buffer->data[buffer->size] = 0;
    
    if (buffer->size >= http->content_length)
    {
        return _http_load_finished(http);
    }
    return NULL;
}

void http_init(http_t* http, url_t* url)
{
    g_assert(HttpModule);
    tcp_socket_init(TSOCK(http), "HttpSocket", url->port);
    object_add_type(OBJ(http), "HttpSocket");
    OBJ(http)->destroy = (ObjectFunc)http_destroy;
    http->url = url;

    socket_set_end(SOCK(http), (SockEndFunc)http_end);
    b_socket_set_consume
        (BSOCK(http), (BConsumeInputFunc)http_consume_header);
    
    http->process_header = (BConsumeInputFunc)http_process_header;
    http->process_body = (BConsumeInputFunc)http_process_body;
    
    if (pref_get_ref("/core/http/proxy/enabled", gboolean))
    {
        http->proxy =
            g_strdup(pref_get_ptr("/core/http/proxy/host", char));
        http->proxy_port =
            pref_get_ref("/core/http/proxy/port", int);
    }
    else
    {
        http->proxy = NULL;
        http->proxy_port = 0;
    }
    
    http->header = NULL;
    http->header_info = http_tags_new();
    
    HttpModule->https = g_list_append(HttpModule->https, http);
}

http_t* http_new_from_url(const url_t* url)
{
    http_t* result = g_malloc0(sizeof(http_t));
    g_assert(HttpModule);
    http_init(result, url?url_new(url->proto, url->host, url->port, url->doc):NULL);
    return result;
}

http_t* http_new_from_string(const char* url)
{
    http_t* result = g_malloc0(sizeof(http_t));
    g_assert(HttpModule);
    http_init(result, url?url_parse(url):NULL);
    return result;
}

static void http_request(http_t* http)
{
    g_log("HTTP", G_LOG_LEVEL_DEBUG,
          "getting '%s'", url_format(http->url));

    if (http->proxy)
    {
        tcp_socket_send_str
            (TSOCK(http), l_printf("GET http://%s/%s HTTP/1.0"CRLF, 
                                   http->url->host, http->url->doc));
    }
    else
    {
        tcp_socket_send_str
            (TSOCK(http), l_printf("GET /%s HTTP/1.1"CRLF, http->url->doc));
    }
    tcp_socket_send_str
        (TSOCK(http), l_printf("Host: %s"CRLF, http->url->host));
    tcp_socket_send_str(TSOCK(http), "Connection: close"CRLF);
    tcp_socket_send_str(TSOCK(http), "Cache-Control: no-cache"CRLF);
    tcp_socket_send_str(TSOCK(http), CRLF);
    
    b_socket_start_receiving(BSOCK(http), 1024);
}

static void http_resolved(http_t* http, unsigned int ip)
{
    if (!g_list_find(HttpModule->https, http)) return;
    
    if (ip == 0)
    {
        socket_done(SOCK(http), 0, "Could not resolve address");
        return;
    }
    
    if (!tcp_socket_connect(TSOCK(http), ip, TSOCK(http)->port))
        return;
    http_request(http);
}

void http_get(http_t* http)
{
    g_assert(HttpModule);
    
    if (!http) return;
    if (!http->proxy) TSOCK(http)->port = htons(http->url->port); // htons(80);
    else TSOCK(http)->port = htons(http->proxy_port);
    
    if (!http->proxy)
        resolve_address(http->url->host, (call_func_t)http_resolved, http);
    else
        resolve_address(http->proxy, (call_func_t)http_resolved, http);
}

//////////////
void http_tags_init(http_tags_t* tags) {
  g_assert(HttpModule);
  g_assert(tags);
  tags->tags = NULL;
}
http_tags_t* http_tags_new(void) {
  http_tags_t* tags;

  g_assert(HttpModule);
  tags = g_malloc0(sizeof(*tags));
  http_tags_init(tags);
  return tags;
}
static http_tag_t*
http_tags_find_internal(http_tags_t* tags, const char* name) {
  GList* dlist;

  g_assert(tags && name);
  for (dlist = tags->tags; dlist; dlist = dlist->next) {
    http_tag_t* tag = dlist->data;
    if (!strcmp(tag->name, name)) return tag;
  }
  return NULL;
}
const char* http_tags_find(http_tags_t* tags, const char* name) {
  http_tag_t* tag;

  g_assert(HttpModule);
  tag = http_tags_find_internal(tags, name);
  if (tag) return tag->value;
  else return NULL;
}
void http_tags_add(http_tags_t* tags, const char* name, const char* value) {
  http_tag_t* tag;

  g_assert(HttpModule);
  g_assert(tags);
  g_assert(name && value);
  tag = http_tags_find_internal(tags, name);
  if (tag) {
    g_free(tag->value);
  } else {
    tag = g_malloc0(sizeof(*tag));
    tag->name = g_strdup(name);
    tags->tags = g_list_append(tags->tags, tag);
  }
  tag->value = g_strdup(value);
}
void http_tags_destroy(http_tags_t* tags) {
  g_assert(HttpModule);
  http_tags_clear(tags);
  g_free(tags);
}
void http_tags_clear(http_tags_t* tags) {
  GList* dlist;
  g_assert(HttpModule);
  g_assert(tags);
  for (dlist = tags->tags; dlist; dlist = dlist->next) {
    http_tag_t* tag = dlist->data;
    g_free(tag->name);
    g_free(tag->value);
    g_free(tag);
  }
  g_list_free(tags->tags);
  tags->tags = NULL;
}


static void https_init() {
  pref_register_bool  ("/core/http/proxy/enabled", TRUE, FALSE);
  pref_register_string("/core/http/proxy/host", TRUE, NULL);
  pref_register_int   ("/core/http/proxy/port", TRUE, 0);
  pref_add_constraint_port("/core/http/proxy/port");
}
static void https_run() {
}
static void https_exit() {
}

static void http_module_init(http_module_t* module) {
  module_init(MODULE(module), "Http Module",
	      (ModuleFunc)https_init,
	      (ModuleFunc)https_run,
	      (ModuleFunc)https_exit);

  module->https = NULL;

  module_require(pref_module_instance);
  module_require(tsocket_module_instance);
  module_require(resolve_module_instance);
  
}


module_t* http_module_instance() {
  if (!HttpModule) {
    HttpModule = g_malloc0(sizeof(*HttpModule));
    http_module_init(HttpModule);
  }
  return MODULE(HttpModule);
}

#ifdef WIN32
#define _L_HTTP_C_
#include "lcorealiasdef.c"
#endif
