/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.8 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include <string.h>

#include <glib.h>

#include "l_limit.h"
#include "preferences.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
  GHashTable* limits;
} limit_module_t;

static limit_module_t* LimitModule = NULL;

static void on_limit_changed(pref_node_t* node, limit_t* limit) {
  int* val = node->value;
  limit->limit = *val;
  object_set_dirty(OBJ(limit));
}

static void limit_destroy(limit_t* limit) {
  if (limit->pref)
    pref_unobserve(limit->pref, (ObjectFunc)on_limit_changed, limit);
  g_free(limit->pref);
  object_destroy(OBJ(limit));
}

/** Initializes the limit. Call from you derived init method.
 * @param lim the limit
 * @param name the names as passed to object_init()
 * @param pref the setting which representated the limit and which is monitored for changes, or NULL
 * @param def the default value, if settings value is given.
 */
void limit_init(limit_t* lim, const char* name,
                const char* pref, unsigned def) {
  g_assert(LimitModule);
  object_init(OBJ(lim), name, TRUE);
  object_add_type(OBJ(lim), "Limit");
  OBJ(lim)->destroy = (ObjectFunc)limit_destroy;

  lim->pref = g_strdup(pref);
  lim->limit = def;
  lim->used = 0;
  if (pref) {
    lim->limit = pref_get_ref(pref, int);
    pref_observe(pref, (ObjectFunc)on_limit_changed, lim);
  }
}

/** Create a new limit.
 * @param name the name of the limit.
 * @param pref the preferences value which defines the limit.
 * @return the new limit
 */
limit_t* limit_new(const char* name, const char* pref) {
  g_assert(LimitModule);
  if (g_hash_table_lookup(LimitModule->limits, name)) {
    g_log("LIMIT", G_LOG_LEVEL_WARNING,
          "limit '%s' already defined", name);
    return NULL;
  } else {
    limit_t* lim = g_malloc0(sizeof(*lim));
    limit_init(lim, name, pref, 3);
    g_hash_table_insert(LimitModule->limits, OBJ(lim)->name, lim);
    return lim;
  }
}

/** Find a limit by name
 * @param name the name of the limit to find.
 * @return the limit if found, NULL otherwise
 */
limit_t* limit_find(const char* name) {
  g_assert(LimitModule);
  return g_hash_table_lookup(LimitModule->limits, name);
}

/** Set the limit.
 * @param limit the limit which whould be changed.
 * @param val the limit to set.
 */
void limit_set_limit(limit_t* limit, int val) {
  g_assert(LimitModule);
  if (limit->limit == val) return;
  if (limit->pref) {
    pref_store(limit->pref, val);
  } else {
    limit->limit = val;
    object_set_dirty(OBJ(limit));
  }
}

/** Get the limit.
 * @param limit the limit.
 * @return the value of the limit.
 */
int limit_get_limit(limit_t* limit) {
  g_assert(LimitModule);
  return limit->limit;
}

/** Get the usage.
 * @param limit the limit.
 * @return the usage of the limit object.
 */
int limit_get_used(limit_t* limit) {
  g_assert(LimitModule);
  return limit->used;
}

/** Check whether the limit is reached
 * @param limit the limit object to look at.
 * @return TRUE if limit is reached, FALSE otherwise
 * @note this function basically returns limit_get_used() >= limit_get_limit()
 */
gboolean limit_reached(limit_t* limit) {
  g_assert(LimitModule);
  return (limit->used >= limit->limit);
}


/** Increase the usage of the limit(s) by 1
 * @param limits a list of limits
 */
void limits_increase(GList* limits) {
  GList* dlist;
  for (dlist = limits; dlist; dlist = dlist->next) {
    limit_t* limit = dlist->data;
    limit->used++;
  }
}

/** Decrease the usage of the limit(s) by 1
 * @param limits a list of limits
 */
void limits_decrease(GList* limits) {
  GList* dlist;
  for (dlist = limits; dlist; dlist = dlist->next) {
    limit_t* limit = dlist->data;
    g_assert(limit->used > 0);
    limit->used--;
  }
}

/** Check a list of limits whether at least one of them has reached the limit.
 * @param limits a list of limits
 * @return TRUE of at least one of them has its limit reached, FALSE otherwise
 */
gboolean limits_reached(GList* limits) {
  GList* dlist;
  for (dlist = limits; dlist; dlist = dlist->next) {
    limit_t* limit = dlist->data;
    if (limit->used >= limit->limit) {
/*       g_log("LIMIT", G_LOG_LEVEL_DEBUG, "limit '%s' reached %d/%d", */
/*             OBJ(limit)->name, limit->used, limit->limit); */
      return TRUE;
/*     } else { */
/*       g_log("LIMIT", G_LOG_LEVEL_DEBUG, "limit '%s' not reached %d/%d", */
/*             OBJ(limit)->name, limit->used, limit->limit); */
    }
  }
  return FALSE;
}

/*
static void limits_clear(GList* limits) {
  GList* dlist;
  for (dlist = limits; dlist; dlist = dlist->next) {
    limit_t* limit = dlist->data;
    limit->used = 0;
  }
}
*/




static void limits_exit() {
  g_hash_table_destroy(LimitModule->limits);
  LimitModule->limits = NULL;
  LimitModule = NULL;
}

static void limit_module_init(limit_module_t* module) {
  module_init(MODULE(module), "Limit Module",
	      NULL, NULL, (ModuleFunc)limits_exit);

  module->limits =
    g_hash_table_new_full
    ((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
     (GDestroyNotify)NULL, (GDestroyNotify)object_kill);

  module_require(pref_module_instance);
  
}


/** If you use limits, require this module.
 * @return the module
 * @see module_require()
 */
module_t* limit_module_instance() {
  if (!LimitModule) {
    LimitModule = g_malloc0(sizeof(*LimitModule));
    limit_module_init(LimitModule);
  }
  return MODULE(LimitModule);
}

#ifdef WIN32
# define _L_LIMIT_C_
# include "lcorealiasdef.c"
#endif
