/* Copyright (C) 2005-2006 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.9 $
 * $Date: 2006/09/26 16:41:38 $
 * $Author: sgop $
 */

#ifndef WIN32
#include <unistd.h>
#endif
#include <errno.h>
#include <fcntl.h>

#include "l_listen.h"
#include "utils.h"
#include "preferences.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
    module_t _module;
    GList* active;
    GList* inactive;
} listen_module_t;

static listen_module_t* ListenModule = NULL;

static void listen_port_start_listening_real(listen_t* listen)
{
    g_assert(ListenModule);
    
    if (socket_connected(SOCK(listen))) return;

    if (pref_get_ref("/core/listen/firewalled", gboolean)) return;

    if (socket_activate(SOCK(listen)))
    {
        ListenModule->inactive = g_list_remove(ListenModule->inactive, listen);
        ListenModule->active = g_list_append(ListenModule->active, listen);

        if (listen->change) listen->change(listen);
    }
}

static void listen_port_stop_listening_real(listen_t* l, gboolean apply)
{
    g_assert(ListenModule);
    
    if (!socket_connected(SOCK(l))) return;

    if (g_list_find(ListenModule->active, l))
    {
        g_log("LISTEN", G_LOG_LEVEL_DEBUG,
              "stop listening on port [%d] (TCP)", ntohs(l->port));

        socket_suspend(SOCK(l), DIR_IN|DIR_OUT);
        socket_set_fd(SOCK(l), -1, TRUE);

        ListenModule->active = g_list_remove(ListenModule->active, l);
        ListenModule->inactive = g_list_append(ListenModule->inactive, l);
    
        if (apply && l->change) l->change(l);
    }
}

void listen_port_stop_listening(listen_t* l)
{
    l->activated = FALSE;
    listen_port_stop_listening_real(l, TRUE);
}

static void listen_ports_settings_changed()
{
    gboolean fw = pref_get_ref("/core/listen/firewalled", gboolean);
    GList* dlist;
    
    g_assert(ListenModule);
    
    if (fw)
    {
        dlist = ListenModule->active;
        while (dlist)
        {
            listen_t* l = dlist->data;
            dlist = dlist->next;
            listen_port_stop_listening_real(l, TRUE);
        }
    }
    else
    {
        dlist = ListenModule->inactive;
        while (dlist)
        {
            listen_t* l = dlist->data;
            dlist = dlist->next;
            if (l->activated) listen_port_start_listening_real(l);
        }
    }
}

/** Start listening.
 * @param listen the listen object
 * @param max_con max pending connections allowed.
 */
void listen_port_start_listening(listen_t* listen, int max_con)
{
    if (max_con > 0) listen->max_con = max_con;
    listen->activated = TRUE;
    listen_port_start_listening_real(listen);
}

/** Accept the connection. This function is usually called from the
 * accept function (see listen_port_new())
 * @param socket the incoming tcp socket
 * @param fd the filedescriptor of the socket.
 * @param ip the incoming ip
 * @param port the incoming port
 */
void listen_init_accepted(tcp_socket_t* socket, int fd,
                          ip_t ip, port_t port)
{
    socket_set_fd(SOCK(socket), fd, TRUE);
    socket_set_nonblock(SOCK(socket), TRUE);
    socket->ip = ip;
    socket->port = port;
}

static void listen_handle_incoming(listen_t* list)
{
    struct sockaddr_in from;
    int len = sizeof(from);
    int fd = accept(socket_get_fd(SOCK(list)), (struct sockaddr*) &from, &len);
    tcp_socket_t* socket;

    if (fd < 0)
    {
        g_log("LISTEN", G_LOG_LEVEL_DEBUG,
              "failed to accept connection: %s (%d)", g_strerror(errno), errno);
        return;
    }
  
    socket = list->accept(fd, from.sin_addr.s_addr, from.sin_port);
    if (!socket) return;
  
    g_log("LISTEN", G_LOG_LEVEL_DEBUG,
          "accepted new connection %p [%s:%d]", socket,
          ntoa(socket->ip), ntohs(socket->port));
  
    b_socket_start_receiving(BSOCK(socket), 0);
}

/** End the listen port. Call from your ending function of the derived class.
 * @param listen the listen socket.
 * @param message ending message.
 * @return TRUE
 */
gboolean listen_port_end(listen_t* listen, const char* message)
{
    listen_port_stop_listening(listen);
    socket_end(SOCK(listen), message);
    return TRUE;
/*     return FALSE;  // test. keep socket in memory */
}

static int listen_port_activate(listen_t* list)
{
    int len = 1;
    socket_t* s = SOCK(list);
    unsigned short port;

    if (!list) return 0;
  
    g_log("LISTEN", G_LOG_LEVEL_DEBUG,
          "start listening [%s] on port %d (TCP)",
          OBJ(s)->name, ntohs(list->port));

    if (!socket_create(s, SOCK_STREAM, IPPROTO_TCP)) return 0;

    setsockopt(socket_get_fd(s), SOL_SOCKET,
               SO_REUSEADDR, (char*)&len, sizeof(len));
    setsockopt(socket_get_fd(s), SOL_SOCKET, 
               SO_KEEPALIVE, (char*)&len, sizeof(len));

    port = socket_bind(s, htonl(INADDR_ANY), list->port);
    if (!port) return 0;

    list->port = port;  // network
    port = ntohs(port); // host

    if (listen(socket_get_fd(s), list->max_con) < 0)
    {
        g_log("LISTEN", G_LOG_LEVEL_DEBUG,
              "failed to listen on port %d, aborting", port);
        socket_done(s, ST_CANT_LISTEN, "failed to listen");
        return 0;
    }

    g_log("LISTEN", G_LOG_LEVEL_DEBUG,
          "started listening on port %d", port);
    socket_resume(SOCK(list), DIR_IN);
    return 1;
}

static void listen_port_changed(pref_node_t* node, listen_t* list)
{
    int* port = node->value;
    listen_port_change_port(list, *port);
}

/** Destroy the listen socket. Call from the destroy function of your
 * derived class
 * @param l the listen socket.
 */
void listen_port_destroy(listen_t* l)
{
    g_assert(ListenModule);

    if (l->pref)
    {
        pref_unobserve(l->pref, (ObjectFunc)listen_port_changed, l);
        g_free(l->pref);
    }
    ListenModule->inactive = g_list_remove(ListenModule->inactive, l);
    socket_destroy(SOCK(l));
}

/** Initialize the listen socket. Call from the init function of your
 * derived class
 * @param list the listen socket.
 * @param pref the preferences value that specifies the port of the
 * listen socket. The listen socket will automatically react on port
 * changes.
 * @param name the name of the socket.
 * @param accept the function that is called when a new connection arrives.
 * @param change the function that is called when the port has changed, or NULL
 */
void listen_port_init(listen_t* list, const char* pref,
                      const char* name, ListenAcceptFunc accept,
                      ListenChangeFunc change)
{
    port_t port = (port_t)pref_get_ref(pref, int);

    g_assert(ListenModule);

    socket_init(SOCK(list), name, 0);
    object_add_type(OBJ(list), "ListenPort");
    OBJ(list)->destroy = (ObjectFunc)listen_port_destroy;
    socket_set_end(SOCK(list), (SockEndFunc)listen_port_end);
    socket_set_activate(SOCK(list), (SockActivateFunc)listen_port_activate);
    socket_set_handler(SOCK(list), (LIOFunc)listen_handle_incoming, DIR_IN);

    list->port = htons(port);
    list->pref = g_strdup(pref);
    list->activated = FALSE;
    list->accept = accept;
    list->change = change;

    ListenModule->inactive = g_list_append(ListenModule->inactive, list);

    if (pref)
        pref_observe(pref, (ObjectFunc)listen_port_changed, list);
}

/** Create a new tcp listen socket.
 * @param pref passed to listen_port_init()
 * @param name passed to listen_port_init()
 * @param accept passed to listen_port_init()
 * @param change passed to listen_port_init()
 * @return a new listen socket
 */
listen_t* listen_port_new(const char* pref, const char* name,
                          ListenAcceptFunc accept, ListenChangeFunc change)
{
    listen_t* list = g_malloc0(sizeof(*list));
    listen_port_init(list, pref, name, accept, change);
    return list;
}

/** Get the port of the listen socket.
 * @param list the listen socket
 */
port_t listen_port_get_port(listen_t* list)
{
    if (!list || !socket_connected(SOCK(list))) return 0;
    return ntohs(list->port);
}
 
/** Change the port of the listen socket.
 * @param l the listen socket.
 * @param port the new port
 * @note the port must be in the list of allowed ports (see port_is_allowed())
 * @note the preferences setting (if there is any) is not changed.
 */
void listen_port_change_port(listen_t* l, port_t port)
{
    if (l->port == htons(port)) return;
    
    g_log("LISTEN", G_LOG_LEVEL_DEBUG,
          "changing listen port %d -> %d", ntohs(l->port), port);
    
    if (socket_connected(SOCK(l)))
    {
        listen_port_stop_listening_real(l, FALSE);
        l->port = htons(port);
        listen_port_start_listening_real(l);
    }
    else
    {
        l->port = htons(port);
    }
}

/////////////////////////////////////////////////////////

static void listen_init()
{
    pref_register_string("/core/listen/range", TRUE, "1025-65535");
    pref_register_bool  ("/core/listen/firewalled", TRUE, FALSE);

    pref_observe("/core/listen/firewalled",
                 (ObjectFunc)listen_ports_settings_changed, NULL);
    pref_observe("/core/listen/range",
                 (ObjectFunc)listen_ports_settings_changed, NULL);
}

static void listen_run()
{
}

static void listen_exit()
{
    while (ListenModule->active)
    {
        listen_t* list = ListenModule->active->data;
        socket_done(SOCK(list), ST_DONE, "ShutDown");
    }
    while (ListenModule->inactive)
    {
        listen_t* list = ListenModule->inactive->data;
        socket_done(SOCK(list), ST_DONE, "ShutDown");
    }
    ListenModule = NULL;
}

static void listen_module_init(listen_module_t* module)
{
    module_init(MODULE(module), "Listen Module",
                (ModuleFunc)listen_init,
                (ModuleFunc)listen_run,
                (ModuleFunc)listen_exit);

    module->active = NULL;
    module->inactive = NULL;
  
    module_require(pref_module_instance);
    module_require(tsocket_module_instance);
}

/** If using tcp listen sockets, require this module 
 * @see module_require()
 */
module_t* listen_module_instance()
{
    if (!ListenModule)
    {
        ListenModule = g_malloc0(sizeof(*ListenModule));
        listen_module_init(ListenModule);
    }
    return MODULE(ListenModule);
}

#ifdef WIN32
#define _L_LISTEN_C_
#include "lcorealiasdef.c"
#endif
