/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.8 $
 * $Date: 2007/11/22 08:42:39 $
 * $Author: sid4u $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define RESOLVE_INT  1800   // 30 minutes

#ifndef WIN32
# include <stdio.h>
# include <sys/types.h>
# include <sys/socket.h>
# include <netinet/in.h>
# include <arpa/inet.h>
# include <netdb.h>
#endif

#include <glib.h>

#include "l_resolve.h"
#include "l_base.h"
#include "utils.h"
#include "l_i18n.h"
#include "thread.h"

#ifdef WIN32
#include "lcorealias.h"
#endif


typedef struct {
  thread_t _thread;
  call_func_t func;
  void* data;
  char* address;
  ip_t ip;
} resolve_t;

typedef struct {
  time_t last_resolved;
  ip_t ip;
} resolve_hist_t;

typedef struct {
  module_t _module;
  GHashTable* resolves;
} resolve_module_t;

static resolve_module_t* ResolveModule = NULL;
static GStaticMutex resolver = G_STATIC_MUTEX_INIT;


static void resolve_cache(const char* address, ip_t ip) {
  resolve_hist_t* hist;

  g_log("RESOLVE", G_LOG_LEVEL_DEBUG,
        "caching [%s]->[%s]", address, ntoa(ip));

  hist = g_hash_table_lookup(ResolveModule->resolves, address);
  if (!hist) {
    hist = g_malloc(sizeof(resolve_hist_t));
    g_hash_table_insert(ResolveModule->resolves, g_strdup(address), hist);
  }
  hist->last_resolved = CurrentTime;
  hist->ip = ip;
}

static ip_t resolve_lookup(const char* address) {
  resolve_hist_t* hist;

  hist = g_hash_table_lookup(ResolveModule->resolves, address);
  
  if (!hist) {
    // not entry found
    return 0;
  } else if (hist->last_resolved + RESOLVE_INT < CurrentTime) {
    // out of date entry found
    g_log("RESOLVE", G_LOG_LEVEL_DEBUG,
          "host [%s] cached but outdated", address);
    return 0;
  } else {
    g_log("RESOLVE", G_LOG_LEVEL_DEBUG,
          "host [%s] recently cached", address);
    return hist->ip;
  }
}

static unsigned resolve_main(resolve_t* res) {
  struct hostent* h;

  g_static_mutex_lock(&resolver);
  h = gethostbyname(res->address);
  g_static_mutex_unlock(&resolver);
  
  if (h) res->ip = ((struct in_addr *) (h->h_addr_list[0]))->s_addr;
  else res->ip = 0;
  return THREAD_COMPLETED;
}
 
static void resolve_destroy(resolve_t* res) {
  g_log("RESOLVE", G_LOG_LEVEL_DEBUG,
        "destroy resolve %p", res);
  if (thread_get_status(THREAD(res)) == THREAD_COMPLETED && res->ip)
    resolve_cache(res->address, res->ip);
  res->func(res->data, res->ip);
  g_free(res->address);
  thread_destroy(THREAD(res));
}

static void resolve_init(resolve_t* res, const char* address,
                         call_func_t func, void* data) {
  thread_init(THREAD(res), _("Resolving address"), 0);
  OBJ(res)->destroy = (ObjectFunc)resolve_destroy;

  thread_set_description(THREAD(res), address);

  res->func = func;
  res->data = data;
  res->address = g_strdup(address);
  res->ip = 0;
}

static resolve_t* resolve_new(const char* address, call_func_t func, void* data) {
  resolve_t* res = g_malloc0(sizeof(*res));
  resolve_init(res, address, func, data);
  return res;
}

void resolve_address(const char* address, call_func_t func, void* data) {
  ip_t ip;
  guint32 addr;

  g_assert(ResolveModule);
  if ((ip = resolve_lookup(address)) != 0) {
    // use cached ip
    func(data, ip);
  } else if ((addr = inet_addr(address)) != INADDR_NONE) {
    // got numbers-and-dot notation
    func(data, addr);
  } else {
    // need to resolve, use thread
    resolve_t* res = resolve_new(address, func, data);
    thread_start(THREAD(res), (ThreadMainFunc)resolve_main);
  }
}

static void resolves_init() {
  ResolveModule->resolves = g_hash_table_new_full
    (g_str_hash, g_str_equal, g_free, g_free);
}

static void resolves_exit() {
  g_hash_table_destroy(ResolveModule->resolves);
  ResolveModule->resolves = NULL;
  ResolveModule = NULL;
}

static void resolve_module_init(resolve_module_t* module) {
  module_init(MODULE(module), "Resolve Module",
	      (ModuleFunc)resolves_init, NULL, (ModuleFunc)resolves_exit);

  module->resolves = NULL;

  module_require(thread_module_instance);
  module_require(base_module_instance);
}

module_t* resolve_module_instance() {
  if (!ResolveModule) {
    ResolveModule = g_malloc0(sizeof(*ResolveModule));
    resolve_module_init(ResolveModule);
  }
  return MODULE(ResolveModule);
}

#ifdef WIN32
#define _L_RESOLVE_C_
#include "lcorealiasdef.c"
#endif
