/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.26 $
 * $Date: 2007/11/21 20:18:23 $
 * $Author: sgop $
 */

#include <string.h>
#include <errno.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <fcntl.h> 
#include <stdio.h>

#include <glib.h>

#include "l_socket.h"
#include "l_limit.h"
#include "l_listen.h"
#include "l_base.h"
#include "l_timer.h"
#include "l_i18n.h"
#include "utils.h"
#include "load.h" 
#include "preferences.h"
#include "ports.h"
#include "ui.h"
 
#ifdef WIN32 
#include "lcorealias.h"
#endif

#ifndef MSG_NOSIGNAL
# define MSG_NOSIGNAL 0
#endif

struct _rate_t {
  unsigned interval;
  file_size_t* bytes;
  time_t history_time;
  time_t start_time;
};

struct _socket_priv_t {
  int fd;
  LIOFunc in_func;              // the input function
  LIOFunc out_func;             // the output function
  int timestamp;                // time of last event
  int timeout;                  // timeout, or 0
  int paused_dir;
  GList* load[2];
  GList* volume;
  rate_t* rate;
  GList* rates;
  SockEndFunc end;              // ending function
  SockActivateFunc activate;    // activate the socket
  void* data;
  unsigned status;
  unsigned dir;
};

struct _volume_t {
  limit_t _limit;
  file_size_t sum;
  int refs;
};
 
typedef struct {
    module_t _module;
    GList* sockets;
    GList* tsockets;
    GList* psockets;
    GList* volumes;
} lsocket_module_t;
 
static lsocket_module_t* LSocketModule = NULL;

// select backend. 
#ifdef WIN32
 // #ifdef USE_UI_SOCKET_WATCHER
 //   # include "l_socket_ui.h" // standard usage of select()
 // #else
    # include "l_socket_win32.h" // standard usage of select()
 // #endif
#else
# include "l_socket_unix.h"   // special usage of glib
#endif
/* # include "l_socket_std.h"    // standard usage of glib */
/* # include "l_socket_poll.h"   // standard usage of poll() */


#define DEFAULT_RATE_INTERVAL 12

static void volume_ref(volume_t* vol);
static void volume_unref(volume_t* vol);
static void rate_add(rate_t* rate, file_size_t bytes);


/** Get the file descriptor of the socket.
 * @param s the socket
 * @return the file descriptor, or -1 if no connection
 */
int socket_get_fd(socket_t* s) {
  g_assert(LSocketModule);
  return s->priv->fd;
}

/** Set the file descriptor of the socket.  This function usually is
 * used when attaching a file descriptor of an incoming connection to
 * a socket_t. Use socket_move() if you have a socket_t instead of a
 * plain file descriptor as source. The old file descriptor can be
 * closed. Also some internal things are done.
 * @param s the target socket.
 * @param fd the file descriptor, or -1 to unset.
 * @param close_old whether to close the old file descriptor
 */
void socket_set_fd(socket_t* s, int fd, gboolean close_old) {
  int add = 0;

  g_assert(LSocketModule);

  if (s->priv->fd == fd) return;

  if (s->priv->fd >= 0) {
    backend_remove_socket(s);
#ifdef WIN32
    if (close_old) closesocket(s->priv->fd);
#else
    if (close_old) close(s->priv->fd);
#endif
    s->priv->fd = -1;
    add--;
  }
  if (fd != -1) {
    add++;
    s->priv->fd = fd;
    backend_add_socket(s);
  }

  // now update the reference counter
  if (add != 0) {
    GList* dlist;
    for (dlist = s->priv->volume; dlist; dlist = dlist->next) {
      volume_t* vol = dlist->data;
      if (add > 0) volume_ref(vol);
      else volume_unref(vol);
    }
  }
}

/** The (virtual) function clean up a socket.  Suspends the socket and
 * closes the connection. If returning TRUE, the socket will be
 * destroyed. 
 * @param s the socket
 * @param message reason of ending the socket.
 * @return TRUE
 * @note Derived classes always must call this function from their
 * class_end()-functions. These end()-function are never called directly,
 * instead use socket_done()
 */
gboolean socket_end(socket_t* s, const char* message)
{
    g_assert(LSocketModule);
    g_log("SOCK", G_LOG_LEVEL_DEBUG, "ending [%s:%s] : (%s) status:%d",
          object_get_type(OBJ(s)), OBJ(s)->name, message, s->priv->status);
    socket_suspend(s, DIR_IN|DIR_OUT);
    g_assert(s->priv->paused_dir == 0);
    socket_set_fd(s, -1, TRUE);
    return TRUE;
}

/** The (virtual) destroys function. Frees the socket.
 * @param s the socket to destroy
 * @note this function may only be called from derived classes in their
 * destroy method. TO close a socket use socket_done() instead.
 */
void socket_destroy(socket_t* s) {
  g_assert(LSocketModule);
/*   g_log("SOCK", G_LOG_LEVEL_INFO, "destroying [%s:%s]", */
/*         OBJ(s)->type, OBJ(s)->name); */

  g_assert(s->priv->fd == -1);

  if (s->priv->timeout > 0)
    LSocketModule->tsockets = g_list_remove(LSocketModule->tsockets, s);
  LSocketModule->sockets = g_list_remove(LSocketModule->sockets, s);
  g_list_free(s->priv->load[0]);
  g_list_free(s->priv->load[1]);
  g_list_free(s->priv->volume);
  g_list_free(s->priv->rates);
  if (s->priv->rate) rate_destroy(s->priv->rate);
  g_free(s->priv);
  
  object_destroy(OBJ(s));
}

/** must be called exactly one time for every socket.
 * Derived classes usually call this function in their init() method.
 * @param s the socket.
 * @param name the name of the object
 * @param timeout number of seconds before socket_done() will be
 * called if doesnt show any activity. 0 for no timeout.
 */
void socket_init(socket_t* s, const char* name, int timeout) {
  g_assert(LSocketModule);
  object_init(OBJ(s), name, TRUE);
  object_add_type(OBJ(s), "Socket");
  OBJ(s)->destroy = (ObjectFunc)socket_destroy;
  s->priv = g_malloc0(sizeof(socket_priv_t));
  s->priv->fd = -1;
  s->priv->in_func = NULL;
  s->priv->out_func = NULL;
  s->priv->timestamp = CurrentTime;
  s->priv->timeout = timeout;
  s->priv->load[0] = NULL;
  s->priv->load[1] = NULL;
  s->priv->volume = NULL;
  s->priv->rate = NULL;
  s->priv->rates = NULL;
  s->priv->paused_dir = 0;
  s->priv->end = socket_end;
  s->priv->activate = NULL;
  s->priv->status = ST_UNKNOWN;
  s->priv->dir = DIR_NONE;

  socket_add_load(s, "Summary");

  LSocketModule->sockets = g_list_append(LSocketModule->sockets, s);
  if (timeout > 0) LSocketModule->tsockets = g_list_append(LSocketModule->tsockets, s);
}

/** Moves (copies) the connection from one socket to another.
 * @param dest destination socket
 * @param src source socket
 * @note If the destination has an active connection it will be
 * closed.
 */
void socket_move(socket_t* dest, socket_t* src) {
  int fd;

  g_assert(LSocketModule);
  g_log("SOCK", G_LOG_LEVEL_DEBUG, "moving [%s:%s] to [%s:%s]",
        object_get_type(OBJ(src)), OBJ(src)->name,
        object_get_type(OBJ(dest)), OBJ(dest)->name);

  socket_suspend(src, DIR_IN|DIR_OUT);
  fd = src->priv->fd;
  socket_set_fd(src, -1, FALSE);
  socket_set_fd(dest, fd, TRUE);

  // we've changed the fd of src manually
  // so we need to update the volumes manually if it was active
  if (dest->priv->fd >= 0) {
    GList* dlist;
    for (dlist = src->priv->volume; dlist; dlist = dlist->next) {
      volume_t* vol = dlist->data;
      volume_unref(vol);
    }
  }
}

/** Activate the socket. This calls the activation function.  TCP
 * sockets for example connect to an IP.
 * @param s the socket
 * @return 1 on success 0 otherwise.
 * @see socket_set_activate()
 */
int socket_activate(socket_t* s)
{
    g_assert(LSocketModule);
    g_assert(s->priv->activate);
    return s->priv->activate(s);
}

/** Set a socket blocking or non-blocking.
 * @param s the socket.
 * @param set TRUE to set non-blocking, FALSE otherwise
 */
void socket_set_nonblock(socket_t* s, gboolean set) {
  g_assert(LSocketModule);
  if (s->priv->fd < 0) return;
#ifndef WIN32
  fcntl(s->priv->fd, F_SETFL, set?O_NONBLOCK:0);
#else
  {
    u_long cmd = set;
    ioctlsocket(s->priv->fd, FIONBIO, &cmd);
  }
#endif
}

static void socket_touch(socket_t* s) {
  s->priv->timestamp = CurrentTime;
}

/** Creates a file descriptor for the socket.
 * @param s the socket.
 * @param type the type of the socket, e.g. SOCK_STREAM
 * @param proto the protocol, e.g. IPPROTO_TCP or IPPROTO_UDP
 * @return 1 on success, 0 otherwise
 * @note the socket must not be created yet.
 */
int socket_create(socket_t* s, int type, int proto) {
  g_assert(LSocketModule);
  if (s->priv->fd != -1) {
    g_log("SOCK", G_LOG_LEVEL_WARNING, "socket already created");
    return 0;
  }

  socket_set_fd(s, socket(AF_INET, type, proto), TRUE);
  
  if (s->priv->fd < 0) {
    socket_done(s, ST_CON_ERROR, "Could not create system socket");
    return 0;
  }
  socket_touch(s);
  return 1;
}

/** Set the timeout for the socket. If no data is sent or received on
 * this socket for the given time, socket_done() is called with
 * argument ST_TIMEOUT.
 * @param s the socket
 * @param timeout the timeout.
 */
void socket_set_timeout(socket_t* s, int timeout) {
  g_assert(LSocketModule);
  if (s->priv->timeout > 0 && timeout == 0)
    LSocketModule->tsockets = g_list_remove(LSocketModule->tsockets, s);
  else if (s->priv->timeout == 0 && timeout > 0)
    LSocketModule->tsockets = g_list_append(LSocketModule->tsockets, s);
  s->priv->timeout = timeout;
  socket_touch(s);
}

/** Suspend the socket. The socket no longer will send/receive data.
 * @param s the socket
 * @param dir DIR_IN | DIR_OUT
 * @note This function normally is not directly used by the end-user
 * @see socket_resume(), socket_set_handler()
 */
void socket_suspend(socket_t* s, int dir) {

  if (!s || s->priv->fd < 0) return;

  g_assert(LSocketModule);
  //  s->timestamp = CurrentTime;
  if (dir & DIR_IN) {
    if (s->priv->paused_dir & DIR_IN) {
      s->priv->paused_dir &= ~DIR_IN;
      if (s->priv->paused_dir == 0) {
        LSocketModule->psockets = g_list_remove(LSocketModule->psockets, s);
      }
    } else {
      backend_clear_read(s);
    }
  }
  if (dir & DIR_OUT) {
    if (s->priv->paused_dir & DIR_OUT) {
      s->priv->paused_dir &= ~DIR_OUT;
      if (s->priv->paused_dir == 0)
        LSocketModule->psockets = g_list_remove(LSocketModule->psockets, s);
    } else {
      backend_clear_write(s);
    }
  }
}

/** Resume sending/receiving data.  Start processing events on that
 * socket.
 * @param s the socket
 * @param dir DIR_IN | DIR_OUT
 * @note this function normally is not used by the end-user
 * @see socket_suspend(), socket_set_handler()
 */
void socket_resume(socket_t* s, int dir) {
  g_assert(LSocketModule);
  if (!s || s->priv->fd < 0) return;

  //  s->timestamp = CurrentTime;
  // only start socket if it is not paused
  if ((dir & DIR_IN) && !(s->priv->paused_dir & DIR_IN)) {
    backend_set_read(s);
  }
  if ((dir & DIR_OUT) && !(s->priv->paused_dir & DIR_OUT)) {
    backend_set_write(s);
    s->priv->timestamp = CurrentTime; // touch
  }
}

/** Set the callback function to handle events on the sockets.
 * The socket gets suspended.
 * @param s the socket.
 * @param function the function
 * @param dir DIR_IN or DIR_OUT
 * @see socket_suspend(), socket_resume()
 */
void socket_set_handler(socket_t* s, LIOFunc function, int dir) {
  
  g_assert(LSocketModule);
  if (!s || !function) return;

  if (dir != DIR_IN && dir != DIR_OUT) return;

  socket_suspend(s, dir);

  // set new functions
  if (dir == DIR_IN) {
    s->priv->in_func = function;
  } else if (dir == DIR_OUT) {
    s->priv->out_func = function;
  }
}

/** Set the ending function of the socket. This function is used by
 * socket_done() to close the socket. The result of the ending
 * function is used to decide whether the socket will be destroy or
 * not.
 * @param s the socket
 * @param func the ending function.
 * @see socket_done()
 */
void socket_set_end(socket_t* s, SockEndFunc func) {
  s->priv->end = func;
}

/** Set the activation function for this socket.  This function is
 * used by socket_activate() which usually makes the socket ready for
 * transfer after it was initialized.
 * @param s the socket.
 * @param func the activation function.
 * @note this function usually is used by derived socket classes only.
 */
void socket_set_activate(socket_t* s, SockActivateFunc func) {
  s->priv->activate = func;
}

/** Set the data of the socket. This is a user value only and is not
 * used by the socket directly.
 * @param s the socket.
 * @param data the user data
 */
void socket_set_data(socket_t* s, void* data) {
  s->priv->data = data;
}

/** Get the data of the socket.
 * @param s the socket.
 * @return the data.
 * @see socket_set_data()
 */
void* socket_get_data(socket_t* s) {
  return s->priv->data;
}

/** Get the status of the socket.
 * @param s the socket.
 * @return the current status of the socket
 */
unsigned socket_get_status(socket_t* s) {
  return s->priv->status;
}

/** Set the status of the socket.
 * @param s the socket.
 * @param status the new status.
 */
void socket_set_status(socket_t* s, unsigned status) {
  s->priv->status = status;
}


/** Shut a socket down.
 * The virtual function socket_end() is called and on return
 * value TRUE the virtual function socket_destroy() is called.
 * @param s the socket.
 * @param status the status
 * @param message shutdown information
 * @note always use this function the end/shutdown/free a socket.
 * @note the socket status only is saved if it is greate than the
 * old status (which usually is 0)
 */
void socket_done(socket_t* s, unsigned status, const char* message) {
  g_assert(LSocketModule);
  if (status > s->priv->status) s->priv->status = status;
  if (s->priv->end(s, message)) object_kill(OBJ(s));
  //FIXME: this could be a good line !? auto-reset status.
  /*   else socket_set_status(SOCK(source), ST_UNKNOWN); */
}

/** Get the socket status.
 * @param s the socket.
 * @return TRUE when connected, FALSE otherwise
 * @note incomplete implementation
 */
gboolean socket_connected(socket_t* s) {
  g_assert(LSocketModule);
  /*
  int err;
  socklen_t errlen = sizeof(err);
#ifdef _WIN32
  getsockopt(socket->priv->priv->fd, SOL_SOCKET, SO_ERROR,
	     (char*)&err, &errlen);
#else
  getsockopt(socket->priv->priv->fd, SOL_SOCKET, SO_ERROR,
	     &err, &errlen);
#endif
  if (err) return FALSE;
  else return TRUE;
  */
  if (!s || s->priv->fd == -1) return FALSE;
  else return TRUE;  // FIXME
}

static void socket_sent(socket_t* s, unsigned bytes) {
  GList* dlist;

  for (dlist = s->priv->load[1]; dlist; dlist = dlist->next) {
    load_t* load = dlist->data;
    load_add(load, bytes);
  }
  if (s->priv->dir == DIR_OUT) {
    for (dlist = s->priv->volume; dlist; dlist = dlist->next) {
      volume_t* vol = dlist->data;
      LIMIT(vol)->used += bytes;
      vol->sum += bytes;
    }
    for (dlist = s->priv->rates; dlist; dlist = dlist->next) {
      rate_t* rate = dlist->data;
      rate_add(rate, bytes);
    }
  }
}

static int socket_volume_allowed(socket_t* s, unsigned min, unsigned init) {
  GList* dlist;

  for (dlist = s->priv->volume; dlist; dlist = dlist->next) {
    volume_t* vol = dlist->data;
    limit_t* limit = dlist->data;
    int temp;
    
    g_assert(vol->refs);
    if (limit->used >= limit->limit) return min;
    temp = (limit->limit)/4/vol->refs;
    if (temp > limit->limit - limit->used) 
      temp = limit->limit-limit->used;
    if ((unsigned)temp < init) init = temp;
  }
  return init<min?min:init;
}

static void socket_unpause(socket_t* s) {
  int old_pause = s->priv->paused_dir;

  // we have to reset the variable first as socket_resume()
  // does nothing if the socket is paused.
  if (old_pause) {
    s->priv->paused_dir = 0;
    socket_resume(s, old_pause);
    LSocketModule->psockets = g_list_remove(LSocketModule->psockets, s);
  }
}

static void socket_pause(socket_t* s, int dir) {
  //  printf("socket %p pause\n", socket);
  // we could simply call socket_suspend() but we dont, because
  // it would probably remove the socket from the list of
  // paused sockets which means CPU overhead
  if (!(s->priv->paused_dir & DIR_IN) && (dir & DIR_IN)) {
    socket_suspend(s, DIR_IN);
  }
  if (!(s->priv->paused_dir & DIR_OUT) && (dir & DIR_OUT)) {
    socket_suspend(s, DIR_OUT);
  }
  if (s->priv->paused_dir == 0 && dir) {
    LSocketModule->psockets = g_list_prepend(LSocketModule->psockets, s);
  }
  s->priv->paused_dir |= dir;
}

/** Send data on the socket. Usually only used by b_socket.
 * @param s the socket
 * @param buf the buffer to send.
 * @param min min bytes to send before returning from this function
 * @param max max bytes to send
 * @return number of bytes sent, -1 on error
 * @note on error socket_done() will be called, so NEVER access the
 * socket again when this function returns -1 because the socket
 * might be destroyed by then.
 */
int socket_send(socket_t* s, const char* buf, unsigned min, unsigned max) {
  unsigned sent = 0;
  int res;
  int allowed;

  g_assert(LSocketModule);
  g_assert(s->priv->fd >= 0);
  g_assert(min <= max);

  socket_touch(s);
  while (1) {
    allowed = socket_volume_allowed(s, min, max);

    if (allowed == 0) {
      socket_pause(s, DIR_OUT);
      break;
    }

    res = send(s->priv->fd, buf, allowed, MSG_NOSIGNAL);
    if (res == -1) {
#ifndef WIN32    	
      socket_done(s, ST_ERROR, l_printf("send: %s", g_strerror(errno)));
      return res;
#else
      gboolean err;
      if (WSAGetLastError() == WSAEWOULDBLOCK)
      {
//#ifdef USE_UI_SOCKET_WATCHER      	
         socket_pause(s, DIR_OUT);
//#endif          
        break;
      }
      g_log("SOCK_UI", G_LOG_LEVEL_DEBUG, "send WSAEWOULDBLOCK ");
      err = (WSAGetLastError() != WSAEWOULDBLOCK);
      if (err)
      {
        socket_done(s, ST_ERROR, l_printf("send: %s", g_strerror(errno)));
        return res;
      }            
#endif
    }
    sent += res;

    if (sent >= min) break;
    min -= res;
    max -= res;
    buf += res;
  }
#ifdef WIN32
  //  #ifdef USE_UI_SOCKET_WATCHER
      if (allowed != 0)
      {
        socket_resume(s, DIR_OUT);
      }
 //   #endif 
#endif

  if (sent > 0) socket_sent(s, sent);

  return sent;
}

/** Same as socket_send(), but used for protocols like UDP
 * @param s the socket.
 * @param buf the buffer to send
 * @param buf_len the size of the buffer.
 * @param ip the ip of the destination
 * @param port the port of the destination
 * @return number of bytes sent, -1 on error
 */
int socket_send_to(socket_t* s, const char* buf, unsigned buf_len,
                   ip_t ip, port_t port) {
  int res;
  struct sockaddr_in to;
  int len;
  
  g_assert(LSocketModule);
  socket_touch(s);

  len = sizeof(to);
  memset(&to, 0, len);
  to.sin_family = AF_INET;
  to.sin_addr.s_addr = ip;
  to.sin_port = port;
  res = sendto(s->priv->fd, buf, buf_len, MSG_NOSIGNAL,(struct sockaddr*)&to, len);
  if (res == -1) {
    socket_done(s, ST_ERROR,
		l_printf("send: %s", g_strerror(errno)));
    return res;
  }

  if (res > 0) socket_sent(s, res);

  return res;
}

static void socket_received(socket_t* s, unsigned bytes) {
  GList* dlist;
  for (dlist = s->priv->load[0]; dlist; dlist = dlist->next) {
    load_t* load = dlist->data;
    load_add(load, bytes);
  }
  if (s->priv->dir == DIR_IN) {
    for (dlist = s->priv->volume; dlist; dlist = dlist->next) {
      volume_t* vol = dlist->data;
      LIMIT(vol)->used += bytes;
      vol->sum += bytes;
    }
    for (dlist = s->priv->rates; dlist; dlist = dlist->next) {
      rate_t* rate = dlist->data;
      rate_add(rate, bytes);
    }
  }
}

/** Similar to socket_send(). Receive some data from the socket.
 * @param s the socket
 * @param buf the buffer to read to
 * @param min min bytes to receive before returning from this function
 * @param max max bytes to receive
 * @return number of bytes received, -1 on error, 0 on close
 * @note on error socket_done() will be called, so NEVER access the
 * socket again when this function returns -1 or 0 because the socket
 * might be destroyed by then.
 */
int socket_recv(socket_t* s, char* buf, unsigned min, unsigned max) {
  int have_read = 0;
  int res;
  unsigned allowed;

  g_assert(LSocketModule);
  socket_touch(s);
  while (1) {
    allowed = socket_volume_allowed(s, min, max);

    if (allowed == 0) {
      socket_pause(s, DIR_IN);
      break;
    }

    res = recv(s->priv->fd, buf, allowed, 0);
    if (res == -1) 
    {
      gboolean err;
#ifdef WIN32
      err = (WSAGetLastError() != WSAEWOULDBLOCK);
  //  #ifdef USE_UI_SOCKET_WATCHER
      if (!err) {
         socket_pause(s, DIR_IN);
         break;
      }      
 //   #endif
#else
      err = (errno != EINPROGRESS);
#endif
      if (err)
      {
          socket_done(s, ST_ERROR,
              l_printf("recv: %s", g_strerror(errno)));
          return res;
      }

    } else if (res == 0) {
      socket_done(s, ST_CLOSED, "closed by peer");
      return res;
    }
    have_read += res;

    if ((unsigned)have_read >= min) break;
    min -= res;
    max -= res;
    buf += res;
  }
#ifdef WIN32
  //  #ifdef USE_UI_SOCKET_WATCHER
      if (allowed != 0)
      {
        socket_resume(s, DIR_IN);
      }
  //  #endif 
#endif

  if (have_read > 0) socket_received(s, have_read);

  return have_read;
}

/** Same as socket_recv(), but used for protocols like UDP
 * @param s the socket.
 * @param buf the buffer to receive to
 * @param buf_len the size of the buffer.
 * @param ip the ip of the source
 * @return number of bytes received, -1 on error, 0 on close
 * @note never acces the socket again if this function returns <= 0, because
 * the socket is destroyed in that case.
 */
int socket_recv_from(socket_t* s, char* buf, unsigned buf_len, ip_t* ip) {
  int res;
  struct sockaddr_in from;
  int len;

  g_assert(LSocketModule);
  socket_touch(s);

  len = sizeof(from);
  res = recvfrom(s->priv->fd, buf, buf_len, MSG_NOSIGNAL, (struct sockaddr*)&from, &len);
  if (res == -1) {
    socket_done(s, ST_ERROR,
                l_printf("recvfrom: %s", g_strerror(errno)));
    return res;
  } else if (res == 0) {
    socket_done(s, ST_CLOSED, "closed by peer");
    return res;
  }

  if (res > 0) {
    socket_received(s, res);
    *ip = from.sin_addr.s_addr;
  } else {
    *ip = 0;
  }

  return res;
}

/** Bind a socket to a ip and port. If the port is not in the list of
 * allowed ports, an alternative one is searched
 * @param s the socket.
 * @param ip the ip
 * @param port the port to bind
 * @return the port that was used, or 0 on error.
 */
port_t socket_bind(socket_t* s, ip_t ip, port_t port)
{
    int save;
    struct sockaddr_in server;
    
    g_assert(LSocketModule);
    g_assert(s->priv->fd != -1);
    
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = ip;
    
    port = ntohs(port);
    
    if (!port_is_allowed(port))
    {
        g_log("SOCK", G_LOG_LEVEL_DEBUG,
              "port %d is not in list of allowed port, searching...", port);
        port = port_get_next(port);
    }
    if (port == 0)
    {
        socket_done(SOCK(s), ST_CANT_BIND,
                    "could not find allowed port");
        return 0;
    }
    save = port;
    server.sin_port = htons(port);
    
    while (bind(s->priv->fd, (struct sockaddr*) &server, sizeof(server)) < 0)
    {
        g_log("SOCK", G_LOG_LEVEL_DEBUG,
              "Cannot bind port %d, searching...", port);
        port = port_get_next(port);
        if (port == save)
        {
            socket_done(SOCK(s), ST_CANT_BIND,
                        "could not find allowed port");
            return 0;
        }
        server.sin_port = htons(port);
    }
    return htons(port);
}

/** Keep track of the load. Loads are used to group sockets and
 * monitor the bandwidth for statistical purpose.
 * @param s the socket to add to the load group
 * @param name the name of the load group
 * @see load_register()
 */
void socket_add_load(socket_t* s, const char* name)
{
    load_t* load;

    g_assert(LSocketModule);
    g_assert(s && name);
    
    if ((load = load_find(name, TRUE)))
        s->priv->load[0] = g_list_append(s->priv->load[0], load);
    if ((load = load_find(name, FALSE)))
        s->priv->load[1] = g_list_append(s->priv->load[1], load);
}

/** Limit the bandwidth of the socket.  You can add as many volumes as
 * you like (for example a global and a per user volume).
 * @param s the socket.
 * @param name the name of the volume
 * @see volume_new()
 */
void socket_add_volume(socket_t* s, const char* name) {
  volume_t* volume = volume_find(name);
  if (volume)
    s->priv->volume = g_list_append(s->priv->volume, volume);
}

static gboolean socket_check_all() {
  GList* dlist;

  g_assert(LSocketModule);
  // first check all sockets that have a timeout,
  // these are mainly the running and waiting sockets
  dlist = LSocketModule->tsockets;
  while (dlist) {
    socket_t* s = dlist->data;
    time_t idle;

    dlist = dlist->next;
    g_assert(s->priv->timeout);
    idle = CurrentTime - s->priv->timestamp;
    // check timeout
    if (idle >= s->priv->timeout) {
      socket_done(s, ST_TIMEOUT,
                  l_printf("Timeout (%d secs)", s->priv->timeout));
      /*
    } else if (s->priv->output && idle > 30) {
      // couldnt send stuff for 30 seconds
      socket_done(s, ST_TIMEOUT,
		  l_printf("Fast timeout (30 secs)"));
      */
    }
  }
  // enable the paused sockets.
  while (LSocketModule->psockets) {
    socket_t* s = LSocketModule->psockets->data;
    // socket_unpause only removes from LSocketModule->psockets if socket is paused
    // so avoid an endless loop and assert if a socket is not paused
    g_assert(s->priv->paused_dir != 0);
    socket_unpause(s);
  }

  // clear volumes
  for (dlist = LSocketModule->volumes; dlist; dlist = dlist->next) {
    limit_t* volume = dlist->data;
    volume->used = 0;
    object_set_dirty(OBJ(volume));
  }
  return TRUE;
}

/** Get the local ip.
 * @param s a socket which is used to detect the dynamic ip
 * @return either the user defined static ip or the dynamic ip.
 */
ip_t get_local_ip(socket_t* s) {
  ip_t ip;
  struct sockaddr_in localaddr;
  int len = sizeof(struct sockaddr_in);

  g_assert(LSocketModule);
  ip = pref_get_ref("/core/local/ip", ip_t);
  if (pref_get_ref("/core/local/force_ip", gboolean) || !s) {
    return ip;
  } else if (ip == 0) {
    if (getsockname(s->priv->fd, (struct sockaddr *)&localaddr, &len) == 0) {
      ip = localaddr.sin_addr.s_addr;
    }
    pref_store("/core/local/ip", ip);
  }  
  return ip;
}

/** Set the local ip.  This function has no effect if lopster uses a
 * static ip.
 * @param ip the new ip.
 */
void set_local_ip(ip_t ip) {
  ip_t my_ip;

  g_assert(LSocketModule);
  my_ip = pref_get_ref("/core/local/ip", ip_t);
  if (!pref_get_ref("/core/local/force_ip", gboolean)) {
    if (my_ip == 0 || my_ip != ip) {
      if (my_ip) {
        char* str = l_strdup(ntoa(my_ip));
        g_log("SOCK", G_LOG_LEVEL_DEBUG,
              "our ip seems to have changed %s -> %s", str, ntoa(ip));
      }
      pref_store("/core/local/ip", ip);
    }
  } else if (my_ip != ip) {
    g_log("SOCK", G_LOG_LEVEL_WARNING,
          "static ip %s != received ip %s", ntoa(my_ip), ntoa(ip));
  }
}




static void volume_init(volume_t* vol, const char* name, const char* pref) {
  g_assert(LSocketModule);
  limit_init(LIMIT(vol), name, pref, 1024);
  vol->refs = 0;
}

/** Create a new volume. This is usually done in the init function of
 * modules, that defined their own volumes.
 * @param name the name of the volume for later access
 * @param pref the preferences value name which is monitored for the current limit
 * @return a freh volume
 */
volume_t* volume_new(const char* name, const char* pref) {
  volume_t* vol;

  g_assert(LSocketModule);
  g_assert(name);

  vol = g_malloc0(sizeof*(vol));
  volume_init(vol, name, pref);

  LSocketModule->volumes = g_list_append(LSocketModule->volumes, vol);
  return vol;
}

static void volume_ref(volume_t* vol) {
  vol->refs++;
}

static void volume_unref(volume_t* vol) {
  vol->refs--;
  if (vol->refs == 0) {
    LIMIT(vol)->used = 0;
    object_set_dirty(OBJ(vol));
  }
}

/** Find a registered volume by name.
 * @param name the name of the volume
 * @return the volume, or NULL if not found
 */
volume_t* volume_find(const char* name) {
  GList* dlist;

  g_assert(LSocketModule);
  for (dlist = LSocketModule->volumes; dlist; dlist = dlist->next) {
    limit_t* limit = dlist->data;
    if (!strcmp(OBJ(limit)->name, name)) return dlist->data;
  }
  g_log("VOLUME", G_LOG_LEVEL_WARNING, "volume '%s' not found", name);
  return NULL;
}

/** Get the sum of bytes that were transferred by sockets using this
 * volume.
 * @param volume the volume
 * @return the sum of bytes transferred
 */
file_size_t volume_get_sum(volume_t* volume) {
  return volume->sum;
}



static void rate_init(rate_t* rate, unsigned interval) {
  rate->interval = interval?interval:DEFAULT_RATE_INTERVAL;
  if (rate->interval < 2) rate->interval = 2;
  rate->bytes = g_malloc0(sizeof(file_size_t) * rate->interval);
  rate->history_time = rate->start_time = CurrentTime;
}

static void rate_add(rate_t* rate, file_size_t bytes) {
  unsigned pos = rate->history_time%rate->interval;
  file_size_t size = rate->bytes[pos];
  
/*   while (rate->history_time + 2*rate->interval < CurrentTime) */
/*     rate->history_time += rate->interval; */

  while (rate->history_time < CurrentTime) {
    if (++pos == rate->interval) pos = 0;
    rate->history_time++;
    rate->bytes[pos] = size;
  }
  rate->bytes[pos] += bytes;
}

/** Create a new rate object.  These object must be added to a socket
 * using socket_add_rate(). Please note that the socket does not own
 * the rate object unless you use socket_add_default_rate() which is a
 * the usual way to monitor the rate of a socket. socket_add_rate() is
 * good for grouping several sockets in one rate object. A download
 * for example groups all its sources.
 * @param interval interval the rate is averaged.
 * @return a new rate object
 */
rate_t* rate_new(unsigned interval) {
  rate_t* rate = g_malloc0(sizeof(rate_t));
  rate_init(rate, interval);
  return rate;
}

/** Destroy a rate object.
 * @param rate the rate to destroy
 */
void rate_destroy(rate_t* rate) {
  if (rate) {
    g_free(rate->bytes);
    g_free(rate);
  }
}

/** Get the transfer rate.
 * @param rate the rate object
 * @return the transfer rate in B/s
 */
int rate_get_rate(rate_t* rate) {
  if (CurrentTime <= rate->start_time) {
    return 0;
  } else if (CurrentTime - rate->history_time > (int)rate->interval-2) {
    return 0;
  } else {
    unsigned pos1 = CurrentTime - 1;
    unsigned pos2 = CurrentTime + 1 - rate->interval;
    unsigned interval;
    file_size_t bytes;

    if (pos2 < rate->start_time) pos2 = rate->start_time;
    if (pos1 > rate->history_time - 1) pos1 = rate->history_time-1;
    if (pos1 <= pos2) return 0;
    interval = pos1 - pos2;
    
    bytes = rate->bytes[pos1 % rate->interval] - rate->bytes[pos2 % rate->interval];
    
    return bytes / interval;
  }
}


/** Handle a rate object. Use this to monitor that transfered data of
 * the socket.
 * @param s the socket
 * @param rate the rate object
 * @see rate_new()
 */
void socket_add_rate(socket_t* s, rate_t* rate) {
  if (s->priv->dir == DIR_NONE) {
    g_log("SOCKET", G_LOG_LEVEL_WARNING,
          "socket has no direction set");
    g_log("SOCKET", G_LOG_LEVEL_WARNING,
          "don't forget to call socket_add_default_rate()");
  }
  s->priv->rates = g_list_append(s->priv->rates, rate);
}

/** Start monitoring the rate of the socket.
 * @param s the socket
 * @param dir DIR_IN or DIR_OUT
 * @note this function may be called only once per socket
 */
void socket_add_default_rate(socket_t* s, unsigned dir) {
  if (dir != DIR_IN && dir != DIR_OUT) {
    g_log("SOCKET", G_LOG_LEVEL_WARNING,
          "invalid direction (%u)", dir);
  } else if (s->priv->rate) {
    g_log("SOCKET", G_LOG_LEVEL_WARNING,
          "default rate already added");
  } else {
    s->priv->dir = dir;
    s->priv->rate = rate_new(0);
    socket_add_rate(s, s->priv->rate);
  }
}

/** Get the transfer rate. Before you use this function you should
 * call socket_add_default_rate().
 * @param s the socket
 * @return the rate in B/s
 */
int socket_get_rate(socket_t* s) {
  if (!s || !s->priv->rate) return 0;
  return rate_get_rate(s->priv->rate);
}


static void lsockets_init() {
  pref_register_bool  ("/core/local/force_ip", TRUE, FALSE);
  pref_register_ip    ("/core/local/ip", TRUE, 0);

/*   load_register(N_("Summary")); */
/*   load_register(N_("UDP")); */
/*   load_register(N_("TCP")); */
/*   load_register(N_("Server")); */
/*   load_register(N_("Peer")); */
/*   load_register(N_("Transfer")); */
 
  backend_init();
}
static void lsockets_run() {
  lopster_call_periodic(1000, (CallFunc)socket_check_all, NULL);
}
static void lsockets_exit() {
  g_list_foreach(LSocketModule->volumes, (GFunc)object_kill, NULL);
  g_list_free(LSocketModule->volumes);
  LSocketModule->volumes = NULL;
  /*
  while (LSocketModule->sockets) {
    socket_t* s = LSocketModule->sockets->data;
    s->status = S_CLOSED;
    s->end(s, "Shutdown");
    object_kill(OBJ(s));
  }
  g_list_free(LSocketModule->sockets);
  LSocketModule->sockets = NULL;
  g_assert(LSocketModule->psockets == NULL);
  g_assert(LSocketModule->tsockets == NULL);
  */
#ifdef WIN32
  backend_exit();
#endif
  LSocketModule = NULL;
}

static void lsocket_module_init(lsocket_module_t* module) {
  module_init(MODULE(module), "LSocket Module",
              (ModuleFunc)lsockets_init,
              (ModuleFunc)lsockets_run,
              (ModuleFunc)lsockets_exit);

  module->sockets = NULL;
  module->tsockets = NULL;
  module->psockets = NULL;
  module->volumes = NULL;

  module_require(base_module_instance);
  module_require(pref_module_instance);
  module_require(load_module_instance);
  module_require(port_module_instance);
  module_require(limit_module_instance);

}


/** Require this function if you directly use this module.
 * @return the module instance
 * @see module_require()
 */
module_t* lsocket_module_instance() {
  if (!LSocketModule) {
    LSocketModule = g_malloc0(sizeof(*LSocketModule));
    lsocket_module_init(LSocketModule);
  }
  return MODULE(LSocketModule);
}

#ifdef WIN32
#define _L_SOCKET_C_
#include "lcorealiasdef.c"
#endif
