/* Copyright (C) 2005 sgop@users.sourceforge.net
   This is free software distributed under the terms of the
   GNU Public License.  See the file COPYING for details. */

#ifndef _SOCKET_H_
#define _SOCKET_H_

//FIXME : Make some defines to handle these for mignw, and msvc
#ifndef WIN32
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include <glib.h>

#include "utils.h"
#include "buffer.h"
#include "l_limit.h"
#include "module.h"

G_BEGIN_DECLS

#define SOCK_READ      0x01
#define SOCK_WRITE     0x02
#define SOCK_ERROR     0x04

/** @defgroup SocketClasses The socket framework
 * Class hierarchy.
 * <div style='border:1px solid #aaa;padding:10px;background-color:#eee'><pre>
 * socket_t
 *   |
 *   +--- b_socket_t
 *   |      |
 *   |      +--- udp_socket_t
 *   |      |      |
 *   |      |      +--- udp_listen_t
 *   |      |
 *   |      +--- tcp_socket_t
 *   |
 *   +--- listen_t
 * </pre></div>
 */
/*@{*/
/*@}*/

/** @defgroup LSocketModule Base socket class
 * The base class of all sockets.
 * @ingroup SocketClasses
 */
/*@{*/

/** @name Base socket status */
/*@{*/
#define ST_UNKNOWN     0  ///< no special status
#define ST_CON_ERROR   1  ///< connection error
#define ST_TIMEOUT     2  ///< timeout
#define ST_ERROR       3  ///< general error
#define ST_CANT_BIND   4  ///< cannot bind socket
#define ST_CLOSED      5  ///< peer closed the connection
#define ST_DONE        6  ///< transfer is done
#define ST_DELETE      7  ///< user deleted the socket
/*@}*/

/** @name Base socket transfer direction */
/*@{*/
#define DIR_NONE    0x0  ///< don't monitor any transferred data
#define DIR_IN      0x1  ///< incoming direction
#define DIR_OUT     0x2  ///< outgoing direction
/*@}*/

typedef struct _rate_t rate_t;
typedef struct _volume_t volume_t;
typedef struct _socket_t socket_t;
typedef struct _socket_priv_t socket_priv_t;

typedef void (*LIOFunc)(socket_t*);
/** Socket read write backend function */
typedef gboolean (*SockReadWriteFunc)(socket_t*, gboolean);
/** Socket end function */
typedef gboolean (*SockEndFunc)(socket_t*, const char*);
/** Socket activate function */
typedef int (*SockActivateFunc)(socket_t*);

#define SOCK(arg) ((socket_t*)(arg))
/** The socket class */
struct _socket_t {
  object_t _object;      ///< the base class
  socket_priv_t* priv;   ///< private members
};

module_t* lsocket_module_instance();

// construction/destruction used by derived classes
void socket_init   (socket_t* s, const char* name, int timeout);
gboolean socket_end(socket_t* s, const char* message);
void socket_destroy(socket_t* s);

void socket_set_activate(socket_t* s, SockActivateFunc func);
void socket_set_handler (socket_t* s, LIOFunc function, int dir);
void socket_set_end     (socket_t* s, SockEndFunc func);
void socket_set_data    (socket_t* s, void* data);
void* socket_get_data   (socket_t* s);
unsigned socket_get_status(socket_t* s);
void socket_set_status(socket_t* s, unsigned status);

// action
int socket_activate(socket_t* s);
void socket_resume (socket_t* s, int dir);
void socket_suspend(socket_t* s, int dir);
void socket_done   (socket_t* s, unsigned status, const char* message);
void socket_move   (socket_t* dest, socket_t* src);

gboolean socket_connected(socket_t* s);
int socket_get_fd(socket_t* s);
void socket_set_fd(socket_t* s, int fd, gboolean close_old);

int socket_send(socket_t* s, const char* buf, unsigned min, unsigned max);
int socket_recv(socket_t* s, char* buf, unsigned min, unsigned max);
int socket_send_to(socket_t* s, const char* buf, unsigned buf_len,
                   ip_t ip, port_t port);
int socket_recv_from(socket_t* s, char* buf, unsigned buf_len, 
                     ip_t* ip);

port_t socket_bind(socket_t* s, ip_t ip, port_t port);

void socket_set_nonblock(socket_t* s, gboolean set);
int socket_create(socket_t* s, int type, int proto);
void socket_set_timeout(socket_t* s, int timeout);

void socket_add_load  (socket_t* s, const char* name);
void socket_add_volume(socket_t* s, const char* name);
void socket_add_rate  (socket_t* s, rate_t* rate);
void socket_add_default_rate(socket_t* s, unsigned dir);
int socket_get_rate(socket_t* s);


ip_t get_local_ip(socket_t* s);
void set_local_ip(ip_t ip);

/*@}*/

/** @defgroup SocketUtils Monitor and limit the bandwidth of sockets
 * @ingroup SocketClasses
 */
/*@{*/

/** @defgroup VolumeModule Limit the bandwidth of sockets.
 * Volumes are used to limit sockets to a certain bandwidth. The
 * volumes automatically monitor a preferences value which specifies
 * the limit. To attach a volume to a socket use
 * socket_add_volume(). If a volume is added to several sockets, then
 * the bandwidth is shared between them.
 */
/*@{*/
volume_t* volume_new(const char* name, const char* pref);
volume_t* volume_find(const char* name);
file_size_t volume_get_sum(volume_t* volume);
/*@}*/

/** @defgroup RateModule Monitor the transfer rate of sockets.
 * A plugin developer normally does not need to use these
 * functions. Instead they use and socket_get_rate() or
 * download_get_rate().
 */
/*@{*/
rate_t* rate_new(unsigned interval);
void rate_destroy(rate_t* rate);
int rate_get_rate(rate_t* rate);
/*@}*/

/*@}*/


G_END_DECLS
#endif
