/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.3 $
 * $Date: 2006/07/23 00:10:47 $
 * $Author: sgop $
 */

#ifdef WIN32
# include "poll.h"
#else
# include <sys/poll.h>
#endif // WIN32

typedef struct pollfd poll_t;

static poll_t* Polls = NULL;
static socket_t** Sockets = NULL;
static unsigned PollsAlloc = 0;
static unsigned PollsUsed = 0;
static int* FdMap = NULL;
static unsigned FdsAlloc = 0;
static unsigned FdsUsed = 0;

#ifdef WIN32
static int to_read = 0;
static int to_write = 0;
static thread_running;
static thread_t* t_socket_manager;
static GStaticMutex socket_lock = G_STATIC_MUTEX_INIT;
#endif


static void backend_add_socket(socket_t* s) {
  g_static_mutex_lock(&socket_lock);
  g_assert(s->priv->fd >= 0);
  while (s->priv->fd >= (int)FdsAlloc) {
    int num = FdsAlloc?FdsAlloc*2:16;
    int i1;
    FdMap = g_realloc(FdMap, sizeof(*FdMap)*num);
    for (i1 = FdsAlloc; i1 < num; i1++) FdMap[i1] = -1;
    FdsAlloc = num;
  }
  if (PollsUsed == PollsAlloc) {
    int num = PollsAlloc?PollsAlloc*2:16;
    Polls = g_realloc(Polls, sizeof(*Polls)*num);
    Sockets = g_realloc(Sockets, sizeof(*Sockets)*num);
    PollsAlloc = num;
  }
  // insert in socket and poll array
  Sockets[PollsUsed] = s;
  Polls[PollsUsed].fd = s->priv->fd;
  Polls[PollsUsed].events = 0;
  // store index in FdMap
  FdMap[s->priv->fd] = PollsUsed;
  // Update limits
  if (s->priv->fd >= (int)FdsUsed) FdsUsed = s->priv->fd+1;
  PollsUsed++;
  g_static_mutex_unlock(&socket_lock);
}

static void backend_remove_socket(socket_t* s) {
  int index;
  g_static_mutex_lock(&socket_lock);
  g_assert(s->priv->fd >= 0 && s->priv->fd < (int)FdsUsed);
  g_assert(PollsUsed > 0);
  index = FdMap[s->priv->fd];
  g_assert(index >= 0 && index < (int)PollsUsed);

  // move last entry to deleted index
  if (index != (int)PollsUsed-1) {
    Sockets[index] = Sockets[PollsUsed-1];
    Polls[index] = Polls[PollsUsed-1];
    // update index of moved poll
    FdMap[Polls[index].fd] = index;
  }
  // remove last entry
  Sockets[PollsUsed-1] = NULL;
  Polls[PollsUsed-1].fd = -1;
  // remove fd from map
  FdMap[s->priv->fd] = -1;
  // Update limits
  while (FdsUsed && (FdMap[FdsUsed-1] == -1)) FdsUsed--;
  PollsUsed--;
  g_static_mutex_unlock(&socket_lock);
}

static void backend_set_read(socket_t* s) {
  int index;
  g_static_mutex_lock(&socket_lock);
  to_read++;
  g_assert(s->priv->fd >= 0 && s->priv->fd < (int)FdsUsed);
  g_assert(PollsUsed > 0);
  index = FdMap[s->priv->fd];
  g_assert(index >= 0 && index < (int)PollsUsed);
  Polls[index].events |= POLLIN;
  g_static_mutex_unlock(&socket_lock);
}

static void backend_clear_read(socket_t* s) {
  int index;
  g_static_mutex_lock(&socket_lock);
  to_read--;
  g_assert(s->priv->fd >= 0 && s->priv->fd < (int)FdsUsed);
  g_assert(PollsUsed > 0);
  index = FdMap[s->priv->fd];
  g_assert(index >= 0 && index < (int)PollsUsed);
  Polls[index].events &= ~POLLIN;
  g_static_mutex_unlock(&socket_lock);
}

static void backend_set_write(socket_t* s) {
  int index;
  g_static_mutex_lock(&socket_lock);
  to_write++;
  g_assert(s->priv->fd >= 0 && s->priv->fd < (int)FdsUsed);
  g_assert(PollsUsed > 0);
  index = FdMap[s->priv->fd];
  g_assert(index >= 0 && index < (int)PollsUsed);

  Polls[index].events |= POLLOUT;
  g_static_mutex_unlock(&socket_lock);
}

static void backend_clear_write(socket_t* s) {
  int index;
  g_static_mutex_lock(&socket_lock);
  to_write--;
  if (s->priv->fd < 0 || s->priv->fd >= (int)FdsUsed) {
    g_log("SOCK", G_LOG_LEVEL_WARNING,
          "condition failed: %d/%u %s", s->priv->fd, FdsUsed,
	  OBJ(s)->name);
    g_assert(FALSE);
  }
  //  g_assert(s->priv->fd >= 0 && s->priv->fd < (int)FdsUsed);
  g_assert(PollsUsed > 0);
  index = FdMap[s->priv->fd];
  g_assert(index >= 0 && index < (int)PollsUsed);

  Polls[index].events &= ~POLLOUT;
  g_static_mutex_unlock(&socket_lock);
}

#ifdef WIN32
int poll (struct pollfd *p, int num, int timeout)
{
  struct timeval tv;
  fd_set read, write, except;
  int    i, n, ret;

  FD_ZERO (&read);
  FD_ZERO (&write);
  FD_ZERO (&except);

  n = -1;
  for (i = 0; i < num; i++)
  {
    if (p[i].fd < 0)
       continue;
    if (p[i].events & POLLIN)
       FD_SET (p[i].fd, &read);
    if (p[i].events & POLLOUT)
       FD_SET (p[i].fd, &write);
    if (p[i].events & POLLERR)
       FD_SET (p[i].fd, &except);
    if (p[i].fd > n)
       n = p[i].fd;
  }

  if (n == -1)
     return (0);

  if (timeout < 0)
     ret = select (n+1, &read, &write, &except, NULL);
  else
  {
    tv.tv_sec  = timeout / 1000;
    tv.tv_usec = 1000 * (timeout % 1000);
    ret = select (n+1, &read, &write, &except, &tv);
  }

  for (i = 0; ret >= 0 && i < num; i++)
  {
    p[i].revents = 0;
    if (FD_ISSET (p[i].fd, &read))
       p[i].revents |= POLLIN;
    if (FD_ISSET (p[i].fd, &write))
       p[i].revents |= POLLOUT;
    if (FD_ISSET (p[i].fd, &except))
       p[i].revents |= POLLERR;
  }
  return (ret); 
}
#endif

static gboolean poll_wait_and_dispatch(int* ms) {
  int ready;
  unsigned i1;
  GArray* derr;
  GArray* din;
  GArray* dout;

  /*
  static GList* OldClients = NULL;
  static GList* Clients = NULL;
  static time_t LastTime = 0;

  if (LastTime < CurrentTime) {
    int old = 0;
    int num = g_list_length(Clients);
    for (dlist = Clients; dlist; dlist = dlist->next) {
      socket_t* s = dlist->data;
      if (g_list_find(OldClients, s)) old++;
    }
    g_log("SOCK", G_LOG_LEVEL_DEBUG,
	  "-- %u/%u sockets got events (%4.1f %%)",
	  num, PollsUsed, num?(double)old*100/num:0);
    g_list_free(OldClients);
    OldClients = Clients;
    Clients = NULL;
    LastTime = CurrentTime;
  }
  */
  g_static_mutex_lock(&socket_lock);
  ready = poll(Polls, PollsUsed, *ms);
 // g_static_mutex_unlock(&socket_lock);
  if (ready == -1) {
#ifndef WIN32 // Seems any g_log here in a thread result in a crash of the apps. Why ?? FIXME
    g_log("SOCK", G_LOG_LEVEL_WARNING,
          "poll(2) failed due to %s (%d)", g_strerror(errno), errno);
#else
    printf("poll(2) failed\n");
#endif
    g_static_mutex_unlock(&socket_lock);
    return TRUE;
  }
  /*
  g_log("SOCK", G_LOG_LEVEL_DEBUG,
	"*** got %d events from %u fds", ready, PollsUsed);
  */
 // g_static_mutex_lock(&socket_lock);
  derr = g_array_new(FALSE, FALSE, sizeof(int));
  din  = g_array_new(FALSE, FALSE, sizeof(int));
  dout = g_array_new(FALSE, FALSE, sizeof(int));

  // now dispatch
  for (i1 = 0; i1 < PollsUsed; i1++) {
    if (Polls[i1].revents & (POLLNVAL | POLLHUP | POLLERR)) {
      g_array_append_val(derr, Polls[i1].fd);
      //      if (!g_list_find(Clients, s)) Clients = g_list_prepend(Clients, s);
    } else {
      if (Polls[i1].revents & POLLIN) {
        g_assert(Polls[i1].events & POLLIN);
        g_array_append_val(din, Polls[i1].fd);
        //	if (!g_list_find(Clients, s)) Clients = g_list_prepend(Clients, s);
      }
      if (Polls[i1].revents & POLLOUT) {
        g_assert(Polls[i1].events & POLLOUT);
        g_array_append_val(dout, Polls[i1].fd);
        //	if (!g_list_find(Clients, s)) Clients = g_list_prepend(Clients, s);
      }
    }
  }
  for (i1 = 0; i1 < derr->len; i1++) {
    int index = g_array_index(derr, int, i1);
    index = FdMap[index];
    if (index != -1) {
     socket_done(Sockets[index], ST_ERROR, "socket error");
    }
  }
  g_array_free(derr, TRUE);

  for (i1 = 0; i1 < din->len; i1++) {
    int index = g_array_index(din, int, i1);
    index = FdMap[index];
    if (index != -1 && (Polls[index].events && POLLIN))
    {
      UI_thread_enter();
      Sockets[index]->priv->in_func(Sockets[index]);
      UI_thread_leave();
    }
  }
  g_array_free(din, TRUE);

  for (i1 = 0; i1 < dout->len; i1++) {
    int index = g_array_index(dout, int, i1);
    index = FdMap[index];
    if (index != -1 && (Polls[index].events && POLLOUT))
    {
      UI_thread_enter();
      Sockets[index]->priv->out_func(Sockets[index]);
      UI_thread_leave();
    }
  }
  g_array_free(dout, TRUE);
  g_static_mutex_unlock(&socket_lock);

  return TRUE;
}

#ifdef WIN32
static unsigned Run(thread_t* socket_manager ) {
  static int timeout = 200;
  
  while (/*thread_running &&*/ !thread_interrupted(t_socket_manager) )
  {
      // FIXME : find a way to sleep until lopster_module starts 
      while (Polls == NULL && ( to_write == 0 && to_read == 0 ))
      {
          if (/*thread_running == FALSE ||*/ thread_interrupted(t_socket_manager)) 
              return THREAD_COMPLETED;
          Sleep(10);
      }
      poll_wait_and_dispatch(&timeout);       
  }
  return THREAD_COMPLETED;
}
#endif

static void backend_init() {
#ifdef WIN32
  thread_running = TRUE;
  //  _beginthreadex(NULL,0, (LPTHREAD_START_ROUTINE) Run, &timeout,0,&id); 
  t_socket_manager = thread_new(0);
  thread_set_force_interrupt(t_socket_manager, TRUE);
  thread_start(t_socket_manager, (ThreadMainFunc)Run);
#else
  static int timeout = 200;
  g_idle_add((GSourceFunc)poll_wait_and_dispatch, &timeout);
#endif
}

#ifdef WIN32
static void backend_end() {
  thread_running = FALSE;
}
#endif
