/* Copyright (C) 2005-2007 Lopster Team
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.3 $
 * $Date: 2007/11/21 20:18:23 $
 * $Author: sgop $
 */

#include "ui.h"
#include "l_timer.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
    PeriodicFunc func;
    void* data;
    int id;
} lopster_call_t;

typedef struct {
    module_t _module;
    GHashTable* periodic_funcs;       // periodically called functions
} timer_module_t;

#ifdef WIN32
time_t wCurrentTime = 0;
#else
time_t CurrentTime = 0;
#endif

static timer_module_t* TimerModule = NULL;

guint lopster_timeout_add (guint interval, GSourceFunc function, gpointer data)
{
    return UI_timeout_add(interval,function,data);
}

guint lopster_idle_add (GSourceFunc function, gpointer data)
{
    return UI_idle_add(function,data);
}

void lopster_timeout_remove (guint id)
{
    UI_timeout_remove(id);
}

void lopster_idle_remove(guint id)
{
    UI_idle_remove(id);
}

static void lopster_call_destroy(lopster_call_t* call)
{
    lopster_timeout_remove(call->id);
    g_free(call);
}

static gboolean on_call_periodic(lopster_call_t* call)
{
    gboolean ret = call->func(call->data);
    if (!ret) g_hash_table_remove(TimerModule->periodic_funcs, call->func);
    return ret;
}

/** Call a function periodically. The function is called until it
 * returns FALSE.
 * @param timeout interval in milliseconds.
 * @param func function to call
 * @param data userdata passed to the function
 * @note you cannot add more than one handler for a func.
 */
void lopster_call_periodic(unsigned timeout, PeriodicFunc func, void* data)
{
    g_assert(TimerModule);
    if (TimerModule)
    {
        lopster_call_t* call = g_hash_table_lookup(TimerModule->periodic_funcs, func);
        if (!call)
        {
            call = g_malloc0(sizeof(*call));
            call->func = func;
            call->data = data;
            call->id = lopster_timeout_add(timeout, (GSourceFunc)on_call_periodic, call);
            g_hash_table_insert(TimerModule->periodic_funcs, call->func, call);
        }
    }
}

/** Cancel a periodic call. 
 * @param func the function that should be canceled.
 * @see lopster_call_periodic()
 */
void lopster_cancel_call(PeriodicFunc func)
{
    g_assert(TimerModule);
    if (TimerModule)
    {
        lopster_call_t* call = g_hash_table_lookup(TimerModule->periodic_funcs, func);
        if (call) g_hash_table_remove(TimerModule->periodic_funcs, call->func);
    }
}



#ifdef WIN32
time_t get_current_time()
{
    return wCurrentTime;
}
#endif
static gboolean update_time() {
#ifdef WIN32
    wCurrentTime = time(NULL);
#else
    CurrentTime = time(NULL);
#endif
    return TRUE;
}

static void timer_init()
{
    TimerModule->periodic_funcs =
        g_hash_table_new_full(g_direct_hash, g_direct_equal,
                              NULL, (GDestroyNotify)lopster_call_destroy);
}

static void timer_run()
{
    lopster_call_periodic(1000, (PeriodicFunc)update_time, NULL);
}

static void timer_exit()
{
    // cleaning delayed calls
    g_hash_table_destroy(TimerModule->periodic_funcs);
    TimerModule->periodic_funcs = NULL;
    TimerModule = NULL;
}


static void timer_module_init(timer_module_t* module) {
    module_init(MODULE(module), "Timer Module",
                (ModuleFunc)timer_init, (ModuleFunc)timer_run, (ModuleFunc)timer_exit);
#ifdef WIN32
    wCurrentTime  = time(NULL);
#else
    CurrentTime = time(NULL);
#endif

    module_require(ui_module_instance);
}

/** Require this timer module always.
 * @return the timer module
 */
module_t* timer_module_instance(void) {
    if (!TimerModule) {
        TimerModule = g_malloc0(sizeof(*TimerModule));
        timer_module_init(TimerModule);
    }
    return MODULE(TimerModule);
}


#ifdef WIN32
# define _L_BASE_C_
# include "lcorealiasdef.c"
#endif
