/* Copyright (C) 2005-2006 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.9 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include "lib_search.h"
#include "library.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
    module_t _module;
    GHashTable* keys;
    GHashTable* files;
} lib_search_module_t;

typedef struct {
    char* key;
    GPtrArray* files;
} ls_key_t;

typedef struct {
    file_t* file;
    ls_key_t** keys;
} ls_file_t;


static lib_search_module_t* LibSearchModule = NULL;


static void lib_search_key_destroy(ls_key_t* keys)
{
    g_ptr_array_free(keys->files, TRUE);
    g_free(keys->key);
    g_free(keys);
}

static void lib_search_file_destroy(ls_file_t* entry)
{
    g_free(entry->keys);
    g_free(entry);
}

static void lib_search_remove_all()
{
    g_hash_table_destroy(LibSearchModule->keys);
    g_hash_table_destroy(LibSearchModule->files);

    LibSearchModule->keys =
        g_hash_table_new_full
        ((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
         (GDestroyNotify)NULL, (GDestroyNotify)lib_search_key_destroy);

    LibSearchModule->files =
        g_hash_table_new_full
        ((GHashFunc)g_direct_hash, (GEqualFunc)g_direct_equal,
         (GDestroyNotify)NULL, (GDestroyNotify)lib_search_file_destroy);

}

static void lib_search_remove_file(file_t* file)
{
    if (!file)
    {
        lib_search_remove_all();
    }
    else
    {
        ls_key_t** kptr;
        ls_file_t* entry;
    
        entry = g_hash_table_lookup(LibSearchModule->files, file);
        if (!entry)
        {
            g_log("LIB_SEARCH", G_LOG_LEVEL_WARNING,
                  "cannot remove unknown file: %s", file->filename);
            return;
        }
        g_assert(entry);

        for (kptr = entry->keys; *kptr; kptr++)
        {
            ls_key_t* cont = *kptr;
            g_ptr_array_remove_fast(cont->files, entry);
            if (cont->files->len == 0)
                g_hash_table_remove(LibSearchModule->keys, cont->key);
        }

        g_hash_table_remove(LibSearchModule->files, file);
    }
}

static void lib_search_index_file(file_t* file)
{
    GSList* token;
    GSList* dlist;
    unsigned token_num;
    ls_key_t** kptr;
    ls_file_t* entry;

    token_num = search_file_tokenize(file, &token);

    // for each key insert file in key table
    entry = g_malloc(sizeof(*entry));
    entry->file = file;
    entry->keys = kptr = g_malloc(sizeof(*kptr)*(token_num+1));
    for (dlist = token; dlist; dlist = dlist->next)
    {
        char* key = dlist->data;
        ls_key_t* keys;
        GPtrArray* array;

        keys = g_hash_table_lookup(LibSearchModule->keys, key);
        if (!keys)
        {
            keys = g_malloc0(sizeof(*keys));
            keys->files = g_ptr_array_new();
            keys->key = key;
            g_hash_table_insert(LibSearchModule->keys, keys->key, keys);
        }
        else
        {
            g_free(key);
        }
        array = keys->files;
        if (array->len == 0 || g_ptr_array_index(array, array->len-1) != entry)
        {
            g_ptr_array_add(array, entry);
            *kptr++ = keys;
        }
    }
    *kptr++ = NULL;
    g_slist_free(token);

    g_hash_table_insert(LibSearchModule->files, file, entry);
}

static int _keys_sort(ls_key_t* c1, ls_key_t* c2)
{
    if (c1->files->len < c2->files->len) return -1;
    if (c1->files->len > c2->files->len) return 1;
    return 0;
}

/** Search in the library for shared files.
 * @param search the pattern the files should match
 * @return a list files of type file_t
 */
GList* lib_search_do(search_pattern_t* search)
{
    GSList* slist;
    unsigned i1;
    ls_key_t* min_cont = NULL;
    GSList* conts = NULL;
    GList* results = NULL;

    g_assert(LibSearchModule);

    for (slist = search->include; slist; slist = slist->next)
    {
        char* key = slist->data;
        ls_key_t* cont;

        if (!key[0] || !key[1]) continue;
        cont = g_hash_table_lookup(LibSearchModule->keys, key);
        if (!cont)
        {
            g_slist_free(conts);
            return NULL;
        }
    
        if (!min_cont)
        {
            min_cont = cont;
        }
        else if (cont->files->len < min_cont->files->len)
        {
            conts = g_slist_insert_sorted(conts, min_cont, (GCompareFunc)_keys_sort);
            min_cont = cont;
        }
        else
        {
            conts = g_slist_insert_sorted(conts, cont, (GCompareFunc)_keys_sort);
        }
    }
    if (!min_cont) return NULL;

    //FIXME: handle also exclude words

    for (i1 = 0; i1 < min_cont->files->len; i1++)
    {
        ls_file_t* file = g_ptr_array_index(min_cont->files, i1);

        // check whether this file contains all searched keys
        for (slist = conts; slist; slist = slist->next)
        {
            ls_key_t* c = slist->data;
            ls_key_t** kptr;
            for (kptr = file->keys; *kptr; kptr++)
            {
                if (*kptr == c) break;  // file has key
            }
            if (*kptr == 0) break;  // key in file not found
        }
        if (slist) continue;  // not all keys found in file.

        //FIXME: handle excludes

        if (search_pattern_check_file(SEARCHP(search), file->file, FALSE))
            results = g_list_prepend(results, file->file);
    }
    g_slist_free(conts);
    return results;
}

static void lib_search_change_file(file_t* file)
{
    lib_search_remove_file(file);
    lib_search_index_file(file);
}

static void lib_searches_init()
{
    LibSearchModule->keys =
        g_hash_table_new_full
        ((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
         (GDestroyNotify)NULL, (GDestroyNotify)lib_search_key_destroy);

    LibSearchModule->files =
        g_hash_table_new_full
        ((GHashFunc)g_direct_hash, (GEqualFunc)g_direct_equal,
         (GDestroyNotify)NULL, (GDestroyNotify)lib_search_file_destroy);

    library_register_actions
        (lib_search_index_file, lib_search_remove_file, lib_search_change_file);
}

static void lib_searches_run()
{
}

static void lib_searches_exit()
{
    //FIXME: free hash tables.
}

static void lib_search_module_init(lib_search_module_t* module)
{
    module_init(MODULE(module), "LibSearch Module",
                (ModuleFunc)lib_searches_init,
                (ModuleFunc)lib_searches_run,
                (ModuleFunc)lib_searches_exit);

    module->keys = NULL;

    module_require(library_module_instance);
    module_require(search_module_instance);

}


/** Require this module if you like to search in the library.
 * @return this module instance
 * @see module_require()
 */
module_t* lib_search_module_instance()
{
    if (!LibSearchModule)
    {
        LibSearchModule = g_malloc0(sizeof(*LibSearchModule));
        lib_search_module_init(LibSearchModule);
    }
    return MODULE(LibSearchModule);
}

#ifdef WIN32
#define _LIB_SEARCH_C_
#include "lcorealiasdef.c"
#endif

