/* Copyright (C) 2005-2006 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.25 $
 * $Date: 2008/12/23 13:38:14 $
 * $Author: sgop $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <unistd.h>
#endif

#include <libxml/parser.h>

#include "l_base.h"
#include "xml_handler.h"
#include "files.h"
#include "mtypes.h"
#include "file_tree.h"
#include "library.h"
#include "file_tree.h"
#include "utils.h"
#include "preferences.h"
#include "attribute.h"
#include "search.h"
#include "plugin.h"
#include "mp3_ext.h"
#include "ogg_ext.h"
#include "flac_ext.h"
#include "action_item.h"
#include "thread.h"
#include "ui.h"
#include "exec.h"
#include "l_i18n.h"
#include "open_actions.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  LibNodeFunc shared;
  LibNodeFunc unshared;
  LibNodeFunc changed;
} lib_actions_t;

typedef struct {
    thread_t _thread;
    library_t* lib;
    char* item;
    gboolean rescan;
    GList* to_add;
    GList* changed;
    struct timeval last_item;
    unsigned progress;
    unsigned left;
} lib_scan_t;

static GStaticMutex process_mutex = G_STATIC_MUTEX_INIT;
typedef struct {
  module_t _module;
  library_t* lib;
  GList* process;
  lib_scan_t* thread;
  GList* locked_nodes;
  gboolean loading_disabled;
  GList* actions;
} lib_module_t;

static lib_module_t* LibModule = NULL;
static library_t* Lib = NULL;


#define DEFAULT_LIB   "shared.xml"


/** Remove all shared folders.
 * @param lib the library
 */
void library_clear_shared(library_t* lib) {
  if (!lib) return;
  g_list_foreach(lib->shared_folder, (GFunc)g_free, NULL);
  g_list_free(lib->shared_folder);
  lib->shared_folder = NULL;
}


/** Add a folder that should be shared.
 * @param lib the library.
 * @param folder the new folder to add.
 */
void library_add_shared(library_t* lib, const char* folder) {
  g_assert(folder);
  if (!lib) return;
  lib->shared_folder = g_list_append(lib->shared_folder, g_strdup(folder));
}

static void  xml_get_shared(library_t* lib, xmlDocPtr xfile, xmlNodePtr xnode) {
  xmlNodePtr xsub;
  xmlChar* attr;

  xsub = xnode->xmlChildrenNode;

  while (xsub) {
    if (!strcmp(xsub->name, "Folder")) {
      attr = xmlNodeListGetString(xfile, xsub->xmlChildrenNode, 1);
      library_add_shared(lib, attr);
      xmlFree(attr);
    }
    xsub = xsub->next;
  }
}

static void xml_get_file(file_tree_t* tree, file_node_t* parent,
                         xmlDocPtr xfile, xmlNodePtr xnode)
{
    xmlNodePtr xsub;
    xmlChar* attr;
    file_t* file;
    int shared;
    time_t mod;
    int len;
    file_node_t* child = NULL;
    int flags;
    char* fname;
    attribute_t* fattr;
    
    file = file_new();
    
    attr = xmlGetProp(xnode, "Size");
    file->size = strtoll(attr, NULL, 10);
    xmlFree(attr);
    
    attr = xmlGetProp(xnode, "Shared");
    if (attr)
    {
        shared = strcmp(attr, "No");
        xmlFree(attr);
    }
    else
    {
        shared = 0;
    }

    attr = xmlGetProp(xnode, "Modified");
    mod = strtoul(attr, NULL, 10);
    xmlFree(attr);
    
    attr = xmlGetProp(xnode, "Name");
    len = strlen(attr); // tell the next function to alloc more bytes
    fname = file_node_get_folder(parent, &len);
    strcpy(fname+len, attr);
    file_set_origname(file, g_strdup(fname));
    file_set_locname(file, fname);
    
    flags = LIB_NEW;
    if (mod >= CurrentTime - LIB_TREE(tree)->mark) flags |= FN_FLAG_RECENT;
    if (shared) flags |= FN_FLAG_SHARED;
    
    child = file_tree_insert_file_in(tree, parent, file, flags, FALSE);
    LIB_NODE(child)->mod_time = mod;
    xmlFree(attr);
    
    xsub = xnode->xmlChildrenNode;
    while (xsub)
    {
        if (!strcmp(xsub->name, "Attr"))
        {
            xmlChar* name = xmlGetProp(xsub, "Name");
            time_t refreshed = 0;
            
            attr = xmlGetProp(xsub, "Time");
            if (attr)
            {
                refreshed = strtoul(attr, NULL, 10);
                xmlFree(attr);
            }
            
            attr = xmlNodeListGetString(xfile, xsub->xmlChildrenNode, 1);
            fattr = attr_store_str(ATTRC(file), "file", name, attr);
            xmlFree(name);
            xmlFree(attr);
            if (fattr) fattr->refreshed = refreshed;
        }
        xsub = xsub->next;
    }
}

static void xml_get_scanned(file_tree_t* tree, file_node_t* parent,
                            xmlDocPtr xfile, xmlNodePtr xnode)
{
    xmlNodePtr xsub;
    xmlChar* attr;
    file_node_t* child;
    
    xsub = xnode->xmlChildrenNode;
    
    while (xsub)
    {
        if (!strcmp(xsub->name, "Directory"))
        {
            attr = xmlGetProp(xsub, "Name");
            child = file_tree_insert_folder(tree, parent, attr);
            xmlFree(attr);
            
            xml_get_scanned(tree, child, xfile, xsub);
        }
        else if (!strcmp(xsub->name, "File"))
        {
            xml_get_file(tree, parent, xfile, xsub);
        }
        xsub = xsub->next;
    }
}

static void xml_load_lib(xmlDocPtr xfile, xmlNodePtr xroot)
{
    xmlChar* attr = xmlGetProp(xroot, "Refreshed");
    xmlNodePtr xsub;
    
    FILE_TREE(Lib)->last_refreshed = strtoul(attr, NULL, 10);
    xmlFree(attr);

    attr_allow_unregistered(TRUE);
  
    xsub = xroot->xmlChildrenNode;
    while (xsub != NULL)
    {
        if (!strcmp(xsub->name, "Shared"))
        {
            xml_get_shared(Lib, xfile, xsub);
        }
        else if (!strcmp(xsub->name, "Scanned"))
        {
            xml_get_scanned(FILE_TREE(Lib), NULL, xfile, xsub);
        }
        xsub = xsub->next;
    }
    
    attr_allow_unregistered(FALSE);
}

// LOAD/SAVE FUNCS
static void xml_save_node(file_node_t* node, xmlNodePtr xnode) {
  file_t* file;
  xmlNodePtr xsub;
  xmlNodePtr xsub2;
  GList* dlist;

  while (node) {
    file = node->file;
    if (file) {
      xsub = xmlNewChild(xnode, NULL, "File", NULL);
      xmlNewProp(xsub, "Name", file->filename);
      xmlNewProp(xsub, "Size", l_printf("%llu", node->file->size));
      xmlNewProp(xsub, "Shared", (node->flags & FN_FLAG_SHARED)?"Yes":"No");
      xmlNewProp(xsub, "Modified", l_printf("%ld", LIB_NODE(node)->mod_time));

      for (dlist = ATTRC(file)->attr; dlist; dlist = dlist->next) {
        attribute_t* attr = dlist->data;
        const char* value;
        
        value = attr_xml_print(attr);
        if (value) {
          xsub2 = xmlNewTextChild(xsub, NULL, "Attr", value);
          xmlNewProp(xsub2, "Name", attr_get_name(attr));
          xmlNewProp(xsub2, "Time", l_printf("%ld", attr->refreshed));
        }
      }
    } else if (node->child) {
      xsub = xmlNewChild(xnode, NULL, "Directory", NULL);
      xmlNewProp(xsub, "Name", OBJ(node)->name);
      xml_save_node(node->child, xsub);
    }
    node = node->next;
  }
}

static void xml_save_folder(char* sf, xmlNodePtr xroot) {
  xmlNewTextChild(xroot, NULL, "Folder", sf);
}
  
static void xml_save_lib(xmlDocPtr xfile, xmlNodePtr xroot) {
  library_t* lib = LibModule->lib;
  xmlNodePtr xsub;

  (void)xfile;

  if (LibModule->thread) thread_lock(THREAD(LibModule->thread));

  xmlNewProp(xroot, "Refreshed", l_printf("%ld", FILE_TREE(lib)->last_refreshed));
  if (FILE_TREE(lib)->description) 
    xmlNewProp(xroot, "Description", FILE_TREE(lib)->description);
  
  xsub = xmlNewChild(xroot, NULL, "Shared", NULL);
  g_list_foreach(lib->shared_folder, (GFunc)xml_save_folder, xsub);

  xsub = xmlNewChild(xroot, NULL, "Scanned", NULL);
  xml_save_node(FILE_TREE(lib)->files, xsub);

  // FIXME: this might dangerous because the 'LibModule->thread' var
  // could have changed meanwhile.
  if (LibModule->thread) thread_unlock(THREAD(LibModule->thread));

}

static void lib_save(unsigned when) {
  char* filename;

  g_assert(LibModule);
  if (!LibModule->lib) return;

  filename = g_strdup_printf
    (SDIR("%s") SDIR("share") "%s", lopster_get_home(), OBJ(LibModule->lib)->name);

  lopster_xml_change("core:library", filename, FALSE);
  lopster_xml_save("core:library", when);

  g_free(filename);
}

/** Save the current library under a certain name.
 * @param name the (new) name
 */
void library_save_as(const char* name) {
  g_assert(LibModule);
  if (!LibModule->lib) return;
  object_set_name(OBJ(LibModule->lib), name);
  lib_save(0);
}

/** Register callback for the shared file to monitor the sharing status.
 * @param shared callback that is called whenever a file is shared.
 * @param unshared callback that is called whenever a file is
 * unshared. If NULL is passed to the function then the whole library
 * was unshared.
 * @param changed callback that is called whenever a file was changed
 * (its attributes or name).
 */
void library_register_actions(LibNodeFunc shared, LibNodeFunc unshared,
                              LibNodeFunc changed)
{
    lib_actions_t* actions = g_malloc(sizeof(*actions));
    
    g_assert(LibModule);
    
    actions->shared = shared;
    actions->unshared = unshared;
    actions->changed = changed;
    LibModule->actions = g_list_append(LibModule->actions, actions);
}

/** Register a function to determine meta data for a file.  Every
 * plugin that supports additional meta data has to install a
 * precessing function.
 * @param name the name of the function
 * @param func the function which computes the meta data.
 */
void library_register_process_func(const char* name, FileProcessFunc func) {
  g_assert(LibModule);
  if (func) {
    LibModule->process = g_list_append(LibModule->process, func);
    g_log("LIB", G_LOG_LEVEL_DEBUG,
          "registered new file processing function '%s'", name);
  }
}

/** Delete a library on disk.
 * @param name the filename of the library.
 */
void library_delete_on_disk(const char *name) {
  char filename[1024];

  if (!name || !(*name)) return;
  
  sprintf(filename, SDIR("%s") SDIR("share") "%s", lopster_get_home(), name);
  unlink(filename); // FIXME : WIN32 
}

/** Return the current library 
 * @return the current library
 */
library_t* library_get_current() {
  g_assert(LibModule);
  return LibModule->lib;
}

/***********************************************************************/
/* refreshing a library                                                */
/***********************************************************************/

static file_node_t* set_remove_cb(file_tree_t* tree, file_node_t* node) {
  (void)tree;
  if (node->file) {
    node->flags &= ~(LIB_DELETE|LIB_CHANGED|LIB_NEW);
    node->flags |= LIB_REMOVE;
  }
  return NULL;
}

static void lib_scan_set_item(lib_scan_t* thread, const char* item)
{
    thread_lock(THREAD(thread));
    gettimeofday(&thread->last_item, 0);
    g_free(thread->item);
    thread->item = g_strdup(item);
    //  thread_set_dirty(thread);
    thread_unlock(THREAD(thread));
}

static void lib_scan_set_progress(lib_scan_t* thread, 
                                  unsigned cnt, unsigned size)
{
    thread_lock(THREAD(thread));
    thread->progress = cnt*100/size;
    thread->left = size-cnt;
    thread_unlock(THREAD(thread));
}

static void lib_scan_folder(lib_scan_t* thread, const char* folder)
{
    const char* name;
    GDir* dir;
    
    lib_scan_set_item(thread, folder);
    
    dir = g_dir_open(folder, 0, NULL);
    if (!dir) return;
    
    while ((name = g_dir_read_name(dir)) != NULL)
    {
        char* nname = g_strdup_printf(SDIR("%s") "%s", folder, name);
        if (g_file_test(nname, G_FILE_TEST_IS_DIR))
        {
            //      printf(" - dir %d %s\n", type, nname); 
            lib_scan_folder(thread, nname);
            g_free(nname);
        }
        else
        {
            thread_lock(THREAD(thread));
            thread->to_add = g_list_append(thread->to_add, nname);
            thread_unlock(THREAD(thread));
        }
        if (thread_interrupted(THREAD(thread))) break;
    }
    g_dir_close(dir);
}

static file_node_t* _collect_all(file_tree_t* tree, file_node_t* node, GList** list) {
  (void)tree;
  if (node->file) *list = g_list_prepend(*list, node);
  return NULL;
}

static void lib_collect_all(lib_scan_t* thread, GList** list)
{
    file_tree_foreach(FILE_TREE(thread->lib), (FileNodeTraverse)_collect_all, list);
}

static file_t* _file_create_from_local(const char *longname)
{
    file_t* file = file_new();
    file_set_locname(file, g_strdup(longname));
    file_set_origname(file, g_strdup(longname));
    return file;
}

static file_size_t _file_check_time(const char* filename, time_t* mod)
{
    struct stat stats;
    if (stat(filename, &stats) < 0) return 0;
    if (mod) *mod = stats.st_mtime;
    return stats.st_size;
}

static file_node_t* _collect_to_delete(file_tree_t* tree, file_node_t* node, GList** list)
{
    (void)tree;
    if (node->file && (node->flags & LIB_REMOVE))
        *list = g_list_prepend(*list, node);
    return NULL;
}

static void lib_clean(lib_scan_t* thread)
{
    file_tree_t* tree = FILE_TREE(thread->lib);
    GList* list = NULL;
    GList* dlist;
    
    file_tree_foreach(tree, (FileNodeTraverse)_collect_to_delete, &list);
    for (dlist = list; dlist; dlist = dlist->next)
    {
        file_node_t* node = dlist->data;
        file_tree_remove_node(tree, node);
    }
    g_list_free(list);
}


static void set_change(file_t* file, file_node_t* node)
{
    (void)file;
    node->flags |= LIB_CHANGED;
}

/** Lock a node. Nodes must be locked if a thread accesses that node.
 * @param node the node of a library.
 * @return TRUE
 */
static gboolean lib_node_lock(lib_node_t* node) {
  g_static_mutex_lock(&process_mutex);
  LibModule->locked_nodes = g_list_append(LibModule->locked_nodes, node);
  g_static_mutex_unlock(&process_mutex);
  return TRUE;
}
/** Unlock a previously locked node.
 * @param node the node of a library.
 * @return TRUE
 */
static void lib_node_unlock(lib_node_t* node) {
  g_static_mutex_lock(&process_mutex);
  LibModule->locked_nodes = g_list_remove(LibModule->locked_nodes, node);
  g_static_mutex_unlock(&process_mutex);
}

static void lib_scan_update(lib_scan_t* thread);

static unsigned lib_scan_main(lib_scan_t* thread)
{
    library_t* lib = thread->lib;
    file_tree_t* tree = FILE_TREE(lib);
    GList* dlist;
    int cnt, size;
    GList* schedule = NULL;
    
    if (thread->rescan)
    {
        file_tree_foreach(tree, (FileNodeTraverse)set_remove_cb, NULL);

        // setup the folder scan functions
        thread_set_description(THREAD(thread), _("scanning folders..."));
        thread_set_progress(THREAD(thread), 0);
        
        size = g_list_length(lib->shared_folder);
        cnt = 0;
        for (dlist = lib->shared_folder; dlist; dlist = dlist->next)
        {
            char* folder = dlist->data;
            
            lib_scan_folder(thread, folder);
            
            if (thread_interrupted(THREAD(thread))) break;
            thread_set_progress(THREAD(thread), ++cnt*100/size);
        }
        if (thread_interrupted(THREAD(thread))) return THREAD_INTERRUPTED;
    }
  
    // no do some things in main thread:
    // 1. make sure all files were added
    thread_sync(THREAD(thread), (ThreadSyncFunc)lib_scan_update, NULL);
    // 2. remove files that are no longer part of the library
    thread_sync(THREAD(thread), (ThreadSyncFunc)lib_clean, NULL);
    // 3. collect all files.
    thread_sync(THREAD(thread), (ThreadSyncFunc)lib_collect_all, &schedule);

    thread_set_progress(THREAD(thread), 0.);
    
    cnt = 0;
    size = g_list_length(schedule);
    for (dlist = schedule; dlist; dlist = dlist->next)
    {
        lib_node_t* node = dlist->data;
        GList* dlist2;
        
        lib_scan_set_item(thread, FILE_NODE(node)->file->filename);
        
        lib_node_lock(node);
        object_monitor_d(OBJ(FILE_NODE(node)->file), (ObjectFunc)set_change, node);
        
        for (dlist2 = LibModule->process; dlist2; dlist2 = dlist2->next)
        {
            FileProcessFunc func = dlist2->data;
            func(THREAD(thread), node);
            if (thread_interrupted(THREAD(thread))) break;
        }
        
        object_unmonitor_d(OBJ(FILE_NODE(node)->file), (ObjectFunc)set_change, node);
        lib_node_unlock(node);
        
        if (FILE_NODE(node)->flags & (LIB_CHANGED|LIB_NEW))
        {
            thread_lock(THREAD(thread));
            thread->changed = g_list_append(thread->changed, node);
            thread_unlock(THREAD(thread));
        }
        
        if (thread_interrupted(THREAD(thread))) break;
        
        lib_scan_set_progress(thread, ++cnt, size);
        
        lib_save(120000);
    }
    
    lib_scan_set_item(thread, NULL);
    
    // make sure everything is cleaned up.
    thread_sync(THREAD(thread), (ThreadSyncFunc)lib_scan_update, NULL);

    g_list_free(schedule);
    if (thread_interrupted(THREAD(thread))) return THREAD_INTERRUPTED;
    
    tree->last_refreshed = CurrentTime;
    return THREAD_COMPLETED;
}

static void lib_propagate_share(file_node_t* node)
{
    GList* dlist;
    g_assert(node->file);

    g_message("propagate share: %s", OBJ(node)->name);

    if (node->flags & LIB_NEW) return;
    for (dlist = LibModule->actions; dlist; dlist = dlist->next)
    {
        lib_actions_t* actions = dlist->data;
        if (actions->shared) actions->shared(node->file);
    }
    lib_save(120000);
}

static void lib_propagate_change(file_node_t* node)
{
    GList* dlist;
    if (!node->file) return;

    g_message("propagate change: %s", node?OBJ(node)->name:"ALL");

    for (dlist = LibModule->actions; dlist; dlist = dlist->next)
    {
        lib_actions_t* actions = dlist->data;
        if (actions->changed) actions->changed(node->file);
    }
    lib_save(120000);
}

static void lib_propagate_unshare(file_node_t* node)
{
    GList* dlist;

    g_message("propagate unshare: %s", node?OBJ(node)->name:"ALL");

    if (node && node->flags & LIB_NEW) return;
    for (dlist = LibModule->actions; dlist; dlist = dlist->next)
    {
        lib_actions_t* actions = dlist->data;
        if (actions->unshared) actions->unshared(node?node->file:NULL);
    }
    lib_save(120000);
}

static void _lib_process_changed(lib_scan_t* thread)
{
    file_tree_t* tree = FILE_TREE(thread->lib);
    GList* dlist;

    for (dlist = thread->changed; dlist; dlist = dlist->next)
    {
        file_node_t* node = dlist->data;
        int flags = node->flags;
        
        // we need to clear these flags cause lib_propagate_xxx don't allow them.
        node->flags &= ~(LIB_NEW|LIB_CHANGED);
        file_node_update(tree, node);
        
        if (node->flags & FN_FLAG_SHARED)
        {
            if (flags & LIB_NEW) lib_propagate_share(node);
            else if (flags & LIB_CHANGED) lib_propagate_change(node);
        }
    }
    g_list_free(thread->changed);
    thread->changed = NULL;
}

static void _lib_process_new(lib_scan_t* thread)
{
    file_tree_t* tree = FILE_TREE(thread->lib);
    GList* dlist;

    for (dlist = thread->to_add; dlist; dlist = dlist->next)
    {
        char* filename = dlist->data;
        char* name = locale_to_utf8(filename);
        file_node_t* node;
        time_t mod;
        
        if (!name)
        {
            g_log("LIB", G_LOG_LEVEL_WARNING,
                  "could not convert filename to utf8: %s", filename);
            continue; //FIXME: check error?
        }
        node = file_tree_search_filename(tree, name);
        if (!node)
        {
            file_t* file = _file_create_from_local(name);
            int flags;
            
            file->size = _file_check_time(filename, &mod);
            
            flags = LIB_NEW;
            if (mod >= CurrentTime - LIB_TREE(tree)->mark) flags |= FN_FLAG_RECENT;
            //	printf("inserting '%s'\n", file->longname);
            node = file_tree_insert_file_inherited(tree, file, flags);
            LIB_NODE(node)->mod_time = mod;
        }
        else if (node->file)
        {
            //	printf("updating '%s'\n", node->file->longname);
            node->file->size = _file_check_time(filename, &mod);
            node->flags &= ~LIB_REMOVE;
            LIB_NODE(node)->mod_time = mod;
        }
        g_free(name);
    }
    g_list_foreach(thread->to_add, (GFunc)g_free, NULL);
    g_list_free(thread->to_add);
    thread->to_add = NULL;
}

static void lib_scan_end(lib_scan_t* thread)
{
    file_tree_t* lib = FILE_TREE(thread->lib);
    
    UI_file_tree_set_progress(lib, 0.);
    UI_file_tree_set_progress_name(lib, NULL);
    UI_file_tree_update_label(lib, NULL);

    switch (thread_get_status(THREAD(thread)))
    {
    case THREAD_FAILED:
        lopster_notify(LOPSTER_MESSAGE_INFO, _("Warning"),
                       _("Failed to refresh the library. The reason is unknown."));
        break;
    case THREAD_INTERRUPTED:
        UI_file_tree_update_label(lib, _("interrupted"));
        break;
    case THREAD_COMPLETED:
        // make sure we dont miss anything
        g_assert(!thread->changed);
        g_assert(!thread->to_add);
        break;
    default:
        break;
    }

    // in case we aborted and there is pending work
    _lib_process_new(thread);
    _lib_process_changed(thread);


    file_tree_mark(lib, 1);
    
    LibModule->thread = NULL;
    thread_destroy(THREAD(thread));
    
    UI_lib_working(FALSE);
}

static void lib_scan_update(lib_scan_t* thread)
{
    file_tree_t* tree = FILE_TREE(thread->lib);
    static char* file = NULL;
    static char* des = NULL;
    static unsigned progress = 0;
    static unsigned left = 0;
    struct timeval tv;
    const char* temp = 0;
    

    thread_lock(THREAD(thread));
    
    _lib_process_new(thread);
    _lib_process_changed(thread);
    
    if (file != thread->item || (file && strcmp(thread->item, file)))
    {
        g_free(file);
        file = g_strdup(thread->item);
        UI_file_tree_update_label(tree, file);
    }
    
    gettimeofday(&tv, 0);
    if ((tv.tv_sec - thread->last_item.tv_sec) * 1000 +
        (tv.tv_usec - thread->last_item.tv_usec) / 1000 > 2000)
    {
        temp = thread_get_description(THREAD(thread));
        progress = thread_get_progress(THREAD(thread));
    }
    else
    {
        if (!temp || left != thread->left)
        {
            // number of files left to process in library
            temp = l_printf(_("%u remaining"), left);
            left = thread->left;
        }
        progress = thread->progress;
    }
    if (des != temp || (temp && strcmp(des, temp)))
    {
        g_free(des);
        des = g_strdup(temp);
        UI_file_tree_set_progress_name(tree, des);
    }
    thread_unlock(THREAD(thread));

    UI_file_tree_set_progress(tree, (double)progress/100.);
}

/** Refresh the library.
 * @param rescan TRUE to resan the disk for files, FALSE to only
 * update the existing files.
 */
void library_refresh(gboolean rescan)
{
    file_tree_t* tree;
    
    g_assert(LibModule);
    
    tree = FILE_TREE(LibModule->lib);
    if (LibModule->thread)
    {
        lopster_notify(LOPSTER_MESSAGE_INFO, _("Cannot refresh library"),
                       _("Reason: the library is busy"));
        return;
    }
    else if (!LibModule->lib)
    {
        g_log("LIB", G_LOG_LEVEL_DEBUG, "no lib!");
        return;
    }
  
    g_log("LIB", G_LOG_LEVEL_DEBUG, "refreshing library: %s", OBJ(tree)->name);
    
    LibModule->thread = g_malloc0(sizeof(*LibModule->thread));
    
    thread_init(THREAD(LibModule->thread), _("Refreshing library"), 100);
    OBJ(LibModule->thread)->destroy = (ObjectFunc)lib_scan_end;
    LibModule->thread->lib = LibModule->lib;
    LibModule->thread->rescan = rescan;
    object_monitor_d(OBJ(LibModule->thread), (ObjectFunc)lib_scan_update, NULL);
    
    UI_lib_working(TRUE);
    thread_start(THREAD(LibModule->thread), (ThreadMainFunc)lib_scan_main);
}

/** Stop refreshing the library.*/
void library_abort_refresh()
{
    g_assert(LibModule);
    if (LibModule->thread)
    {
        thread_interrupt(THREAD(LibModule->thread));
    }
    else
    {
        lopster_notify(LOPSTER_MESSAGE_WARNING, _("Operation invalid"),
                       _("Reason: the library is not working"));
    }
}

static gboolean lib_node_is_locked(file_node_t* node)
{
    gboolean res;
    g_static_mutex_lock(&process_mutex);
    res = (g_list_find(LibModule->locked_nodes, node) != NULL);
    g_static_mutex_unlock(&process_mutex);
    return res;
}

static file_node_t* lib_shared_action(file_tree_t* tree, file_node_t* node, 
                                      FileNodeTraverse action, void* data) {
  file_node_t* temp;
  file_node_t* temp2;
  file_node_t* result;
  
  if (!tree) return NULL;
  if (node && node->file && (node->flags & FN_FLAG_SHARED) &&
      !(node->flags & (LIB_NEW|LIB_CHANGED)) && !lib_node_is_locked(node)) {
    result = action(tree, node, data);
    if (result) return result;
  }

  if (node) temp = node->child;
  else temp = tree->files;

  while (temp) {
    temp2 = temp->next;
    result = lib_shared_action(tree, temp, action, data);
    if (result) return result;
    temp = temp2;
  }
  return NULL;
}

/** Call a function foreach shared node that contains a file.
 * @param func the function to call
 * @param data the user data for the function
 * @return the node that was return by the func, or NULL
 */
file_node_t* library_foreach_shared(FileNodeTraverse func, void* data) {
  g_assert(LibModule);
  if (!LibModule->lib) return NULL;
  return lib_shared_action(FILE_TREE(LibModule->lib), NULL, func, data);
}

/** Find a shared file.
 * @param winname the windows name of the file
 * @return the shared file if found, or NULL
 */
file_t* library_winname_requested(const char* winname) {
  char* filename;
  file_node_t* node;
  file_t* file = NULL;

  g_assert(LibModule);
  if (!LibModule->lib) return NULL;
  filename = win2local(g_strdup(winname));
  
  node = file_tree_search_filename(FILE_TREE(LibModule->lib), filename);
  
  if (node && node->file && (node->flags & FN_FLAG_SHARED) &&
      !lib_node_is_locked(node)) {
    file = node->file;
  }
  g_free(filename);
  return file;
}


















static void lib_node_destroy(file_node_t* node) {
  if (node->file && (node->flags & LIB_DELETE)) {
    g_log("LIB", G_LOG_LEVEL_DEBUG,
          "deleting file %s (deactivated)\n", OBJ(node)->name);
    //    unlink(node->file->longname);
  }
  file_node_destroy(node);
}

static void lib_node_init(lib_node_t* node, const char* name) {
  file_node_init(FILE_NODE(node), name);
  object_add_type(OBJ(node), "LibraryNode");
  OBJ(node)->destroy = (ObjectFunc)lib_node_destroy;
  node->mod_time = 0;
}

static file_node_t* lib_node_new(const char* name) {
  lib_node_t* node = g_malloc0(sizeof(*node));
  lib_node_init(node, name);
  return FILE_NODE(node);
}

/////////////////

static int lib_node_recent(file_tree_t* tree, file_node_t* node) {
  if (LIB_NODE(node)->mod_time + LIB_TREE(tree)->mark > CurrentTime)
    return LIB_TREE(tree)->mark + LIB_NODE(node)->mod_time - CurrentTime;
  else 
    return 0;
}









  

static gboolean lib_start_refresh(void)
{
    if (!LibModule->thread) library_refresh(FALSE);
    return FALSE;
}


static void lib_destroy(library_t* lib)
{
    if (!lib) return;
    
    if (LibModule->lib == lib)
    {
        // propagate if we destroy current lib
        lib_propagate_unshare(NULL);
        // make sure we dont save delayed
        lopster_xml_change("core:library", NULL, TRUE);
        LibModule->lib = NULL;
    }
    g_list_foreach(lib->shared_folder, (GFunc)g_free, NULL);
    g_list_free(lib->shared_folder);
    
    file_tree_destroy(FILE_TREE(lib));
}

static void lib_init(library_t* lib, const char* name)
{
    file_tree_init(FILE_TREE(lib), name, NULL);
    object_add_type(OBJ(lib), "Library");
    OBJ(lib)->destroy = (ObjectFunc)lib_destroy;
    FILE_TREE(lib)->node_new = lib_node_new;
    FILE_TREE(lib)->node_recent = lib_node_recent;
    FILE_TREE(lib)->node_shared = lib_propagate_share;
    FILE_TREE(lib)->node_unshared = lib_propagate_unshare;
    
    lib->mark = 0;
    lib->shared_folder = NULL;
}

static library_t* lib_new(const char* name) {
  library_t* lib = g_malloc0(sizeof(library_t));
  lib_init(lib, name);
  return lib;
}

static void on_lib_mark_changed(pref_node_t* node, library_t* lib)
{
    lib->mark = *((int*)(node->value));
    file_tree_mark(FILE_TREE(lib), 1);
}


static void on_lib_destroyed(library_t* lib)
{
    pref_unobserve("/core/lib/mark_val", (ObjectFunc)on_lib_mark_changed, lib);
}

static library_t* lib_create_new(const char* name)
{
    library_t* lib = lib_new(name);
    
    if (lib)
    {
        lib->mark = pref_get_ref("/core/lib/mark_val", int);
        pref_observe("/core/lib/mark_val", (ObjectFunc)on_lib_mark_changed, lib);
        object_notify_d(OBJ(lib), (ObjectFunc)on_lib_destroyed, NULL);
    }

    return lib;
}

static library_t* lib_load(const char* fname)
{
    char* filename = l_printf(SDIR("%s") SDIR("share") "%s",
                              lopster_get_home(), fname);
    
    lopster_xml_change("core:library", filename, TRUE);
    
    Lib = lib_create_new(fname);
    if (!lopster_xml_load("core:library"))
    {
        object_kill(OBJ(Lib));
        Lib = NULL;
    }
    return Lib;
}

/** loads library and replaces old if existent */
static void lib_try_load(const char* name)
{
    library_t* lib;
    
    g_assert(LibModule);
    //  return;
    if (LibModule->thread)
    {
        lopster_notify(LOPSTER_MESSAGE_INFO, _("Cannot load library"),
                       _("Reason: the current library is busy"));
        return;
    }
    
    if (!name || *name == 0) name = DEFAULT_LIB;
    
    if ((lib = lib_load(name)) != NULL)
    {
        if (LibModule->lib)
        {
            g_log("LIB", G_LOG_LEVEL_DEBUG,
                  "destroying old lib: %s", OBJ(LibModule->lib)->name);
            object_kill(OBJ(LibModule->lib));
        }
        LibModule->lib = lib;
        lopster_timeout_add(10000, (GSourceFunc)lib_start_refresh, NULL);
        object_make_public(OBJ(lib), NULL);
    }
    else if (!LibModule->lib)
    {
        LibModule->lib = lib_create_new(name);
        object_make_public(OBJ(LibModule->lib), NULL);
    }
    // prevent that we load the library cause we change the
    // pref value.
    LibModule->loading_disabled = TRUE;
    pref_store("/core/lib/current", OBJ(LibModule->lib)->name);
    LibModule->loading_disabled = FALSE;
}

static void lib_load_current() {
  const char* name = pref_get_ptr("/core/lib/current", char);
  if (!LibModule->loading_disabled) lib_try_load(name);
}





// actions funcs
static file_t* file_node_to_file(file_node_t* node) {
  return node?node->file:NULL;
}

static void on_share(GList* nodes, file_tree_t* tree)
{
    GList* dlist;
    for (dlist = nodes; dlist; dlist = dlist->next)
    {
        file_node_t* node = dlist->data;
        file_tree_share_node(tree, node, MEDIA_ANY, FN_GROUP_ALL);
    }
}

static gboolean on_check_not_shared(file_node_t* node, file_tree_t* tree) {
  (void)tree;
  if (node->fci[MEDIA_ANY][FN_GROUP_SHARED] <
      node->fci[MEDIA_ANY][FN_GROUP_ALL]) return TRUE;
  return FALSE;
}

static void on_unshare(GList* nodes, file_tree_t* tree)
{
    GList* dlist;
    for (dlist = nodes; dlist; dlist = dlist->next)
    {
        file_node_t* node = dlist->data;
        file_tree_unshare_node(tree, node, MEDIA_ANY, FN_GROUP_ALL);
    }
}

static gboolean on_check_shared(file_node_t* node, file_tree_t* tree) {
  (void)tree;
  if (node->fci[MEDIA_ANY][FN_GROUP_SHARED] > 0) return TRUE;
  return FALSE;
}
 
static void library_actions_init(void) {
  action_item_t* item;
  //  action_item_t* sub;

  item = action_item_root("Library file actions");
  action_item_open_file(N_("Open file"), item, 
                        (ActionObjectConvert)file_node_to_file);
  action_item_sep(item);
  action_item_new(N_("Share file"), item, 0,
                  (ActionItemFunc)on_share,
                  (ActionItemCheck)on_check_not_shared);
  action_item_new(N_("Unshare file"), item, 0,
                  (ActionItemFunc)on_unshare,
                  (ActionItemCheck)on_check_shared);
  action_item_sep(item);
  action_item_simple(N_("Rename file"), item, AI_RESTRICT_1,
                     (ActionItemFunc)NULL);
  action_item_simple(N_("Delete file"), item, 0,
                     (ActionItemFunc)NULL);

  item = action_item_root("Library folder actions");
  action_item_open_folder(N_("Open folder"), item, NULL);
  action_item_sep(item);
  action_item_new(N_("Share subtree"), item, 0,
                  (ActionItemFunc)on_share,
                  (ActionItemCheck)on_check_not_shared);
  action_item_new(N_("Unshare subtree"), item, 0,
                  (ActionItemFunc)on_unshare,
                  (ActionItemCheck)on_check_shared);
  action_item_sep(item);
  action_item_simple(N_("Rename file"), item, AI_RESTRICT_1,
                     (ActionItemFunc)NULL);
  action_item_simple(N_("Delete file"), item, 0,
                     (ActionItemFunc)NULL);
}

static void library_init()
{
    g_assert(LibModule);
    
    lopster_xml_register_home("core:library", SDIR("share") DEFAULT_LIB,
                              "Library", xml_load_lib, xml_save_lib);
    
    pref_register_int("/core/lib/mark_val", TRUE, 0);
    pref_add_constraint_min("/core/lib/mark_val", 0);
    pref_register_string("/core/lib/current", TRUE, DEFAULT_LIB);
    
    attr_register_int ("file", N_("Bitrate"), ATTR_IMPORTANT, (UIPrintFunc)print_bitrate);
    attr_register_int ("file", N_("Frequency"), ATTR_IMPORTANT, (UIPrintFunc)print_frequency);
    attr_register_bool("file", N_("VBR"), ATTR_USEFUL, (UIPrintFunc)print_vbr);
    attr_register_int ("file", N_("Duration"), ATTR_IMPORTANT, (UIPrintFunc)print_duration);
    attr_register_str ("file", N_("Title"),  ATTR_USEFUL, NULL);
    attr_register_str ("file", N_("Artist"), ATTR_USEFUL, NULL);
    attr_register_str ("file", N_("Album"),  ATTR_USEFUL, NULL);
    attr_register_int ("file", N_("Year"),   ATTR_USEFUL, NULL);
    
    library_actions_init();
    
    library_register_process_func("file_parse_mp3", file_parse_mp3);
#ifdef HAVE_OGG
    library_register_process_func("file_parse_ogg", file_parse_ogg);
#endif
#ifdef HAVE_FLAC
    library_register_process_func("file_parse_flac", file_parse_flac);
#endif
}

static void library_run()
{
    pref_observe("/core/lib/current", (ObjectFunc)lib_load_current, NULL);
    lib_load_current();
}

static void library_exit()
{
    object_kill(OBJ(LibModule->lib));
    g_list_foreach(LibModule->actions, (GFunc)g_free, NULL);
    g_list_free(LibModule->actions);
    LibModule = NULL;
}

static void library_module_init(lib_module_t* module)
{
    module_init(MODULE(module), "Library Module",
                (ModuleFunc)library_init,
                (ModuleFunc)library_run,
                (ModuleFunc)library_exit);
    
    module->lib = NULL;
    module->process = NULL;
    module->thread = NULL;
    module->locked_nodes = NULL;
    module->loading_disabled = FALSE;
    module->actions = NULL;
    
    module_require(pref_module_instance);
    module_require(thread_module_instance);
    module_require(mtype_module_instance);
    module_require(search_module_instance);
    module_require(base_module_instance);
    module_require(xml_module_instance);
}


/** Require this module if you like to acces the library.
 * @return the library module.
 * @see module_require()
 */
module_t* library_module_instance()
{
    if (!LibModule)
    {
        LibModule = g_malloc0(sizeof(*LibModule));
        library_module_init(LibModule);
    }
    return MODULE(LibModule);
}

#ifdef WIN32
#define _LIBRARY_C_
#include "lcorealiasdef.c"
#endif

