/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.9 $
 * $Date: 2007/11/21 20:18:23 $
 * $Author: sgop $
 */

#include <string.h>

#include <glib.h>

#include "l_base.h"
#include "load.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
  GList* loads;
} load_module_t;

static load_module_t* l_LoadModule = NULL;

const int LoadSteps[BAND_SIZE] = {
  1, 5, 30, 180, 720, 4320
};


static void load_destroy(load_t* load) {
  int i1;

  g_free(load->name);
  for (i1 = 0; i1 < BAND_SIZE; i1++) g_free(load->val[i1]);
  g_free(load->val);
  g_free(load->pos);
  g_free(load->tno);
  g_free(load);
}

static void load_init(load_t* load, const char* name, gboolean in) {
  int i1;

  g_assert(load && name);

  load->name = g_strdup(name);
  load->val  = g_malloc(sizeof(double*) * BAND_SIZE);
  load->pos = g_malloc0(sizeof(int) * BAND_SIZE);
  load->tno = g_malloc0(sizeof(int) * BAND_SIZE);
  load->timestamp = CurrentTime;
  load->in = in;
  load->first = 0;

  for (i1 = 0; i1 < BAND_SIZE; i1++) {
    load->val[i1] = g_malloc0(sizeof(double)*BAND_WIDTH);
  }
}

static load_t* load_new(const char* name, gboolean in) {
  load_t* load;

  g_assert(l_LoadModule);
  load = g_malloc0(sizeof(*load));
  load_init(load, name, in);
  l_LoadModule->loads = g_list_append(l_LoadModule->loads, load);
  return load;
}

/** Bring the load up date (to the current time) manually.
 * @param load the load.
 */
void load_step(load_t* load) {
  int i1;

  g_assert(load && load->timestamp <= CurrentTime);

  while (load->timestamp < CurrentTime) {
    for (i1 = 0; i1 < BAND_SIZE; i1++) {
      load->tno[i1]++;
      if (load->tno[i1] % LoadSteps[i1] == 0) {
        load->val[i1][load->pos[i1]] /= LoadSteps[i1];
        load->tno[i1] = 0;
        load->pos[i1]++;
        if (load->pos[i1] >= BAND_WIDTH) load->pos[i1] = 0;
      }
    }
    load->timestamp++;
  }
}

/** Add some transferred bytes to the load.
 * @param load the load object
 * @param val the value to add.
 * @see load_find()
 */
void load_add(load_t* load, int val) {
  int i1;

  g_assert(load);
  
  if (load->timestamp < CurrentTime) load_step(load);
  if (!load->first) load->first = CurrentTime;

  for (i1 = 0; i1 < BAND_SIZE; i1++) {
    load->val[i1][load->pos[i1]] += val;
  }    
}

load_t* load_find(const char* name, gboolean in)
{
    GList* dlist;
    
    g_assert(l_LoadModule);
    g_assert(name);
    for (dlist = l_LoadModule->loads; dlist; dlist = dlist->next) {
        load_t* load = dlist->data;
        if (load->in == in && !strcmp(load->name, name)) return load;
    }
    return NULL;
}

/** Find a registered load by name.
 * @param name the name of the load.
 * @param in TRUE to get the incoming load, FALSE to get the outgoing one.
 * @return the load if found, NULL otherwise
 */
load_t* load_get(const char* name, gboolean in)
{
    load_t* load;
    
    g_assert(name);
    load = load_find(name, in);
    if (!load)
    {
        g_log("LOAD", G_LOG_LEVEL_WARNING,
              "load '%s' not registered", name);
        g_assert(FALSE);
    }
    return load;
}

/** Register a new load. It is registered for both the incoming and
 * outgoing transfer.
 * @param name of the load, for later access, must be unique
 */
void load_register(const char* name) {
  load_t* load;

  g_log("LOAD", G_LOG_LEVEL_DEBUG, "register load '%s'", name);
  load = load_find(name, TRUE);
  if (load) {
    g_log("LOAD", G_LOG_LEVEL_WARNING,
          "in-load '%s' already registered", name);
    g_assert(FALSE);
  } else {
    load = load_new(name, TRUE);
  }

  load = load_find(name, FALSE);
  if (load) {
    g_log("LOAD", G_LOG_LEVEL_WARNING,
          "out-load '%s' already registered", name);
    g_assert(FALSE);
  } else {
    load = load_new(name, FALSE);
  }
}

/** Get all registered loads. 
 * @return a list of all registered loads.
 */
GList* loads_get() {
  g_assert(l_LoadModule);
  return l_LoadModule->loads;
}


static void loads_init() {
}
static void loads_run() {
}
static void loads_exit() {
  if (l_LoadModule) {
    g_list_foreach(l_LoadModule->loads, (GFunc)load_destroy, NULL);
    g_list_free(l_LoadModule->loads);
    l_LoadModule->loads = NULL;
  }
}

static void load_module_init(load_module_t* module) {
  module_init(MODULE(module), "Load Module",
              (ModuleFunc)loads_init,
              (ModuleFunc)loads_run,
              (ModuleFunc)loads_exit);
  module->loads = NULL;
}

/** If you use transfer loads, you have to require this module.
 * @see module_require()
 */
module_t* load_module_instance() {
  if (!l_LoadModule) {
    l_LoadModule = g_malloc0(sizeof(*l_LoadModule));
    load_module_init(l_LoadModule);
  }
  return MODULE(l_LoadModule);
}

#ifdef WIN32
# define _LOAD_C_
# include "lcorealiasdef.c"
#endif
