/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.12 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>

#include "module.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

#define MODULE_INACTIVE      0
#define MODULE_INITIALIZING  1
#define MODULE_INITIALIZED   2
#define MODULE_RUNNING       3
#define MODULE_EXITING       4

static GList* Modules = NULL;
static GList* RequireStack = NULL;

struct _module_priv_t {
  int status;
  ModuleFunc init;
  ModuleFunc run;
  ModuleFunc done;
  GList* deps;
  GList* parents;
};

#include "dynamic.h"

static gboolean module_is_new(module_t* dep, module_t* last) {
  GList* dlist;
  for (dlist = Modules; dlist; dlist = dlist->next) {
    module_t* module = dlist->data;
    if (module == dep) return FALSE;
    if (module == last) return TRUE;
  }
  g_assert(FALSE);
  return TRUE;
}

void module_destroy(module_t* module) {
  g_list_free(module->priv->deps);
  g_list_free(module->priv->parents);
  g_free(module->priv);
  object_destroy(OBJ(module));
}

static void module_dummy() { }

/** You must call this function from your derived module before you
 * require any other module.
 * @param module the module
 * @param name the name of the module.
 * @param init the init method
 * @param run the run method
 * @param done the exit method. Make sure you free all additional
 * resources here. The module base is destructed later.
 * @see module_require()
 */
void module_init(module_t* module, const char* name,
                 ModuleFunc init, ModuleFunc run, ModuleFunc done) {
  object_init(OBJ(module), name, TRUE);
  object_add_type(OBJ(module), "Module");
  OBJ(module)->destroy = (ObjectFunc)module_destroy;
  module->priv = g_malloc0(sizeof(*module->priv));
  module->priv->status  = MODULE_INACTIVE;
  module->priv->init    = init?init:(ModuleFunc)module_dummy;
  module->priv->run     = run ?run :(ModuleFunc)module_dummy;
  module->priv->done    = done?done:(ModuleFunc)module_dummy;
  module->priv->deps    = NULL;
  module->priv->parents = NULL;
  
  RequireStack = g_list_prepend(RequireStack, module);

  Modules = g_list_append(Modules, module);
/*   g_log("MODULE", G_LOG_LEVEL_DEBUG, */
/*         "created module instance '%s'", OBJ(module)->name); */

  // require the dynamic_module for all modules to make sure that
  // the plugins get not unloaded before all plugins are shut down
  module_require(dynamic_module_instance);
}

static void module_require_direct_real(module_t* module, module_t* dep) {
  if (module == dep) return;
  if (g_list_find(module->priv->deps, dep)) return;
  module->priv->deps = g_list_append(module->priv->deps, dep);
  dep->priv->parents = g_list_append(dep->priv->parents, module);
}

void module_require_direct(module_t* dep, ModuleFunc init) {
  module_t* module;

  if (!RequireStack) {
    g_log("MODULE", G_LOG_LEVEL_WARNING,
          "RequireStack empty, did you forget to call 'module_init()?");
    g_assert(FALSE);
  }
  module = RequireStack->data;

  init(dep);

  g_assert(RequireStack->data == dep);
  RequireStack = g_list_delete_link(RequireStack, RequireStack);

  module_require_direct_real(module, dep);
}

/** Require another module. In your derived class you must call
 * module_init() before you require any modules.
 * @param func the module instance function
 */
void module_require(ModuleInstanceFunc func) {
  module_t* dep;
  module_t* parent;
  module_t* last;
  
  if (!RequireStack) {
    g_log("MODULE", G_LOG_LEVEL_WARNING,
          "RequireStack empty, did you forget to call 'module_init()?");
    g_assert(FALSE);
  }
  parent = RequireStack->data;

  // remember the latest module instance
  g_assert(Modules);  // there is at least the root module
  last = g_list_last(Modules)->data;
  dep = func();

  // check until 'last', whether module already existed
  if (module_is_new(dep, last)) {
    g_assert(dep == RequireStack->data);
    RequireStack = g_list_delete_link(RequireStack, RequireStack);
  }
  module_require_direct_real(parent, dep);
}

static void  _check_list_for(GList* list, int status) {
  for ( ; list; list = list->next) {
    module_t* module = list->data;
    if (module->priv->status != status) {
      g_log("MODULE", G_LOG_LEVEL_WARNING,
            "module '%s' (%p) is expected to have status %d, not %d",
	    OBJ(module)->name, module, status, module->priv->status);
    }
  }
}

static void  _check_stack_for(GList* stack, module_t* module) {
  if (!g_list_find(stack, module)) return;
  g_log("MODULE", G_LOG_LEVEL_WARNING,
        "module '%s' is in a loop", OBJ(module)->name);
  for ( ; stack; stack = stack->next) {
    module_t* m = stack->data;
    g_log("MODULE", G_LOG_LEVEL_WARNING,
          " - required from '%s'", OBJ(m)->name);
  }
}

static void module_initialize(module_t* module, GList** stack) {
  _check_stack_for(*stack, module);
  if (module->priv->status != MODULE_INACTIVE) return;
  module->priv->status = MODULE_INITIALIZING;
  *stack = g_list_prepend(*stack, module);
  g_list_foreach(module->priv->deps, (GFunc)module_initialize, stack);
  *stack = g_list_remove(*stack, module);
  _check_list_for(module->priv->deps, MODULE_INITIALIZED);
  if (module->priv->init) {
    g_log("MODULE", G_LOG_LEVEL_DEBUG,
          "initializing '%s'", OBJ(module)->name);
    module->priv->init(module);
  } else {
    g_log("MODULE", G_LOG_LEVEL_WARNING,
          "no init func defined for '%s'", OBJ(module)->name);
  }
  module->priv->status = MODULE_INITIALIZED;
}

static void module_run(module_t* module) {
  if (module->priv->status != MODULE_INITIALIZED) return;
  g_list_foreach(module->priv->deps, (GFunc)module_run, NULL);
  _check_list_for(module->priv->deps, MODULE_RUNNING);
  if (module->priv->run) {
    g_log("MODULE", G_LOG_LEVEL_DEBUG,
          "running '%s'", OBJ(module)->name);
    module->priv->run(module);
  } else {
    g_log("MODULE", G_LOG_LEVEL_WARNING,
          "no run func defined for '%s'", OBJ(module)->name);
  }
  module->priv->status = MODULE_RUNNING;
}

static void module_shutdown(module_t* module) {
  if (module->priv->status != MODULE_RUNNING) return;
  g_list_foreach(module->priv->parents, (GFunc)module_shutdown, NULL);
  _check_list_for(module->priv->parents, MODULE_INACTIVE);
  if (module->priv->done) {
    g_log("MODULE", G_LOG_LEVEL_DEBUG,
	  "shutting down '%s'", OBJ(module)->name);
    module->priv->done(module);
  } else {
    g_log("MODULE", G_LOG_LEVEL_WARNING,
	  "no exit func defined for '%s'", OBJ(module)->name);
  }
  module->priv->status = MODULE_INACTIVE;
}

void modules_initialize() {
  GList* stack = NULL;
  GList* dlist;
  g_log("MODULE", G_LOG_LEVEL_DEBUG, "Initializing modules...");
  for (dlist = Modules; dlist; dlist = dlist->next) {
    module_t* module = dlist->data;
    module_initialize(module, &stack);
  }
}
void modules_run() {
  GList* dlist;
  g_log("MODULE", G_LOG_LEVEL_DEBUG, "Running modules...");
  for (dlist = Modules; dlist; dlist = dlist->next) {
    module_t* module = dlist->data;
    module_run(module);
  }
}

void modules_shutdown() {
  GList* dlist;
  g_log("MODULE", G_LOG_LEVEL_DEBUG, "Shutting down modules...");
  for (dlist = Modules; dlist; dlist = dlist->next) {
    module_t* module = dlist->data;
    module_shutdown(module);
  }
  for (dlist = Modules; dlist; dlist = dlist->next) {
    module_t* module = dlist->data;
    object_kill(OBJ(module));
  }
  g_list_free(Modules);
  Modules = NULL;
}

/** Checks whether the module is running. This could be useful if you
 * like to know whether the module is currently destructing.
 * @param module the module
 * @return TRUE if module is running, FALSE otherwise
 */
gboolean module_running(module_t* module) {
  return (module && module->priv->status == MODULE_RUNNING);
}

#ifdef WIN32
# define _MODULE_C_
# include "lcorealiasdef.c"
#endif
