/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.10 $
 * $Date: 2006/07/22 17:11:47 $
 * $Author: sid4u $
 */

#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <glib.h>

#include "library.h"
#include "files.h"
#include "l_io.h"

#ifdef WIN32
#include "win32_defs.h"
#endif

#define FRAMES_FLAG     0x0001
#define BYTES_FLAG      0x0002
#define TOC_FLAG        0x0004
#define VBR_SCALE_FLAG  0x0008

static const short br[2][3][15] = {
  {				// mpeg 2
    {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256},	// 
    {0,  8, 16, 24, 32, 40, 48,  56,  64,  80,  96, 112, 128, 144, 160},	// 
    {0,  8, 16, 24, 32, 40, 48,  56,  64,  80,  96, 112, 128, 144, 160}	// 
  },
  {				// mpeg1
    {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448}, // 
    {0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384}, // 
    {0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320}  // 
  }
};
static const int freq[3][3] = {
  {22050, 24000, 16000}, // mpeg 2
  {44100, 48000, 32000}, // mpeg 1
  {11025, 12000, 8000},	 // mpeg 2.5
};
static const int bs[3] = { 384, 1152, 1152 };
static const int s1[3] = { 48000, 144000, 144000 };

static int HeaderPos = 0;

static int ExtractI4(unsigned char *buf) {
  int x;
  /* big endian extract */
  x = buf[0];
  x <<= 8;
  x |= buf[1];
  x <<= 8;
  x |= buf[2];
  x <<= 8;
  x |= buf[3];

  return x;
}

#define CHUNK_SIZE 1024

#define M_SYNC      0x7ff
#define M_VERSION   0x003
#define M_LAYER     0x003
#define M_PROTECT   0x001
#define M_BITRATE   0x00f
#define M_FREQUENCY 0x003
#define M_PADDING   0x001
#define M_PRIVATE   0x001
#define M_CHANNEL   0x003
#define M_MODE_EXT  0x003
#define M_COPYRIGHT 0x001
#define M_ORIGINAL  0x001
#define M_EMPHASIS  0x003

#define O_SYNC      21
#define O_VERSION   19
#define O_LAYER     17
#define O_PROTECT   16
#define O_BITRATE   12
#define O_FREQUENCY 10
#define O_PADDING    9
#define O_PRIVATE    8
#define O_CHANNEL    6
#define O_MODE_EXT   4
#define O_COPYRIGHT  3
#define O_ORIGINAL   2
#define O_EMPHASIS   0

#define SYNC(header)      ((header>>O_SYNC)      & M_SYNC)
#define MVERSION(header)  ((header>>O_VERSION)   & M_VERSION)
#define LAYER(header)     ((header>>O_LAYER)     & M_LAYER)
#define PROTECT(header)   ((header>>O_PROTECT)   & M_PROTECT)
#define BITRATE(header)   ((header>>O_BITRATE)   & M_BITRATE)
#define FREQUENCY(header) ((header>>O_FREQUENCY) & M_FREQUENCY)
#define PADDING(header)   ((header>>O_PADDING)   & M_PADDING)
#define PRIVATE(header)   ((header>>O_PRIVATE)   & M_PRIVATE)
#define CHANNEL(header)   ((header>>O_CHANNEL)   & M_CHANNEL)
#define MODE_EXT(header)  ((header>>O_MODE_EXT)  & M_MODE_EXT)
#define COPYRIGHT(header) ((header>>O_COPYRIGHT) & M_COPYRIGHT)
#define ORIGINAL(header)  ((header>>O_ORIGINAL)  & M_ORIGINAL)
#define EMPHASIS(header)  ((header>>O_EMPHASIS)  & M_EMPHASIS)

static int header_valid(guint32 header) {
  if (SYNC(header) != 0x7ff) return 0;
  if (MVERSION(header) == 0x1) return 0;
  if (LAYER(header) == 0x0) return 0;
  if (BITRATE(header) == 0xf) return 0;
  if (FREQUENCY(header) == 0x3) return 0;
  if (BITRATE(header) == 0x0) return 0;
  return 1;
}

static guint32 mpeg_audio_find_header(FILE* file, int pos) {
  char buffer[CHUNK_SIZE+3];
  int i1;

  FSEEK(file, pos, SEEK_SET);
  if (fread(buffer, CHUNK_SIZE+3, 1, file) != 1) return 0;
  for (i1 = 0; i1 < CHUNK_SIZE; i1++) {
    guint32 header;
    memcpy(&header, buffer+i1, sizeof(header));      /* take care of alignment */
    header = GUINT32_FROM_BE(header);
    if (header_valid(header)) {
      HeaderPos = pos+i1+4;
      return header;
    }
  }
  return 0;
}

void file_parse_mp3(thread_t* thread, lib_node_t* node) {
  FILE *fd;
  file_t* file;
  unsigned char xing[120];
  unsigned char id3[128];
  long totalframes;
  int stereo;
  double bpf;
  double tpf;
  int xing_offset;
  int pos;
  char* temp;
  int bitrate;
  int frequency;
  gboolean vbr;
  int duration;
  guint32 header = 0;
  int layer;
  int version;
  attribute_t* attr;

  file = FILE_NODE(node)->file;
  temp = get_suffix(file->filename);
  if (!temp) return;
  if (strcasecmp(temp, "mp3") && strcasecmp(temp, "mp2"))
    return;

  attr = attr_find(ATTRC(file), "file", "Bitrate");
  if (attr && attr->refreshed >= node->mod_time) return;

  temp = locale_from_utf8(file->localname);
  if (!temp) return;

  fd = fopen(temp, "r");
  g_free(temp);

  if (fd == NULL) {
    g_log("MP3", G_LOG_LEVEL_WARNING, "could not open mp3 audio file '%s'", temp);
    return;
  }
  
  pos = 0;
  while (pos < 1024*1024) {
    header = mpeg_audio_find_header(fd, pos);
    if (header) break;
    pos += CHUNK_SIZE;
  }
  if (!header) {
    g_log("MP3", G_LOG_LEVEL_WARNING,
          "could not find mp3 audio file header in '%s'", file->filename);
  }
  if (MVERSION(header) & 0x2) {
    version = (MVERSION(header)&0x1);
  } else {
    version = 2;
  }
  layer = 3-LAYER(header);
  
  if (version == 1)
    bitrate = br[1][layer][BITRATE(header)];
  else
    bitrate = br[0][layer][BITRATE(header)];
  frequency = freq[version][FREQUENCY(header)];

  // Mode
  if (CHANNEL(header) == 0x3) stereo = 0;
  else stereo = 1;

  if (version == 1) {
    if (stereo) xing_offset = 32;
    else xing_offset = 17;
  } else {
    if (stereo) xing_offset = 17;
    else xing_offset = 9;
  }

  tpf = (double)bs[layer] / frequency;

  if ((FSEEK(fd, HeaderPos+xing_offset, SEEK_SET) == 0) &&
      (fread(xing, 120, 1, fd) == 1) &&
      (memcmp(xing, "Xing", 4) == 0)) {
    unsigned char * buf = xing + 4;
    int head_flags;
    int vbr_scale = -1;
    int bytes;

    head_flags = ExtractI4(buf);
    buf += 4;

    if (head_flags & FRAMES_FLAG) {
      totalframes = ExtractI4(buf);
      buf += 4;
    } else totalframes = 1;

    if (head_flags & BYTES_FLAG) {
      bytes = ExtractI4(buf);
      buf += 4;
    } else bytes = 1;

    if (head_flags & TOC_FLAG) {
      buf += 100;
    }

    if (head_flags & VBR_SCALE_FLAG) {
      vbr_scale = ExtractI4(buf);
      buf += 4;
    }

    bitrate = (gint) ((bytes * 8) / (tpf * totalframes * 1000));
    bpf = (double) (file->size - HeaderPos) / totalframes;
    vbr = TRUE;
  } else {
    bpf = bitrate * s1[layer] / frequency;
    totalframes = (int) ((file->size - HeaderPos) / bpf);
    vbr = FALSE;
  }
  duration = totalframes * tpf;
  

  // always lock node if we do write access to the library
  thread_lock(thread);

  attr_store(ATTRC(file), "file", "Bitrate", bitrate);
  if (vbr) attr_store(ATTRC(file), "file", "VBR", TRUE);
  attr_store(ATTRC(file), "file", "Frequency", frequency);
  attr_store(ATTRC(file), "file", "Duration", duration);

  // now read id3v1
  FSEEK(fd, -128, SEEK_END);
  if (fread(id3, 128, 1, fd) == 1 && !strncmp(id3, "TAG", 3)) {
    char tag[31] = "";
    char* res;
    char* temp;

    tag[30] = 0;

    memcpy(tag, id3+3, 30); res = g_strstrip(tag);
    temp = auto_text(to_utf8(res, -1, NULL, NULL, 2, "utf-8", "latin1"));
    if (temp && *temp) attr_store(ATTRC(file), "file", "Title", temp);
    
    memcpy(tag, id3+33, 30); res = g_strstrip(tag);
    temp = auto_text(to_utf8(res, -1, NULL, NULL, 2, "utf-8", "latin1"));
    if (temp && *temp) attr_store(ATTRC(file), "file", "Artist", temp);
    
    memcpy(tag, id3+63, 30); res = g_strstrip(tag);
    temp = auto_text(to_utf8(res, -1, NULL, NULL, 2, "utf-8", "latin1"));
    if (temp && *temp) attr_store(ATTRC(file), "file", "Album", temp);
    
    memcpy(tag, id3+93, 4); tag[4] = 0;
    if (atoi(tag) > 1000) attr_store(ATTRC(file), "file", "Year", atoi(tag));

    // we are not interested in the rest of the ID3v1 atm
  }
  thread_unlock(thread);
  
  fclose(fd);

}

