/* Copyright (C) 2005-2006 sgop@users.sourceforge.net.
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.15 $
 * $Date: 2007/06/02 18:41:24 $
 * $Author: sgop $
 */

#include <string.h>

#include "object.h"
#include "l_timer.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

#define OBJ_FLAG_SYNC     0x01
#define OBJ_FLAG_DIRTY    0x02
#define OBJ_FLAG_DESTROY  0x04
#define OBJ_FLAG_CLEAN    0x08

typedef struct {
    ObjectFunc func;
    gpointer data;
} object_func_t;

typedef struct {
    char* name;
    void* data;
} object_data_t;

struct _object_priv_t {
    GSList* type;
    GList* update;
    GList* notify;
    GList* data;
    char flags;
};

static GHashTable* InstanceMonitor = NULL;  //FIXME: free on exit.
static GHashTable* InstanceMonitorInh = NULL;  //FIXME: free on exit.


#define OBJ_DESTROYING(obj) (object->priv->flags & OBJ_FLAG_DESTROY)
#define OBJ_CLEANING(obj) (object->priv->flags & OBJ_FLAG_CLEAN)
#define OBJ_SYNC(obj) (object->priv->flags & OBJ_FLAG_SYNC)
#define OBJ_DIRTY(obj) (object->priv->flags & OBJ_FLAG_DIRTY)

#define OBJ_ASSERT(object, cond, message)                             \
    if (!(cond)) {                                                    \
        g_log("OBJECT", G_LOG_LEVEL_WARNING, "%s: (%s) %s", message,  \
              object_get_type(object),                                \
              object->name?object->name:"<none>");                    \
        g_assert(FALSE);                                              \
    }
#define OBJ_WARN(object, cond, message)                               \
    if (cond) {                                                       \
        g_log("OBJECT", G_LOG_LEVEL_WARNING, "%s: (%s) %s", message,  \
              object_get_type(object),                                \
              object->name?object->name:"<none>");                    \
    }
#define OBJ_WARN_RETURN(object, cond, message)                        \
    if (cond) {                                                       \
        g_log("OBJECT", G_LOG_LEVEL_WARNING, "%s: (%s) %s", message,  \
              object_get_type(object),                                \
              object->name?object->name:"<none>");                    \
        return;                                                       \
    }

static GList* DirtyObjects = NULL;
static int Timer = 0;


/** Add object type string.
 * The string has to be static, it will not be freed by the object.
 * @param object the object
 * @param type the type of the object
 */
void object_add_type(object_t* object, const char* type)
{
    object->priv->type = g_slist_prepend(object->priv->type, (char*)type);
}

/** Get the latest type of the object.
 * @param object the object
 * @return the type of the object
 */
const char* object_get_type(object_t* object)
{
    if (object && object->priv && object->priv->type)
        return (const char*)(object->priv->type->data);
    else
        return "<unknown type>";
}

/** Destroy an object. This function must be called in the destroy
 * method of derived classes only.
 * @param object the object
 */
void object_destroy(object_t* object)
{
    OBJ_ASSERT(object, OBJ_DESTROYING(object),
               "always use object_kill() to destroy");
    OBJ_WARN(object, object->priv->data, "gui_data is not expected");

    while (object->priv->data)
        object_free_data_common(object, ((object_data_t*)object->priv->data)->name);

    g_list_foreach(object->priv->notify, (GFunc)g_free, NULL);
    g_list_free(object->priv->notify);
    g_list_foreach(object->priv->update, (GFunc)g_free, NULL);
    g_list_free(object->priv->update);
    g_slist_free(object->priv->type);
    if (OBJ_DIRTY(object))
    {
        DirtyObjects = g_list_remove(DirtyObjects, object);
    }
    g_free(object->priv);
    g_free(object->name);
    g_free(object);
}

/** Constructor. Must used only in the init method of derived classes.
 * @param object the object.
 * @param name the name of the object.
 * @param sync TRUE to do synchronous updates, FALSE to do delayed updates.
 * @note you always should set the destroy function of the object
 * @note you always should assign a static string to type of the object
 */
void object_init(object_t* object, const char* name, gboolean sync)
{
    object->name = g_strdup(name);
    object->destroy = (ObjectFunc)object_destroy;
    object->priv = g_malloc0(sizeof(*object->priv));
    object->priv->update = NULL;
    object->priv->notify = NULL;
    object->priv->data = NULL;
    object->priv->flags = 0;
    object_add_type(object, "Object");
    if (sync) object->priv->flags |= OBJ_FLAG_SYNC;
}

/** Copy the object. Only the object name is copied.
 * @param dest the destination object.
 * @param src the source object.
 */
void object_copy(object_t* dest, object_t* src)
{
    g_free(dest->name);
    dest->name = g_strdup(src->name);
}

/** Change the syncing of the object.
 * @param object the object.
 * @param sync TRUE to update synchronous, FALSE otherwise.
 */
void object_set_sync(object_t* object, gboolean sync)
{
    if (sync) object->priv->flags |= OBJ_FLAG_SYNC;
    else object->priv->flags &= ~OBJ_FLAG_SYNC;
}

static void object_notify_add(object_t* object, ObjectFunc func, gpointer data) 
{
    object_func_t* f;

    OBJ_WARN_RETURN(object, OBJ_DESTROYING(object),
                    "notify add while destroying");

    f = g_malloc(sizeof(*f));
    f->func = func;
    f->data = data;
    object->priv->notify = g_list_append(object->priv->notify, f);
}

/** Remove a function previously added with object_notify_add().  Call
 * this function if you don't want to monitor the object any longer or
 * when the user data was destroyed.
 * @param object the object to monitor.
 * @param func the function to call when object gets destroyed.
 * @param data the data passed to the function
 */
static void object_notify_remove(object_t* object, ObjectFunc func, gpointer data)
{
    GList* dlist;

/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_notify_remove: (%s) %s", */
/*           object->type, object->name); */

    // remove warning later, for now monitor critical observers
    OBJ_WARN_RETURN(object, OBJ_DESTROYING(object),
                    "notify remove while destroying");

    for (dlist = object->priv->notify; dlist; )
    {
        object_func_t* f = dlist->data;
        GList* next = dlist->next;
        if (f->data == data && (!func || f->func == func))
        {
            object->priv->notify = g_list_delete_link(object->priv->notify, dlist);
            g_free(f);
            if (func) return;
        }
        dlist = next;
    }
    if (func) OBJ_WARN(object, TRUE, "destroy not found, cannot remove");
}


/** Call a function when (before) an object is destroyed.
 * @param object the object to monitor
 * @param func the function to call before destroying object.
 * @param data the object which monitores.
 * @see object_notify_o()
 * @note the difference to object_notify_o() is that you have to take
 * care yourself if data gets invalid. In such a case you have to use
 * object_unnotify_d().
 */
void object_notify_d(object_t* object, ObjectFunc func, gpointer data) 
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_notify_d: (%s) %s", */
/*           object->type, object->name); */
    object_notify_add(object, func, data);
}

/** Remove a function previously added with object_notify_d().  Call
 * this function if you don't want to monitor the object any longer or
 * when the user data was destroyed.
 * @param object the object to monitor.
 * @param func the function to call when object gets destroyed.
 * @param data the data passed to the function
 */
void object_unnotify_d(object_t* object, ObjectFunc func, gpointer data)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_unnotify_d: (%s) %s", */
/*           object->type, object->name); */
    object_notify_remove(object, func, data);
}


static void on_observer2_destroyed(object_t* observer, object_t* object)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "on_observer2_destroyed: (%s) %s - (%s) %s", */
/*           object->type, object->name, observer->type, observer->name); */
    object_notify_remove(object, NULL, observer);
}

static void on_object2_destroyed(object_t* object, object_t* observer)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "on_object2_destroyed: (%s) %s - (%s) %s", */
/*           object->type, object->name, observer->type, observer->name); */
    object_notify_remove(observer, (ObjectFunc)on_observer2_destroyed, object);
}

/** Notify an observer before the object gets destroyed.
 * @param object the object o monitor
 * @param func the function to call before destroying object.
 * @param observer the object which monitores.
 */
void object_notify_o(object_t* object, ObjectFunc func, object_t* observer)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_notify_o: (%s) %s - (%s) %s", */
/*           object->type, object->name, observer->type, observer->name); */
    // notify observer
    object_notify_add(object, func, observer);
    // stop notifying observer if observer is destroyed
    object_notify_add(observer, (ObjectFunc)on_observer2_destroyed, object);
    // stop notifying observer if object is destroyed
    object_notify_add(object, (ObjectFunc)on_object2_destroyed, observer);
}

/** Remove a function previously added with object_notify_o().  Call
 * this function if you don't want to monitor the object any longer or
 * when the user data was destroyed.
 * @param object the object to monitor.
 * @param func the function to call when object gets destroyed.
 * @param observer the object which monitores.
 */
void object_unnotify_o(object_t* object, ObjectFunc func, object_t* observer)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_unnotify_o: (%s) %s - (%s) %s", */
/*           object->type, object->name, observer->type, observer->name); */
    object_notify_remove(object, func, observer);
    object_notify_remove(observer, (ObjectFunc)on_observer2_destroyed, object);
    object_notify_remove(object, (ObjectFunc)on_object2_destroyed, observer);
}



static void object_monitor_add(object_t* object, ObjectFunc func, gpointer data)
{
    object_func_t* f;

    OBJ_WARN_RETURN(object, OBJ_DESTROYING(object),
                    "monitor add while destroying");

    f = g_malloc(sizeof(*f));
    f->func = func;
    f->data = data;

    //NOTE: it is ok if object is currently cleaning
    object->priv->update = g_list_append(object->priv->update, f);
}

static void object_monitor_remove(object_t* object, ObjectFunc func, gpointer data)
{
    GList* dlist;
  
/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_monitor_remove: (%s) %s", */
/*           object->type, object->name); */

    // remove warning later, for now monitor critical observers
    OBJ_WARN_RETURN(object, OBJ_DESTROYING(object),
                    "monitor remove while destroying");

    for (dlist = object->priv->update; dlist; )
    {
        object_func_t* f = dlist->data;
        GList* next = dlist->next;

        if (f->data == data && (!func || f->func == func))
        {
            // if object is cleaning we may not modify the update list
            // -> disable the entry
            if (OBJ_CLEANING(object))
            {
                OBJ_WARN(object, TRUE, "virtual update remove");
                f->func = NULL;
                f->data = NULL;
            }
            else
            {
                object->priv->update = g_list_delete_link(object->priv->update, dlist);
                g_free(f);
            }
            if (func) return;
        }
        dlist = next;
    }
    if (func) OBJ_WARN(object, TRUE, "monitor not found, cannot remove");
}


/** Call a function when an object gets dirty.
 * @param object the object to monitor
 * @param func the function to call if object gets dirty
 * @param data the object which monitores.
 * @see object_monitor_o()
 * @note the difference to object_monitor_o() is that you have to take
 * care yourself if data gets invalid. In such a case you have to use
 * object_unmonitor_d().
 */
void object_monitor_d(object_t* object, ObjectFunc func, gpointer data)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_monitor_d: (%s) %s", */
/*           object->type, object->name); */
    object_monitor_add(object, func, data);
}

/** Remove a function previously added with object_monitor_d().  Call
 * this function if you don't want to monitor the object any longer or
 * when the user data was destroyed.
 * @param object the object to monitor.
 * @param func the function to call when object gets destroyed.
 * @param data the data passed to the function
 */
void object_unmonitor_d(object_t* object, ObjectFunc func, gpointer data)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_unmonitor_d: (%s) %s", */
/*           object->type, object->name); */
    object_monitor_remove(object, func, data);
}


static void on_observer_destroyed(object_t* observer, object_t* object)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "on_observer_destroyed: (%s) %s - (%s) %s", */
/*           object->type, object->name, observer->type, observer->name); */
    object_monitor_remove(object, NULL, observer);
}

static void on_object_destroyed(object_t* object, object_t* observer)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "on_object_destroyed: (%s) %s - (%s) %s", */
/*           object->type, object->name, observer->type, observer->name); */
    object_notify_remove(observer, (ObjectFunc)on_observer_destroyed, object);
}

/** Notify an observer before if the object gets dirty.
 * @param object the object o monitor
 * @param func the function to call if object gets dirty.
 * @param observer the object which monitores.
 */
void object_monitor_o(object_t* object, ObjectFunc func, object_t* observer)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_monitor_o: (%s) %s - (%s) %s", */
/*           object->type, object->name, observer->type, observer->name); */
    object_monitor_add(object, func, observer);
    object_notify_add(observer, (ObjectFunc)on_observer_destroyed, object);
    object_notify_add(object, (ObjectFunc)on_object_destroyed, observer);
}

/** Remove a function previously added with object_monitor_o().  Call
 * this function if you don't want to monitor the object any longer or
 * when the user data was destroyed.
 * @param object the object to monitor.
 * @param func the function to call when the object gets dirty
 * @param observer the object which monitores.
 */
void object_unmonitor_o(object_t* object, ObjectFunc func, object_t* observer)
{
/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_unmonitor_o: (%s) %s - (%s) %s", */
/*           object->type, object->name, observer->type, observer->name); */
    object_monitor_remove(object, func, observer);
    object_notify_remove(observer, (ObjectFunc)on_observer_destroyed, object);
    object_notify_remove(object, (ObjectFunc)on_object_destroyed, observer);
}

gboolean object_monitor_find(object_t* object, ObjectFunc func, gpointer data)
{
    GList* dlist;

    g_assert(object);
    for (dlist = object->priv->update; dlist; dlist = dlist->next)
    {
        object_func_t* f = dlist->data;
        if (f->func == func && f->data == data) return TRUE;
    }
    return FALSE;
}


static void object_clean(object_t* object)
{
    GList* dlist;

    OBJ_ASSERT(object, OBJ_DIRTY(object), "object should be dirty");

    object->priv->flags |= OBJ_FLAG_CLEAN;

    for (dlist = object->priv->update; dlist; )
    {
        object_func_t* f = dlist->data;
        GList* next = dlist->next;
        // if the entry was disabled during this function call,
        // remove it now
        if (!f->func)
        {
            OBJ_WARN(object, TRUE, "delayed update remove");
            object->priv->update = g_list_delete_link(object->priv->update, dlist);
            g_free(f);
        }
        else
        {
            f->func(object, f->data);
        }
        dlist = next;
    }
    object->priv->flags &= ~(OBJ_FLAG_DIRTY|OBJ_FLAG_CLEAN);
}

static gboolean dirty_timer()
{
    if (Timer) lopster_timeout_remove(Timer);
    Timer = 0;
  
    while (DirtyObjects)
    {
        object_t* object = DirtyObjects->data;
        object_clean(object);
        DirtyObjects = g_list_remove(DirtyObjects, object);
    }
    return FALSE;
}

/** Set an object dirty. The effect is that all observers are informed
 * either instantly or after some time (depending on the sync status
 * of the object).
 * @param object the object which got dirty.
 * @see object_monitor(), object_monitor_add()
 */
void object_set_dirty(object_t* object)
{

    if (!object->priv) return;  // make sure we don't get dirty while we are constructing
    if (OBJ_DESTROYING(object)) return;
    if (!object->priv->update) return; // dont set dirty if no one monitors
    if (OBJ_DIRTY(object)) return;

    object->priv->flags |= OBJ_FLAG_DIRTY;

    if (OBJ_SYNC(object))
    {
        object_clean(object);
    }
    else
    {
        DirtyObjects = g_list_prepend(DirtyObjects, object);
        if (Timer) return;
        Timer = lopster_timeout_add(100, (GSourceFunc)dirty_timer, NULL);
    }
}


static GList* object_get_data_link(object_t* object, const char* name)
{
    GList* dlist;
    
    OBJ_ASSERT(object, name, "name==NULL not allowed");

    if (!object || !object->priv) return NULL;

    for (dlist = object->priv->data; dlist; dlist = dlist->next)
    {
        object_data_t* data = dlist->data;
        if (!strcmp(data->name, name)) return dlist;
    }
    return NULL;
}

/**
 * Check for  labeled data from the object.
 * @param object the object. 
 * @param name the name of the gui data. 
 */
void* object_get_data(object_t* object, const char* name)
{
    GList* dlist;

    dlist = object_get_data_link(object, name);

    if (dlist)
    {
        object_data_t* data = dlist->data;
        return data->data;
    }
    return NULL;
}

/**
 * Set labeled data for the object. This function is similar to g_object_set_data().
 * @param object the object.
 * @param name the name of the gui data.
 * @param data the data, the object takes ownership.
 */
void object_set_data(object_t* object, const char* name, void* user_data)
{
    object_data_t* data;
    
    // make sure this data does not exist already..
    g_assert(object_get_data_link(object, name) == NULL);
    // some other sanity checks.
    g_assert(name && user_data);
    
    data = g_malloc0(sizeof(*data));
    data->name = g_strdup(name);
    data->data = user_data;
    object->priv->data = g_list_prepend(object->priv->data, data);
}

/**
 * free the gui data field and set it to 0.
 * @param object the object.
 * @param name the name of the gui data.
 * @param func function to free the data
 */
void object_free_data(object_t* object, const char* name, FreeFunc func)
{
    GList* dlist = object_get_data_link(object, name);

    if (dlist)
    {
        object_data_t* data = dlist->data;
        if (func) func(data->data);
        g_free(data->name);
        g_free(data);
        object->priv->data = g_list_delete_link(object->priv->data, dlist);
    }
    else
    {
        g_log("OBJECT", G_LOG_LEVEL_WARNING,
              "cannot free data: '%s' not found in object (%s) %s",
              name, object_get_type(object),
              object->name?object->name:"<none>");
    }
}

/**
 * g_free the gui data field and set it to 0.
 * @param object the object.
 * @param name the name of the gui data.
 * @param func function to free the data
 */
void object_free_data_common(object_t* object, const char* name)
{
    object_free_data(object, name, g_free);
}


/** Destroy a object. You always must use this function (or
 * object_destroy()) to destroy an object which is derived from this
 * class. Before the object gets destroy, all observers are informed.
 * @param object the object to destroy
 * @param data user data for the observers.
 * @see object_notify_d(), object_notify_o()
 */
void object_kill_full(object_t* object, void* data)
{
    GList* dlist;

/*     g_log("====", G_LOG_LEVEL_DEBUG, "object_kill_full: (%s) %s", */
/*           object->type, object->name); */

    OBJ_ASSERT(object, !OBJ_DESTROYING(object),
               "don't destroy object during destruction");

    object->priv->flags |= OBJ_FLAG_DESTROY;

    for (dlist = object->priv->notify; dlist; dlist = dlist->next)
    {
        object_func_t* func = dlist->data;
        func->func(object, func->data);
    }
    object->destroy(object, data);
}

/** Destroy the object. Calls object_destroy_full() 
 * @param object the object.
 * @see object_destroy_full()
 */
void object_kill(object_t* object)
{
    object_kill_full(object, NULL);
}



static void object_reset_ref(object_t* object, object_t** pointer)
{
    g_assert(object && pointer && object==*pointer);
    *pointer = NULL;
}

void object_set_ptr(object_t** object)
{
    g_assert(object);
    object_notify_add(*object, (ObjectFunc)object_reset_ref, object);
}

void object_unset_ptr(object_t** object)
{
    g_assert(object);
    object_notify_remove(*object, (ObjectFunc)object_reset_ref, object);
}

void object_set_name(object_t* object, const char* name)
{
    g_free(object->name);
    object->name = name?g_strdup(name):NULL;
}

/** Make a new object public.
 * All observers get notified.
 * @param object the new object
 * @see object_register_instance
 * @param data user data to be passed to func
 */
void object_make_public(object_t* object, void* data)
{
    if (InstanceMonitor)
    {
        GSList* slist = object->priv->type;
        const char* type = slist->data;
        GList* observers = g_hash_table_lookup(InstanceMonitor, type);
        GList* dlist;

        for (dlist = observers; dlist; dlist = dlist->next)
        {
            ObjectFunc func = dlist->data;
            func(object, data);
        }
    }
    if (InstanceMonitorInh)
    {
        GSList* slist;
        for (slist = object->priv->type ; slist; slist = slist->next)
        {
            const char* type = slist->data;
            GList* observers = g_hash_table_lookup(InstanceMonitorInh, type);
            GList* dlist;
            
            for (dlist = observers; dlist; dlist = dlist->next)
            {
                ObjectFunc func = dlist->data;
                func(object, data);
            }
        }
    }
}


/** Monitor for new instances of a particular object type
 * @param type the type of the object
 * @param func function to call on new instance
 * @param inherited TRUE to allow super classes too
 */
void object_register_instance(const char* type, ObjectFunc func, gboolean inherited)
{
    GList* dlist;
    GHashTable** table;

    if (inherited) table = &InstanceMonitorInh;
    else table = &InstanceMonitor;

    if (*table == NULL)
    {
        *table = g_hash_table_new_full
            ((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
             (GDestroyNotify)g_free, (GDestroyNotify)NULL);
    }

    dlist = g_hash_table_lookup(*table, type);

    if (dlist)
    {
        dlist = g_list_append(dlist, func);
    }
    else
    {
        dlist = g_list_append(NULL, func);
        g_hash_table_insert(*table, g_strdup(type), dlist);
    }
}

#ifdef WIN32
# define _OBJECT_C_
# include "lcorealiasdef.c"
#endif
