/* Copyright (C) 2005-2007 The Lopster Team
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.22 $
 * $Date: 2007/11/21 20:18:24 $
 * $Author: sgop $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>

#include <glib.h>

#include "plugin.h"
#include "version.h"
#include "preferences.h"
#include "utils.h"
#include "xml_handler.h"
#include "dynamic.h"
#include "l_base.h"
#include "string_list.h"
#include "ui.h"
#include "l_i18n.h"

#ifdef WIN32
#include "lcorealias.h"
#endif


typedef struct {
  char* filename;
  gboolean active;
  gboolean existent;
} plugin_known_t;

typedef struct {
  module_t _module;
  int NOI;
  GList* known_plugin;
  GList* plugins;
  GList* active;
  gboolean all_loaded;
} plugin_module_t;

static GList* Folders = NULL;
static plugin_module_t* PluginModule = NULL;

static void plugins_save(void)
{
    if (lopster_xml_loaded("core:plugins"))
        lopster_xml_save("core:plugins", 20000);
}

static void plugin_known_free(plugin_known_t* known)
{
    g_free(known->filename);
    g_free(known);
}

static plugin_known_t* plugin_known_find(const char* filename)
{
    GList* dlist;
    for (dlist = PluginModule->known_plugin; dlist; dlist = dlist->next)
    {
        plugin_known_t* known = dlist->data;
        if (!strcmp(known->filename, filename)) return known;
    }
    return NULL;
}

gboolean plugin_known_as_active(const char* filename)
{
    plugin_known_t* known = plugin_known_find(filename);
    return known?known->active:TRUE;
}

void plugin_known_active(const char* filename, gboolean active)
{
    plugin_known_t* known = plugin_known_find(filename);
    
    if (!known)
    {
        g_log("PLUGIN", G_LOG_LEVEL_WARNING,
              "plugin '%s' is not known", filename);
        return;
    }
    if (known->active != active)
    {
        known->active = active;
        plugins_save();
    }
}

static void plugin_known_existent(const char* filename, gboolean existent)
{
    plugin_known_t* known = plugin_known_find(filename);
    
    if (!known)
    {
        known = g_malloc0(sizeof(*known));
        known->filename = g_strdup(filename);
        PluginModule->known_plugin = g_list_append(PluginModule->known_plugin, known);
        known->existent = existent;
        plugins_save();
    }
}


/** Get a status string for the plugin.
 * @param plugin the plugin
 * @return the status of the plugin (whether loaded or not, etc.) 
 */
const char* plugin_status_string(plugin_t* plugin)
{
    g_assert(plugin);
    switch (plugin->status) {
    case PLUGIN_INACTIVE: return _("inactive");
    case PLUGIN_ACTIVE: return _("active");
    case PLUGIN_VERSION: return _("incompatible");
    case PLUGIN_DUPLICATE: return _("duplicate");
    default: return "__error__";
    }
}

static void plugin_online_status_destroy(online_status_t* status)
{
    object_destroy(OBJ(status));
}

static online_status_t* plugin_online_status_new()
{
    online_status_t* status = g_malloc0(sizeof(*status));
    object_init(OBJ(status), "Online status", TRUE);
    object_add_type(OBJ(status), "OnlineStatus");
    OBJ(status)->destroy = (ObjectFunc)plugin_online_status_destroy;
    return status;
}

/** Set the status of your plugin.
 * @param id the plugin id
 * @param status the new status
 */
void plugin_set_online_status(int id, int status)
{
    plugin_t* plugin = plugin_get(id);
    g_assert(plugin);
    if (plugin->online->status == status) return;
    plugin->online->status = status;
    object_set_dirty(OBJ(plugin->online));
}

/** Set the network stats.
 * @param id the plugin id,
 * @param users number of users on that network, or 0
 * @param files of files on that network, or 0
 * @param gigs of GiB on that network, or 0
 */
void plugin_set_online_stats(int id, unsigned users, unsigned files, unsigned gigs)
{
    plugin_t* plugin = plugin_get(id);
    g_assert(plugin);
    plugin->online->users = users;
    plugin->online->files = files;
    plugin->online->gigs = gigs;
    object_set_dirty(OBJ(plugin->online));
}

static void plugin_info_destroy(plugin_info_t* info) {
  g_free(info->name);
  g_free(info->sname);
  g_free(info->copyright);
  g_free(info->icon);
  g_free(info->license);
  g_free(info->website);
  g_free(info->comment);
  g_list_foreach(info->authors, (GFunc)g_free, NULL);
  g_list_free(info->authors);
  g_free(info);
}

static void plugin_destroy(plugin_t* plugin) {
  // this function normally is called on shutdown
  // (when this module doesn't exist anymore)
  if (PluginModule) {
    PluginModule->plugins = g_list_remove(PluginModule->plugins, plugin);
  }
  g_free(plugin->filename);
  if (plugin->online) object_kill(OBJ(plugin->online));
  if (plugin->info) plugin_info_destroy(plugin->info);
  module_destroy(MODULE(plugin));
}

/** Initialize the plugin. You must call this function in your start
 * method of the plugin.
 * @param plugin the plugin
 * @param name the name of the plugin (passed to module_init())
 * @param init init method (passed to module_init())
 * @param run run method (passed to module_init())
 * @param done done method (passed to module_init())
 */
void plugin_init(plugin_t* plugin, const char* name,
                 ModuleFunc init, ModuleFunc run, ModuleFunc done) {
  module_init(MODULE(plugin), name, init, run, done);
  object_add_type(OBJ(plugin), "Plugin");
  OBJ(plugin)->destroy = (ObjectFunc)plugin_destroy;

  plugin->online = plugin_online_status_new();

}

static gboolean plugin_already_loaded(plugin_t* plugin) {
  GList* dlist;

  g_assert(PluginModule);
  for (dlist = PluginModule->plugins; dlist; dlist = dlist->next) {
    plugin_t* p = dlist->data;
    if (p->status == PLUGIN_ACTIVE &&
	!strcmp(p->info->name, plugin->info->name)) return TRUE;
  }
  return FALSE;
}

static void add_folder(const char* folder)
{
    g_log("PLUGIN", G_LOG_LEVEL_DEBUG,
          "adding to library path: %s", folder);
    Folders = g_list_append(Folders, g_strdup(folder));
}

static void setup_folders(void)
{
    static gboolean called = FALSE;
    
    if (called) return;
    else called = TRUE;
    
#ifdef PACKAGE_SOURCE_DIR
    // the plugins
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("napster") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("soulseek") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("edonkey2000") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("lobster") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("directconnect") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("bittorrent") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("gnutella") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("fasttrack") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("lastfm") ".libs");
    
    // the plugin gtk2 uis
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("napster") SDIR("gtk2_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("soulseek") SDIR("gtk2_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("edonkey2000") SDIR("gtk2_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("lobster") SDIR("gtk2_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("directconnect") SDIR("gtk2_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("bittorrent") SDIR("gtk2_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("gnutella") SDIR("gtk2_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("fasttrack") SDIR("gtk2_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("lastfm") SDIR("gtk2_ui") ".libs");
    
    // the plugin qt4 uis
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("napster") SDIR("qt4_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("soulseek") SDIR("qt4_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("edonkey2000") SDIR("qt4_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("lobster") SDIR("qt4_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("directconnect") SDIR("qt4_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("bittorrent") SDIR("qt4_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("gnutella") SDIR("qt4_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("fasttrack") SDIR("qt4_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("plugin") SDIR("lastfm") SDIR("qt4_ui") ".libs");
    
    // the core gui
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("lcore") SDIR("gtk2_ui") ".libs");
    add_folder(SDIR(PACKAGE_SOURCE_DIR) SDIR("lcore") SDIR("qt4_ui") ".libs");
#else
    add_folder(PACKAGE_LIB_DIR);
#endif
}

const char* plugin_foreach_folder(FolderSearchFunc func, gpointer data)
{
    GList* dlist;

    setup_folders();

    for (dlist = Folders; dlist; dlist = dlist->next)
    {
        char* folder = dlist->data;
        const char* res = func(folder, data);
        if (res) return res;
    }
    return NULL;
}

static void plugins_search(void)
{
    GList* dlist;
    
    setup_folders();
    
    g_log("PLUGIN", G_LOG_LEVEL_DEBUG, "searching for plugins");
    
    for (dlist = Folders; dlist; dlist = dlist->next)
    {
        const char* f = dlist->data;
        GDir* dir = g_dir_open(f, 0, NULL);
        const char* entry;
        if (!dir)
        {
            g_log("PLUGIN", G_LOG_LEVEL_WARNING, "cannot open folder '%s'", f);
            continue;
        }
        while ((entry = g_dir_read_name(dir)) != NULL)
        {
            GModule* handle;
            plugin_t* plugin;
            PluginInitFunc p_init;
            char* temp = l_printf(SDIR("%s")"%s", f, entry);
            plugin_info_t* info;
            unsigned len = strlen(temp);
            gpointer symbol;

            if (len < 9) continue;
            if (g_file_test(temp, G_FILE_TEST_IS_DIR)) continue;
            if (g_file_test(temp, G_FILE_TEST_IS_SYMLINK)) continue;
            if (strncmp(entry, "libplugin", 9)) continue;
            if (!strcmp(temp+(len-4), ".lai")) continue;
            if (!strcmp(temp+(len-3), ".la")) continue;
            if (!strcmp(temp+(len-2), ".a")) continue;
            if (!strcmp(temp+(len-2), ".o")) continue;
#ifdef WIN32
            if (!strcmp(temp+(len-4), ".ilk")) continue;
            if (!strcmp(temp+(len-4), ".exp")) continue;
            if (!strcmp(temp+(len-4), ".pdb")) continue;
            if (!strcmp(temp+(len-4), ".lib")) continue;
#endif
            handle = lopster_dll_open(temp, G_MODULE_BIND_LAZY);
            if (!handle)
            {
                g_log("PLUGIN", G_LOG_LEVEL_INFO, "%s => skip", g_module_error());
                continue;
            }
            if (!g_module_symbol(handle, "plugin_initialize", &symbol))
            {
                g_log("PLUGIN", G_LOG_LEVEL_WARNING, "symbol 'plugin_initialize' not found");
                lopster_dll_close(handle);
                continue;
            } else p_init = symbol;
      
            plugin = g_malloc0(sizeof(*plugin));
            plugin->filename = g_strdup(temp);
            plugin->info = g_malloc0(sizeof(*info));
            PluginModule->plugins = g_list_append(PluginModule->plugins, plugin);
            
            p_init(plugin->info);
            
            if (!plugin->info->name)
            {
                g_log("PLUGIN", G_LOG_LEVEL_WARNING, "plugin does not set name");
                plugin->info->name = _("UnknownPlugin");
            }
            if (!plugin->info->sname)
            {
                g_log("PLUGIN", G_LOG_LEVEL_WARNING, "plugin does not set short name");
                plugin->info->sname = _("???");
            }
            if (!plugin->info->comment)
            {
                g_log("PLUGIN", G_LOG_LEVEL_WARNING, "plugin does not set description");
                plugin->info->comment = _("Unknown plugin which has no appropiate info set");
            }
            
            if ((plugin->info->if_version>>16) != (IF_VERSION>>16))
            {
                // major/minor version differs, patches are allowed.
                g_log("PLUGIN", G_LOG_LEVEL_WARNING,
                      "plugin '%s' has invalid interface version", plugin->info->name);
                plugin->status = PLUGIN_VERSION;
            }
            else if (plugin->info->pl_version == 0)
            {
                g_log("PLUGIN", G_LOG_LEVEL_WARNING,
                      "plugin '%s' has no interface version", plugin->info->name);
                plugin->status = PLUGIN_VERSION;
            }
            else if (plugin_already_loaded(plugin))
            {
                g_log("PLUGIN", G_LOG_LEVEL_WARNING,
                      "plugin with name '%s' already loaded", plugin->info->name);
                plugin->status = PLUGIN_DUPLICATE;
            }
            else
            {
                plugin_known_t* known = plugin_known_find(plugin->filename);
                gboolean load;
                if (!known)
                {
                    g_log("PLUGIN", G_LOG_LEVEL_INFO,
                          "new plugin '%s' will be loaded by default", plugin->info->name);
                    load = TRUE;
                }
                else if (known->active)
                {
                    g_log("PLUGIN", G_LOG_LEVEL_INFO,
                          "known plugin '%s' will be loaded", plugin->info->name);
                    load = TRUE;
                }
                else
                {
                    g_log("PLUGIN", G_LOG_LEVEL_INFO,
                          "known plugin '%s' will not be loaded", plugin->info->name);
                    load = FALSE;
                }
                // now load that thing.
                if (load)
                {
                    plugin->status = PLUGIN_ACTIVE;
                    plugin->ID = PluginModule->NOI++;
                    PluginModule->active = g_list_append(PluginModule->active, plugin);
                    
                    module_require_direct(MODULE(plugin), (ModuleFunc)plugin->info->start);
                }
                else
                {
                    plugin->status = PLUGIN_INACTIVE;
                }
            }

            // now save some info for the plguin.
            plugin_known_existent(plugin->filename, TRUE);
            if (plugin->status != PLUGIN_ACTIVE)
            {
                PluginModule->all_loaded = FALSE;
                lopster_dll_close(handle);
                plugin_known_active(plugin->filename, FALSE);
            }
            else
            {
                plugin_known_active(plugin->filename, TRUE);
            }
        }
        g_dir_close(dir);
    }
}

/** Check wehther all plugin that were found are actually loaded.
 * @return true if all plugins were loaded, false otherwise
 */
gboolean plugin_all_loaded() {
  g_assert(PluginModule);
  return PluginModule->all_loaded;
}

/** Get the plugin from the list of active plugins.
 * @param index index of the plugin in the list.
 * @return the plugin, or NULL if invalid index
 */
plugin_t* plugin_get(int index) {
  GList* dlist;

  g_assert(PluginModule);

  dlist = g_list_nth(PluginModule->active, index);
  if (dlist) return dlist->data;
  else return NULL;
}

/** Get the number of loaded plugins
 * @return the number of plugins loaded
 */
int plugin_get_number() {
  g_assert(PluginModule);
  return PluginModule->NOI;
}

/** Get a full list of plugins, including the not loaded ones.
 * @return list of plugin
 */
GList* plugin_get_list() {
  g_assert(PluginModule);
  return PluginModule->plugins;
}

/** Get the name of the plugin.
 * @param plugin the plugin
 * @return name of the plugin
 */
const char* plugin_get_name(plugin_t* plugin) {
  return plugin?plugin->info->name:"Core";
}

/** Get the short name of the plugin.
 * @param plugin the plugin
 * @return short name of the plugin
 */
const char* plugin_get_short_name(plugin_t* plugin) {
  return plugin?plugin->info->sname:"Core";
}

/** Get the interger (encoded) version of the plugin.
 * @param plugin the plugin
 * @return encoded version as unsigned
 */
unsigned plugin_get_version(plugin_t* plugin) {
  return plugin?plugin->info->pl_version:(unsigned)IF_VERSION;
}

/** Get the version as string.
 * @param plugin the plugin
 * @return the version as string
 */
const char* plugin_get_version_string(plugin_t* plugin) {
  if (!plugin) {
    if (V_COMMIT) {
      return l_printf("%s/%u", PURE_VERSION, V_COMMIT);
    } else {
      return PURE_VERSION;
    }
  } else {
    unsigned version = plugin_get_version(plugin);
    if (version & 0xff) {
      return l_printf("%u.%u.%u/%u", (version>>24)&0xff,
                      (version>>16)&0xff, (version>>8)&0xff, (version>>0)&0xff);
    } else {
      return l_printf("%u.%u.%u",
                      (version>>24)&0xff, (version>>16)&0xff, (version>>8)&0xff);
    }
  }
}

/** Get the version of the plugin on it's last run.  This function is
 * useful to detect whether there was an upgrade or whether it is
 * necessary to start a setup wizard.
 * @param plugin the plugin
 * @return the version of plugin during its last run 
 */
int plugin_get_old_version(plugin_t* plugin) {
  char* pref = l_printf("/version/%s", plugin_get_short_name(plugin));
  return pref_get_ref(pref, int);
}

/** Get the id of a plugin.
 * @param name the name of the plugin
 * @return the idnex of the plugin with the according name, or -2 if none found.
 * @note the name 'core' returns -1
 */
int plugin_get_id(const char* name) {
  GList* dlist;
  int num = 0;
  
  g_assert(PluginModule);

  if (!name) return -2;

  if (!strncasecmp(name, "core", 4) &&
      (name[4] == 0 || name[4] == ' ')) return -1;

  for (dlist = PluginModule->active; dlist; dlist = dlist->next, num++) {
    plugin_t* plugin = dlist->data;
    int len;
    len = strlen(plugin->info->sname);
    if (!strncasecmp(plugin->info->sname, name, len) &&
	(name[len] == 0 || name[len] == ' ')) return num;
    len = strlen(plugin->info->name);
    if (!strncasecmp(plugin->info->name, name, len) &&
	(name[len] == 0 || name[len] == ' ')) return num;
  }
  return -2;
}

static void xml_plugin_load(xmlDocPtr xfile, xmlNodePtr xnode)
{
    xmlChar* attr = xmlGetProp(xnode, "active");
    gboolean active = FALSE;
    
    if (attr)
    {
        if (!strcasecmp(attr, "true") || !strcasecmp(attr, "yes") || !strcmp(attr, "1"))
            active = TRUE;
        xmlFree(attr);
    }
    
    attr = xmlNodeListGetString(xfile, xnode->xmlChildrenNode, 1);
    if (attr)
    {
        plugin_known_existent(attr, FALSE);
        plugin_known_active(attr, active);
        xmlFree(attr);
    }
}

static void xml_plugins_load(xmlDocPtr xfile, xmlNodePtr xnode) {
  xmlNodePtr xsub = xnode->xmlChildrenNode;
  while (xsub != NULL) {
    if (!strcmp(xsub->name, "Plugin")) {
      xml_plugin_load(xfile, xsub);
    }
    xsub = xsub->next;
  }
}

static void xml_plugin_save(plugin_known_t* known, xmlNodePtr xnode)
{
    if (known->existent)
    {
        xmlNodePtr xsub = xmlNewTextChild(xnode, NULL, "Plugin", known->filename);
        xmlNewProp(xsub, "active", known->active?"yes":"no");
    }
}

static void xml_plugins_save(xmlDocPtr xfile, xmlNodePtr xnode)
{
    (void)xfile;
    g_list_foreach(PluginModule->known_plugin, (GFunc)xml_plugin_save, xnode);
}

void plugin_connect_change(int id, unsigned status)
{
    if (id < 0)
    {
        GList* dlist;
        for (dlist = PluginModule->active; dlist; dlist = dlist->next)
        {
            plugin_t* plugin = dlist->data;
            if (plugin && plugin->connect) plugin->connect(status);
        }
    }
    else
    {
        plugin_t* plugin = plugin_get(id);
        if (plugin && plugin->connect) plugin->connect(status);
    }
}



static void plugins_init() {
  int i1;

  g_assert(PluginModule);

  if (PluginModule->all_loaded) string_lists_save_unregistered(FALSE);

  // register version numbers
  for (i1 = -1; i1 < plugin_get_number(); i1++) {
    char* pref = l_printf("/version/%s", plugin_get_short_name(plugin_get(i1)));
    pref_register_int(pref, TRUE, 0);
  }
  
}

static void plugins_run() {
  int i1;

  for (i1 = -1; i1 < plugin_get_number(); i1++) {
    plugin_t* plugin = plugin_get(i1);
    char* pref = l_printf("/version/%s", plugin_get_short_name(plugin_get(i1)));
    pref_store(pref, plugin_get_version(plugin));
  }
  
}

static void plugins_done() {
  g_list_foreach(PluginModule->known_plugin, (GFunc)plugin_known_free, NULL);
  g_list_free(PluginModule->known_plugin);
  g_list_free(PluginModule->plugins);
  g_list_free(PluginModule->active);
  PluginModule = NULL;
}

static void plugin_module_init(plugin_module_t* module)
{
    module_init(MODULE(module), "Plugin Module",
                (ModuleFunc)plugins_init, (ModuleFunc)plugins_run, 
                (ModuleFunc)plugins_done);
    module->NOI = 0;
    module->plugins = NULL;
    module->active = NULL;
    module->known_plugin = NULL;
    module->all_loaded = TRUE;
    
    module_require(dynamic_module_instance);
    module_require(xml_module_instance);
    module_require(string_list_module_instance);
    // better require the base module because we call an init function of
    // every plugin before it has the chance to require any module.
    module_require(base_module_instance);
    
    lopster_xml_register_home
        ("core:plugins", "plugins.xml", "Plugins",
         (XmlNodeFunc)xml_plugins_load, (XmlNodeFunc)xml_plugins_save);
    lopster_xml_load("core:plugins");
    
    if (g_module_supported())
    {
        plugins_search();
    }
    else
    {
        g_log("PLUGIN", G_LOG_LEVEL_CRITICAL,
              "plugins not supported by system");
    }
}


module_t* plugin_module_instance()
{
    if (!PluginModule)
    {
        PluginModule = g_malloc0(sizeof(*PluginModule));
        plugin_module_init(PluginModule);
    }
    return MODULE(PluginModule);
}

const char* plugin_get_icon(plugin_t* plugin)
{
    if (!plugin) return NULL;  //FIXME: create lopster icon
    return plugin->info->icon;
}

#ifdef WIN32
#define _PLUGIN_C_
#include "lcorealiasdef.c"
#endif
