/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.9 $
 * $Date: 2007/11/21 20:18:24 $
 * $Author: sgop $
 */

#ifndef _PLUGIN_H_
#define _PLUGIN_H_

#include <glib.h>
#include <gmodule.h>

#include "module.h"
G_BEGIN_DECLS

/** @defgroup PluginModule Creating new plugins for lopster
 *
 * To create plugin for lopster you have to build a shared library which exports the symbol
 * plugin_initialize() of type PluginInitFunc.
 *
 * @ingroup BasicFramework
 * @{
 */

#define PLUGIN_INACTIVE        0
#define PLUGIN_ACTIVE          1
#define PLUGIN_VERSION         2
#define PLUGIN_DUPLICATE       3

/** @name Status flags of the plugins */
/*@{*/
#define ONLINE_STATUS_IDLE   0x0  ///< Plugin is idle
#define ONLINE_STATUS_ONLINE 0x1  ///< Plugin is online
/*@}*/

/** @name Flags to control the connection status of a plugin */
/*@{*/
#define CONNECT_START          1  ///< Start auto-connecting
#define CONNECT_STOP           2  ///< Stop auto-connecting
/*@}*/

typedef struct _plugin_priv_t plugin_priv_t;
typedef struct _plugin_info_t plugin_info_t;

#define ACTION(f) static void (f)(va_list ap)

/** The entry point for a plugin */
typedef void (*PluginInitFunc)(plugin_info_t* info);
/** Function to control the connect status of a plugin */
typedef void (*PluginConnectFunc)(unsigned mode);

typedef const char* (*FolderSearchFunc)(const char* folder, gpointer data);


/** The online status of the plugin. Never set the values of this
 * struct manually, instead use plugin_set_online_status() and
 * plugin_set_online_stats()
 */
typedef struct {
  object_t _object;   ///< the base class
  int status;         ///< the status
  unsigned users;     ///< the number of users online, or 0 if unknown
  unsigned files;     ///< the number of files in the network, or 0 if unknown
  unsigned gigs;      ///< the number of GiB in the network, or 0 if unknown
} online_status_t;

/** The plugin class. */
typedef struct {
  // initialize
  module_t _module;          ///< the base class
  PluginConnectFunc connect; ///< function to control the connection status
  // read
  online_status_t* online;   ///< the online status, see plugin_set_online_status()
  int ID;                    ///< the unique ID > 0 of the plugin
  // private
  plugin_info_t* info;
  char* filename;
  int status;
} plugin_t;

/** The real plugin start function. Make sure that you call
 * plugin_init() inside this function before you require any modules.
 * @note At this point, the plugin already has a unique ID assigned
 * which you can save for later use.
 */
typedef void (*PluginStartFunc)(plugin_t* plugin);

/** The plugin information. This is the first thing the core will know
 * from you.  Once the plugin was selected for startup, a plugin of
 * type plugin_t will be created and loaded like a normal module via
 * module_require().
 */
struct _plugin_info_t {
  char* name;            ///< the name of the plugin, e.g. OpenNap
  char* sname;           ///< the short name of the plugin, e.g. Nap
  char* copyright;       ///< the copyright information, utf8
  unsigned if_version;   ///< the interface version, should always be set to BUILD_VERSION
  unsigned pl_version;   ///< the plugin version
  char* icon;            ///< the filename of the icon.
  char* license;         ///< license information
  GList* authors;        ///< the authors
  char* website;         ///< a link to the website if available
  char* comment;         ///< a short description of the plugin/network
  PluginStartFunc start; ///< the start function, this is similar to the module_instance functions.
};


module_t* plugin_module_instance();


void plugin_init(plugin_t* plugin, const char* name,
                 ModuleFunc init, ModuleFunc run, ModuleFunc done);

void plugin_set_online_status(int id, int status);

void plugin_set_online_stats(int id, unsigned users, unsigned files, unsigned gigs);

int plugin_get_number(void);

plugin_t* plugin_get(int index);

unsigned plugin_get_version(plugin_t* plugin);

int plugin_get_old_version(plugin_t* plugin);

const char* plugin_get_name(plugin_t* plugin);

const char* plugin_get_short_name(plugin_t* plugin);

const char* plugin_get_icon(plugin_t* plugin);

const char* plugin_get_version_string(plugin_t* plugin);

const char* plugin_status_string(plugin_t* plugin);

gboolean plugin_all_loaded(void);
int plugin_get_id(const char* name);

GList* plugin_get_list(void);

/*@}*/

void plugin_known_active(const char* filename, gboolean active);
gboolean plugin_known_as_active(const char* filename);
void plugin_connect_change(int id, unsigned status);
const char* plugin_foreach_folder(FolderSearchFunc func, gpointer data);

G_END_DECLS
#endif
