/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.11 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>

#include <libxml/parser.h>

#include "xml_handler.h"
#include "preferences.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
  pref_node_t* prefs;
} pref_module_t;

typedef struct {
  constraint_t _c;
  int value;
} constraint_int_t;

typedef struct {
  constraint_t _c;
  char from, to;
} constraint_replace_t;

static pref_module_t* PrefModule = NULL;

pref_node_t* __node__;



static void pref_save(pref_node_t* node) {
  g_assert(PrefModule);
  if (node && !node->save) return;
  if (!module_running(MODULE(PrefModule))) return;
  lopster_xml_save("core:prefs", 10000);
}


static void pref_node_destroy(pref_node_t* node) {
  if (!node) return;

  g_list_foreach(node->children, (GFunc)object_kill, NULL);
  g_list_free(node->children);

  g_list_foreach(node->constraint, (GFunc)g_free, NULL);
  g_list_free(node->constraint);

  if (node->value) node->free(node->value);
  object_destroy(OBJ(node));
}

static void
pref_node_init(pref_node_t* node, const char* name, gboolean save,
	       FreeFunc free, StoreFunc store,
	       StoreStrFunc store_str, XMLPrintFunc xml_print) {
  object_init(OBJ(node), name, TRUE);
  object_add_type(OBJ(node), "PrefNode");
  OBJ(node)->destroy = (ObjectFunc)pref_node_destroy;
  node->save = save;
  node->free = free;
  node->store = store;
  node->store_str = store_str;
  node->xml_print = xml_print;
}

static pref_node_t* 
pref_node_new(const char* name, gboolean save, FreeFunc free, StoreFunc store,
              StoreStrFunc store_str, XMLPrintFunc xml_print) {
  pref_node_t* node;

  node = g_malloc0(sizeof(*node));
  pref_node_init(node, name, save, free, store, store_str, xml_print);
  return node;
}

static pref_node_t* pref_search_child(pref_node_t* parent, const char* name) {
  GList* dlist;
  pref_node_t* child;

  if (!parent || !name) return NULL;
  for (dlist = parent->children; dlist; dlist = dlist->next) {
    child = dlist->data;
    if (!strcmp(OBJ(child)->name, name)) return child;
  }
  return NULL;
}

static pref_node_t*
pref_search_rec(pref_node_t* node, const char* name, gboolean only_leaf) {
  unsigned l1 = strlen(name);
  unsigned l2 = strlen(OBJ(node)->name);

  if (l1 < l2) return NULL;

  if (strncmp(OBJ(node)->name, name, l2)) return NULL;
  if (node->children) {
    GList* dlist;
    if (!only_leaf && name[l2] == 0) return node;
    if (name[l2] != '/') return NULL;
    for (dlist = node->children; dlist; dlist = dlist->next) {
      pref_node_t* child = dlist->data;
      node = pref_search_rec(child, name+l2+1, only_leaf);
      if (node) return node;
    }
  } else {
    if (l1 != l2) return NULL;
    return node;
  }
  return NULL;
}

/** Search for a value by name.
 * @param name the name of the value.
 * @param only_leaf whether or not inner nodes are returned.
 * @return the node if found, or NULL
 */
pref_node_t* pref_search(const char* name, gboolean only_leaf) {
  pref_node_t* node;
  
  g_assert(PrefModule);
  if (!name) return NULL;
  if (!PrefModule->prefs) return NULL;

  node = pref_search_rec(PrefModule->prefs, name, only_leaf);
  if (!node) {
    g_log("PREF", G_LOG_LEVEL_WARNING, "value '%s' not found", name);
  }
  return node;
}

static gboolean pref_node_correct(pref_node_t* node) {
  gboolean ret = FALSE;
  GList* dlist;
  for (dlist = node->constraint; dlist; dlist = dlist->next) {
    constraint_t* constraint = dlist->data;
    if (constraint->func) ret |= constraint->func(node, constraint);
  }
  return ret;
}

/** Store a registered value.  The argument(s) depend on the type of
 * the value. They are passed to a function of type StoreFunc. Make
 * sure you pass the correct arguments.
 * @param name the name of the value.
 * @param ... additional arguments.
 */
void pref_store(const char* name, ...) {
  pref_node_t* node = pref_search(name, TRUE);
  va_list ap;
  void* res;

  if (!node) g_assert(FALSE);  // we dont allow errors
  
  va_start(ap, name);
  res = node->store(node->value, ap);
  va_end(ap);
  if (node->value != res) {
    node->value = res;
    pref_node_correct(node);
    object_set_dirty(OBJ(node));
  }
}

static void pref_store_str_real(pref_node_t* node, const char* value) {
  void* res = node->store_str(node->value, value);
  if (node->value != res) {
    node->value = res;
    pref_node_correct(node);
    object_set_dirty(OBJ(node));
  }
}

/** Store a value from a string.
 * @param name the name of the value.
 * @param value the value representated by a string
 */
void pref_store_str(const char* name, const char* value) {
  pref_node_t* node = pref_search(name, TRUE);
  if (!node) return;
  pref_store_str_real(node, value);
}

/** Register a value.
 * @param name the name of the value.
 * @param save whether or not to save this value
 * @param xml_print function to print the value into a (ASCII) string
 * @param free function that frees the value
 * @param store function to store the value
 * @param store_str function to store the value from a string
 * @param ... inital values as accepted by the store function.
 */
void pref_register(const char* name, gboolean save,
                   XMLPrintFunc xml_print,
                   FreeFunc free,
                   StoreFunc store,
                   StoreStrFunc store_str, ...) {
  pref_node_t* parent;
  pref_node_t* child;
  char* rname = l_strdup(name);
  char* pos;
  va_list ap;

  g_assert(PrefModule);
  g_assert(store && store_str);
  if (!PrefModule->prefs) {
    g_log("PREF", G_LOG_LEVEL_ERROR, "no root node available");
    return;
  }

  // search position to insert new node
  parent = PrefModule->prefs;
  pos = strchr(rname, '/');
  if (!pos) {
    g_log("PREF", G_LOG_LEVEL_WARNING, "no insertion node found");
    return;
  }
  *pos++ = 0;
  if (strcmp(OBJ(parent)->name, rname)) {
    g_log("PREF", G_LOG_LEVEL_WARNING, "no insertion node found");
    return;
  }
  rname = pos;
  while ((pos = strchr(rname, '/')) != NULL) {
    GList* dlist;

    *pos = 0;
    for (dlist = parent->children; dlist; dlist = dlist->next) {
      pref_node_t* child = dlist->data;
      if (!strcmp(OBJ(child)->name, rname)) {
	parent = child;
	rname = pos +1;
	break;
      }
    }
    *pos = '/';
    if (!dlist) break;
  }

  if (*rname == 0) {
    g_log("PREF", G_LOG_LEVEL_WARNING,
	  "value '%s' already registered", name);
    return;
  }

  while (rname) {
    pos = strchr(rname, '/');
    if (pos) {
      *pos++ = 0;
      child = pref_node_new(rname, save, NULL, NULL, NULL, NULL);
      parent->children = g_list_append(parent->children, child);
      child->parent = parent;
      parent = child;
      rname = pos;
    } else {
      child =
	pref_node_new(rname, save, free?free:(FreeFunc)g_free,
		      store, store_str, xml_print);
      parent->children = g_list_append(parent->children, child);
      child->parent = parent;
      va_start(ap, store_str);
      child->value = child->store(child->value, ap);
      va_end(ap);

      pref_node_correct(child);
      break;
    }
  }

  {
    const char* val =
      child->xml_print?child->xml_print(child->value):NULL;
    
    g_log("PREF", G_LOG_LEVEL_DEBUG,
	  "registered value '%s' : '%s'", name, val?val:"(null)");
    parent = child->parent;
    while (parent) {
      if (save) parent->save = TRUE;
      parent = parent->parent;
    }
  }
  
  // now keep track of changes and save to disk
  object_monitor_d(OBJ(child), (ObjectFunc)pref_save, NULL);
  object_set_dirty(OBJ(child));
}

/** Register an int value. This is a wrapper for pref_register().
 * @param name the name of the value
 * @param save whether to save the value with the settings or not
 * @param value the int value
 */
void pref_register_int(const char* name, gboolean save, int value) {
  pref_register(name, save, 
		(XMLPrintFunc)print_int,
		(FreeFunc)NULL,
		(StoreFunc)store_int,
		(StoreStrFunc)store_str_int,
		value);
}

/** Register an ip value. This is a wrapper for pref_register().
 * @param name the name of the value
 * @param save whether to save the value with the settings or not
 * @param value the ip
 */
void pref_register_ip(const char* name, gboolean save,  ip_t value) {
  pref_register(name, save, 
                (XMLPrintFunc)print_ip,
                (FreeFunc)NULL,
                (StoreFunc)store_ip,
                (StoreStrFunc)store_str_ip,
                value);
}

/** Register a string. This is a wrapper for pref_register().
 * @param name the name of the value
 * @param save whether to save the value with the settings or not
 * @param value the string
 */
void pref_register_string(const char* name, gboolean save, const char* value) {
  pref_register(name, save, 
                (XMLPrintFunc)print_string,
                (FreeFunc)NULL,
                (StoreFunc)store_string,
                (StoreStrFunc)store_str_string,
                value);
}

/** Register a boolean value. This is a wrapper for pref_register().
 * @param name the name of the value
 * @param save whether to save the value with the settings or not
 * @param value the boolean value
 */
void pref_register_bool(const char* name, gboolean save, gboolean value) {
  pref_register(name, save, 
                (XMLPrintFunc)print_bool,
                (FreeFunc)NULL,
                (StoreFunc)store_bool,
                (StoreStrFunc)store_str_bool,
                value);
}

/** Observer a value for changes.
 * @param name the name of the value
 * @param func the function to be called when the values has changed
 * @param data user data passed to func
 */
void pref_observe(const char* name, ObjectFunc func, void* data)
{
    pref_node_t* node = pref_search(name, TRUE);
    
    g_assert(func);
    if (!node)
    {
        g_log("PREF", G_LOG_LEVEL_ERROR,
              "value '%s' not found", name?name:"(null)");
        return;
    }
    object_monitor_d(OBJ(node), func, data);
}

/** Stop observing a value.
 * @param name the name of the value
 * @param func the function that was called on changes.
 * @param data the user data that was passed to func
 */
void pref_unobserve(const char* name, ObjectFunc func, void* data)
{
    pref_node_t* node = pref_search(name, TRUE);
    
    g_assert(func);
    if (!node)
    {
        g_log("PREF", G_LOG_LEVEL_ERROR,
              "value '%s' not found", name?name:"(null)");
        return;
    }
    object_unmonitor_d(OBJ(node), func, data);
}

static void xml_get_children(xmlDocPtr xdoc, xmlNodePtr xnode,
                             pref_node_t* parent, gboolean skip_unknown) {
  pref_node_t* child;
  xmlChar* attr;
  xmlChar* attr2;
  xmlNodePtr xsub;

  xsub = xnode->xmlChildrenNode;
  while (xsub != NULL) {
    if (!strcmp(xsub->name, "Section")) {
      attr = xmlGetProp(xsub, "Name");
      child = pref_search_child(parent, attr);
      if (!child && skip_unknown) {
        g_log("PREF", G_LOG_LEVEL_WARNING,
              "Discarding unknown section '%s'", attr);
      } else {
        if (!child) {
          child = pref_node_new(attr, TRUE, NULL, NULL, NULL, NULL);
          parent->children = g_list_append(parent->children, child);
          child->parent = parent;
        }	
        xml_get_children(xdoc, xsub, child, skip_unknown);
      }
      xmlFree(attr);
    } else if (!strcmp(xsub->name, "Value")) {
      attr = xmlGetProp(xsub, "Name");
      child = pref_search_child(parent, attr);
      if (!child && skip_unknown) {
        g_log("PREF", G_LOG_LEVEL_WARNING,
              "Discarding unknown value '%s'", attr);
      } else {
        attr2 = xmlNodeListGetString(xdoc, xsub->xmlChildrenNode, 1);
        if (!child) {
          child =
            pref_node_new(attr, TRUE, (FreeFunc)g_free,
                          (StoreFunc)store_string,
                          (StoreStrFunc)store_str_string,
                          (XMLPrintFunc)print_string);
          parent->children = g_list_append(parent->children, child);
          child->parent = parent;
        }
        pref_store_str_real(child, attr2);
        xmlFree(attr2);
      }
      xmlFree(attr);
    }
    xsub = xsub->next;
  }
}

static void xml_load_prefs(xmlDocPtr xfile, xmlNodePtr xroot) {
  xml_get_children(xfile, xroot, PrefModule->prefs,
                   FALSE/*plugin_all_loaded()*/);
}  

static void prefs_load() {

  lopster_xml_load("core:prefs");

  // need to reset the ip, since we assume we have no static one
  if (!pref_get_ref("/core/local/force_ip", gboolean))
    pref_store("/core/local/ip", 0);
}

static void xml_save_node(pref_node_t* value, xmlNodePtr xnode) {
  xmlNodePtr xsub;

  if (!value->save) return;
  
  if (value->xml_print && value->children) {
    g_log("PREF", G_LOG_LEVEL_DEBUG, "should not happen");
  }
  if (value->xml_print) {
    xsub = xmlNewTextChild(xnode, NULL, "Value", 
                           value->xml_print(value->value));
    xmlNewProp(xsub, "Name", OBJ(value)->name);
  } else if (value->children) {
    xsub = xmlNewChild(xnode, NULL, "Section", NULL);
    xmlNewProp(xsub, "Name", OBJ(value)->name);
    g_list_foreach(value->children, (GFunc)xml_save_node, xsub);
  }
}

static void xml_save_prefs(xmlDocPtr xfile, xmlNodePtr xroot) {
  (void)xfile;
  g_list_foreach(PrefModule->prefs->children, (GFunc)xml_save_node, xroot);
}


static void prefs_init() {
  lopster_xml_register_home("core:prefs", "preferences.xml",
                            "Preferences", xml_load_prefs, xml_save_prefs);
}

static void prefs_run() {
  prefs_load();
}

static void prefs_exit() {
  object_kill(OBJ(PrefModule->prefs));
  PrefModule = NULL;
}

static void pref_module_init(pref_module_t* module) {
  module_init(MODULE(module), "Preferences Module",
              (ModuleFunc)prefs_init, (ModuleFunc)prefs_run, (ModuleFunc)prefs_exit);
  module->prefs = pref_node_new("", FALSE, NULL, NULL, NULL, NULL);

  module_require(xml_module_instance);
  //  module_require(MODULE(module), base_module_instance);
}

/** If you access the preferences in your module, you must require
 * this function.
 * @return the preferences module
 * @see module_require()
 */
module_t* pref_module_instance() {
  if (!PrefModule) {
    PrefModule = g_malloc0(sizeof(*PrefModule));
    pref_module_init(PrefModule);
  }
  return MODULE(PrefModule);
}

/** Add a constraint for a value. Every time the value changes, the
 * constraint is ask to adjust it if it's not valid
 * @param name the name of the value.
 * @param constraint the constraint.
 */
void pref_add_constraint(const char* name, constraint_t* constraint) {
  pref_node_t* node = pref_search(name, TRUE);
  
  if (!node) {
    g_log("PREF", G_LOG_LEVEL_ERROR,
          "value '%s' not found", name?name:"(null)");
    g_free(constraint);
    return;
  }
  node->constraint = g_list_append(node->constraint, constraint);
  if (pref_node_correct(node)) object_set_dirty(OBJ(node));
}

static gboolean pref_clip_min(pref_node_t* node, constraint_int_t* c) {
  int* val = node->value;
  if (val && *val < c->value) {
    g_log("PREF", G_LOG_LEVEL_WARNING, "correcting '%s' from %d to %d",
	  OBJ(node)->name, *val, c->value);
    *val = c->value;
    return TRUE;
  } else {
    return FALSE;
  }
}
static gboolean pref_clip_max(pref_node_t* node, constraint_int_t* c) {
  int* val = node->value;
  if (val && *val > c->value) {
    g_log("PREF", G_LOG_LEVEL_WARNING,
	  "correcting '%s' from %d to %d",
	  OBJ(node)->name, *val, c->value);
    *val = c->value;
    return TRUE;
  } else {
    return FALSE;
  }
}

static gboolean pref_replace(pref_node_t* node, constraint_replace_t* c) {
  char* str = node->value;
  gboolean ret = FALSE;

  while (str && *str) {
    if (*str == c->from) {
      *str = c->to;
      ret = TRUE;
    }
    str++;
  }
  return ret;
}

/** Limit a int value to a lower bound.
 * @param name the name of the value.
 * @param min the minimum value allowed.
 */
void pref_add_constraint_min(const char* name, int min) {
  constraint_int_t* c = g_malloc(sizeof(*c));
  CONSTRAINT(c)->func = (PrefFunc)pref_clip_min;
  c->value = min;
  pref_add_constraint(name, CONSTRAINT(c));
}

/** Limit a int value to an upper bound.
 * @param name the name of the value.
 * @param max the maximum value allowed.
 */
void pref_add_constraint_max(const char* name, int max) {
  constraint_int_t* c = g_malloc(sizeof(*c));
  CONSTRAINT(c)->func = (PrefFunc)pref_clip_max;
  c->value = max;
  pref_add_constraint(name, CONSTRAINT(c));
}

/** Limit a int value to a lower and upper bound.
 * @param name the name of the value.
 * @param min the minimum value allowed.
 * @param max the maximum value allowed.
 */
void pref_add_constraint_minmax(const char* name, int min, int max) {
  pref_add_constraint_min(name, min);
  pref_add_constraint_max(name, max);
}

/** Disallow a certain character in a string value.
 * @param name the name of the value.
 * @param from the character that is not allowed.
 * @param to the character that replaces the not allowed one.
 */
void pref_add_constraint_replace(const char* name, char from, char to) {
  constraint_replace_t* c = g_malloc(sizeof(*c));
  CONSTRAINT(c)->func = (PrefFunc)pref_replace;
  c->from = from;
  c->to = to;
  pref_add_constraint(name, CONSTRAINT(c));
}

/** Limit an int value to the range of possible ports [0,65535]
 * @param name the name of the value.
 */
void pref_add_constraint_port(const char* name) {
  pref_add_constraint_min(name, 0);
  pref_add_constraint_max(name, 65535);
}

static void pref_foreach_rec(pref_node_t* node, ObjectFunc func,
			     void* data) {
  if (node->children) {
    GList* dlist;
    for (dlist = node->children; dlist; dlist = dlist->next) {
      pref_node_t* child = dlist->data;
      pref_foreach_rec(child, func, data);
    }
  } else {
    func(OBJ(node), data);
  }
}

/** Apply a function to all values.
 * @param func the function to apply.
 * @param data the data to pass to func
 */
void pref_foreach(ObjectFunc func, void* data) {
  g_assert(PrefModule);
  if (PrefModule->prefs && func)
    pref_foreach_rec(PrefModule->prefs, func, data);
}

/** Get the full name of preferences node.
 * @param node the node.
 * @return the full name.
 */
const char* pref_node_fullname(pref_node_t* node) {
  char* result = l_strdup(OBJ(node)->name);
  while (node->parent) {
    node = node->parent;
    result = l_printf("%s/%s", OBJ(node)->name, result);
  }
  return result;
}

#ifdef WIN32
# define _PREFERENCES_C_
# include "lcorealiasdef.c"
#endif
