#ifndef _UI_PREFERENCES_H
#define _UI_PREFERENCES_H

#include <QtGui>
#include "lcore/preferences.h"
#include "lcore/mtypes.h"
#include "lcore/library.h"
#include "lcore/chat.h"
#include "lcore/plugin.h"
#include "lcore/string_list.h"
#include "lcore/l_base.h"
#include "lcore/l_i18n.h"
#include "lcore/ui.h"
#include "LibraryPref.hpp"
#include "ui_utils.hpp"
#include "ui_main.hpp"
#include "ui_style.hpp"
#include "ui_support.hpp"
#include "widgets/PreferencesDialog.h"


typedef void (*ApplyFunc)(char *pref, char* value);

class QT4_UI_EXPORT PrefUpdater : public QObject {
public:
    PrefUpdater(char* preference) 
        : m_preference(preference) {}
    virtual void apply_pref() = 0;
    virtual void cancel_pref() = 0;
protected:
    char* m_preference;
};

class QT4_UI_EXPORT GroupCheckPrefUpdater : public PrefUpdater
{
public:
    GroupCheckPrefUpdater(char* preference, QGroupBox *box) :
        PrefUpdater(preference), m_box(box) { 
            m_box->setChecked(pref_get_ref(m_preference, gboolean));
        }
    void apply_pref() {
        pref_store(m_preference, m_box->isChecked() );
        }
    void cancel_pref() {
        m_box->setChecked(pref_get_ref(m_preference, gboolean));
        }
private:
    QGroupBox *m_box;
};

class QT4_UI_EXPORT CheckBoxPrefUpdater : public PrefUpdater
{
public:
    CheckBoxPrefUpdater(char* preference, QCheckBox *checkbox) :
        PrefUpdater(preference), m_checkbox(checkbox) { 
            m_checkbox->setChecked(pref_get_ref(m_preference, gboolean));
        }
    void apply_pref() {
        pref_store(m_preference, m_checkbox->isChecked() );
        }
    void cancel_pref() {
        m_checkbox->setChecked(pref_get_ref(m_preference, gboolean));
        }
private:
    QCheckBox *m_checkbox;
};

class QT4_UI_EXPORT LineEditPrefUpdater : public PrefUpdater
{
public:
    LineEditPrefUpdater(char* preference, QLineEdit *line) :
        PrefUpdater(preference), m_line(line) {
            if (preference)
            {
              pref_node_t* node = pref_search(preference, TRUE);
              if (!node) { return;}
              const char* text = node->xml_print?node->xml_print(node->value):NULL;
              if (text)
              {
                line->setText( qfu("%1").arg(text));
              }
            }
        }
    void apply_pref() {
            if (m_preference)
            {
                pref_store_str(m_preference, qtu(m_line->text()));
            }
        }
    void cancel_pref() {
            if (m_preference)
            {
                pref_node_t* node = pref_search(m_preference, TRUE);
                m_line->setText( qfu("%1").arg(node->xml_print(node->value))) ;
            }
        }
protected:
     void set_text(QString text) {
        m_line->setText( text) ;    
     }
    QLineEdit *m_line;
};

class QT4_UI_EXPORT FolderLinePrefUpdater : public LineEditPrefUpdater
{
    Q_OBJECT
public:
    FolderLinePrefUpdater(char* preference, QLineEdit *line, QToolButton *choose) :
        LineEditPrefUpdater(preference,line), m_choose(choose) {
            connect(choose,SIGNAL(clicked()), SLOT(update_folder()));
        }
private slots:
    void update_folder() {
        QString folder = QFileDialog::getOpenFileName(m_choose,
                                    qtr("Choose folder"));
        if (!folder.isEmpty())
        {
            set_text(folder);
        }
        
    }
private:
    QToolButton *m_choose;
};

class QT4_UI_EXPORT MediaLinePrefUpdater : public FolderLinePrefUpdater
{
public:
    MediaLinePrefUpdater(char* preference, QLineEdit *line, QToolButton *choose, int media) :
        FolderLinePrefUpdater(preference,line,choose) {     
            mtype = mtype_find(media);
            if (mtype)
            {
                set_text(qfu("%1").arg(mtype->download));
            }
        }
    void apply_pref() {
            mtype_free(mtype);
            if (!m_line->text().isEmpty())
            {
                mtype->download = qtu(m_line->text());
            }
            mtype_save();            
        }
    void cancel_pref() {
            if (mtype)
            {
                set_text(qfu("%1").arg(mtype->download));
            }
        }

private:
    mtype_t* mtype;
};

/*class QT4_UI_EXPORT UiPrefUpdater : public PrefUpdater
{
public:
    UiPrefUpdater(char* preference, QLineEdit *line) :
        PrefUpdater(preference), m_line(line) { }
    void apply_pref() {       
        ui_set_name(qtu(m_line->text()), TRUE);
        }
    void cancel_pref() {
        m_line->setText( qfu("%1").arg(ui_get_list(NULL)?ui_get_list(NULL)->name:"")) ;
        }
private:
    QLineEdit *m_line;
};*/

class QT4_UI_EXPORT SpinBoxPrefUpdater : public PrefUpdater
{
public:
    SpinBoxPrefUpdater(char* preference, QSpinBox *spin) :
        PrefUpdater(preference), m_spin(spin){
            if (preference)
            {
              const int val = pref_get_ref(preference, int);
              if (val)
              {
                m_spin->setValue(val) ;
              }
            }
        }
    void apply_pref() { 
        pref_store(m_preference, m_spin->value());
        }
    void cancel_pref() {
        m_spin->setValue(pref_get_ref(m_preference, int));
        }
private:
    QSpinBox *m_spin;
};

class QT4_UI_EXPORT TextEditPrefUpdater : public PrefUpdater
{
public:
    TextEditPrefUpdater(char* preference, QTextEdit *text) :
        PrefUpdater(preference), m_text(text){
            if (preference)
            {
              pref_node_t* node = pref_search(preference, TRUE);
              if (!node) { return;}
              const char* text = node->xml_print?node->xml_print(node->value):NULL;
              if (text)
              {
                m_text->insertPlainText(qfu("%1").arg(text));
              }
            }
        }
    void apply_pref() { 
            pref_store_str(m_preference, qtu(m_text->toPlainText()));
        }
    void cancel_pref() {
        const char* text = pref_get_ptr(m_preference, char);
        m_text->insertPlainText(QString("%1").arg(text));
        }
private:
    QTextEdit *m_text;
};

class QT4_UI_EXPORT ImageButtonPrefUpdater : public PrefUpdater
{
    Q_OBJECT
public:
    ImageButtonPrefUpdater(char* preference, QToolButton *button) :
        PrefUpdater(preference), m_button(button){
            connect(button, SIGNAL(clicked()), this, SLOT(selectImage()));
            setImage();
        }
    void apply_pref() { 
              pref_store_str(m_preference, qtu(m_image_name));
        }
    void cancel_pref() {
              setImage();
        }
    void setImage() {
            if (m_preference)
            {
              QIcon icon;
              const char* file = pref_get_ptr(m_preference, char);
              if (!file) { 
                  const char* pathname = lopster_pixmap_find("user_icon.png");
                  icon = create_icon(pathname); 
                  m_image_name = qfu("%1").arg(pathname); 
              }
              else {
                  icon = create_icon(file);
                  m_image_name = qfu("%1").arg(file); 
              }
              m_button->setIcon(icon);
              m_button->setIconSize(m_button->size());
            }
        }
    void setImage(QString image) {
              QIcon icon = create_icon(image); 
              m_button->setIcon(icon);
              m_button->setIconSize(m_button->size());
        }
private slots:
    void selectImage() { 
              QString image_name = QFileDialog::getOpenFileName(m_button,
                                    qtr("Open Images"), "",
                                    qtr("Images (*.png *.xpm *.jpg);;"
                                       "All Files (*)"));
              if (!image_name.isEmpty())
              {
                  m_image_name = image_name;
                  setImage(m_image_name);
              }
        }
private:
    QToolButton *m_button;
    QString m_image_name;
};

class QT4_UI_EXPORT ComboPrefUpdater : public PrefUpdater
{
public:
    ComboPrefUpdater(char* preference, QComboBox *combo, QStringList list) :
        PrefUpdater(preference), m_combo(combo) {
            combo->addItems(list);
            if (preference)
            {
              const int val = pref_get_ref(preference, int);
              combo->setCurrentIndex(val);
            }
        }
    void apply_pref() { 
             pref_store(m_preference,  m_combo->currentIndex());
        }
    void cancel_pref() {
        m_combo->setCurrentIndex(pref_get_ref(m_preference, int));
        }
private:
    QComboBox *m_combo;
};

class QT4_UI_EXPORT UiComboPrefUpdater : public PrefUpdater
{
public:
    UiComboPrefUpdater(QComboBox *combo) :
        PrefUpdater(NULL), m_combo(combo) {
            GList* dlist;
            unsigned cnt = 0;
            m_current_index = cnt;
            m_current = ui_get_list(&dlist);
            while (dlist)
            {
                ui_entry_t* entry = (ui_entry_t*) dlist->data;
                m_combo->addItem(qtr(entry->name));
                m_combo->setItemData(cnt, QVariant::fromValue((void*)entry));
                if (entry == m_current) { m_current_index = cnt;}
                dlist = dlist->next;
                cnt++;
            }
            m_combo->setCurrentIndex(m_current_index);
        }
    void apply_pref() { 
        QVariant v = m_combo->itemData(m_combo->currentIndex());
        if ( !v.isValid()) return;
        m_current = (ui_entry_t*) v.value<void *>();
        ui_set_name(m_current->name, TRUE);
        }
    void cancel_pref() {
            m_combo->setCurrentIndex(m_current_index);
        }
private:
    QComboBox *m_combo;
    int m_current_index;
    ui_entry_t* m_current;
};

class QT4_UI_EXPORT DropButtonValue : QObject
{
	Q_OBJECT
public:
    DropButtonValue(int c, QString i, unsigned int f, QObject *parent=0, bool s=false) 
            : QObject(parent), child(c), item(i), flag(f), sep(s)  {}
    int child;
    QString item;
    unsigned int flag;
    bool sep;
};

class QT4_UI_EXPORT DropButtonPrefUpdater : public PrefUpdater
{
    Q_OBJECT
public:
        DropButtonPrefUpdater(char* preference, QPushButton *button, QList <DropButtonValue *> items);
        void apply_pref();
        void cancel_pref();
        QMenu * create_menu(/*QList <DropButtonValue *> items,*/ int no_loop);
private slots:
    void updateFlag(QAction *action);

private:
    void updateMenu( );
    QPushButton *m_button;
    QMenu *m_menu;
    int m_val;
    QList <DropButtonValue *> m_drop_list;
    QHash <QAction *, DropButtonValue *> m_items;
    QList <QAction *> m_actions;
};

class QT4_UI_EXPORT SharedFileTreePrefUpdater : public PrefUpdater
{
public:
    SharedFileTreePrefUpdater(char* preference, LibraryPref *pref_widget) :
        PrefUpdater(preference), m_pref_widget(pref_widget){
            set_library(library_get_current());
        }
    void apply_pref() { 
            m_pref_widget->share_on_library_apply();
        }
    void cancel_pref() {
            m_pref_widget->share_on_library_cancel();
    }
    void set_library(library_t* lib) {
        m_pref_widget->share_init_from_library(lib);
    }
private:
    LibraryPref *m_pref_widget;
};

class QT4_UI_EXPORT CharsetPrefUpdater : public PrefUpdater
{
public:
    CharsetPrefUpdater(char* preference, CharsetWidget *charset) 
                : PrefUpdater(preference), m_charset(charset) {
        set_charset();
    }
    void set_charset() {
        const char* text = pref_get_ptr(m_preference, char);
        m_charset->charset_set_selection(qfu(text));
    }
    void apply_pref() {
        pref_store_str(m_preference, qtu(m_charset->charset_get_selection()));
    }
    void cancel_pref() {
        set_charset();
    }

private:
    CharsetWidget *m_charset;
};

class QT4_UI_EXPORT StringListPrefUpdater : public LineEditPrefUpdater
{
    /*Q_OBJECT */
public:
    StringListPrefUpdater(  char* preference, char* name, 
                            const QString &title, QLineEdit *entry,
                            StringListWidget *widget);
    ~StringListPrefUpdater();
    void apply_pref();
    void cancel_pref();
/*    void create_list();
 private slots:
    void update_strings_list();
    void update_entry(QListWidgetItem * item); */
private:
    StringListWidget *str_widget;
    char* string_list;
    char* tmp_string_list;
};

class QT4_UI_EXPORT StylePrefUpdater : public PrefUpdater {
public:
    StylePrefUpdater(StyleWidget *style_page)   
        : PrefUpdater(NULL) {
        this->m_style_page = style_page;
    }
    void apply_pref() {
        m_style_page->ui_style_apply_changes();
    }
    void cancel_pref() {}
private:
    StyleWidget *m_style_page;
};

typedef QWidget* (*CreatePrefPageFunc)(void);

class PluginPageInfo
{
public:
    PluginPageInfo(int i, QString l, QString d, 
					CreatePrefPageFunc f1) :
            id(i), label(l), desc(d),
			create_func(f1) {}
    int id;
    QString label;
    QString desc;
    CreatePrefPageFunc create_func;
};

class PrefTreeWidgetItem : public QTreeWidgetItem
{
public:
    PrefTreeWidgetItem(QTreeWidget *tree)
        : QTreeWidgetItem(tree) {m_page = 0;} // 0 point to default page
    PrefTreeWidgetItem(QTreeWidgetItem *item)
        : QTreeWidgetItem(item) {m_page = 0;}
    void set_page(int i) { m_page = i;}
    int page() { return m_page;}
private:
    int m_page;
};

class QT4_UI_EXPORT PreferencesDialog : public QDialog, private Ui::PreferencesDialog
{
   Q_OBJECT
public:
    PreferencesDialog(QWidget *parent = 0);
    ~PreferencesDialog();
    void ui_preferences_init();

    void ui_preferences_id_page( );
    void ui_preferences_share_page( );
    void ui_preferences_transfer_page( );

    void ui_preferences_basic_page( );
    void ui_preferences_pathapp_page( );
    void ui_preferences_chatfunc_page( );
    void ui_preferences_chatappear_page( );
    void ui_preferences_style_page();

    static void ui_preferences_apply_cancel_add(PrefUpdater *updater);
    void ui_preferences_init_plugins();
    void ui_preferences_show();

  
    static void ui_preferences_plugin_add_page(int plugin_id, QString label, 
                                            QString desc, 
											CreatePrefPageFunc f1);
    int pageno;
    QTreeWidgetItem *root_item;
    QTreeWidgetItem *plugins_item;
private slots:
    void ui_preferences_apply_all();
    void ui_preferences_cancel_all();
    void ui_preferences_ok_all();
    void ui_preferences_display_page(QTreeWidgetItem *);
private:
    bool init_prefs;
};

#endif //_UI_PREFERENCES_H
