/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.24 $
 * $Date: 2007/11/21 20:18:24 $
 * $Author: sgop $
 */

#include <ctype.h>
#include <string.h>

#include <libxml/parser.h>

#include "mtypes.h"
#include "preferences.h"
#include "plugin.h"
#include "files.h"
#include "xml_handler.h"
#include "ui.h"
#include "action_item.h"
#include "l_i18n.h"
#include "l_base.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
  GList* pattern;
  GList* searches;
  GList* auto_searches;
  int pnumber;
  GList* fast_funcs;
  GList* slow_funcs;
  GList* actions;
} search_module_t;

static search_module_t* SearchModule = NULL;


#define MASTER(arg) ((master_t*)(arg))

typedef struct {
  user_t _user;
  int weight;
} master_t;

typedef struct {
  search_t* search;
  ObjectFunc func;
  void* data;
  int timer;
  int minutes;
} search_timer_t;


static void search_pattern_save(void) {
  lopster_xml_save("core:patterns", 1000);
}

static void search_pattern_load(void) {
  attr_allow_unregistered(TRUE);
  lopster_xml_load("core:patterns");
  attr_allow_unregistered(FALSE);
}

static gboolean on_research(search_timer_t* t) {
  t->func(OBJ(t->search), t->data);
  return TRUE;
}
static void on_auto_remove(search_t* search, search_timer_t* t) {
  g_assert(search == t->search);
  SearchModule->auto_searches = g_list_remove(SearchModule->auto_searches, t);
  if (t->timer) lopster_timeout_remove(t->timer);
  g_free(t);
}

/** Activate autosearch. A plugin has to call this function once to
 * activate autosearch.
 * @param search the search
 * @param minutes the autosearch interval in minutes
 * @param func the function to call when the research should be done.
 * @param data user data for func
 */
void search_set_continuous(search_t* search, int minutes, 
                           ObjectFunc func, void* data)
{
    search_timer_t* t = g_malloc(sizeof(*t));
    t->search = search;
    t->func = func;
    t->data = data;
    t->minutes = minutes;
    t->timer = lopster_timeout_add(minutes*60*1000, (GSourceFunc)on_research, t);

    SearchModule->auto_searches = g_list_append(SearchModule->auto_searches, t);
    object_notify_d(OBJ(search), (ObjectFunc)on_auto_remove, t);
}

/** Force research. As search_set_continuous() must be called only
 * once per search, this function usually is used when a plugin joins
 * a new server and likes to immediately search for sources on that
 * network.
 * @param func the function previously passed to search_set_continuous()
 * @param data user data which is passed to that function.
 */
void search_foreach_continuous(ObjectFunc func, void* data) {
  GList* dlist;
  g_assert(SearchModule);
  for (dlist = SearchModule->auto_searches; dlist; dlist = dlist->next) {
    search_timer_t* t = dlist->data;
    if (t->func == func) {
      func(OBJ(t->search), data);
      g_assert(t->timer);
      // reset the timer.
      lopster_timeout_remove(t->timer);
      t->timer = lopster_timeout_add(t->minutes*60*1000, (GSourceFunc)on_research, t);
    }
  }
}

/** Register a search handler. Call this function if your network
 * supports searching.
 * @param action function to call when a search is done.
 */
void search_register_action(SearchFunc action) {
  g_assert(SearchModule);
  if (action) {
    SearchModule->actions = g_list_append(SearchModule->actions, action);
  }
}

/** Register a compare function for search results. This function
 * specifies how search results are grouped in a search page.
 * @param compare the function which compares two search_result_t
 * @param slow TRUE if the compare function is relatively slow.
 * @note only file attributes that were introduced by the plugin
 * should be compared.
 */
void search_register_result_compare(GCompareFunc compare, gboolean slow) {
  g_assert(SearchModule);

  if (!compare) return;
  if (slow) SearchModule->slow_funcs = g_list_append(SearchModule->slow_funcs, compare);
  else SearchModule->fast_funcs = g_list_append(SearchModule->fast_funcs, compare);
}

#define WORD_CHAR(c) \
        (isalnum((unsigned char)(c))||(c)=='\''||(c)=='@'||(unsigned char)(c) > 128)

/** Convert a string into search terms and excluded search terms.
 * @param term the string to parse
 * @param[out] include list of include terms, must be initialized
 * @param[out] exclude list of exclude terms, must be initialized
 * @return the number of include terms
 */
unsigned search_string_tokenize(char* term, GSList** include, GSList** exclude) {
  char* pos;
  gboolean ex;
  unsigned res = 0;

  while (*term) {
    ex = FALSE;
    while (*term && !WORD_CHAR(*term)) {
      if (exclude) {
        if (*term == '-') ex = TRUE;
        else if (ex) ex = FALSE;
      }
      term++;
    }

    pos = term;
    while (WORD_CHAR(*pos)) pos++;
    if (*pos) *pos++ = 0;

    if (term[0] && term[1]) {
      if (ex)
        *exclude = g_slist_append(*exclude, g_ascii_strdown(term, -1));
      else {
        *include = g_slist_append(*include, g_ascii_strdown(term, -1));
        res++;
      }
    }
    term = pos;
  }
  return res;
}

/** Tokenize a filename.
 * @param file the file which name should be tokenized.
 * @param[out] tokens the resulting tokes.
 * @return the number if tokens
 */
unsigned search_file_tokenize(file_t* file, GSList** tokens) {
  char* sname = l_strdup(file->filename);
  //  char* pos = strrchr(sname, '.');
  //  if (pos) *pos = 0;
  *tokens = NULL;
  return search_string_tokenize(sname, tokens, NULL);
}

static gboolean _check_size(search_pattern_t* pattern, file_t* file) {
  attribute_t* attr = attr_find(ATTRC(pattern), "pattern", "Size");
  file_size_t min, max;

  if (!attr) return TRUE;
  min = attr_get_refi(attr, file_size_t, 0);
  max = attr_get_refi(attr, file_size_t, 1);
  if (min && file->size < min) return FALSE;
  if (max && file->size > max) return FALSE;
  return TRUE;
}

static gboolean _check_bitrate(search_pattern_t* pattern, file_t* file) {
  attribute_t* attr;
  attribute_t* fattr;
  int min, max, val;

  attr = attr_find(ATTRC(pattern), "pattern", "Bitrate");
  if (!attr) return TRUE;

  fattr = attr_find(ATTRC(file), "file", "Bitrate");
  if (!fattr) return TRUE;

  min = attr_get_refi(attr, int, 0);
  max = attr_get_refi(attr, int, 1);
  val = attr_get_ref(fattr, int);
  if (min && val < min) return FALSE;
  if (max && val > max) return FALSE;
  return TRUE;
}

static gboolean _check_frequency(search_pattern_t* pattern, file_t* file) {
  attribute_t* attr;
  attribute_t* fattr;
  int min, max, val;

  attr = attr_find(ATTRC(pattern), "pattern", "Frequency");
  if (!attr) return TRUE;

  fattr = attr_find(ATTRC(file), "file", "Frequency");
  if (!fattr) return TRUE;

  min = attr_get_refi(attr, int, 0);
  max = attr_get_refi(attr, int, 1);
  val = attr_get_ref(fattr, int);
  if (min && val < min) return FALSE;
  if (max && val > max) return FALSE;
  return TRUE;
}

static gboolean _check_duration(search_pattern_t* pattern, file_t* file) {
  attribute_t* attr;
  attribute_t* fattr;
  int min, max, val;

  attr = attr_find(ATTRC(pattern), "pattern", "Duration");
  if (!attr) return TRUE;

  fattr = attr_find(ATTRC(file), "file", "Duration");
  if (!fattr) return TRUE;

  min = attr_get_refi(attr, int, 0);
  max = attr_get_refi(attr, int, 1);
  val = attr_get_ref(fattr, int);
  if (min && val < min) return FALSE;
  if (max && val > max) return FALSE;
  return TRUE;
}

/** Check a file whether it matches a pattern.
 * @param pattern the search pattern.
 * @param file the file to check
 * @param check_inex TRUE to also check include/exclude terms
 * @return TRUE if file matches, FALSE otherwise
 */
gboolean search_pattern_check_file(search_pattern_t* pattern, file_t* file, gboolean check_inex) {
  g_assert(SearchModule);
  
  if (pattern->media != MEDIA_ANY && file->media != pattern->media)
    return FALSE;

  if (!_check_size(pattern, file)) return FALSE;
  if (!_check_bitrate(pattern, file)) return FALSE;
  if (!_check_frequency(pattern, file)) return FALSE;
  if (!_check_duration(pattern, file)) return FALSE;
  
  if (check_inex) {
    GSList* slist;
    for (slist = pattern->include; slist; slist = slist->next) {
      char* token = slist->data;
      if (!strcasestr(file->localname, token)) return FALSE;
    }
    for (slist = pattern->exclude; slist; slist = slist->next) {
      char* token = slist->data;
      if (strcasestr(file->localname, token)) return FALSE;
    }
  }

  return TRUE;
}

static const char* master_print(master_t* master) {
  return l_printf(ngettext("%d source", "%d sources", master->weight), master->weight);
}

static void master_init(master_t* master) {
  user_init(USER(master), "MasterUser", -2);
  USER(master)->print = (UserPrint)master_print;
  master->weight = 0;
}

static master_t* master_new(void) {
  master_t* master = g_malloc0(sizeof(*master));
  master_init(master);
  return master;
}



/** Copy a search pattern. The destination pattern must be freshly
 * initialized.
 * @dest the destination pattern
 * @source the source pattern
 */
void search_pattern_copy(search_pattern_t* dest, search_pattern_t* source) {
  GSList* dlist;

  attr_container_copy(ATTRC(dest), ATTRC(source), NULL);

  g_slist_foreach(dest->include, (GFunc)g_free, NULL);
  g_slist_free(dest->include);
  dest->include = NULL;
  for (dlist = source->include; dlist; dlist = dlist->next) {
    char* token = dlist->data;
    dest->include = g_slist_append(dest->include, g_strdup(token));
  }
  g_slist_foreach(dest->exclude, (GFunc)g_free, NULL);
  g_slist_free(dest->exclude);
  dest->exclude = NULL;
  for (dlist = source->exclude; dlist; dlist = dlist->next) {
    char* token = dlist->data;
    dest->exclude = g_slist_append(dest->exclude, g_strdup(token));
  }
  dest->media = source->media;
}

// used by CORE only
GList* search_get_searches() {
  g_assert(SearchModule);
  return SearchModule->searches;
}

static void search_pattern_destroy(search_pattern_t* pattern) {
  if (!pattern) return;
  g_slist_foreach(pattern->include, (GFunc)g_free, NULL);
  g_slist_free(pattern->include);
  pattern->include = NULL;
  g_slist_foreach(pattern->exclude, (GFunc)g_free, NULL);
  g_slist_free(pattern->exclude);
  pattern->exclude = NULL;
  attr_container_destroy(ATTRC(pattern));
}

// used by UI only
void search_pattern_remove(search_pattern_t* pattern) {
  GList* dlist;

  g_assert(SearchModule);

  dlist = g_list_find(SearchModule->pattern, pattern);
  if (!dlist) return;
  SearchModule->pattern = g_list_delete_link(SearchModule->pattern, dlist);
  object_kill(OBJ(pattern));

  search_pattern_save();
}

static void search_pattern_init(search_pattern_t* pattern) {
  g_assert(SearchModule);

  attr_container_init(ATTRC(pattern), NULL);
  object_add_type(OBJ(pattern), "SearchPattern");
  OBJ(pattern)->destroy = (ObjectFunc)search_pattern_destroy;
  pattern->include = NULL;
  pattern->exclude = NULL;
  pattern->media = MEDIA_AUDIO;
}

/** Create a new search pattern.
 * @return the new pattern
 */
search_pattern_t* search_pattern_new()
{
    search_pattern_t* pattern = g_malloc0(sizeof(*pattern));
    search_pattern_init(pattern);
    return pattern;
}

/*
static search_pattern_t* 
search_pattern_clone(search_pattern_t* pattern) {
  search_pattern_t* result;
  
  result = search_pattern_new();
  search_pattern_copy(result, pattern);
  return result;
}
*/

static void xml_save_pattern(search_pattern_t* pattern, xmlNodePtr xroot) {
  xmlNodePtr xsub = xmlNewChild(xroot, NULL, "Pattern", NULL);
  xmlNewProp(xsub, "Name", OBJ(pattern)->name);
  xmlNewProp(xsub, "Media", int2media(pattern->media, 0));
  if (pattern->include) {
    char* str = make_string_from_list(pattern->include, " ");
    xmlNewTextChild(xsub, NULL, "SearchFor", str);
    g_free(str);
  }
  if (pattern->exclude) {
    char* str = make_string_from_list(pattern->exclude, " ");
    xmlNewTextChild(xsub, NULL, "SearchNot", str);
    g_free(str);
  }
  
  ATTR_XML_PRINT_ALL(ATTRC(pattern), xsub);
}
static void xml_save_patterns(xmlDocPtr xfile, xmlNodePtr xroot) {
  (void)xfile;
  g_list_foreach(SearchModule->pattern, (GFunc)xml_save_pattern, xroot);
}

static int search_pattern_compare(search_pattern_t* p1, search_pattern_t* p2) {
  int result;
  result = attr_container_compare(ATTRC(p1), ATTRC(p2));
  if (result) return result;
  if (p1->media != p2->media) return 1;
  return 0;
}

// used by UI only
search_pattern_t* search_pattern_find(search_pattern_t* pattern) {
  GList* dlist;

  if (!pattern) return NULL;
  for (dlist = SearchModule->pattern; dlist; dlist = dlist->next) {
    search_pattern_t* p2 = dlist->data;
    if (p2 == pattern || !search_pattern_compare(p2, pattern))
      return p2;
  }
  return NULL;
}

// used by UI only
void search_pattern_add(search_pattern_t* pattern) {
  g_assert(SearchModule);
  g_assert(pattern);
  SearchModule->pattern = g_list_append(SearchModule->pattern, pattern);
  object_make_public(OBJ(pattern), NULL);
  search_pattern_save();
}

static search_pattern_t* xml_get_pattern(xmlDocPtr xdoc, xmlNodePtr xnode) {
  search_pattern_t* pattern;
  xmlChar* attr;
  xmlChar* attr2;
  xmlNodePtr xsub;

  pattern = search_pattern_new();

  attr = xmlGetProp(xnode, "Name");
  if (attr) {
    object_set_name(OBJ(pattern), attr);
    xmlFree(attr);
  }
  attr = xmlGetProp(xnode, "Media");
  if (attr) {
    pattern->media = media2int(attr);
    xmlFree(attr);
  }

  xsub = xnode->xmlChildrenNode;
  while (xsub != NULL) {
    if (!strcmp(xsub->name, "Attr")) {
      attr = xmlGetProp(xsub, "Name");
      if (attr) {
        attr2 = xmlNodeListGetString(xdoc, xsub->xmlChildrenNode, 1);
        attr_store_str(ATTRC(pattern), "pattern", attr, attr2);
        xmlFree(attr);
        xmlFree(attr2);
      }
    } else if (!strcmp(xsub->name, "SearchFor")) {
      attr = xmlNodeListGetString(xdoc, xsub->xmlChildrenNode, 1);
      if (attr && *attr) {
        make_list_from_string(&pattern->include, attr, " ");
        xmlFree(attr);
      }
    } else if (!strcmp(xsub->name, "SearchNot")) {
      attr = xmlNodeListGetString(xdoc, xsub->xmlChildrenNode, 1);
      if (attr && *attr) {
        make_list_from_string(&pattern->exclude, attr, " ");
        xmlFree(attr);
      }
    }
    xsub = xsub->next;
  }
  return pattern;
}

static void xml_load_patterns(xmlDocPtr xfile, xmlNodePtr xroot) {
  xmlNodePtr xsub = xroot->xmlChildrenNode;
  while (xsub != NULL) {
    if (!strcmp(xsub->name, "Pattern")) {
      search_pattern_t* pattern = xml_get_pattern(xfile, xsub);
      if (pattern) {
        SearchModule->pattern = g_list_append(SearchModule->pattern, pattern);
        object_make_public(OBJ(pattern), NULL);
      }
    }
    xsub = xsub->next;
  }
}

static int search_result_compare_size(file_t* file1, file_t* file2) {
  if (file1->size > file2->size) return -1;
  return (file1->size < file2->size);
}

// download from file
static void on_download_from_file(GList* files) {
  GList* dlist;
  for (dlist = files; dlist; dlist = dlist->next) {
    file_t* file = dlist->data;
    download_from_file(file);
  }
}

// download from source
static void on_download_from_result(GList* results) {
  GList* dlist;
  for (dlist = results; dlist; dlist = dlist->next) {
    search_result_t* result = dlist->data;
    download_t* download = download_from_file(FILE(result));
    if (!download) continue;
    download_add_source(download, FILE(result), result->user);
  }
}
static gboolean on_check_source(search_result_t* result) {
  return (result && result->user && source_factory_get(result->user->ID));
}

// resume from source.
static int TempNumber = 0;
static void on_resume_sources(download_t* download, GList* results) {
  GList* dlist;
  for (dlist = results; dlist; dlist = dlist->next) {
    search_result_t* result = dlist->data;
    if (download_matches_source(download, FILE(result), result->user))
      download_add_source(download, FILE(result), result->user);
  }
}
static void on_resume_from_result(GList* results) {
  downloads_foreach((ObjectFunc)on_resume_sources, results);
}
static void on_check_single_source(download_t* download, search_result_t* result) {
  if (download_matches_source(download, FILE(result), result->user))
    TempNumber++;
}
static gboolean on_check_source2(search_result_t* result) {
  TempNumber = 0;
  downloads_foreach((ObjectFunc)on_check_single_source, result);
  return (TempNumber > 0);
}

static void on_search_remove_all(void) {
  GList* dlist = SearchModule->searches;
  while (dlist) {
    search_t* search = dlist->data;
    dlist = dlist->next;
    if (!search->download) object_kill(OBJ(search));
  }
}

/*
static void on_pattern_search(search_pattern_t* pattern) {
  search_new_from_pattern(pattern, -1);
}
*/

static action_item_t* on_get_protocol_user_menu(GList* selection) {
  GList* dlist;
  action_item_t* result = NULL;

  for (dlist = selection; dlist; dlist = dlist->next) {
    search_result_t* res = dlist->data;
    action_item_t* item = user_action_result_find(res->user->ID);
    if (!item) return NULL;
    if (!result) result = item;
    else if (result != item) return NULL;
  }
  return result;
}

static void search_actions_init(void) {
  action_item_t* item;
  
  item = action_item_root("Search result actions");
  action_item_simple(N_("Download"), item, 0,
                     (ActionItemFunc)on_download_from_file);
/*   action_item_simple(N_("Download as..."), item, 0, */
/*                      (ActionItemFunc)on_download_folder); */
  action_item_new(N_("Download from source"), item, 0,
                  (ActionItemFunc)on_download_from_result,
                  (ActionItemCheck)on_check_source);
  action_item_sep(item);
  action_item_new(N_("Resume from source"), item, 0,
                  (ActionItemFunc)on_resume_from_result,
                  (ActionItemCheck)on_check_source2);
  action_item_sep(item);
  action_item_sub(N_("User menu"), item,
                  (ActionItemLink)on_get_protocol_user_menu,
                  FALSE);
  /* 
     action_item_simple("Download from source as...", item, 0,
     (ActionItemFunc)NULL);
  */
  action_item_sep(item);

  action_item_simple(N_("Remove all searches"), item, AI_GLOBAL,
                     (ActionItemFunc)on_search_remove_all);
  
}

static void search_destroy(search_t* search) {
  if (!search) return;
  SearchModule->searches = g_list_remove(SearchModule->searches, search);
  storage_destroy(search->results);
  g_free(search->sources);
  search_pattern_destroy(SEARCHP(search));
}

static void search_init(search_t* search, int network)
{
    GList* dlist;
    
    search_pattern_init(SEARCHP(search));
    object_add_type(OBJ(search), "Search");
    OBJ(search)->destroy = (ObjectFunc)search_destroy;
    search->network = network;
    search->download = NULL;
    search->results = storage_new((GDestroyNotify)object_kill);

    for (dlist = SearchModule->fast_funcs; dlist; dlist = dlist->next)
    {
        GCompareFunc func = dlist->data;
        storage_add_compare_func(search->results, func, FALSE);
    }
    for (dlist = SearchModule->slow_funcs; dlist; dlist = dlist->next)
    {
        GCompareFunc func = dlist->data;
        storage_add_compare_func(search->results, func, TRUE);
    }
    
    if (SearchModule->pnumber < 0) SearchModule->pnumber = plugin_get_number();
    search->sources = g_malloc0(sizeof(int)*(SearchModule->pnumber+2));
}

static search_t* search_new(int network)
{
    search_t* search = g_malloc0(sizeof(*search));
    search_init(search, network);
    
    SearchModule->searches = g_list_append(SearchModule->searches, search);
    return search;
}

static void search_queue(search_t* search)
{
    GList* dlist;
    for (dlist = SearchModule->actions; dlist; dlist = dlist->next)
    {
        SearchFunc func = dlist->data;
        func(search);
    }
}

/** Create a new search using the given pattern.
 * @param pattern patter to search for.
 * @param network unused
 * @return a new search
 * @note the search is automatically started.
 */
search_t* search_new_from_pattern(search_pattern_t* pattern, int network)
{
    search_t* search = search_new(network);
    search_pattern_copy(SEARCHP(search), pattern);
    
    object_make_public(OBJ(search), NULL);
    
    search_queue(search);
    
    return search;
}

// used by core only
search_t* search_new_from_download(download_t* download) {
  if (!download->search) {
    search_t* search = download->search = search_new(-1);
    search_pattern_t* pattern = SEARCHP(search);

    search->download = download;
    object_set_name(OBJ(pattern), "Resume search");
    pattern->media = FILE(download)->media;
    search->network = -1;
  }
  
  search_queue(download->search);
  
  return download->search;
}

// used by core only
void search_refresh(search_t* search) {
  if (!search) return;
  storage_clear(search->results);
  search_queue(search);
}

static void
search_result_set_master(search_result_t* result, search_result_t* master) {
  g_assert(result->master == NULL);
  master->children = g_list_append(master->children, result);
  result->master = master;
  master->weight += result->weight;
  ((master_t*)master->user)->weight += result->weight;
  master->available |= result->available;
}

static void search_result_destroy(search_result_t* result) {
  g_list_foreach(result->children, (GFunc)object_kill, NULL);
  g_list_free(result->children);
  result->children = NULL;

  user_unref(result->user);
  file_destroy(FILE(result));
}

static void search_result_init(search_result_t* sr, file_t* file,
                               user_t* user, int weight) {
  file_init(FILE(sr));
  file_copy(FILE(sr), file, "file");
  object_add_type(OBJ(sr), "SearchResult");
  OBJ(sr)->destroy = (ObjectFunc)search_result_destroy;
  sr->children = NULL;
  sr->master = NULL;
  sr->user = user_ref(user);
  sr->weight = weight;
  sr->available = FALSE;
}

static search_result_t* search_result_new(file_t* file, user_t* user, int weight) {
  search_result_t* result = g_malloc0(sizeof(*result));
  search_result_init(result, file, user, weight);
  return result;
}

static void search_result_make_master(search_t* search, search_result_t* master)
{
    search_result_t* result;
    
    // make new child from master
    // and make master a real master.
    result = search_result_new(FILE(master), master->user, master->weight);
    result->children = master->children;
    result->available = master->available;
    master->children = NULL;
    
    user_unref(master->user);
    master->user = user_ref(USER(master_new()));
    
    // add new child to the master
    search_result_set_master(result, master);
    
    // show new child
    object_make_public(OBJ(result), search);
}

/** Add a new result to a search.
 * @param search the search
 * @param file the new file.
 * @param user the user providing this file.
 * @param weight the number of sources for this file, often == 1
 * @param available TRUE if file is available
 */
void search_add_result(search_t* search, file_t* file,
                       user_t* user, int weight, gboolean available) {
  search_result_t* master;
  search_result_t* result;
  
  g_assert(SearchModule);
  g_assert(search && file);
  g_assert(!search->download);
  
  // search master for this file
  master = storage_find(search->results, file);
  if (!master) {
    // if not found, insert new entry
    result = search_result_new(file, user, weight);
    result->available = available;
    storage_insert(search->results, result);
  } else {
    // if found, check whether it is a real master.
    g_assert(!master->master);
    if (!master->children) {
      // make it a real master, but check for same user
      if (master->user->ip && user->ip) {
        if (master->user->ip == user->ip) return;
      } else {
        if (!strcmp(OBJ(master->user)->name, OBJ(user)->name)) return;
      }
      search_result_make_master(search, master);
    } else {
      // got a real master, check children for same user
      GList* dlist;
      for (dlist = master->children; dlist; dlist = dlist->next) {
        result = dlist->data;
        if (result->user->ip && user->ip) {
          if (result->user->ip == user->ip) return;
        } else {
          if (!strcmp(OBJ(result->user)->name, OBJ(user)->name)) return;
        }
      }
    }
    g_assert(master->children);
    // we finally have a new child for this master node
    result = search_result_new(file, user, weight);
    result->available = available;
    attr_container_adapt(ATTRC(master), ATTRC(result), "file");
    search_result_set_master(result, master);
    // update master
    object_make_public(OBJ(master), search);
  }
  
  search->sources[0] += weight;
  if (result->user->ID < 0)
    search->sources[SearchModule->pnumber+1] += weight;
  else
    search->sources[result->user->ID+1] += weight;

  // show new result
  object_make_public(OBJ(result), search);
}

// used by CORE
int search_get_number_results(search_t* search) {
  return storage_count(search->results);
}

// used by CORE
void search_foreach_result(search_t* search, StorageTraverseFunc func, gpointer data) {
  storage_foreach(search->results, func, data);
}

// used by CORE
void search_patterns_foreach(ObjectFunc func, void* data) {
  GList* dlist;
  g_assert(func);
  g_assert(SearchModule);
  for (dlist = SearchModule->pattern; dlist; dlist = dlist->next) {
    search_pattern_t* pattern = dlist->data;
    func(OBJ(pattern), data);
  }
}


const char* print_filesize_minmax_ui(file_size_t* bytes) {
  const char* str1 = "";
  const char* str2 = "";

  if (bytes[0] == 0 && bytes[1] == 0) return "";

  if (bytes[0]) str1 = print_bytes_exact(bytes);
  if (bytes[1]) str2 = print_bytes_exact(bytes+1);

  if (bytes[0] == bytes[1]) {
    return l_printf(_("exactly %s"), str1);
  } else if (*str1 && *str2) {
    return l_printf(_("%s to %s"), str1, str2);
  } else if (*str1) {
    return l_printf(_("at least %s"), str1);
  } else if (*str2) {
    return l_printf(_("at most %s"), str2);
  } else {
    g_assert(FALSE);
    return "";
  }
}
const char* print_bitrate_minmax_ui(int* value) {
  const char* str1 = "";
  const char* str2 = "";

  if (value[0] == 0 && value[1] == 0) return "";

  if (value[0]) str1 = print_bitrate(value);
  if (value[1]) str2 = print_bitrate(value+1);

  if (value[0] == value[1]) {
    return l_printf(_("exactly %s"), str1);
  } else if (*str1 && *str2) {
    return l_printf(_("%s to %s"), str1, str2);
  } else if (*str1) {
    return l_printf(_("at least %s"), str1);
  } else if (*str2) {
    return l_printf(_("at most %s"), str2);
  } else {
    g_assert(FALSE);
    return "";
  }
}
const char* print_frequency_minmax_ui(int* value) {
  const char* str1 = "";
  const char* str2 = "";

  if (value[0] == 0 && value[1] == 0) return "";

  if (value[0]) str1 = print_frequency(value);
  if (value[1]) str2 = print_frequency(value+1);

  if (value[0] == value[1]) {
    return l_printf(_("exactly %s"), str1);
  } else if (*str1 && *str2) {
    return l_printf(_("%s to %s"), str1, str2);
  } else if (*str1) {
    return l_printf(_("at least %s"), str1);
  } else if (*str2) {
    return l_printf(_("at most %s"), str2);
  } else {
    g_assert(FALSE);
    return "";
  }
}
const char* print_duration_minmax_ui(int* value) {
  const char* str1 = "";
  const char* str2 = "";

  if (value[0] == 0 && value[1] == 0) return "";

  if (value[0]) str1 = print_duration(value);
  if (value[1]) str2 = print_duration(value+1);

  if (value[0] == value[1]) {
    return l_printf(_("exactly %s"), str1);
  } else if (*str1 && *str2) {
    return l_printf(_("%s to %s"), str1, str2);
  } else if (*str1) {
    return l_printf(_("at least %s"), str1);
  } else if (*str2) {
    return l_printf(_("at most %s"), str2);
  } else {
    g_assert(FALSE);
    return "";
  }
}

static void searches_init() {
  attr_register("pattern", N_("Size"), ATTR_IMPORTANT,
                (UIPrintFunc)print_filesize_minmax_ui,
                (XMLPrintFunc)print_filesize_minmax,
                (CompareFunc)compare_filesize_minmax,
                (CopyFunc)copy_filesize_minmax,
                (FreeFunc)NULL,
                (StoreStrFunc)store_str_filesize_minmax,
                (StoreFunc)store_filesize_minmax);
  attr_register("pattern", N_("Bitrate"), ATTR_IMPORTANT,
                (UIPrintFunc)print_bitrate_minmax_ui,
                (XMLPrintFunc)print_int_minmax,
                (CompareFunc)compare_int_minmax,
                (CopyFunc)copy_int_minmax,
                (FreeFunc)NULL,
                (StoreStrFunc)store_str_int_minmax,
                (StoreFunc)store_int_minmax);
  attr_register("pattern", N_("Frequency"), ATTR_IMPORTANT,
                (UIPrintFunc)print_frequency_minmax_ui,
                (XMLPrintFunc)print_int_minmax,
                (CompareFunc)compare_int_minmax,
                (CopyFunc)copy_int_minmax,
                (FreeFunc)NULL,
                (StoreStrFunc)store_str_int_minmax,
                (StoreFunc)store_int_minmax);
  attr_register("pattern", N_("Duration"), ATTR_IMPORTANT,
                (UIPrintFunc)print_duration_minmax_ui,
                (XMLPrintFunc)print_int_minmax,
                (CompareFunc)compare_int_minmax,
                (CopyFunc)copy_int_minmax,
                (FreeFunc)NULL,
                (StoreStrFunc)store_str_int_minmax,
                (StoreFunc)store_int_minmax);
  
  search_register_result_compare((GCompareFunc)search_result_compare_size, FALSE);


  search_actions_init();
}

static void searches_run() {
  lopster_xml_register_home("core:patterns", "search_pattern.xml",
                            "SearchPattern", xml_load_patterns, xml_save_patterns);
  // load the pattern here cause plugins might register attributes
  search_pattern_load();
}
static void searches_exit() {
  g_list_free(SearchModule->fast_funcs);
  g_list_free(SearchModule->slow_funcs);
  g_list_free(SearchModule->actions);
  //FIXME: free
  //  GList* pattern;
  //  GList* searches;
  //  GList* auto_searches;
}

static void search_module_init(search_module_t* module) {
  module_init(MODULE(module), "Search Module",
              (ModuleFunc)searches_init,
              (ModuleFunc)searches_run,
              (ModuleFunc)searches_exit);

  module->pattern = NULL;
  module->searches = NULL;
  module->pnumber = -1;
  module->fast_funcs = NULL;
  module->slow_funcs = NULL;
  module->actions = NULL;

  module_require(pref_module_instance);
  module_require(xml_module_instance);

}

/** Require this module if you support searches
 * @return search module
 */
module_t* search_module_instance() {
  if (!SearchModule) {
    SearchModule = g_malloc0(sizeof(*SearchModule));
    search_module_init(SearchModule);
  }
  return MODULE(SearchModule);
}

#ifdef WIN32
#define _SEARCH_C_
#include "lcorealiasdef.c"
#endif
