/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.8 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include <string.h>

#include <libxml/parser.h>
#include <glib.h>

#include "xml_handler.h"
#include "utils.h"
#include "string_list.h"
#ifdef WIN32
#include "win32_defs.h"
#endif

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  StringListFunc added;
  StringListFunc removed;
  void* data;
} sl_func_t;

typedef struct {
  module_t _module;
  GList* lists;
  gboolean save_unregistered;
} string_list_module_t;

typedef struct _string_list_t {
  GList* items;
  char* name;
  char* label;
  gboolean save;
  gboolean allow_ws;
  gboolean case_sensitive;
  GList* funcs;
  gboolean registered;
} string_list_t;

static string_list_module_t* StringListModule = NULL;


static void string_list_load(void) {
  lopster_xml_load("core:string_lists");
}

static void string_list_save(void) {
  lopster_xml_save("core:string_lists", 10000);
}


static void string_list_destroy(string_list_t* list) {
  g_list_foreach(list->items, (GFunc)g_free, NULL);
  g_list_foreach(list->funcs, (GFunc)g_free, NULL);
  g_free(list->name);
  g_free(list->label);
  g_free(list);
}

static string_list_t* string_list_find(const char* name) {
  GList* dlist;

  g_assert(StringListModule);
  for (dlist = StringListModule->lists; dlist; dlist = dlist->next) {
    string_list_t* sl = dlist->data;
    if (!strcmp(sl->name, name)) return sl;
  }
  return NULL;
}

static string_list_t* string_list_get_and_warn(const char* name) {
  string_list_t* list = string_list_find(name);
  if (!list) {
    g_log("SLIST", G_LOG_LEVEL_WARNING,
          "string list '%s' not found", name);
  }
  return list;
}

/** Whether to allow whitespaces.
 * @param name the name of the string list.
 * @param csens TRUE to be case sensitive, FALSE otherwise
 */
void string_list_set_case(const char* name, gboolean csens) {
  string_list_t* list = string_list_get_and_warn(name);
  if (list) list->case_sensitive = csens;
}

/** Get the string list items.
 * @param name the name of the list.
 * @return the items, or NULL
 * @note this function might be removed soon, use
 * string_list_foreach()
 */
const GList* string_list_get_items(const char* name) {
  string_list_t* list = string_list_get_and_warn(name);
  return list?list->items:NULL;
}

/** Get the label of a string list.
 * @param name the name of the stringlist
 * @return the label of the string list, or NULL if not found.
 */
const char* string_list_get_label(const char* name) {
  string_list_t* list = string_list_get_and_warn(name);
  return list?list->label:NULL;
}

/** Check whether a string list allows white spaces.
 * @param name the name of the string list.
 * @return TRUE if white spaces are allowed, FALSE otherwise.
 */
gboolean string_list_get_allow_ws(const char* name) {
  string_list_t* list = string_list_get_and_warn(name);
  return list?list->allow_ws:TRUE;
}

/** Monitor a string list for changes.
 * @param name the name of the string list.
 * @param added function to call when an item was remove.
 * @param removed function to call before an item is removed.
 * @param data user data for the callbacks.
 */
void string_list_monitor(const char* name, StringListFunc added,
                         StringListFunc removed, void* data) {
  string_list_t* list = string_list_get_and_warn(name);
  sl_func_t* func;

  if (!list) return;

  func = g_malloc(sizeof(*func));
  func->added = added;
  func->removed = removed;
  func->data = data;
  list->funcs = g_list_append(list->funcs, func);
}


static string_list_t* string_list_new(const char* name, const char* label,
                                      gboolean allow_ws) {
  string_list_t* sl;

  g_assert(StringListModule);

  sl = g_malloc(sizeof(*sl));
  sl->name = g_strdup(name);
  sl->label = g_strdup(label);
  sl->items = NULL;
  sl->funcs = NULL;
  sl->allow_ws = allow_ws;
  sl->case_sensitive = FALSE;
  sl->save = TRUE;
  sl->registered = FALSE;
  StringListModule->lists = g_list_append(StringListModule->lists, sl);
  return sl;
}

/** Check whehter a GList contains an item.
 * @param sl the list
 * @param item the item
 * @return the GList element, or NULL if not found
 */
GList* string_list_contains(GList* sl, const char* item) {
  GList* dlist;
  for (dlist = sl; dlist; dlist = dlist->next) {
    char* text = dlist->data;
    //    if (sl->case_sensitive) {
    //      if (!strcmp(text, item)) return dlist;
    //    } else {
    if (!strcasecmp(text, item)) return dlist;
    //    }
  }
  return NULL;
}

/** Check whehter a GSList contains an item.
 * @param sl the list
 * @param item the item
 * @return the GSList element, or NULL if not found
 */
GSList* string_slist_contains(GSList* sl, const char* item) {
  GSList* dlist;
  for (dlist = sl; dlist; dlist = dlist->next) {
    char* text = dlist->data;
    //    if (sl->case_sensitive) {
    //      if (!strcmp(text, item)) return dlist;
    //    } else {
    if (!strcasecmp(text, item)) return dlist;
    //    }
  }
  return NULL;
}

/** Register a new string list. The string list by default is not
 * case sensitive and automatically saved to disk.
 * @param name the name of the string list.
 * @param label a short desciption of the string list (for UI)
 * @param allow_ws TRUE to allow white spaces in entries, FALSE otherwise
 * @see string_list_set_case(), string_list_set_save()
 */
void string_list_register(const char* name, const char* label, gboolean allow_ws) {
  string_list_t* sl;

  g_assert(name);
  g_assert(label);
  sl = string_list_find(name);
  if (sl) {
    g_free(sl->label);
    sl->label = g_strdup(label);
    sl->allow_ws = allow_ws;
  } else {
    sl = string_list_new(name, label, allow_ws);
  }
  sl->registered = TRUE;
}

static gboolean string_list_add_real(string_list_t* sl, const char* item) {
  GList* dlist;
  
  if (!string_list_contains(sl->items, item)) {
    char* nitem = g_strdup(item);
    sl->items = g_list_append(sl->items, nitem);

    for (dlist = sl->funcs; dlist; dlist = dlist->next) {
      sl_func_t* func = dlist->data;
      if (func->added) func->added(nitem, func->data);
    }
    return TRUE;
  }
  return FALSE;
}

/** Add a new item to a string list. If the item is already in the list.
 * nothing is inserted. If the string list does not allow white spaces
 * but the item contains some, the item is split into several items which
 * all are inserted.
 * @param name the name of the string list.
 * @param item the item to add.
 * @return TRUE if insertion succeeded, FALSE otherwise
 */
gboolean string_list_add(const char* name, const char* item) {
  string_list_t* sl;
  gboolean ret = FALSE;
  
  sl = string_list_get_and_warn(name);
  if (!sl) return FALSE;

  if (sl->allow_ws) {
    ret |= string_list_add_real(sl, item);
  } else {
    char* temp = g_strdup(item);
    item = arg(temp, 0);
    while (item) {
      ret |= string_list_add_real(sl, item);
      item = arg(NULL, 0);
    }
    g_free(temp);
  }
  if (ret) string_list_save();
  return ret;
}

static void string_list_remove_real(string_list_t* list, GList* item) {
  GList* dlist;

  for (dlist = list->funcs; dlist; dlist = dlist->next) {
    sl_func_t* func = dlist->data;
    if (func->removed) func->removed(item->data, func->data);
  }
  g_free(item->data);
  list->items = g_list_delete_link(list->items, item);
}

/** Remove an item from a string list.
 * @param name the name of the string list.
 * @param item the item to remove.
 * @return TRUE if item was removed, FALSE otherwise
 */
gboolean string_list_remove(const char* name, const char* item) {
  string_list_t* sl;
  GList* link;

  sl = string_list_get_and_warn(name);
  if (!sl) return FALSE;

  link = string_list_contains(sl->items, item);
  if (!link) return FALSE;

  string_list_remove_real(sl, link);

  string_list_save();
  return TRUE;
}

/** Check whether a string list contains a certain item.
 * @param name the name of the stringlist
 * @param item the item to look for.
 * @return TRUE if the stringlist contains the item, FALSE otherwise
 */
gboolean string_list_check(const char* name, const char* item) {
  string_list_t* sl;

  sl = string_list_find(name);
  if (!sl || !string_list_contains(sl->items, item)) return FALSE;
  else return TRUE;
}

/** Iterate through the items of a string list.
 * @param name the name of the string list.
 * @param func the function to call on each item
 * @param data the user data for the function
 */
void string_list_foreach(const char* name, StringListFunc func, void* data) {
  string_list_t* sl = string_list_get_and_warn(name);
  GList* dlist;

  if (!sl) return;
  g_assert(func);

  for (dlist = sl->items; dlist; dlist = dlist->next) {
    const char* item = dlist->data;
    func(item, data);
  }
}

/** Whether to save the string list to disk.
 * @param name the name of the string list.
 * @param save TRUE to save to disk, FALSE otherwise
 */
void string_list_set_save(const char* name, gboolean save) {
  string_list_t* sl = string_list_get_and_warn(name);
  if (!sl) return;
  if (sl->save == save) return;
  sl->save = save;
  if (save) string_list_save();
}

/** Copy the items from one list to another.
 * @param dest the name of the destination list.
 * @param src the name of the source list
 * @note Only items in the destination list are only changed if it is
 * necessary, for example, if source and destination contain the same
 * items, nothing is copied at all, thus the destination does not get
 * dirty (see string_list_monitor())
 */
void string_list_copy_items(const char* dest, const char* src) {
  string_list_t* sl1 = string_list_get_and_warn(dest);
  string_list_t* sl2 = string_list_get_and_warn(src);
  GList* dlist;

  if (!sl1 || !sl2) return;
  if (sl1->allow_ws != sl1->allow_ws) {
    g_log("SLIST", G_LOG_LEVEL_WARNING,
          "cannot copy items from '%s' to '%s': incompatible lists",
          src, dest);
    return;
  }
  sl2->case_sensitive = sl2->case_sensitive;

  dlist = sl1->items;
  while (dlist) {
    GList* link = dlist;
    const char* item = link->data;
    
    dlist = dlist->next;
    if (!string_list_contains(sl2->items, item))
      string_list_remove_real(sl1, link);
  }

  for (dlist = sl2->items; dlist; dlist = dlist->next) {
    const char* item = dlist->data;

    if (!string_list_contains(sl1->items, item))
      string_list_add_real(sl1, item);
  }
  if (sl1->save) string_list_save();
}
  
void string_lists_foreach(StringListFunc func, void* data) {
  GList* dlist;

  g_assert(StringListModule);

  for (dlist = StringListModule->lists; dlist; dlist = dlist->next) {
    string_list_t* sl = dlist->data;
    if (!sl->save) continue;
    func(sl->name, data);
  }
}


static void xml_get_list(xmlDocPtr xdoc, xmlNodePtr xnode) {
  xmlChar* attr;
  xmlNodePtr xsub;
  string_list_t* list;

  attr = xmlGetProp(xnode, "Name");
  if (!attr) return;

  list = string_list_find(attr);
  if (!list) list = string_list_new(attr, NULL, FALSE);

  xmlFree(attr);
  
  xsub = xnode->xmlChildrenNode;
  while (xsub != NULL) {
    if (!strcmp(xsub->name, "Item")) {
      attr = xmlNodeListGetString(xdoc, xsub->xmlChildrenNode, 1);
      if (attr) {
        string_list_add_real(list, attr);
        xmlFree(attr);
      }
    }
    xsub = xsub->next;
  }
}

static void xml_get_lists(xmlDocPtr xfile, xmlNodePtr xnode) {
  xmlNodePtr xsub = xnode->xmlChildrenNode;
  while (xsub != NULL) {
    if (!strcmp(xsub->name, "StringList")) {
      xml_get_list(xfile, xsub);
    }
    xsub = xsub->next;
  }
}

static void xml_save_item(const char* item, xmlNodePtr xnode) {
  xmlNewTextChild(xnode, NULL, "Item", item);
}

static void xml_save_list(string_list_t* list, xmlNodePtr xnode) {
  if (list->save && (list->registered || StringListModule->save_unregistered)) {
    xmlNodePtr xsub = xmlNewChild(xnode, NULL, "StringList", NULL);
    xmlNewProp(xsub, "Name", list->name);
    g_list_foreach(list->items, (GFunc)xml_save_item, xsub);
  }
}

static void xml_save_lists(xmlDocPtr xfile, xmlNodePtr xroot) {
  (void)xfile;
  g_list_foreach(StringListModule->lists, (GFunc)xml_save_list, xroot);
}

void string_lists_save_unregistered(gboolean save) {
  StringListModule->save_unregistered = save;
  string_list_save();
}


static void string_lists_init() {
}
static void string_lists_run() {
}
static void string_lists_exit() {
  if (StringListModule) {
    g_list_foreach(StringListModule->lists, (GFunc)string_list_destroy, NULL);
    g_list_free(StringListModule->lists);
    StringListModule->lists = NULL;
  }
}

static void string_list_module_init(string_list_module_t* module) {
  module_init(MODULE(module), "String List Module",
              (ModuleFunc)string_lists_init,
              (ModuleFunc)string_lists_run,
              (ModuleFunc)string_lists_exit);
  module->lists = NULL;
  module->save_unregistered = TRUE;

  module_require(xml_module_instance);

  lopster_xml_register_home("core:string_lists", "string_lists.xml",
                            "StringLists", xml_get_lists, xml_save_lists);
  string_list_load();
}


/** You need to require this module if you like to use string lists.
 * @return the mdoule instance
 */
module_t* string_list_module_instance() {
  if (!StringListModule) {
    StringListModule = g_malloc0(sizeof(*StringListModule));
    string_list_module_init(StringListModule);
  }
  return MODULE(StringListModule);
}

#ifdef WIN32
# define _STRING_LIST_C_
# include "lcorealiasdef.c"
#endif
