/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.11 $
 * $Date: 2007/11/23 08:19:57 $
 * $Author: sid4u $
 */

#include <string.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif
#include <errno.h>

#include "utils.h"
#include "tcp_socket.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
} tsocket_module_t;

static tsocket_module_t* TSocketModule = NULL;

/** Connect the socket.  This function sets the ip/port and then calls
 * socket_activate()
 * @param sock the socket
 * @param ip the destination ip in network byte order.
 * @param port the port in network byte order.
 * @return 1 on success, 0 otherwise.
 */
int tcp_socket_connect(tcp_socket_t* sock, ip_t ip, port_t port) {
  g_assert(TSocketModule);
  sock->ip = ip;
  sock->port = port;
  return socket_activate(SOCK(sock));
}

/** Moves (switches) the tcp socket.
 * @param dest the destination
 * @param src the source
 * @param copy_buffer whether or not to copy the in and out buffer.
 */
void tcp_socket_move(tcp_socket_t* dest, tcp_socket_t* src,
                     gboolean copy_buffer) {
  g_assert(TSocketModule);
  b_socket_move(BSOCK(dest), BSOCK(src), copy_buffer);
  dest->ip = src->ip;
  dest->port = src->port;
}

static int tcp_socket_activate(tcp_socket_t* sock) {
  socket_t* s = SOCK(sock);  
  struct sockaddr_in server;
  int res;
  int len = 1;
  gboolean err;

  g_log("SOCK", G_LOG_LEVEL_DEBUG,
	"connecting %p to [%s:%d] (TCP)", s, ntoa(sock->ip),
	ntohs(sock->port));
  
  if (!socket_create(SOCK(sock), SOCK_STREAM, IPPROTO_TCP)) {
    socket_done(s, ST_CON_ERROR, "Could not connect");
    return 0;
  }
  socket_set_nonblock(s, TRUE);

  server.sin_family = AF_INET;
  server.sin_addr.s_addr = sock->ip;
  server.sin_port = sock->port;

  setsockopt(socket_get_fd(s), SOL_SOCKET, SO_REUSEADDR,
	     &len, sizeof(len));
  setsockopt(socket_get_fd(s), SOL_SOCKET, SO_KEEPALIVE,
	     &len, sizeof(len));

  res = connect(socket_get_fd(s), (struct sockaddr *) &server, sizeof(server));
  if (res == 0) return 1;
  
#ifdef _WIN32
  err = (WSAGetLastError() != WSAEWOULDBLOCK);
#else
  err = (errno != EINPROGRESS);
#endif
  if (err) {
    socket_done(s, ST_CON_ERROR, "Could not connect");
    return 0;
  }
  return 1;
}

/** uses socket_recv() */
static int tcp_socket_read(tcp_socket_t* s) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));
  int read;
  int bytes;
  
  buffer_need_space(buffer, 2);

  bytes = buffer_space(buffer) -1;
  if (bytes > 20480) bytes = 20480;

  read = socket_recv(SOCK(s), buffer_write_pos(buffer), 0, bytes);
  if (read > 0) {
    buffer->size += read;
    buffer->data[buffer->size] = 0;
  }
  return read;
}

/** uses socket_send() */
static int tcp_socket_write(tcp_socket_t* s) {
  buffer_t* buffer = b_socket_get_out_buffer(BSOCK(s), 0);
  int bytes = buffer_avail(buffer);
  int written;

  g_assert(buffer);
  if (bytes > 20480) bytes = 20480;
  written = socket_send(SOCK(s), buffer_read_pos(buffer), 0, bytes);
  
  if (written > 0) buffer_consume_virt(buffer, written);
  return written;
}

/** Ends the socket. Call from the end function of the derived class only.
 * @param sock the socket.-
 * @param message the ending message
 * @return TRUE if socket should be destroyed, FALSE otherwise
 */
gboolean tcp_socket_end(tcp_socket_t* sock, const char* message) {
  g_assert(TSocketModule);
  return b_socket_end(BSOCK(sock), message);
}

/** Destroys the socket. Call from destroy function of the derived
 * class only.
 * @param sock the socket
 */
void tcp_socket_destroy(tcp_socket_t* sock) {
  g_assert(TSocketModule);
  b_socket_destroy(BSOCK(sock));
}

/** Initializes the socket. Call from init function of the derived
 * class only.
 * @param s the socket
 * @param name see b_socket_init()
 * @param timeout see b_socket_init()
 */
void tcp_socket_init(tcp_socket_t* s, const char* name, int timeout) {
  g_assert(TSocketModule);
  b_socket_init(BSOCK(s), name, timeout);
  object_add_type(OBJ(s), "TcpSocket");
  OBJ(s)->destroy = (ObjectFunc)tcp_socket_destroy;

  socket_set_end(SOCK(s), (SockEndFunc)tcp_socket_end);
  socket_set_activate(SOCK(s), (SockActivateFunc)tcp_socket_activate);
  b_socket_set_read_write(BSOCK(s), (BReadFunc)tcp_socket_read, (BWriteFunc)tcp_socket_write);

  s->ip = 0;
  s->port = 0;
  socket_add_load(SOCK(s), "TCP");
}

/** Create a new tcp socket.
 * @param name passed to tcp_socket_init()
 * @param timeout passed to tcp_socket_init()
 */
tcp_socket_t* tcp_socket_new(const char* name, int timeout) {
  tcp_socket_t* s = g_malloc0(sizeof(*s));
  tcp_socket_init(s, name, timeout);
  return s;
}

/** Send a string. 
 * @param s the socket.
 * @param message the string to send
 */
void tcp_socket_send_str(tcp_socket_t* s, const char* message)
{
    g_assert(TSocketModule);
    tcp_socket_send_bin(s, message, strlen(message));
}

/** Send binary data.
 * @param s the socket.
 * @param message the data
 * @param len the length of the data.
 */
void tcp_socket_send_bin(tcp_socket_t* s, const char* message, int len) {
  buffer_t* buffer = b_socket_get_out_buffer(BSOCK(s), len+1);
  memcpy(buffer_write_pos(buffer), message, len);
  buffer->size += len;
  b_socket_start_sending(BSOCK(s));
}


static void tsockets_init() {
}
static void tsockets_run() {
}
static void tsockets_exit() {
}

static void tsocket_module_init(tsocket_module_t* module) {
  module_init(MODULE(module), "TSocket Module",
	      (ModuleFunc)tsockets_init,
	      (ModuleFunc)tsockets_run,
	      (ModuleFunc)tsockets_exit);

  module_require(bsocket_module_instance);
  module_require(lsocket_module_instance);
  
}


/** Require this function if you like to use tcp socket. 
 * @return the module instance
 * @see module_require()
 */
module_t* tsocket_module_instance() {
  if (!TSocketModule) {
    TSocketModule = g_malloc0(sizeof(*TSocketModule));
    tsocket_module_init(TSocketModule);
  }
  return MODULE(TSocketModule);
}

#ifdef WIN32
#define _TCP_SOCKET_C_
#include "lcorealiasdef.c"
#endif

