/* Copyright (C) 2005-2006 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.12 $
 * $Date: 2007/11/25 22:18:53 $
 * $Author: sgop $
 */

#include <string.h>

#include <glib.h>

#include "thread.h"
#include "l_base.h"
#include "l_i18n.h"
#include "ui.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

struct _thread_priv_t {
    unsigned status;
    unsigned progress;
    gboolean interrupt;
    char* des;
    gboolean dirty;
    unsigned update_interval;
    int timer;
    GMutex* mutex;
    GThread* handle;
    ThreadMainFunc main;
    
    GCond* cond;
    ThreadSyncFunc cond_func;
    void* cond_data;
    gboolean can_force_interrupt;
};

static GStaticMutex ListMutex = G_STATIC_MUTEX_INIT;

typedef struct {
    module_t _module;
    GList* running;
    GList* threads;
    gboolean exit_requested;
} thread_module_t;

static thread_module_t* ThreadModule = NULL;


#define thread_list_remove(list, item)          \
    g_static_mutex_lock(&ListMutex);            \
    list = g_list_remove(list, item);           \
    g_static_mutex_unlock(&ListMutex);

#define thread_list_append(list, item)            \
    g_static_mutex_lock(&ListMutex);              \
    list = g_list_append(list, item);             \
    g_static_mutex_unlock(&ListMutex);


// thread safe
static void thread_set_status(thread_t* thread, unsigned status)
{
    if (thread->priv->status != status)
    {
        //    printf("-thread_set_status %d\n", status);
        if (status == THREAD_RUNNING)
        {
            thread_list_append(ThreadModule->running, thread);
        }
        else if (thread->priv->status == THREAD_RUNNING)
        {
            thread_list_remove(ThreadModule->running, thread);
        }
        thread->priv->status = status;
        thread->priv->dirty = TRUE;
    }
}

/** Get the status of the thread. This function is thread-safe.
 * @param thread the thread.
 * @return the status.
 */
unsigned thread_get_status(thread_t* thread) {
  return thread->priv->status;
}

/** Interrupt a thread.  The thread's status is set to
 * THREAD_INTERRUPTED and the interrupt member of the class is set to
 * TRUE. The main thread must read that member from time to time and
 * abort the calculation, so the thread gets really interrupted.
 * @param thread the thread
 */
void thread_interrupt(thread_t* thread)
{
    if (thread->priv->status != THREAD_RUNNING) return;
    //  printf("-thread_interrupt\n");
    thread_set_status(thread, THREAD_INTERRUPTED);
    thread->priv->interrupt = TRUE;
}

static void thread_wait(thread_t* thread) {
  if (thread->priv->handle) {
    //    printf("-thread_wait\n");
    g_thread_join(thread->priv->handle);
    thread->priv->handle = NULL;
  }
}

static void thread_remove_source(thread_t* thread) {
  if (thread->priv->timer) {
    lopster_timeout_remove(thread->priv->timer);
    thread->priv->timer = 0;
  }
}
static gboolean thread_free(thread_t* thread) {
  //  printf("-thread_free\n");
  thread_remove_source(thread);
  thread_wait(thread);
  // do the last update.
  if (thread->priv->dirty) object_set_dirty(OBJ(thread));
  object_kill(OBJ(thread));
  return FALSE;
}

// thread safe
static void thread_done(thread_t* thread, unsigned status) {
  //  printf("-thread_done\n");
  thread_set_status(thread, status);
  // remove update timer
  thread_remove_source(thread);
  // and new source, that calls thread_free() from main loop
  thread->priv->timer = lopster_idle_add((GSourceFunc)thread_free, thread);
}

/** Destroy the thread. Call from base class destroy only.
 * @param thread the thread to destroy
 */
void thread_destroy(thread_t* thread)
{
    thread_list_remove(ThreadModule->threads, thread);

    //  printf("-thread_destroy\n");
    g_free(thread->priv->des);
    g_mutex_free(thread->priv->mutex);
    g_free(thread->priv);
    object_destroy(OBJ(thread));
    
    if (!ThreadModule->threads && ThreadModule->exit_requested)
        lopster_exit_task_terminated();
}

/** Initialize a thread. Call this function from the init method of
 * your derive class only.
 * @param thread the thread
 * @param name the name of the thread.
 * @param update interval in milliseconds the thread gets dirty, use
 * object_monitor() or object_monitor_add() to get a function called
 * every 'update' ms. The updates are done in the thread you've called
 * thread_start() from.
 */
void thread_init(thread_t* thread, const char* name, unsigned update) {
  //  printf("-thread_init\n");
  object_init(OBJ(thread), name, TRUE);
  object_add_type(OBJ(thread), "Thread");
  OBJ(thread)->destroy = (ObjectFunc)thread_destroy;

  thread->priv = g_malloc0(sizeof(thread_priv_t));
  thread->priv->mutex = g_mutex_new();
  thread->priv->progress = 0;
  thread->priv->des = NULL;
  thread->priv->status = THREAD_NOT_STARTED;
  thread->priv->interrupt = FALSE;
  thread->priv->dirty = FALSE;
  thread->priv->update_interval = update?update:200;
  thread->priv->timer = 0;
  thread->priv->main = NULL;
  thread->priv->cond = NULL;
  thread->priv->can_force_interrupt = FALSE;
  thread_list_append(ThreadModule->threads, thread);
  object_make_public(OBJ(thread), NULL);
}

/** Create a simple threaded calculation.
 * @param update interval in milliseconds the thread gets dirty.
 * @return a new thread
 */
thread_t* thread_new(unsigned update) {
  thread_t* thread;
  //  printf("-thread_new\n");
  thread = g_malloc0(sizeof(*thread));
  thread_init(thread, "Default operation", update);

  return thread;
}

/** Lock the thread.
 * @param thread the thread.
 * @note The mutex used here is the same that is used by other thread
 * functions, so make sure you don't run into a deadlock
 */
void thread_lock(thread_t* thread) {
  //  printf("-thread_lock\n");
  g_mutex_lock(thread->priv->mutex);
}

/** Unlock the thread which was locked with thread_lock().
 * @param thread the thread.
 */
void thread_unlock(thread_t* thread) {
  //  printf("-thread_unlock\n");
  g_mutex_unlock(thread->priv->mutex);
}

/** Set the description of the thread. This function is thread safe.
 * @param thread the thread.
 * @param des the new description.
 */
void thread_set_description(thread_t* thread, const char* des) {
  thread_lock(thread);
  if (strcmp(des?des:"", thread->priv->des?thread->priv->des:"")) {
    //    printf("-thead_set_des %s\n", des);
    if (thread->priv->des) g_free(thread->priv->des);
    thread->priv->des = des?g_strdup(des):NULL;
    thread->priv->dirty = TRUE;
  }
  thread_unlock(thread);
}

/** Get the description of the thread. FIXME: this function is not
 * thread-safe.
 * @param thread the thread.
 * @return the description.
 */
const char* thread_get_description(thread_t* thread) {
  return thread->priv->des;
}

/** Set the progress of the thread. This function is thread-safe.
 * @param thread the thread
 * @param val the progress from 0 to 100
 */
void thread_set_progress(thread_t* thread, unsigned val) {
  if (thread->priv->progress != val) {
    //    printf("-thead_set_progress %d\n", val);
    thread->priv->progress = val;
    thread->priv->dirty = TRUE;
  }
}

/** Get the progress of the thread. This function is thread-safe.
 * @param thread the thread.
 * @return the progress from 0 to 100
 */
unsigned thread_get_progress(thread_t* thread) {
  return thread->priv->progress;
}

/** Get the name of the thread.
 * @param thread the thread.
 * @return the name of the thread.
 */
const char* thread_get_name(thread_t* thread) {
  return OBJ(thread)->name;
}

static void* thread_main(thread_t* thread) {
  //  printf("-thread_main\n");
  thread->priv->main(thread);

  thread_set_progress(thread, 0);
  if (thread->priv->status == THREAD_RUNNING) {
    thread_done(thread, THREAD_COMPLETED);
  } else {
    thread_done(thread, thread->priv->status);
  }
  return NULL;
}

static gboolean thread_updater(thread_t* thread) {
  if (thread->priv->dirty) {
    object_set_dirty(OBJ(thread));
    thread->priv->dirty = FALSE;
  }
  return TRUE;
}

/** Start the thread.
 * @param thread
 * @param main the function to execute
 */
void thread_start(thread_t* thread, ThreadMainFunc main) {
  //  printf("-thread_start\n");
  if (!thread) return;
  if (thread->priv->status == THREAD_RUNNING) return;

  thread->priv->main = main;

  if (!thread->priv->main) {
    thread_done(thread, THREAD_FAILED);
    return;
  }
  
  thread_set_status(thread, THREAD_RUNNING);

  thread->priv->handle =
    g_thread_create((GThreadFunc)thread_main, thread, TRUE, NULL);

  if (!thread->priv->handle) {
    thread_done(thread, THREAD_FAILED);
  } else {
    thread->priv->timer =
      lopster_timeout_add(thread->priv->update_interval,
                    (GSourceFunc)thread_updater, thread);
  }
}

static gboolean _wakeup_thread(thread_t* thread)
{
    g_assert(thread->priv->cond);
    if (thread->priv->cond)
    {
        if (thread->priv->cond_func)
            thread->priv->cond_func(thread, thread->priv->cond_data);
        
        // lock to make sure the thread called g_cond_wait() already
        thread_lock(thread);
        g_cond_broadcast(thread->priv->cond);
        thread_unlock(thread);
    }
    return FALSE;
}

/** Pauses thread and executes function in main thread.  May only be
 * called from within the thread.
 * @param thread the thread.
 * @param sync the function to execute
 * @param data the data that should be passed to the function
 */
void thread_sync(thread_t* thread, ThreadSyncFunc sync, void* data)
{
    thread_lock(thread);
    
    g_assert(!thread->priv->cond);
    if (g_thread_self() != thread->priv->handle)
    {
        g_log("THREAD", G_LOG_LEVEL_WARNING,
              "thread_sync() may only be called from within thread '%s'",
              OBJ(thread)->name);
    }
    else
    {
        thread->priv->cond = g_cond_new();
        thread->priv->cond_func = sync;
        thread->priv->cond_data = data;
        
        lopster_idle_add((GSourceFunc)_wakeup_thread, thread);
        g_cond_wait(thread->priv->cond, thread->priv->mutex);
        g_cond_free(thread->priv->cond);
        thread->priv->cond = NULL;
    }

    thread_unlock(thread);
}

void thread_set_force_interrupt(thread_t* thread, gboolean value)
{
    thread->priv->can_force_interrupt = value;
}

/** Check whether the thread was interrupted by user.
 * Call this function from the thread main function to check whether
 * the calculation should be canceled.
 * @param thread the thread
 * @return TRUE if thread was interrupted.
 */
gboolean thread_interrupted(thread_t* thread)
{
    return thread->priv->interrupt;
}

static gboolean threads_exit_request(int mode)
{
    GList* dlist;
    gboolean ret;
    GList* wthreads = NULL;
    
    if (!ThreadModule) return TRUE;
    ThreadModule->exit_requested = TRUE;
    
    //FIXME: we don't expect new threads during exit. So make this
    //implementation easy and dont take care about un/locking thread
    //lists.

    // Interrupt threads that don't need user confirmation. 
    for (dlist = ThreadModule->running; dlist; dlist = dlist->next)
    {
        thread_t* thread = dlist->data;
        if (thread->priv->can_force_interrupt) 
        {
            wthreads = g_list_append(wthreads, thread);
            thread_interrupt(thread);
        }
    }
    
    //  printf("-threads_close\n");
    if (mode == LOPSTER_EXIT_SAFE || mode == LOPSTER_EXIT_SOON)
    {
        if (ThreadModule->running)
        {
            UI_threads_show(_("Waiting for tasks to be finished"));
            ret = FALSE;
        } else {
            ret = TRUE;
        }
    }
    else
    { // LOPSTER_EXIT_NOW
        // interrupt all threads (non-blocking)
        while (ThreadModule->running)
        {
            thread_t* thread = ThreadModule->running->data;
            wthreads = g_list_append(wthreads, thread);
            thread_interrupt(thread);
        }
        // wait for all threads (they all will add an idle function
        // which will be hopefully be executed before the main
        // programm exits) FIXME: verify
        for (dlist = wthreads; dlist; dlist = dlist->next)
        {
            thread_t* thread = dlist->data;
            thread_wait(thread);
        }
        ret = TRUE;
    }

    return ret;
}


static void threads_init()
{
    g_thread_init(NULL);
    lopster_exit_task_register(threads_exit_request);
}

static void threads_run()
{
}

static void threads_exit()
{
    //FIXME: implement.
}

static void thread_module_init(thread_module_t* module) {
  module_init(MODULE(module), "Thread Module",
	      (ModuleFunc)threads_init,
	      (ModuleFunc)threads_run,
	      (ModuleFunc)threads_exit);
  module->running = NULL;
  module->threads = NULL;
  module->exit_requested = FALSE;
}

/** Require this module if you are using threads.
 * @return the thread module
 * @see module_require()
 */
module_t* thread_module_instance() {
  if (!ThreadModule) {
    ThreadModule = g_malloc0(sizeof(*ThreadModule));
    thread_module_init(ThreadModule);
  }
  return MODULE(ThreadModule);
}

#ifdef WIN32
#define _THREAD_C_
#include "lcorealiasdef.c"
#endif
