/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.9 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include <string.h>

#include "utils.h"
#include "preferences.h"
#include "udp_socket.h"
#include "l_listen.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
  GList* active;
  GList* inactive;
} usocket_module_t;

static usocket_module_t* USocketModule = NULL;


static int udp_socket_activate(udp_socket_t* sock) {

  g_log("SOCK", G_LOG_LEVEL_DEBUG, "activating %p (UDP)", sock);
  
  if (!socket_create(SOCK(sock), SOCK_DGRAM, IPPROTO_UDP)) return 0;
  socket_set_nonblock(SOCK(sock), TRUE);

  return 1;
}

static int udp_socket_read(udp_socket_t* s) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));
  int read;
  int bytes;
  int hdr = sizeof(unsigned)+sizeof(ip_t);
  ip_t ip;
  
  buffer_need_space(buffer, hdr+4097);
  
  bytes = buffer_space(buffer)-1;

  read = socket_recv_from(SOCK(s), buffer_write_pos(buffer)+hdr,
                          bytes-hdr, &ip);
  if (read > 0) {
    memcpy(buffer_write_pos(buffer), &read, sizeof(unsigned));
    buffer->size += sizeof(unsigned);
    memcpy(buffer_write_pos(buffer), &ip, sizeof(ip_t));
    buffer->size += sizeof(ip_t);
    buffer->size += read;
    buffer->data[buffer->size] = 0;
  }
  return read;
}

static int udp_socket_write(socket_t* s) {
  buffer_t* buffer = b_socket_get_out_buffer(BSOCK(s), 0);
  int written;
  unsigned hdr = sizeof(unsigned)+sizeof(ip_t)+sizeof(port_t);
  unsigned len;
  ip_t ip;
  port_t port;

  g_assert(buffer);
  g_assert(buffer_avail(buffer) >= hdr);

  memcpy(&len, buffer_read_pos(buffer), sizeof(len));
  g_assert(buffer_avail(buffer) >= hdr+len);
  buffer_consume_virt(buffer, sizeof(len));

  memcpy(&ip, buffer_read_pos(buffer), sizeof(ip));
  buffer_consume_virt(buffer, sizeof(ip));
  
  memcpy(&port, buffer_read_pos(buffer), sizeof(port));
  buffer_consume_virt(buffer, sizeof(port));
  
  written = socket_send_to(SOCK(s), buffer_read_pos(buffer),
                           len, ip, port);

  buffer_consume_virt(buffer, len);
  return written;
}

/** Destroy the udp socket. Call from the destroy function of a
 * derived class only.
 * @param s the socket.
 */
void udp_socket_destroy(udp_socket_t* s) {
  b_socket_destroy(BSOCK(s));
}

/** Initialize the udp socket. Call from the init function of a
 * derived class only.
 * @param s the socket.
 * @param name the name of the socket, passed to b_socket_init()
 * @param timeout passed to b_socket_init()
 */
void udp_socket_init(udp_socket_t* s, const char* name, int timeout) {
  b_socket_init(BSOCK(s), name, timeout);
  object_add_type(OBJ(s), "UdpSocket");
  OBJ(s)->destroy = (ObjectFunc)udp_socket_destroy;
  socket_set_activate(SOCK(s), (SockActivateFunc)udp_socket_activate);
  b_socket_set_read_write
    (BSOCK(s), (BReadFunc)udp_socket_read,(BWriteFunc)udp_socket_write);
  
  socket_add_load(SOCK(s), "UDP");
}

/** Create a new UDP socket.
 * @param name passed to udp_socket_init()
 * @param timeout passed to udp_socket_init()
 * @return a new udp socket.
 */
udp_socket_t* udp_socket_new(const char* name, int timeout) {
  udp_socket_t* s = g_malloc0(sizeof(*s));
  udp_socket_init(s, name, timeout);
  return s;
}

/** Send a string to a destination address.
 * @param s the udp socket.
 * @param message the message to send
 * @param ip the ip to send to.
 * @param port the port to send to.
 */
void udp_socket_send_str_to(udp_socket_t* s, const char* message,
                            ip_t ip, port_t port) {
  udp_socket_send_bin_to(s, message, strlen(message), ip, port);
}

/** Send binary data to a destination address.
 * @param s the udp socket.
 * @param message the data to send
 * @param len the length of the data
 * @param ip the ip to send to.
 * @param port the port to send to.
 */
void udp_socket_send_bin_to(udp_socket_t* s, const char* message, int len,
                            ip_t ip, port_t port) {
  buffer_t* buffer;
  int hdr = sizeof(ip_t)+sizeof(unsigned)+sizeof(port_t);

  g_assert(len>=0);

  buffer = b_socket_get_out_buffer(BSOCK(s), len+hdr+1);
  memcpy(buffer_write_pos(buffer), &len, sizeof(len));
  buffer->size += sizeof(len);
  memcpy(buffer_write_pos(buffer), &ip, sizeof(ip));
  buffer->size += sizeof(ip);
  memcpy(buffer_write_pos(buffer), &port, sizeof(port));
  buffer->size += sizeof(port);
  memcpy(buffer_write_pos(buffer), message, len);
  buffer->size += len;
  
  b_socket_start_sending(BSOCK(s));
}




static void udp_listen_start_listening_real(udp_listen_t* list) {
  g_assert(USocketModule);

  if (socket_connected(SOCK(list))) return;

  if (pref_get_ref("/core/listen/firewalled", gboolean)) return;

  socket_activate(SOCK(list));
  
  if (socket_connected(SOCK(list))) {
    USocketModule->inactive = g_list_remove(USocketModule->inactive, list);
    USocketModule->active = g_list_append(USocketModule->active, list);
    //    if (listen->change) listen->change(listen);
  }
}

/** Start listening.
 * @param list the listen object
 */
void udp_listen_start_listening(udp_listen_t* list) {
  list->activated = TRUE;
  udp_listen_start_listening_real(list);
}

static void udp_listen_stop_listening_real(udp_listen_t* l) {
  g_assert(USocketModule);

  if (!socket_connected(SOCK(l))) return;
  g_log("LISTEN", G_LOG_LEVEL_DEBUG,
	"stop listening on port [%d] (UDP)", ntohs(l->port));
  socket_suspend(SOCK(l), DIR_IN|DIR_OUT);
  socket_set_fd(SOCK(l), -1, TRUE);

  USocketModule->active = g_list_remove(USocketModule->active, l);
  USocketModule->inactive = g_list_append(USocketModule->inactive, l);
}

/** Stop listening.
 * @param l the udp listen socket.
 */
void udp_listen_stop_listening(udp_listen_t* l) {
  l->activated = FALSE;
  udp_listen_stop_listening_real(l);
}


static int udp_listen_activate(udp_listen_t* list) {
  socket_t* s = SOCK(list);
  unsigned short port;

  if (!list) return 0;
  
  g_log("LISTEN", G_LOG_LEVEL_DEBUG,
	"start listening [%s] on port %d (UDP)",
	OBJ(s)->name, ntohs(list->port));

  socket_create(s, SOCK_DGRAM, IPPROTO_UDP);

  port = socket_bind(s, htonl(INADDR_ANY), list->port);
  if (!port) return 0;
  list->port = port;
  port = ntohs(port);

  g_log("LISTEN", G_LOG_LEVEL_DEBUG,
	"started listening on port %d", port);
  b_socket_start_receiving(BSOCK(list), 0);
  return 1;
}

/** Change the port of the listen socket.
 * @param l the listen socket.
 * @param port the new port
 * @note the port must be in the list of allowed ports (see port_is_allowed())
 * @note the preferences setting (if there is any) is not changed.
 */
void udp_listen_change_port(udp_listen_t* l, port_t port) {
  if (l->port == htons(port)) return;

  g_log("LISTEN", G_LOG_LEVEL_DEBUG,
        "changing listen port %d -> %d", ntohs(l->port), port);

  if (socket_connected(SOCK(l))) {
    udp_listen_stop_listening_real(l);
    l->port = htons(port);
    udp_listen_start_listening_real(l);
  } else {
    l->port = htons(port);
  }
}

static void udp_listen_changed(pref_node_t* node, udp_listen_t* list) {
  int* port = node->value;
  udp_listen_change_port(list, *port);
}

/** Destroy the listen socket. Call from the destroy function of your
 * derived class only.
 * @param l the listen socket.
 */
void udp_listen_destroy(udp_listen_t* l) {
  if (l->pref) {
    pref_unobserve(l->pref, (ObjectFunc)udp_listen_changed, l);
    g_free(l->pref);
  }
  USocketModule->inactive = g_list_remove(USocketModule->inactive, l);
  udp_socket_destroy(USOCK(l));
}

/** End the listen port. Call from your ending function of the derived
 * class only
 * @param list the listen socket.
 * @param message ending message.
 * @return TRUE
 */
gboolean udp_listen_end(udp_listen_t* list, const char* message) {
  udp_listen_stop_listening(list);
  socket_end(SOCK(list), message);
  return TRUE;
}

/** Initialize the udp listen socket. Call from the init function of
 * your derived class
 * @param list the listen socket.
 * @param pref the preferences value that specifies the port of the
 * listen socket. The listen socket will automatically react on port
 * changes.
 * @param name the name of the socket.
 */
void udp_listen_init(udp_listen_t* list, const char* pref, const char* name) {
  port_t port = (port_t)pref_get_ref(pref, int);

  udp_socket_init(USOCK(list), name, 0);
  object_add_type(OBJ(list), "UdpListenPort");
  OBJ(list)->destroy = (ObjectFunc)udp_listen_destroy;
  socket_set_end(SOCK(list), (SockEndFunc)udp_listen_end);
  socket_set_activate(SOCK(list), (SockActivateFunc)udp_listen_activate);

  list->port = htons(port);
  list->activated = FALSE;
  list->pref = g_strdup(pref);

  USocketModule->inactive = g_list_append(USocketModule->inactive, list);
  if (pref) pref_observe(pref, (ObjectFunc)udp_listen_changed, list);
}

/** Create a new udp listen socket.
 * @param pref passed to listen_port_init()
 * @param name passed to listen_port_init()
 * @return a new udp listen socket
 */
udp_listen_t* udp_listen_new(const char* pref, const char* name) {
  udp_listen_t* list = g_malloc0(sizeof(*list));
  udp_listen_init(list, pref, name);
  return list;
}

/** Get the port of the listen socket.
 * @param list the listen socket
 */
port_t udp_listen_get_port(udp_listen_t* list) {
  if (!list) return 0;
  return ntohs(list->port);
}



static void udp_listen_fw_toggle(pref_node_t* node)
{
    gboolean* fw = node->value;
    GList* dlist;
    
    g_assert(USocketModule);
    
    if (*fw)
    {
        dlist = USocketModule->active;
        while (dlist)
        {
            udp_listen_t* l = dlist->data;
            dlist = dlist->next;
            udp_listen_stop_listening_real(l);
        }
    }
    else
    {
        dlist = USocketModule->inactive;
        while (dlist)
        {
            udp_listen_t* l = dlist->data;
            dlist = dlist->next;
            if (l->activated)
                udp_listen_start_listening_real(l);
        }
    }
}


static void usocket_init()
{
    pref_observe("/core/listen/firewalled",
                 (ObjectFunc)udp_listen_fw_toggle, NULL);
}

static void usocket_run()
{
}

static void usocket_exit()
{
    while (USocketModule->active)
    {
        udp_listen_t* list = USocketModule->active->data;
        socket_done(SOCK(list), ST_DONE, "ShutDown");
    }
    while (USocketModule->inactive)
    {
        udp_listen_t* list = USocketModule->inactive->data;
        socket_done(SOCK(list), ST_DONE, "ShutDown");
    }
    USocketModule = NULL;
}

static void usocket_module_init(usocket_module_t* module)
{
    module_init(MODULE(module), "Udp Module",
                (ModuleFunc)usocket_init,
                (ModuleFunc)usocket_run,
                (ModuleFunc)usocket_exit);
    
    module->active = NULL;
    module->inactive = NULL;
    
    module_require(pref_module_instance);
    module_require(bsocket_module_instance);
    module_require(listen_module_instance);
}

/** If using udp listen sockets, require this module 
 * @see module_require()
 */
module_t* usocket_module_instance()
{
    if (!USocketModule)
    {
        USocketModule = g_malloc0(sizeof(*USocketModule));
        usocket_module_init(USocketModule);
    }
    return MODULE(USocketModule);
}

#ifdef WIN32
#define _UDP_SOCKET_C_
#include "lcorealiasdef.c"
#endif

