/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.20 $
 * $Date: 2007/11/21 20:18:24 $
 * $Author: sgop $
 */

#include <string.h>

#include "ui.h"
#include "module.h"
#include "xml_handler.h"
#include "plugin.h"
#include "dynamic.h"
#include "legacy.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

static ui_interface_t UI;


typedef struct {
    module_t _module;
    void (*init)(ui_interface_t*);
    void (*run)(void);
    void (*done)(void);
} ui_module_t;

static ui_entry_t* TheUI = NULL;
static gboolean UICommandLine = FALSE;
static gboolean UIFailed = FALSE;
static ui_module_t* UIModule = NULL;
static GMainLoop* MainLoop = NULL;
static GList* AvailableUIs = NULL;

static ui_entry_t* ui_entry_new(const char* name, const char* iname,
                                const char* filename)
{
    ui_entry_t* entry = g_malloc0(sizeof(*entry));
    entry->name = name?g_strdup(name):g_strdup(iname);
    entry->iname = g_strdup(iname);
    entry->filename = g_strdup(filename);
    return entry;
}

static void ui_entry_destroy(ui_entry_t* entry)
{
    g_free(entry->name);
    g_free(entry->iname);
    g_free(entry->filename);
    g_free(entry);
}

static ui_entry_t* ui_entry_find(const char* pname, const char* iname)
{
    GList* dlist;
    for (dlist = AvailableUIs; dlist; dlist = dlist->next)
    {
        ui_entry_t* entry = dlist->data;
        if (pname && !strcmp(pname, entry->name)) return entry;
        if (iname && !strcmp(iname, entry->iname)) return entry;
    }
    return NULL;
}

ui_entry_t* ui_get_list(GList** list)
{
    if (list) *list = AvailableUIs;
    return TheUI;
}

static const char* _check_for_ui(const char* folder, const char* network)
{
    const char* iface = TheUI->iname;
    char* temp;

    temp = l_printf(SDIR("%s")"lib%s_%s.so", folder, iface, network);
    if (g_file_test(temp, G_FILE_TEST_EXISTS)) return temp;
    // try different one
    temp = l_printf(SDIR("%s")"lib%s_%s.so.0", folder, iface, network);
    if (g_file_test(temp, G_FILE_TEST_EXISTS)) return temp;
    // try another one
    temp = l_printf(SDIR("%s")"lib%s_%s.0.so", folder, iface, network);
    if (g_file_test(temp, G_FILE_TEST_EXISTS)) return temp;
#ifdef WIN32
    // try another one (dll)
    temp = l_printf(SDIR("%s")"lib%s_%s.dll", folder, iface, network);
/*     g_log("UI", G_LOG_LEVEL_WARNING, */
/*           "Test for lib%s_%s.dll in folder "SDIR("%s"), iface, network, folder);     */
    if (g_file_test(temp, G_FILE_TEST_EXISTS)) return temp;
#endif

    return 0;
}

/**
 * Find a library in the library paths.
 * @param network the network
 * @return pathname of a shared library, or NULL if not found
 */
const char* ui_find_library_file(const char* network)
{
    if (!TheUI) return NULL;
    if (network)
    {
        const char* res = plugin_foreach_folder
            ((FolderSearchFunc)_check_for_ui, (gpointer)network);
        if (!res)
        {
            g_log("UI", G_LOG_LEVEL_WARNING,
                  "%s lib for '%s' not found",
                  TheUI->name, network);
        }
        return res;
    }
    return NULL;
}


static const char* _scan_for_uis(const char* folder)
{
    const char* name;
    GDir* dir;

    dir = g_dir_open(folder, 0, NULL);
    if (!dir) return NULL;

    while ((name = g_dir_read_name(dir)) != NULL)
    {
        GModule* handle;
        char* fname;
        gpointer symbol;
        char* pos;
        ui_entry_t* entry;
        unsigned len = strlen(name);

        if (strncmp(name, "lib", 3)) continue;
        pos = strchr(name+3, '_');
        if (!pos) continue;
        if (strncmp(pos, "_core", 5)) continue;
        
        if (!strcmp(name+(len-4), ".ilk")) continue;

#ifdef WIN32
        if (!strcmp(name+(len-4), ".ilk")) continue;
        if (!strcmp(name+(len-4), ".exp")) continue;
        if (!strcmp(name+(len-4), ".pdb")) continue;
        if (!strcmp(name+(len-4), ".lib")) continue;
#else
        // @Sid: this section should also work for win.
        if (!strstr(pos, ".so") && strcasecmp(name+(len-4), ".dll")) continue;
#endif

        fname = l_printf(SDIR("%s")"%s", folder, name);
        
        handle = lopster_dll_open(fname, G_MODULE_BIND_LAZY);
        if (!handle) 
        {
            g_log("UI", G_LOG_LEVEL_WARNING, "%s", g_module_error());
            continue;
        }
        
        if (g_module_symbol(handle, "ui_init", &symbol) &&
            g_module_symbol(handle, "ui_run", &symbol) &&
            g_module_symbol(handle, "ui_done", &symbol))
        {
            const char** pname;
            const char* iname;
            if (!g_module_symbol(handle, "ThePluginName", &symbol))
            {
                g_log("UI", G_LOG_LEVEL_WARNING,
                      "plugin %s has no symbol 'ThePluginName' defined", name);
                pname = NULL;
            }
            else
            {
                pname = symbol;
            }
            iname = auto_text(g_strndup(name+3, pos-name-3));
            
            entry = ui_entry_find(pname?*pname:NULL, iname);
            if (!entry)
            {
                entry = ui_entry_new(pname?*pname:NULL, iname, fname);
                AvailableUIs = g_list_append(AvailableUIs, entry);
            }
            else if (strlen(entry->filename) > strlen(fname))
            {
                g_free(entry->filename);
                entry->filename = g_strdup(fname);
                if (pname)
                {
                    g_free(entry->name);
                    entry->name = g_strdup(*pname);
                }
            }
        }
        
        lopster_dll_close(handle);
    }

    return NULL;
}

static void ui_find_available(void)
{
    static gboolean AlreadySearched = FALSE;
    GList* dlist;

    if (AlreadySearched) return;

    if (!g_module_supported())
    {
        g_log("UI", G_LOG_LEVEL_WARNING, "plugins not supported by system");
        return;
    }

    plugin_foreach_folder((FolderSearchFunc)_scan_for_uis, NULL);

    for (dlist = AvailableUIs; dlist; dlist = dlist->next)
    {
        ui_entry_t* entry = dlist->data;
        g_log("UI", G_LOG_LEVEL_INFO,
              "new UI found: [%s] %s", entry->name, entry->filename);
    }
    AlreadySearched = TRUE;
}

static void ui_run_daemon(void)
{
    if (!MainLoop) MainLoop = g_main_loop_new(NULL, FALSE);
    g_main_loop_run(MainLoop);
}

static void ui_quit_daemon(void)
{
    if (MainLoop)
    {
        g_main_loop_quit(MainLoop);
        g_main_loop_unref(MainLoop);
    }
}

#define CHECK_SYMBOL(sym, func)                                              \
    if (!(sym))                                                         \
    {                                                                   \
        g_log("UI", G_LOG_LEVEL_WARNING, "could not find symbol '%s'", #sym); \
        lopster_dll_close(handle);                                      \
        g_assert(FALSE);                                                \
        return;                                                   \
    } else func = symbol;

static void ui_load_syms(void)
{
    GModule* handle;
    gpointer symbol;
    
    UI.enter_main_loop = ui_run_daemon;
    UI.leave_main_loop = ui_quit_daemon;
    
    if (!g_module_supported())
    {
        g_log("UI", G_LOG_LEVEL_WARNING, "plugins not supported by system");
        return;
    }
    
    if (!TheUI)
    {
        g_log("UI", G_LOG_LEVEL_INFO, "starting deamon mode");
        return;
    }
    
    // RTLD_GLOBAL: allow plugin to export symbols, since the
    // user interface of other plugins might want to access
    handle = lopster_dll_open(TheUI->filename, G_MODULE_BIND_LAZY);
    if (!handle)
    {
        g_log("UI", G_LOG_LEVEL_WARNING, "%s", g_module_error());
        g_log("UI", G_LOG_LEVEL_DEBUG, "fallback to deamon mode");
        return;
    }
    
    CHECK_SYMBOL(g_module_symbol(handle, "ui_init", &symbol), UIModule->init);
    CHECK_SYMBOL(g_module_symbol(handle, "ui_run", &symbol), UIModule->run);
    CHECK_SYMBOL(g_module_symbol(handle, "ui_done", &symbol), UIModule->done);
}



static void xml_load_ui(xmlDocPtr xfile, xmlNodePtr xroot)
{
    xmlChar* attr = xmlGetProp(xroot, "Name");
    (void)xfile;
    ui_set_name(attr, FALSE);
    if (attr) xmlFree(attr);
}  

static void xml_save_ui(xmlDocPtr xfile, xmlNodePtr xroot)
{
    (void)xfile;
    xmlNewProp(xroot, "Name", TheUI?TheUI->iname:"");
}

void ui_set_name(const char* ui, gboolean save)
{
    ui_find_available();

    if (!UIModule) UICommandLine = TRUE;

    if (ui && *ui)
    {
        ui_entry_t* entry = ui_entry_find(ui, ui);
        if (!entry)
        {
            g_log("UI", G_LOG_LEVEL_WARNING, 
                  "cannot find UI '%s'", ui);
        }
        TheUI = entry;
        UIFailed = (entry==NULL);
    }
    else
    {
        TheUI = NULL;
        UIFailed = FALSE;
    }
    
    // only save if user changes it successfully.
    if (save && !UIFailed)
    {
        lopster_xml_save("core:ui", 0);
    }
}



void UI_enter_main_loop(void) {
    if (UI.enter_main_loop) UI.enter_main_loop();
}
void UI_leave_main_loop(void) {
    if (UI.leave_main_loop) UI.leave_main_loop();
}
void UI_download_show_details(download_t* download) {
    if (UI.download_show_details) UI.download_show_details(download);
}
void UI_file_tree_update_label(file_tree_t* tree , const char* str) {
    if (UI.file_tree_update_label) UI.file_tree_update_label(tree, str);
}
void UI_file_tree_set_progress(file_tree_t* tree, double prog) {
    if (UI.file_tree_set_progress) UI.file_tree_set_progress(tree, prog);
}
void UI_file_tree_set_progress_name(file_tree_t* tree, const char* str) {
    if (UI.file_tree_set_progress_name) UI.file_tree_set_progress_name(tree, str);
}
void UI_file_tree_set_abort(file_tree_t* tree, GFunc func, void* data) {
    if (UI.file_tree_set_abort) UI.file_tree_set_abort(tree, func, data);
}
void UI_lib_working(gboolean val) {
    if (UI.lib_working) UI.lib_working(val);
}
void UI_chat_switch_page(chat_page_t* page) {
    if (UI.chat_switch_page) UI.chat_switch_page(page);
}
gboolean UI_chat_page_printable(chat_page_t* page) {
    if (UI.chat_page_printable) return UI.chat_page_printable(page);
    else return TRUE;
}
void UI_chat_create_simple_page(chat_page_t* page) {
    if (UI.chat_create_simple_page) UI.chat_create_simple_page(page);
}
void UI_chat_print(chat_page_t* page, int index, va_list ap) {
    if (UI.chat_print) UI.chat_print(page, index, ap);
}
void UI_threads_show(const char* message) {
    if (UI.threads_show) UI.threads_show(message);
}

guint UI_idle_add(GSourceFunc function, gpointer data)
{
    if (UI.idle_timeout_add)
    {
        return UI.idle_timeout_add(function, data);
    }
    else
    {
        g_log("UI", G_LOG_LEVEL_WARNING,
              "idle function not set by UI");
        function(data);
        return 0;
    }
}

guint UI_timeout_add(guint interval, GSourceFunc function, gpointer data)
{
    if (UI.timeout_add) 
    {
        return UI.timeout_add(interval, function, data);
    }
    else
    {
        g_log("UI", G_LOG_LEVEL_WARNING,
              "timeout function not set by UI");
        return 0;
    }
}

void UI_timeout_remove(guint id)
{
    if (UI.timeout_remove) UI.timeout_remove(id);
}

void UI_idle_remove(guint id)
{
    if (UI.idle_timeout_remove) UI.idle_timeout_remove(id);
}

static void ui_init()
{
    if (UIModule->init)
        UIModule->init(&UI);

    if (!UI.idle_timeout_add) 
        g_log("UI", G_LOG_LEVEL_WARNING, "idle_timeout_add() no defined by UI");
    if (!UI.idle_timeout_remove) 
        g_log("UI", G_LOG_LEVEL_WARNING, "idle_timeout_remove() no defined by UI");
    if (!UI.timeout_add) 
        g_log("UI", G_LOG_LEVEL_WARNING, "timeout_add() no defined by UI");
    if (!UI.timeout_remove) 
        g_log("UI", G_LOG_LEVEL_WARNING, "idle_timeout_remove() no defined by UI");
}
static void ui_run()
{
    if (UIModule->run) UIModule->run();
}
static void ui_done()
{
    if (UIModule->done) UIModule->done();

    g_list_foreach(AvailableUIs, (GFunc)ui_entry_destroy, NULL);
    AvailableUIs = NULL;
}

static void ui_module_init(ui_module_t* module)
{
    module_init(MODULE(module), "User Interface Module",
                (ModuleFunc)ui_init, (ModuleFunc)ui_run, (ModuleFunc)ui_done);
    module->init = NULL;
    module->run = NULL;
    module->done = NULL;

    module_require(xml_module_instance);
    lopster_xml_register_home
        ("core:ui", "ui.xml", "UserInterface", xml_load_ui, xml_save_ui);

    ui_find_available();

    // if UI was not set successfully on command line,
    // choose now the right one.
    if (!UICommandLine || UIFailed)
    {
        if (UICommandLine)
            g_log("UI", G_LOG_LEVEL_WARNING,
                  "ignoring command line UI argument");

        if (!lopster_xml_load("core:ui") || UIFailed)
        {
            if (UIFailed)
                g_log("UI", G_LOG_LEVEL_WARNING,
                      "ignoring saved UI setting");

            // could not load/set saved setting, fallback
            // to random UI
            if (AvailableUIs->data)
            {
                ui_entry_t* entry = AvailableUIs->data;
                g_log("UI", G_LOG_LEVEL_INFO,
                      "using first available UI '%s'", entry->name);
                ui_set_name(entry->name, TRUE);
            }
            else
            {
                g_log("UI", G_LOG_LEVEL_INFO,
                      "no UI available -> running in deamon mode");
            }
        }
    }
    module_require(dynamic_module_instance);
    
    ui_load_syms();
    
    // require some other modules
    //FIXME: should we really require them all here??
    module_require(download_module_instance);
    module_require(upload_module_instance);
    module_require(library_module_instance);
    module_require(legacy_module_instance);
}

module_t* ui_module_instance()
{
    if (!UIModule)
    {
        UIModule = g_malloc0(sizeof(*UIModule));
        ui_module_init(UIModule);
    }
    return MODULE(UIModule);
}

#ifdef WIN32
#define _UI_C_
#include "lcorealiasdef.c"
#endif
