/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.16 $
 * $Date: 2007/11/21 20:18:24 $
 * $Author: sgop $
 */

#include <string.h>

#include "upload.h"
#include "preferences.h"
#include "l_limit.h"
#include "ui.h"
#include "action_item.h"
#include "plugin.h"
#include "l_base.h"
#include "l_timer.h"
#include "l_i18n.h"
#include "l_io.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
  GList* remote;
  GList* queue;
  GList* active;
  GList* uploads;
  gboolean queue_check;
  volume_t* bandwidth;
  GList* limits;
  rate_t* rate;
} upload_module_t;

static upload_module_t* UploadModule = NULL;


static void upload_sort_in(upload_t* upload);
static void upload_sort_out(upload_t* upload);

static void upload_limit_changed(void) {
  UploadModule->queue_check = TRUE;
}

/** This function normally is not used by plugins.  Plugins should use
 * socket_done() instead.
 * @param upload the upload to delete.
 */
void upload_delete(upload_t* upload) {
  g_assert(UploadModule);
  socket_done(SOCK(upload), ST_UL_DELETE, "User deleted");
}

static action_item_t* on_get_protocol_user_menu(GList* selection) {
  GList* dlist;
  action_item_t* result = NULL;

  for (dlist = selection; dlist; dlist = dlist->next) {
    upload_t* upload = dlist->data;
    action_item_t* item = user_action_upload_find(upload->user->ID);
    if (!item) return NULL;
    if (!result) result = item;
    else if (result != item) return NULL;
  }
  return result;
}

static void on_upload_delete(GList* uploads) {
  GList* dlist;
  for (dlist = uploads; dlist; dlist = dlist->next) {
    upload_t* upload = dlist->data;
    upload_delete(upload);
  }
}

static void upload_actions_init(void) {
  action_item_t* item;

  item = action_item_root("Upload actions");
  action_item_simple(N_("Delete upload"), item, 0,
                     (ActionItemFunc)on_upload_delete);
  action_item_sub(N_("User menu"), item,
                  (ActionItemLink)on_get_protocol_user_menu,
                  FALSE);
}

/** Get the number of overall upload slots.
 * @return the number of slots
 */
int uploads_slots_number(void) {
  g_assert(UploadModule);
  return limit_get_limit(limit_find("Upload Limit"));
}

/** Get the number of free upload slots.
 * @return the number of free slots
 * @note the result can be lower than zero if the limit is not only
 * hit but exceeded
 */ 
int uploads_slots_free(void) {
  limit_t* limit = limit_find("Upload Limit");
  g_assert(UploadModule);
  return limit_get_limit(limit)-limit_get_used(limit);
}

/** Get the number of files in queue.
 * @return the number of queued files.
 */
unsigned uploads_queue_size(void) {
  g_assert(UploadModule);
  return g_list_length(UploadModule->queue);
}

void upload_add_limit_real(upload_t* upload, limit_t* limit)
{
    upload->limits = g_list_append(upload->limits, limit);
    if (!object_monitor_find(OBJ(limit), (ObjectFunc)upload_limit_changed, NULL))
        object_monitor_d(OBJ(limit), (ObjectFunc)upload_limit_changed, NULL);
}
/** Add an upload limit.
 * @param upload the upload
 * @param name the name of the limit
 * @see limit_new()
 */
void upload_add_limit(upload_t* upload, const char* name) {
  limit_t* limit = limit_find(name);
  g_assert(UploadModule);
  if (!limit) {
    g_log("UPLOAD", G_LOG_LEVEL_WARNING, "limit '%s' not found", name);
    return;
  }
  upload_add_limit_real(upload, limit);
}

static int upload_priority_compare(upload_t* u1, upload_t* u2) {
  if (u1->priority < u2->priority) return 1;  // u2 first
  if (u1->priority > u2->priority) return -1; // u1 first
  if (u1->request_time < u2->request_time) return -1; // u1 first
  return (u1->request_time > u2->request_time);
}

static void upload_sort_in(upload_t* upload) {
  if (upload->status == UPLOAD_QUEUED) {
    UploadModule->queue = g_list_insert_sorted
      (UploadModule->queue, upload, (GCompareFunc)upload_priority_compare);
  } else if (upload->status == UPLOAD_REMOTE) {
    UploadModule->remote = g_list_prepend(UploadModule->remote, upload);
  } else if (upload->status == UPLOAD_STARTED) {
    UploadModule->active = g_list_prepend(UploadModule->active, upload);
    socket_set_timeout(SOCK(upload), 30);
    limits_increase(upload->limits);
  }
}

static void upload_sort_out(upload_t* upload) {
  if (upload->status == UPLOAD_QUEUED) {
    UploadModule->queue = g_list_remove(UploadModule->queue, upload);
  } else if (upload->status == UPLOAD_REMOTE) {
    UploadModule->remote = g_list_remove(UploadModule->remote, upload);
  } else if (upload->status == UPLOAD_STARTED) {
    UploadModule->active = g_list_remove(UploadModule->active, upload);
    socket_set_timeout(SOCK(upload), 0);
    UploadModule->queue_check = TRUE;
    limits_decrease(upload->limits);
  }
}

static void upload_set_status(upload_t* upload, int status) {
  if (upload->status == status) return;
  upload_sort_out(upload);
  upload->status = status;
  upload_sort_in(upload);
}

gboolean upload_end(upload_t* upload, const char* message) {

  g_assert(UploadModule);
  upload_set_status(upload, UPLOAD_UNKNOWN);

  if (upload->fd) fclose(upload->fd);
  upload->fd = NULL;
  segments_destroy(upload->segments);
  upload->segments = NULL;
  upload->cur_segment = NULL;

  tcp_socket_end(TSOCK(upload), message);
  return TRUE;
}

void upload_destroy(upload_t* upload) {

  g_assert(UploadModule);
  UploadModule->uploads = g_list_remove(UploadModule->uploads, upload);
  upload_sort_out(upload);

  g_assert(segments_count(upload->segments) == 0);

  segments_destroy(upload->finished_segments);
  upload->finished_segments = NULL;

  g_assert(!upload->fd);

  user_unref(upload->user);

  object_kill(OBJ(upload->file));
  g_list_free(upload->limits);
  
  tcp_socket_destroy(TSOCK(upload));
}

static segment_t* upload_select_segment(upload_t* upload) {
  segment_t* segment = segments_get_first(upload->segments);
  if (segment) {
    /* FIXME 
    socket_rate_set_todo
      (SOCK(upload), DIR_OUT, segment->stop - segment->start);
    */
  }
  return segment;
}
static void on_upload_destroyed(upload_t* upload, user_t* user) {
  user->uploads = g_list_remove(user->uploads, upload);
}

static void upload_set_user(upload_t* upload, user_t* user)
{
    const char* name = l_printf
        ("u:%s:%s:ulimit", plugin_get_short_name(plugin_get(user->ID)), user->print(user));
    limit_t* limit = limit_find(name);
    
    if (!limit) limit = limit_new(name, "/core/upload/limit_per_user");
    upload_add_limit_real(upload, limit);
    
    upload->user = user_ref(user);
    user->uploads = g_list_append(user->uploads, upload);
    object_notify_o(OBJ(upload), (ObjectFunc)on_upload_destroyed, OBJ(user));
}

static gboolean upload_prepare_out(upload_t* upload);
void upload_init(upload_t* upload, file_t* file, user_t* user) {
  g_assert(UploadModule);
  
  tcp_socket_init(TSOCK(upload), user->print?user->print(user):"Upload", 0);
  object_add_type(OBJ(upload), "Upload");
  OBJ(upload)->destroy = (ObjectFunc)upload_destroy;
  socket_set_end(SOCK(upload), (SockEndFunc)upload_end);
  b_socket_set_prepare_out(BSOCK(upload), (BPrepareFunc)upload_prepare_out);

  upload->file = file_clone(file, NULL);
  upload->limits = NULL;
  upload->segments = segments_new();
  upload->fd = NULL;
  upload->priority = UPLOAD_PRI_NORMAL;
  upload->base_priority = UPLOAD_PRI_NORMAL;
  upload->request_time = CurrentTime;
  upload->status = UPLOAD_UNKNOWN;
  upload->select = (UploadSegSelFunc)upload_select_segment;
  upload->requested = FALSE;

  upload_set_user(upload, user);

  UploadModule->uploads = g_list_append(UploadModule->uploads, upload);

  // add global limit
  upload_add_limit(upload, "Upload Limit");
  // add global bandwidth limit
  socket_add_volume(SOCK(upload), "Upload Bandwidth Limit");
  // register bandwidth load
  socket_add_load(SOCK(upload), "Transfer");
  // monitor the rate
  socket_add_default_rate(SOCK(upload), DIR_OUT);
  socket_add_rate(SOCK(upload), UploadModule->rate);

}

/*
static upload_t* upload_new(file_t* file) {
  upload_t* upload;

  upload = g_malloc0(sizeof(*upload));
  upload_init(upload, file);
  return upload;
}
*/

/** Attach a new segment to the upload.  These segments are used to
 * upload.
 * @param upload the upload.
 * @param start the segment start position (inclusive)
 * @param stop the segment stop position (exclusive)
 * @param begin TRUE to prepend to the list of segments, FALSE otherwise
 * @return the appended segment, or NULL on error (if start >= stop)
 */
segment_t* upload_add_segment(upload_t* upload, file_size_t start,
                              file_size_t stop, gboolean begin) {
  segment_t* segment;
  g_assert(UploadModule);
  if (start >= stop) return NULL;
  segment = segment_new(start, stop, 0, SEG_ERR_NONE);
  if (begin) segments_prepend(upload->segments, segment);
  else segments_append(upload->segments, segment);
  return segment;
}

/** Remove a segment. The segment doesn't need to be one of the whole
 * attached segments. A part from a segment may be cut, but you cannot
 * cut from more than one segments.
 * @param upload the upload
 * @param start the start position (inclusive)
 * @param stop the stop position (exclusive)
 */
void upload_remove_segment(upload_t* upload, file_size_t start, file_size_t stop) {
  g_assert(UploadModule);
  segment_destroy(segments_cut(upload->segments, start, stop));
}

static void upload_calc_priority(upload_t* upload) {
  // for now simply copy the base priority
  upload->priority = upload->base_priority;
}

/** Change the bas priority of uploads.
 * @param upload the upload.
 * @param priority the new base priority
 */
void upload_change_base_priority(upload_t* upload, double priority) {
  g_assert(UploadModule);

  if (upload->status == UPLOAD_QUEUED) upload_sort_out(upload);
  upload->base_priority = priority;
  upload_calc_priority(upload);
  if (upload->status == UPLOAD_QUEUED) upload_sort_in(upload);
  UploadModule->queue_check = TRUE;
}

static gboolean upload_allowed(upload_t* upload) {
  // allow urgent upload always.
  if (upload->priority >= UPLOAD_PRI_URGENT) return TRUE;

  if (limits_reached(upload->limits)) return FALSE;
  return TRUE;
}

static void upload_queue(upload_t* upload) {
  int index;
  
  if (upload->status == UPLOAD_STARTED) {
    g_log("UPLOAD", G_LOG_LEVEL_WARNING,
          "cannot queue already started upload '%s'", OBJ(upload)->name);
    return;
  }
  
  upload_set_status(upload, UPLOAD_QUEUED);

  // debug
  index = g_list_index(UploadModule->queue, upload);
  g_assert(index >= 0);
  // debug end

  upload->queue(upload, index+1, uploads_slots_number());

}

static void upload_start_real(upload_t* upload) {

  if (upload->status == UPLOAD_STARTED) {
    g_log("UPLOAD", G_LOG_LEVEL_WARNING,
          "upload '%s' already started", OBJ(upload)->name);
    return;
  }

  upload_set_status(upload, UPLOAD_STARTED);

  if (!upload->start(upload)) {
    socket_done(SOCK(upload), ST_UL_START_ERR, "start error");
  }
}

/** Use this function to notify that the upload was requested again.
 * @param upload the upload
 */
void upload_renew(upload_t* upload) {
  g_assert(UploadModule);
  if (upload->status == UPLOAD_STARTED) return;
  if (upload->status != UPLOAD_QUEUED) upload_set_status(upload, UPLOAD_QUEUED);
  upload->requested = TRUE;
  UploadModule->queue_check = TRUE;
}

/** Use this function to put an upload on queue.
 * @param upload the upload
 */
void upload_wake_up(upload_t* upload) {
  g_assert(UploadModule);
  if (upload->status != UPLOAD_REMOTE) return;
  upload_set_status(upload, UPLOAD_QUEUED);
  UploadModule->queue_check = TRUE;
}

/** Use this function to insert a new upload.
 * @param upload the upload
 * @param start whether to start the upload.
 * @note if start is FALSE then you have to use
 * upload_set_remote(upload, FALSE) later
 */
void upload_insert(upload_t* upload, gboolean start)
{
    g_assert(UploadModule);
    upload_calc_priority(upload);
    if (start) upload_renew(upload);
    else upload_set_status(upload, UPLOAD_REMOTE);
    object_make_public(OBJ(upload), NULL);
}

static gboolean upload_prepare_out(upload_t* upload)
{
    b_socket_t* s = BSOCK(upload);
    buffer_t* buffer = b_socket_get_out_buffer(s, 0);
    segment_t* segment = upload->cur_segment;
    unsigned avail;
    file_size_t seg_remain;
    
    
    // free some buffer space if it was almost completely sent
    if (buffer->max - buffer->consumed <= 1024)
        buffer_consume(buffer, 0);
    
    if (!segment) return TRUE;
    
    // lets fill the buffer, if there's room
    avail = buffer_space(buffer);
    if (!avail) return TRUE;
    
    // NOTE: segment->size can be greater than the segment
    seg_remain = segment->stop - segment->start - segment->size;
    if (seg_remain < 0)
    {
        // ops, we've already sent to much...
        // FIXME: we should try to remove the according data from
        // the buffer if possible
        b_socket_close_when_sent(BSOCK(upload), ST_DONE);
    }
    else if (seg_remain == 0)
    {
        const char* res;
        // move the segment to the finished list
        segment = segments_cut(upload->segments, segment->start, segment->stop);
        g_assert(segment->size == segment->stop-segment->start);
        segments_insert(upload->finished_segments, segment->start, 
                        segment->stop, segment->size, segment->error);
        upload->cur_segment = NULL;
        segment_destroy(segment);
        
        res = upload_prepare_segment(upload);
        if (res)
        {
            socket_suspend(SOCK(upload), DIR_OUT);
            socket_done(SOCK(upload), ST_ERROR, res);
            return FALSE;
        }
        else if (!upload->cur_segment)
        {
            b_socket_close_when_sent(BSOCK(upload), ST_UL_SEG_FIN);
        }
    }
    else if (!upload->fd)
    {
        g_log("UPLOAD", G_LOG_LEVEL_WARNING,
              "upload '%s' has no open fd", upload->file->filename);
        return TRUE; //FIXME: return FALSE?
    }
    else
    {
        int n;
        if (avail > seg_remain) avail = seg_remain;
        n = fread(buffer_write_pos(buffer), 1, avail, upload->fd);
        if (n <= 0)
        {
            g_log("UPLOAD", G_LOG_LEVEL_WARNING,
                  "could not read from '%s' at pos "FF,
                  upload->file->localname, segment->start+segment->size);
            socket_done(SOCK(upload), ST_UL_IO_ERR, "file read error");
            return FALSE;
        }
        buffer->size += n;
        segment->size += n;
    }
    return TRUE;
}

/** Prepare a segment for upload. The select function of the upload is
 * called to determine the next segment to upload. Then the local file
 * is prepared for upload and the header function of the upload is
 * called.
 * @param upload the upload
 * @return a string if an error has occurred, NULL on success.
 * @note this function is normally only used once at the beginning of
 * upload. All subsequent segments are prepared automatically.
 */
const char* upload_prepare_segment(upload_t* upload) {
  file_t* file;

  g_assert(UploadModule);
  // return if we are already sending
  if (upload->cur_segment) return NULL;

  // socket must be connected
  g_assert(socket_connected(SOCK(upload)));
  upload->cur_segment = upload->select(upload);
  if (!upload->cur_segment) return NULL;

  file = upload->file;
  if (!upload->fd) {
    char* temp = locale_from_utf8(file->localname);
    upload->fd = fopen(temp, "rb");
    g_free(temp);
  }
    
  if (!upload->fd) {
    g_log("UPLOAD", G_LOG_LEVEL_WARNING,
          "could not open file '%s'", file->localname);
    socket_set_status(SOCK(upload), ST_UL_IO_ERR);
    return "could not open upload file";
  }

  /*
  g_log("UPLOAD", G_LOG_LEVEL_DEBUG,
	"seeking to %llu in file '%s'",
	upload->cur_segment->start, file->localname);
  */
  // FIXME: we could check whether we are already at the correct file position
  if (FSEEK(upload->fd, upload->cur_segment->start, SEEK_SET) == -1)
  {
      g_log("UPLOAD", G_LOG_LEVEL_WARNING, "could not seek to "FF" in file '%s'",
            upload->cur_segment->start, file->localname);
      socket_set_status(SOCK(upload), ST_UL_IO_ERR);
      return "could seek in upload file";
  }

  /* FIXME: check buffer data first..
  {
    buffer_t* buffer = b_socket_get_out_buffer(BSOCK(upload), 8192);
    if (buffer_avail(buffer) > 8192) buffer_resize(buffer, 8192);
  }
  */
  // make sure we have a reasonable outbuffer size
  b_socket_get_out_buffer(BSOCK(upload), 10240);  // 10k
  upload->cur_segment->size = 0;
  
  /*
  g_log("UPLOAD", G_LOG_LEVEL_DEBUG,
	"start sending file '%s' [%llu-%llu=%llu]", 
	upload->file->filename,
	upload->cur_segment->start, upload->cur_segment->stop,
	upload->cur_segment->stop - upload->cur_segment->start);
  */

  if (upload->header && !upload->header(upload))
    return "error sending header";

  return NULL;
}

/** Change the stop position of the current segment.
 * @param upload the upload
 * @param stop the new stop position (exclusive)
 * @return TRUE on success, FALSE otherwise
 */
gboolean upload_change_segment_stop(upload_t* upload, file_size_t stop) {
  segment_t* segment = upload->cur_segment;
  if (!segment) return FALSE;

  // we have already sent more bytes, failed!
  if (stop < segment->start + segment->size) return FALSE;
  segment->stop = stop;
  /* FIXME
  socket_rate_set_todo
    (SOCK(upload), DIR_OUT, segment->stop - segment->start);
  */
  return TRUE;
}

static gboolean uploads_check(void) {
  unsigned cnt = 0;
  GList* dlist;

  g_assert(UploadModule);
  
  if (UploadModule->queue_check && UploadModule->queue) {
    for (dlist = UploadModule->queue; dlist; ) {
      upload_t* upload = dlist->data;
      dlist = dlist->next;
      if (upload->status != UPLOAD_REMOTE && upload_allowed(upload)) {
        upload_start_real(upload);
        cnt++;
      } else if (upload->requested) {
        g_assert(upload->status == UPLOAD_QUEUED);
        upload_queue(upload);
      }
      upload->requested = FALSE;
    }
  }

  // if not upload was allowed, disable check.
  if (!cnt) UploadModule->queue_check = FALSE;

  return TRUE;
}

/** Iterate through upload.
 * @param func the callback
 * @param data the user data for the callback
 */
void uploads_foreach(ObjectFunc func, void* data) {
  g_assert(func);
  g_assert(UploadModule);
  g_list_foreach(UploadModule->uploads, (GFunc)func, data);
}

/** Find an upload from a user.
 * @param user the user requesting a file.
 * @param winname the windows filename to search for.
 * @return the upload of that user/file if found, NULL otherwise.
 * @note before creating a new upload you always should search for the
 * already existing upload to avoid duplicates
 */
upload_t* upload_find(user_t* user, const char* winname) {
  GList* dlist;
  for (dlist = user->uploads; dlist; dlist = dlist->next) {
    upload_t* upload = dlist->data;
    g_assert(upload->user == user);
    if (!strcmp(upload->file->netname, winname)) return upload;
  }
  return NULL;
}

/** Get the next upload from a user.
 * @param upload the current upload.
 * @return the next upload.
 */
upload_t* upload_get_next_from_user(upload_t* upload) {
  GList* dlist;
  for (dlist = upload->user->uploads; dlist; dlist = dlist->next) {
    upload_t* next = dlist->data;
    if (next != upload) return upload;
  }
  return NULL;
}

/** Get the rate object for all uploads. Plugins are normally not
 * interested in the rate of the uploads.  this function.
 * @return the rate object
 */
rate_t* uploads_get_rate() {
  g_assert(UploadModule);
  return UploadModule->rate;
}


static void uploads_init() {
  pref_register_int   ("/core/upload/limit", TRUE, 3);
  pref_add_constraint_min("/core/upload/limit", 0);
  pref_register_int   ("/core/upload/limit_per_user", TRUE, 1);
  pref_add_constraint_min("/core/upload/limit_per_user", 1);
  //  pref_register_int   ("/core/upload/hard_limit", TRUE, 50);
  //  pref_add_constraint_min("/core/upload/hard_limit", 0);
  pref_register_int   ("/core/upload/bandlimit", TRUE, 1024*14);
  pref_add_constraint_minmax("/core/upload/bandlimit", 0, 1<<30);
  
  UploadModule->rate = rate_new(0);

  upload_actions_init();
}
static void uploads_run() {
  // register limits
  limit_new("Upload Limit", "/core/upload/limit");
  // register bandwidth limits
  volume_new("Upload Bandwidth Limit", "/core/upload/bandlimit");

  lopster_call_periodic(1000, (CallFunc)uploads_check, NULL);
}

static void upload_module_init(upload_module_t* module) {
  module_init(MODULE(module), "Upload Module",
              (ModuleFunc)uploads_init,
              (ModuleFunc)uploads_run, NULL);

  module->remote = NULL;
  module->queue  = NULL;
  module->active = NULL;
  module->uploads = NULL;
  module->queue_check = TRUE;
  
  module_require(pref_module_instance);
  module_require(lsocket_module_instance);
  module_require(tsocket_module_instance);
  module_require(limit_module_instance);
  module_require(base_module_instance);
  
}


/** Require this module if you use uploads.
 * @return the upload module.
 * @see module_require()
 */
module_t* upload_module_instance() {
  if (!UploadModule) {
    UploadModule = g_malloc0(sizeof(*UploadModule));
    upload_module_init(UploadModule);
  }
  return MODULE(UploadModule);
}

#ifdef WIN32
#define _UPLOAD_C_
#include "lcorealiasdef.c"
#endif
