/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.7 $
 * $Date: 2006/07/22 17:11:48 $
 * $Author: sid4u $
 */

#ifndef _UPLOAD_H_
#define _UPLOAD_H_

#include <stdio.h>

typedef struct _upload_t upload_t;

#include "tcp_socket.h"
#include "file_segment.h"
#include "files.h"
#include "action_item.h"
#include "user.h"
#include "module.h"

G_BEGIN_DECLS

/** \defgroup UploadModule Uploads in lopster
 * @ingroup FileSharingFramework
 */
/*@{*/

/** @name Basic upload priorities */
/*@{*/
#define UPLOAD_PRI_NULL      0.0    ///< quite no priority
#define UPLOAD_PRI_LOW       1.0    ///< low priority
#define UPLOAD_PRI_NORMAL    2.0    ///< normal priority
#define UPLOAD_PRI_HIGH      3.0    ///< high priority
#define UPLOAD_PRI_URGENT    4.0    ///< quite urgent (always start)
/*@}*/

// hide these from the user?
#define UPLOAD_UNKNOWN     0
#define UPLOAD_QUEUED      1
#define UPLOAD_REMOTE      2
#define UPLOAD_STARTED     3

/** @name Upload status */
/*@{*/
#define ST_UL_IO_ERR        300  ///< IO error, normally error reading local file
#define ST_UL_START_ERR     301  ///< start error, if the start function fails
#define ST_UL_SEG_FIN       302  ///< segment finished, this happens if subsequent calls to upload_prepare_segment() gives no new segment
#define ST_UL_DELETE        303  ///< status if upload gets user deleted.
/*@}*/


#define UPLOAD(arg) ((upload_t*)(arg))

/** Function that is called when the uploads gets locally queued.
 * @param upload the upload
 * @param pos the queue position
 * @param slots the overall slots.
 */
typedef void (*UploadQueueFunc)(upload_t* upload, int pos, int slots);
/** Function that is called when the uploads got un-queued 
 * @param upload the upload
 * @return FALSE on error, TRUE if started
 */
typedef gboolean (*UploadStartFunc)(upload_t* upload);
/** Function that is called right before segment is ready to send 
 * @param upload the upload
 * @return FALSE on error, TRUE otherwise
 */
typedef gboolean (*UploadHeaderFunc)(upload_t* upload);
/** Function to select the next segment for the upload.
 * @param upload the upload
 * @return a new segment, or NULL if no segment can be found.
 */
typedef segment_t* (*UploadSegSelFunc)(upload_t* upload);

#define UPLOAD_HISTORY  30

/** The upload class. */
struct _upload_t {
  tcp_socket_t _socket;      ///< the base class
  file_t* file;              ///< the file that gets uploaded.
  user_t* user;              ///< the requester of the file
  GList* limits;
  segments_t* finished_segments;
  segments_t* segments;      ///< the segment that are attached to this upload, see upload_add_segment()
  segment_t* cur_segment;    ///< the current segment, see upload_prepare_segment()
  gboolean requested;

  FILE* fd;
  double priority;           // dont change
  double base_priority;      // only change manually before insertion
  time_t request_time;
  UploadQueueFunc queue;     ///< the queue function, set if you need to take action if the upload gets queued
  UploadStartFunc start;     ///< the start function, normally is set by plugins
  UploadHeaderFunc header;   ///< the header function, set this function if you need to send some header before the transfer starts.
  UploadSegSelFunc select;   ///< the segment selection function. This function selects a segment from the attached segments. Set this function if you don't like to sequentially process the segments.
  int status;                ///< status of the upload, rarely used (make private?)
};

/** @name Basic upload functionality */
/*@{*/
module_t* upload_module_instance();
int uploads_slots_number(void);
int uploads_slots_free(void);
unsigned uploads_queue_size(void);
upload_t* upload_find(user_t* user, const char* winname);
upload_t* upload_get_next_from_user(upload_t* upload);
/*@}*/

/** @name Creating and destroying uploads */
/*@{*/
void upload_init(upload_t* upload, file_t* file, user_t* user);
gboolean upload_end(upload_t* upload, const char* message);
void upload_destroy(upload_t* upload);
/*@}*/

/** @name Basic upload control */
/*@{*/
void upload_insert(upload_t* upload, gboolean start);
void upload_renew(upload_t* upload);
void upload_wake_up(upload_t* upload);
segment_t* upload_add_segment(upload_t* upload, file_size_t start, file_size_t stop, gboolean begin);
const char* upload_prepare_segment(upload_t* upload);
/*@}*/

/** @name Advanced upload control */
/*@{*/
void upload_remove_segment(upload_t* upload, file_size_t start, file_size_t stop);
gboolean upload_change_segment_stop(upload_t* upload, file_size_t stop);
/*@}*/

/** @name Functions not used by plugins */
/*@{*/
void upload_change_base_priority(upload_t* upload, double priority);
void upload_delete(upload_t* upload);
rate_t* uploads_get_rate();
void uploads_foreach(ObjectFunc func, void* data);
/*@}*/


/*@}*/

G_END_DECLS
#endif
