/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.7 $
 * $Date: 2006/07/22 17:11:48 $
 * $Author: sid4u $
 */

#include "user.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  module_t _module;
  GList* monitored;
} user_module_t;

static user_module_t* UserModule = NULL;

/** Destroy a user. Call from derives destroy only.
 * @param user the user to destroy.
 */
void user_destroy(user_t* user)
{
    if (user->refs)
    {
        g_message("never use object_kill() to destroy a user, use user_unref() instead");
        return;
    }
    g_free(user->sources);
    g_free(user->uploads);
    object_destroy(OBJ(user));
}

static const char* user_print(user_t* user)
{
    return OBJ(user)->name;
}

void user_init(user_t* user, const char* name, int ID)
{
    object_init(OBJ(user), name, TRUE);
    object_add_type(OBJ(user), "User");
    OBJ(user)->destroy = (ObjectFunc)user_destroy;
    user->print = (UserPrint)user_print;
    user->ip = 0;
    user->port = 0;
    user->ID = ID;
    user->firewalled = FALSE;
    user->uploads = NULL;
    user->sources = NULL;
    user->refs = 0;
}

user_t* user_new(const char* name, int ID)
{
    user_t* user = g_malloc0(sizeof(*user));
    user_init(user, name, ID);
    return user;
}

/** Whenever you link to the user you have to call this function
 * @param user the user
 * @return the user as generic pointer
 */
void* user_ref(user_t* user)
{
    user->refs++;
    return user;
}

/** When you no longer link to a user, call this function. If noone else links
 * to this user, it is destroyed.
 * @param user the user as generic pointer
 * @return the user, or NULL if it was destroyed.
 */
void* user_unref(user_t* user)
{
    g_assert(user->refs);
    user->refs--;
    if (!user->refs)
    {
        object_kill(OBJ(user));
        return NULL;
    }
    else
    {
        return user;
    }
}

ip_t user_ip(void* user)
{
    return USER(user)->ip;
}

port_t user_port(void* user)
{
    return USER(user)->port;
}

void user_set_ip(void* user, ip_t ip)
{
    if (USER(user)->ip == ip) return;
    USER(user)->ip = ip;
    object_set_dirty(OBJ(user));
}

void user_set_port(void* user, port_t port)
{
    if (USER(user)->port == port) return;
    USER(user)->port = port;
    object_set_dirty(OBJ(user));
}

void user_set_ip_port(void* user, ip_t ip, port_t port)
{
    if (USER(user)->ip == ip && USER(user)->port == port) return;
    USER(user)->ip = ip;
    USER(user)->port = port;
    object_set_dirty(OBJ(user));
}

gboolean user_get_firewalled(user_t* user)
{
    return user->firewalled;
}

void user_set_firewalled(user_t* user, gboolean wall)
{
    if (user->firewalled == wall) return;
    user->firewalled = wall;
    object_set_dirty(OBJ(user));
}

void user_start_monitor(user_t* user)
{
    g_assert(UserModule);
    if (!user->mon_cnt)
    {
        user_ref(user);
        UserModule->monitored = g_list_append(UserModule->monitored, user);
        if (user->monitor) user->monitor(user);
    }
    user->mon_cnt++;
}

void user_stop_monitor(user_t* user)
{
    g_assert(UserModule);
    user->mon_cnt--;
    if (!user->mon_cnt)
    {
        if (user->unmonitor) user->unmonitor(user);
        UserModule->monitored = g_list_remove(UserModule->monitored, user);
        user_unref(user);
    }
}

void user_set_online(user_t* user, gboolean setting)
{
    if (user->online == setting) return;
    user->online = setting;
    object_set_dirty(OBJ(user));
}

int user_get_monitor(user_t* user)
{
    return user->mon_cnt;
}

void users_foreach_monitored(int ID, UserForeach func, void* data)
{
    GList* dlist;
    
    g_assert(UserModule);
    for (dlist = UserModule->monitored; dlist; dlist = dlist->next)
    {
        user_t* user = dlist->data;
        g_assert(user->mon_cnt > 0);
        if (user->ID != ID) continue;
        if (func(user, data)) return;
    }
}

static void user_module_init(user_module_t* module)
{
    module_init(MODULE(module), "User Module", NULL, NULL, NULL);
    module->monitored = NULL;
}

module_t* user_module_instance()
{
    if (!UserModule)
    {
        UserModule = g_malloc0(sizeof(*UserModule));
        user_module_init(UserModule);
    }
    return MODULE(UserModule);
}

#ifdef WIN32
#define _USER_C_
#include "lcorealiasdef.c"
#endif
