/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.6 $
 * $Date: 2006/07/22 17:11:48 $
 * $Author: sid4u $
 */

#ifndef _USER_H_
#define _USER_H_

typedef struct _user_t user_t;

#include "object.h"
#include "files.h"
#include "download.h"
#include "module.h"

G_BEGIN_DECLS

/** \defgroup UserModule Base class for users
 * In Lopster users are persistent objects which means that two users
 * are that same if and only if the pointers to the user object are
 * equal. Whenever you hold a pointer to a user you have to add a
 * reference to it using user_ref(). When you release a user, call
 * user_unref(). The user automatically is destroyed if the reference
 * counter drops to 0.
 *
 * Users can be monitored, which means that you get informed when the
 * user get online or offline. If someone start monitoring a user, its
 * monitor member function is called. If the last stops monitoring the
 * user, the unmonitor member function is called. The @ref BuddyModule
 * for example monitores users, another application is private chat
 * pages, which track users.
 * 
 * The un/monitor member function should execute network specific
 * stuff to un/monitor a user.
 *
 * @ingroup FileSharingFramework
 */
/*@{*/

/** User print function */
typedef const char* (*UserPrint)(user_t*);
/** User monitor function */
typedef void (*UserMonitor)(user_t*);
/** User traverse function */
typedef gboolean (*UserForeach)(user_t*, void* data);

#define USER(arg) ((user_t*)(arg))
/** Basic user class */
struct _user_t {
  object_t _obj;             ///< base class
  int ID;                    ///< the id of the plugin which owns the user
  UserPrint print;           ///< the user print function, or NULL to use the object name
  UserMonitor monitor;       ///< called when the user is monitored
  UserMonitor unmonitor;     ///< called when the user is unmonitored.

  // private
  gboolean firewalled;
  ip_t ip;
  port_t port;
  int mon_cnt;
  unsigned refs;
  gboolean online;
  GList* uploads;
  GList* sources;
};

module_t* user_module_instance();

void user_destroy(user_t* user);
void user_init(user_t* user, const char* name, int ID);
user_t* user_new(const char* name, int ID);

ip_t user_ip(void* user);
port_t user_port(void* user);
gboolean user_get_firewalled(user_t* user);
void* user_ref(user_t* user);
void* user_unref(user_t* user);

void user_set_ip(void* user, ip_t ip);
void user_set_port(void* user, port_t port);
void user_set_ip_port(void* user, ip_t ip, port_t port);
void user_set_firewalled(user_t* user, gboolean wall);

void user_start_monitor(user_t* user);
void user_stop_monitor(user_t* user);
int user_get_monitor(user_t* user);
void user_set_online(user_t* user, gboolean setting);
void users_foreach_monitored(int ID, UserForeach func, void* data);

/*@}*/

G_END_DECLS
#endif
