/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.17 $
 * $Date: 2007/11/21 20:18:24 $
 * $Author: sgop $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <unistd.h>
#endif

#include "l_base.h"
#include "l_listen.h"
#include "l_http.h"
#include "l_io.h"
#include "preferences.h"
#include "plugin.h"
#include "library.h"
#include "upload.h"
#include "download.h"
#include "web_network.h"
#include "search.h"
#include "version.h"

#ifdef WIN32
#include "lcorealias.h"
#endif


#define LF "\n"
#define CR "\r"
#define CRLF "\r\n"

typedef struct {
  http_t _http;
  // library
  void* lib_folder;
  int lib_media;
  int lib_time;
  // search
  void* search_page;
} web_t;

typedef struct {
  char* name;
  unsigned pos;
  unsigned len;
} web_content_t;

typedef struct {
  char* name;
  char* content;
  GList* replace;
  GList* insert;
} web_template_t;

typedef struct {
  module_t _module;
  GList* templates;
  time_t last_loaded;
  const char* selection;
  listen_t* listen;
  int tab;
  int oddeven;
  file_node_t* folder;
  int media;
  int time;
  search_t* search;
} web_module_t;

static web_module_t* WebModule = NULL;

typedef void (*BufferFillFunc)(buffer_t* buffer, void* data);

#define TAB_NONE     -1
#define TAB_NETWORK   0
#define TAB_SEARCH    1
#define TAB_LIBRARY   2
#define TAB_DOWNLOAD  3
#define TAB_UPLOAD    4
#define TAB_SETTINGS  5

static const char* CodeName(int no) {
  switch (no) {
  case 200: return "OK";
  case 302: return "Found";
  case 400: return "Bad Request";
  case 401: return "Unauthorized";
  case 404: return "Not Found";
  default: return "Invalid";
  };
}
static const char* TNames(int no) {
  switch (no) {
  case 0:  return "Show unshared";
  case 1:  return "Show shared";
  case 2:  return "Show new";
  default: return "Show all";
  };
}

static const char* MNames(int no) {
  switch (no) {
  case MEDIA_FOLDER:      return "Folders";
  case MEDIA_AUDIO:       return "Audio files";
  case MEDIA_VIDEO:       return "Videos";
  case MEDIA_APPLICATION: return "Applications";
  case MEDIA_IMAGE:       return "Images";
  case MEDIA_DOCUMENT:    return "Documents";
  case MEDIA_UNKNOWN:     return "Other";
  default:                return "Files";
  };
}


static const char*
HttpSpecialChars(char** buffer, unsigned* len, unsigned* res_len) {

  *res_len = 0;
  if (*len == 0) return NULL;
  
  switch ((*buffer)[0]) {
  case '&':
    *res_len = 5; *buffer += 1; *len -= 1;
    return "&amp;";
  case '<':
    *res_len = 4; *buffer += 1; *len -= 1;
    return "&lt;";
  case '>':
    *res_len = 4; *buffer += 1; *len -= 1;
    return "&gt;";
  case '"':
    *res_len = 6; *buffer += 1; *len -= 1;
    return "&quot;";
  default:
    *res_len = 0;
    while (*len > 0 &&
	   **buffer != '&' &&
	   **buffer != '<' &&
	   **buffer != '>' &&
	   **buffer != '"') {
      *res_len += 1; *buffer += 1; *len -= 1;
    }
    return *buffer;
  }
}

static void
web_content_destroy(web_content_t* content) {
  g_free(content->name);
  g_free(content);
}
static web_content_t*
web_content_new(const char* name, unsigned pos, unsigned len) {
  web_content_t* content = g_malloc0(sizeof(*content));
  content->name = g_strdup(name);
  content->pos = pos;
  content->len = len;
  return content;
}

static void
web_template_destroy(web_template_t* template) {
  g_free(template->name);
  g_free(template->content);
  g_list_foreach(template->replace, (GFunc)web_content_destroy, NULL);
  g_list_free(template->replace);
  g_list_foreach(template->insert, (GFunc)web_content_destroy, NULL);
  g_list_free(template->insert);
  g_free(template);
}

static web_template_t*
web_template_new(const char* name, char* buffer) {
  web_template_t* template = g_malloc0(sizeof(*template));
  char* pos;
  char* tname;
  int size=strlen(buffer);

  while (size > 0 && isspace((int)buffer[size-1])) size--;
  buffer[size] = 0;
  while (isspace((int)(*buffer))) buffer++;

  template->name = g_strdup(name);
  template->content = g_strdup(buffer);

  pos = template->content;
  while ((tname = strstr(pos, "[insert:")) != NULL) {
    char* tstart = tname+8;
    char* temp = NULL;
    web_content_t* content;

    while (*tstart) {
      if (isspace((int)(*tstart))) break;
      if (*tstart == ']') {
	temp = tstart;
	break;
      }
      tstart++;
    }
    if (!temp) {
      g_log("WEB", G_LOG_LEVEL_WARNING,
	    "no ']' in insert header : skipping");
      pos += 8;
      continue;
    }
    *temp = 0;
    content = web_content_new
      (tname+8, tname-template->content, temp-tname+1);
    template->insert = g_list_append(template->insert, content);
    *temp = ']';
    pos = temp+1;
  }

  pos = template->content;
  while ((tname = strstr(pos, "[replace:")) != NULL) {
    char* tstart = tname+9;
    char* temp = NULL;
    web_content_t* content;

    while (*tstart) {
      if (isspace((int)(*tstart))) break;
      if (*tstart == ']') {
	temp = tstart;
	break;
      }
      tstart++;
    }
    if (!temp) {
      g_log("WEB", G_LOG_LEVEL_WARNING,
	    "no ']' in replace header : skipping");
      pos += 9;
      continue;
    }
    *temp = 0;
    content = web_content_new
      (tname+9, tname-template->content, temp-tname+1);
    template->replace = g_list_append(template->replace, content);
    *temp = ']';
    pos = temp+1;
  }
  return template;
}

static gboolean web_reload_templates() {
  const char* dir = pref_get_ptr("/core/web/resource_path", char);
  const char* fname;
  buffer_t* buffer;
  size_t size;
  FILE* file;
  struct stat stats;

  fname = l_printf(SDIR("%s") "web.templates", dir);
  
  if (stat(fname, &stats) < 0) {
    g_log("WEB", G_LOG_LEVEL_WARNING,
	  "can not stat '%s'", fname);
    return FALSE;
  }
  if (stats.st_mtime < WebModule->last_loaded) return TRUE;
  WebModule->last_loaded = CurrentTime;
  g_log("WEB", G_LOG_LEVEL_DEBUG, "reload templates '%s'", fname);
  
  size = stats.st_size;
  file = fopen(fname, "r");
  if (!file) {
    g_log("WEB", G_LOG_LEVEL_WARNING,
	  "can not open '%s'", fname);
    return FALSE;
  }

  buffer = buffer_new(size+1);
  if (fread(buffer_write_pos(buffer), 1, size, file) != size) {
    g_log("WEB", G_LOG_LEVEL_WARNING,
	  "can not load '%s'", fname);
    fclose(file);
    buffer_destroy(buffer);
    return FALSE;
  }
  buffer->size = size;
  buffer->data[size] = 0;

  fclose(file);

  g_list_foreach(WebModule->templates, (GFunc)web_template_destroy, NULL);
  g_list_free(WebModule->templates);
  WebModule->templates = NULL;

  while (buffer_avail(buffer)) {
    char* pos = buffer_read_pos(buffer);
    char* tname = strstr(pos, "[template:");
    char* tstart;
    char* temp;

    if (!tname) break;

    tstart = strchr(tname, '\n');
    if (!tstart) tstart = strchr(tname, '\r');
    if (!tstart) {
      g_log("WEB", G_LOG_LEVEL_WARNING,
	    "no line end after template header : skipping");
      buffer_consume_virt(buffer, 10);
      continue;
    }
    tstart++;
    temp = strchr(tname, ']');
    if (!temp || temp >= tstart) {
      g_log("WEB", G_LOG_LEVEL_WARNING,
	    "no ']' in template header : skipping");
      buffer_consume_virt(buffer, 10);
      continue;
    }
    *temp = 0;
    tname += 10;
    temp = strstr(tstart, "[template");
    if (!temp || temp[9] != ']') {
      g_log("WEB", G_LOG_LEVEL_WARNING,
	    "no [template] end tag found : aborting");
      break;
    }
    *temp = 0;

    g_log("WEB", G_LOG_LEVEL_DEBUG,
	  "new template '%s' found", tname);
    WebModule->templates = 
      g_list_append(WebModule->templates, web_template_new(tname, tstart));
    buffer_consume_virt(buffer, (temp+10-pos));
  }
  g_log("WEB", G_LOG_LEVEL_DEBUG,
	"no more templates found");

  buffer_destroy(buffer);
  return TRUE;
}

static web_template_t* web_template_find(const char* name) {
  GList* dlist;
  for (dlist = WebModule->templates; dlist; dlist = dlist->next) {
    web_template_t* template = dlist->data;
    if (!strcmp(template->name, name)) return template;
  }
  return NULL;
}

static void
web_buffer_insert_template(buffer_t* buffer, web_template_t* template,
			   GHashTable* replace, GHashTable* replace_str, void* data) {
  unsigned cur_pos = 0;
  unsigned size = strlen(template->content);
  GList* replace_list = template->replace;
  GList* insert_list = template->insert;

  while (cur_pos < size) {
    unsigned insert_pos;
    unsigned replace_pos;

    if (!replace_list) replace_pos = size;
    else replace_pos = ((web_content_t*)replace_list->data)->pos;
    if (!insert_list) insert_pos = size;
    else insert_pos = ((web_content_t*)insert_list->data)->pos;

    if (replace_pos == cur_pos) {
      web_content_t* content = replace_list->data;
      BufferFillFunc func = g_hash_table_lookup(replace, content->name);
      const char* str = g_hash_table_lookup(replace_str, content->name);
      if (str)
	buffer_add_string(buffer, str);
      else if (func)
	func(buffer, data);
      else
	buffer_add_string(buffer, l_printf("Replacement for '%s' not found", content->name));

      replace_list = replace_list->next;
      cur_pos += content->len;
    } else if (insert_pos == cur_pos) {
      web_content_t* content = insert_list->data;
      BufferFillFunc func = g_hash_table_lookup(replace, content->name);
      const char* str = g_hash_table_lookup(replace_str, content->name);
      web_template_t* sub_template = web_template_find(content->name);

      if (str) {
	//override insert with string
	buffer_add_string(buffer, str);
      } else if (func) {
	//override insert with function
	func(buffer, data);
      } else if (sub_template) {
	// use original template
	web_buffer_insert_template(buffer, sub_template, replace,
				   replace_str, data);
      } else {
	buffer_add_string(buffer, l_printf("Replacement for '%s' not found", content->name));
      }
      
      insert_list = insert_list->next;
      cur_pos += content->len;
    } else {
      unsigned bytes;
      if (replace_pos < insert_pos) bytes = replace_pos - cur_pos;
      else bytes = insert_pos - cur_pos;
      buffer_add_data(buffer, template->content+cur_pos, bytes);
      cur_pos += bytes;
    }
  }
}


static void
web_pack_template(buffer_t* buffer, const char* t, void* data,
		  ...) {
  web_template_t* template = web_template_find(t);
  GHashTable* replace =
    g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
  GHashTable* replace_str =
    g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
  va_list ap;
  char* name;

  if (template) {
    va_start(ap, data);
    while ((name = va_arg(ap, char*)) != NULL) {
      BufferFillFunc func = va_arg(ap, void*);
      if (*name == 1)
	g_hash_table_insert(replace_str, g_strdup(name+1), func);
      else
	g_hash_table_insert(replace, g_strdup(name), func);
    }
    va_end(ap);
    web_buffer_insert_template(buffer, template, replace,
			       replace_str, data);
  } else {
    buffer_add_string(buffer, l_printf("template '%s' not found", t));
  }
  g_hash_table_destroy(replace);
}


static buffer_t* web_load_page(const char* fname, gboolean binary)
{
    FILE* file;
    buffer_t* buffer;
    file_size_t size;
    
    if (!fname) return NULL;
    if (binary) file = fopen(fname, "rb");
    else file = fopen(fname, "r");
    if (!file)
    {
        g_log("WEB", G_LOG_LEVEL_WARNING,
              "could not open '%s'", fname);
        return NULL;
    }
    
    FSEEK(file, 0, SEEK_END);
    size = FTELL(file);
    FSEEK(file, 0, SEEK_SET);
    buffer = buffer_new(size+1);
    if (fread(buffer_write_pos(buffer), size, 1, file) != 1)
    {
        buffer_destroy(buffer);
        buffer = NULL;
    }
    else
    {
        buffer->data[size] = 0;
        buffer->size = size;
    }
    fclose(file);
    return buffer;
}

static void
web_send_page(web_t* web, buffer_t* buffer, 
	      int code, const char* mime, const char* add_header) {
  tcp_socket_send_str
    (TSOCK(web), l_printf("HTTP/1.1 %d %s"CRLF, code, CodeName(code)));
  tcp_socket_send_str
    (TSOCK(web), l_printf("Content-length: %d"CRLF, buffer->size));
  tcp_socket_send_str
    (TSOCK(web), "Cache-Control: no-cache, no-store, must-revalidate"CRLF);
  tcp_socket_send_str
    (TSOCK(web), "Expires: 0"CRLF);
  if (add_header) tcp_socket_send_str(TSOCK(web), add_header);
  tcp_socket_send_str
    (TSOCK(web), "Server: Lopster/2.0.0"CRLF);
  tcp_socket_send_str
    (TSOCK(web), l_printf("Content-Type: %s"CRLF CRLF, mime));

  tcp_socket_send_bin(TSOCK(web), buffer->data, buffer->size);
}

static void
web_send_simple_page(web_t* web, const char* header,
		     int code, const char* add, const char* message) {
  buffer_t* buffer = buffer_new(1024);
  
  web_pack_template(buffer, "WebSimplePage", NULL,
		    "\1TheHeadLine", header,
		    "\1TheCode", l_printf("%d", code),
		    "\1TheMessage", message,
		    "\1TheVersion", FULL_VERSION,
		    NULL);
  web_send_page(web, buffer, code, "text/html", add);
  buffer_destroy(buffer);
}

static void web_send_bad_request(web_t* web) {
  web_send_simple_page
    (web, "Bad Request", 400, NULL,
     "Your browser has sent a request this server could not understand.");
}
static void web_send_unauthorized(web_t* web) {
  web_send_simple_page
    (web, "Unauthorized", 401, 
     "WWW-Authenticate: Basic realm=\"Lopster Web Interface\""CRLF,
     "Your client is not authorized to view this page.");
}
static void web_send_not_found(web_t* web) {
  web_send_simple_page
    (web, "Not Found", 404, NULL,
     "The requested page was not found on this server");
}
static void web_send_redirect(web_t* web, const char* dest) {
  web_send_simple_page
    (web, "Created Page", 302,
     l_printf("Location: %s"CRLF, dest),
     l_printf("Please click <a href='%s'>here</a>, if your browser doesn't redirect you automatically.", dest));
}

static void
web_send_file(web_t* web, const char* filename, const char* type) {
  buffer_t* buffer = NULL;
  
  buffer = web_load_page(filename, TRUE);
  
  if (!buffer) {
    g_log("WEB", G_LOG_LEVEL_WARNING,
	  "file '%s' not found", filename);
    web_send_not_found(web);
    return;
  }
  web_send_page(web, buffer, 200, type, NULL);
  buffer_destroy(buffer);
}

static void web_send_css(web_t* web, const char* page) {
  const char* dir = pref_get_ptr("/core/web/resource_path", char);
  const char* filename = l_printf(SDIR("%s") "%s", dir, page);
  web_send_file(web, filename, "text/css");
}
static void web_send_javascript(web_t* web, const char* page) {
  const char* dir = pref_get_ptr("/core/web/resource_path", char);
  const char* filename = l_printf(SDIR("%s") "%s", dir, page);
  web_send_file(web, filename, "text/javascript");
}
static void web_send_png(web_t* web, const char* page) {
  const char* filename = lopster_pixmap_find(page);
  web_send_file(web, filename, "image/png");
}

/////////////////////////////////////////////
//  SEARCH
/////////////////////////////////////////////
static void 
web_add_search_result(search_result_t* result, buffer_t* buffer) {
  const char* type;

  if (WebModule->oddeven == 0) type = "WebSearchFileOdd";
  else type = "WebSearchFileEven";
  WebModule->oddeven = 1-WebModule->oddeven;
  
  web_pack_template
    (buffer, type, NULL,
     "\1TheFilename", FILE(result)->filename,
     "\1TheFilesize", print_filesize(&(FILE(result)->size)),
     "\1TheMeta", attr_print_all(ATTRC(result)),
     "\1TheSource", result->user->print?result->user->print(result->user):"",
     /* 		    "\1TheIcon", plugin_get_icon(plugin_get(result->user->ID)), */
     // FIXME: need better impl
     "\1TheIcon", result->user->ID>=0?plugin_get_icon(plugin_get(result->user->ID)):"search.png",
     "\1TheId", l_printf("%p", result),
     NULL);
}

static void web_add_search_files(buffer_t* buffer) {
  if (WebModule->search) {
    search_foreach_result(WebModule->search, (StorageTraverseFunc)web_add_search_result, buffer);
  }
}

static void web_add_search_tabs(buffer_t* buffer) {
  GList* dlist;
  
  for (dlist = search_get_searches(); dlist; dlist = dlist->next) {
    search_t* search = dlist->data;
    if (search->download) continue;

    web_pack_template(buffer, (search==WebModule->search)?"WebSearchActiveTab":"WebSearchTab", NULL,
		      "\1TheSearchName", OBJ(search)->name,
		      "\1TheSearchPointer", l_printf("%p", search),
		      NULL);
  }
}

static void
on_add_pattern(search_pattern_t* pattern, buffer_t* buffer) {
  attribute_t* attr_rate = attr_find(ATTRC(pattern), "pattern", "Bitrate");
  attribute_t* attr_size = attr_find(ATTRC(pattern), "pattern", "Size");
  int* vec_rate = attr_rate?attr_get_ptr(attr_rate, int):NULL;
  file_size_t* vec_size = attr_size?attr_get_ptr(attr_size, file_size_t):NULL;
  
  web_pack_template(buffer, "WebSearchPatternItem", NULL,
                    "\1TheName", OBJ(pattern)->name,
                    "\1TheMedia", l_printf("%d", pattern->media),
                    "\1TheMinSize", vec_size?print_bytes_exact(vec_size):"0B",
                    "\1TheMaxSize", vec_size?print_bytes_exact(vec_size+1):"0B",
                    "\1TheMinRate", vec_rate?l_printf("%d", vec_rate[0]):"0",
                    "\1TheMaxRate", vec_rate?l_printf("%d", vec_rate[1]):"0",
                    NULL);
  buffer_add_string(buffer, "\n");
}
static void web_add_patterns(buffer_t* buffer) {
  search_patterns_foreach((ObjectFunc)on_add_pattern, buffer);
}
static void web_add_search(buffer_t* buffer) {
  web_pack_template(buffer, "WebSearch", NULL,
		    "TheFileList", web_add_search_files,
		    "TheSearchTabs", web_add_search_tabs,
		    "ThePatternItems", web_add_patterns,
		    NULL);
}  

static search_t* web_find_search(void* pointer) {
  GList* dlist;

  for (dlist = search_get_searches(); dlist; dlist = dlist->next) {
    search_t* pattern = dlist->data;
    if (pointer == pattern) return pattern;
  }
  return NULL;
}
static void web_add_search_actions(buffer_t* buffer) {
  web_pack_template(buffer, "WebSelectionAction", NULL,
		    "\1TheAction", "download",
		    "\1TheActionName", "download",
		    NULL);
}
static void web_send_search(web_t* web, http_tags_t* tags) {
  buffer_t* buffer = buffer_new(1024);
  const char* value;
  
  if ((value = http_tags_find(tags, "id")) != NULL) {
    web->search_page = (void*)strtoul(value, NULL, 16);
  }
  WebModule->search = web_find_search(web->search_page);
  
  WebModule->tab = TAB_SEARCH;
  web_pack_template(buffer, "WebPage", NULL,
		    "\1TheHeadLine", "Lopster Search Interface",
		    "TheContent", web_add_search,
		    "\1WebSubNavigation", "",
		    "TheSelActions", web_add_search_actions,
		    "\1TheVersion", FULL_VERSION,
		    NULL);
  web_send_page(web, buffer, 200, "text/html", NULL);
  buffer_destroy(buffer);

}



/////////////////////////////////////////////
//  LIBRARY
/////////////////////////////////////////////
static gboolean expand(file_node_t* node) {
  file_node_t* trace;

  if (!node->parent) return TRUE;  // exand if root
  trace = WebModule->folder;
  while (trace) {
    if (node == trace) return TRUE;
    trace = trace->parent;
  }
  return FALSE;
}

static file_node_t*
show_children(file_tree_t* tree, file_node_t* node, buffer_t* buffer) {
  (void)tree;
  if (node != WebModule->folder) return NULL;

  node = node->child;
  while (node) {
    if (node->file && node->fci[WebModule->media][WebModule->time] > 0) {
      const char* pix;
      const char* type;
      if (node->fci[WebModule->media][1] == 0) {
        pix = "share_none.png";
      } else if (WebModule->time == 0) {
        pix = "share_none.png";
      } else if (WebModule->time == 1) {
        pix = "share_all.png";
      } else if (node->fci[WebModule->media][1] < node->fci[WebModule->media][3]) {
        pix = "share_part.png";
      } else {
        pix = "share_all.png";
      }
      
      if (WebModule->oddeven == 0) type = "WebLibFileOdd";
      else type = "WebLibFileEven";
      
      web_pack_template(buffer, type, NULL,
                        "\1ThePixmap", pix,
                        "\1TheFilename", OBJ(node)->name,
                        "\1TheFilesize", print_filesize(&(node->file->size)),
                        "\1TheMeta", attr_print_all(ATTRC(node->file)),
                        "\1TheId", l_printf("%p", node),
                        NULL);
      WebModule->oddeven = 1-WebModule->oddeven;
    }
    node = node->next;
  }

  return node;
}

static void web_add_lib_files(buffer_t* buffer) {
  library_t* lib = library_get_current();
  if (lib && WebModule->folder) {
    file_tree_foreach(FILE_TREE(lib), (FileNodeTraverse)show_children, buffer);
  }
}
/*
static void web_add_lib_options(buffer_t* buffer) {
  file_tree_t* tree = FILE_TREE(lib_get_current());
  int i1, i2;

  if (!tree) return;

  buffer_add_string(buffer, 
		    l_printf("<form action='/library' method='get'>\n", WebModule->folder));
  buffer_add_string
    (buffer, "<select name='show_time' class='options' size='1'>\n");
  i1 = WebModule->media;
  for (i2 = 0; i2 <= 3; i2++) {
    char* text;
    text =
      l_printf("<option %svalue='%d'>%s: %d (%.1f%%) %s (%.1f%%)</option>\n",
	       WebModule->time==i2?"selected='selected' ":"",
	       i2, TNames(i2),  tree->fci[i1][i2],
	       (double)(tree->fci[i1][i2])/
	       (double)(tree->fci[i1][3])*100.,
	       print_filesize(&(tree->fsi[i1][i2])),
	       (double)tree->fsi[i1][i2]/
	       (double)tree->fsi[i1][3]*100.);
    buffer_add_string(buffer, text);
  }
  buffer_add_string(buffer, "</select>\n");

  buffer_add_string
    (buffer, "<select name='show_media' class='options' size='1'>\n");
  i2 = WebModule->time;
  for (i1 = 0; i1 <= MEDIA_ANY; i1++) {
    char* text;
    text =
      l_printf("<option %svalue='%d'>%s: %d (%.1f%%) %s (%.1f%%)</option>\n",
	       WebModule->media==i1?"selected='selected' ":"",
	       i1, MNames(i1),  tree->fci[i1][i2],
	       (double)(tree->fci[i1][i2])/
	       (double)(tree->fci[MEDIA_ANY][i2])*100.,
	       print_filesize(&(tree->fsi[i1][i2])),
	       (double)tree->fsi[i1][i2]/
	       (double)tree->fsi[MEDIA_ANY][i2]*100.);
    buffer_add_string(buffer, text);
  }
  buffer_add_string(buffer, "</select>\n");
  buffer_add_string(buffer, l_printf("<input type='hidden' name='folder' value='%p'/>\n", WebModule->folder));
  buffer_add_string(buffer, "<input type='submit' class='button' value='Change View'/>\n");
  buffer_add_string(buffer, "</form>\n");
}
*/

static void web_add_lib_folder(buffer_t* buffer, file_node_t* node) {

  while (node) {
    if (!node->file && node->fci[WebModule->media][WebModule->time] > 0) {
      const char* pix;
      const char* type;
      gboolean expand_sub;

      if (node->fci[WebModule->media][1] == 0) {
	pix = "share_none.png";
      } else if (WebModule->time == 0) {
	pix = "share_none.png";
      } else if (WebModule->time == 1) {
	pix = "share_all.png";
      } else if (node->fci[WebModule->media][1] < node->fci[WebModule->media][3]) {
	pix = "share_part.png";
      } else {
	pix = "share_all.png";
      }
      if (node->fci[MEDIA_FOLDER][3] > 1) {
	if (expand(node)) {
	  expand_sub = TRUE;
	  type = "WebLibNormalFolder";
	} else {
	  expand_sub = FALSE;
	  type = "WebLibExpandableFolder";
	}
      } else {
	expand_sub = FALSE;
	type = "WebLibNormalFolder";
      }
      if (node == WebModule->folder) type = "WebLibSelectedFolder";
      web_pack_template(buffer, type, NULL,
			"\1ThePixmap", pix,
			"\1TheFolderPointer", l_printf("%p", node),
			"\1ShowMedia", l_printf("%d", WebModule->media),
			"\1ShowTime", l_printf("%d", WebModule->time),
			"\1TheName", OBJ(node)->name,
			NULL);
      if (expand_sub) {
	web_pack_template(buffer, "WebLibFolderList", node->child,
			  "TheSubFolders", web_add_lib_folder,
			  NULL);
      }
    }
    node = node->next;
  }
}

static void web_add_lib(buffer_t* buffer) {
  library_t* lib = library_get_current();
  file_node_t* node;
  
  if (lib) node = FILE_TREE(lib)->files;
  else node = NULL;

  web_pack_template(buffer, "WebLibrary", node,
		    "TheSubFolders", web_add_lib_folder,
		    "TheFileList", web_add_lib_files,
		    NULL);
}  

static file_node_t*
find_folder(file_tree_t* tree, file_node_t* node, void* folder) {
  (void)tree;
  if (!node->file && node == folder) {
    WebModule->folder = folder;
    return node;
  }
  return NULL;
}

static void web_add_lib_actions(buffer_t* buffer) {
  web_pack_template(buffer, "WebAction", NULL,
		    "\1TheAction", "refresh_lib",
		    "\1TheActionName", "refresh library",
		    NULL);
}
static void web_add_lib_sel_actions(buffer_t* buffer) {
  web_pack_template(buffer, "WebSelectionAction", NULL,
		    "\1TheAction", "share",
		    "\1TheActionName", "share",
		    NULL);
  web_pack_template(buffer, "WebSelectionAction", NULL,
		    "\1TheAction", "unshare",
		    "\1TheActionName", "unshare",
		    NULL);
}
static void web_send_library(web_t* web, http_tags_t* tags) {
  buffer_t* buffer = buffer_new(1024);
  const char* value;

  if ((value = http_tags_find(tags, "folder")) != NULL) {
    web->lib_folder = (void*)strtoul(value, NULL, 16);
  }
  if ((value = http_tags_find(tags, "show_media")) != NULL) {
    web->lib_media = strtoul(value, NULL, 10);
    if (web->lib_media < MEDIA_UNKNOWN) web->lib_media = MEDIA_ANY;
    else if (web->lib_media > MEDIA_ANY) web->lib_media = MEDIA_ANY;
  }
  if ((value = http_tags_find(tags, "show_time")) != NULL) {
    web->lib_time = strtoul(value, NULL, 10);
    if (web->lib_time < 0) web->lib_time = 3;
    else if (web->lib_time > 3) web->lib_time = 3;
  }

  WebModule->media = web->lib_media;
  WebModule->time = web->lib_time;
  WebModule->folder = NULL;
  file_tree_foreach(FILE_TREE(library_get_current()),
                    (FileNodeTraverse)find_folder, web->lib_folder);

  WebModule->tab = TAB_LIBRARY;
  web_pack_template(buffer, "WebPage", NULL,
		    "\1TheHeadLine", "Lopster Library Interface",
		    "TheContent", web_add_lib,
		    "TheActions", web_add_lib_actions,
		    "TheSelActions", web_add_lib_sel_actions,
		    "\1TheVersion", FULL_VERSION,
		    NULL);
  web_send_page(web, buffer, 200, "text/html", NULL);
  buffer_destroy(buffer);
}

/////////////////////////////////////////////
//  SEARCH
/////////////////////////////////////////////

/////////////////////////////////////////////
//  NETWORK
/////////////////////////////////////////////
static void web_add_plugin_info(buffer_t* buffer, plugin_t* plugin)
{
    online_status_t* status = plugin->online;
    gboolean online = (status->status & ONLINE_STATUS_ONLINE);
    
    web_pack_template(buffer, "WebPluginInfo", plugin,
                      "\1ThePluginInfoItem", online?"Online":"Offline",
                      NULL);

    if (online)
    {
        if (status->users)
            web_pack_template(buffer, "WebPluginInfo", plugin,
                              "\1ThePluginInfoItem", l_printf("%u users", status->users),
                              NULL);
        if (status->files)
            web_pack_template(buffer, "WebPluginInfo", plugin,
                              "\1ThePluginInfoItem", l_printf("%u files", status->files),
                              NULL);
        if (status->gigs)
            web_pack_template(buffer, "WebPluginInfo", plugin,
                              "\1ThePluginInfoItem", l_printf("%u gigs", status->gigs),
                              NULL);
    }
}

static void web_add_plugins(buffer_t* buffer) {
  int i1;
  
  for (i1 = 0; i1 < plugin_get_number(); i1++) {
    plugin_t* plugin = plugin_get(i1);
    const char* action;

    if (plugin->online->status & ONLINE_STATUS_ONLINE)
      action = "disconnect";
    else
      action = "connect";

    if (!plugin) continue;
    web_pack_template(buffer, "WebPlugin", plugin,
		      "\1ThePluginName", plugin->info->name,
		      "\1ThePluginId", l_printf("%d", plugin->ID),
		      "\1ThePluginAction", action,
		      "ThePluginInfo", web_add_plugin_info,
		      NULL);
  }
}

static void web_add_plugin_actions(buffer_t* buffer) {
  web_pack_template(buffer, "WebAction", NULL,
		    "\1TheAction", "connect",
		    "\1TheActionName", "connect plugins",
		    NULL);
  web_pack_template(buffer, "WebAction", NULL,
		    "\1TheAction", "disconnect",
		    "\1TheActionName", "disonnect plugins",
		    NULL);
}

static void web_send_network(web_t* web, http_tags_t* tags) {
  buffer_t* buffer = buffer_new(1024);

  (void)tags;
  WebModule->tab = TAB_NETWORK;
  web_pack_template(buffer, "WebPage", NULL,
		    "\1TheHeadLine", "Lopster Network Interface",
		    "TheContent", web_add_plugins,
		    "TheActions", web_add_plugin_actions,
		    "\1WebActionBar", "",
		    "\1TheVersion", FULL_VERSION,
		    NULL);
  web_send_page(web, buffer, 200, "text/html", NULL);
  buffer_destroy(buffer);
}


/////////////////////////////////////////////
//  DOWNLOADS
/////////////////////////////////////////////
static void web_add_download(download_t* download, buffer_t* buffer) {
  const char* type;
  file_size_t prog;
  int rate = download_get_rate(download);
  file_size_t todo = FILE(download)->size - download_get_progress(download);
  int left = rate>0?todo/rate:0;
  
  if (WebModule->oddeven == 0) type = "WebDownFileOdd";
  else type = "WebDownFileEven";
  WebModule->oddeven = 1-WebModule->oddeven;
  
  prog = 100*download_get_progress(download)/FILE(download)->size;
  web_pack_template
    (buffer, type, NULL,
     "\1TheFilename", FILE(download)->filename,
     "\1TheFilesize", print_filesize(&(FILE(download)->size)),
     "\1TheSources", l_printf("%d/%u", download->no_active, g_list_length(download->sources)),
     "\1TheId", l_printf("%p", download),
     "\1TheProgress", l_printf("%d", prog),
     "\1TheProgressLeft", l_printf("%d", 100-prog),
     "\1TheRate", rate?print_speed(&rate):"",
     "\1TheETA", left?print_duration_unit(&left):"",
     NULL);
}
static void web_add_download_list(buffer_t* buffer) {
  downloads_foreach((ObjectFunc)web_add_download, buffer);
}
static void web_add_downloads(buffer_t* buffer) {
  web_pack_template(buffer, "WebDownload", NULL,
		    "TheFileList", web_add_download_list,
		    NULL);
}

static void web_add_download_actions(buffer_t* buffer) {
  web_pack_template(buffer, "WebSelectionAction", NULL,
		    "\1TheAction", "delete_download",
		    "\1TheActionName", "delete",
		    NULL);
}
static void web_send_download(web_t* web, http_tags_t* tags) {
  buffer_t* buffer = buffer_new(1024);

  (void)tags;
  WebModule->tab = TAB_DOWNLOAD;
  web_pack_template(buffer, "WebPage", NULL,
		    "\1TheHeadLine", "Lopster Download Interface",
		    "TheContent", web_add_downloads,
		    "\1WebSubNavigation", "",
		    "TheSelActions", web_add_download_actions,
		    "\1TheVersion", FULL_VERSION,
		    NULL);
  web_send_page(web, buffer, 200, "text/html", NULL);
  buffer_destroy(buffer);
}

/////////////////////////////////////////////
//  UPLOADS
/////////////////////////////////////////////
static void web_add_upload(upload_t* upload, buffer_t* buffer) {
  const char* type;
  int prog;
  segment_t* seg = upload->cur_segment;
  int rate = socket_get_rate(SOCK(upload));
  int left = rate>0?(seg->stop - seg->start - seg->size)/rate:0;

  if (WebModule->oddeven == 0) type = "WebUpFileOdd";
  else type = "WebUpFileEven";
  WebModule->oddeven = 1-WebModule->oddeven;
  
  if (seg) {
    prog = 100*seg->size / (seg->stop - seg->start);
    if (prog > 100) prog = 100;
    else if (prog < 0) prog = 0;
  } else {
    prog = 0;
  }

  web_pack_template
    (buffer, type, NULL,
     "\1TheFilename", upload->file->filename,
     "\1TheFilesize", print_filesize(&(upload->file->size)),
     "\1TheSource", upload->user->print(upload->user),
     "\1TheIcon", plugin_get_icon(plugin_get(upload->user->ID)),
     "\1TheId", l_printf("%p", upload),
     "\1TheProgress", l_printf("%d", prog),
     "\1TheProgressLeft", l_printf("%d", 100-prog),
     "\1TheRate", seg?print_speed(&rate):"",
     "\1TheETA", seg?print_duration_unit(&left):"",
     NULL);
}
static void web_add_upload_list(buffer_t* buffer) {
  uploads_foreach((ObjectFunc)web_add_upload, buffer);
}
static void web_add_uploads(buffer_t* buffer) {
  web_pack_template(buffer, "WebUpload", NULL,
		    "TheFileList", web_add_upload_list,
		    NULL);
}
static void web_add_upload_actions(buffer_t* buffer) {
  web_pack_template(buffer, "WebSelectionAction", NULL,
		    "\1TheAction", "delete_upload",
		    "\1TheActionName", "delete",
		    NULL);
}

static void web_send_upload(web_t* web, http_tags_t* tags) {
  buffer_t* buffer = buffer_new(1024);

  (void)tags;
  WebModule->tab = TAB_UPLOAD;
  web_pack_template(buffer, "WebPage", NULL,
		    "\1TheHeadLine", "Lopster Upload Interface",
		    "TheContent", web_add_uploads,
		    "\1WebSubNavigation", "",
		    "TheSelActions", web_add_upload_actions,
		    "\1TheVersion", FULL_VERSION,
		    NULL);
  web_send_page(web, buffer, 200, "text/html", NULL);
  buffer_destroy(buffer);
}

static void web_add_setting(pref_node_t* node, buffer_t* buffer) {
  const char* text;
  const char* type;
  
  printf("pref add setting %s\n", OBJ(node)->name);
  if (node->children) return;
  if (!node->xml_print || !node->store_str || !node->save) return;

  if (WebModule->oddeven == 0) type = "WebSettingOdd";
  else type = "WebSettingEven";
  WebModule->oddeven = 1-WebModule->oddeven;
  
  text = node->xml_print(node->value);
  web_pack_template(buffer, type, NULL,
		    "\1TheName", pref_node_fullname(node),
		    "\1TheValue", text?text:"",
		    NULL);
}
static void web_add_settings_list(buffer_t* buffer) {
  pref_foreach((ObjectFunc)web_add_setting, buffer);
}
static void web_add_settings(buffer_t* buffer) {
  web_pack_template(buffer, "WebSettings", NULL,
		    "TheSettings", web_add_settings_list,
		    NULL);
}
static void web_send_settings(web_t* web, http_tags_t* tags) {
  buffer_t* buffer = buffer_new(1024);

  (void)tags;
  WebModule->tab = TAB_SETTINGS;
  web_pack_template(buffer, "WebPage", NULL,
		    "\1TheHeadLine", "Lopster Settings Interface",
		    "TheContent", web_add_settings,
		    "\1WebSubNavigation", "",
		    "TheSelActions", web_add_upload_actions,
		    "\1TheVersion", FULL_VERSION,
		    NULL);
  web_send_page(web, buffer, 200, "text/html", NULL);
  buffer_destroy(buffer);
}




static void web_send_start_plugin(web_t* web, http_tags_t* tags) {
  const char* value = http_tags_find(tags, "id");

  plugin_connect_change(atoi(value), CONNECT_START);

  value = http_tags_find(HTTP(web)->header_info, "Host");
  if (value)
    web_send_redirect(web, l_printf("http://%s/network", value));
  else
    web_send_redirect(web, "/network");
}

static void web_send_stop_plugin(web_t* web, http_tags_t* tags)
{
    const char* value = http_tags_find(tags, "id");
    
    plugin_connect_change(atoi(value), CONNECT_STOP);

    value = http_tags_find(HTTP(web)->header_info, "Host");
    if (value)
        web_send_redirect(web, l_printf("http://%s/network", value));
    else
        web_send_redirect(web, "/network");
}

static void web_send_close_search(web_t* web, http_tags_t* tags) {
  GList* dlist;
  const char* value;

  if ((value = http_tags_find(tags, "id")) != NULL) {
    WebModule->search = (void*)strtoul(value, NULL, 16);
  }

  web->search_page = NULL;
  for (dlist = search_get_searches(); dlist; dlist = dlist->next) {
    search_t* search = dlist->data;
    if (search == WebModule->search) {
      if (!web->search_page && dlist->next) {
        web->search_page = dlist->next->data;
      }
      object_kill(OBJ(search));
      break;
    }
    web->search_page = search;
  }

  value = http_tags_find(HTTP(web)->header_info, "Host");
  if (web->search_page)
    web_send_redirect(web, l_printf("%s%s/search?id=%p", 
				    value?"http://":"",
				    value?value:"",
				    web->search_page));
  else
    web_send_redirect(web, l_printf("%s%s/search",
				    value?"http://":"",
				    value?value:""));
}
static void web_send_new_search(web_t* web, http_tags_t* tags) {
  search_pattern_t* pattern;
  search_t* search;
  const char* value;
  int min, max;
  
  pattern = search_pattern_new();

  if ((value = http_tags_find(tags, "term")) != NULL && *value) {
    make_list_from_string(&pattern->include, value, " ");
    object_set_name(OBJ(pattern), value);
  } else {
    object_set_name(OBJ(pattern), "-");
  }
  if ((value = http_tags_find(tags, "exclude")) != NULL && *value) {
    make_list_from_string(&pattern->exclude, value, " ");
  }

  if ((value = http_tags_find(tags, "size_min")) != NULL)
    min = extract_bytes(value); else min = 0;
  if ((value = http_tags_find(tags, "size_max")) != NULL)
    max = extract_bytes(value); else max = 0;
  if (min || max)
    attr_store(ATTRC(pattern), "pattern", "Size", min, max);

  if ((value = http_tags_find(tags, "dur_min")) != NULL)
    min = extract_duration(value); else min = 0;
  if ((value = http_tags_find(tags, "dur_max")) != NULL)
    max = extract_duration(value); else max = 0;
  if (min || max)
    attr_store(ATTRC(pattern), "pattern", "Duration", min, max);

  if ((value = http_tags_find(tags, "rate_min")) != NULL)
    min = atoi(value); else min = 0;
  if ((value = http_tags_find(tags, "rate_max")) != NULL)
    max = atoi(value); else max = 0;
  if (min || max)
    attr_store(ATTRC(pattern), "pattern", "Bitrate", min, max);

  if ((value = http_tags_find(tags, "freq_min")) != NULL)
    min = atoi(value); else min = 0;
  if ((value = http_tags_find(tags, "freq_max")) != NULL)
    max = atoi(value); else max = 0;
  if (min || max)
    attr_store(ATTRC(pattern), "pattern", "Frequency", min, max);

  if ((value = http_tags_find(tags, "media")) != NULL)
    pattern->media = media2int(value);
  else pattern->media = MEDIA_ANY;

  if (pattern) {
    search = search_new_from_pattern(pattern, -1);
    object_kill(OBJ(pattern));
  } else {
    search = NULL;
  }
  
  value = http_tags_find(HTTP(web)->header_info, "Host");
  web_send_redirect(web, l_printf("%s%s/search?id=%p",
				  value?"http://":"",
				  value?value:"", search));
}
static void web_send_refresh_lib(web_t* web, http_tags_t* tags) {
  const char* value;

  (void)tags;
  library_refresh(TRUE);

  value = http_tags_find(HTTP(web)->header_info, "Host");
  web_send_redirect(web, l_printf("%s%s/library?folder=%p",
                                  value?"http://":"",
                                  value?value:"", WebModule->folder));
}
static file_node_t*
un_share_nodes(file_tree_t* tree, file_node_t* node, gboolean* share) {
  static char str[100];
  sprintf(str, "%p", node);
  if (strstr(WebModule->selection, str)) {
    if (*share) file_tree_share_node(tree, node, MEDIA_ANY, FN_GROUP_ALL);
    else file_tree_unshare_node(tree, node, MEDIA_ANY, FN_GROUP_ALL);
  }
  return NULL;
}
static void web_send_un_share(web_t* web, http_tags_t* tags, gboolean share) {
  library_t* lib = library_get_current();
  const char* value;

  value = http_tags_find(tags, "sel");
  if (!value) {
    web_send_bad_request(web);
    return;
  }
  WebModule->selection = value;
  
  file_tree_foreach(FILE_TREE(lib), (FileNodeTraverse)un_share_nodes, &share);
  
  value = http_tags_find(HTTP(web)->header_info, "Host");
  web_send_redirect(web, l_printf("%s%s/library?folder=%p",
				  value?"http://":"",
				  value?value:"", WebModule->folder));
}
static void collect_items(void* item, GList** list) {
  static char str[100];
  sprintf(str, "%p", item);
  if (strstr(WebModule->selection, str)) *list = g_list_append(*list, item);
}
static void web_send_delete_down(web_t* web, http_tags_t* tags) {
  const char* value;
  GList* list = NULL;
  GList* dlist;

  value = http_tags_find(tags, "sel");
  if (!value) {
    web_send_bad_request(web);
    return;
  }
  WebModule->selection = value;
  
  downloads_foreach((ObjectFunc)collect_items, &list);
  for (dlist = list; dlist; dlist = dlist->next) {
    download_t* download = dlist->data;
    object_kill(OBJ(download));
  }
  g_list_free(list);

  value = http_tags_find(HTTP(web)->header_info, "Host");
  web_send_redirect
    (web, l_printf("%s%s/download", value?"http://":"", value?value:""));
}
static void web_send_delete_up(web_t* web, http_tags_t* tags) {
  const char* value;
  GList* list = NULL;
  GList* dlist;

  value = http_tags_find(tags, "sel");
  if (!value) {
    web_send_bad_request(web);
    return;
  }
  WebModule->selection = value;
  
  uploads_foreach((ObjectFunc)collect_items, &list);
  for (dlist = list; dlist; dlist = dlist->next) {
    upload_t* upload = dlist->data;
    upload_delete(upload);
  }
  g_list_free(list);

  value = http_tags_find(HTTP(web)->header_info, "Host");
  web_send_redirect
    (web, l_printf("%s%s/upload", value?"http://":"", value?value:""));
}
static void web_process_action(web_t* web, http_tags_t* tags) {
  const char* name = http_tags_find(tags, "name");

  if (!name)
    web_send_bad_request(web);
  else if (!strcmp(name, "connect"))
    web_send_start_plugin(web, tags);
  else if (!strcmp(name, "disconnect"))
    web_send_stop_plugin(web, tags);
  else if (!strcmp(name, "close_search"))
    web_send_close_search(web, tags);
  else if (!strcmp(name, "new_search"))
    web_send_new_search(web, tags);
  else if (!strcmp(name, "refresh_lib"))
    web_send_refresh_lib(web, tags);
  else if (!strcmp(name, "share"))
    web_send_un_share(web, tags, TRUE);
  else if (!strcmp(name, "unshare"))
    web_send_un_share(web, tags, FALSE);
  else if (!strcmp(name, "delete_upload"))
    web_send_delete_up(web, tags);
  else if (!strcmp(name, "delete_download"))
    web_send_delete_down(web, tags);
  else
    web_send_bad_request(web);
}





static gboolean web_authorized(web_t* web)
{
    char* require;
    char* user;
    char* pass;
    const char* value;
    
    if (!web) return FALSE;
    pass = pref_get_ptr("/core/web/password", char);
    if (!pass || *pass == 0) pass = pref_get_ptr("/core/user/pass", char);
    if (!pass || *pass == 0) return FALSE;
    
    user = pref_get_ptr("/core/web/username", char);
    if (!user || *user == 0) user = pref_get_ptr("/core/user/nick", char);
    if (!user || *user == 0) return FALSE;
    
    require = l_printf("%s:%s", user, pass);
    
    value = http_tags_find(HTTP(web)->header_info, "Authorization");
    if (value && !strncmp(value, "Basic ", 6))
    {
        int len;
        unsigned char* res = base64_decode(value+6, &len);
        if (res && !strcmp(require, res))
        {
            g_free(res);
            return TRUE;
        }
    }
    return FALSE;
}

static http_tags_t* web_tags_get(char* doc) {
  http_tags_t* tags;
  char* pos1;
  char* pos2;
  char* data;
  
  tags = http_tags_new();

  if (!doc) return tags;
  pos1 = strchr(doc, '?');
  if (!pos1) return tags;
  
  *pos1 = 0;
  data = l_strdup(pos1+1);
  while (data) {
    pos1 = data;
    data = strchr(data, '&');
    if (data) {
      *data = 0;
      data++;
    }
    pos2 = strchr(pos1, '=');
    if (!pos2) break;
    *pos2 = 0;
    pos2++;
    http_tags_add(tags, pos1, pos2);
    //    printf("new tag '%s' '%s'\n", pos1, pos2);
  }
  
  return tags;
}
static void web_process_incoming(web_t* web) {
  //  GList* dlist;
  WebModule->oddeven = 0;

  web_reload_templates();

  g_log("WEB", G_LOG_LEVEL_DEBUG,
	"process request: %s", HTTP(web)->header);
  /*
  for (dlist = HTTP(web)->header_info->tags; dlist; dlist = dlist->next) {
    http_tag_t* h = dlist->data;
    printf("    %20s : %s\n", h->name, h->value);
  }
  */
  if (!strncmp(HTTP(web)->header, "GET ", 4)) {
    char* temp = l_strdup(HTTP(web)->header);
    char* doc = arg(temp, 0);
    http_tags_t* tags;

    doc = arg(NULL, 0);
    
    if (!doc) {
      web_send_bad_request(web);
      return;
    }
    tags = web_tags_get(doc);
    
    if (!web_authorized(web)) {
      g_log("WEB", G_LOG_LEVEL_WARNING,
	    "rejected request: client is not authorized");
      web_send_unauthorized(web);
    } else if (!strcmp(doc, "/lopster.css")) {
      web_send_css(web, "lopster.css");
    } else if (!strcmp(doc, "/lopster.js")) {
      web_send_javascript(web, "lopster.js");
    } else if (!strcmp(doc, "/") || !strcmp(doc, "/network")) {
      web_send_network(web, tags);
    } else if (!strcmp(doc, "/library")) {
      web_send_library(web, tags);
    } else if (!strcmp(doc, "/search")) {
      web_send_search(web, tags);
    } else if (!strcmp(doc, "/download")) {
      web_send_download(web, tags);
    } else if (!strcmp(doc, "/upload")) {
      web_send_upload(web, tags);
/*     } else if (!strcmp(doc, "/settings")) { */
/*       web_send_settings(web, tags); */
    } else if (!strcmp(doc, "/action")) {
      web_process_action(web, tags);
    } else {
      unsigned len = strlen(doc);
      if (len >= 4 && !strchr(doc+1, DIR_SEP) &&
          !strcmp(doc+len-4, ".png")) {
        web_send_png(web, doc+1);
      } else {
        web_send_not_found(web);
      }
    }
    http_tags_destroy(tags);
  } else {
    web_send_bad_request(web);
  }
  return;
}

gboolean web_end(web_t* web, const char* message) {
  //  printf("*** web end %s, %d\n", message, SOCK(web)->status);

  g_assert(WebModule);

  if (socket_get_status(SOCK(web)) == ST_INPUT_COMPLETE) {
    web_process_incoming(web); // always writes something
  }
  return http_end(HTTP(web), message);
}

static const char* web_process_header(web_t* web) {
  if (!HTTP(web)->header) {
    web_send_bad_request(web);
    return NULL;
  }
  return NULL;
}

static void web_init(web_t* web)
{
    http_init(HTTP(web), url_parse("http://localhost")/*FIXME*/);
    object_add_type(OBJ(web), "WebSocket");
    socket_set_end(SOCK(web), (SockEndFunc)web_end);
    HTTP(web)->process_header = (BConsumeInputFunc)web_process_header;
    web->lib_folder = NULL;
    web->lib_media = MEDIA_ANY;
    web->lib_time = 3;
}

static web_t* web_new(void)
{
    web_t* web = g_malloc0(sizeof(*web));
    web_init(web);
    return web;
}

static tcp_socket_t* web_accept_con(int fd, ip_t ip, port_t port) {
  web_t* web = web_new();
  listen_init_accepted(TSOCK(web), fd, ip, port);
  return TSOCK(web);
}


static void web_stop_listening(void) {
  if (!WebModule->listen) return;
  socket_done(SOCK(WebModule->listen), ST_DONE, "stopping web service");
}

static void web_start_listening(void) {
  if (WebModule->listen) return;
  WebModule->listen =
    listen_port_new("/core/web/port", "WebListen",
		    (ListenAcceptFunc)web_accept_con, NULL);
  object_set_ptr((object_t**)&WebModule->listen);
  listen_port_start_listening(WebModule->listen, 10);
}

static void web_daemon_toggle(pref_node_t* node) {
  gboolean* active = node->value;
  if (*active) web_start_listening();
  else web_stop_listening();
}


static void web_daemon_init() {
  pref_register_int   ("/core/web/port", TRUE, 1212);
  pref_add_constraint_port("/core/web/port");
  pref_register_bool  ("/core/web/active", TRUE, FALSE);
  pref_register_string("/core/web/username", TRUE, NULL);
  pref_register_string("/core/web/password", TRUE, NULL);
#ifdef PACKAGE_SOURCE_DIR
  pref_register_string("/core/web/resource_path", TRUE, 
                       SDIR(PACKAGE_SOURCE_DIR) SDIR("lcore") "html");
#else
  pref_register_string("/core/web/resource_path", TRUE, 
                       SDIR(PACKAGE_DATA_DIR) "html");
#endif
}
static void web_daemon_run() {
  if (pref_get_ref("/core/web/active", gboolean))
    web_start_listening();
  pref_observe("/core/web/active", 
	       (ObjectFunc)web_daemon_toggle, NULL);
}
static void web_daemon_exit() {
}

static void web_module_init(web_module_t* module) {
  module_init(MODULE(module), "Web Module",
	      (ModuleFunc)web_daemon_init,
	      (ModuleFunc)web_daemon_run,
	      (ModuleFunc)web_daemon_exit);

  module->templates = NULL;
  module->last_loaded = 0;
  module->selection = NULL;
  module->listen = NULL;
  module->tab = -1;
  module->oddeven = 0;
  module->folder = NULL;
  module->media  = MEDIA_ANY;
  module->time   = 3;
  module->search = NULL;

  module_require(pref_module_instance);
  module_require(listen_module_instance);
  module_require(plugin_module_instance);
  module_require(library_module_instance);
  module_require(search_module_instance);
  module_require(upload_module_instance);
  module_require(download_module_instance);
  module_require(http_module_instance);
}


module_t* web_module_instance() {
  if (!WebModule) {
    WebModule = g_malloc0(sizeof(*WebModule));
    web_module_init(WebModule);
  }
  return MODULE(WebModule);
}

#ifdef WIN32
#define _WEB_NETWORK_C_
#include "lcorealiasdef.c"
#endif
