/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.12 $
 * $Date: 2008/12/23 13:38:14 $
 * $Author: sgop $
 */

#include <string.h>

#include <glib.h>

#include <libxml/parser.h>

#include "xml_handler.h"
#include "utils.h"
#include "l_base.h"
#ifdef WIN32
#include "lcorealias.h"
#endif

typedef struct {
  char* name;
  char* filename;
  char* root;
  time_t last_saved;
  gboolean loaded;
  XmlNodeFunc load;
  XmlNodeFunc save;
  int timer;
} xml_file_t;

typedef struct {
  module_t _module;
  GHashTable* files;
} xml_module_t;

static xml_module_t* XmlModule = NULL;

/** Register an xml file. A name is associated with a filename,
 * read/write function and a root node defined. You always must
 * register the xml files before you can call other functions on it.
 * @param name the name which is use later to access the xml file.
 * @param filename the full name of the file.
 * @param root the name of the root node in the xml file
 * @param load the load function
 * @param save the save function
 * @note Usually lopster_xml_register_home() is used instead of this function
 * @see lopster_xml_save(), lopster_xml_load()
 */
void lopster_xml_register(const char* name, const char* filename, const char* root,
                          XmlNodeFunc load, XmlNodeFunc save) {
  g_assert(XmlModule);
  if (XmlModule) {
    xml_file_t* file = g_hash_table_lookup(XmlModule->files, name);
    if (file) {
      g_log("XML", G_LOG_LEVEL_WARNING,
            "'%s' already registered as '%s'", name, file->filename);
      return;
    }
    file = g_malloc0(sizeof(*file));
    file->name = g_strdup(name);
    file->filename = g_strdup(filename);
    file->root = g_strdup(root);
    file->load = load;
    file->save = save;
    g_log("XML", G_LOG_LEVEL_DEBUG,
          "registered '%s' as '%s'", file->name, file->filename);
    g_hash_table_insert(XmlModule->files, file->name, file);
  }
}

/** This is a wrapper for lopster_xml_register(). Instead of a full
 * pathname you only give the filename relative to lopster's home
 * directory.
 * @param name the name which is use later to access the xml file.
 * @param filename the name of the file relative to lopster's home
 * @param root the name of the root node in the xml file
 * @param load the load function
 * @param save the save function
 */
void lopster_xml_register_home(const char* name, const char* filename, const char* root,
                               XmlNodeFunc load, XmlNodeFunc save) {
  g_assert(XmlModule);
  if (XmlModule) {
    const char* home = lopster_get_home();
    if (!home) {
      g_log("XML", G_LOG_LEVEL_WARNING,
            "lopster has no home directory set, did you require 'base_module_instance'?");
    } else {
      char* fname = l_printf(SDIR("%s") "%s", home, filename);
      lopster_xml_register(name, fname, root, load, save);
    }
  }
}

static void 
lopster_xml_save_file(const char* filename, const char* root, XmlNodeFunc func)
{
    xmlDocPtr xdoc;
    xmlNodePtr xroot;
    
    g_log("XML", G_LOG_LEVEL_INFO, "saving '%s'", filename);
    
    xdoc = xmlNewDoc("1.0");
    xroot = xmlNewNode(NULL, root);
    xmlDocSetRootElement(xdoc, xroot);
    
    func(xdoc, xroot); 
    
    create_dir(l_printf(filename));
    
    xmlSaveFormatFileEnc(filename, xdoc, "utf-8", 1);
    xmlFreeDoc(xdoc);
}

static gboolean _on_xml_save(xml_file_t* file) {
  lopster_xml_save_file(file->filename, file->root, file->save);
  file->last_saved = CurrentTime;
  if (file->timer) {
    lopster_timeout_remove(file->timer);
    file->timer = 0;
  }
  return FALSE;
}

/** Change the xml filename of a previously registered file.
 * @param name the name associated with the xml file
 * @param filename the new filename
 * @param save TRUE if the file should be saved to disk before changing the name
 * @return TRUE if filename was changed, FALSE otherwise
 * @see lopster_xml_register(), lopster_xml_register_home()
 */
gboolean lopster_xml_change(const char* name, const char* filename, gboolean save) {
  g_assert(XmlModule);
  if (XmlModule) {
    xml_file_t* file = g_hash_table_lookup(XmlModule->files, name);
    if (!file) {
      g_log("XML", G_LOG_LEVEL_WARNING,
            "cannot change '%s': not registered", name);
      return FALSE;
    }
    if (file->filename == filename) return FALSE;
    if (file->filename && filename && !strcmp(file->filename, filename))
      return FALSE;
    
    if (file->timer) {
      if (save) lopster_xml_save_file(file->filename, file->root, file->save);
      lopster_timeout_remove(file->timer);
      file->timer = 0;
    }
    
    g_free(file->filename);
    file->filename = g_strdup(filename);
    file->loaded = FALSE;
    file->last_saved = 0;
  }
  return (XmlModule != NULL);
}

static gboolean
lopster_xml_load_file(const char* filename, const char* root, XmlNodeFunc func) {
  xmlDocPtr xfile;
  xmlNodePtr xroot;

  g_assert(func);
  if (!g_file_test(filename, G_FILE_TEST_EXISTS)) 
    return FALSE;
  g_log("XML", G_LOG_LEVEL_INFO, "loading '%s'", filename);

  xfile = xmlParseFile(filename);

  if (!xfile) return FALSE;

  xroot = xmlDocGetRootElement(xfile);
  if (xroot == NULL) {
    g_log("XML", G_LOG_LEVEL_WARNING, "no root element in file: %s", filename);
    xmlFreeDoc(xfile);
    return FALSE;
  }
  
  if (strcmp(xroot->name, root)) {
    g_log("XML", G_LOG_LEVEL_WARNING, "Root element != %s", root);
    xmlFreeDoc(xfile);
    return FALSE;
  }
  
  func(xfile, xroot);
  
  xmlFreeDoc(xfile);

  return TRUE;
}

/** Load the xml file.
 * @param name the name associated with the xml file.
 * @return TRUE if successfully loaded, FALSE otherwise.
 * @see lopster_xml_register(), lopster_xml_register_home()
 */
gboolean lopster_xml_load(const char* name)
{
    g_assert(XmlModule);
    if (XmlModule)
    {
        xml_file_t* file = g_hash_table_lookup(XmlModule->files, name);
        if (!file)
        {
            g_log("XML", G_LOG_LEVEL_WARNING, "cannot load '%s': not registered", name);
            return FALSE;
        }
        if (lopster_xml_load_file(file->filename, file->root, file->load))
        {
            file->loaded = TRUE;
            return TRUE;
        }
        else
        {
            return FALSE;
        }
    }
    return FALSE;
}

static void xml_file_destroy(xml_file_t* file) {
  if (file->timer) _on_xml_save(file);
  g_free(file->name);
  g_free(file->filename);
  g_free(file->root);
  g_free(file);
}

/** Save the xml file. The file normally is not saved instantly but
 * after some time. If you call this function while a save is still
 * pending, the new save request is ignored.
 * @param name the name associated with the filename
 * @param timeout time in milliseconds after which the file is saved, 0 to save instantly
 * @see lopster_xml_register(), lopster_xml_register_home()
 * @note This module automatically takes care that the pending saves
 * are flushed to disk before the program exits.
 */
void lopster_xml_save(const char* name, unsigned timeout)
{
    g_assert(XmlModule);
    if (XmlModule)
    {
        xml_file_t* file = g_hash_table_lookup(XmlModule->files, name);
        if (!file)
        {
            g_log("XML", G_LOG_LEVEL_WARNING, "cannot save '%s': not registered", name);
            return;
        }
    
        if (!timeout)
        {
            _on_xml_save(file);
        }
        else if (!file->timer)
        {
            file->timer = lopster_timeout_add(timeout, (GSourceFunc)_on_xml_save, file);
            if (file->timer == 0)
            {
                // timer is not set yet so force save
                g_log("XML", G_LOG_LEVEL_WARNING, "timer was not set, force save");
                _on_xml_save(file);
            }
        }
    }
}

/** Find the filename of a register xml file.
 * @param name the name associated with the xml file.
 * @return the complete filename of the file.
 * @see lopster_xml_register(), lopster_xml_register_home()
 */
const char* lopster_xml_find(const char* name) {
  g_assert(XmlModule);
  if (XmlModule) {
    xml_file_t* file = g_hash_table_lookup(XmlModule->files, name);
    return file?file->filename:NULL;
  }
  return NULL;
}

/** Check whether an xml file was loaded.
 * @param name the name associated with the xml file.
 * @return TRUE if file was loaded, FALSE otherwise
 * @see lopster_xml_register(), lopster_xml_register_home(), lopster_xml_load()
 */
gboolean lopster_xml_loaded(const char* name) {
  g_assert(XmlModule);
  if (XmlModule) {
    xml_file_t* file = g_hash_table_lookup(XmlModule->files, name);
    if (!file) {
      g_log("XML", G_LOG_LEVEL_WARNING, "cannot find '%s': not registered", name);
      return FALSE;
    }
    return file->loaded;
  }
  return FALSE;
}

static void _on_xml_flush(char* name, xml_file_t* file) {
  (void)name;
  if (file->timer) _on_xml_save(file);
}

static void lopster_xml_flush() {
  g_assert(XmlModule);
  g_log("CORE", G_LOG_LEVEL_DEBUG, "flushing pending saves");
  g_hash_table_foreach(XmlModule->files, (GHFunc)_on_xml_flush, NULL);
}

void xml_set_bool_attr(xmlNodePtr xnode, const char* name, gboolean val)
{
    xmlNewProp(xnode, name, val?"yes":"no");
}

gboolean xml_get_bool_attr(xmlNodePtr xnode, const char* name, gboolean def)
{
    gboolean res = def;
    xmlChar* attr = xmlGetProp(xnode, name);
    if (attr)
    {
        res = (strcmp(attr, "yes") == 0);
        xmlFree(attr);
    }
    return res;
}


static void xml_init() {
  // need to call that early in shutdown process, since we get destroyed
  // _after_ our users get destroyed.
  lopster_call_at_exit(lopster_xml_flush);
}

static void xml_run() {
}

static void xml_exit() {
  g_hash_table_destroy(XmlModule->files);
  XmlModule->files = NULL;
  XmlModule = NULL;
}

static void xml_module_init(xml_module_t* module) {

  module_init(MODULE(module), "XML Module",
              (ModuleFunc)xml_init, (ModuleFunc)xml_run, (ModuleFunc)xml_exit);

  module->files =
    g_hash_table_new_full((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
                          NULL, (GDestroyNotify)xml_file_destroy);

  //  module_require(MODULE(module), base_module_instance);
}

/** This module is one of the very few modules that is ready for use
 * as soon as you've requested it.  No need to wait for
 * module_init() */
module_t* xml_module_instance() {
  if (!XmlModule) {
    XmlModule = g_malloc0(sizeof(*XmlModule));
    xml_module_init(XmlModule);
  }
  return MODULE(XmlModule);
}

#ifdef WIN32
# define _XML_HANDLER_C_
# include "lcorealiasdef.c"
#endif
