/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.6 $
 * $Date: 2007/05/23 20:21:53 $
 * $Author: sid4u $
 */

#include <stdio.h>  // needed for bzlib
#include <string.h>  // needed for memset

#include <zlib.h>
#include <bzlib.h>
#include <glib.h>

#include "buffer.h"
#include "zbuffer.h"
#include "huffman.h"

#ifdef WIN32
#include "lcorealias.h"
#endif

struct _compress_t {
  buffer_t* in;
  buffer_t* out;
  gboolean (*init)(compress_t*);
  int (*action)(compress_t*, gboolean finish);
  void (*destroy)(compress_t*);
  union {
    z_stream* zstream;
    bz_stream* bzstream;
  } stream;
};

typedef struct _hm_node_t hm_node_t;

struct _hm_node_t {
  int index;
  int cnt;
  hm_node_t* left;
  hm_node_t* right;
};


static void z_decompress_destroy(compress_t* decompress);
static gboolean z_decompress_init(compress_t* decompress);
static int z_decompress_action(compress_t* infl, gboolean finish);

static void z_compress_destroy(compress_t* compress);
static gboolean z_compress_init(compress_t* compress);
static int z_compress_action(compress_t* defl, gboolean finish);

static void bz_decompress_destroy(compress_t* decompress);
static gboolean bz_decompress_init(compress_t* decompress);
static int bz_decompress_action(compress_t* infl, gboolean finish);

static void bz_compress_destroy(compress_t* compress);
static gboolean bz_compress_init(compress_t* compress);
static int bz_compress_action(compress_t* defl, gboolean finish);

static void hm_compress_destroy(compress_t* compress);
static gboolean hm_compress_init(compress_t* compress);
static int hm_compress_action(compress_t* defl, gboolean finish);


/** Create a new compression object.
 * @param type the type of the compression
 * @param compress TRUE to compress, FALSE to decompress
 * @param in the data to process.
 * @param out a buffer to write the result to
 * @return the compression object
 */
compress_t* compress_new(int type, gboolean compress,
                         buffer_t* in, buffer_t* out) {
  compress_t* comp;

  comp = g_malloc(sizeof(*comp));
  comp->in = in;
  comp->out = out;
  if (type == COMPRESS_ZLIB && !compress) {
    comp->init = z_decompress_init;
    comp->action = z_decompress_action;
    comp->destroy = z_decompress_destroy;
  } else if (type == COMPRESS_ZLIB && compress) {
    comp->init = z_compress_init;
    comp->action = z_compress_action;
    comp->destroy = z_compress_destroy;
  } else if (type == COMPRESS_BZIP2 && !compress) {
    comp->init = bz_decompress_init;
    comp->action = bz_decompress_action;
    comp->destroy = bz_decompress_destroy;
  } else if (type == COMPRESS_BZIP2 && compress) {
    comp->init = bz_compress_init;
    comp->action = bz_compress_action;
    comp->destroy = bz_compress_destroy;
  } else if (type == COMPRESS_HUFFMAN && compress) {
    comp->init = hm_compress_init;
    comp->action = hm_compress_action;
    comp->destroy = hm_compress_destroy;
  } else {
    g_assert(FALSE);
    g_free(comp);
    comp = NULL;
  }
  if (comp && !comp->init(comp)) {
    comp->destroy(comp);
    comp = NULL;
  }
  return comp;
}

/** Start the de/compression.
 * @param comp the compression object
 * @param finish TRUE if finalize the compression, FALSE otherwise.
 * @return 1 on success, 0 otherwise
 */
int compress_action(compress_t* comp, gboolean finish) {
  return comp->action(comp, finish);
}

/** Destroy the compression object.
 * @param comp the compression object.
 */
void compress_destroy(compress_t* comp) {
  comp->destroy(comp);
  g_free(comp);
}



static void z_decompress_destroy(compress_t* decompress) {
  inflateEnd(decompress->stream.zstream);
  g_free(decompress->stream.zstream);
}

static gboolean z_decompress_init(compress_t* decompress) {
  int res;

  decompress->stream.zstream = g_malloc0(sizeof(z_stream));
  res = inflateInit(decompress->stream.zstream);
  if (res != Z_OK) {
    g_log("COMPRESS", G_LOG_LEVEL_WARNING,
	  "inflateInit error: %s (%d)", decompress->stream.zstream->msg, res);
    decompress->destroy(decompress);
    return FALSE;
  }
  return TRUE;
}

static int z_decompress_action(compress_t* infl, gboolean finish) {
  int res;
  buffer_t* in = infl->in;
  buffer_t* out = infl->out;
  z_stream* zip = infl->stream.zstream;

  zip->next_in = buffer_read_pos(in);
  zip->avail_in = buffer_avail(in);

  do {
    /* if there is no more output space left, create some more */
    if (!zip->avail_out) buffer_resize(out, out->max*2);
    zip->next_out = buffer_write_pos(out);
    zip->avail_out = buffer_space(out);
    
    res = inflate(zip, (!zip->avail_in && finish)?Z_FINISH:0);
    out->size = out->max - zip->avail_out;
    if (res == Z_STREAM_END) {
      break;
    } else if (res != Z_OK) {
      g_log("COMPRESS", G_LOG_LEVEL_WARNING,
            "decompress error: %s (error %d)\n", zip->msg, res);
      return 0;
    }
  } while (zip->avail_in > 0 || finish);

  buffer_consume_virt(in, in->size - zip->avail_in);
  return 1;
}






static void z_compress_destroy(compress_t* compress) {
  deflateEnd(compress->stream.zstream);
  g_free(compress->stream.zstream);
}

static gboolean z_compress_init(compress_t* compress) {
  int res;

  compress->stream.zstream = g_malloc0(sizeof(z_stream));
  res = deflateInit(compress->stream.zstream, Z_DEFAULT_COMPRESSION);
  if (res != Z_OK) {
    g_log("COMPRESS", G_LOG_LEVEL_WARNING,
	  "deflateInit error: %s (%d)", compress->stream.zstream->msg, res);
    return FALSE;
  }
  return TRUE;
}

static int z_compress_action(compress_t* defl, gboolean finish) {
  int res;
  buffer_t* in = defl->in;
  buffer_t* out = defl->out;
  z_stream* zip = defl->stream.zstream;

  zip->next_in = buffer_read_pos(in);
  zip->avail_in = buffer_avail(in);

  do {
    /* if there is no more output space left, create some more */
    if (!zip->avail_out) buffer_resize(out, out->max*2);
    zip->next_out = buffer_write_pos(out);
    zip->avail_out = buffer_space(out);
    
    res = deflate(zip, finish?Z_FINISH:0);
    out->size = out->max - zip->avail_out;
    if (res == Z_STREAM_END) {
      break;
    } else if (res != Z_OK) {
      g_log("COMPRESS", G_LOG_LEVEL_WARNING,
	    "deflate error: %s (%d)\n", zip->msg, res);
      return 0;
    }
  } while (zip->avail_in > 0 || finish);

  buffer_consume_virt(in, in->size - zip->avail_in);
  return 1;
}



static void bz_decompress_destroy(compress_t* decompress) {
  BZ2_bzDecompressEnd(decompress->stream.bzstream);
  g_free(decompress->stream.bzstream);
}

static gboolean bz_decompress_init(compress_t* decompress) {
  int res;

  decompress->stream.bzstream = g_malloc0(sizeof(bz_stream));
  res = BZ2_bzDecompressInit(decompress->stream.bzstream, 0, 0);
  if (res != BZ_OK) {
    g_log("COMPRESS", G_LOG_LEVEL_WARNING,
	  "bzDecompressInit error: (%d)", res);
    decompress->destroy(decompress);
    return FALSE;
  }
  return TRUE;
}

static int bz_decompress_action(compress_t* infl, gboolean finish) {
  int res;
  buffer_t* in = infl->in;
  buffer_t* out = infl->out;
  bz_stream* zip = infl->stream.bzstream;

  zip->next_in = buffer_read_pos(in);
  zip->avail_in = buffer_avail(in);

  do {
    /* if there is no more output space left, create some more */
    if (!zip->avail_out) buffer_resize(out, out->max*2);
    zip->next_out = buffer_write_pos(out);
    zip->avail_out = buffer_space(out);
    
    res = BZ2_bzDecompress(zip);
    out->size = out->max - zip->avail_out;
    if (res == BZ_STREAM_END) {
      break;
    } else if (res != BZ_OK) {
      g_log("COMPRESS", G_LOG_LEVEL_WARNING,
	    "decompress error: %d\n", res);
      return 0;
    }
  } while (zip->avail_in > 0 || finish);

  buffer_consume_virt(in, in->size - zip->avail_in);
  return 1;
}



static void bz_compress_destroy(compress_t* compress) {
  BZ2_bzCompressEnd(compress->stream.bzstream);
  g_free(compress->stream.bzstream);
}

static gboolean bz_compress_init(compress_t* compress) {
  int res;

  compress->stream.bzstream = g_malloc0(sizeof(bz_stream));
  res = BZ2_bzCompressInit(compress->stream.bzstream, 6, 0, 0);
  if (res != BZ_OK) {
    g_log("COMPRESS", G_LOG_LEVEL_WARNING,
	  "bzCompressInit error: (%d)", res);
    compress->destroy(compress);
    return FALSE;
  }
  return TRUE;
}

static int bz_compress_action(compress_t* defl, gboolean finish) {
  int res;
  buffer_t* in = defl->in;
  buffer_t* out = defl->out;
  bz_stream* zip = defl->stream.bzstream;

  zip->next_in = buffer_read_pos(in);
  zip->avail_in = buffer_avail(in);

  do {
    /* if there is no more output space left, create some more */
    if (!zip->avail_out) buffer_resize(out, out->max*2);
    zip->next_out = buffer_write_pos(out);
    zip->avail_out = buffer_space(out);
    
    res = BZ2_bzCompress(zip, finish?BZ_FINISH:BZ_RUN);
    out->size = out->max - zip->avail_out;
    if (res == BZ_STREAM_END) {
      break;
    } else if (res != (finish?BZ_FINISH_OK:BZ_RUN_OK)) {
      g_log("COMPRESS", G_LOG_LEVEL_WARNING,
	    "compress error: (%d)\n", res);
      return 0;
    }
  } while (zip->avail_in > 0 || finish);

  buffer_consume_virt(in, in->size - zip->avail_in);
  return 1;
}



static void hm_compress_destroy(compress_t* compress) {
  (void)compress;
}

static gboolean hm_compress_init(compress_t* compress) {
  (void)compress;
  return TRUE;
}

static int hm_compress_action(compress_t* defl, gboolean finish) {
  (void)finish;
  if (huffman_encode_memory(defl->in, defl->out)) {
    g_log("COMPRESS", G_LOG_LEVEL_WARNING,
	  "could not huffman encode");
    buffer_consume_virt(defl->in, buffer_avail(defl->in));
    return 0;
  }
  return 1;
}

#ifdef WIN32
# define _ZBUFFER_C_
# include "lcorealiasdef.c"
#endif
