/* Copyright (C) 2005-2006 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.4 $
 * $Date: 2007/11/15 14:18:12 $
 * $Author: sid4u $
 */

#include <string.h>
#include <glib.h>
#include <stdio.h>

#include "bt_benc.h"

/**
 * This is a beconding utility to decode from bencoded data,
 * all relative Strings, Integers, Lists and Dictionnaries
*/

static bt_value_t* bt_benc_parse_rec(const char** buffer, unsigned* size);

static bt_value_t* bt_benc_value_new(unsigned type)
{
    bt_value_t* value = g_malloc0(sizeof(*value));
    value->type = type;
    return value;
}

bt_value_t* bt_benc_value_find_dic(bt_value_t* value, char* name)
{
    GList* dlist;

    if (value->type != BT_BENC_DIC) return NULL;
    for (dlist = value->pairs; dlist; dlist = dlist->next->next)
    {
        bt_value_t* v1 = dlist->data;
        bt_value_t* v2 = dlist->next->data;
        if (!strcmp(v1->str, name)) return v2;
    }
    return NULL;
}

gboolean bt_benc_value_get_string(bt_value_t* value, char** str)
{
    if (value->type != BT_BENC_STR) return FALSE;
    if (*str) g_free(*str);
    *str = g_strdup(value->str);
    return TRUE;
}

gboolean bt_benc_value_get_int(bt_value_t* value, gint64* val)
{
    if (value->type != BT_BENC_INT) return FALSE;
    *val = value->number;
    return TRUE;
}


static void bt_benc_value_free(bt_value_t* value)
{
    switch (value->type) {
    case BT_BENC_INT:
        break;
    case BT_BENC_STR:
        g_free(value->str);
        break;
    case BT_BENC_DIC:
        g_list_foreach(value->pairs, (GFunc)bt_benc_value_free, NULL);
        break;
    case BT_BENC_LIST:
        g_list_foreach(value->pairs, (GFunc)bt_benc_value_free, NULL);
        break;
    default:
        g_assert(FALSE);
    }
    g_free(value);
}



static unsigned D = 0;

static bt_value_t* bt_benc_parse_str(const char** buffer, unsigned* size)
{
    bt_value_t* str;
    unsigned res = 0;

    while (*size && **buffer != ':')
    {
        res = res*10 + (**buffer)-'0';
        (*buffer)++; (*size)--;
    }

    if (*size == 0)
    {
        // g_message("no size");
        return NULL;
    }
    (*buffer)++; (*size)--;
  
    if (res > *size)
    {
        // g_message("size %u / %u", res, *size);
        return NULL;
    }
    
    str = bt_benc_value_new(BT_BENC_STR);
    str->str = g_malloc0(res+1);
    memcpy(str->str, *buffer, res);
    str->str_size = res;
    // g_message("%*s>>%s<<", D, " ", res>100?"<data>":str->str);

    (*buffer) += res;
    (*size) -= res;
    
    return str;
}

static bt_value_t* bt_benc_parse_int(const char** buffer, unsigned* size)
{
    bt_value_t* value;
    gboolean sign = FALSE;
    
    if (*size == 0) return NULL;

    value = bt_benc_value_new(BT_BENC_INT);
    value->number = 0;

    if (**buffer == '-')
    {
        sign = TRUE;
        (*buffer)++; (*size)--;
    }

    while (*size && **buffer != 'e')
    {
        value->number = value->number*10 + (**buffer)-'0';
        (*buffer)++; (*size)--;
    }
    if (sign) value->number = -value->number;
    
    if (**buffer != 'e')
    {
        bt_benc_value_free(value);
        value = NULL;
    }
    if (*size) 
    {
        (*buffer)++; (*size)--;
    }
    // g_message("%*s>>%ld<<", D, " ", value->number);
    
    return value;
}

static bt_value_t* bt_benc_parse_list(const char** buffer, unsigned* size)
{
    bt_value_t* list = bt_benc_value_new(BT_BENC_LIST);

    // g_message("%*sparse list (%u %p)", D, " ", *size, *buffer);
    while (*size && **buffer != 'e')
    {
        bt_value_t* v1 = bt_benc_parse_rec(buffer, size);
        if (v1) list->list = g_list_append(list->list, v1);
    }
    // g_message("%*sparse list end (%u %p)", D, " ", *size, *buffer);
    if (**buffer != 'e')
    {
        bt_benc_value_free(list);
        list = NULL;
    }
    if (*size)
    {
        (*buffer)++; (*size)--;
    }
    return list;
}

static bt_value_t* bt_benc_parse_dic(const char** buffer, unsigned* size)
{
    bt_value_t* dic = bt_benc_value_new(BT_BENC_DIC);
    
    // g_message("%*sparse dic (%u %p)", D, " ", *size, *buffer);
    while (*size && **buffer != 'e')
    {
        bt_value_t* v1 = bt_benc_parse_rec(buffer, size);
        bt_value_t* v2 = bt_benc_parse_rec(buffer, size);
        
        if (v1 && v2 && v1->type == BT_BENC_STR)
        {
            dic->pairs = g_list_append(dic->pairs, v1);
            dic->pairs = g_list_append(dic->pairs, v2);
        }
        else
        {
            if (v1) bt_benc_value_free(v1);
            if (v2) bt_benc_value_free(v2);
        }
    }
    // g_message("%*sparse dic end (%u %p)", D, " ", *size, *buffer);
    if (**buffer != 'e')
    {
        bt_benc_value_free(dic);
        dic = NULL;
    }
    if (*size)
    {
        (*buffer)++; (*size)--;
    }
    return dic;
}

static bt_value_t* bt_benc_parse_rec(const char** buffer, unsigned* size)
{
    const char* dic_begin_memory_offset;
    bt_value_t* v;

    if (*size == 0) return NULL;
    
    D += 4;
    dic_begin_memory_offset = *buffer;

    switch (**buffer)
    {
    case 'd':
        (*buffer)++; (*size)--;
        v = bt_benc_parse_dic(buffer, size);
        break;
    case 'l':
        (*buffer)++; (*size)--;
        v = bt_benc_parse_list(buffer, size);
        break;
    case 'i':
        (*buffer)++; (*size)--;
        v = bt_benc_parse_int(buffer, size);
        break;
    default:
        v = bt_benc_parse_str(buffer, size);
    }

    if (v)
    {
        v->dic_begin_memory_offset = dic_begin_memory_offset;
        v->dic_end_memory_offset = (*buffer);
    }
    else
    {
        // g_message("%*sNO VALUE!", D, " ");
    }
    D -= 4;
    return v;
}

void bt_benc_destroy(bt_value_t* torrent)
{
    bt_benc_value_free(torrent);
}

static void bt_benc_value_dump(bt_value_t* value)
{
    GList* dlist;

    if (!value) return;

    switch (value->type)
    {
    case BT_BENC_INT:
        printf("%ld", value->number);
        break;
    case BT_BENC_STR:
        if (strlen(value->str) > 50) printf("'<data>'");
        else printf("'%s'", value->str);
        break;
    case BT_BENC_DIC:
        printf("{");
        for (dlist = value->pairs; dlist; dlist = dlist->next)
        {
            bt_value_t* v = dlist->data;
            bt_benc_value_dump(v);
            printf(" => ");
            dlist = dlist->next;
            v = dlist->data;
            bt_benc_value_dump(v);
            printf("\n");
        }
        printf("}\n");
        break;
    case BT_BENC_LIST:
        printf("[");
        for (dlist = value->list; dlist; dlist = dlist->next)
        {
            bt_value_t* v = dlist->data;
            bt_benc_value_dump(v);
        }
        printf("]");
        break;
    default:
        g_assert(FALSE);
  }
}


bt_value_t* bt_benc_parse_from_data(const char* buffer, unsigned size)
{
    bt_value_t* value = bt_benc_parse_rec(&buffer, &size);
    return value;
}

/* This method can be used to display the parsing results (testing purposes) */
void bt_benc_parse_from_data_dumper(const char* buffer, unsigned size)
{
    bt_value_t* value = bt_benc_parse_rec(&buffer, &size);
    bt_benc_value_dump(value);
    bt_benc_destroy(value);
}

