/* Copyright (C) 2005-2006 (C) Lopster Team
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.3 $
 * $Date: 2006/07/30 18:39:15 $
 * $Author: sid4u $
 */

#include <stdlib.h>
#include <string.h>

#include "bt_torrent.h"
#include "bt_peer.h"
#include "bt_peer_message.h"
#include "bt_users.h"
#include "bt_download.h"
#include "bt_peer_protocol.h"
#include "bt.h"

typedef struct {
    module_t _module;
    GList* sources;
} bt_download_module_t; 

static bt_download_module_t* BtDownloadModule = NULL;

#define BT_SIZE_REQUEST 1024*32

#define HANDLER(f) static const char* f (bt_source_t* source, bt_peer_message_t* message)
#define BT_CALL(f) f(source, command)

static void bt_source_destroy(bt_source_t* source)
{
    BtDownloadModule->sources = g_list_remove(BtDownloadModule->sources, source);
    source_destroy(SOURCE(source));
}
static void bt_download_clean_and_peering(bt_source_t* source );
static const char* bt_source_consume_input(bt_source_t* source);
static segment_t* bt_download_check_available_segment(bt_source_t* source);
static void bt_download_send_peer_request_real(bt_source_t* source, segment_t* segment );
static void bt_download_check_and_update_pieces(bt_peer_t* peer);


/*
    FIXME : each source_end result in a check if download is finished.
    in case of bt downloads, that is quite different.
    a download is a file_t. 
    A file_t here is composed of one ore more pieces..and also part of pieces known as blocks (segments) .
    In order to say that a download is done, we need to validate each pieces it's composed of (SHA1)
    if a download is composed of full pieces, every thing is okay, as pieces are checked at every ST_DL_SEG_FIN.
    Else, in case of a download composed of multiple pieces and some of them are truncated :
    let's say a piece is 5 length... where a, b ,c , d , e , f are pieces.
    the following file is composed of these pieces as followed : 
    aaabbbbbccccceeeeeff
    
    we can see that piece "a" and piece "f" are not fully complete in this file...
    so we need also the file(s) preceding and following this one.
    we say file(s) with an "s" here, as these can be small files, smaller than a piece length.
    And we should find a way to read a whole segment from different files or memory mapped files.

    So here's the deal...if we have a segment(block) that is part of a piece, and this segment is last of this piece 
    if this piece is covered by the file (meaning in between)...just check it and on succcess source_end.
    if we get a segment , we know is last part of a piece...cf first piece of example upper
    we should take care not to call source_end...make a pending source_end
    until this piece get done on the other file(s)
    
    Make a static list of pending sources to check and end
    each incomplete piece, need to check for there pending 

*/
static gboolean bt_source_end(bt_source_t* source, const char* message)
{
    segment_t* segment;
    unsigned status = socket_get_status(SOCK(source));
    
    switch (status) {
    case ST_SOURCE_REMOTE:
    case ST_DL_SEG_FIN:
        // Check if source->last_segment (current ending segment) is last of one (or two) piece's block
        // do a hash check of piece(s) in case
        // how to get the piece(s) assiociated with this segment.
        // segment is from a download. so check for related pieces of a download.
        // then check the segment->start+size and segment->end position 
        // else just update piece information
        g_assert(source->last_segment);
                    
        segment = bt_download_check_available_segment(source);
        if (segment)
        { 
            source->last_segment = segment;
            b_socket_set_consume(BSOCK(source), (BConsumeInputFunc)bt_source_consume_input);
            bt_download_send_peer_request_real(source, segment);
            b_socket_start_receiving(BSOCK(source), 10240);
            return FALSE;
        }
        // FIXME : check if the current segment's piece is done (last block) and start a sha1 check, 
        // send HAVE in case ,Update Bitfield, Cancel any other requests for that piece,
        // Update interested to PEER
        bt_download_check_and_update_pieces(source->peer);
        bt_download_clean_and_peering(source);
        return FALSE;
    default:
        return source_end(SOURCE(source), message);
    } 
}

// Move back socket to peer and go back to consuming peer messages with bt_peer_consume
static void bt_download_clean_and_peering(bt_source_t* source )
{
    tcp_socket_move(TSOCK(source->peer), TSOCK(source) , TRUE);
    b_socket_set_consume(BSOCK(source->peer), (BConsumeInputFunc)bt_peer_consume);
    b_socket_start_receiving(BSOCK(source->peer), 0);    
}

static void bt_download_check_and_update_pieces(bt_peer_t* peer)
{
    bt_user_t* user = peer->user;
    bt_torrent_t* torrent = user->am_interested_torrent->torrent;
    
    g_assert(torrent);

    // now we have the torrent we share thoughts on, let's check for finished pieces
    
}

HANDLER(handle_block_message)
{
    (void)message;
    g_message("handle_block");
    // FIXME :  peer_message have also 2 other parameters that are begin and index.
    //          We need to check wether the current used segment match the one peer sent
    return source_start_downloading(SOURCE(source));
}

// Fixme : Is this handling needed here ? We could get suddenly a choke message, so maybe needed. 
// To be double check
HANDLER(handle_ckoke_message)
{
    (void)message;
    g_message("handle_ckoke_message");
    // Stop requesting, clean the source and go back to peer discussion, until next unchoke
    bt_download_clean_and_peering(source);
    return NULL;
}

static const char* bt_handle_source_message(bt_source_t* source, bt_peer_message_t* command)
{
  if (command->id == BT_MESSAGE_UNKNOWN)
  {
      g_log("BT_DOWN", G_LOG_LEVEL_DEBUG,
            "got invalid peer message during source download");
      return "got invalid command";
  }
  else if (command->id == BT_MESSAGE_PIECE)
  {
      return BT_CALL(handle_block_message);
  }
  else if (command->id == BT_MESSAGE_CHOKE)
  {
      return BT_CALL(handle_ckoke_message);
  }
  else
  {
      g_log("BT_DOWN", G_LOG_LEVEL_DEBUG, "SOURCE sent an unhandled message while downloading with id '%d' ", command->id);
  }
  return NULL;
}

static const char* bt_source_consume_input(bt_source_t* source)
{
    int last_id = -1;
    buffer_t* buffer;
    bt_peer_message_t* result;
    
    buffer = b_socket_get_in_buffer(BSOCK(source));
 
    if (!buffer) return NULL;
    while ((result = bt_consume_message(buffer, -1, last_id)) != NULL)
    {
        const char* message = bt_handle_source_message(source, result);
        last_id = result->id;
        bt_peer_message_free(result);
        if (message) return message;
    }
    buffer_consume(buffer, 0);
    return NULL;
}

static segment_t* bt_download_check_available_segment(bt_source_t* source)
{
   return source_next_segment(SOURCE(source));
}

static void bt_download_send_peer_request_real(bt_source_t* source, segment_t* segment)
{
    guint32 index = segment->start / 32768;
    guint32 begin = segment->start % 32768;
    guint32 size = segment->stop-segment->start;

    //FIXME: make sure that segment is not larger than 32 bits.
    bt_peer_send(BSOCK(source), BT_MESSAGE_REQUEST, index, begin, size);
}

//FIXME : make the max_nb_request_per_peer dynamic
static gboolean bt_download_send_peer_request(bt_peer_t* peer, bt_source_t* source)
{
    int length;
    int i;
    int nb_request;
    int length_request;
    segment_t* segment;

    length_request = BT_SIZE_REQUEST;

    if ( (BT_SIZE_REQUEST * 5) <= (source->req_stop - source->req_start))
    {
        nb_request = 5;
    }
    else 
    {
        nb_request = (source->req_stop - source->req_start) / ((file_size_t) BT_SIZE_REQUEST );
        if (nb_request == 0)
        {
            nb_request = 1;
            length_request = (source->req_stop - source->req_start) % ((file_size_t) BT_SIZE_REQUEST );
        }
    }        

    // We know how many request we should queue nb_request segments
    length = 0;
    for (i = 0 ; i < nb_request ; i++)
    { 
        source_use_segment(SOURCE(source), source->req_start+length, source->req_start+length+length_request, TRUE);
        length += BT_SIZE_REQUEST;
    }
    
    segment = bt_download_check_available_segment(source);
    if (!segment) return FALSE;
    // save a record of last segment used (for piece checking)
    source->last_segment = segment;
    bt_download_send_peer_request_real(source, segment);
    return TRUE;
}


//TODO :    this kind od segment request would get us random blocks and pieces.
//          There are different kind of strategies when downloading pieces.
//          Rarest first is one of the most used, when a download_request_segment(offset) will be ready
//          this kind of strategy will be implemented and deployed.
//          For now we use what download_request_segment gives us
//          But in fact requesting a segment to peer that peer doesn't have, will just close the connection
static const char* on_peer_connected(bt_peer_t* peer, bt_source_t* source)
{
    segment_t* seg;
    source->peer = peer;
    // Fixme : request segment before moving socket in case we don't get anything we just return
    tcp_socket_move(TSOCK(source), TSOCK(peer), TRUE);
    seg = source_request_segment(SOURCE(source), MAX_PEER_BLOCK_REQUEST * (32 * 1024) );
    if (!seg) {
        socket_done(SOCK(source), ST_ERROR, "no segment");
        return "no segment";
    }
    source->req_start = seg->start;
    source->req_stop = seg->stop;
    if( bt_download_send_peer_request( peer, source) == FALSE )
    {
        tcp_socket_move(TSOCK(peer),TSOCK(source), TRUE);
        //FIXME : should maybe free the segment used
        return "Couldn't request piece to peer";
    }
    b_socket_set_consume(BSOCK(source), (BConsumeInputFunc)bt_source_consume_input);
    b_socket_start_receiving(BSOCK(source), 10240);
    
    return "handover->bt_source";
}


static gboolean bt_source_start(bt_source_t* source)
{
    bt_torrent_t* torrent;
    bt_user_t* user = BT_USER(SOURCE(source)->user);   
    torrent = bt_user_get_am_interested_torrent(user);
    // TODO: check if not already connected to user
   // if (user->connected == TRUE) return TRUE;
    if (user->state == PEER_CONNECTED) return TRUE;
    // Send handshake (expect user to send back handshake)
    // and if user send back a good handshake, the callback bt_on_peer_connected is called
    bt_peer_request(user, torrent,
                    (BtPeerConnected)on_peer_connected, OBJ(source));
    return TRUE;
}


static void bt_source_init(bt_source_t* source, file_t* file, bt_user_t* user)
{
    source_init(SOURCE(source), file, USER(user), FALSE);
    object_add_type(OBJ(source), "BtSource");
    OBJ(source)->destroy = (ObjectFunc)bt_source_destroy;
    socket_set_end(SOCK(source), (SockEndFunc)bt_source_end);
    SOURCE(source)->start = (SourceStartFunc)bt_source_start;
    socket_add_rate(SOCK(source), user->rate);
    BtDownloadModule->sources = g_list_append(BtDownloadModule->sources, source);
}

static bt_source_t* bt_source_new(file_t* file, bt_user_t* user)
{
    bt_source_t* source = g_malloc0(sizeof(*source));
    bt_source_init(source, file, user);
    return source;
}

static bt_source_t*
bt_file_to_source(const download_t* download, file_t* file, bt_user_t* user)
{
    bt_source_t* source = bt_source_new(file, user);
    (void)download;
    return source;
}


static void bt_downloads_init()
{
    source_factory_register(BT_ID, (FileToSourceFunc)bt_file_to_source);
}

static void bt_downloads_run()
{
}

static void bt_downloads_exit()
{
}

static void bt_download_module_init(bt_download_module_t* module)
{
    module_init(MODULE(module), "Bt Download Module",
                (ModuleFunc)bt_downloads_init,
                (ModuleFunc)bt_downloads_run,
                (ModuleFunc)bt_downloads_exit);
    
    module_require(download_module_instance);
}

module_t* bt_download_module_instance()
{
    if (!BtDownloadModule)
    {
        BtDownloadModule = g_malloc0(sizeof(*BtDownloadModule));
        bt_download_module_init(BtDownloadModule);
    }
    return MODULE(BtDownloadModule);
}

#undef HANDLER
#undef BT_CALL
