/* Copyright (C) 2005 The Lopster Team
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */


/* $Revision: 1.7 $
 * $Date: 2007/09/24 20:46:59 $
 * $Author: sid4u $
 */

#include <string.h>
#include <sys/types.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#endif

#include <libxml/parser.h>

#include "lcore/version.h"
#include "lcore/l_base.h"
#include "lcore/xml_handler.h"
#include "lcore/b_socket.h"
#include "lcore/l_listen.h"
#include "lcore/file_tree.h"
#include "lcore/library.h"
#include "lcore/preferences.h"
#include "lcore/l_resolve.h"
#include "lcore/ui.h"
#include "lcore/l_http.h"
#include "lcore/l_i18n.h"
#include "bt.h"
#include "bt_torrent.h"
#include "bt_tracker.h"
#include "bt_users.h"
#include "bt_peer.h"
#include "bt_download.h"


typedef struct {
    module_t _module;
    GList* torrents;
    gboolean auto_connect;
    gboolean timer_active;
    unsigned cnt_online;
    unsigned cnt_all;
    listen_t* listen;
    gboolean save;
    //GList* users; // List of updated user_t (each user are related to one or more torrents
} bt_net_module_t;


static bt_net_module_t* BtNetModule = NULL;
static long scheduler_count = 0; 
#define SCHEDULER_ONE_SECOND_INTERVAL   1
#define SCHEDULER_TEN_SECOND_INTERVAL   10

static void bt_update_connect_status(void)
{
    int status = ONLINE_STATUS_IDLE;
    if (BtNetModule->auto_connect) status |= ONLINE_STATUS_AUTO;
    if (BtNetModule->cnt_online > 0) status |= ONLINE_STATUS_ONLINE;
    plugin_set_online_status(BT_ID, status);
}

static void bt_net_set_torrent(bt_torrent_t* torrent, int status)
{
    if (torrent->status == BT_TORRENT_RUNNING) BtNetModule->cnt_online--;
    torrent->status = status;
    if (torrent->status == BT_TORRENT_RUNNING) BtNetModule->cnt_online++;
    bt_update_connect_status();
}

void bt_net_start_torrent(bt_torrent_t* torrent)
{
    if (torrent->status == BT_TORRENT_RUNNING) return;

    bt_net_set_torrent(torrent, BT_TORRENT_RUNNING);
    // Start each trackers (if multiples).
    // each tracker would set itself to STOPPED in case of error and update won't be made anymore (until next restart)
    g_list_foreach(torrent->trackers, (GFunc)bt_tracker_start_request, NULL);
}

void bt_net_stop_torrent(bt_torrent_t* torrent)
{
    if (torrent->status == BT_TORRENT_STOPPED) return;

    bt_net_set_torrent(torrent, BT_TORRENT_STOPPED);
    
    g_list_foreach(torrent->trackers, (GFunc)bt_tracker_stop_tracker_request, NULL);
    // TODO: Close each connection from each peers properly
    // TODO: Close the download (file descriptor)
    // TODO: XML Save state
}

// What difference between PAUSE and STOP ? (Many bittorrent clients make the difference)
// Sgop: i currently implemented it this way: paused torrents are not affected when the plugin
// is started or stopped. So, paused is something like 'deactivated'
void bt_net_pause_torrent(bt_torrent_t* torrent)
{
    if (torrent->status == BT_TORRENT_PAUSED) return;
    torrent->status = BT_TORRENT_PAUSED;

    g_list_foreach(torrent->trackers, (GFunc)bt_tracker_stop_tracker_request, NULL);
}


/* 
 * Stop torrent with STOP flag
 * Would PAUSE ? torrent with PAUSE flags
 */
static gboolean bt_net_connect_next(void)
{
    GList* dlist;
    
    g_assert(BtNetModule);
    for (dlist = BtNetModule->torrents; dlist; dlist = dlist->next)
    {
        bt_torrent_t* torrent = dlist->data;
        if (torrent->status == BT_TORRENT_STOPPED)
        {
            bt_net_start_torrent(torrent);
            return TRUE;
        }
    }
    BtNetModule->timer_active = FALSE;
    return FALSE;
}



// If a torrent exist this is the next function to be called 
// to check wether the torrent is running and incomplete(then no need to add the torrent)
// If it's complete (not running then) user is the king if he wants to add it another time let it be
gboolean bt_network_torrent_can_redownload( bt_torrent_t* torrent) {
		//TODO: Check the diffrent position we're in, when should we say the torrent can be redownloaded ?
		return TRUE;
}


static void bt_network_torrent_remove(bt_torrent_t* torrent)
{
    bt_net_stop_torrent(torrent);
    BtNetModule->torrents = g_list_remove(BtNetModule->torrents, torrent);
    BtNetModule->cnt_all--;
    bt_update_connect_status();
}

static void bt_net_start_timer(void)
{
    if (!BtNetModule->timer_active)
    {
        BtNetModule->timer_active = TRUE;
        lopster_call_periodic(1000, (CallFunc)bt_net_connect_next, NULL);
    }
}

static void bt_net_stop_timer(void)
{
    if (BtNetModule->timer_active)
    {
        lopster_cancel_call((CallFunc)bt_net_connect_next);
        BtNetModule->timer_active = FALSE;
    }
}

static void bt_net_start(void)
{
    BtNetModule->auto_connect = TRUE;
    bt_net_start_timer();
}

static void bt_net_stop(void)
{
    GList* dlist;

    BtNetModule->auto_connect = FALSE;
    bt_net_stop_timer();

    for (dlist = BtNetModule->torrents; dlist; dlist = dlist->next)
    {
        bt_torrent_t* torrent = dlist->data;
        if (torrent->status == BT_TORRENT_RUNNING)
            bt_net_stop_torrent(torrent);
    }
}

void bt_connect_change(unsigned status)
{
    switch (status)
    {
    case CONNECT_START:
        bt_net_start();
        break;
    case CONNECT_STOP:   // the core doesnt support this value currently
        bt_net_stop();
        break;
    case CONNECT_DISCONNECT:
        bt_net_stop();
        break;
    }
}

static void bt_network_torrent_new(bt_torrent_t* torrent)
{
    g_assert(BtNetModule);
    g_assert(torrent);
    //FIXME: set status according to autconnect
    BtNetModule->torrents = g_list_append(BtNetModule->torrents, torrent);
    object_notify_d(OBJ(torrent), (ObjectFunc)bt_network_torrent_remove, NULL);

    // we have new stopped torrent, so activate timer if in autoconnect mode
    BtNetModule->cnt_all++;
    bt_update_connect_status();

    if (BtNetModule->auto_connect) bt_net_start_timer();
}


static void bt_network_load() {
    BtNetModule->save = FALSE;

    if (!lopster_xml_load("bt:torrents")) {
        BtNetModule->save = TRUE;
    }

    BtNetModule->save = TRUE;
}


static tcp_socket_t* bt_accept_con(int fd, ip_t ip, port_t port) {
    tcp_socket_t* socket = tcp_socket_new("BtTemp", 30);
  
    listen_init_accepted(socket, fd, ip, port);
    // TODO : start incoming peer dialog here
    return socket;
}

port_t bt_get_listen_port() {
  g_assert(BtNetModule);
  return listen_port_get_port(BtNetModule->listen);
}

static void bt_change_port() {
    // TODO : change port listenning parameter on http tracker request, 
    // and also send optional listenning port message to connected peers
}

static void bt_start_listen_service() {
    g_assert(BtNetModule);
    BtNetModule->listen =
        listen_port_new("/bt/listen/port", "BtNetModule->listen",
                        (ListenAcceptFunc)bt_accept_con,
                        (ListenChangeFunc)bt_change_port);
    object_set_ptr((object_t**)&BtNetModule->listen);
    listen_port_start_listening(BtNetModule->listen, 10);
}


static void bt_network_do_peers_connection_checks(bt_torrent_t* torrent)
{
    GList* dlist;
    GList* dlist2;
    int peer_connecting = 0;
    int allowed;

    if (scheduler_count % SCHEDULER_ONE_SECOND_INTERVAL == 0)
    {
        // every one second : connect to some peers
            // get a number of peers to connect to
            // check for each tracker number of users
            // divide the number of users to connect to, getting them from the tracker
            // for each download from the torrent add some sources
        //g_log("BT_NET", G_LOG_LEVEL_DEBUG,"bt_network_do_peers_connection_checks scheduler_count % SCHEDULER_ONE_SECOND_INTERVAL == 0 ");    

        for (dlist = torrent->trackers; dlist; dlist = dlist->next)
        {
            bt_tracker_t* tracker = (bt_tracker_t*) dlist->data;
            for (dlist2 = tracker->peers; dlist2; dlist2 = dlist2->next)
            {
                bt_user_t* user = (bt_user_t*) dlist2->data;
                if (user->state == PEER_CONNECTING)
                {
                    peer_connecting++;
                }
            }
        }
        //g_log("BT_NET", G_LOG_LEVEL_DEBUG,"bt_network_do_peers_connection_checks peer_connecting DONE ");    

        // FIXME : Add a plugin method plugin_get_allowed_connection(ID) 
        // to get the max number of connection allowed for this plugin
    #define MAX_PLUGIN_ALLOWED  300
        allowed = MAX_PLUGIN_ALLOWED;

        if( allowed < 0 || allowed > 1000 )  allowed = 1000;  

        // FIXME : Add a method in core to check if plugin listener is healthy meaning we can accept incoming
    #define INCOMING_AVAILABLE  TRUE
        if( INCOMING_AVAILABLE ) {  
            int free = MAX_PLUGIN_ALLOWED / 20;  //leave 5%
            allowed = allowed - free;
        }

        //FIXME : Add a parameter in plugin preferences MAX_SIMULTANIOUS_CONNECT_ATTEMPTS
    #define MAX_SIMULTANIOUS_CONNECT_ATTEMPTS   50
        if( allowed > 0 ) {
            int wanted = MAX_SIMULTANIOUS_CONNECT_ATTEMPTS - peer_connecting;
            if( wanted > allowed ) {
                peer_connecting += wanted - allowed;
            }
           //g_log("BT_NET", G_LOG_LEVEL_DEBUG,"allowed > 0  ");    

            //load stored peer-infos to be established
            while( peer_connecting < MAX_SIMULTANIOUS_CONNECT_ATTEMPTS ) {
                download_t* download;
                // FIXME : for the moment get_next_optimistic_user return first non connected user from list
                bt_user_t* user = bt_torrent_get_next_optimistic_user(torrent);
                if( user == NULL )  break;
                if( user->state < PEER_START && user->ip_resolved == TRUE  ) {
                    source_t* source;
                    user->state = PEER_START;
                    // FIXME : for the moment bt_torrent_get_download_to_process return random download...change that
                    download = bt_torrent_get_download_to_process(torrent);
                    if (download == NULL){
                        break;
                    }
                    bt_user_set_am_interested_torrent(user, torrent);
                    source = download_add_source(download, FILE(download), USER(user));
                    if( source != NULL ) {
                        peer_connecting++;
                    }
                }  
            } 
        } 
    } 


    // TODO: Every 5 seconds : send a keep alive to the user we should stay connected to

    // TODO: Every 10 seconds : check for user to ban (bad guys)

    // TODO: Every 30 seconds : if too much connection opened, close some optimistics

    // TODO : every 60 seconds : implement the exchange of peers in protocol
}


static gboolean bt_torrent_check( bt_torrent_t* torrent )
{
    bt_network_do_peers_connection_checks(torrent);
    
    // Most important part, is the un/choke, here all is done
    //FIXME : this will be added later

    if (scheduler_count % SCHEDULER_TEN_SECOND_INTERVAL == 0)
    {
   //       g_log("BT_NET", G_LOG_LEVEL_DEBUG,"bt_peer_do_unchokes %d", scheduler_count);    

   //     bt_peer_do_unchokes(torrent);
    }
    return TRUE;
}

static gboolean bt_torrent_scheduler( )
{
    GList* dlist;
    g_assert(BtNetModule);
    // upgrade the count for each seconds
    scheduler_count++;
    for (dlist = BtNetModule->torrents; dlist; dlist = dlist->next)
    {
        bt_torrent_t* torrent = dlist->data;
        if (torrent->status == BT_TORRENT_RUNNING)
        {
            bt_torrent_check(torrent);
        }
    }
    return TRUE;
}

static void bt_nets_init()
{
    object_register_instance("Torrent", (ObjectFunc)bt_network_torrent_new, FALSE);
    // TODO : register xml preference, bt_id, port to listen to, actions ?
/*
    lopster_xml_register_home("bt:torrents", SDIR("dc") "torrents.xml",
                              "Torrents", xml_load_torrents, xml_save_torrents);

    pref_register_string("/bt/username", TRUE, NULL);
    pref_add_constraint_replace("/bt/username", ' ', '_');

    pref_register_string("/bt/password", TRUE, NULL);
    pref_add_constraint_replace("/bt/password", ' ', '_');

    pref_register_int("/bt/listen/tcp_port", TRUE, 60001); // Fixme, will try to get user listen port first ?
    pref_add_constraint_port("/bt/listen/tcp_port");
*/
    lopster_call_at_exit(bt_net_stop);
}

static void bt_nets_run()
{
    // bt_network_load();
    // TODO: call function to load the network loop, listen on bt port (add a function to check from outside visibility of the port)
    //   bt_start_listen_service();
    
    // FIXME : uncomment it, when we are sure, parsing and tracker(connection-hhttp,https,udp- and parsing) are correct
    // lopster_call_periodic(1000,bt_torrent_scheduler,NULL); 
}

static void bt_nets_exit()
{
    //FIXME: free mem (that's for sure i forgot it on many places :))
    BtNetModule = NULL;
}

static void bt_net_module_init(bt_net_module_t* module)
{
    module_init(MODULE(module), "Bt Network Module",
                (ModuleFunc)bt_nets_init,
                (ModuleFunc)bt_nets_run,
                (ModuleFunc)bt_nets_exit);

    module->listen = NULL;
    module->save = TRUE;
    module->torrents = NULL;

    module_require(base_module_instance);
    module_require(xml_module_instance);
    module_require(tsocket_module_instance);
    module_require(resolve_module_instance);
    module_require(listen_module_instance);
    module_require(bt_tracker_module_instance);
    module_require(bt_torrent_module_instance);
    module_require(bt_download_module_instance);
}


module_t* bt_net_module_instance()
{
    if (!BtNetModule)
    {
        BtNetModule = g_malloc0(sizeof(*BtNetModule));
        bt_net_module_init(BtNetModule);
    }
    return MODULE(BtNetModule);
}
