/* Copyright (C) 2005-2006 The Lopster Team.
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.3 $
 * $Date: 2006/07/30 18:39:15 $
 * $Author: sid4u $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <stdlib.h>

#include "lcore/utils.h"
#include "lcore/library.h"

#include "bt_torrent.h"
#include "bt_peer.h"
#include "bt_utils.h"
#include "bt.h"
#include "bt_peer_message.h"
#include "bt_peer_protocol.h"

#define BT_REC(proto, level) \
   g_log(proto, level, "R: %d %s", peer_message->id, peer_message->message?peer_message->message:"NO MESSAGE");


#define HANDLER(f) static const char* f (bt_peer_t* peer, bt_peer_message_t* peer_message)
#define BT_CALL(f) f(peer, peer_message)
#define REQUIRE(expr, proto, message) \
  if (!(expr)) {g_log(proto, G_LOG_LEVEL_WARNING, message); return message;}


typedef struct {
  bt_user_t* user;
  BtPeerConnected connected;
  object_t* obj;
} bt_peer_req_t;

static GList* PeerRequests = NULL;
/* static GList* Peers = NULL; */

static void bt_peer_request_delete(bt_peer_req_t* req);
static bt_peer_req_t* bt_peer_request_find(bt_peer_t* peer);


static gboolean bt_peer_end(bt_peer_t* peer, const char* message)
{
    tcp_socket_end(TSOCK(peer), message);
    return TRUE;
}

static void bt_peer_destroy(bt_peer_t* peer) {
  if (peer->user) user_unref(USER(peer->user));
  tcp_socket_destroy(TSOCK(peer));
}

static const char* bt_handle_peer_message(bt_peer_t* peer, bt_peer_message_t* command);

const char* bt_peer_consume(bt_peer_t* peer)
{
    int length = -1 ;
    buffer_t* buffer = b_socket_get_in_buffer(BSOCK(peer));
    bt_peer_message_t* result;

    g_log("BT_PEER", G_LOG_LEVEL_DEBUG, "bt_handle_peer_message");

    if (!buffer) return NULL;
    // length here is the size of old data to consume
    while ((result = bt_consume_message(buffer, length, BT_MESSAGE_UNKNOWN)) != NULL)
    {
        const char* message = bt_handle_peer_message(peer, result);
        length =  result->length;
        bt_peer_message_free(result);
        if (message) return message;
    }
    buffer_consume(buffer, 0);
    return NULL;
}

void bt_peer_init(bt_peer_t* peer) {
  tcp_socket_init(TSOCK(peer), "BtPeer", 30);
  object_add_type(OBJ(peer), "BtPeer");
  OBJ(peer)->destroy = (ObjectFunc)bt_peer_destroy;
  socket_set_end(SOCK(peer), (SockEndFunc)bt_peer_end);
  b_socket_set_consume(BSOCK(peer), (BConsumeInputFunc)bt_peer_consume);
  peer->flags = 0;
  peer->user = NULL;
 // peer->state = PEER_INCOMING;
  socket_add_load(SOCK(peer), "Peer");
}

bt_peer_t* bt_peer_new() {
  bt_peer_t* peer = g_malloc0(sizeof(*peer));
  bt_peer_init(peer);
  return peer;
}

// FIXME: can be called directly from bt_peer_init as we have the user
static void bt_peer_set_user(bt_peer_t* peer, bt_user_t* user) {
  if (peer->user) user_unref(USER(peer->user));
  peer->user = user_ref(USER(user));
  object_set_name(OBJ(peer), OBJ(user)->name);
}

static gboolean bt_peer_check_interest_in_bitfield_piece( bt_torrent_t* torrent, int piece_index )
{
    bt_bitfield_t* torrent_bitfield;
    g_assert(torrent);
    torrent_bitfield = torrent->bitfield;
    g_assert(torrent_bitfield);
    return bt_bitfield_has_piece(torrent_bitfield,piece_index);
}

static gboolean bt_peer_check_interest_in_bitfield( bt_torrent_t* torrent, bt_bitfield_t* bitfield )
{
    bt_bitfield_t* torrent_bitfield;
    g_assert(torrent);
    g_assert(bitfield);
    torrent_bitfield = torrent->bitfield;
    g_assert(torrent_bitfield);
    return bt_bitfield_compare(torrent_bitfield,bitfield);
}

gboolean bt_peer_is_uploadable (bt_peer_t* peer)
{
    bt_user_t* user;
    bt_torrent_t* torrent;
    
    user = peer->user;
    torrent = bt_user_get_am_interested_torrent(user);

    g_assert(user);
    g_assert(torrent);

    if (user->am_choking || user->choking_me || !user->interested_me)
    {
       return FALSE;
    }
    return TRUE;
}

//static void bt_peer_send_bitfield( bt_peer_t* peer, bt_torrent_t* torrent );
HANDLER(handle_incoming_handshake) {
    char* info_hash;
    bt_user_t* user;
    bt_torrent_t* user_torrent;
    bt_bitfield_t* user_torrent_bitfield;

    g_message("handle_incoming_handshake");

    user = peer->user;
    user_torrent = bt_user_get_am_interested_torrent(user);
    user_torrent_bitfield = bt_user_get_am_interested_torrent_bitfield(user);

    if (user_torrent == NULL)
    {
        return "user_torrent null"; //Fixme : return maybe a message
    }
    if (user_torrent->status != BT_TORRENT_RUNNING)
    {
        return "This torrent is not running"; //Fixme : return maybe a message and drop connection, this torrent is not currently running
    }

    // Parse info_hash from peer handshake message
    info_hash = g_malloc(sizeof(char*) * SHA_DIGEST_LENGTH + 1);
    memcpy(info_hash,peer_message->message,SHA_DIGEST_LENGTH);
    info_hash[SHA_DIGEST_LENGTH] = 0;

    // Fixme : strcmp does not work sometimes, wonder why
    if (strcmp(user_torrent->info_hash, info_hash) != 0) 
    {
        return "the user info hash for this torrent doesn't match ours"; 
    }
    
    // TODO : math peer id with current connected ip (this is from tracker data)

    // now we know we share the same thoughts
   // user->connected = TRUE;
    user->state = PEER_CONNECTED;
    
    if( (bt_bitfield_empty(user_torrent->bitfield) == FALSE)  && ( bt_bitfield_has_all(user_torrent_bitfield) == FALSE ) )
    {
        g_log("BT_PEER", G_LOG_LEVEL_DEBUG,"handle_incoming_handshake -> bt_peer_send_bitfield");
/*FIXME: implement         bt_peer_send_bitfield(peer, user_torrent); */
    }

    // Don't send choke and not interested yet...let the scheduler handle that
/*
    bt_peer_send_choke(peer);
    bt_peer_send_not_interested(peer);
*/

    // TODO : 
    // Normally we should send every 2 minutes , a KEEP_ALIVE message.
    // This should be done normally if you still want to be connected to user (great bandwidth/latency user etc...)
    // g_timeout_add(keep_alive every 2 mn)??? only if peer is worth it
    g_free(info_hash);
    return NULL;
}

HANDLER(handle_incoming_keep_alive)
{
    (void)peer;
    (void)peer_message;

    g_message("handle_incoming_choke");
    //TODO : We got a keep alive from user, don't drop connection 
    // For now it's not dropping at all :)
    return NULL;
}

HANDLER(handle_incoming_choke)
{
    bt_user_t* user = peer->user;

    (void)peer_message;

    g_message("handle_incoming_choke");
    // TODO : peer choked us, we need to update it
    user->choking_me = TRUE;

    // TODO : clear any queued upload or download request on peer
    //          the choke message is also managed in download module

    // Any current downloading segment from peer may not be stopped, as it's an already sent message.
    // But it can be dropped from peer so we should handle that

    return NULL;
}

HANDLER(handle_incoming_unchoke)
{
    const char* res = NULL;
    bt_peer_req_t* req;
    bt_user_t* user = peer->user;

    (void)peer_message;

    g_message("handle_incoming_unchoke");
    user->choking_me = FALSE;

   // TODO : peer unchoked us
   // At this point we need to know if we still have some used segment in our files to that peer.
   // This needs to be checked

    (void)peer;
    //if (peer->flags & DC_PEER_DOWNLOAD) {
    req = bt_peer_request_find(peer);

    if (!req) return "don't need anything";

    if (user->am_interested == TRUE)
    {
        res = req->connected(peer, req->obj);
    }
    //   bt_peer_request_delete(req); // don't drop connection 
    //}
    return NULL;
    //   return res;
}

HANDLER(handle_incoming_interested)
{
    bt_user_t* user = peer->user;

    (void)peer_message;

    g_message("handle_incoming_interested");
    user->interested_me = TRUE;

    return NULL;
}

HANDLER(handle_incoming_not_interested)
{
    bt_user_t* user = peer->user;

    (void)peer_message;
    g_message("handle_incoming_not_interested");

    user->interested_me = FALSE;
    return NULL;
}

HANDLER(handle_incoming_have) {
    unsigned int raw_index_piece;
    int piece_index;
    bt_user_t* user;
    bt_torrent_t* user_torrent;
    bt_bitfield_t* user_torrent_bitfield;

    user = peer->user;
    user_torrent = bt_user_get_am_interested_torrent(user);
    user_torrent_bitfield = bt_user_get_am_interested_torrent_bitfield(user);
    if (user_torrent == NULL) 
    {
        return "user_torrent null"; //Fixme : return maybe a message
    }
    if (user_torrent->status != BT_TORRENT_RUNNING)
    {
        return NULL; //Fixme : return maybe a message and drop connection, this torrent is not currently running
    }

    // Parse the index of pieces sent
    
    memcpy(&raw_index_piece,peer_message->message,sizeof(raw_index_piece));
    piece_index=ntohl(raw_index_piece);
    bt_bitfield_set(user_torrent_bitfield,piece_index);

    // Always keep peer that we're interested if it's the case
    if ( bt_peer_check_interest_in_bitfield_piece(user_torrent,piece_index) )
    {
        //TODO : if peer become seeder with that piece , update status of that user accordingly

        bt_peer_send(BSOCK(peer), BT_MESSAGE_INTERESTED);
        user->am_interested = TRUE;
        
        // FIXME : unchoke needs to be send after checking some information on peer
        // For example the latency, number of bytes downloaded 
        // and also check reciprocity as did peer unchoked use, have we downloaded from him etc
        bt_peer_send(BSOCK(peer), BT_MESSAGE_UNCHOKE);
        user->am_choking = FALSE;
    }
    else
    {
        user->am_interested = FALSE;
    }
    return NULL;
}

HANDLER(handle_incoming_bitfield) {
    //TODO : peer sent his bitfield, update the whole bitfield lot
    char* bitfield;
    bt_user_t* user;
    bt_torrent_t* user_torrent;
    bt_bitfield_t* user_torrent_bitfield;

    user = peer->user;
    g_message("handle_incoming_bitfield");

    user_torrent = bt_user_get_am_interested_torrent(user);

    if (user_torrent == NULL) 
    {
        return "user_torrent null"; //Fixme : return maybe a message
    }
    if (user_torrent->status != BT_TORRENT_RUNNING)
    {
        return "This torrent is not running"; //Fixme : return maybe a message and drop connection, this torrent is not currently running
    }

    // Parse bitfield
    bitfield = malloc(sizeof(char*)*peer_message->length + 1);
    memcpy(bitfield,peer_message->message,peer_message->length);
	bitfield[peer_message->length] = '\0';
    user_torrent_bitfield = bt_bitfield_setup_from_peer(user_torrent->no_pieces, bitfield, peer_message->length ); 
    bt_user_set_interested_torrent_bitfield(user,user_torrent_bitfield);
    


    // Always keep peer that we're interested if it's the case
    if ( bt_peer_check_interest_in_bitfield(user_torrent,user_torrent_bitfield) )
    {
        g_log("BT_PEER", G_LOG_LEVEL_DEBUG,"handle_incoming_bitfield user->am_interested = TRUE"); 
        if (bt_bitfield_complete(user_torrent_bitfield))
        {
            user->seeder = TRUE;
            // Unchoke seeder right away as anytime it want to unchoke us we should be ready
            // For other peers, let the unchoker manager handle that
            bt_peer_send(BSOCK(peer), BT_MESSAGE_UNCHOKE);
            user->am_choking = FALSE;
        }
        
        bt_peer_send(BSOCK(peer), BT_MESSAGE_INTERESTED);
        user->am_interested = TRUE;   
    }
    else
    {
        user->am_interested = FALSE;
    }
    return NULL;
}

HANDLER(handle_incoming_request) {
/*     bt_user_t* user; */
/*     bt_torrent_t* torrent; */
 //   bt_upload_t* upload;
/*     file_t* file; */
/*     char* res; */
/*     file_size_t offset; */
/*     file_size_t start; */
/*     file_size_t stop; */

    g_message("handle_incoming_request");
    
    // FIXME : This part is quite uncommon to implement in the plugin. upload is NOT compilable for the moment
    /* This is how this should normally be handled : 
        - First check wether peer has no more than MAX_PEER_BLOCK_REQUEST pending else return NULL.
        - Then create if doesn't exist the upload_list (as defined in upload part, the upload_list use is necessary).
        - Fire upload directly.
        - As we should be aware, more message of request or other messages maybe sent from peer
          In that case in the upload part, there should be a handling to consume incoming messages too
          or other types of message. So handle them as they should be by reattaching the peer_consume_message to the SOCKET(upload)
    */
/*
    user = peer->user;
    torrent = bt_user_get_am_interested_torrent(user);

    if ( bt_peer_is_uploadable (peer) == FALSE )
    {
        return "request not compatible with peer state";
    }
    
    // TODO : Check wether the requested piece is available from us.
    //          Also the size requested
    
    // TODO : Check wether maximum requests per peer and max total request is reached 

    // TODO :   Others checks, like for example peer rate etc ? 
    //          should it be done here or during unchoke ?       
    
    // TODO : update file from torrent with piece parameter
    // file = bt_torrent_get_file_from_infos(index,begin,length);
    // Now we know peer can download from us. Fire the upload
    upload = bt_upload_new(peer->user, file);
    UPLOAD(upload)->base_priority = UPLOAD_PRI_URGENT;
    upload_insert(UPLOAD(upload), FALSE);

    // TODO : set start and stop according to index, begin and length parameters from peer_message
    bt_upload_set_offset(upload, start, stop);

    // This should only be done when there are no more request coming from buffer.
    res = bt_upload_set_connection(upload, peer);

    // TODO : DO NOT FORGET TO UPDATE TOTAL UPLOADED
*/
    return NULL;
}


HANDLER(handle_incoming_cancel) {
    //TODO : Someone has requested a block from us, but wants to cancel that request
    //TODO : Cancel this pending request in the queue if it still exists
    return NULL;
}

static const char*
bt_handle_peer_message(bt_peer_t* peer, bt_peer_message_t* peer_message)
{  
    if (peer_message->id == BT_MESSAGE_HANDSHAKE)
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_handshake);
    }
    else if (peer_message->id == BT_MESSAGE_KEEP_ALIVE)
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_keep_alive);
    }
    else if (peer_message->id == BT_MESSAGE_CHOKE)
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_choke);
    }
    else if (peer_message->id == BT_MESSAGE_UNCHOKE)
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_unchoke);
    }
    else if (peer_message->id == BT_MESSAGE_INTERESTED)
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_interested);
    }
    else if (peer_message->id == BT_MESSAGE_NOT_INTERESTED)
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_not_interested);
    }
    else if (peer_message->id == BT_MESSAGE_HAVE)
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_have);
    }
    else if (peer_message->id == BT_MESSAGE_BITFIELD)
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_bitfield);
    }
    else if (peer_message->id == BT_MESSAGE_REQUEST)
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_request);
    }
/*    else if (peer_message->id == BT_MESSAGE_PIECE) //This is handled in the bt_download module
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_piece);
    } */
    else if (peer_message->id == BT_MESSAGE_CANCEL) //Fixme : Should this also handled in bt_download_module ?
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_DEBUG);
        return BT_CALL(handle_incoming_cancel);
    }
    else
    {
        BT_REC("BT_PEER", G_LOG_LEVEL_WARNING);
    }
    return NULL;
}


static void bt_peer_connect_to_peer(bt_user_t* user, bt_torrent_t* torrent) 
{
  bt_peer_t* peer = bt_peer_new();
  bt_peer_set_user(peer,user);

  if (!tcp_socket_connect(TSOCK(peer), user_ip(user), user_port(user))) return;
  
  user->state = PEER_CONNECTING;

  // update list of connected peers. We'll use that list to un/choke through manager below  
  torrent->peers = g_list_append(torrent->peers, peer);

  bt_peer_send(BSOCK(peer), BT_MESSAGE_HANDSHAKE, torrent->info_hash, BT_PEER_ID);

  b_socket_start_receiving(BSOCK(peer), 0);
}

static void on_request_destroy(object_t* object, bt_peer_req_t* req) {
  (void)object;
  req->obj = NULL;
  bt_peer_request_delete(req);
}

static void bt_peer_request_delete(bt_peer_req_t* req)
{
    GList* dlist = g_list_find(PeerRequests, req);
    if (!dlist) return;  // not existen anymore.
    
    PeerRequests = g_list_delete_link(PeerRequests, dlist);
    if (req->obj)
    {
        object_unnotify_d(req->obj, (ObjectFunc)on_request_destroy, req);
    }
    user_unref(USER(req->user));
    g_free(req);
}

void bt_peer_request(bt_user_t* user, bt_torrent_t* torrent,
                     BtPeerConnected func, object_t* obj)
{
    bt_peer_req_t* req = g_malloc0(sizeof(*req));
    
    req->user = user_ref(USER(user));
    req->obj = obj;
    req->connected = func;
    PeerRequests = g_list_append(PeerRequests, req);
    
    if (obj)
    {
        object_notify_d(obj, (ObjectFunc)on_request_destroy, req);
    }
    
    bt_peer_connect_to_peer(user,torrent);
}

static bt_peer_req_t* bt_peer_request_find(bt_peer_t* peer) 
{
  GList* dlist;
  if (!peer) return NULL;
  for (dlist = PeerRequests; dlist; dlist = dlist->next) {
    bt_peer_req_t* req = dlist->data;
    if (req->user == peer->user) return req;
  }
  return NULL;
}



/* This is the peer manager and utils functions, it would choke/unchoke peer every 10 seconds
   depending on upload/download speed, interested or not
   Mostly copied and adapted to C from azureus client (from java code)
*/

//FIXME : MAX_TO_UNCHOKE would be equals to a bt plugin preference = MAX_BT_UPLOAD
#define MAX_TO_UNCHOKE          20

static gboolean bt_peer_is_unchokable( bt_peer_t* peer, gboolean allow_snubbed ) {
    bt_user_t* user;
    user = peer->user;
    return user->state == PEER_CONNECTED && !user->seeder && user->interested_me && ( !user->is_snubbed || allow_snubbed );
}

// UnchokerUtil.updateLargestValueFirstSort( rate, bests, peer, best_peers, start_pos );
static GList* bt_unchoker_update_largest_value_first_sort( int new_value, int values [] , bt_peer_t* new_item, GList* items, int start_pos ) {  
    int i, j;

    for( i = start_pos; i < MAX_TO_UNCHOKE ; i++ ) {
      if( new_value >= values[ i ] ) {
        for( j = MAX_TO_UNCHOKE - 2; j >= i; j-- ) {  //shift displaced values to the right
          values[j + 1] = values[ j ];
        }
        
        values[ i ] = new_value;
        g_list_insert(items, new_item, i );
        
        if( g_list_length(items) > MAX_TO_UNCHOKE ) {  //throw away last item if list too large 
          GList* list = g_list_last(items);
          bt_peer_t* peer = list->data;
          g_list_remove(items, peer );
        }
        return items;
      }
    }
    return items;
}

/*
public static bt_peer_t* bt_unchoker_get_next_optimistic_peer( GList* all_peers, gboolean factor_reciprocated, gboolean allow_snubbed ) {
    //find all potential optimistic peers
  //  ArrayList optimistics = new ArrayList();
    GList* optimistics; 
    GList* dlist;
    GList* result;
    for( dlist =  all_peers; dlist ; dlist = dlist->next ) {
      bt_peer_t* peer = (bt_peer_t*)dlist->data;
      
      if( bt_peer_is_unchokable( peer, FALSE ) && peer->am_choking ) {
        g_list_append(optimistics, peer); 
      }
    }

    if( g_list_length(optimistics) == 0 && allow_snubbed ) {  //try again, allowing snubbed peers as last resort
      for( dlist = all_peers; dlist ; dlist = dlist->next ) {
        bt_peer_t* peer = (bt_peer_t*)dlist->data;
        
        if( bt_peer_is_unchokable( peer, TRUE ) && peer->am_choking ) {
            g_list_append(optimistics, peer); 
        }
      }
    }

    if( g_list_length(optimistics) == 0 )  return NULL;  //no unchokable peers avail

    //factor in peer reciprocation ratio when picking optimistic peers
    if( factor_reciprocated ) {
        GList* ratioed_peers;
        int* ratios; //table of optimitics size
 

      //order by upload ratio
      for( dlist = optimistics; dlist ; dlist = dlist->next )  {
        bt_peer_t* peer = (bt_peer_t*)dlist->data;

        //score of >0 means we've uploaded more, <0 means we've downloaded more
        //FIXME C
       // long score = peer.getStats().getTotalDataBytesSent() - peer.getStats().getTotalDataBytesReceived();

        bt_unchoker_update_largest_value_first_sort( score, ratios, peer, ratioed_peers, 0 );  //higher value = worse score
      }
      //FIXME
      // double factor = 1F / ( 0.8 + 0.2 * Math.pow( random.nextFloat(), -1 ) );  //map to sorted list using a logistic curve 
      
      int pos = (int)(factor * g_list_length(ratioed_peers));
      result = g_list_nth(ratioed_peers, pos);
      return (bt_peer_t*) result->data;;
    }

    //FIXME
    int rand_pos = new Random().nextInt( g_list_length(optimistics) );
    result = g_list_nth(optimistics, rand_pos);
    return (bt_peer_t*) result->data;;
}
*/

/*
static gboolean bt_peer_is_unchokable( bt_peer_t*  peer, gboolean allow_snubbed ) {
    return peer->state == PEER_CONNECTED &&  peer.isInterested() && ( !peer.isSnubbed() || allow_snubbed );
}
  */


gboolean bt_peer_do_unchokes( bt_torrent_t* torrent )
{
    static int no_calling = 0 ;
    GList* dlist;
    GList* chokes = NULL;
    GList* unchokes = NULL;
    GList* optimistic_unchokes = NULL;
    GList* best_peers = NULL;
    int optimistic_unchokes_size = 0;
    int start_pos;
    int bests[MAX_TO_UNCHOKE]; // = // new long[ max_to_unchoke ];  //ensure we never pick more slots than allowed to unchoke
    

    gboolean optimistic = FALSE;

    int max_optimistic = ((MAX_TO_UNCHOKE - 1) / 10) + 1;  //one optimistic unchoke for every 10 upload slots

    if (PeerRequests == NULL)
    {
      return TRUE;
    }

    no_calling++;

    // Every 3 times (30 seconds) we add a peer to be unchoked as optimistic
    if ((no_calling - 3)  == 0)
    {
      no_calling = 0;
      optimistic = TRUE;
    }

    //get all the currently unchoked peers
    for (dlist = torrent->peers; dlist; dlist = dlist->next) {
     bt_peer_t* peer = (bt_peer_t* )dlist->data;
     bt_user_t* user = peer->user;

     if(user->seeder) continue; // never add a seeder to this treatment, as he will always be unchoked from the first time til disconnect
     
     if( !user->am_choking ) { 
       if( bt_peer_is_unchokable( peer, TRUE ) ) {
         unchokes = g_list_append(unchokes, peer );
         if( user->is_optimistic_unchoke &&  !user->am_choking ) {
           optimistic_unchokes = g_list_append(optimistic_unchokes, peer );
         }
       }
       else {  //should be immediately choked
         chokes = g_list_append(chokes, peer );
       }
     }
    }

   // optimistic_unchokes_size = g_list_length(optimistic_unchokes)
    if( !optimistic ) {  //ensure current optimistic unchokes remain unchoked
     int i = 0;
     for (dlist = optimistic_unchokes, i=0; dlist; dlist = dlist->next, i++) {
       bt_peer_t* peer = (bt_peer_t*)dlist->data;
       bt_user_t* user = peer->user;
       if( i < max_optimistic ) {
         best_peers = g_list_append(best_peers, peer );
       }
       else { //too many optimistics
         user->is_optimistic_unchoke = FALSE;
       }
     }
    }   
    
    //fill slots with peers who we are currently downloading the fastest from
    start_pos = g_list_length(best_peers);
    for (dlist = torrent->peers ;  dlist ; dlist = dlist->next )
    {
        bt_peer_t* peer = (bt_peer_t*)dlist->data;
        bt_user_t* user = peer->user;
        if(user->seeder) continue; // never add a seeder to this treatment, as he will always be unchoked from the first time til disconnect
    
        if( user->am_interested && bt_peer_is_unchokable ( peer , FALSE ) && (g_list_find(best_peers, peer) != NULL) ) {  //viable peer found
            int rate = rate_get_rate(user->rate);  
            if( rate > 256 ) {  //filter out really slow peers
                //FIXME : Sorting 
                best_peers = bt_unchoker_update_largest_value_first_sort( rate, bests, peer, best_peers, start_pos );
            }
        }
    }
    /*
    //if we havent yet picked enough slots
    if( g_list_length(best_peers) < MAX_TO_UNCHOKE ) {  
      start_pos = g_list_length(best_peers);
      
      //fill the remaining slots with peers that we have downloaded from in the past
      for( dlist = torrent->peers; dlist ; dlist = dlist->next  ) {
        bt_peer_t* peer = (bt_peer_t*)dlist->data;
        bt_user_t* user = peer->user;

        if( peer->am_interested && bt_peer_is_unchokable ( peer , FALSE ) && (g_list_find(best_peers, peer) == NULL) ) {  //viable peer found
          //FIXME
          //long uploaded_ratio = peer.getStats().getTotalDataBytesSent() / (peer.getStats().getTotalDataBytesReceived() + (DiskManager.BLOCK_SIZE-1));
          //make sure we haven't already uploaded several times as much data as they've sent us
          if( uploaded_ratio <3) {
            //FIXME : another rate for uplaoding ?
            int rate = rate_get_rate(user->rate);
            best_peers = bt_unchoker_update_largest_value_first_sort(rate, bests, peer, best_peers, start_pos );
          }  
        }
      }
    }
     */
    if( optimistic ) {
      //make space for new optimistic unchokes
      while( g_list_length(best_peers) > MAX_TO_UNCHOKE - max_optimistic ) {
        GList* list = g_list_last(best_peers);
        bt_peer_t* peer = (bt_peer_t*) list->data;
        g_list_remove( best_peers, peer );
      }
    }


    /*
    
    //if we still have remaining slots
    while( g_list_length(best_peers) < max_to_unchoke ) { 
      bt_peer_t* peer =bt_unchoker_get_next_optimistic_peer( torrent->peers, TRUE, TRUE );  //just pick one optimistically
      if( peer == NULL )  break;  //no more new unchokes avail
      
      if( g_list_find ( best_peers, peer) == NULL ) {
        g_list_append(best_peers, peer);
        peer->is_optimistic_unchoke = TRUE;
      }
      else {
        //we're here because the given optimistic peer is already "best", but is choked still,
        //which means it will continually get picked by the getNextOptimisticPeer() method,
        //and we'll loop forever if there are no other peers to choose from
        bt_peer_send(BSOCK(peer), BT_MESSAGE_UNCHOKE);
      }
    }
    */

    //update chokes
    for( dlist = unchokes; dlist ; dlist = dlist->next ) {
      bt_peer_t* peer = (bt_peer_t*)dlist->data;

      if( g_list_find(best_peers, peer) == NULL ) {  //should be choked
        if( g_list_length(best_peers) < MAX_TO_UNCHOKE ) {  //but there are still slots needed (no optimistics avail), so don't bother choking them
          g_list_append(best_peers, peer );
        }
        else {
           g_list_append(chokes,peer);
           g_list_remove(unchokes, peer);
        }
      }
    }
    
    //update unchokes
    for( dlist = best_peers; dlist ; dlist = dlist->next ) {
      bt_peer_t* peer = (bt_peer_t*)dlist->data;
      if( g_list_find(unchokes, peer) == NULL ) {
          g_list_append(unchokes, peer );
      }
    }

    // now everything is settled, we can send choke and unchoke (if we are in the timing)

    for(dlist = chokes; dlist; dlist=dlist->next)
    {
        bt_peer_t* peer = (bt_peer_t*) dlist->data;
        bt_peer_send(BSOCK(peer), BT_MESSAGE_CHOKE);
    }
    for(dlist = unchokes; dlist; dlist=dlist->next)
    {
        bt_peer_t* peer = (bt_peer_t*) dlist->data;
        bt_peer_send(BSOCK(peer), BT_MESSAGE_UNCHOKE);
    }

    //TODO : clean unchokes and chokes
    return TRUE;
}

//TODO : make a un/choker manager for seeding mode

#undef HANDLER
#undef BT_CALL
