/* Copyright (C) 2005-2006 (C) Lopster Team
   This is free software distributed under the terms of the
   GNU Public License.  See the file COPYING for details. */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <stdlib.h>

#include "bt_torrent.h"
#include "bt_users.h"
#include "bt_peer_message.h"

void bt_peer_message_free(bt_peer_message_t* peer_message)
{
    g_free(peer_message);
}

static void bt_peer_message_init( bt_peer_message_t* peer_message)
{
    peer_message->id = BT_MESSAGE_UNKNOWN;
    peer_message->message = NULL;
    peer_message->length = 0;
}


static	bt_peer_message_t* bt_peer_message_new( ) 
{
    bt_peer_message_t* peer_message = g_malloc0(sizeof(*peer_message));
    bt_peer_message_init(peer_message);
    return peer_message;
}

static	void bt_peer_message_setup_id(bt_peer_message_t* peer_message, int id) 
{
    g_assert(peer_message);
    peer_message->id = id;
}

static	void bt_peer_message_setup_length(bt_peer_message_t* peer_message, int length) 
{
    g_assert(peer_message);
    peer_message->length = length;
}

static	void bt_peer_message_setup_message(bt_peer_message_t* peer_message, char* message) 
{
    g_assert(peer_message);
    peer_message->message = message;
}

static bt_peer_message_t* bt_peer_message_parse_handshake( buffer_t* buffer ) {
    bt_peer_message_t* result;
    int length;
    char* data;
    char* pos = NULL;

    unsigned int pstrlen;

/*
    The handshake is a required message and must be the first message transmitted by the client.
    handshake: <pstrlen><pstr><reserved><info_hash><peer_id>
    o          19BitTorrent protocol00000000
    o   pstrlen: string length of <pstr>, as a single raw byte
    o   pstr: string identifier of the protocol
    o   reserved: eight (8) reserved bytes. zero!!
    o   info_hash: 20-byte SHA1 hash of the info key in the metainfo file. 
        This is the same info_hash that is transmitted in tracker requests.
    o   peer_id: 20-byte string used as a unique ID for the client. 
        This is the same peer_id that is transmitted in tracker requests.
    In version 1.0 of the BitTorrent protocol, pstrlen=19, and pstr="BitTorrent protocol".
*/      //  if (buffer_avail(buffer) < 1) return NULL;

    //data should be at least 68 bytes
    if (buffer_avail(buffer) < 68) return NULL;

	data = buffer_read_pos(buffer);

    // find "bittorrent protocol" string in buffer else rerturn NULL
    pos = strstr(data,"BitTorrent protocol");
    if ( pos == NULL) {
        return NULL;
    }
    pos = data;
    // ok parse pstrlen normally 19 
    pstrlen=*data;
    if (pstrlen != 19) {
        return NULL;
    }
    pos += 20; 

    // Skip 8 reserved bytes 
    // TODO : In the futur, check these bytes
    pos += 7; // positionned on the last byte of reserved bytes
    
    length = pos-data;
    buffer_consume_virt(buffer, length + 1); // + the last reserved byte
    *pos = 0;
    data = buffer_read_pos(buffer);
    result = bt_peer_message_new( );	
    bt_peer_message_setup_id(result,BT_MESSAGE_HANDSHAKE);
	bt_peer_message_setup_length(result,40); // peer_id + torrent_hash
	bt_peer_message_setup_message(result,data);
    return result;
}

bt_peer_message_t* bt_consume_message(buffer_t* buffer, int length_to_consume, int last_id)
{
    bt_peer_message_t* result;
    guint16 consume_len;
    char* data;
    char* pos;
    unsigned int raw_len;
    int length;
    int id;

    if (!buffer) return NULL;
    
    if (last_id == BT_MESSAGE_PIECE) 
    {
         return NULL;
    }

    if (length_to_consume != -1) 
    {
         buffer_consume_virt(buffer, length_to_consume);
    }

    while (TRUE)
    {
        if (buffer_avail(buffer) < 5) return NULL; //at least we need the length (4 bytes) and the id (1 byte)
        data = buffer_read_pos(buffer);
        // Check for handshake message (begin normally with signle byte 19
        if (*data == 0x13 && buffer_avail(buffer) >= 68 )
        {
            // call the parse hanshake message, if yes return it , else continue;
            result = bt_peer_message_parse_handshake(buffer);
            if (result) 
            {
                g_log("BT_PEERMSG", G_LOG_LEVEL_DEBUG, "HANDSHAKE Message detected" );
                return result;
            }
        }

        pos = data;
        memcpy(&raw_len,data,4);
        length = ntohl(raw_len);       
        
        // This is where PIECE message need to be checked before any other messages
        
        id = *(data+4);
        if (id == BT_MESSAGE_PIECE)
        {

            int raw_index;
            int raw_begin;
            if ( buffer_avail(buffer) < 13) return NULL;
            pos = data + 4;
                     
            consume_len = pos - data;
            buffer_consume_virt(buffer, consume_len + 1);
            *pos = 0;
            data = buffer_read_pos(buffer);

            result = bt_peer_message_new( );
            result->id = id;

            memcpy(&raw_index,data,4);
            result->index = ntohl(raw_index);
            memcpy(&raw_begin,data+4,4);
            result->begin = ntohl(raw_begin);
            result->length = length-9; // Skip the ID, INDEX and BEGIN information
            pos = data+7;
            buffer_consume_virt(buffer,8);
            *pos = 0;
            return result;
        }
        
        // Not a piece message also
        if (buffer_avail(buffer) < length+4)
        {
            // Can be a Keep Alive Message (that is 4 bytes length zeroed)
            if (length == 0)
            {
                // this is a keep alive message, let's send it
                result = bt_peer_message_new( );
                bt_peer_message_setup_id(result, BT_MESSAGE_KEEP_ALIVE);
                return result;
            }
            else
            {
                return NULL;
            }
        }
        pos = data + 3;

        // It's not a keep alive message, and len match the length of the rest of the buffer 
        // (there maybe more data in the buffer though)
        
        consume_len = pos - data;
        buffer_consume_virt(buffer, consume_len + 1);
        *pos = 0;

        // Needed ??
        data = buffer_read_pos(buffer);

        if (id < BT_MESSAGE_CHOKE || id > BT_MESSAGE_PORT )
        {
            return NULL;
        }        

        // Now we know we have a real peer message. 
        result = bt_peer_message_new( );
        bt_peer_message_setup_id(result, id);
        bt_peer_message_setup_length(result, length-1); // skip the 1 len for the id
        
        pos = data;
        buffer_consume_virt(buffer, 1);
        *pos = 0;
        bt_peer_message_setup_message(result, data);
        return result;
    }
}
