/* Copyright (C) 2005 The Lopster Team
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */


/* $Revision: 1.6 $
 * $Date: 2007/11/19 11:12:55 $
 * $Author: sid4u $
 */

#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <stdlib.h>   // for srand() and rand()
#include <glib.h>
#include <string.h>

#include "lcore/l_base.h"
#include "lcore/l_i18n.h"
#include "lcore/files.h"
#include "lcore/download.h"
#include "bt_torrent.h"
#include "bt_benc.h"
#include "bt_piece.h"
#include "bt_tracker.h"
#include "bt_network.h"
#include "bt_users.h"
#ifdef WIN32
#include "btalias.h"
#endif

#define MAX_TORRENT_SIZE  1024*1024

#define ANNOUNCE			"announce"
#define ANNOUNCE_LIST	"announce-list"
#define CREATED_BY		"created by"
#define CREATION_DATE	"creation date"
#define INFO					"info"

#define FILES					"files"
#define LENGTH				"length"
#define NAME					"name"
#define PIECE_LENGTH	"piece length"
#define PIECES				"pieces"

#define PATH					"path"

#define RAND(min,max) ((rand()/(float)0x7fffffff)*(max-min))+min

typedef struct {
    module_t _module;
    GList* torrents;
} bt_torrent_module_t;

static bt_torrent_module_t* BtTorrentModule = 0;

// Before adding a torrent, this should be called to check for the existence of a torrent in the list.
/* static gboolean bt_torrent_exists(bt_torrent_t* torrent) */
/* { */
/*     GList* dlist; */
/*     g_assert(BtTorrentModule); */

/*     for (dlist = BtTorrentModule->torrents; dlist; dlist = dlist->next) */
/*     { */
/*         bt_torrent_t* t = dlist->data; */
/*         if (strcmp(t->info_hash, torrent->info_hash) == 0) return TRUE; */
/*     } */
/* 	return FALSE; */
/* } */


static void bt_torrent_destroy(bt_torrent_t* torrent)
{
    BtTorrentModule->torrents = g_list_append(BtTorrentModule->torrents, torrent);
    g_list_foreach(torrent->trackers, (GFunc)object_kill, NULL);
    //FIXME: handle files,
    g_free(torrent->main_directory);
    g_free(torrent->created_by);
    object_destroy(OBJ(torrent));
}

static void bt_torrent_init(bt_torrent_t* torrent)
{
    object_init(OBJ(torrent), "Torrent", TRUE);
    object_add_type(OBJ(torrent), "Torrent");
    OBJ(torrent)->destroy = (ObjectFunc)bt_torrent_destroy;
    torrent->status = BT_TORRENT_STOPPED;
}

static bt_torrent_t* bt_torrent_new(void)
{
    bt_torrent_t* torrent = g_malloc0(sizeof(*torrent));
    bt_torrent_init(torrent);
    return torrent;
}


// this function creates a new file if there is not any yet.
static file_t* bt_torrent_get_unique_file(bt_torrent_t* torrent)
{
    if (!torrent->files) 
    {
        file_t* unique_file = file_new();
        torrent->files = g_list_append(torrent->files, unique_file);
    }
    if (torrent->files->next) return NULL;
    else return torrent->files->data;
}
 
static void bt_torrent_generate_hashinfo(bt_torrent_t* torrent, bt_value_t* value)
{
	// It's a dictionnary and we made a record of the offset where the result begins in the buffer
	SHA1(value->dic_begin_memory_offset,
		value->dic_end_memory_offset - value->dic_begin_memory_offset,
		torrent->info_hash);
}

static gboolean bt_torrent_parse_main_tracker(bt_torrent_t* torrent, bt_value_t* value)
{
    bt_tracker_t* tracker;

    if (value->type != BT_BENC_STR)	return FALSE;

    // create new tracker, but dont allow duplicates
    tracker = bt_tracker_new_from_url(value->str, FALSE);
    if (tracker) 
    {
        tracker->torrent = torrent;
        torrent->trackers = g_list_append(torrent->trackers,  tracker);
    }

    return TRUE;
}

static gboolean bt_torrent_parse_trackers_list(bt_torrent_t* torrent, bt_value_t* value)
{
    GList* dlist;

// Tracker list is a BT_BENC_LIST format
    if (value->type != BT_BENC_LIST) return FALSE;

    for (dlist = value->list; dlist; dlist = dlist->next)
    {
        bt_value_t* v = dlist->data;
        GList* dlist2;
        bt_value_t* tvalue;
        bt_tracker_t* tracker;

        // Each tracker is in a single list itself
        if (v->type != BT_BENC_LIST) return FALSE;

        dlist2 = v->list;
        tvalue = dlist2->data;
        if (tvalue->type != BT_BENC_STR) return FALSE;

				tracker = bt_tracker_new_from_url(tvalue->str, FALSE);
        if (tracker) 
        {
            tracker->torrent = torrent;
            torrent->trackers = g_list_append(torrent->trackers,  tracker);
        }
		}

    return TRUE;
}

static gboolean bt_torrent_parse_created_by(bt_torrent_t* torrent, bt_value_t* value)
{
    return bt_benc_value_get_string(value, &torrent->created_by);
}

static gboolean bt_torrent_parse_creation_date(bt_torrent_t* torrent, bt_value_t* value)
{
    gint64 val;
    if (bt_benc_value_get_int(value, &val))
    {
        torrent->creation_date = val;
        return TRUE;
    }
    return FALSE;
}

static gboolean bt_torrent_parse_unique_file_length(bt_torrent_t* torrent, bt_value_t* value)
{
    file_t* file = bt_torrent_get_unique_file(torrent);
    return bt_benc_value_get_int(value, &file->size);
}

// Method for multiple files , a big Bencoded List of dictionnaries to parse
static gboolean bt_torrent_parse_files(bt_torrent_t* torrent, bt_value_t* value)
{
    GList* dlist;

    if (value->type != BT_BENC_LIST) return FALSE;

    for (dlist = value->list; dlist; dlist = dlist->next)
    {
        bt_value_t* dic = dlist->data;
        file_t* new_file;
        GList* dlist2;
        
        if (dic->type != BT_BENC_DIC) return FALSE;

        new_file = file_new();
        torrent->files = g_list_append(torrent->files, new_file);	

        for (dlist2 = dic->pairs; dlist2; dlist2 = dlist2->next)
        {
            bt_value_t* file_info_key = dlist2->data;
            bt_value_t* file_info_value;
            GList* dlist3;
            
            g_assert(file_info_key->type == BT_BENC_STR);
            g_assert(dlist2->next);
            dlist2 = dlist2->next;
            file_info_value = dlist2->data;

            if (!strcmp(file_info_key->str,LENGTH))
            {
                if (!bt_benc_value_get_int(file_info_value, &new_file->size)) return FALSE;
            }
            else if (!strcmp(file_info_key->str,PATH))
            {
                char* filename = NULL;
                if (file_info_value->type != BT_BENC_LIST) return FALSE;
                for (dlist3 = file_info_value->list; dlist3; dlist3 = dlist3->next)
                {
                    bt_value_t* v1 = dlist3->data;
                    char* current_path;
                    
                    if (v1->type != BT_BENC_STR) return FALSE;
                    current_path = v1->str;
                    if (filename) filename = l_printf("%s/%s", filename, current_path);
                    else filename = g_strdup(current_path);
                }
                file_set_origname(new_file, g_strdup(filename));
                file_set_locname(new_file, g_strdup(filename));
            }
            else
            { 
                // TODO: handle other tags
                continue;
            }
        }
    }
    return TRUE;
}

static gboolean bt_torrent_parse_main_directory(bt_torrent_t* torrent, bt_value_t* value)
{
    return bt_benc_value_get_string(value, &torrent->main_directory);
}

static gboolean bt_torrent_parse_main_filename(bt_torrent_t* torrent, bt_value_t* value )
{
    char* filename = 0;
    file_t* file = bt_torrent_get_unique_file(torrent);
    if (bt_benc_value_get_string(value, &filename))
    {
        file_set_origname(file, g_strdup(filename));
        file_set_locname(file, g_strdup(filename));
        return TRUE;
    }
    return FALSE;
}

static gboolean
bt_torrent_parse_pieces(bt_torrent_t* torrent, bt_value_t* value, guint64 plen)
{
    unsigned len;
    unsigned num;
    GList* dlist;
    guint64 offset = 0;
    int size;

    if (value->type != BT_BENC_STR)	return FALSE;
    size = value->str_size;
    if ((len = size) % 20 != 0 || (num = len / 20) == 0)
        return FALSE;
    torrent->no_pieces = (int) num;
    // Fixme : SGop this seems not to be working , can you check that out ?
 /*   for (dlist = torrent->files; dlist; dlist = dlist->next)
    {
        file_t* file = dlist->data;
        int spiece = offset / plen;
        int epiece = (offset+file->size+plen-1) / plen;
        if (epiece > num) return FALSE;
        attr_store(ATTRC(file), "urn:sha1:pieces", value->str+spiece*20,
                   (epiece-spiece)*20, plen);
        offset += file->size;
    }
    if ((offset+plen-1) / plen != num) return FALSE;
    */
    return TRUE;
}


static gboolean bt_torrent_parse_info(bt_torrent_t* torrent, bt_value_t* value)
{
    // Now the main part of parsing, we won't go again in the upper loop..all should end here
    GList* dlist;
    guint64 plen = 0;
    gboolean is_single;

    // We expect a dictionnary here
    if (value->type != BT_BENC_DIC) return FALSE;


    // first check for single or multi file torrent
    if (bt_benc_value_find_dic(value, LENGTH))
    {
        if (bt_benc_value_find_dic(value, FILES)) return FALSE;  // inconsistent
        is_single = TRUE;
    }
    else if (bt_benc_value_find_dic(value, FILES))
    {
        is_single = FALSE;
    }
    else
    {
        return FALSE;  // incomplete torrent??
    }

    // We need to also make a hash of the info structure (now we know it's a real torrent file)
    /* Generate the hash_info for this torrent */
    bt_torrent_generate_hashinfo(torrent, value);

    for (dlist = value->pairs; dlist; dlist = dlist->next->next)
    {
        bt_value_t* v1 = dlist->data;
        bt_value_t* v2 = dlist->next->data;

        if (!strcmp(v1->str,LENGTH))
        {
            if (!bt_torrent_parse_unique_file_length(torrent, v2)) return FALSE;
        }
        else if (!strcmp(v1->str,FILES))
        {
            // We got FILES , that means, it's a multiple files torrent
            // So we'll have a big ?  list of dictionnaries (as much files the torrent has)
            if (!bt_torrent_parse_files(torrent, v2)) return FALSE;
        }
        else if (!strcmp(v1->str,NAME))
        {
            if (is_single)
            {
                if(!bt_torrent_parse_main_filename(torrent, v2)) return FALSE;
            }
            else
            {
                if(!bt_torrent_parse_main_directory(torrent, v2)) return FALSE;
            }
        }
        else if (!strcmp(v1->str,PIECE_LENGTH))
        {
            if (!bt_benc_value_get_int(v2, &plen)) return FALSE;
            torrent->piece_length =  plen;
        }
        else if (!strcmp(v1->str,PIECES))
        {
            // Finally the different pieces 20 bytes SHA1, divide by 20 gives also 
            // the same results as divide the whole (torrent size / PIECE_LENGTH) + 1 (if has rest)
            if(!bt_torrent_parse_pieces(torrent, v2, plen)) return FALSE;
        }
        else
        {
            // TODO: we've got other keys to check like *-utf
            continue;
        }
    } //endfor
    return TRUE;
}
            
static void bt_torrent_update_bitfield( bt_torrent_t* torrent )
{
    char* str_bitfield;
    int no_pieces = torrent->no_pieces; 
    int no_bits;
    no_bits = no_pieces / 8 ;
    if (no_pieces % 8 != 0) 
        no_bits++;

    str_bitfield = g_strnfill(no_bits*8,'0');
    // Update bitfield record.
    // Fixme :  Bitfield here is created from 0, but if we load the torrent from the saved xml state. 
    //          Get the ascii bitfield from the BT xml setting for this torrent and pass it to the parameter
    torrent->bitfield =  bt_bitfield_setup_from_client(str_bitfield);
}

static bt_torrent_t* bt_torrent_new_from_data(const char* buffer, unsigned size)
{
    bt_value_t* value = bt_benc_parse_from_data(buffer, size);
    bt_torrent_t* torrent;
    GList* dlist;
    
    if (!value) return NULL;
    
    // Value is normally positionned on Main Dictionnary,
    if (value->type != BT_BENC_DIC)
    {
        g_log("BT_TORRENT", G_LOG_LEVEL_WARNING,
              "Dictionary expected at toplevel");
        bt_benc_destroy(value);
        return NULL;
    }

    torrent = bt_torrent_new();
    
    // NOTE: we are sure that there are really an even number of items in value->pairs
    // and the odd items are always strings.
    for (dlist = value->pairs; dlist; dlist = dlist->next)
    {
        bt_value_t* v1 = dlist->data;
        bt_value_t* v2;

        g_assert(dlist->next);
        g_assert(v1->type == BT_BENC_STR);

        dlist = dlist->next;
        v2 = dlist->data;

        if (!strcmp(v1->str, ANNOUNCE))
        {
            if (!bt_torrent_parse_main_tracker(torrent, v2)) break;
        }
        else if (!strcmp(v1->str, ANNOUNCE_LIST))
        {
            if (!bt_torrent_parse_trackers_list(torrent, v2)) break;
        }
        else if (!strcmp(v1->str,CREATED_BY))
        {
            if (!bt_torrent_parse_created_by(torrent, v2)) break;
        }
        else if (!strcmp(v1->str,CREATION_DATE))
        {
            if (!bt_torrent_parse_creation_date(torrent, v2)) break;
        }		
        else if (!strcmp(v1->str,INFO))
        {
            if (!bt_torrent_parse_info(torrent, v2)) break;
        }		
        else
        {
            // TODO: we've got other keys to check like *-utf
            continue;
        }		//endif
    } //endfor

    //  We should have now a torrent structure filled with the needed data (any error upper should return to parent)
    bt_torrent_update_bitfield(torrent);
    // Let's now send back the structure to the caller (display )
    bt_benc_destroy(value);

    // on error: delete torrent
    if (dlist)
    {
        object_kill(OBJ(torrent));
        torrent = NULL;
    }
    return torrent;
}


bt_torrent_t* bt_torrent_new_from_file(const char* filename)
{
    char* buffer;
    gsize len;
    bt_torrent_t* torrent;
    GError* error = NULL;

    g_log("BT_TORRENT", G_LOG_LEVEL_DEBUG, "parsing file '%s'", filename);

    // TODO : check if the torrent already exist, and warn user before re-download

    if (!g_file_get_contents(filename, &buffer, &len, &error))
    {
        lopster_notify(LOPSTER_MESSAGE_ERROR, _("Failed to load .torrent"),
                       error?error->message:_("The file could not be loaded for unknown reason."));
        if (error) g_error_free(error);
        return NULL;
    }
    
    torrent = bt_torrent_new_from_data(buffer, len);
    g_free(buffer);

    if (torrent)
        BtTorrentModule->torrents = g_list_append(BtTorrentModule->torrents, torrent);
    else
        lopster_notify
            (LOPSTER_MESSAGE_ERROR, _("Failed to parse .torrent"),
             l_printf(_("File %s could not be parsed. It seems not to be valid torrent file."), filename));

    return torrent;
}


// This is called if bt_torrent_search_for_torrent return TRUE, and user still want to download
gboolean bt_torrent_check_elligibility(bt_torrent_t* torrent)
{
    g_assert(torrent);
    return bt_network_torrent_can_redownload(torrent);
}

/*
GList* bt_torrent_file_to_pieces( bt_torrent_t* torrent, file_t* file)
{
    GList* pieces_from_file;
    GList* dlist;
    int piece_index_begin;
    int piece_index_end;

    file_size_t begin = 0;
    file_sze_t end = 0;
    g_assert(torrent);
    g_assert(file);

    // Would create a list of all pieces related to a file
    for (dlist = torrent->files; dlist ; dlist = dlist->next)
    {
        file_t* current = (file_t*) dlist->data;
        end += current->size - ((file_size_t)1);
        if (current == file)
        {
            break;
        }
        begin+= end + ((file_size_t)1);
    }
    if (end == 0) {
        // error !!! log data here
        return;
    }
    // We have the begin and end ...
    // all pieces related to that range are to go to the list
    
    piece_index_begin = bt_utils_file_size_to_piece_index(torrent, begin);
    piece_index_end = bt_utils_file_size_to_piece_index(torrent, end);
    
    if ( (piece_index_end - piece_index_begin) > 0) 
    {
        // at least, more than one piece found
        
    }
    
    
}

gboolean bt_torrent_check_piece_hash(bt_torrent_t* torrent, int piece)
{
    // TODO : implement the hashing of a piece and check for it's validity
    // Get the piece hash from the list of pieces.
    // make the hash of the current piece from data collected
    // compare both and return valid of not
    // this function would update bitfield ?? or let another module do it
    return FALSE;
}
*/
    //FIXME : this function would check for a HIGH priority download
    // if exist, return it, else choose an unfinished one randomly (maybe check for nearly finished one ?)
    // or avilable segments one so no round up 
download_t* bt_torrent_get_download_to_process(bt_torrent_t* torrent)
{

    GList* dlist;
    GList* dlist2 = NULL;
    GList* dlist3;
    int size = 0 ;
    int choosed = 0 ;

    //TODO : check for HIGH Priority download
    for (dlist = torrent->downloads; dlist; dlist = dlist->next)
    {
        download_t* download = (download_t*) dlist->data;
        if (download->flags & DOWNLOAD_FINISHED) continue;
        dlist2 = g_list_append(dlist2,download);
    }
    if (dlist2 == NULL) return NULL;
    size = g_list_length(dlist2);

    if (size == 0) return NULL;
    srand( time( 0 ) );
    choosed = RAND(0,size-1);
    dlist3 =  g_list_nth(dlist2,choosed);
    return ((download_t*) dlist3->data);
}

    // FIXME : This function should test the popularity of a peer 
    // For the moment it will return the first non connected user
bt_user_t* bt_torrent_get_next_optimistic_user(bt_torrent_t* torrent)
{
    GList* dlist;
    GList* dlist2;

    for (dlist = torrent->trackers; dlist; dlist = dlist->next)
    {
        bt_tracker_t* tracker = (bt_tracker_t*) dlist->data;

        for (dlist2 = tracker->peers; dlist2; dlist2=dlist2->next)
        {
            bt_user_t* user = (bt_user_t*) dlist2->data;
            if (user->state == PEER_INCOMING) return user;
        }
    }
    
    return NULL;
}

void bt_torrent_accept_file(bt_torrent_t* torrent, file_t* file)
{
    if (g_list_find(torrent->files, file))
    {
        //FIXME: check for existing download
        download_t* download = download_from_file(file);
        torrent->files = g_list_remove(torrent->files, file);
        torrent->downloads = g_list_append(torrent->downloads, download);
        object_kill(OBJ(file));
    }
}


static void bt_torrents_init()
{
}

static void bt_torrents_run()
{
}

static void bt_torrents_exit()
{
    //FIXME: destroy all torrents.
}

static void bt_torrent_module_init(bt_torrent_module_t* module)
{
    module_init(MODULE(module), "Bt Torrent Module",
                (ModuleFunc)bt_torrents_init,
                (ModuleFunc)bt_torrents_run,
                (ModuleFunc)bt_torrents_exit);

    module->torrents = NULL;
    
    module_require(bt_tracker_module_instance);
    module_require(bt_pieces_module_instance);
/*     module_require(xml_module_instance); */
/*     module_require(tsocket_module_instance); */
/*     module_require(resolve_module_instance); */
/*     module_require(listen_module_instance); */
}


module_t* bt_torrent_module_instance()
{
    if (!BtTorrentModule)
    {
        BtTorrentModule = g_malloc0(sizeof(*BtTorrentModule));
        bt_torrent_module_init(BtTorrentModule);
    }
    return MODULE(BtTorrentModule);
}

#ifdef WIN32
#define _BT_C_
#include "btaliasdef.c"
#endif

