/* Copyright (C) 2005 The Lopster Team
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */


/* $Revision: 1.8 $
 * $Date: 2007/11/19 11:12:55 $
 * $Author: sid4u $
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <openssl/sha.h>
#include <glib.h>
#include "lcore/l_base.h"
#include "lcore/l_http.h"
#include "lcore/download.h"
#include "bt.h"
#include "bt_download.h"
#include "bt_benc.h"
#include "bt_users.h"
#include "bt_utils.h"
#include "bt_tracker.h"

typedef struct {
    module_t _module;
    GList* trackers;
} bt_tracker_module_t;

static bt_tracker_module_t* BtTrackerModule = 0;


static gboolean bt_tracker_find(bt_tracker_t* tracker)
{
    GList* dlist;

    g_assert(BtTrackerModule);

    for (dlist = BtTrackerModule->trackers; dlist; dlist = dlist->next)
    {
        bt_tracker_t* t = dlist->data;
        if (url_compare(tracker->url, t->url)) return TRUE;
    }
    return FALSE;
} 

//TODO : transform into a module for having a list of already created tracker and track;) them
static url_t* bt_parse_url_from_long_host(const char* long_host)
{
    char* host;
    char* doc;
    char* temp = l_strdup(long_host);

    // TODO: Manage udp:// type of tracker
    if (!strncmp(temp, "http://", 7))
        host = temp+7;
    else if (!strncmp(temp, "udp://", 6))
        host = temp+6;
    else
        return NULL;

    doc = strchr(host, '/');
    if (!doc) return NULL;
    *doc = 0;
    host = l_strdup(host);
    *doc = '/';
    
    return url_new(host, doc);
}


static void bt_tracker_destroy(bt_tracker_t* tracker)
{
    object_kill(OBJ(tracker->url));
    tracker->torrent = NULL;
    g_free(tracker->error_message);
    g_free(tracker->warning_message);
    g_free(tracker->id_old);
    g_free(tracker->id_new);
    // give freedom to the peers again :)
    g_list_foreach(tracker->peers, (GFunc)user_unref, NULL);
    g_list_free(tracker->peers);
    tracker->peers = NULL;
    object_destroy(OBJ(tracker));
}

static void bt_tracker_init(bt_tracker_t* tracker, const char* url)
{
    object_init(OBJ(tracker), url, TRUE);
    object_add_type(OBJ(tracker), "Tracker");
    OBJ(tracker)->destroy = (ObjectFunc)bt_tracker_destroy;
    tracker->url = bt_parse_url_from_long_host(url);
    tracker->pending_callback = -1;
}

static bt_tracker_t* bt_tracker_new(const char* url)
{
    bt_tracker_t* tracker = g_malloc0(sizeof(*tracker));
    bt_tracker_init(tracker , url);
    return tracker;
}

static void bt_tracker_add_user(bt_tracker_t* tracker, bt_user_t* user)
{
    if (!g_list_find(tracker->peers, user))
    {
        tracker->peers = g_list_append(tracker->peers, user_ref(USER(user)));
    }
}

bt_tracker_t* bt_tracker_new_from_url(const char* url, gboolean allow_duplicates)
{
    bt_tracker_t* tracker = bt_tracker_new(url);
    if (!allow_duplicates && bt_tracker_find(tracker))
    {
        object_kill(OBJ(tracker));
        tracker = 0;
    }
    return tracker;
}

#define FAILURE_REASON		"failure reason"
#define WARNING_MESSAGE		"warning message"
#define INTERVAL					"interval"
#define MIN_INTERVAL			"min interval"
#define TRACKER_ID				"tracker id"
#define COMPLETE					"complete"
#define INCOMPLETE				"incomplete"
#define PEERS							"peers"
#define	PEER_ID						"peer id"
#define	PEER_IP						"ip"
#define	PEER_PORT					"port"

static gboolean bt_tracker_update_request(bt_tracker_t* tracker);

static gboolean
bt_tracker_update_from_data(bt_tracker_t* tracker, const char* buffer, unsigned size)
{
    GList* dlist;
    GList* dlist2;
    GList* dlist3;
    gboolean fresh_tracker = FALSE; 
    gboolean success = TRUE; //SET to FALSE (TRUE HERE FOR A TEST)
    bt_value_t* value = bt_benc_parse_from_data(buffer, size);
  
    if (!value) return FALSE;

    if (tracker->state != BT_TRACKER_STARTED)
    {
        fresh_tracker = TRUE;
        tracker->state = BT_TRACKER_STARTED;
    }
    // TODO: Check for any error message from tracker (first value in list would be in string format
    if (value->type != BT_BENC_DIC)
    {
        success = FALSE;
        // TODO: Very important here to check what kind of error we got here
    }
	
    // We got a dictionnary as a result, that's what we were waiting for, let's go inside and check for the keys
    //TODO: Parse data from tracker
    for (dlist = value->pairs; success && dlist; dlist = dlist->next)
    {
        bt_value_t* v;
        bt_value_t* to_check = dlist->data;
        dlist = dlist->next;
        v = dlist->data;

        if (to_check->type == BT_BENC_STR && strcmp(to_check->str,FAILURE_REASON)==0)
        {
            // Error message, parse it and leave the for
            tracker->error_message = g_strdup(v->str);
            success = FALSE; // ? or TRUE, we got result , what should we get here ?
            break;
        }
        else if (to_check->type == BT_BENC_STR && strcmp(to_check->str,WARNING_MESSAGE)==0)
        {
            // Error message, parse it and leave the for
            tracker->warning_message = g_strdup(v->str);
        }
        else if (to_check->type == BT_BENC_STR && strcmp(to_check->str,INTERVAL)==0)
        {
            tracker->interval = v->number;
        }
        else if (to_check->type == BT_BENC_STR && strcmp(to_check->str,MIN_INTERVAL)==0)
        {
            tracker->min_interval = v->number;
        }
        else if (to_check->type == BT_BENC_STR && strcmp(to_check->str,TRACKER_ID)==0)
        {
            char* tracker_id = v->str;
            if (tracker->id_old == NULL)
            {
                tracker->id_old = g_strdup(tracker_id);
                tracker->id_new = g_strdup(tracker_id);
            }
            else
            {
                g_free(tracker->id_old);
                tracker->id_old = tracker->id_new;
                tracker->id_new = g_strdup(tracker_id);
            }
        }
        else if (to_check->type == BT_BENC_STR && strcmp(to_check->str,COMPLETE)==0)
        {
            tracker->complete = v->number;
        }
        else if (to_check->type == BT_BENC_STR && strcmp(to_check->str,INCOMPLETE)==0)
        {
            tracker->incomplete = v->number;
        }
        else if (to_check->type == BT_BENC_STR && strcmp(to_check->str,PEERS)==0)
        {
            //Parse the list of peers (list of dictionnaries)
            if (v->type != BT_BENC_LIST)
            {
                break;
            }
            for (dlist2 = v->list; dlist2; dlist2 = dlist2->next)
            {
                bt_value_t* dic_peer;
                
                char* peer_id = NULL;
                char* ip_address = NULL;
                int port = -1;				
                dic_peer = dlist2->data;
                if (dic_peer->type != BT_BENC_DIC)
                {
                    success = FALSE;
                    break;
                }
                for (dlist3 = dic_peer->pairs; dlist3 ; dlist3 = dlist3->next)
                {
                    bt_value_t* key = dlist3->data;
                    bt_value_t* result;
                    dlist3 = dlist3->next;
                    result = dlist3->data;

                    if (key->type == BT_BENC_STR && strcmp(key->str,PEER_ID)==0)
                    {
                        // Try to find that ID from the bt_network list of peers.
                        // If true then we get the pointer (and update ip and port) else create (and update ip and port)
                        //  user_peer = bt_user_find(result->str);

                        //FIXME: this is binary string, handle correctly.
                        peer_id = result->str;
                    }
                    else if (key->type == BT_BENC_STR && strcmp(key->str,PEER_IP)==0)
                    {
                        ip_address  = result->str;
                    }
                    else if (key->type == BT_BENC_STR && strcmp(key->str,PEER_PORT)==0)
                    {
                        port  = result->number;
                    }
                    else
                    {
                        //TODO: What else should it send (still those exotic trackers ?)
                        g_log("BT_TRACKER", G_LOG_LEVEL_DEBUG,
                              "TRACKER SENT A NON KNOWN DATA (TO BE ADDED ?) %s", key->str );
                    }
                    
                    // if (user_peer && port != -1 && ip_address != NULL)
                    if (peer_id != NULL && port != -1 && ip_address != NULL)
                    {
                        bt_user_t* user_peer;
                       // Let's get it and wait until we connect, if it matches our client just drop the connection
                       /* if (strcmp(BT_PEER_ID, peer_id) == 0 ) { */
                       /*     // Matching my peer_id, skip it */
                       /*     continue; */
                       /* }*/
                        
                        user_peer = bt_user_find(peer_id);
                        bt_tracker_add_user(tracker, user_peer);
                        bt_user_setup(user_peer, ip_address, port);
                        /* peer_id = NULL;*/
                        /* port = -1;*/
                        /* ip_address = NULL;*/
                    }
                } // FOR 
            } // FOR
        } // Else if
        else
        {
            // TODO: there shouldn't be anymore keys as BT protocol V1, but we need to log in case of exotic trackers
            g_log("BT_TRACKER", G_LOG_LEVEL_DEBUG,
                  "TRACKER SENT A NON KNOWN DATA (TO BE ADDED ?) %s", to_check->str );
        }
    }
	
    if (success)
    {
        // Start a timing for a periodic call of updating that tracker
        if (tracker->min_interval > 0)
        {
            tracker->pending_callback = g_timeout_add(tracker->min_interval* 1000,
                                  ( GSourceFunc)bt_tracker_update_request, tracker);
        }
        else if (tracker->interval > 0)
        {
            tracker->pending_callback = g_timeout_add(tracker->interval* 1000,
                                  ( GSourceFunc)bt_tracker_update_request ,tracker);
        }
        object_set_dirty(OBJ(tracker));
    }

    // Finally destroy that result
    //TODO : Free some memory here (damn i always forget that !!)
    bt_benc_destroy(value); 

    return success;
}

// Tracker will return a bencoded body or if error happened a message
static void bt_tracker_result_process_body(http_t* http)
{
    bt_tracker_t* tracker = socket_get_data(SOCK(http));
    buffer_t* buffer;
  
    if (!tracker) return;

    buffer = b_socket_get_in_buffer(BSOCK(http));
    if (!buffer)
    {
        g_log("BT_TRACKER", G_LOG_LEVEL_DEBUG, "http return null buffer" );
        return;
    }
    //  print_raw("peer list", buffer->data, buffer->size);

    // Check if it's a request to stop , we normally don't have to parse the data !! (NEED TO DOUBLE CHECK THAT)
    if (tracker->state == BT_TRACKER_STOPPED) return;

    // TODO: Header should already be consumed here, and so valid html data should come next, need to check that with sgop
    // 				Also check if gziped data un gunzipped before coming here
    if (!bt_tracker_update_from_data(tracker, buffer_read_pos(buffer), buffer_avail(buffer)))
    {
        //TODO: Handle error from tracker here 
        g_log("BT_TRACKER", G_LOG_LEVEL_WARNING, 
              "something went wrong in parsing tracker data !! to implement !! :)" );
    }
    
    // This used to be the place where users where started automatically..
    // But now the scheduler will connect to some peers automatically
/*
    else 
    {
        // Fixme : this only works for unique file torrent. Need to add a flag for the user attached as a source
        GList* dlist;
        GList* dlist2;
        bt_torrent_t* torrent = tracker->torrent;

        // Let's add the new peer we got from tracker to the list of sources 
        // for each of the different download_t
        for (dlist = torrent->downloads; dlist ; dlist = dlist->next)
        {
            download_t* download = dlist->data;
            for (dlist2 = tracker->peers; dlist2 ; dlist2 = dlist2->next)
            {
                source_t* source;
                bt_user_t* user = dlist2->data;
                if (user->connected == TRUE) continue;
                // TODO : check if user is not already source for another torrent else return;
                bt_user_set_am_interested_torrent(user, torrent);
                source = download_add_source(download, FILE(download), USER(user));
            }
        }
        
    } 
*/
}


static gboolean bt_tracker_request_result(http_t* http, const char* message)
{
    if (socket_get_status(SOCK(http)) == ST_INPUT_COMPLETE)
    {
        bt_tracker_result_process_body(http);
    }
    else
    {
        //TODO: Notify the torrent, that one of it's tracker return an error (timeout ..)
        g_log("BT_TRACKER", G_LOG_LEVEL_WARNING,
              "something went wrong in http_get tracker !!! http://%s:%d/%s", http->host, http->port, http->doc?http->doc:"NO DOC" );
    }
/*   socket_set_status(SOCK(http), ST_UNKNOWN); */
    return http_end(http, message);
}

//TODO: Add an error parameter message, in case tracker send something terrible !! :)
//TODO: Manage protocol , should call automatically http, udp or https (bctp ?)
static void bt_tracker_make_http_request(bt_tracker_t* tracker, const char* doc)
{
    url_t* url = tracker->url;
    http_t* http = http_new(url->host, url->port, doc);

    socket_set_data(SOCK(http), tracker);

    if (!http) return;

    socket_set_end(SOCK(http), (SockEndFunc)bt_tracker_request_result);

    g_log("BT_TRACKER", G_LOG_LEVEL_DEBUG, "getting %s", doc);
    http_get(http);
}


#define TRACKER_DOC_START_REQUEST  "%s?info_hash=%s&peer_id=%s&port=%d&uploaded=%d&downloaded=%d&left=%d&event=started&numwant=50"
#define TRACKER_DOC_UPDATE_REQUEST "%s?info_hash=%s&peer_id=%s&port=%d&uploaded=%d&downloaded=%d&left=%d&numwant=50"
#define TRACKER_DOC_STOP_REQUEST   "%s?info_hash=%s&peer_id=%s&port=%d&uploaded=%d&downloaded=%d&left=%d&event=stopped"


//TODO: add some otehr parameters to the tracker request
// Called only once for each trackers (just to start the engine for each tracker)
// Then each tracker , if no error happens, will set bt_tracker_update_request(itself) in a callback 
// that will be called within tracker->interval  
void bt_tracker_start_request(bt_tracker_t* tracker)
{
    char* buffer;
    char* escaped_info_hash =
        bt_utils_hexdigest(tracker->torrent->info_hash, SHA_DIGEST_LENGTH);

    tracker->state = BT_TRACKER_STARTED;

    buffer = g_strdup_printf
        (TRACKER_DOC_START_REQUEST,	tracker->url->doc?tracker->url->doc:"/",
         escaped_info_hash, BT_PEER_ID, 60098, tracker->torrent->uploaded,
         tracker->torrent->downloaded, tracker->torrent->full_size);
    bt_tracker_make_http_request(tracker, buffer);

    g_free(escaped_info_hash);
    g_free(buffer);
}

static gboolean bt_tracker_update_request(bt_tracker_t* tracker)
{
    char* buffer;
    char* escaped_info_hash =
        bt_utils_hexdigest(tracker->torrent->info_hash, SHA_DIGEST_LENGTH);
	
    buffer = g_strdup_printf
        (TRACKER_DOC_UPDATE_REQUEST, tracker->url->doc?tracker->url->doc:"/", 
         escaped_info_hash, BT_PEER_ID, 60098, tracker->torrent->uploaded,
         tracker->torrent->downloaded, tracker->torrent->left);
    bt_tracker_make_http_request(tracker, buffer);

    g_free(escaped_info_hash);
    g_free(buffer);
    // Always return false, as we want this function not to be called again in the same timing..
    // The function upper will create a new call in timing (new timing sent from tracker)
    return FALSE;
}

void bt_tracker_stop_request(bt_tracker_t* tracker)
{
    char* buffer;
    char* escaped_info_hash =
        bt_utils_hexdigest(tracker->torrent->info_hash,SHA_DIGEST_LENGTH);
	
    buffer = g_strdup_printf
        (TRACKER_DOC_STOP_REQUEST, tracker->url->doc?tracker->url->doc:"/", 
         escaped_info_hash, BT_PEER_ID, 60098, tracker->torrent->uploaded,
         tracker->torrent->downloaded, tracker->torrent->left);
    bt_tracker_make_http_request(tracker, buffer);

    g_free(escaped_info_hash);
    g_free(buffer);
}

void bt_tracker_stop_tracker_request(bt_tracker_t* tracker)
{
    // stopping shoukd work fine now
    if (tracker->pending_callback > 0) {
        g_source_remove (tracker->pending_callback);
    }
    tracker->state = BT_TRACKER_STOPPED;
    bt_tracker_stop_request(tracker);
}

//TODO: this is called from the ui interface, and we should find a way not to let user for the update manually 
//			if the min_interval we got from tracker is not reached yet. 
//SOLUTION: Disable the manual update until min interval is reached ?
void bt_tracker_manual_request(bt_tracker_t* tracker )
{
    //TODO: Check if user can make a manual request (check min_interval) and last update time
    //TODO: then cancel a pending call in lopster_periodic (there should only be one)
    //TODO: Then call bt_tracker_update_request
    (void)tracker;
}

void bt_tracker_do_trackers_checks(bt_torrent_t* torrent)
{
    /*TODO : check if client is firewalled
        if not then make a number of wanted clients to get from tracker
        else use a strict limit (<= maxNumberOfAllowedConnection from client)

        At the end anyway, we should at least have a timing for the tracker to update themselves
    */

    // FIXME : maybe not really needed at first here, as the tracker has a timeout to update itself automatically
    // FIXME : tracker maybe shouldn't directly start connection when receiving users.
    //          instead leave the bt_network to start on scheduler for each torrent
}

void bt_tracker_display_tracker(bt_tracker_t* tracker)
{
    if (!tracker) return;
}



static void bt_trackers_init()
{
}

static void bt_trackers_run()
{
}

static void bt_trackers_exit()
{
    //FIXME: destroy all trackers.
}

static void bt_tracker_module_init(bt_tracker_module_t* module)
{
    module_init(MODULE(module), "Bt Tracker Module",
                (ModuleFunc)bt_trackers_init,
                (ModuleFunc)bt_trackers_run,
                (ModuleFunc)bt_trackers_exit);

    module->trackers = NULL;

    module_require(http_module_instance);
    module_require(bt_download_module_instance);
/*     module_require(xml_module_instance); */
/*     module_require(tsocket_module_instance); */
/*     module_require(resolve_module_instance); */
/*     module_require(listen_module_instance); */
}


module_t* bt_tracker_module_instance()
{
    if (!BtTrackerModule)
    {
        BtTrackerModule = g_malloc0(sizeof(*BtTrackerModule));
        bt_tracker_module_init(BtTrackerModule);
    }
    return MODULE(BtTrackerModule);
}
