/* Copyright (C) 2005-2006 sgop@users.sourceforge.net.
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.5 $
 * $Date: 2007/09/05 14:34:07 $
 * $Author: sid4u $
 */

#include <stdlib.h>

#include <libxml/parser.h>

#include "lcore/browse.h"
#include "lcore/tcp_socket.h"
#include "lcore/zbuffer.h"
#include "lcore/l_i18n.h"
#ifdef WIN32
#include "lcore/win32_defs.h"
#endif

#include "dc_proto.h"
#include "dc_peer.h"
#include "dc_browse.h"

#ifdef WIN32
#include "dcalias.h"
#endif

#define DC_BROWSE_INACTIVE    0
#define DC_BROWSE_WAITING     1
#define DC_BROWSE_RUNNING     2
#define DC_BROWSE_DONE        3

#define DC_TYPE_DCLST         0
#define DC_TYPE_XMLBZ         1
#define DC_TYPE_BZ            2

#define DC_BROWSE(arg) ((dc_browse_t*)arg)
typedef struct {
    browse_t _browse;
    tcp_socket_t* socket;
    unsigned todo;
    int type;
} dc_browse_t;

typedef struct {
    module_t _module;
} dc_browse_module_t;

static dc_browse_module_t* DcBrowseModule = NULL;


#define HANDLER(f) static const char* f (tcp_socket_t* s, char* data)
#define DC_CALL(f) f(s, command->message)
#define REQUIRE(expr, proto, message)                                   \
    if (!(expr)) {g_log(proto, G_LOG_LEVEL_WARNING, message); return message;}



static void dc_bsock_destroy(tcp_socket_t* s) {
    dc_browse_t* browse = socket_get_data(SOCK(s));
    browse->socket = NULL;
    tcp_socket_destroy(TSOCK(s));
}

static gboolean dc_bsock_end(tcp_socket_t* s, const char* message) {
    browse_t* browse = socket_get_data(SOCK(s));
    //  dc_browse_t* dbrowse = socket_get_data(SOCK(s));

    switch (socket_get_status(SOCK(s))) {
    case S_BROWSE_COMPLETE:
        browse_set_status(BROWSE(browse), NULL);
        browse_set_progress(BROWSE(browse), 0, "Complete");
        break;
    default:
        browse_set_status(browse, message);
        browse_set_progress(browse, 0., NULL);
    }
    tcp_socket_end(s, message);
    return TRUE;
}

static tcp_socket_t* dc_bsock_new(dc_browse_t* browse) {
    tcp_socket_t* s = tcp_socket_new(OBJ(BROWSE(browse)->user)->name, 30);
    object_add_type(OBJ(s), "DcBSock");
    OBJ(s)->destroy = (ObjectFunc)dc_bsock_destroy;
    socket_set_end(SOCK(s), (SockEndFunc)dc_bsock_end);
    socket_set_data(SOCK(s), browse);

    socket_add_load(SOCK(s), "Peer");

    return s;
}

static void dc_browse_destroy(dc_browse_t* browse) {
    if (browse->socket) {
        socket_done(SOCK(browse->socket), ST_DELETE, "closed");
    }
    browse_destroy(BROWSE(browse));
}

static void dc_browse_init(dc_browse_t* browse, dc_user_t* user)
{
    browse_init(BROWSE(browse), USER(user));
    object_add_type(OBJ(browse), "DcBrowse");
    OBJ(browse)->destroy = (ObjectFunc)dc_browse_destroy;
    browse->socket = dc_bsock_new(browse);
}

static dc_browse_t* dc_browse_new(dc_user_t* user)
{
    dc_browse_t* browse = g_malloc0(sizeof(*browse));
    dc_browse_init(browse, user);
    return browse;
}

static void xml_get_node(browse_t* browse, xmlDocPtr xdoc,
                         xmlNodePtr xnode, const char* base) {
    xmlNodePtr xsub = xnode->xmlChildrenNode;
  
    while (xsub != NULL) {
        if (!strcmp(xsub->name, "Directory")) {
            xmlChar* name = xmlGetProp(xsub, "Name");
            if (name && *name) {
                char* temp = g_strdup_printf("%s%s\\", base, name);
                xml_get_node(browse, xdoc, xsub, temp);
                g_free(temp);
            }
            if (name) xmlFree(name);
        } else if (!strcmp(xsub->name, "File")) {
            xmlChar* name = xmlGetProp(xsub, "Name");
            xmlChar* size = xmlGetProp(xsub, "Size");
            xmlChar* tth = xmlGetProp(xsub, "TTH");
            if (name && *name && size && *size) {
                file_t* file = file_new();
                // FIXME: utf8?
                file_set_origname(file, g_strdup_printf("%s%s", base, name));
                file_set_locname(file, g_strdup(file->netname));
                file->size = strtoll(size, NULL, 10);
                if (tth && *tth) {
                    attr_store_str(ATTRC(file), "file", "urn:tree:tiger", tth);
                }
                browse_insert_file(browse, file);
            }
            if (name) xmlFree(name);
            if (size) xmlFree(size);
            if (tth) xmlFree(tth);
        }
        xsub = xsub->next;
    }
}

static gboolean xml_get_listing(browse_t* browse, xmlDocPtr xdoc) {
    xmlNodePtr xroot = xmlDocGetRootElement(xdoc);
    xmlNodePtr xsub;
    xmlChar* base;
    char* b;

    if (xroot == NULL) {
        g_log("DC_BROWSE", G_LOG_LEVEL_WARNING, 
              "no root element in listing");
        return FALSE;
    }
    if (strcmp(xroot->name, "FileListing")) {
        g_log("DC_BROWSE", G_LOG_LEVEL_DEBUG,
              "Root xml element != FileListing");
        return FALSE;
    }
    base = xmlGetProp(xroot, "Base");
    if (base && base[strlen(base)-1] != '/') {
        b = g_strdup_printf("%s/", base);
        xmlFree(base);
    } else if (base) {
        b = g_strdup(base);
        xmlFree(base);
    } else {
        b = g_strdup("");
    }

    xsub = xroot->xmlChildrenNode;
    while (xsub != NULL) {
        if (strcmp(xsub->name, "")) {
            xml_get_node(browse, xdoc, xsub, unix2win(b));
        }
        xsub = xsub->next;
    }
    return TRUE;
}

static unsigned
dc_browse_parse_xml(dc_browse_t* browse, buffer_t* buffer) {
    xmlDocPtr xdoc;
  
    xdoc = xmlParseMemory(buffer_read_pos(buffer), buffer_avail(buffer));
    if (!xdoc) return S_BROWSE_FAILED;

    if (!xml_get_listing(BROWSE(browse), xdoc)) {
        xmlFreeDoc(xdoc);
        return S_BROWSE_FAILED;
    }
  
    return S_BROWSE_COMPLETE;
}

static const char* dc_bsock_get_list(tcp_socket_t* s) {
    buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));
    dc_browse_t* browse = socket_get_data(SOCK(s));
    unsigned done = buffer_avail(buffer);
    unsigned todo = browse->todo;
    double prog = (double)done/(double)todo;

    browse_set_progress(BROWSE(browse), prog, _("Browsing..."));
  
    if (done >= todo) {
        buffer_t* decomp;
        unsigned res = S_BROWSE_FAILED;
        if (browse->type == DC_TYPE_XMLBZ || browse->type == DC_TYPE_BZ) {
            compress_t* c;

            decomp = buffer_new(1024);
            c = compress_new(COMPRESS_BZIP2, FALSE, buffer, decomp);
            if (!c || !compress_action(c, TRUE)) {
                if (c) compress_destroy(c);
                buffer_destroy(decomp);
                socket_set_status(SOCK(s), S_BROWSE_FAILED);
                return "decompression error";
            }
            compress_destroy(c);
        } else {
            decomp = buffer;
        }
        switch (browse->type) {
        case DC_TYPE_XMLBZ:
            //      dc_write_share_buffer("browse.xml", decomp);
            res = dc_browse_parse_xml(browse, decomp);
            break;
        case DC_TYPE_BZ:
        case DC_TYPE_DCLST:
            res = S_BROWSE_FAILED;
            //FIXME: implement
            break;
        default:
            g_assert(FALSE);
        }
        if (decomp != buffer) buffer_destroy(decomp);
        socket_set_status(SOCK(s), res);
        return "done";
    }
    return NULL;
}

HANDLER(handle_file_length) {
    char* length = arg(data, 0);
    dc_browse_t* browse = socket_get_data(SOCK(s));
  
    REQUIRE(length, "DC_BROWSE", "invalid $FileLength");
    browse->todo = atoi(length);
    REQUIRE(browse->todo > 0, "DC_BROWSE", "invalid file length");

    dc_send(BSOCK(s), FALSE, "$Send");
    b_socket_set_consume(BSOCK(s), (BConsumeInputFunc)dc_bsock_get_list);
    browse_set_status(BROWSE(browse), NULL);
    return "AGAIN";
}

HANDLER(handle_maxed_out) {
    (void)data;
    socket_set_status(SOCK(s), S_BROWSE_FAILED);
    return "connection denied";
}

static const char*
dc_handle_bsock_message(tcp_socket_t* s, dc_message_t* command) {
    if (!command->type) {
        DC_REC("DC_BROWSE", G_LOG_LEVEL_WARNING);
        return "got invalid command";
    } else if (!strcmp(command->type, "$FileLength")) {
        DC_REC("DC_BROWSE", G_LOG_LEVEL_DEBUG);
        return DC_CALL(handle_file_length);
    } else if (!strcmp(command->type, "$MaxedOut")) {
        DC_REC("DC_BROWSE", G_LOG_LEVEL_DEBUG);
        return DC_CALL(handle_maxed_out);
    } else {
        DC_REC("DC_BROWSE", G_LOG_LEVEL_WARNING);
    }
    return NULL;
}

static const char* dc_bsock_consume_input(tcp_socket_t* s) {
    buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));
    dc_message_t* result;

    if (!buffer) return NULL;
    while ((result = dc_consume_message(buffer)) != NULL) {
        const char* message = dc_handle_bsock_message(s, result);
        dc_message_destroy(result);
        if (message) return message;
    }
    buffer_consume(buffer, 0);
    return NULL;
}

static const char* on_peer_connected(dc_peer_t* peer, tcp_socket_t* s) {
    dc_browse_t* browse = socket_get_data(SOCK(s));

    browse_set_status(BROWSE(browse), _("Getting file list"));
    tcp_socket_move(s, TSOCK(peer), TRUE);

    b_socket_set_consume
        (BSOCK(s), (BConsumeInputFunc)dc_bsock_consume_input);

    if (peer->features & DC_FEATURE_XMLBZLIST) {
        browse->type = DC_TYPE_XMLBZ;
        dc_send(BSOCK(s), FALSE, "$Get files.xml.bz2$1");
    } else if (peer->features & DC_FEATURE_BZLIST) {
        browse->type = DC_TYPE_BZ;
        dc_send(BSOCK(s), FALSE, "$Get MyList.bz2$1");
    } else {
        browse->type = DC_TYPE_DCLST;
        dc_send(BSOCK(s), FALSE, "$Get MyList.DcLst$1");
    }
  
    b_socket_start_receiving(BSOCK(s), 10240000);

    return "handover->dc_browse";
}

void dc_browse_user(dc_user_t* user, dc_hub_t* hub) {
    dc_browse_t* browse = DC_BROWSE(browse_find(USER(user)));

    if (browse) {
        // FIXME
        g_log("BROWSE", G_LOG_LEVEL_WARNING,
              "browse '%s' already existent", OBJ(user)->name);
        return;
    }
    /*
      if (browse->socket) {
      g_log("BROWSE", G_LOG_LEVEL_WARNING,
      "browse '%s' has not socket", OBJ(user)->name);
      return;
      }
      if (socket_connected(SOCK(browse->socket))) {
      g_log("BROWSE", G_LOG_LEVEL_WARNING,
      "browse '%s' already active", OBJ(user)->name);
      return;
      }
    */
    if (!hub && user->hubs) hub = user->hubs->data;
    if (!hub) {
        g_log("BROWSE", G_LOG_LEVEL_WARNING,
              "no hub for browsing found");
        return;
    }
    browse = dc_browse_new(user);
    browse_set_status(BROWSE(browse), _("Waiting for response"));
    dc_peer_request(user, hub, (DcPeerConnected)on_peer_connected,
                    OBJ(browse->socket));
}


static void dc_browses_init()
{
}

static void dc_browses_exit()
{
}

static void dc_browse_module_init(dc_browse_module_t* module)
{
    module_init(MODULE(module), "Dc browse Module",
                (ModuleFunc)dc_browses_init, NULL, (ModuleFunc)dc_browses_exit);
  
    module_require(browse_module_instance);
}


module_t* dc_browse_module_instance()
{
    if (!DcBrowseModule)
    {
        DcBrowseModule = g_malloc0(sizeof(*DcBrowseModule));
        dc_browse_module_init(DcBrowseModule);
    }
    return MODULE(DcBrowseModule);
}



#undef HANDLER
#undef DC_CALL
#undef REQUIRE

#ifdef WIN32
#define _DC_BROWSE_C_
#include "dcaliasdef.c"
#endif


