/* Copyright (C) 2005-2007 sgop@users.sourceforge.net.
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.6 $
 * $Date: 2007/05/23 20:42:28 $
 * $Author: sid4u $
 */

#include <string.h>

#include "lcore/chat.h"
#include "lcore/preferences.h"
#include "lcore/string_list.h"
#include "lcore/ui.h"
#include "lcore/completion.h"
#include "lcore/l_i18n.h"
#include "dc_network.h"
#include "dc_chat.h"
#include "dc_proto.h"
#include "dc.h"
#include "dc_ui.h"

#ifdef WIN32
#include "dcalias.h"
#endif

static GList* DcPages = NULL;

static void string_convert_cr(char* message)
{
    if (!message) return;
    while (*message)
    {
        if (*message == '\r') *message = ' ';
        message++;
    }
}


void dc_server_message(dc_hub_t* hub, char* message)
{
    string_convert_cr(message);
    server_message(hub->short_name, message);
}


static dc_cp_t* dc_cp_search_hub(dc_hub_t* hub, const char* name, int type)
{
    GList* dlist;
  
/*   g_message("searching in %u pages for (%s)[%s]",  */
/*             g_list_length(DcPages), name, OBJ(hub)->name); */
    for (dlist = DcPages; dlist; dlist = dlist->next)
    {
        dc_cp_t* dcp = dlist->data;
/*     g_message("  -  for (%s)[%s]", OBJ(dcp)->name, OBJ(dcp->hub)->name); */
        if (dcp->type == type && dcp->hub == hub &&
            (!name || !strcmp(OBJ(dcp)->name, name))) return dcp;
    }
    return NULL;
}

static dc_cp_t* dc_cp_search(const char* channel, int type)
{
    GList* dlist;
    
    for (dlist = DcPages; dlist; dlist = dlist->next)
    {
        dc_cp_t* dcp = dlist->data;
        if (dcp->type == type &&
            !strcmp(OBJ(dcp)->name, channel)) return dcp;
    }
    return NULL;
}

static gboolean dc_cp_closing(dc_cp_t* page)
{
    if (page->type == DC_PAGE_PRIVATE) return TRUE;
    if (page->type == DC_PAGE_PUBLIC) return TRUE;
    return TRUE;
}

static void on_hub_destroyed(dc_hub_t* hub, dc_cp_t* page)
{
    (void)hub;
    object_kill(OBJ(page));
}

static void dc_cp_update_hub(dc_cp_t* dcp, dc_hub_t* hub)
{
    if (dcp->hub != hub)
    {
        if (dcp->hub)
            object_unnotify_o(OBJ(dcp->hub), (ObjectFunc)on_hub_destroyed, OBJ(dcp));

        dcp->hub = hub;

        if (dcp->hub)
            object_notify_o(OBJ(dcp->hub), (ObjectFunc)on_hub_destroyed, OBJ(dcp));
    }
    chat_page_set_active(PAGE(dcp), (hub->status == DC_HUB_ONLINE));
}

static void dc_cp_remove_user(dc_cp_t* page, dc_user_t* user);
static void on_user_changed(dc_user_t* user, dc_cp_t* page)
{
    if (page->type == DC_PAGE_PUBLIC)
    {
        if (!g_list_find(user->hubs, page->hub) && g_list_find(page->users, user))
            dc_cp_remove_user(page, user);
    }
    else if (page->type == DC_PAGE_PRIVATE)
    {
        if (g_list_find(user->hubs, page->hub))
            chat_page_set_active(PAGE(page), TRUE);
        else if (user->hubs)
            dc_cp_update_hub(page, DC_HUB(user->hubs->data));
        else
            chat_page_set_active(PAGE(page), FALSE);
    }
}

static void dc_cp_remove_user(dc_cp_t* page, dc_user_t* user)
{
    page->users = g_list_remove(page->users, user);
    object_unmonitor_d(OBJ(user), (ObjectFunc)on_user_changed, page);
    
    if (page->type == DC_PAGE_PUBLIC) DCUI.user_vanished(page, user);
    
    user_unref(USER(user));
}

static void dc_cp_add_user(dc_cp_t* page, dc_user_t* user)
{
    user_ref(USER(user));
    page->users = g_list_append(page->users, user);
    
    if (page->type == DC_PAGE_PUBLIC) DCUI.user_appeared(page, user);
    
    object_monitor_d(OBJ(user), (ObjectFunc)on_user_changed, page);
}


static void dc_cp_destroy(dc_cp_t* page)
{
    DcPages = g_list_remove(DcPages, page);
    while (page->users)
        dc_cp_remove_user(page, page->users->data);
    
    chat_page_destroy(PAGE(page));
}


static void dc_cp_init(dc_cp_t* dcp, dc_hub_t* hub, const char* name,
                       const char* label, const char* title, int type)
{
    chat_page_init(PAGE(dcp), name, label, title, DC_ID);
    object_add_type(OBJ(dcp), "DcChatPage");
    OBJ(dcp)->destroy = (ObjectFunc)dc_cp_destroy;
    PAGE(dcp)->close = (ChatCloseFunc)dc_cp_closing;
    dcp->type = type;
    dcp->hub = NULL;
    DcPages = g_list_append(DcPages, dcp);
    
    dc_cp_update_hub(dcp, hub);
}

static dc_cp_t* dc_cp_new(dc_hub_t* hub, const char* name,
                          const char* label, const char* title, int type)
{
    dc_cp_t* dcp = g_malloc0(sizeof(*dcp));
    dc_cp_init(dcp, hub, name, label, title, type);
    return dcp;
}
  
dc_cp_t* dc_cp_new_private(dc_hub_t* hub, const char* nick)
{
    dc_user_t* user;
    dc_cp_t* dcp;
    
    dcp = dc_cp_new(hub, nick, nick,
                    l_printf(_("Private conversation with %s"), nick),
                    DC_PAGE_PRIVATE);
    object_add_type(OBJ(dcp), "DcPrivatePage");
    
    user = dc_user_find(nick);
    object_monitor_o(OBJ(user), (ObjectFunc)on_user_changed, OBJ(dcp));
    
    object_make_public(OBJ(dcp), NULL);

    return dcp;
}

static void on_hub_changed(dc_hub_t* hub, dc_cp_t* dcp)
{
    chat_page_t* page = PAGE(dcp);

    dc_cp_update_hub(dcp, hub);

    if (strcmp(page->label, hub->name))
    {
        chat_page_set_label(page, dcp->hub->name);
        chat_page_set_title(page, l_printf("Public chat on %s", dcp->hub->name));
    }
}


static char* dc_user_name(dc_user_t* user) {
  return OBJ(user)->name;
}

static gint dc_user_compare(const gchar* u1, const gchar* u2, gsize len) {
  int res = strncasecmp(u1, u2, len);
  
  if (!res) return res;

  // skip prefix if there is any.
  if (*u2 != '[') return res;
  u2 = strchr(u2, ']');
  if (!u2) return res;
  u2++;
  if (!strncasecmp(u1, u2, len)) return 0;

  return res;
}

static char* dc_pub_expand(dc_cp_t* page, const char* text)
{
    const char* pos;
    GList* results;
    char* result;
    int offset;
    completion_t* c;
    
    c = completion_new((CompletionFunc)dc_user_name);
    completion_set_compare(c, (CompletionStrncmpFunc)dc_user_compare);
    completion_set_items(c, &page->users);
    
    if (!text)
    {
        results = completion_complete(c, "", &result);
        offset = 0;
        pos = NULL;
    }
    else
    {
        pos = strrchr(text, ' ');
        if (!pos) pos = text;
        else pos++;
        
        if (*pos == '@')
        {
            results = completion_complete(c, pos+1, &result);
            offset = (pos-text)+1;
        }
        else
        {
            results = completion_complete(c, pos, &result);
            offset = (pos-text);
        }
    }

    if (g_list_length(results) > 1)
    {
        GList* dlist;
        unsigned num = g_list_length(results);
        
        chat_print(PAGE(page), "+\0client",
                   "+\1c_client",
                   l_printf(ngettext("%d matching user found", "%d matching users found", num), num),
                   "+\0nl", NULL);
        for (dlist = results; dlist; dlist = dlist->next)
        {
            dc_user_t* user = dlist->data;
            chat_print(PAGE(page), "+\0client",
                       "+\1c_text", "  ",
                       "+\1nick", OBJ(user)->name, "\0\0",
                       "+\0nl", NULL);
        }
        if (result)
        {
            char* t = result;
            result = g_malloc(offset+strlen(t)+1);
            strncpy(result, text, offset);
            sprintf(result+offset, "%s", t);
            g_free(t);
        }
    }
    else if (g_list_length(results) == 1)
    {
        dc_user_t* user = results->data;
        g_free(result);
        result = g_malloc(offset+strlen(OBJ(user)->name)+3);
        strncpy(result, text, offset);
        if (offset < 2) 
            sprintf(result+offset, "%s: ", OBJ(user)->name);
        else
            sprintf(result+offset, "%s ", OBJ(user)->name);
    }
    else
    {
        chat_print(PAGE(page), "+\0client",
                   "+\1c_client", _("No matching user found"),
                   "+\0nl", NULL);
    }
    completion_free(c);
    return result;
}

dc_cp_t* dc_cp_new_public(dc_hub_t* hub)
{
    dc_cp_t* dpp = dc_cp_new(hub, hub->name, hub->name,
                             l_printf("Public chat on %s", hub->name),
                             DC_PAGE_PUBLIC);
    PAGE(dpp)->expand = (ChatExpandFunc)dc_pub_expand;
    object_add_type(OBJ(dpp), "DcPublicPage");
    
    object_monitor_o(OBJ(hub), (ObjectFunc)on_hub_changed, OBJ(dpp));

    object_make_public(OBJ(dpp), NULL);
    return dpp;
}

void dc_motd_message(dc_hub_t* hub, char* message)
{
    string_convert_cr(message);
    motd_message(hub->short_name, message);
}

void dc_emote_message(dc_hub_t* hub, char* nick, char* message)
{
    dc_cp_t* page = dc_chat_active(hub);
    dc_user_t* user;
    
    if (!page) return;
    emote_message(PAGE(page), NULL, NULL, nick, message);
    
    user = dc_user_find(nick);
    if (!g_list_find(page->users, user)) dc_cp_add_user(page, user);
}

void dc_public_message(dc_hub_t* hub, char* nick, char* message)
{
    dc_cp_t* page = dc_chat_active(hub);
    dc_user_t* user;
    
    if (!page) return;
    
    string_convert_cr(message);
    
/*     if (*nick && nick[1] && nick[2]) { */
/*         if (*nick == '<') nick++; */
/*         if (nick[strlen(nick)-1] == '>') nick[strlen(nick)-1] = 0; */
/*     } */
    
    user_message(PAGE(page), NULL, NULL, nick, message);
    
    user = dc_user_find(nick);
    if (!g_list_find(page->users, user)) dc_cp_add_user(page, user);
}



void dc_private_message(dc_hub_t* hub, char* nick, char* message)
{
    chat_page_t *page;
    
    string_convert_cr(message);
    
    // search for an existing priv page and force pipe if found
    if ((page = PAGE(dc_cp_search(nick, DC_PAGE_PRIVATE))))
    {
        if (DCP(page)->hub != hub) dc_cp_update_hub(DCP(page), hub);
        user_message(page, NULL, NULL, nick, message);
    }
    // create new priv page if we want to pipe
    else if ((pref_get_ref("/core/chat/pipe", int) & PIPE_PRIVATE))
    {
        page = PAGE(dc_cp_new_private(hub, nick));
        user_message(page, NULL, NULL, nick, message);
    }
    else
    {
        private_message(NULL, "DC", NULL, nick, message);
    }
}

void dc_send_private_message(dc_hub_t* hub, chat_page_t* page,
                             const char* nick, char *message)
{
    char* pos;
    int private;
    const char* my_nick;
    
    if (!hub) return;
    my_nick = hub->c_nick;
    
    if (page && page->ID == DC_ID && 
        DCP(page)->type == DC_PAGE_PRIVATE &&
        !strcasecmp(OBJ(page)->name, nick))
    {
        private = 1;
    }
    else
    {
        private = 0;
    }
  
    line_insert_lf(message, 0);
    do
    {
        pos = strchr(message, '\n');
        if (pos) *pos = 0;
        dc_hub_send(hub, TRUE, "$To: %s From: %s $<%s> %s",
                    nick, my_nick, my_nick, auto_text(dc_escape(message)));
        if (private) user_message(page, NULL, NULL, hub->c_nick, message);
        else  private_message(page, "to", nick, hub->c_nick, message);
        if (!pos) break;
        message = pos + 1;
    }
    while (*message);
}

void dc_send_public_message(dc_cp_t* dcp, char* message)
{
    char* pos;
    const char* my_nick;
    
    if (!message) return;

    if (!dcp || !dcp->hub) return;
    my_nick = dcp->hub->c_nick;
    
    line_insert_lf(message, 0);
    do
    {
        pos = strchr(message, '\n');
        if (pos) *pos = 0;
        dc_hub_send(dcp->hub, TRUE, "<%s> %s", my_nick, auto_text(dc_escape(message)));
        if (!pos) break;
        message = pos + 1;
    }
    while (*message);
}

dc_cp_t* dc_chat_active(dc_hub_t* hub)
{
    // don't care about name, dc hubs only have one public page
    return dc_cp_search_hub(hub, NULL, DC_PAGE_PUBLIC);
}


#ifdef WIN32
#define _DC_CHAT_C_
#include "dcaliasdef.c"
#endif

