/* Copyright (C) 2005-2006 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.2 $
 * $Date: 2006/05/07 16:51:57 $
 * $Author: sgop $
 */

#include <stdlib.h>

#include "dc_download.h"
#include "dc_proto.h"
#include "dc_peer.h"
#include "dc.h"

typedef struct {
    module_t _module;
    GList* sources;
} dc_download_module_t;

static dc_download_module_t* DcDownloadModule = NULL;

#define HANDLER(f) static const char* f (dc_source_t* source, char* data)
#define DC_CALL(f) f(source, command->message)
#define REQUIRE(expr, proto, message)                                   \
    if (!(expr)) {g_log(proto, G_LOG_LEVEL_WARNING, message); return message;}


static void dc_source_destroy(dc_source_t* source)
{
    DcDownloadModule->sources = g_list_remove(DcDownloadModule->sources, source);
    source_destroy(SOURCE(source));
}

static gboolean dc_source_end(dc_source_t* source, const char* message)
{
    unsigned status = socket_get_status(SOCK(source));
    switch (status) {
    case ST_SOURCE_REMOTE:
    case ST_DL_SEG_FIN:
        source_end(SOURCE(source), message);
        socket_set_status(SOCK(source), ST_UNKNOWN);
        source_set_remote(SOURCE(source), 5, 0);  // restart in 5 secs
        return FALSE;
    default:
        return source_end(SOURCE(source), message);
    }
}

HANDLER(handle_file_length)
{
    char* length = arg(data, 0);
    file_size_t size;
    
    REQUIRE(length, "DC_DOWN", "invalid $FileLength");
    size = strtoll(length, NULL, 10);
    REQUIRE(size == SOURCE(source)->file->size,
            "DC_DOWN", "file size mismatch");
    
    // we hopefully can still use this segment.
    source_use_segment(SOURCE(source), source->req_start, source->req_stop, TRUE);
    if (!source_next_segment(SOURCE(source)))
        return "could not grab next segment";
    
    dc_send(BSOCK(source), FALSE, "$Send");
    
    return source_start_downloading(SOURCE(source));
}

HANDLER(handle_maxed_out)
{
    (void)data;
    source_set_remote(SOURCE(source), 60*5, 0);
    socket_set_status(SOCK(source), ST_SOURCE_REMOTE);
    return "remotely queued";
}

static const char* dc_handle_source_message(dc_source_t* source, dc_message_t* command)
{
  if (!command->type)
  {
      DC_REC("DC_DOWN", G_LOG_LEVEL_WARNING);
      return "got invalid command";
  }
  else if (!strcmp(command->type, "$FileLength"))
  {
      DC_REC("DC_DOWN", G_LOG_LEVEL_DEBUG);
      return DC_CALL(handle_file_length);
  }
  else if (!strcmp(command->type, "$MaxedOut"))
  {
      DC_REC("DC_DOWN", G_LOG_LEVEL_DEBUG);
      return DC_CALL(handle_maxed_out);
  }
  else
  {
      DC_REC("DC_DOWN", G_LOG_LEVEL_WARNING);
  }
  return NULL;
}

static const char* dc_source_consume_input(dc_source_t* source)
{
    buffer_t* buffer = b_socket_get_in_buffer(BSOCK(source));
    dc_message_t* result;
    
    if (!buffer) return NULL;
    while ((result = dc_consume_message(buffer)) != NULL)
    {
        const char* message = dc_handle_source_message(source, result);
        dc_message_destroy(result);
        if (message) return message;
    }
    buffer_consume(buffer, 0);
    return NULL;
}


static const char* on_peer_connected(dc_peer_t* peer, dc_source_t* source)
{
    segment_t* seg;
    
    g_log("DC_DOWN", G_LOG_LEVEL_DEBUG, "peer response");
    tcp_socket_move(TSOCK(source), TSOCK(peer), TRUE);
    
    seg = source_request_segment(SOURCE(source), 0);
    if (!seg) {
        socket_done(SOCK(source), ST_ERROR, "no segment");
        return "no segment";
    }
    
    // send request (DC counts offset from 1, not 0)
    dc_send(BSOCK(source), FALSE, "$Get %s$"FF,
            SOURCE(source)->file->netname, seg->start+seg->size+1);
    // save requested segment
    source->req_start = seg->start;
    source->req_stop = seg->stop;
    
    b_socket_set_consume(BSOCK(source), (BConsumeInputFunc)dc_source_consume_input);
    
    b_socket_start_receiving(BSOCK(source), 10240);
    
    return "handover->dc_source";
}

static gboolean dc_source_start(dc_source_t* source)
{
    dc_user_t* user = DC_USER(SOURCE(source)->user);
    if (!user->hubs) return FALSE;
    
    dc_peer_request(user, (dc_hub_t*)(user->hubs->data),
                    (DcPeerConnected)on_peer_connected, OBJ(source));
    
    return TRUE;
}


static void dc_source_init(dc_source_t* source, file_t* file, dc_user_t* user)
{
    source_init(SOURCE(source), file, USER(user), TRUE);
    object_add_type(OBJ(source), "DcSource");
    OBJ(source)->destroy = (ObjectFunc)dc_source_destroy;
    socket_set_end(SOCK(source), (SockEndFunc)dc_source_end);
    SOURCE(source)->start = (SourceStartFunc)dc_source_start;
    
    DcDownloadModule->sources = g_list_append(DcDownloadModule->sources, source);
}

static dc_source_t* dc_source_new(file_t* file, dc_user_t* user)
{
    dc_source_t* source = g_malloc0(sizeof(*source));
    dc_source_init(source, file, user);
    return source;
}

static dc_source_t*
dc_file_to_source(const download_t* download, file_t* file, dc_user_t* user)
{
    dc_source_t* source = dc_source_new(file, user);
    (void)download;
    return source;
}


static gboolean
dc_download_matches_source(download_t* download, file_t* file, dc_user_t* user)
{
    gboolean same_hash = FALSE;
    attribute_t* attr;
    
    attr = attr_find(ATTRC(download), "file", "urn:tree:tiger");
    if (attr)
    {
        attribute_t* hash = attr_find(ATTRC(file), "file", "urn:tree:tiger");
        if (hash && compare_hash(attr->value, hash->value)) return FALSE;
        else if (hash) same_hash = TRUE;
    }
    
    if (same_hash)
    {
        // check size only if hash matches
        if (FILE(download)->size != file->size) return FALSE;
    }
    else
    {
        if (!download_check_file(download, file)) return FALSE;
    }

    return TRUE;
}


static void dc_downloads_init()
{
    source_factory_register(DC_ID, (FileToSourceFunc)dc_file_to_source);
    download_register_match(DC_ID, (DownloadMatchFunc)dc_download_matches_source);
}

static void dc_downloads_run()
{
}

static void dc_downloads_exit()
{
}

static void dc_download_module_init(dc_download_module_t* module)
{
    module_init(MODULE(module), "Dc Download Module",
                (ModuleFunc)dc_downloads_init,
                (ModuleFunc)dc_downloads_run,
                (ModuleFunc)dc_downloads_exit);
    
    module_require(download_module_instance);
}

module_t* dc_download_module_instance()
{
    if (!DcDownloadModule)
    {
        DcDownloadModule = g_malloc0(sizeof(*DcDownloadModule));
        dc_download_module_init(DcDownloadModule);
    }
    return MODULE(DcDownloadModule);
}
