/* Copyright (C) 2005-2006 sgop@users.sourceforge.net.
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.2 $
 * $Date: 2006/05/07 16:51:57 $
 * $Author: sgop $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <stdlib.h>

#include "lcore/utils.h"
#include "lcore/library.h"

#include "dc_peer.h"
#include "dc.h"
#include "dc_upload.h"
#include "dc_share.h"

#define HANDLER(f) static const char* f (dc_peer_t* peer, char* data)
#define DC_CALL(f) f(peer, command->message)
#define REQUIRE(expr, proto, message) \
  if (!(expr)) {g_log(proto, G_LOG_LEVEL_WARNING, message); return message;}


typedef struct {
  dc_user_t* user;
  DcPeerConnected connected;
  object_t* obj;
} dc_peer_req_t;

static GList* PeerRequests = NULL;

static void dc_peer_request_delete(dc_peer_req_t* req);
static dc_peer_req_t* dc_peer_request_find(dc_peer_t* peer);


static gboolean dc_peer_end(dc_peer_t* peer, const char* message)
{
    tcp_socket_end(TSOCK(peer), message);
    return TRUE;
}

static void dc_peer_destroy(dc_peer_t* peer) {
  if (peer->user) user_unref(USER(peer->user));
  tcp_socket_destroy(TSOCK(peer));
}

static const char* dc_handle_peer_message(dc_peer_t* peer, dc_message_t* command);
static const char* dc_peer_consume(dc_peer_t* peer) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(peer));
  dc_message_t* result;

  if (!buffer) return NULL;
  while ((result = dc_consume_message(buffer)) != NULL) {
    const char* message = dc_handle_peer_message(peer, result);
    dc_message_destroy(result);
    if (message) return message;
  }
  buffer_consume(buffer, 0);
  return NULL;
}

void dc_peer_init(dc_peer_t* peer) {
  tcp_socket_init(TSOCK(peer), "DcPeer", 30);
  object_add_type(OBJ(peer), "DcPeer");
  OBJ(peer)->destroy = (ObjectFunc)dc_peer_destroy;
  socket_set_end(SOCK(peer), (SockEndFunc)dc_peer_end);
  b_socket_set_consume(BSOCK(peer), (BConsumeInputFunc)dc_peer_consume);
  peer->flags = 0;
  peer->user = NULL;
  socket_add_load(SOCK(peer), "Peer");
}

dc_peer_t* dc_peer_new() {
  dc_peer_t* peer = g_malloc0(sizeof(*peer));
  dc_peer_init(peer);
  return peer;
}

static void dc_peer_set_user(dc_peer_t* peer, dc_user_t* user) {
  if (peer->user) user_unref(USER(peer->user));
  peer->user = user_ref(USER(user));
  object_set_name(OBJ(peer), OBJ(user)->name);

  user_set_ip(user, TSOCK(peer)->ip);

  if (dc_peer_request_find(peer))
    peer->flags |= DC_PEER_DOWNLOAD;
  else
    peer->flags &= ~DC_PEER_DOWNLOAD;
}

static int dc_peer_get_number(dc_peer_t* peer) {
  guint32 num = (guint32)GPOINTER_TO_INT(peer);
  return ((num>>2) & 0x7fff);
}

/* static void dc_peer_send(dc_peer_t* peer, gboolean convert, const char* fmt, ...) { */
/*   va_list ap; */
/*   char* message; */
/*   buffer_t* buffer; */

/*   g_assert(DcHubModule); */

/*   va_start(ap, fmt); */
/*   message = g_strdup_vprintf(fmt, ap); */
/*   va_end(ap); */

/*   buffer = b_socket_get_out_buffer(BSOCK(hub), 1024); */
/*   if (convert) */
/*     dc_pack_command(buffer, "DC_PEeR", auto_text */
/*                     (from_utf8(message, -1, NULL, NULL, 2, "windows-1252", "utf-8"))); */
/*   else */
/*     dc_pack_command(buffer, "DC_PEER", message); */
/*   b_socket_start_sending(BSOCK(hub)); */

/*   g_free(message); */
/* } */

static void dc_peer_greet(dc_peer_t* peer, const char* nick) {
  const char* lock =
    "EXTENDEDPROTOCOL@LII@testversionII Pk=L@O@P@S@T@E@R@-@2@0@0";
  dc_send(BSOCK(peer), FALSE, "$MyNick %s", nick);
  dc_send(BSOCK(peer), FALSE, "$Lock %s", lock);
}


HANDLER(handle_my_nick) {
  char* nick = arg(data, 0);
  dc_user_t* user;

  REQUIRE(nick, "DC_PEER", "invalid $MyNick");

  user = dc_user_find(nick);
  dc_peer_set_user(peer, user);

  if (peer->flags & DC_PEER_INCOMING) {
    if (user->hubs) {
      dc_hub_t* hub = user->hubs->data;
      dc_peer_greet(peer, hub->c_nick);
    } else {
      dc_peer_greet(peer, dc_hub_get_nick(NULL));
    }
  }

  return NULL;
}

HANDLER(handle_lock) {
  char* lock = arg(data, 0);
  char* pk = arg(NULL, 1);
  char* key;

  REQUIRE(lock, "DC_PEER", "invalid $Lock");
  
  if (pk && !strcmp(pk, "Pk=")) pk += 3;

  if (!strncmp(lock, "EXTENDEDPROTOCOL", 16)) {
    // just testing for now
    dc_send(BSOCK(peer), FALSE, "$Supports BZList XmlBZList");
    //FIXME: is UGetBlock implied?
    //MiniSlots
  }

  if (peer->flags & DC_PEER_DOWNLOAD) {
    dc_send(BSOCK(peer), FALSE, "$Direction Download %d",
            dc_peer_get_number(peer));
  } else {
    dc_send(BSOCK(peer), FALSE, "$Direction Upload %d",
            dc_peer_get_number(peer));
  }

  key = dc_generate_key(lock);
  dc_send(BSOCK(peer), FALSE, "$Key %s", key);
  g_free(key);
  return NULL;
}


HANDLER(handle_supports)
{
    char* feature;
    
    feature = arg(data, 0);
    peer->features = 0;
    while (feature)
    {
        if (!strcmp(feature, "BZList"))
            peer->features |= DC_FEATURE_BZLIST;
        else if (!strcmp(feature, "MiniSlots"))
            peer->features |= DC_FEATURE_MINISLOTS;
        else if (!strcmp(feature, "XmlBZList"))
            peer->features |= DC_FEATURE_XMLBZLIST;
        else if (!strcmp(feature, "GetZBlock"))
            peer->features |= DC_FEATURE_GETZBLOCK;
        else if (strcmp(feature, "ADCGet") &&
                 strcmp(feature, "ZLIG") &&
                 strcmp(feature, "TTHL") &&
                 strcmp(feature, "TTHF"))
            g_log("DC_PEER", G_LOG_LEVEL_WARNING,
                  "feature '%s' not supported", feature);
        feature = arg(NULL, 0);
  }
  return NULL;
}

HANDLER(handle_direction) {
  char* direction = arg(data, 0);
  char* num = arg(NULL, 0);

  REQUIRE(num, "DC_PEER", "invalid $Direction");

  if (!strcmp(direction, "Upload")) {
    /*
    REQUIRE(peer->flags & DC_PEER_DOWNLOAD, "DC_PEER",
	    "Both wish to upload");
    */
    // we dont expect data, but lets see what we receive
    peer->flags |= DC_PEER_DOWNLOAD;
  } else if (!strcmp(direction, "Download")) {
    if ((peer->flags & DC_PEER_DOWNLOAD) && dc_peer_get_number(peer) < atoi(num)) {
      peer->flags &= ~DC_PEER_DOWNLOAD;
    }
  } else {
    REQUIRE(FALSE, "DC_PEER", "unknown direction");
  }
  return NULL;
}

HANDLER(handle_key) {
  char* lock = arg(data, 0);
  const char* res = NULL;

  (void)peer;
  REQUIRE(lock, "DC_PEER", "invalid $Lock");

  if (peer->flags & DC_PEER_DOWNLOAD) {
    dc_peer_req_t* req = dc_peer_request_find(peer);

    if (!req) return "don't need anything";
    
    res = req->connected(peer, req->obj);
    dc_peer_request_delete(req);
  }

  return res;
}

HANDLER(handle_get) {
  char* filename;
  char* offset;
  dc_upload_t* upload;
  file_t* file;
  file_size_t off;

  REQUIRE(!(peer->flags & DC_PEER_DOWNLOAD),
          "DC_PEER", "this is not an upload");

  filename = data;
  offset = strrchr(data, '$');
  REQUIRE(offset, "DC_PEER", "invalid $Get : no '$' found");
  *offset++ = 0;
  
  REQUIRE(filename, "DC_PEER", "invalid $Get : no filename given");

  off = strtoll(offset, NULL, 10);
  REQUIRE(off>0, "DC_PEER", "invalid $Get : invalid offset");
  off--;

  file = dc_get_special_file(filename);
  if (file) {
    if (off >= file->size) {
      dc_send(BSOCK(peer), FALSE, "$Error Invalid offset");
      object_kill(OBJ(file));
      return "invalid offset";
    }
    upload = dc_upload_new(peer->user, file);
    UPLOAD(upload)->base_priority = UPLOAD_PRI_URGENT;
    upload_insert(UPLOAD(upload), FALSE);
    object_kill(OBJ(file));
  } else {
    upload = DC_UPLOAD(upload_find(USER(peer->user), filename));
    if (!upload) {
      file = library_winname_requested(filename);
      if (!file) {
        dc_send(BSOCK(peer), FALSE, "$Error File not found");
        return "file no found";
      } else if (off >= file->size) {
        dc_send(BSOCK(peer), FALSE, "$Error Invalid offset");
        return "invalid offset";
      }
      upload = dc_upload_new(peer->user, file);
      upload_insert(UPLOAD(upload), FALSE);
    }
  }

  dc_upload_set_offset(upload, off, 0);
  return dc_upload_set_connection(upload, peer);
}

HANDLER(handle_uget_block) {
  char* offset = arg(data, 0);
  char* size = arg(NULL, 0);
  char* filename = arg(NULL, 1);
  dc_upload_t* upload;
  file_t* file;
  file_size_t start;
  file_size_t stop;

  REQUIRE(!(peer->flags & DC_PEER_DOWNLOAD),
          "DC_PEER", "this is not an upload");

  REQUIRE(filename, "DC_PEER", "invalid $UGetBlock : not enough arguments");

  start = strtoll(offset, NULL, 10);
  stop = start + strtoll(size, NULL, 10);
  
  file = dc_get_special_file(filename);
  if (file) {
    if (start < 0 || stop <= start || stop > file->size) {
      dc_send(BSOCK(peer), FALSE, "$Error Invalid offset");
      object_kill(OBJ(file));
      return "invalid offset";
    }
    upload = dc_upload_new(peer->user, file);
    UPLOAD(upload)->base_priority = UPLOAD_PRI_URGENT;
    upload_insert(UPLOAD(upload), FALSE);
    object_kill(OBJ(file));
  } else {
    //FIXME: this is a workaround for files.xml.bz2 requests.
    char* fname = g_strdup_printf("\\%s", filename);
    upload = DC_UPLOAD(upload_find(USER(peer->user), fname));
    if (!upload) {
      file = library_winname_requested(fname);
      g_free(fname);
      if (!file) {
        dc_send(BSOCK(peer), FALSE, "$Error File not found");
        return "file no found";
      } else if (start < 0 || stop <= start || stop > file->size) {
        dc_send(BSOCK(peer), FALSE, "$Error Invalid offset");
        return "invalid offset";
      }
      upload = dc_upload_new(peer->user, file);
      upload_insert(UPLOAD(upload), FALSE);
    } else {
      g_free(fname);
    }
  }

  dc_upload_set_offset(upload, start, stop);
  return dc_upload_set_connection(upload, peer);
}

HANDLER(handle_get_list_len)
{
    file_t* file = dc_get_special_file("MyList.DcLst");
    if (file)
    {
        dc_send(BSOCK(peer), FALSE, "$ListLen "FF, file->size);
        object_kill(OBJ(file));
        return NULL;
    }
    else
    {
        dc_send(BSOCK(peer), FALSE, "$Error File not found");
        return "MyList.DcLst not found";
    }
}

static const char*
dc_handle_peer_message(dc_peer_t* peer, dc_message_t* command)
{
    if (!command->type)
    {
        DC_REC("DC_PEER", G_LOG_LEVEL_WARNING);
    }
    else if (!strcmp(command->type, "$MyNick"))
    {
        DC_REC("DC_PEER", G_LOG_LEVEL_DEBUG);
        return DC_CALL(handle_my_nick);
    }
    else if (!strcmp(command->type, "$Lock"))
    {
        DC_REC("DC_PEER", G_LOG_LEVEL_DEBUG);
        return DC_CALL(handle_lock);
    }
    else if (!strcmp(command->type, "$Key"))
    {
        DC_REC("DC_PEER", G_LOG_LEVEL_DEBUG);
        return DC_CALL(handle_key);
    }
    else if (!strcmp(command->type, "$Supports"))
    {
        DC_REC("DC_PEER", G_LOG_LEVEL_DEBUG);
        return DC_CALL(handle_supports);
    }
    else if (!strcmp(command->type, "$Get"))
    {
        DC_REC("DC_PEER", G_LOG_LEVEL_DEBUG);
        return DC_CALL(handle_get);
    }
    else if (!strcmp(command->type, "$UGetBlock"))
    {
        DC_REC("DC_PEER", G_LOG_LEVEL_DEBUG);
        return DC_CALL(handle_uget_block);
    }
    else if (!strcmp(command->type, "$GetListLen"))
    {
        DC_REC("DC_PEER", G_LOG_LEVEL_DEBUG);
        return DC_CALL(handle_get_list_len);
    }
    else if (!strcmp(command->type, "$Direction"))
    {
        DC_REC("DC_PEER", G_LOG_LEVEL_DEBUG);
        return DC_CALL(handle_direction);
    }
    else
    {
        DC_REC("DC_PEER", G_LOG_LEVEL_WARNING);
    }
    return NULL;
}

void dc_peer_connect_to_user(dc_hub_t* hub, ip_t ip, port_t port) {
  dc_peer_t* peer = dc_peer_new();

  if (!tcp_socket_connect(TSOCK(peer), ip, port)) return;

  // greet the peer
  dc_peer_greet(peer, hub->c_nick);
  b_socket_start_receiving(BSOCK(peer), 0);
}



static void on_request_destroy(object_t* object, dc_peer_req_t* req) {
  (void)object;
  req->obj = NULL;
  dc_peer_request_delete(req);
}

static void dc_peer_request_delete(dc_peer_req_t* req)
{
    GList* dlist = g_list_find(PeerRequests, req);
    if (!dlist) return;  // not existen anymore.
    
    PeerRequests = g_list_delete_link(PeerRequests, dlist);
    if (req->obj)
    {
        object_unnotify_d(req->obj, (ObjectFunc)on_request_destroy, req);
    }
    user_unref(USER(req->user));
    g_free(req);
}

void dc_peer_request(dc_user_t* user, dc_hub_t* hub,
                     DcPeerConnected func, object_t* obj)
{
    dc_peer_req_t* req = g_malloc0(sizeof(*req));
    port_t port;
    
    req->user = user_ref(USER(user));
    req->obj = obj;
    req->connected = func;
    PeerRequests = g_list_append(PeerRequests, req);
    
    if (obj)
    {
        object_notify_d(obj, (ObjectFunc)on_request_destroy, req);
    }
    
    port = dc_tcp_listen_port();
    if (port)
    {
        dc_hub_send(hub, FALSE, "$ConnectToMe %s %s:%u",
                    OBJ(user)->name, ntoa(get_local_ip(SOCK(hub))), port);
    }
    else
    {
        // FIXME: we should probably use another hub->c_nick
        // because we must send the correct greeting when the connection comes in
        dc_hub_send(hub, FALSE, "$RevConnectToMe %s %s",
                    hub->c_nick, OBJ(user)->name);
    }
}

static dc_peer_req_t* dc_peer_request_find(dc_peer_t* peer) {
  GList* dlist;
  if (!peer) return NULL;
  for (dlist = PeerRequests; dlist; dlist = dlist->next) {
    dc_peer_req_t* req = dlist->data;
    if (req->user == peer->user) return req;
  }
  return NULL;
}

#undef HANDLER
#undef DC_CALL
#undef REQUIRE
