/* Copyright (C) 2005-2006 sgop@users.sourceforge.net.
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.2 $
 * $Date: 2006/09/26 16:34:40 $
 * $Author: sgop $
 */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include "lcore/search.h"
#include "lcore/mtypes.h"
#include "lcore/l_base.h"
#include "lcore/udp_socket.h"
#include "lcore/library.h"
#include "lcore/lib_search.h"
#include "lcore/upload.h"

#include "dc_network.h"
#include "dc.h"
#include "dc_proto.h"
#include "dc_search.h"
#include "dc_share.h"

#define DC_SEARCH_LIMIT   50

typedef struct {
  GList* token;
  file_size_t min_size;
  file_size_t max_size;
  int media;

  int max_results;
  GList* results;
} dc_search_request_t;

typedef struct {
  module_t _module;
  GList* searches;
  int search_cnt;
  time_t search_time;
  udp_socket_t* udp;
} dc_search_module_t;

static dc_search_module_t* DcSearchModule = NULL;


static gboolean dc_search_is_allowed() {
    if (CurrentTime > DcSearchModule->search_time)
    {
        DcSearchModule->search_time = CurrentTime;
        DcSearchModule->search_cnt = 0;
    }
    if (DcSearchModule->search_cnt < DC_SEARCH_LIMIT)
    {
        DcSearchModule->search_cnt++;
        return TRUE;
    }
    return FALSE;
}

char* dc_search_create_request(search_t* search)
{
    char* req;
    char* keys = NULL;
    file_size_t size = 0;
    gboolean is_max = FALSE;
    attribute_t* hash;
    int media;
    
    if (search->download)
    {
        // if download search, use tth from download
        hash = attr_find(ATTRC(search->download), "file", "urn:tree:tiger");
        if (!hash)
            g_log("DC_SEARCH", G_LOG_LEVEL_DEBUG, "NO HASH FOUND");
    } 
    else
    {
        // else use tth from search pattern
        hash = attr_find(ATTRC(search), "pattern", "urn:tree:tiger");
    }

    if (hash)
    {
        is_max = TRUE;
        size = 0;
        keys = g_strdup_printf("TTH:%s", attr_xml_print(hash));
        media = DC_MEDIA_TTH;
    }
    else if (search->download)
    {
        is_max = FALSE;
        size = FILE(search->download)->size;
        keys = make_string_from_list(search->download->search_keys, "$");
        media = dc_set_media(FILE(search->download)->media);
    }
    else
    {
        int* vec;
        attribute_t* attr;
        
        if (!SEARCHP(search)->include) return NULL;
        attr = attr_find(ATTRC(search), "pattern", "Size");
        if (attr)
        {
            vec = attr_get_ptr(attr, int);
            if (vec[0] == vec[1])
            {
                is_max = FALSE;
                size = vec[0];
            }
            else if (vec[0])
            {
                is_max = FALSE;
                size = vec[0];
            }
            else if (vec[1])
            {
                is_max = TRUE;
                size = vec[1];
            }
        }
        keys = make_string_from_list(SEARCHP(search)->include, "$");
        media = dc_set_media(SEARCHP(search)->media);
    }
    
    req = g_strdup_printf("%c?%c?"FF"?%d?%s$",
                          size>0?'T':'F', is_max?'T':'F',
                          size, media, keys);
    g_free(keys);
    return req;
}

static void dc_search_queue(search_t* search, dc_hub_t* hub)
{
    dc_hub_send_search(hub, search);
}

void dc_search_hub_entered(dc_hub_t* hub)
{
    search_foreach_continuous((ObjectFunc)dc_search_queue, hub);
}

static void dc_search_remove(search_t* pattern)
{
    DcSearchModule->searches = g_list_remove(DcSearchModule->searches, pattern);
}

static void dc_search_new(search_t* search)
{
    g_assert(DcSearchModule);
    g_assert(search);
    
    if (!g_list_find(DcSearchModule->searches, search))
    {
        object_notify_d(OBJ(search), (ObjectFunc)dc_search_remove, NULL);
        DcSearchModule->searches = g_list_prepend(DcSearchModule->searches, search);
        if (search->download)
            search_set_continuous(search, 30, (ObjectFunc)dc_search_queue, NULL);
    }
    dc_search_queue(search, NULL);
}

void dc_search_insert_result(dc_user_t* user, file_t* file,
                             unsigned fs, unsigned as) {
  GList* dlist;
  
  (void)as;
  for (dlist = DcSearchModule->searches; dlist; dlist = dlist->next) {
    search_t* search = dlist->data;
    attribute_t* attr;
    
    if (search->download) {
      if (download_matches_source(search->download, file, USER(user)))
        download_add_source(search->download, file, USER(user));
    } else {
      if (!search_pattern_check_file(SEARCHP(search), file, TRUE)) continue;

      attr = attr_find(ATTRC(search->download), "pattern", "urn:tree:tiger");
      if (attr) {
        attribute_t* hash = attr_find(ATTRC(file), "file", "urn:tree:tiger");
        if (!hash || compare_hash(attr->value, hash->value)) continue;
      }

      search_add_result(search, file, USER(user), 1, (fs > 0));
    }
  }
  object_kill(OBJ(file));
}

static void
dc_udp_send(ip_t ip, port_t port, const char* command, const char* mes) {
  g_assert(command);
  if (!DcSearchModule->udp) {
    DcSearchModule->udp = udp_socket_new("DcSearchUdp", 0);
    object_set_ptr((object_t**)&DcSearchModule->udp);
    if (!socket_activate(SOCK(DcSearchModule->udp))) return;
  }
  if (DcSearchModule->udp) {
    udp_socket_send_str_to
      (DcSearchModule->udp,
       l_printf("%s%s%s|", command, mes?" ":"", mes?mes:""),
       ip, htons(port));

    //    if (strcmp(command, "$SR"))
    g_log("DC_UDP", G_LOG_LEVEL_DEBUG, "S: %13s [%s]", command, mes?mes:"");
  }
}

void dc_search_answer(dc_hub_t* hub, const char* add1,
                      const char* add2, char* query,
                      int media, file_size_t min, file_size_t max) {
  
  gboolean passive_search = (!strcmp(add1, "Hub"));
  gboolean passive_mode = dc_in_passive_mode();
  GList* results = NULL;

  //FIXME: handle FOLDER searches
  if (media == DC_MEDIA_FOLDER) return;
  // do nothing if both in passive mode
  if (passive_mode && passive_search) return;
  
  if (!dc_search_is_allowed()) {
    //    g_log("DC_SEARCH", G_LOG_LEVEL_DEBUG, "ignoring search");
    return;
  }
  
  if (media == DC_MEDIA_TTH) {
    if (!strncmp(query, "TTH:", 4)) {
      file_t* file = dc_file_find_by_tth(query+4);
      if (file) results = g_list_append(results, file);
    }
  } else {
    search_pattern_t* search = search_pattern_new();

    make_list_from_string(&search->include, query, " $");
    if (min || max)
      attr_store(ATTRC(search), "pattern", "Size", min, max);
    search->media = dc_get_media(media);
    if (search->media == MEDIA_INVALID) search->media = MEDIA_ANY;

    //FIXME:    search->max_results = passive_search?10:20;
    results = lib_search_do(search);
    object_kill(OBJ(search));
  }

  if (results) {
    GList* dlist;
    int fslots = uploads_slots_free();
    int slots = uploads_slots_number();
    if (fslots < 0) fslots = 0;

    for (dlist = results; dlist; dlist = dlist->next) {
      file_t* file = dlist->data;
      attribute_t* attr = attr_find(ATTRC(file), "file", "urn:tree:tiger");
      const char* tth = attr_xml_print(attr);
      
      if (passive_search) {
        dc_hub_send
          (hub, FALSE, "$SR %s %s\x05"FF" %d/%d\x05%s (%s:%d)\x05%s",
           auto_text(from_utf8(hub->c_nick, -1, NULL, NULL, 3,
                               hub->charset, "windows-1252", "utf-8")),
           file->netname, file->size, fslots, slots,
           tth
           ?l_printf("TTH:%s", tth)
           :auto_text(from_utf8(hub->name, -1, NULL, NULL, 3,
                                hub->charset, "windows-1252", "utf-8")),
           ntoa(TSOCK(hub)->ip), ntohs(TSOCK(hub)->port),
           add2?add2:""
           );
      } else if (hub) {
        dc_udp_send
          (inet_addr(add1), add2?atoi(add2):411, "$SR",
           l_printf("%s %s\x05"FF" %d/%d\x05%s (%s:%d)",
                    auto_text(from_utf8(hub->c_nick, -1, NULL, NULL, 3,
                                        hub->charset, "windows-1252", "utf-8")),
                    file->netname, file->size, fslots, slots,
                    tth
                    ?l_printf("TTH:%s", tth)
                    :auto_text(from_utf8(hub->name, -1, NULL, NULL, 3,
                                         hub->charset, "windows-1252", "utf-8")),
                    ntoa(TSOCK(hub)->ip), ntohs(TSOCK(hub)->port)
                    )
           );
      } else {
        dc_udp_send
          (inet_addr(add1), add2?atoi(add2):411, "$SR",
           l_printf("%s %s\x05"FF" %d/%d\x05%s (%s:%d)",
                    auto_text(from_utf8(dc_hub_get_nick(NULL), -1, NULL, NULL, 2,
                                        "windows-1252", "utf-8")),
                    file->netname, file->size, fslots, slots,
                    tth?l_printf("TTH:%s", tth):"unknown",
                    "0.0.0.0", 0
                    )
           );
      }
    }
    g_list_free(results);
  }
}

static int search_result_compare_tth(file_t* file1, file_t* file2) {
  attribute_t* attr1;
  attribute_t* attr2;
  hash_t* hash1;
  hash_t* hash2;

  // file1 gets inserted
  attr2 = attr_find(ATTRC(file2), "file", "urn:tree:tiger");
  hash2 = attr2?attr_get_ptr(attr2, hash_t):NULL;
  attr1 = attr_find(ATTRC(file1), "file", "urn:tree:tiger");
  hash1 = attr1?attr_get_ptr(attr1, hash_t):NULL;
  // take as equal if one file has no hash.
  if (!hash1 || !hash2) return 0;

  return compare_hash(hash1, hash2);
}




static void dc_searches_init() {

  attr_register_hash("pattern", "urn:tree:tiger", ATTR_UNIMPORTANT, NULL);

  search_register_action(dc_search_new);

  search_register_result_compare
    ((GCompareFunc)search_result_compare_tth, TRUE);
  //FIXME: register file check func for TTH
}

static void dc_searches_run() {
}

static void dc_searches_exit() {
}

static void dc_search_module_init(dc_search_module_t* module) {
  module_init(MODULE(module), "Dc Search Module",
              (ModuleFunc)dc_searches_init,
              (ModuleFunc)dc_searches_run,
              (ModuleFunc)dc_searches_exit);
  
  module->searches = NULL;
  module->search_cnt = 0;
  module->search_time = 0;
  module->udp = NULL;

  module_require(base_module_instance);
  module_require(search_module_instance);
  module_require(lib_search_module_instance);
}


module_t* dc_search_module_instance() {
  if (!DcSearchModule) {
    DcSearchModule = g_malloc0(sizeof(*DcSearchModule));
    dc_search_module_init(DcSearchModule);
  }
  return MODULE(DcSearchModule);
}

