/* Copyright (C) 2005-2007 sgop@users.sourceforge.net.
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.5 $
 * $Date: 2007/04/29 09:28:12 $
 * $Author: sgop $
 */

#include <glib.h>

#include "lcore/utils.h"
#include "lcore/l_socket.h"

#include "dc_users.h"
#include "dc.h"

typedef struct {
    module_t _module;
    GHashTable* users;
} dc_user_module_t;

static dc_user_module_t* DcUserModule = NULL;


static void dc_user_destroy(dc_user_t* user)
{
    g_assert(!user->hubs);
    g_assert(!user->opped);

    g_hash_table_remove(DcUserModule->users, OBJ(user)->name);

    g_free(user->description);
    g_free(user->linespeed);
    g_free(user->email);
    g_list_free(user->hubs);

    user_destroy(USER(user));
}

static void dc_user_monitor(dc_user_t* user)
{
    // users on dc always are monitored, do nothing
    (void)user;
}

static void dc_user_unmonitor(dc_user_t* user)
{
    // users on dc always are monitored, do nothing
    (void)user;
}

static void dc_user_init(dc_user_t* user, const char* nick) {
    user_init(USER(user), nick, DC_ID);
    object_add_type(OBJ(user), "DcUser");
    OBJ(user)->destroy = (ObjectFunc)dc_user_destroy;
    USER(user)->monitor = (UserMonitor)dc_user_monitor;
    USER(user)->unmonitor = (UserMonitor)dc_user_unmonitor;
    USER(user)->ip = 0;
    user->description = NULL;
    user->linespeed = NULL;
    user->email = NULL;
    user->bytes = 0;
    user->opped = NULL;
    user->hubs = NULL;
    g_hash_table_insert(DcUserModule->users, OBJ(user)->name, user);
}

static dc_user_t* dc_user_new(const char* nick)
{
    dc_user_t* user = g_malloc0(sizeof(*user));
    dc_user_init(user, nick);
    return user;
}

void dc_user_is_online(dc_user_t* user, dc_hub_t* hub)
{
    if (!hub->users)
    {
        hub->users =
            g_hash_table_new_full
            ((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
             (GDestroyNotify)NULL, (GDestroyNotify)NULL);
    }
    if (!g_hash_table_lookup(hub->users, OBJ(user)->name))
    {
        user_ref(USER(user));
        g_hash_table_insert(hub->users, OBJ(user)->name, user);
        user_set_online(USER(user), TRUE);
        user->hubs = g_list_append(user->hubs, hub);
        hub->bytes += user->bytes;
        object_set_dirty(OBJ(user));
    }
    else if (!(hub->features & DC_FEATURE_NOHELLO))
    {
        g_log("DC_USERS", G_LOG_LEVEL_WARNING,
              "user '%s' already on hub", OBJ(user)->name);
    }
}

static void dc_user_is_offline_real(dc_user_t* user, dc_hub_t* hub)
{
    user->hubs = g_list_remove(user->hubs, hub);
    if (user->opped) user->opped = g_list_remove(user->opped, hub);
    if (!user->hubs) user_set_online(USER(user), FALSE);
    if (user_unref(USER(user))) object_set_dirty(OBJ(user));
}

void dc_user_is_offline(dc_user_t* user, dc_hub_t* hub)
{
    if (hub->users && g_hash_table_lookup(hub->users, OBJ(user)->name))
    {
        hub->bytes -= user->bytes;
        g_hash_table_remove(hub->users, OBJ(user)->name);
        dc_user_is_offline_real(user, hub);
    }
    else
    {
        g_log("DC_USERS", G_LOG_LEVEL_WARNING,
              "user '%s' is not on hub", OBJ(user)->name);
    }
}

void dc_user_set_bytes(dc_user_t* user, file_size_t bytes)
{
    GList* dlist;
    if (user->bytes == bytes) return;
    for (dlist = user->hubs; dlist; dlist = dlist->next)
    {
        dc_hub_t* hub = dlist->data;
        hub->bytes -= user->bytes;
        hub->bytes += bytes;
    }
    user->bytes = bytes;
}

static void _user_offline(const char* nick, dc_user_t* user, dc_hub_t* hub)
{
    (void)nick;
    dc_user_is_offline_real(user, hub);
}

void dc_user_left_hub(dc_hub_t* hub)
{
    if (hub->users)
    {
        g_hash_table_foreach(hub->users, (GHFunc)_user_offline, hub);
        g_hash_table_destroy(hub->users);
        hub->users = NULL;
    }
}

dc_user_t* dc_user_find(const char* nick)
{
    dc_user_t* user = g_hash_table_lookup(DcUserModule->users, nick);
    g_assert(DcUserModule);
    if (!user) user = dc_user_new(nick);
    return user;
}

static void dc_users_init()
{
    DcUserModule->users =
        g_hash_table_new_full
        ((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
         (GDestroyNotify)NULL, (GDestroyNotify)NULL);
}

static void dc_users_exit()
{
    g_log("DC_USERS", G_LOG_LEVEL_DEBUG,
          "we have %u in cache consuming approx. %u bytes", 
          g_hash_table_size(DcUserModule->users),
          (int)(g_hash_table_size(DcUserModule->users)*sizeof(dc_user_t)));
}

static void dc_user_module_init(dc_user_module_t* module)
{
    module_init(MODULE(module), "Dc User Module",
                (ModuleFunc)dc_users_init, NULL, (ModuleFunc)dc_users_exit);
  
    module->users = NULL;

    //  module_require(MODULE(module), dc_net_module_instance);
    module_require(user_module_instance);
}


module_t* dc_user_module_instance()
{
    if (!DcUserModule)
    {
        DcUserModule = g_malloc0(sizeof(*DcUserModule));
        dc_user_module_init(DcUserModule);
    }
    return MODULE(DcUserModule);
}
