/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.8 $
 * $Date: 2007/09/05 14:34:10 $
 * $Author: sid4u $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "lcore/browse.h"
#include "lcore/ui.h"
#include "lcore/l_i18n.h"
#include "lcore/l_base.h"
#include "lcore/preferences.h"
#ifdef WIN32
#include "lcore/win32_defs.h"
#endif
#include "nap_browse.h"
#include "nap_proto.h"
#include "nap_whois.h"

#ifdef WIN32
#include "napalias.h"
#endif

#define NAP_BROWSE_INIT    0
#define NAP_BROWSE_DIRECT  1
#define NAP_BROWSE_SERVER  2
#define NAP_BROWSE_DONE    3
#define NAP_BROWSE_FAILED  4

static GList* NapBrowses = NULL;
static nap_net_t* LastBrowseRequest = NULL;

struct _nap_browse_t {
  browse_t _browse;
  tcp_socket_t* socket;
  int status;
  int progress;
};

static tcp_socket_t* nap_browse_socket_new(nap_browse_t* browse);

static void nap_browse_destroy(nap_browse_t* browse) {

  NapBrowses = g_list_remove(NapBrowses, browse);
  // now check whether there is a direct browse socket
  if (browse->socket)
    socket_done(SOCK(browse->socket), ST_DELETE, "destroy");

  browse_destroy(BROWSE(browse));
}

static void nap_browse_init(nap_browse_t* browse, nap_user_t* user) {
  browse_init(BROWSE(browse), USER(user));
  object_add_type(OBJ(browse), "NapBrowse");
  OBJ(browse)->destroy = (ObjectFunc)nap_browse_destroy;
  browse->status = NAP_BROWSE_INIT;
  nap_whois_user(user, user->net, NAP_WHOIS_QUIET, FALSE);
  browse->progress = 0;
  NapBrowses = g_list_append(NapBrowses, browse);
}

static nap_browse_t* nap_browse_new(nap_user_t* user) {
  nap_browse_t* browse = g_malloc0(sizeof(*browse));
  nap_browse_init(browse, user);
  return browse;
}

void nap_browse_finished(nap_user_t* user, nap_browse_t* browse) {

  if (!browse) browse = NAP_BROWSE(browse_find(USER(user)));
  if (browse) {
    browse_t* b = BROWSE(browse);
    nap_user_t* user = NAP_USER(b->user);

    if (!browse->socket && user->shared > browse->progress) {
      char* s_header =
        g_markup_printf_escaped(_("Browsing %s not complete"), b->user->print(b->user));
      char* s_message =
        g_markup_printf_escaped(_("Network <b>%s</b> provides %d files only "
                                  "while the user appears to share %u files."),
                                user->net?OBJ(user->net)->name:"???", browse->progress, user->shared);
        
      lopster_notify(LOPSTER_MESSAGE_INFO, s_header, s_message);

      g_free(s_header);
      g_free(s_message);
    }

    browse->status = NAP_BROWSE_DONE;
    browse_set_status(b, NULL);
    browse_set_progress(b, 0, NULL);
  }

}

void nap_browse_user(nap_user_t* user) {
  nap_browse_t* browse = NAP_BROWSE(browse_find(USER(user)));

  if (browse) {
    // FIXME
    g_log("BROWSE", G_LOG_LEVEL_WARNING,
          "browse '%s' already existent", OBJ(user)->name);
    return;
  }
  browse = nap_browse_new(user);
  browse->socket = nap_browse_socket_new(browse);
  browse->status = NAP_BROWSE_DIRECT;
  browse_set_status(BROWSE(browse), _("Waiting for direct browse response"));
  
#if 0  // test
  socket_done(SOCK(browse->socket), S_CLOSE, "testing server browse");
#else
  nap_net_send(user->net, NAP_S_BROWSE_DIRECT, TRUE, "%s", OBJ(user)->name);
#endif
}






static gboolean nap_browse_socket_end(tcp_socket_t* s, const char* message) {
  browse_t* browse = socket_get_data(SOCK(s));
  nap_browse_t* nbrowse = socket_get_data(SOCK(s));
  user_t* user = browse->user;

  switch (socket_get_status(SOCK(s))) {
  case S_BROWSE_COMPLETE:
    nap_browse_finished(NAP_USER(user), nbrowse);
    break;
  case S_NOT_SHARING:
    browse_set_status(browse, NULL);
    browse_set_progress(browse, 0., NULL);
    lopster_notify(LOPSTER_MESSAGE_INFO, 
                   l_printf(_("Browsing %s"), user->print(user)), message);
  default:
    g_log("BROWSE", G_LOG_LEVEL_DEBUG,
          "direct browse failed, starting server browse");
    browse_set_status(browse, _("Getting files from server"));
    browse_set_progress(browse, 0., NULL);
    if (NAP_USER(user)->net)
      nap_net_send(NAP_USER(user)->net, NAP_S_BROWSE, TRUE, "%s", OBJ(user)->name);
    nbrowse->status = NAP_BROWSE_SERVER;
  }
  nbrowse->socket = NULL;
  tcp_socket_end(s, message);
  return TRUE;
}

void nap_browse_insert_file(nap_browse_t* browse, file_t* file) {
  nap_user_t* user = NAP_USER(BROWSE(browse)->user);
  browse->progress++;
  browse_insert_file(BROWSE(browse), file);
  if (user->shared > 0)
    browse_set_progress(BROWSE(browse), (double)browse->progress/(double)user->shared, NULL);
}
static const char* nap_browse_get_files(tcp_socket_t* s) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));
  char *pos1;
  char *pos2;
  char *filename;
  char *md5;
  char *size;
  char *bitrate;
  char *frequency;
  char *duration;
  int res;
  file_t* file;
  nap_browse_t* browse = socket_get_data(SOCK(s));

  pos1 = buffer->data;
  while ((pos2 = strchr(pos1, '\n'))) {
    
    *pos2 = 0;
    
    if (!(filename = arg(pos1, 0))) {
      socket_set_status(SOCK(s), S_BROWSE_COMPLETE);
      return "Complete";
    }
    if (!(md5 = arg(NULL, 0))) return "get_files: no md5";
    if (!(size = arg(NULL, 0))) return "get_files: no size";
    if (!(bitrate = arg(NULL, 0))) return "get_files: no bitrate";
    if (!(frequency = arg(NULL, 0))) return "get_files: no frequency";
    if (!(duration = arg(NULL, 0))) return "get_files: no duration";

    file = file_new();
    file_set_origname(file, g_strdup(filename));
    file_set_locname(file, nap_net_to_utf8(NULL, filename));

    file->size = strtoll(size, NULL, 10);
    // ignore md5
    //    file->md5 = g_strdup(md5);
      
    attr_store_str(ATTRC(file), "file", "Bitrate", bitrate);
    attr_store_str(ATTRC(file), "file", "Frequency", frequency);
    attr_store_str(ATTRC(file), "file", "Duration", duration);
    
    nap_browse_insert_file(browse, file);

    pos1 = pos2 + 1;
  }

  res = pos1 - buffer->data;
  buffer_consume(buffer, res);

  return NULL;
}

static const char* nap_browse_get_nick(tcp_socket_t* s) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));
  browse_t* browse = socket_get_data(SOCK(s));
  char* pos;
  int cnt;

  if (buffer->consumed != 0) return "error: buffer error";

  pos = strchr(buffer->data, '\n');
  if (!pos) return NULL;

  *pos = 0;
  cnt = (pos - buffer->data)+1;

  pos = buffer->data;

  /* Basically there should be no space between SENDLIST and the nick
   * but the opennap/slavanap protocol specs say there is one
   * (which is not correct i think).
   * XNap sends one space before the nick, lopster and other
   * clients dont. So get rid of the XNap space here.
   */
  while (*pos == ' ') pos++;

  // we dont need the nickname since we made the connection
  g_log("BROWSE", G_LOG_LEVEL_DEBUG, "got nick '%s'", buffer->data);
  b_socket_set_consume
    (BSOCK(s), (BConsumeInputFunc)nap_browse_get_files);

  buffer_consume(buffer, cnt);
  browse_set_status(browse, _("Getting file list..."));

  if (buffer_avail(buffer)) return "AGAIN";  // lets continue immediately
  else return NULL;
}

static const char* nap_browse_con_ack(tcp_socket_t* s) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));

  if (!buffer)
    return "error: no buffer";
  if (buffer_avail(buffer) != 1)
    return "error: invalid input";
  if (buffer_read_pos(buffer)[0] != '1')
    return "error: other char that '1'";
  buffer_consume(buffer, 1);

  g_log("BROWSE", G_LOG_LEVEL_DEBUG, "got '1'");
  b_socket_set_consume(BSOCK(s), (BConsumeInputFunc)nap_browse_get_nick);
  
  g_log("BROWSE", G_LOG_LEVEL_DEBUG, "sending 'GETLIST'");
  tcp_socket_send_str(TSOCK(s), "GETLIST");

  return NULL;
}

static tcp_socket_t* nap_browse_socket_new(nap_browse_t* browse) {
  tcp_socket_t* s = g_malloc0(sizeof(*s));
  tcp_socket_init(TSOCK(s), "NapDBrowse", 30);
  object_add_type(OBJ(s), "NapDBrowse");
  socket_set_end(SOCK(s), (SockEndFunc)nap_browse_socket_end);
  socket_set_data(SOCK(s), browse);
  b_socket_set_consume(BSOCK(s), (BConsumeInputFunc)nap_browse_con_ack);

  socket_add_load(SOCK(s), "Peer");

  return s;
}

void nap_browse_outgoing_get(nap_user_t* user, ip_t ip, port_t port) {
  nap_browse_t* browse = NAP_BROWSE(browse_find(USER(user)));

  if (!browse || !browse->socket) return;
  user_set_ip(USER(user), ip);

  if (!tcp_socket_connect(browse->socket, ip, port)) return;

  b_socket_start_receiving(BSOCK(browse->socket), 1024000);
  /* Shaking hands means initializing the established connection */
  browse_set_status(BROWSE(browse), _("Shaking hands..."));
}

static const char* nap_browse_get_nick2(tcp_socket_t* s) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));
  char* pos;
  int cnt;
  nap_browse_t* browse;
  char* data;
  nap_user_t* user;

  data = buffer_read_pos(buffer);
  
  pos = strchr(data, '\n');
  if (!pos) return NULL;

  *pos = 0;
  cnt = (pos - data)+1;

  g_log("BROWSE", G_LOG_LEVEL_DEBUG, "got nick '%s'", data);
  user = nap_user_find(data, NULL);
  browse = NAP_BROWSE(browse_find(USER(user)));

  if (!browse || !browse->socket) {
    return "direct browse not found";
  } else if (browse->status != NAP_BROWSE_DIRECT) {
    return "direct browse not active";
  } else {
    // copy the socket
    buffer_consume(buffer, cnt);

    user_set_ip(BROWSE(browse)->user, s->ip);
    tcp_socket_move(browse->socket, s, TRUE);
    
    socket_set_data(SOCK(browse->socket), browse);
    b_socket_set_consume
      (BSOCK(browse->socket), (BConsumeInputFunc)nap_browse_get_files);
    b_socket_start_receiving(BSOCK(browse->socket), 20480);

    browse_set_status(BROWSE(browse), _("Getting file list..."));
    return "tsocket->nap_browse";
  }

}

const char* nap_browse_incoming_get(tcp_socket_t* s) {
  b_socket_set_consume
    (BSOCK(s),(BConsumeInputFunc)nap_browse_get_nick2);
  // allow handshake with max 1024 chars
  b_socket_set_limit(BSOCK(s), 1024);

  if (buffer_avail(b_socket_get_in_buffer(BSOCK(s))))
    return "AGAIN";
  else
    return NULL;
}



void nap_browse_set_last_net(nap_net_t* net) {
  LastBrowseRequest = net;
}

static file_node_t*
add_to_buffer(file_tree_t* tree, file_node_t* node, buffer_t* buffer) {
  file_t* file;
  int added;
  attribute_t* attr1;
  attribute_t* attr2;
  attribute_t* attr3;

  (void)tree;

  file = node->file;
  g_assert(file);

  attr1 = attr_find(ATTRC(file), "file", "Bitrate");
  attr2 = attr_find(ATTRC(file), "file", "Frequency");
  attr3 = attr_find(ATTRC(file), "file", "Duration");

  while (1) {
    added =
      snprintf(buffer->data + buffer->size, buffer->max - buffer->size,
               "\"%s\" 00000000000000000000000000000000 "FF" %d %d %d\n",
               file->netname, file->size,
               attr1?attr_get_ref(attr1, int):24,
               attr2?attr_get_ref(attr2, int):16000,
               attr2?attr_get_ref(attr3, int):600);

    // check for error, different for glibc < and >= 2.1
    if (added > 0 && added <= buffer_space(buffer)) break;
    // we need more space
    buffer_resize(buffer, buffer->max*2);
  }

  buffer->size += added;
  return NULL;
}

static void nap_browse_pack_library(buffer_t* buffer) {
  // now add the shared file list to the buffer
  library_foreach_shared((FileNodeTraverse)add_to_buffer, buffer);
  // and add a newline to mark it finished
  buffer_add_string(buffer, "\n");
}

static const char* nap_browse_send_nick_list(tcp_socket_t* s) {
  char* nick;
  buffer_t* buffer;

  if (!LastBrowseRequest) {
    g_log("BROWSE", G_LOG_LEVEL_WARNING, 
          "don't know the network, guessing our own nickname");
  }

  nick = nap_net_get_nick(LastBrowseRequest);
  if (!nick) {
    g_log("BROWSE", G_LOG_LEVEL_WARNING, 
          "could not detect the nickname");
    return "couldn't detect our nickname";
  }

  g_log("BROWSE", G_LOG_LEVEL_DEBUG, "sending nick and file list");
  tcp_socket_send_str(s, l_printf("%s\n", nick));

  buffer = b_socket_get_out_buffer(BSOCK(s), 1024);
  nap_browse_pack_library(buffer);
  b_socket_close_when_sent(BSOCK(s), ST_DONE);
  return NULL;
}

const char* nap_browse_incoming_send(tcp_socket_t* s) {
  return nap_browse_send_nick_list(s);
}


static const char* nap_share_con_ack(tcp_socket_t* s) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));

  if (!buffer)
    return "error: no buffer";
  if (buffer_avail(buffer) != 1)
    return "error: invalid input";
  if (buffer_read_pos(buffer)[0] != '1')
    return "error: other char that '1'";
  buffer_consume(buffer, 1);

  g_log("BROWSE", G_LOG_LEVEL_DEBUG, "got '1'");
  g_log("BROWSE", G_LOG_LEVEL_DEBUG, "sending 'SENDLIST'");
  tcp_socket_send_str(s, "SENDLIST");
  
  return nap_browse_send_nick_list(s);
}

void nap_browse_outgoing_send(nap_net_t* net, ip_t ip, port_t port) {
  tcp_socket_t* socket = tcp_socket_new("NapTemp", 30);

  b_socket_set_consume
    (BSOCK(socket), (BConsumeInputFunc)nap_share_con_ack);

  if (!tcp_socket_connect(socket, ip, port)) return;

  LastBrowseRequest = net;
  b_socket_start_receiving(BSOCK(socket), 1024);
}

void nap_browse_error(nap_user_t* user, const char* message) {
  nap_browse_t* browse = NAP_BROWSE(browse_find(USER(user)));

  if (!browse || !browse->socket) return;

  if (strstr(message, "is not sharing any files"))
    socket_done(SOCK(browse->socket), S_NOT_SHARING,
                "not sharing anything");
  else
    socket_done(SOCK(browse->socket), S_BROWSE_FAILED, message);
  
}

#ifdef WIN32
#define _NAP_BROWSE_C_
#include "napaliasdef.c"
#endif
