/* Copyright (C) 2005-2006 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.14 $
 * $Date: 2006/10/29 00:52:57 $
 * $Author: sgop $
 */

#include <string.h>

#include "lcore/preferences.h"
#include "lcore/string_list.h"
#include "lcore/ui.h"
#include "lcore/completion.h"
#include "lcore/l_base.h"
#include "lcore/l_i18n.h"
#include "nap_chat.h"
#include "nap.h"
#include "nap_proto.h"
#include "nap_ui.h"
#ifdef WIN32
#include "napalias.h"
#endif

static GList* NapPages = NULL;

static const char* NapChannelMode(int id) {
  switch (id) {
  case NAP_CMODE_TOPIC_REST: return "TOPIC";
  case NAP_CMODE_REGISTERED: return "REGISTERED";
  case NAP_CMODE_PRIVATE:    return "PRIVATE";
  case NAP_CMODE_INVITE:     return "INVITE";
  case NAP_CMODE_MODERATED:  return "MODERATED";
  default: return NULL;
  }
}
static const char* NapChannelModeShort(int id) {
  switch (id) {
  case NAP_CMODE_TOPIC_REST: return "t";
  case NAP_CMODE_REGISTERED: return "r";
  case NAP_CMODE_PRIVATE:    return "p";
  case NAP_CMODE_INVITE:     return "i";
  case NAP_CMODE_MODERATED:  return "m";
  default: return NULL;
  }
}

int NapChannelModeInt(const char* mode) {
  if (*mode == '+' || *mode == '-') mode++;
  if (!strcmp(mode, "TOPIC"))      return NAP_CMODE_TOPIC_REST;
  if (!strcmp(mode, "REGISTERED")) return NAP_CMODE_REGISTERED;
  if (!strcmp(mode, "PRIVATE"))    return NAP_CMODE_PRIVATE;
  if (!strcmp(mode, "INVITE"))     return NAP_CMODE_INVITE;
  if (!strcmp(mode, "MODERATED"))  return NAP_CMODE_MODERATED;
  return 0;
}

nap_cp_t*
nap_cp_search_net(nap_net_t* net, const char* name, int type) {
  GList* dlist;

  for (dlist = NapPages; dlist; dlist = dlist->next) {
    nap_cp_t* ncp = dlist->data;
    if (ncp->type == type && ncp->net == net &&
        !strcasecmp(OBJ(ncp)->name, name)) return ncp;
  }
  return NULL;
}

nap_cp_t* nap_cp_search(const char* name, int type) {
  GList* dlist;

  for (dlist = NapPages; dlist; dlist = dlist->next) {
    nap_cp_t* ncp = dlist->data;
    if (ncp->type == type && !strcmp(OBJ(ncp)->name, name)) 
      return ncp;
  }
  return NULL;
}

static gboolean nap_cp_closing(nap_cp_t* page) {
  (void)page;
  return TRUE;
}

static void nap_cp_destroy(nap_cp_t* page) {
  chat_page_destroy(PAGE(page));
}
static void nap_cp_init(nap_cp_t* ncp, nap_net_t* net, const char* name,
                        const char* label, const char* title, int type)
{
    chat_page_init(PAGE(ncp), name, label, title, NAP_ID);
    object_add_type(OBJ(ncp), "NapChatPage");
    OBJ(ncp)->destroy = (ObjectFunc)nap_cp_destroy;
    PAGE(ncp)->close = (ChatCloseFunc)nap_cp_closing;
    ncp->net = net;
    ncp->type = type;
    NapPages = g_list_append(NapPages, ncp);
}

static nap_cp_t* nap_cp_new(nap_net_t* net, const char* name, const char* label, 
                            const char* title, int type)
{
    nap_cp_t* ncp = g_malloc0(sizeof(*ncp));
    nap_cp_init(ncp, net, name, label, title, type);
    return ncp;
}

static nap_cp_t* nap_cp_new_other(nap_net_t* net, const char* name,
                                  const char* label, const char* title)
{
    nap_cp_t* ncp;
    
    ncp = nap_cp_new(net, name, label, title, NAP_PAGE_OTHER);
    UI_chat_create_simple_page(PAGE(ncp));
    return ncp;
}
			   
static void on_user_changed(nap_user_t* user, nap_cp_t* ncp) {
  chat_page_set_active(PAGE(ncp), USER(user)->online);
}

nap_cp_t* nap_cp_new_private(nap_net_t* net, const char* nick)
{
    nap_cp_t* ncp = nap_cp_new(net, nick, nick,
                               l_printf(_("%s - Private conversation with %s"), OBJ(net)->name, nick),
                               NAP_PAGE_PRIVATE);
    nap_user_t* user;
    
    UI_chat_create_simple_page(PAGE(ncp));
    
    user = nap_user_find(nick, net);
    user_start_monitor(USER(user));  //make sure the online status is monitored.
    object_monitor_o(OBJ(user), (ObjectFunc)on_user_changed, OBJ(ncp));
    
    return ncp;
}

nap_cp_t* nap_cp_new_wallop(nap_net_t* net) {
  nap_cp_t* ncp = nap_cp_new(net, "__NAP__Wallop__",  _("Operator"),
                             l_printf(_("%s - Operator messages"), OBJ(net)->name),
                             NAP_PAGE_WALLOP);
  UI_chat_create_simple_page(PAGE(ncp));
  return ncp;
}


static gboolean nap_pub_closing(nap_pub_t* page) {
  if (!PAGE(page)->active) return TRUE;

  page->leaving = TRUE;
  nap_net_send(NCP(page)->net, NAP_S_PART_CHANNEL, TRUE,
               "%s", OBJ(page)->name);
  chat_page_set_active(PAGE(page), FALSE);
  return FALSE;
}

static void nap_clear_channel_users(nap_pub_t* npp)
{
    if (!npp) return;
    g_list_foreach(npp->users, (GFunc)object_kill, NULL);
    g_list_free(npp->users);
    npp->users = NULL;
    npp->online = 0;
}

static void nap_pub_destroy(nap_pub_t* page)
{
    nap_clear_channel_users(page);
    g_free(page->topic);
    chat_page_destroy(PAGE(page));
}

static char* nap_user_name(nap_cu_t* cu) {
  return OBJ(cu)->name;
}

static char* nap_pub_expand(nap_pub_t* page, const char* text) {
  const char* pos;
  GList* results;
  char* result;
  int offset;
  completion_t* c;

  c = completion_new((CompletionFunc)nap_user_name);
  completion_set_compare
    (c, (GCompletionStrncmpFunc)strncasecmp);
  completion_set_items(c, &page->users);
  
  if (!text) {
    results = completion_complete(c, "", &result);
    offset = 0;
    pos = NULL;
  } else {
    pos = strrchr(text, ' ');
    if (!pos) pos = text;
    else pos++;

    if (*pos == '@') {
      results = completion_complete(c, pos+1, &result);
      offset = (pos-text)+1;
    } else {
      results = completion_complete(c, pos, &result);
      offset = (pos-text);
    }
  }

  if (g_list_length(results) > 1) {
    GList* dlist;
    chat_print(PAGE(page), "+\0client",
               "+\1c_client", l_printf(_("%d matching users found"), 
                                       g_list_length(results)),
               "+\0nl", NULL);
    for (dlist = results; dlist; dlist = dlist->next) {
      nap_cu_t* cu = dlist->data;
      chat_print(PAGE(page), "+\0client",
                 "+\1c_text", "  ",
                 "+\1nick", OBJ(cu)->name, "\0\0",
                 "+\0nl", NULL);
    }
    if (result) {
      char* t = result;
      result = g_malloc(offset+strlen(t)+1);
      strncpy(result, text, offset);
      sprintf(result+offset, "%s", t);
      g_free(t);
    }
  } else if (g_list_length(results) == 1) {
    nap_cu_t* cu = results->data;
    g_free(result);
    result = g_malloc(offset+strlen(OBJ(cu)->name)+3);
    strncpy(result, text, offset);
    if (offset < 2) 
      sprintf(result+offset, "%s: ", OBJ(cu)->name);
    else
      sprintf(result+offset, "%s ", OBJ(cu)->name);
  } else {
    chat_print(PAGE(page), "+\0client",
               "+\1c_client", _("No matching user found"),
               "+\0nl", NULL);
  }
  completion_free(c);
  return result;
}

static void nap_pub_init(nap_pub_t* npp, nap_net_t* net, const char* channel)
{
    nap_cp_init(NCP(npp), net, channel, channel,
                l_printf(_("%s - channel %s"), OBJ(net)->name, channel),
                NAP_PAGE_PUBLIC);
    object_add_type(OBJ(npp), "NapPublicPage");
    OBJ(npp)->destroy = (ObjectFunc)nap_pub_destroy;
    PAGE(npp)->close = (ChatCloseFunc)nap_pub_closing;
    PAGE(npp)->expand = (ChatExpandFunc)nap_pub_expand;
    PAGE(npp)->active = FALSE;
}

static nap_pub_t* nap_pub_new(nap_net_t* net, const char* channel)
{
    nap_pub_t* npp = g_malloc0(sizeof(*npp));
    nap_pub_init(npp, net, channel);
    return npp;
}

nap_pub_t* nap_cp_new_public(nap_net_t* net, const char* channel)
{
    nap_pub_t* npp = nap_pub_new(net, channel);
    object_make_public(OBJ(npp), NULL);
    return npp;
}

static void nap_cp_update_net(nap_cp_t* ncp, nap_net_t* net)
{
    if (!ncp || ncp->net == net) return;
    ncp->net = net;
    if (ncp->type == NAP_PAGE_PRIVATE)
    {
        chat_page_set_title
            (PAGE(ncp), 
             l_printf(_("%s - Private conversation with %s"), OBJ(net)->name, OBJ(ncp)->name));
    }
    else
    {
        g_log("NAP_CHAT", G_LOG_LEVEL_WARNING,
              "nap_cp_update_net: this should not happen");
        g_assert(FALSE);
    }
}

void nap_join_channel(nap_net_t* net, const char* channel) {
  nap_pub_t* npp;

  npp = NPP(nap_cp_search_net(net, channel, NAP_PAGE_PUBLIC));
  if (npp && PAGE(npp)->active) return;

  if (!npp) npp = nap_cp_new_public(net, channel);
  nap_net_send(net, NAP_S_JOIN_CHANNEL, TRUE, "%s", channel);
}

static void nap_cu_destroy(nap_cu_t* cu) {
  object_destroy(OBJ(cu));
}

static void nap_cu_init(nap_cu_t* cu, const char* nick, int shared, int speed)
{
    object_init(OBJ(cu), nick, TRUE);
    object_add_type(OBJ(cu), "NapChannelUser");
    OBJ(cu)->destroy = (ObjectFunc)nap_cu_destroy;
    cu->shared = shared;
    cu->speed = speed;
}

static nap_cu_t* nap_cu_new(const char* nick, int shared, int speed)
{
    nap_cu_t* cu = g_malloc0(sizeof(*cu));
    nap_cu_init(cu, nick, shared, speed);
    return cu;
}

static nap_cu_t* nap_cu_search(nap_pub_t* npp, const char* nick) {
  GList* dlist;
  
  if (!npp || !nick) return NULL;
  for (dlist = npp->users; dlist; dlist = dlist->next) {
    nap_cu_t* cu = dlist->data;
    if (!strcasecmp(OBJ(cu)->name, nick)) return cu;
  }
  return NULL;
}

nap_cu_t* nap_cu_add(nap_pub_t* npp, const char* nick, int shared, int speed)
{
    nap_cu_t* cu;
    
    cu = nap_cu_search(npp, nick);
    if (!cu)
    {
        cu = nap_cu_new(nick, shared, speed);
        npp->users = g_list_append(npp->users, cu);
        npp->online++;
        object_make_public(OBJ(cu), npp);
    }
    else
    {
        cu->shared = shared;
        cu->speed = speed;
        object_set_dirty(OBJ(cu));
    }
    return cu;
}

void nap_cu_joined(nap_pub_t* npp, const char* nick, int shared, int speed)
{
    nap_cu_t* cu = nap_cu_add(npp, nick, shared, speed);
    NAPUI_user_joined(npp, cu);
}

void nap_cu_parted(nap_pub_t* npp, const char* nick, int shared, int speed) {
  nap_cu_t* cu;

  cu = nap_cu_search(npp, nick);
  if (!cu) return;

  cu->shared = shared;
  cu->speed = speed;

  NAPUI_user_parted(npp, cu);

  npp->users = g_list_remove(npp->users, cu);
  npp->online--;
  object_kill(OBJ(cu));
}


void nap_channel_joined(nap_pub_t* npp)
{
    nap_net_t* net = NCP(npp)->net;
    
    chat_page_set_active(PAGE(npp), TRUE);
    npp->opped = FALSE;
    npp->leaving = FALSE;
    nap_clear_channel_users(npp);
    
    NAPUI_channel_joined(npp);
    
    nap_net_send(net, NAP_S_GET_CHMODE, TRUE, "%s", OBJ(npp)->name);
    nap_net_send(net, NAP_S_GET_CHLIMIT, TRUE, "%s", OBJ(npp)->name);
    nap_net_send(net, NAP_S_GET_CHLEVEL, TRUE, "%s", OBJ(npp)->name);
}

void nap_channel_part(nap_pub_t* page) {
  if (!page->leaving) {
    chat_page_set_active(PAGE(page), FALSE);
    if (pref_get_ref("/core/chat/auto_rejoin", gboolean))
      nap_net_send(NCP(page)->net, NAP_S_JOIN_CHANNEL, TRUE,
                   "%s", OBJ(page)->name);
  } else {
    object_kill(OBJ(page));
    //    chat_page_close(PAGE(page));  // should also work
  }
}

void nap_channel_new_topic(nap_pub_t* npp, const char* topic, const char* who)
{
    if (!npp) return;
    g_free(npp->topic);
    npp->topic = g_strdup(topic);
    NAPUI_new_topic(npp);
    
    if (who)
    {
        if (topic)
        {
            chat_print(PAGE(npp),
                       "+\0server",
                       "+\1nick", who, "\0\0",
                       "+\1c_server", _("changed channel topic:"),
                       "+\1c_text", topic,
                       "+\0nl", NULL);
        }
        else
        {
            chat_print(PAGE(npp),
                       "+\0server",
                       "+\1nick", who, "\0\0",
                       "+\1c_server", _("cleared channel topic"),
                       "+\0nl", NULL);
        }
    }
}

void nap_chat_joined_net(nap_net_t* net) {
  GList* dlist;
  GSList* slist;

  if (!net) return;
  for (dlist = NapPages; dlist; dlist = dlist->next) {
    nap_cp_t* ncp = dlist->data;
    if (ncp->net != net) continue;
    if (ncp->type == NAP_PAGE_PUBLIC) {
      if (pref_get_ref("/core/chat/auto_rejoin", gboolean))
        nap_net_send(ncp->net, NAP_S_JOIN_CHANNEL, TRUE,
                     "%s", OBJ(ncp)->name);
    } if (ncp->type == NAP_PAGE_PRIVATE) {
    } if (ncp->type == NAP_PAGE_OTHER) {
      chat_page_set_active(PAGE(ncp), TRUE);
    }
  }
  //FIXME: make sure we don't join an already open channel again..
  for (slist = net->autojoin; slist; slist = slist->next) {
    char* channel = slist->data;
    nap_join_channel(net, channel);
  }

}

void nap_chat_left_net(nap_net_t* net) {
  GList* dlist;

  if (!net) return;
  for (dlist = NapPages; dlist; dlist = dlist->next) {
    nap_cp_t* ncp = dlist->data;
    if (ncp->net != net) continue;
    if (ncp->type == NAP_PAGE_PUBLIC) {
      chat_page_set_active(PAGE(ncp), FALSE);
    } if (ncp->type == NAP_PAGE_PRIVATE) {
    } if (ncp->type == NAP_PAGE_OTHER) {
      chat_page_set_active(PAGE(ncp), FALSE);
    }
  }
}

///////

void nap_server_message(nap_net_t* net, char* message) {
  chat_page_t* page;

  if (!message) message = "";
  if (net->flags & NAP_NET_OWN_SERVER) {
    page = PAGE(nap_cp_search_net(net, "__NAP__Server__", NAP_PAGE_OTHER));
    if (!page)
      page = PAGE(nap_cp_new_other(net, "__NAP__Server__", _("Server"),
                                   l_printf(_("%s - Server messages"), OBJ(net)->name)));
    normal_message(page, NULL, NULL, message);
  } else {
    server_message(OBJ(net)->name, message);
  } 
}

void nap_motd_message(nap_net_t* net, char* message) {
  motd_message(OBJ(net)->name, message);
}

void nap_emote_message(nap_net_t* net, char* channel, char* user, 
		       char* message) {
  chat_page_t* page;
  page = PAGE(nap_cp_search_net(net, channel, NAP_PAGE_PUBLIC));
  if (!page) return;
  emote_message(page, NULL, NULL, user, message);
}

void nap_public_message(nap_net_t* net, char* channel,
                        char* user, char* message) {
  chat_page_t* page;
  page = PAGE(nap_cp_search_net(net, channel, NAP_PAGE_PUBLIC));
  if (!page) return;
  user_message(page, NULL, NULL, user, message);
}

void nap_chop_message(nap_net_t* net, char* channel,
                      char* user, char* message) {
  chat_page_t* page;

  page = PAGE(nap_cp_search_net(net, channel, NAP_PAGE_PUBLIC));
  if (!page) return;

  if ((pref_get_ref("/core/chat/pipe", int) & PIPE_CHAN_OP)) {
    chat_set_buffer(NAP_B_OP);
    user_message(page, NULL, NULL, user, message);
    chat_set_buffer(0);
  } else {
    user_message(page, NULL, _("Op"), user, message);
  }
}

static void
nap_send_quiet_message(nap_net_t* net, const char* user, char* message) {
  char* pos;

  if (!net) return;

  line_insert_lf(message, LFP);
  do {
    pos = strchr(message, '\n');
    if (pos) *pos = 0;
    nap_net_send(net, NAP_S_PRIVATE_MSG, TRUE, "%s %s", user, message);
    if (!pos) break;
    message = pos + 1;
  } while (*message);
}

void nap_private_message(nap_net_t* net, char* user, char* message) {
  chat_page_t *page;
  static time_t last_afk = 0;

  if (message && !strcmp(message, "//WantQueue")) return;

  if ((page = PAGE(nap_cp_search(user, NAP_PAGE_PRIVATE)))) {
    nap_cp_update_net(NCP(page), net);
    user_message(page, NULL, NULL, user, message);
  } else if ((pref_get_ref("/core/chat/pipe", int) & PIPE_PRIVATE)) {
    // create new priv page if we want to pipe
    page = PAGE(nap_cp_new_private(net, user));
    user_message(page, NULL, NULL, user, message);
  } else {
    private_message(NULL, OBJ(net)->name, NULL, user, message);
  }

  if (pref_get_ref("/core/user/away", gboolean) && CurrentTime > last_afk + 10) {
    const char* message = 
      pref_get_ptr("/core/user/away_message", char);
    if (message && *message)
      nap_send_quiet_message(net, user, l_printf(_("AFK: %s"), message));
    else
      nap_send_quiet_message(net, user, "AFK: Please leave a message");
    last_afk = CurrentTime;
  }
}

void nap_global_message(nap_net_t* net, char* user, char* message) {
  if (net->flags & NAP_NET_IGN_GLOBAL) return;
  global_message(OBJ(net)->name, user, message);
}

void nap_wallop_message(nap_net_t* net, char* user, char* message) {
  chat_page_t* page;

  if (!message) message = "";

  if (pref_get_ref("/core/chat/pipe", int) & PIPE_OPERATOR) {
    page = PAGE(nap_cp_search_net(net, "__NAP__Wallop__", NAP_PAGE_WALLOP));
    if (!page) page = PAGE(nap_cp_new_wallop(net));
    user_message(page, NULL, NULL, user, message);
  } else {
    user_message(NULL, OBJ(net)->name, _("operator"), user, message);
  }
}

void nap_send_private_message(nap_net_t* net, chat_page_t* page,
                              const char* user, char* message) {
  char* pos;
  int private;

  if (!net) return;

  if (page && page->ID == NAP_ID && 
      NCP(page)->type == NAP_PAGE_PRIVATE &&
      !strcasecmp(OBJ(page)->name, user)) {
    private = 1;
  } else {
    private = 0;
  }

  line_insert_lf(message, LFP);
  do {
    pos = strchr(message, '\n');
    if (pos) *pos = 0;
    nap_net_send(net, NAP_S_PRIVATE_MSG, TRUE, "%s %s", user, message);
    if (private) user_message(page, NULL, NULL, net->cur_nick, message);
    else private_message(page, _("to"), user, net->cur_nick, message);
    if (!pos) break;
    message = pos + 1;
  } while (*message);

}

void nap_send_public_message(nap_cp_t* ncp, char* message) {
  char* pos;
  
  if (!ncp || !ncp->net) return;
  if (!PAGE(ncp)->active) return;

  line_insert_lf(message, LFP);
  do {
    pos = strchr(message, '\n');
    if (pos) *pos = 0;
    nap_net_send(ncp->net, NAP_S_PUBLIC_MSG, TRUE,
                 "%s %s", OBJ(ncp)->name, message);
    if (!pos) break;
    message = pos + 1;
  } while (*message);
}

void nap_send_chop_message(nap_cp_t* ncp, char* message) {
  char* pos;
  
  if (!ncp || !ncp->net) return;
  if (!PAGE(ncp)->active) return;
  if (ncp->type != NAP_PAGE_PUBLIC) return;

  line_insert_lf(message, LFP);
  do {
    pos = strchr(message, '\n');
    if (pos) *pos = 0;
    nap_net_send(ncp->net, NAP_S_CHOP_MSG, TRUE,
                 "%s %s", OBJ(ncp)->name, message);
    if (!pos) break;
    message = pos + 1;
  } while (*message);
}

void nap_send_emote_message(nap_cp_t* ncp, char* message) {
  char* pos;

  if (!ncp || !ncp->net) return;
  if (!PAGE(ncp)->active) return;

  line_insert_lf(message, LFP);
  do {
    pos = strchr(message, '\n');
    if (pos) *pos = 0;
    nap_net_send(ncp->net, NAP_S_EMOTE_MSG, TRUE,
                 "%s \"%s\"", OBJ(ncp)->name, message);
    if (!pos) break;
    message = pos + 1;
  } while (*message);
}

void nap_send_wallop_message(nap_net_t* net, char* message) {
  char* pos;

  if (!net) return;

  line_insert_lf(message, LFP);
  do {
    pos = strchr(message, '\n');
    if (pos) *pos = 0;
    nap_net_send(net, NAP_S_WALLOP_MSG, TRUE, "%s", message);
    if (!pos) break;
    message = pos + 1;
  } while (*message);
}

void nap_send_global_message(nap_net_t* net, char* message) {
  char* pos;

  if (!net) return;

  line_insert_lf(message, LFP);
  do {
    pos = strchr(message, '\n');
    if (pos) *pos = 0;
    nap_net_send(net, NAP_S_GLOBAL_MSG, TRUE, "%s", message);
    if (!pos) break;
    message = pos + 1;
  } while (*message);
}

void nap_channel_update_op(nap_net_t* net) {
  GList* dlist;
  nap_cp_t* ncp;

  for (dlist = NapPages; dlist; dlist = dlist->next) {
    ncp = dlist->data;
    if (ncp->type == NAP_PAGE_PUBLIC && ncp->net == net) 
      NAPUI_cp_update_op(NPP(ncp));
  }
}

static char* nap_int2chmode(int set, int clear) {
  static char mode[2048];
  int cnt;

  set   &= NAP_CMODE_MASK;
  clear &= NAP_CMODE_MASK;

  // switch topic flag
  if (set & NAP_CMODE_TOPIC_REST) {
    clear |= NAP_CMODE_TOPIC_REST;
    set &= ~NAP_CMODE_TOPIC_REST;
  } else if (clear & NAP_CMODE_TOPIC_REST) {
    set |= NAP_CMODE_TOPIC_REST;
    clear &= ~NAP_CMODE_TOPIC_REST;
  }
  *mode = 0;

  cnt = 0;
  while (set) {
    if (set & 1) {
      if (*mode) strcat(mode, " +");
      else strcat(mode, "+");
      strcat(mode, NapChannelMode(1<<cnt));
    }
    cnt++;
    set >>= 1;
  }
  cnt = 0;
  while (clear) {
    if (clear & 1) {
      if (*mode) strcat(mode, " -");
      else strcat(mode, "-");
      strcat(mode, NapChannelMode(1<<cnt));
    }
    cnt++;
    clear >>= 1;
  }
  if (*mode) return mode;
  else return NULL;
}

void nap_channel_set_mode(nap_pub_t* page, int set, int clear) {

  if (!NCP(page)->net) return;

  if ((page->chmode & set) == set &&
      (page->chmode & clear) == 0) return;
  set = (~page->chmode) & set;
  clear = page->chmode & clear;

  nap_net_send(NCP(page)->net, NAP_S_SET_CHMODE, TRUE, "%s %s",
               OBJ(page)->name, nap_int2chmode(set, clear));
}

void nap_channel_set_level(nap_pub_t* page, int level) {
  if (!NapIsLevel(level)) return;
  if (!NCP(page)->net) return;

  if (page->level == level) return;
  nap_net_send(NCP(page)->net, NAP_S_SET_CHLEVEL, TRUE, "%s %s",
               OBJ(page)->name, NapLevelShort(level));
}

void nap_channel_set_limit(nap_pub_t* page, int limit) {
  if (!NCP(page)->net) return;

  if (page->limit == limit) return;
  if (limit < 0) return;
  nap_net_send(NCP(page)->net, NAP_S_SET_CHLIMIT, TRUE,
               "%s %d", OBJ(page)->name, limit);
}


void nap_channel_mode_changed(nap_pub_t* page, const char* nick,
			      int set, int clear) {
  page->chmode |= set;
  page->chmode &= ~clear;

  NAPUI_cp_update_op(page);
  
  if (nick) {
    int i1;
    chat_print(PAGE(page),
               "+\0server",
               "+\1nick", nick, "\0\0",
               "+\3c_server", _("changed channel mode:"),
               NULL);
    for (i1 = 0; i1 < NAP_CMODE_NUMBER; i1++) {
      if (set & (1<<i1))
        chat_print(PAGE(page),
                   "+\0c_text", "+",
                   "+\0c_text", NapChannelModeShort(1<<i1),
                   NULL);
      if (clear & (1<<i1))
        chat_print(PAGE(page),
                   "+\0c_text", "-",
                   "+\0c_text", NapChannelModeShort(1<<i1),
                   NULL);
    }
    chat_print(PAGE(page),
               "+\0nl", NULL);
  }
}

void nap_channel_limit_changed(nap_pub_t* page, const char* nick,
			       int limit) {
  if (page->limit == limit) return;
  page->limit = limit;

  NAPUI_cp_update_op(page);

  if (nick) {
    chat_print(PAGE(page),
               "+\0server",
               "+\1nick", nick, "\0\0",
               "+\1c_server", _("changed channel limit:"),
               "+\1c_text", l_printf("%d", limit),
               "+\0nl", NULL);
  }
}

void nap_channel_level_changed(nap_pub_t* page, const char* nick, int level) {
  if (page->level == level) return;
  page->level = level;

  NAPUI_cp_update_op(page);

  if (nick) {
    chat_print(PAGE(page),
               "+\0server",
               "+\1nick", nick, "\0\0",
               "+\1c_server", _("changed channel level:"),
               "+\1c_text", NapLevel(level),
               "+\0nl", NULL);
  }
}

void nap_channel_user_flagged(nap_pub_t* page, const char* who,
                              const char* user, gboolean set,
                              int flag, const char* reason) {
  nap_cu_t* cu;
  const char* nick;

  if (!flag || !page || !user) return;
  if (!NCP(page)->net) return;

  if (!strcmp(user, "you")) nick = nap_net_get_nick(NCP(page)->net);
  else nick = user;
  
  cu = nap_cu_search(page, nick);
  if (cu) {
    if (set && (cu->flags & flag)) return;
    if (!set && !(cu->flags & flag)) return;

    if (set) cu->flags |= flag;
    else cu->flags &= ~flag;
    object_set_dirty(OBJ(cu));
  }

  chat_print(PAGE(page), "+\0server",
             "+\1nick", nick, "\0\0",
             NULL);


  if (flag == NAP_CU_OP) {
    if (set) chat_print(PAGE(page), "+\1c_server", _("was opped"), NULL);
    else chat_print(PAGE(page), "+\1c_server", _("was deopped"), NULL);
  } else if (flag == NAP_CU_VOICE) {
    if (set) chat_print(PAGE(page), "+\1c_server", _("was voiced"), NULL);
    else chat_print(PAGE(page), "+\1c_server", _("was devoiced"), NULL);
  } else if (flag == NAP_CU_MUZZLE) {
    if (set) chat_print(PAGE(page), "+\1c_server", _("was muzzled"), NULL);
    else chat_print(PAGE(page), "+\1c_server", _("was unmuzzled"), NULL);
  } else if (flag == NAP_CU_BAN) {
    if (set) chat_print(PAGE(page), "+\1c_server", _("was banned"), NULL);
    else chat_print(PAGE(page), "+\1c_server", _("was unbanned"), NULL);
  } else {
    chat_print(PAGE(page), "+\1c_server", "was --fatal:unknow action--", NULL);
    g_assert(FALSE);
  }

  chat_print(PAGE(page), "+\1c_text", "(", NULL);
  if (who)
    chat_print(PAGE(page), "+\0nick", who, "\0\0", NULL);
  else
    chat_print(PAGE(page), "+\0c_text", OBJ(NCP(page)->net)->name, NULL);

  if (reason) {
    chat_print(PAGE(page),
               "+\0c_text", ":",
               "+\0c_text", reason,
               NULL);
  }
  chat_print(PAGE(page), "+\0c_text", ")", "+\0nl", NULL);

  // finally check whether it was us that got opped
  if (!strcasecmp(nap_net_get_nick(NCP(page)->net), nick) &&
      flag == NAP_CU_OP) {
    page->opped = set;
    NAPUI_cp_update_op(page);
  }

}

#ifdef WIN32
#define _NAP_CHAT_C_
#include "napaliasdef.c"
#endif

