/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.7 $
 * $Date: 2006/07/22 17:14:03 $
 * $Author: sid4u $
 */

#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "lcore/preferences.h"
#include "nap_download.h"
#include "nap.h"
#include "nap_proto.h"
#include "nap_whois.h"

typedef struct {
  module_t _module;
  GList* sources;
} nap_download_module_t;

static nap_download_module_t* NapDownloadModule = NULL;


static void nap_source_update_info(nap_source_t* source) {
  nap_user_t* user = NAP_USER(SOURCE(source)->user);
  if (!user->net) return;
  nap_whois_user(user, user->net, NAP_WHOIS_QUIET, FALSE);
}

static gboolean nap_source_start(nap_source_t* source) {
  nap_user_t* user = NAP_USER(SOURCE(source)->user);
  if (!user->net) return FALSE;

  nap_source_update_info(source);

  nap_net_send(user->net, NAP_S_DOWNLOAD_REQUEST, FALSE, "%s \"%s\"",
               auto_text(nap_net_to_utf8(user->net, OBJ(user)->name)),
               SOURCE(source)->file->netname);
  
  return TRUE;
}

static void nap_source_destroy(nap_source_t* source) {
  NapDownloadModule->sources = g_list_remove(NapDownloadModule->sources, source);
  source_destroy(SOURCE(source));
}

static void nap_source_send_start(nap_source_t* source) {
  segment_t* segment;
  char* str;
  gint32 len;
  nap_user_t* user = NAP_USER(SOURCE(source)->user);

  if (user->client_id != NAP_CLIENT_LOPSTER_ADV) {
    source->advanced = FALSE;
    return;
  }

  if (!socket_connected(SOCK(source))) return;
  source->advanced = TRUE;
  
  segment = SOURCE(source)->cur_segment;
  str = l_printf("....START "FF, segment->start+segment->size);
  len = strlen(str+4);
  pack_u32(str, len);

  g_log("NAP_DOWN", G_LOG_LEVEL_DEBUG, "sending [%s]", str+4);
  tcp_socket_send_bin(TSOCK(source), str, len+4);
}

static void nap_source_send_stop(nap_source_t* source) {
  char* str;
  gint32 len;
  nap_user_t* user = NAP_USER(SOURCE(source)->user);

  if (user->client_id != NAP_CLIENT_LOPSTER_ADV) {
    source->advanced = FALSE;
    return;
  }

  if (!socket_connected(SOCK(source))) return;
  source->advanced = TRUE;

  str = l_printf("....STOP "FF, SOURCE(source)->real_stop);
  len = strlen(str+4);
  pack_u32(str, len);

  g_log("NAP_DOWN", G_LOG_LEVEL_DEBUG, "sending [%s]", str+4);
  tcp_socket_send_bin(TSOCK(source), str, len+4);
}

static void nap_source_send_finished(nap_source_t* source) {
  char* str;
  gint32 len;

  str = l_printf("....FINISHED");
  len = strlen(str+4);
  pack_u32(str, len);
  
  g_log("NAP_DOWN", G_LOG_LEVEL_DEBUG, "sending [%s]", str+4);
  tcp_socket_send_bin(TSOCK(source), str, len+4);
}

static const char* nap_source_start_downloading(nap_source_t* source) {
  if (!source->started) {
    nap_net_send(NULL, NAP_S_DOWNLOAD_STARTED, FALSE, "");
    source->started = TRUE;
  }
  nap_source_send_stop(source);
  return source_start_downloading(SOURCE(source));
}

static const char* nap_source_get_continued(nap_source_t* source) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(source));
  unsigned avail = buffer_avail(buffer);
  const char* res;

  if (avail > 9) avail = 9;
  if (strncmp(buffer_read_pos(buffer), "CONTINUED", avail)) 
    return "'CONTINUED' expected";
  
  if (avail < 9) return NULL;  // need more data.

  buffer_consume_virt(buffer, 9);

  res = nap_source_start_downloading(source);
  if (res) return res;
  else if (buffer_avail(buffer) > 0) return "AGAIN";
  else return NULL;
}

static gboolean nap_source_end(nap_source_t* source, const char* message) {
  unsigned status = socket_get_status(SOCK(source));
  gboolean delete;
  gboolean ended;

  if (status == ST_DL_SEG_FIN) {
    g_assert(source->started);

    // reset status!!
    socket_set_status(SOCK(source), ST_UNKNOWN);
    if (source->advanced) {
      if (!source_attach_segment(SOURCE(source))) {
        nap_source_send_finished(source);
        b_socket_close_when_sent(BSOCK(source), ST_DONE);
      } else {
        nap_source_send_start(source);
        b_socket_set_consume
          (BSOCK(source), (BConsumeInputFunc)nap_source_get_continued);
      }
      ended = FALSE;
    } else {
      source_end(SOURCE(source), message);
      source_set_remote(SOURCE(source), 5, 0);  // restart in 5 secs
      ended = TRUE;
    }
    delete = FALSE;
  } else {
    delete = source_end(SOURCE(source), message);
    ended = TRUE;
  }

  if (source->started && ended) {
    nap_net_send(NULL, NAP_S_DOWNLOAD_STOPPED, FALSE, "");
    source->started = FALSE;
  }
  return delete;
}

static void
nap_source_init(nap_source_t* source, file_t* file, nap_user_t* user) {
  source_init(SOURCE(source), file, USER(user), TRUE);
  object_add_type(OBJ(source), "NapSource");
  OBJ(source)->destroy = (ObjectFunc)nap_source_destroy;
  socket_set_end(SOCK(source), (SockEndFunc)nap_source_end);
  SOURCE(source)->start = (SourceStartFunc)nap_source_start;
  SOURCE(source)->seg_changed = (SourceSegChangedFunc)nap_source_send_stop;
  source->started = FALSE;

  nap_whois_user(user, user->net, NAP_WHOIS_QUIET, FALSE);

  NapDownloadModule->sources = g_list_append(NapDownloadModule->sources, source);
}

static nap_source_t* nap_source_new(file_t* file, nap_user_t* user) {
  nap_source_t* source = g_malloc0(sizeof(*source));
  nap_source_init(source, file, user);
  return source;
}

static const char* nap_source_get_info_out(nap_source_t* source) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(source));
  char* pos;
  file_size_t lsize;
  source_t* src = SOURCE(source);
  int cnt;
  const char* res;

  if (buffer_avail(buffer) < 1) return "invalid input";
  pos = buffer_read_pos(buffer);

  cnt = 0;
  lsize = 0;
  while (isdigit(*pos) && cnt < 15) {
    lsize *= 10;
    lsize += (*pos)-'0';
    cnt++; pos++;
    // break here, ignored additional digits that are file content
    if (src->file->size == lsize) break;
  }

  buffer_consume_virt(buffer, cnt);

  if (src->file->size != lsize) {
    g_log("NAP_DOWN", G_LOG_LEVEL_WARNING,
          "reported size ("FF") != file size ("FF") -> abort [%s]",
          lsize, src->file->size, buffer_read_pos(buffer));
    return "size mismatch";
  }
  
  nap_user_stat_cnt(NAP_USER(src->user), NS_DOWN);
  res = nap_source_start_downloading(source);
  if (res) return res;
  else if (buffer_avail(buffer)) return "AGAIN";
  else return NULL;
}

static const char* nap_source_accept(nap_source_t* source) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(source));
  segment_t* segment;
  char* temp;
  nap_user_t* user = NAP_USER(SOURCE(source)->user);

  if (buffer_avail(buffer) < 1) return "invalid input";
  if (*(buffer_read_pos(buffer)) != '1') return "input != '1'";
  buffer_consume_virt(buffer, 1);

  g_log("NAP_DOWN", G_LOG_LEVEL_DEBUG, "got incoming '1'");

  segment = source_attach_segment(SOURCE(source));
  if (!segment) return "no segment";
  
  temp = l_printf("GET%s \"%s\" "FF, nap_net_get_nick(user->net),
                  SOURCE(source)->file->netname, segment->start+segment->size);

  g_log("NAP_DOWN", G_LOG_LEVEL_DEBUG, "sending [%s]", temp);
  tcp_socket_send_str(TSOCK(source), temp);

  b_socket_set_consume
    (BSOCK(source), (BConsumeInputFunc)nap_source_get_info_out);

  if (buffer_avail(buffer)) return "AGAIN";
  else return NULL;
}

static void nap_source_connect_and_start(source_t* source) {
  b_socket_set_consume
    (BSOCK(source), (BConsumeInputFunc)nap_source_accept);
  if (socket_activate(SOCK(source)))
    b_socket_start_receiving(BSOCK(source), 102400);
}

void nap_source_ack(source_t* source, nap_net_t* net, ip_t ip, port_t port) {
  nap_user_t* user = NAP_USER(source->user);

  NAP_USER(source->user)->net = net;
  user_set_ip_port(user, ip, port);

  TSOCK(source)->ip = ip;
  TSOCK(source)->port = port;

  if (port == 0) {
    if (nap_get_listen_port() == 0) {
      socket_done(SOCK(source), ST_ERROR, "Both firewalled");
    } else {
      nap_net_send(user->net, NAP_S_DOWNLOAD_FW, FALSE, "%s \"%s\"",
                   auto_text(nap_net_to_utf8(user->net, OBJ(user)->name)),
                   source->file->netname);
    }
  } else {
    nap_source_connect_and_start(source);
  }
}

static const char* nap_source_get_info_in(tcp_socket_t* s) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));
  char* nick;
  char* filename;
  char* size;
  char* temp;
  source_t* source;
  segment_t* segment;
  const char* res;
  nap_user_t* user;

  if (buffer_avail(buffer) == 0) return NULL;

  // this function (like many other napster consumer) assumes,
  // that the buffer is 0 terminated. The tcp_socket appends
  // a 0 whenever it reads from the socket :)
  nick = arg(buffer_read_pos(buffer), 0); //FIXME: this modifies the data..
  filename = arg(NULL, 0);
  size = arg(NULL, 0);
  
  if (!size) return NULL;
  
  // there should not be any additional data in the buffer,
  // so simply clear it.
  buffer_clear(buffer);

  g_log("NAP_DOWN", G_LOG_LEVEL_DEBUG,
        "incoming %s '%s'", nick, filename);
  user = nap_user_find(nick, NULL);
  source = source_find(USER(user), filename);
  if (!source) return "download not found";

  if (socket_connected(SOCK(source))) 
    return "download already connected";

  if (source->file->size != strtoll(size, NULL, 10)) {
    return l_printf("size mismatch "FF" - "FF,
		    source->file->size, strtoll(size, NULL, 10));
  }
  
  segment = source_attach_segment(source);
  if (!segment) return "no segment";
  
  tcp_socket_move(TSOCK(source), s, TRUE);
  
  temp = l_printf(FF, segment->start + segment->size);
  tcp_socket_send_str(TSOCK(source), temp);
  
  g_log("NAP_DOWN", G_LOG_LEVEL_DEBUG, "sending [%s]", temp);
  
  nap_user_stat_cnt(user, NS_DOWN);
  res = nap_source_start_downloading(NAP_SOURCE(source));
  if (res) socket_done(SOCK(source), ST_ERROR, res);

  return "tsock->nap_source";
}

const char* nap_download_incoming(tcp_socket_t* s) {

  b_socket_set_consume(BSOCK(s), (BConsumeInputFunc)nap_source_get_info_in);
  // allow handshake with max 1024 chars
  b_socket_set_limit(BSOCK(s), 1024);

  if (buffer_avail(b_socket_get_in_buffer(BSOCK(s))))
    return "AGAIN";
  else
    return NULL;
}

static nap_source_t*
nap_file_to_source(const download_t* download, file_t* file, nap_user_t* user) {
  nap_source_t* source = nap_source_new(file, user);
  (void)download;
  return source;
}


static void nap_downloads_init() {
  source_factory_register(NAP_ID, (FileToSourceFunc)nap_file_to_source);
}

static void nap_downloads_run() {
}

static void nap_downloads_exit() {
}

static void nap_download_module_init(nap_download_module_t* module) {
  module_init(MODULE(module), "Nap Download Module",
              (ModuleFunc)nap_downloads_init,
              (ModuleFunc)nap_downloads_run,
              (ModuleFunc)nap_downloads_exit);

  module_require(pref_module_instance);
  module_require(download_module_instance);

}

module_t* nap_download_module_instance() {
  if (!NapDownloadModule) {
    NapDownloadModule = g_malloc0(sizeof(*NapDownloadModule));
    nap_download_module_init(NapDownloadModule);
  }
  return MODULE(NapDownloadModule);
}
