/* Copyright (C) 2005-2007 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.19 $
 * $Date: 2008/12/23 13:38:14 $
 * $Author: sgop $
 */

#include <string.h>
#include <sys/types.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#endif

#include <libxml/parser.h>

#include "lcore/version.h"
#include "lcore/l_base.h"
#include "lcore/xml_handler.h"
#include "lcore/b_socket.h"
#include "lcore/l_listen.h"
#include "lcore/file_tree.h"
#include "lcore/library.h"
#include "lcore/preferences.h"
#include "lcore/l_resolve.h"
#include "lcore/ui.h"
#include "lcore/l_i18n.h"
#include "lcore/download.h"
#include "nap_network.h"
#include "nap_proto.h"
#include "nap_hotlist.h"
#include "nap.h"
#include "nap_ui.h"
#include "nap_browse.h"
#include "nap_chat.h"
#include "nap_upload.h"
#include "nap_download.h"

#ifdef WIN32
#include "napalias.h"
#endif

typedef struct {
    module_t _module;
    GList* groups;
    GList* online_nets;
    int online_cnt;
    GList* resolver;
    listen_t* listen;
    gboolean save;
    GList* services;
    gboolean auto_connect;
    GList* nets_auto;
} nap_net_module_t;


static nap_net_module_t* NapNetModule = NULL;
static char* CharSet = NULL;


static void nap_ng_destroy(nap_ng_t* ng);
static void nap_net_destroy(nap_net_t* net);
static void nap_server_destroy(nap_server_t* server);

static void nap_net_share_file(nap_net_t* net, file_t* file);


static void nap_charset_observer(pref_node_t* node)
{
    const char* charset = (char*)(node->value);
    g_free(CharSet);
    if (charset && *charset)
        CharSet = g_strdup(charset);
    else
        CharSet = NULL;
}

char* nap_net_to_utf8(nap_net_t* net, const char* data)
{
    return to_utf8(data, -1, NULL, NULL, 4, "utf-8", net?net->charset:NULL, CharSet, "latin1");
}

void nap_network_save()
{
    if (module_running(MODULE(NapNetModule)))
        lopster_xml_save("nap:serverlist", 10000);
}


int NapSpeedFromType(int no)
{
    switch (no) {
    case LINE_MODEM:  return 4;
    case LINE_ISDN:   return 5;
    case LINE_CABLE:
    case LINE_POWER:  return 7;
    case LINE_SATELLITE:
    case LINE_DSL:    return 8;
    case LINE_T1:     return 9;
    case LINE_T3:     return 10;
    default:          return 0;
    }
}

const char* NapSpeedStr(int no)
{
    switch (no) {
    case 0: return N_("unknown");
    case 1: return N_("14.4 kbps");
    case 2: return N_("28.8 kbps");
    case 3: return N_("33.6 kbps");
    case 4: return N_("56 kbps");
    case 5: return N_("64K ISDN");
    case 6: return N_("128K ISDN");
    case 7: return N_("Cable");
    case 8: return N_("DSL");
    case 9: return N_("T1");
    case 10:return N_("T3+");
    default:return "?speed?";
    }
}

const char* NapSpeedStrShort(int no)
{
    switch (no) {
    case 0: return N_("?");
    case 1: return N_("14.4");
    case 2: return N_("28.8");
    case 3: return N_("33.6");
    case 4: return N_("56K");
    case 5: return N_("64K");
    case 6: return N_("128K");
    case 7: return N_("Cable");
    case 8: return N_("DSL");
    case 9: return N_("T1");
    case 10:return N_("T3+");
    default:return "?speed?";
    }
}

int NapIsLevel(int no)
{
    return (no >= NAP_L_LEECH && no <= NAP_L_ELITE);
}

int NapLevelInt(const char* level)
{
    if (!level || strlen(level) < 3) return NAP_L_USER;
    if (!strncmp(level, "Lee", 3)) return NAP_L_LEECH;
    if (!strncmp(level, "Use", 3)) return NAP_L_USER;
    if (!strncmp(level, "Mod", 3)) return NAP_L_MOD;
    if (!strncmp(level, "Adm", 3)) return NAP_L_ADMIN;
    if (!strncmp(level, "Eli", 3)) return NAP_L_ELITE;
    return NAP_L_UNKNOWN;
}

const char* NapLevel(int no)
{
    switch (no) {
    case NAP_L_LEECH: return "Leech";
    case NAP_L_USER:  return "User";
    case NAP_L_MOD:   return "Moderator";
    case NAP_L_ADMIN: return "Administrator";
    case NAP_L_ELITE: return "Elite";
    default:return "?level?";
    }
}

const char* NapLevelShort(int no)
{
    switch (no) {
    case NAP_L_LEECH: return "Leech";
    case NAP_L_USER:  return "User";
    case NAP_L_MOD:   return "Mod";
    case NAP_L_ADMIN: return "Admin";
    case NAP_L_ELITE: return "Elite";
    default:return "?";
    }
}

const char* NapUserModes(int no)
{
    switch (1<<no) {
    case NAP_MODE_ERROR:
        return _("Show server error messages");
    case NAP_MODE_BAN:
        return _("Show bans messages");
    case NAP_MODE_CHANGE:
        return _("Show configuration changed");
    case NAP_MODE_KILL:
        return _("Show kill messages");
    case NAP_MODE_LEVEL:
        return _("Show level changes");
    case NAP_MODE_SERVER:
        return _("Show server messages");
    case NAP_MODE_MUZZLE:
        return _("Show muzzle messages");
    case NAP_MODE_PORT:
        return _("Show dataport changes");
    case NAP_MODE_WALLOP:
        return _("Show wallops");
    case NAP_MODE_CLOAK:
        return _("Show cloaks");
    case NAP_MODE_FLOOD:
        return _("Show floods");
    case NAP_MODE_PING:
        return _("Show pings");
    case NAP_MODE_MSG:
        return _("Allow private message");
    case NAP_MODE_WHOIS:
        return _("Show whois requests");
    case NAP_MODE_ANNOUNCE:
        return _("Show announcements");
    case NAP_MODE_BROWSE:
        return _("Show browse notifications");
    case NAP_MODE_MOTD:
        return _("Show Motd");
    case NAP_MODE_FRIEND:
        return _("Show messages about friend list");
    case NAP_MODE_CHANNEL:
        return _("Show channel operator actions");
    case NAP_MODE_REGISTER:
        return _("Show new user registrations");
    case NAP_MODE_VAR:
        return _("Show server config changes");
    case NAP_MODE_ABUSE:
        return _("Show client abuses");
    default:
        return "?? unknown user mode ??";
    }
}

const char* NapUserMode(int mode, int mask)
{
    int i1;
    char* result = NULL;
    static int pos[NAP_MODE_NUMBER] = 
        {00, 07, 12, 20, 26, 33, 41, 49, 55, 63, 70, 77, 83, 88};
    
    result = l_strdup("-ERROR -BAN -CHANGE -KILL "
                      "-LEVEL -SERVER -MUZZLE -PORT "
                      "-WALLOP -CLOAK -FLOOD -PING "
                      "-MSG -WHOIS");
    
    for (i1 = 0; i1 < NAP_MODE_NUMBER; i1++)
    {
        if ((mask & (1<<i1)) && (mode & (1<<i1)))
            result[pos[i1]] = '+';
    }
    return result;
}

static void nap_update_connect_status()
{
    int status = ONLINE_STATUS_IDLE;
    if (NapNetModule->online_nets) status |= ONLINE_STATUS_ONLINE;
    plugin_set_online_status(NAP_ID, status);
}

void nap_update_stats()
{
    GList* dlist;
    int users=0, files=0, gigs=0;
    
    g_assert(NapNetModule);
    for (dlist = NapNetModule->online_nets; dlist; dlist = dlist->next)
    {
        nap_net_t* net = dlist->data;
        nap_server_t* server = net->active;
        if (!server) continue;
        users += server->users;
        files += server->files;
        gigs += server->gigs;
    }
    plugin_set_online_stats(NAP_ID, users, files, gigs);
}

static url_t* nap_service_find(url_t* url)
{
    GList* dlist;
    
    for (dlist = NapNetModule->services; dlist; dlist = dlist->next)
    {
        url_t* serv = dlist->data;
        if (url_equal(serv, url)) return serv;
    }
    return NULL;
}

static void on_service_destroy(url_t* url)
{
    NapNetModule->services = g_list_remove(NapNetModule->services, url);
}

static url_t* nap_service_new(url_t* url)
{
    object_add_type(OBJ(url), "NapServerService");
    NapNetModule->services = g_list_append(NapNetModule->services, url);
    object_notify_d(OBJ(url), (ObjectFunc)on_service_destroy, NULL);
    nap_network_save();
    object_make_public(OBJ(url), NULL);
    return url;
}

static url_t* nap_service_add(const char* url)
{
    url_t* serv = url_parse(url);
    url_t* serv2;
    
    if (!serv) return NULL;

    serv2 = nap_service_find(serv);
    if (serv2)
    {
        object_kill(OBJ(serv));
        return serv2;
    }
    
    nap_service_new(serv);
    return serv;
}


void nap_service_get_list(const char* url)
{
    url_t* serv = nap_service_add(url);
    if (!serv)
    {
        lopster_notify(LOPSTER_MESSAGE_ERROR, _("Invalid address"),
                       _("Cannot use that address to get the server list."));
    }
    else
    {
        nap_napi_get_list(serv);
    }
}



static nap_ng_t* nap_search_ng(const char* name)
{
    GList* dlist;
    
    g_assert(NapNetModule);
    for (dlist = NapNetModule->groups; dlist; dlist = dlist->next)
    {
        nap_ng_t* ng = dlist->data;
        if (!strcmp(OBJ(ng)->name, name)) return ng;
    }
    return NULL;
}


//// GROUPS
//////////////////
int nap_ng_stats(nap_ng_t* ng, int mode)
{
    GList* dlist;
    nap_net_t* net;
    int result = 0;
    
    g_assert(NapNetModule);
    for (dlist = ng->nets; dlist; dlist = dlist->next)
    {
        net = dlist->data;
        if (!net->active || net->active->status != NAP_SERVER_ONLINE)
            continue;
        result += nap_net_stats(net, mode);
    }
    return result;
}

static nap_net_t* nap_ng_search_net(nap_ng_t* ng, const char* name)
{
    GList* dlist;
    
    g_assert(NapNetModule);
    if (!ng) return NULL;
    
    for (dlist = ng->nets; dlist; dlist = dlist->next)
    {
        nap_net_t* net = dlist->data;
        if (!strcasecmp(OBJ(net)->name, name)) return net;
    }
    return NULL;
}

static void nap_ng_add_net(nap_ng_t* ng, nap_net_t* net)
{
    g_assert(NapNetModule);
    if (!ng || !net) return;
    ng->nets = g_list_append(ng->nets, net);

    if (NAP_NET_ONLINE(net)) ng->online++;

    if (NapNetModule->auto_connect && (ng->flags & NAP_NG_AUTO_CONNECT))
        net->flags |= NAP_NET_AUTO;

    if (net->flags & NAP_NET_AUTO)
        ng->nets_auto = g_list_append(ng->nets_auto, net);
        
    net->group = ng;
    object_set_dirty(OBJ(ng));
    object_set_dirty(OBJ(net));
    nap_network_save();
}

static void nap_ng_remove_net(nap_ng_t* ng, nap_net_t* net)
{
    g_assert(NapNetModule);
    if (!ng || !net) return;
    ng->nets = g_list_remove(ng->nets, net);

    if (NAP_NET_ONLINE(net)) ng->online--;
    
    if (net->flags & NAP_NET_AUTO)
    {
        ng->nets_auto = g_list_remove(ng->nets_auto, net);
        net->flags &= ~NAP_NET_AUTO;
    }

    net->group = NULL;
    object_set_dirty(OBJ(ng));
    object_set_dirty(OBJ(net));
    nap_network_save();
}

static void nap_ng_destroy(nap_ng_t* ng)
{
    while (ng->nets)
    {
        nap_net_t* net = ng->nets->data;
        object_kill(OBJ(net));
    }
    NapNetModule->groups = g_list_remove(NapNetModule->groups, ng);
    object_destroy(OBJ(ng));
    nap_network_save();
}

static void nap_ng_init(nap_ng_t* ng, const char* name)
{
    object_init(OBJ(ng), name, FALSE);
    object_add_type(OBJ(ng), "NapGroup");
    OBJ(ng)->destroy = (ObjectFunc)nap_ng_destroy;
    ng->keep = 5;
}

static nap_ng_t* nap_ng_new(const char* name)
{
    nap_ng_t* ng;
    
    g_assert(NapNetModule);
    ng = g_malloc0(sizeof(*ng));
    nap_ng_init(ng, name);
    object_make_public(OBJ(ng), NULL);
    
    return ng;
}

nap_ng_t* nap_ng_create(const char* name, int flags, int mode)
{
    nap_ng_t* ng = NULL;
    
    g_assert(NapNetModule);
    if (mode == NAP_CREATE_UNIQUE)
    {
        ng = nap_search_ng(name);
        if (ng) return NULL;
    }
    else if (mode == NAP_CREATE_REPLACE)
    {
        ng = nap_search_ng(name);
        if (ng) return ng;
    }
    
    ng = nap_ng_new(name);
    ng->flags = flags;
    NapNetModule->groups = g_list_append(NapNetModule->groups, ng);
    nap_network_save();
    return ng;
}


void nap_ng_start_connect(nap_ng_t* ng)
{
    GList* dlist;
    
    g_assert(NapNetModule);
    for (dlist = ng->nets; dlist; dlist = dlist->next)
    {
        nap_net_t* net = dlist->data;
        nap_net_start_connect(net);
    }
}

void nap_ng_stop_connect(nap_ng_t* ng)
{
    GList* dlist;
    
    g_assert(NapNetModule);
    for (dlist = ng->nets; dlist; dlist = dlist->next)
    {
        nap_net_t* net = dlist->data;
        if (!NAP_NET_ONLINE(net)) nap_net_disconnect(net);
    }
}

void nap_ng_set_auto(nap_ng_t* ng, gboolean value)
{
    if (value)
    {
        ng->flags |= NAP_NG_AUTO_CONNECT;
        if (NapNetModule->auto_connect) nap_ng_start_connect(ng);
    }
    else
    {
        ng->flags &= ~NAP_NG_AUTO_CONNECT;
        nap_ng_stop_connect(ng);
    }
    nap_network_save();
}

static void nap_ng_disconnect(nap_ng_t* ng)
{
    GList* dlist;
    
    g_assert(NapNetModule);
    for (dlist = ng->nets; dlist; dlist = dlist->next)
    {
        nap_net_t* net = dlist->data;
        nap_net_disconnect(net);
    }
}

//// NETS
//////////////////
int nap_net_stats(nap_net_t* net, int mode)
{
    int cnt;
    GList* dlist;
    nap_server_t* server;
    int result;
    
    g_assert(NapNetModule);
    if (!net || !net->servers) return 0;
    
    if (net->active && net->active->status == NAP_SERVER_ONLINE)
    {
        server = net->active;
        switch (mode) {
        case 0: return server->users;
        case 1: return server->files;
        case 2: return server->gigs;
        default:return 0;
        }
    }
    else
    {
        result = cnt = 0;
        for (dlist = net->servers; dlist; dlist = dlist->next)
        {
            server = dlist->data;
            //      if (server->users <= 0) continue;
            cnt++;
            switch (mode) {
            case 0: result += server->users; break;
            case 1: result += server->files; break;
            case 2: result += server->gigs;  break;
            default:return 0;
            }
        }
        if (mode != 3)
        {
            if (cnt > 0) result /= cnt;
            else result = 0;
        }
        return result;
    }
}

void nap_net_set_level(nap_net_t* net, int level)
{
    g_assert(NapNetModule);
    if (!net || net->level == level) return;
    net->level = level;
    nap_channel_update_op(net);
}

static void nap_server_set_status(nap_server_t* server, int status,
                                  const char* message)
{
    g_free(server->message);
    server->message = message?g_strdup(message):NULL;
    server->status = status;
    object_set_dirty(OBJ(server));
    object_set_dirty(OBJ(server->net));
}

static gboolean relogin_timer(nap_server_t* server)
{
    server->relogin = 0;
    nap_server_set_status(server, NAP_SERVER_OFFLINE, "Ready to connect");
    return FALSE;
}

gboolean nap_net_end(nap_net_t* net, const char* message)
{
    nap_server_t* server;
    
    g_assert(NapNetModule);
    if (!net->active) return FALSE;
    
    if (net->active->status != NAP_SERVER_ONLINE)
    {
        net->number_fails++;
    }
    else
    {
        net->last_success = CurrentTime;
        net->group->online--;
        NapNetModule->online_nets = g_list_remove(NapNetModule->online_nets, net);
        nap_update_connect_status();
        nap_update_stats();
        nap_chat_left_net(net);
        nap_hot_left_net(net);
    }

    object_set_dirty(OBJ(net));
    socket_set_timeout(SOCK(net), 0);  // disabled timeout checking
    
    g_list_free(net->active_searches);
    net->active_searches = NULL;
    g_list_free(net->queued_searches);
    net->queued_searches = NULL;
    
    net->skip_search_end = 0;
    net->cur_searches = 0;
    
    buffer_clear(net->temp_buffer);
    
    server = net->active;
    net->active = NULL;
    switch (socket_get_status(SOCK(net)))
    {
    case S_NET_USER_DISCONNECT:
        if (server->relogin)
        {
            lopster_timeout_remove(server->relogin);
            server->relogin = 0;
        }
        break;
    case ST_TIMEOUT:
    case ST_ERROR:
        server->ip = 0;
    default:
        if (!server->relogin)
            server->relogin = lopster_timeout_add(60000, (GSourceFunc)relogin_timer, server);
    }
    nap_server_set_status
        (server, server->relogin?NAP_SERVER_WAITING:NAP_SERVER_OFFLINE,
         message);
    
    tcp_socket_end(TSOCK(net), message);
    socket_set_status(SOCK(net), ST_UNKNOWN);
    
    return FALSE;
}

static const char* nap_net_consume_input(nap_net_t* net)
{
    buffer_t* buffer = b_socket_get_in_buffer(BSOCK(net));
    nap_message_t* result;
    
    if (!buffer) return NULL;
    
    while ((result = nap_consume_message(buffer)) != NULL)
    {
        const char* message = nap_handle_message(net, result);
        g_free(result->message);
        g_free(result);
        if (message) return message;
    }
    buffer_consume(buffer, 0);
    return NULL;
}

static void nap_net_copy_some_unimportant(nap_net_t* net) {
  int max_copy = 10240;
  buffer_t* src;
  buffer_t* dest;

  g_assert(buffer_avail(net->temp_buffer) > 0);
  src = net->temp_buffer;  // cannot be NULL
  dest = b_socket_get_out_buffer(BSOCK(net), 4096);

  while (max_copy > 0 && buffer_avail(src) > 4) {
    int size;
    memcpy(&size, buffer_read_pos(src), sizeof(size));
    
    buffer_consume_virt(src, sizeof(size));
    if (size > buffer_avail(src))
    {
        g_log("NAP_NET", G_LOG_LEVEL_CRITICAL,
              "internal temp buffer error");
        buffer_clear(src);
        return;
    }
    buffer_add_data(dest, buffer_read_pos(src), size);
    buffer_consume_virt(src, size);
    max_copy -= size;
  }
  buffer_consume(src, 0);

  // reduce memory usage
  if (src->max > 4096 && src->size < 2048)
    buffer_resize(src, 4096);
}

static gboolean nap_net_prepare_out(nap_net_t* net) {
  buffer_t* buffer = b_socket_get_out_buffer(BSOCK(net), 0);
  if (!buffer_avail(buffer) && buffer_avail(net->temp_buffer)) {
    nap_net_copy_some_unimportant(net);
  }
  return TRUE;
}

/* static void on_charset_changed(pref_node_t* node, nap_net_t* net) { */
/*   (void)node; */
/*   net->charset_warned = FALSE; */
/* } */

static void nap_net_init(nap_net_t* net, const char* name) {
  tcp_socket_init(TSOCK(net), name, 0);
  object_add_type(OBJ(net), "NapNetwork");
  OBJ(net)->destroy = (ObjectFunc)nap_net_destroy;
  socket_set_end(SOCK(net), (SockEndFunc)nap_net_end);
  b_socket_set_consume(BSOCK(net), (BConsumeInputFunc)nap_net_consume_input);
  b_socket_set_prepare_out(BSOCK(net), (BPrepareFunc)nap_net_prepare_out);

  net->level = NAP_L_USER;
  net->temp_buffer = NULL;
  net->charset = NULL;
  net->charset_warned = FALSE;

  socket_add_load(SOCK(net), "Server");
  
/*   pref_observe("/nap/charset", (ObjectFunc)on_charset_changed, net); */
}

static nap_net_t* nap_net_new(const char* name) {
  nap_net_t* net;
  
  g_assert(NapNetModule);
  net = g_malloc0(sizeof(*net));
  nap_net_init(net, name);
  object_make_public(OBJ(net), NULL);

  return net;
}
nap_net_t* nap_net_create(nap_ng_t* ng, const char* name, int mode)
{
    nap_net_t* net = NULL;
    
    g_assert(NapNetModule);
    if (mode == NAP_CREATE_UNIQUE)
    {
        net = nap_ng_search_net(ng, name);
        if (net) return NULL;
    }
    else if (mode == NAP_CREATE_REPLACE)
    {
        net = nap_ng_search_net(ng, name);
        if (net) return net;
    }
    net = nap_net_new(name);
    nap_ng_add_net(ng, net);
    return net;
}

void nap_net_start_connect(nap_net_t* net)
{
    g_assert(NapNetModule);
    if (net->flags & NAP_NET_AUTO) return;
    net->group->nets_auto = g_list_append(net->group->nets_auto, net);
    net->flags |= NAP_NET_AUTO;
    object_set_dirty(OBJ(net));
}

void nap_net_disconnect(nap_net_t* net)
{
    g_assert(NapNetModule);

    if (net->flags & NAP_NET_AUTO)
    {
        net->group->nets_auto = g_list_remove(net->group->nets_auto, net);
        net->flags &= ~NAP_NET_AUTO;
    }

    if (net->active)
        socket_done(SOCK(net), S_NET_USER_DISCONNECT, "User disconnected");
    object_set_dirty(OBJ(net));
}

static void nap_channel_destroy(nap_ce_t* entry)
{
    g_free(entry->name);
    g_free(entry->topic);
    g_free(entry);
}
static void nap_channels_destroy(nap_channels_t* channels)
{
    g_list_foreach(channels->list, (GFunc)nap_channel_destroy, NULL);
    g_list_free(channels->list);
    object_destroy(OBJ(channels));
}

static nap_channels_t* nap_channels_new(nap_net_t* net)
{
    nap_channels_t* channels = g_malloc(sizeof(*channels));
    object_init(OBJ(channels), OBJ(net)->name, TRUE);
    object_add_type(OBJ(channels), "NapChannels");
    OBJ(channels)->destroy = (ObjectFunc)nap_channels_destroy;
    channels->net = net;
    channels->list = NULL;
    channels->refreshed = 0;
    channels->refreshing = FALSE;
    channels->min = pref_get_ref("/nap/chat/channels/min", int);
    channels->max = pref_get_ref("/nap/chat/channels/max", int);
    return channels;
}

static void nap_net_channels_clear(nap_net_t* net)
{
    if (!net->channel_list) return;
    g_list_foreach(net->channel_list->list, (GFunc)nap_channel_destroy, NULL);
    g_list_free(net->channel_list->list);
    net->channel_list->list = NULL;
}

void nap_net_get_channels(nap_net_t* net)
{
    if (!net || !net->channel_list) return;
    
    if (net->channel_list->refreshing) return; //FIXME: notify user.
    
    net->channel_list->refreshing = TRUE;
    nap_net_channels_clear(net);
    nap_net_send(net, NAP_S_GET_CHANNELS, FALSE, "");
    
    object_set_dirty(OBJ(net->channel_list));
}

void nap_net_show_channels(nap_net_t* net)
{
    if (!net->channel_list)
    {
        net->channel_list = nap_channels_new(net);
    }
    object_make_public(OBJ(net->channel_list), NULL);
  
    if (net->channel_list->list)
        object_set_dirty(OBJ(net->channel_list));
    else
        nap_net_get_channels(net);
}


void nap_net_get_global_users(nap_net_t* net, char *args)
{
    g_assert(NapNetModule);
    if (!NAP_NET_ONLINE(net)) return;
    nap_net_send(net, NAP_S_GET_GUSER, FALSE, "%s", args?args:"* eam");
}


static void nap_net_destroy(nap_net_t* net)
{
/*   pref_unobserve("/nap/charset", (ObjectFunc)on_charset_changed, net); */
    if (net->channel_list) object_kill(OBJ(net->channel_list));
    
    nap_net_disconnect(net);
    while (net->servers)
    {
        object_kill(OBJ(net->servers->data));
    }
    if (net->group) nap_ng_remove_net(net->group, net);
    g_free(net->cur_nick);
    g_free(net->cur_pass);
    g_free(net->nick);
    g_free(net->pass);
    if (net->autojoin)
    {
        g_slist_foreach(net->autojoin, (GFunc)g_free, NULL);
        g_slist_free(net->autojoin);
    }
    
    if (net->temp_buffer) buffer_destroy(net->temp_buffer);
    
    tcp_socket_destroy(TSOCK(net));
}

void nap_net_move(nap_net_t* net, nap_ng_t* to)
{
    nap_ng_t* from;
    
    g_assert(NapNetModule);
    from = net->group;
    if (from == to) return;
    nap_ng_remove_net(from, net);
    nap_ng_add_net(to, net);
}

static void nap_net_add_server(nap_net_t* net, nap_server_t* server) {
  g_assert(NapNetModule);
  if (!net || !server) return;
  net->servers = g_list_append(net->servers, server);
  server->net = net;
  object_set_dirty(OBJ(net));
  object_set_dirty(OBJ(server));
  nap_network_save();
}

static void nap_net_remove_server(nap_net_t* net, nap_server_t* server) {
  g_assert(NapNetModule);
  if (!net || !server) return;
  net->servers = g_list_remove(net->servers, server);
  server->net = NULL;
  object_set_dirty(OBJ(net));
  object_set_dirty(OBJ(server));
  nap_network_save();
}

char* nap_net_print_version(nap_net_t* net) {
  char* result = get_text_buffer();

  g_assert(NapNetModule);
  if (net->type == NAP_TYPE_OPENNAP)
    sprintf(result, "Opennap %d.%d", net->major, net->minor);
  else if (net->type == NAP_TYPE_OPENNAP_NG)
    sprintf(result, "Opennap-ng %d.%d", net->major, net->minor);
  else if (net->type == NAP_TYPE_SLAVANAP)
    sprintf(result, "SlavaNap %d.%d.%d", net->major, net->minor, net->micro);
  else
    strcpy(result, _("Unknown server"));
  return result;
}

void nap_net_send(nap_net_t* net, guint16 type, gboolean convert, 
                  const char* fmt, ...) {
  va_list ap;
  GList* dlist = NULL;
  char* message;

  g_assert(NapNetModule);
  if (!net && NapNetModule->online_nets) {
    dlist = NapNetModule->online_nets;
    net = dlist->data;
  }

  va_start(ap, fmt);
  message = g_strdup_vprintf(fmt, ap);
  va_end(ap);

  while (net) {
    if (net->active) {
      buffer_t* buffer = b_socket_get_out_buffer(BSOCK(net), 1024);
      if (convert)
        nap_pack_command(buffer, type, auto_text
                         (from_utf8(message, -1, NULL, NULL, 4,
                                    net->charset, CharSet, "latin1", "utf-8")));
      else
        nap_pack_command(buffer, type, message);
      b_socket_start_sending(BSOCK(net));
    }
    if (dlist) {
      dlist = dlist->next;
      if (dlist) net = dlist->data;
      else net = NULL;
    } else {
      net = NULL;
    }
  }
  g_free(message);
}

static void
nap_net_send_unimportant(nap_net_t* net, guint16 type, gboolean convert,
                         const char* fmt, ...) {
  va_list ap;
  GList* dlist = NULL;
  char* message;

  g_assert(NapNetModule);
  if (!net && NapNetModule->online_nets) {
    dlist = NapNetModule->online_nets;
    net = dlist->data;
  }

  va_start(ap, fmt);
  message = g_strdup_vprintf(fmt, ap);
  va_end(ap);

  while (net) {
    if (net->active) {
      int pos;
      int size;
      buffer_t* buffer;

      if (!net->temp_buffer) net->temp_buffer = buffer_new(1024);
      buffer = net->temp_buffer;

      pos = buffer->size;
      // leave space for length
      buffer->size += sizeof(size);

      if (convert)
        nap_pack_command(buffer, type, auto_text
                         (from_utf8(message, -1, NULL, NULL, 4,
                                    net->charset, CharSet, "latin1", "utf-8")));
      else
        nap_pack_command(buffer, type, message);

      size = buffer->size - pos - 4;
      if (size > 0) {
        memcpy(buffer->data+pos, &size, sizeof(size));
      } else {
        // command was not written for some reason, reset to old buffer
        buffer->size = pos;
      }
      
      b_socket_start_sending(BSOCK(net));
    }
    if (dlist) {
      dlist = dlist->next;
      if (dlist) net = dlist->data;
      else net = NULL;
    } else {
      net = NULL;
    }
  }
  g_free(message);
}

port_t nap_get_listen_port() {
  g_assert(NapNetModule);
  return listen_port_get_port(NapNetModule->listen);
}

/** really connects the network */
static void nap_net_connect_real_to(nap_net_t* net, nap_server_t* server) {
  socket_t* socket = SOCK(net);
  int speed;

  g_assert(!net->active);
  
  socket_set_timeout(socket, 30);

  net->active = server;
  nap_server_set_status(server, NAP_SERVER_CONNECTING, _("Connecting..."));

  if (!tcp_socket_connect(TSOCK(net), server->ip, htons(server->port))) {
    server->ip = 0;
    return;
  }

  b_socket_start_receiving(BSOCK(net), 1024000);
  
  speed = pref_get_ref("/core/line/type", int);
  nap_net_send(net, NAP_S_LOGIN, TRUE, "%s %s %d \"%s\" %d \"anon@anon\"",
               net->cur_nick, net->cur_pass,
               nap_get_listen_port(), FULL_VERSION, 
               NapSpeedFromType(speed));
}


void nap_server_resolver(nap_server_t* server, unsigned long ip) {
  GList* dlist;

  g_assert(NapNetModule);
  // check whether server is still valid
  dlist = g_list_find(NapNetModule->resolver, server);
  if (!dlist) return;
  NapNetModule->resolver = g_list_delete_link(NapNetModule->resolver, dlist);

  server->ip = ip;
  if (ip == 0) {
    nap_server_set_status(server, NAP_SERVER_WAITING, _("could not resolve"));
    if (!server->relogin)
      server->relogin = lopster_timeout_add(120000, (GSourceFunc)relogin_timer, server);
  } else {
    // now ready to connect
    nap_server_set_status(server, NAP_SERVER_OFFLINE,
                          _("successfully resolved address"));
    if (!server->net->active) {
      nap_net_connect_real_to(server->net, server);
    }
  }
}

char* nap_net_get_nick(nap_net_t* net) {
  char* nick = NULL;

  g_assert(NapNetModule);
  if (net) nick = net->nick;
  if (!nick || *nick == 0 || strchr(nick, ' '))
    nick = pref_get_ptr("/nap/username", char);
  if (!nick || *nick == 0 || strchr(nick, ' '))
    nick = pref_get_ptr("/core/user/nick", char);
  if (!nick || *nick == 0 || strchr(nick, ' '))
    nick = NULL;
  return nick;
}

char* nap_net_get_pass(nap_net_t* net) {
  char* pass = NULL;

  g_assert(NapNetModule);
  if (net) pass = net->pass;
  if (!pass || *pass == 0 || strchr(pass, ' '))
    pass = pref_get_ptr("/nap/password", char);
  if (!pass || *pass == 0 || strchr(pass, ' '))
    pass = pref_get_ptr("/core/user/pass", char);
  if (!pass || *pass == 0 || strchr(pass, ' '))
    pass = NULL;
  return pass;
}


/** Connects the network, but first prepares it for connect */
static void nap_net_connect_to(nap_net_t* net, nap_server_t* server)
{
    char* nick = NULL;
    char* pass = NULL;
    
    g_assert(NapNetModule);
    g_assert(!net->active);
    
    net->servers = g_list_remove(net->servers, server);
    net->servers = g_list_append(net->servers, server);
  
    nick = nap_net_get_nick(net);
    if (!nick)
    {
        g_log("NAP_NET", G_LOG_LEVEL_WARNING,
              "Cannot connect to %s (no username given)",
              OBJ(net)->name);
        return;
    }
    pass = nap_net_get_pass(net);
    if (!pass)
    {
        g_log("NAP_NET", G_LOG_LEVEL_WARNING,
              "Cannot connect to %s (no password given)",
              OBJ(net)->name);
        return;
    }
    g_free(net->cur_nick);
    net->cur_nick = g_strdup(nick);
    g_free(net->cur_pass);
    net->cur_pass = g_strdup(pass);
    
    if (server->relogin)
    {
        lopster_timeout_remove(server->relogin);
        server->relogin = 0;
    }
    if (server->ip == 0)
    {
        NapNetModule->resolver = g_list_append(NapNetModule->resolver, server);
        nap_server_set_status(server, NAP_SERVER_RESOLVING, _("resolving..."));
        resolve_address(OBJ(server)->name, (call_func_t)nap_server_resolver, server);
    }
    else
    {
        nap_net_connect_real_to(net, server);
    }
}

static nap_server_t* nap_net_search_server(nap_net_t* net, const char* address) {
  GList* dlist;
  nap_server_t* server;

  g_assert(NapNetModule);
  if (!net) return NULL;

  for (dlist = net->servers; dlist; dlist = dlist->next) {
    server = dlist->data;
    if (!strcasecmp(OBJ(server)->name, address)) return server;
  }
  return NULL;
}


//// SERVERS
//////////////////
void nap_server_connect(nap_server_t* server)
{
    nap_net_t* net = server->net;
    
    g_assert(NapNetModule);
    if (net->active == server) return;
    if (net->active) nap_net_disconnect(net);
    nap_net_connect_to(net, server);

}

void nap_server_disconnect(nap_server_t* server)
{
    nap_net_t* net = server->net;
    g_assert(NapNetModule);
    nap_net_disconnect(net);
}

static void nap_server_destroy(nap_server_t* server) {
  NapNetModule->resolver = g_list_remove(NapNetModule->resolver, server);
  if (server->net) {
    if (server->net->active == server)
      nap_net_disconnect(server->net);
    nap_net_remove_server(server->net, server);
  }
  if (server->relogin) {
    lopster_timeout_remove(server->relogin);
    server->relogin = 0;
  }

  g_free(server->message);
  g_free(server->admin);
  g_free(server->web);
  object_destroy(OBJ(server));
}

static void nap_server_init(nap_server_t* server, const char* address)
{
    object_init(OBJ(server), address, TRUE);
    object_add_type(OBJ(server), "NapServer");
    OBJ(server)->destroy = (ObjectFunc)nap_server_destroy;
    server->ip = 0;
    server->last_success = 0;
    server->number_fails = 0;
}

static nap_server_t* nap_server_new(const char* address)
{
    nap_server_t* server = g_malloc0(sizeof(*server));
    nap_server_init(server, address);
    object_make_public(OBJ(server), NULL);
    return server;
}

nap_server_t* nap_server_create(nap_net_t* net, const char* address, int mode) {
  nap_server_t* server;

  g_assert(NapNetModule);
  if (mode == NAP_CREATE_UNIQUE) {
    server = nap_net_search_server(net, address);
    if (server) return NULL;
  } else if (mode == NAP_CREATE_REPLACE) {
    server = nap_net_search_server(net, address);
    if (server) return server;
  }
  server = nap_server_new(address);
  nap_net_add_server(net, server);
  return server;
}

////////////////////
////////////////////
static void xml_save_server(nap_server_t* server, xmlNodePtr xnode)
{
    xmlNodePtr xsub = xmlNewChild(xnode, NULL, "Server", NULL);
    xmlNewProp(xsub, "Address", OBJ(server)->name);
    xmlNewProp(xsub, "Port", l_printf("%u", server->port));
    xml_set_bool_attr(xsub, "Meta", (server->flags & NAP_SERVER_META));
    xmlNewTextChild(xsub, NULL, "IP", l_printf("%lu", server->ip));
    xmlNewTextChild(xsub, NULL, "Speed", l_printf("%d", server->linespeed));
    if (server->admin)
        xmlNewTextChild(xsub, NULL, "Admin", server->admin);
    if (server->web)
        xmlNewTextChild(xsub, NULL, "Web", server->web);
    xmlNewProp(xsub, "LastSuccess", l_printf("%u", server->last_success));
    xmlNewProp(xsub, "Fails", l_printf("%u", server->number_fails));
}

static void xml_save_net(nap_net_t* net, xmlNodePtr xnode)
{
    GSList* dlist;
    xmlNodePtr xsub = xmlNewChild(xnode, NULL, "Network", NULL);
    xmlNewProp(xsub, "Name", OBJ(net)->name);
    if (net->nick)
        xmlNewProp(xsub, "Username", net->nick);
    if (net->pass)
        xmlNewProp(xsub, "Password", net->pass);
    if (net->charset && *net->charset)
        xmlNewProp(xsub, "Encoding", net->charset);

    xml_set_bool_attr(xsub, "OwnServer", (net->flags & NAP_NET_OWN_SERVER));
    xml_set_bool_attr(xsub, "IgnoreGlobal", (net->flags & NAP_NET_IGN_GLOBAL));
    xml_set_bool_attr(xsub, "EmulateWinMX", (net->flags & NAP_NET_EMU_WINMX));
    xml_set_bool_attr(xsub, "Private", (net->flags & NAP_NET_PRIVATE));

    xmlNewProp(xsub, "LastSuccess", l_printf("%u", net->last_success));
    xmlNewProp(xsub, "Fails", l_printf("%u", net->number_fails));
    for (dlist = net->autojoin; dlist; dlist = dlist->next) {
        char* channel = dlist->data;
        xmlNewTextChild(xsub, NULL, "AutoJoin", channel);
    }
    g_list_foreach(net->servers, (GFunc)xml_save_server, xsub);
}

static void xml_save_group(nap_ng_t* ng, xmlNodePtr xnode)
{
    xmlNodePtr xsub = xmlNewChild(xnode, NULL, "Group", NULL);
    xmlNewProp(xsub, "Name", OBJ(ng)->name);
    xmlNewProp(xsub, "StayConnected", l_printf("%d", ng->keep));
    if (ng->flags & NAP_NG_AUTO_CONNECT)
        xmlNewProp(xsub, "AutoConnect", "yes");
    g_list_foreach(ng->nets, (GFunc)xml_save_net, xsub);
}

static void xml_save_service(url_t* serv, xmlNodePtr xroot)
{
    xmlNodePtr xsub = xmlNewChild(xroot, NULL, "Service", NULL);
    xmlNewProp(xsub, "Url", url_format(serv));
}
static void xml_save_groups(xmlDocPtr xfile, xmlNodePtr xroot)
{
    (void)xfile;
    g_list_foreach(NapNetModule->services, (GFunc)xml_save_service, xroot);
    g_list_foreach(NapNetModule->groups, (GFunc)xml_save_group, xroot);
}

static nap_server_t* xml_get_server(xmlDocPtr xdoc, xmlNodePtr xnode)
{
    nap_server_t* server;
    xmlChar* attr;
    
    attr = xmlGetProp(xnode, "Address");
    server = nap_server_new(attr);
    xmlFree(attr);
    attr = xmlGetProp(xnode, "Port");
    server->port = atoi(attr);
    xmlFree(attr);
    
    server->flags |=
        xml_get_bool_attr(xnode, "Meta", FALSE) ?NAP_SERVER_META:0;
    
    xnode = xnode->xmlChildrenNode;
    while (xnode != NULL)
    {
        attr = xmlNodeListGetString(xdoc, xnode->xmlChildrenNode, 1);
        if (!strcmp(xnode->name, "IP"))
        {
            server->ip = strtoul(attr, NULL, 10);
        }
        else if (!strcmp(xnode->name, "Speed"))
        {
            server->linespeed = atoi(attr);
        }
        else if (!strcmp(xnode->name, "Admin"))
        {
            server->admin = attr?g_strdup(attr):NULL;
        }
        else if (!strcmp(xnode->name, "Web"))
        {
            server->web = attr?g_strdup(attr):NULL;
        }
        if (attr) xmlFree(attr);
        xnode = xnode->next;
    }
    return server;
}

static nap_net_t* xml_get_net(xmlDocPtr xdoc, xmlNodePtr xnode)
{
    nap_net_t* net;
    xmlChar* attr;
    
    attr = xmlGetProp(xnode, "Name");
    net = nap_net_new(attr);
    xmlFree(attr);
    attr = xmlGetProp(xnode, "Username");
    if (attr)
    {
        net->nick = g_strdup(attr);
        xmlFree(attr);
    }
    attr = xmlGetProp(xnode, "Password");
    if (attr)
    {
        net->pass = g_strdup(attr);
        xmlFree(attr);
    }
    attr = xmlGetProp(xnode, "Encoding");
    if (attr)
    {
        if (*attr) net->charset = g_strdup(attr);
        xmlFree(attr);
    }
    net->flags |=
        xml_get_bool_attr(xnode, "Private", FALSE)?NAP_NET_PRIVATE:0;
    net->flags |=
        xml_get_bool_attr(xnode, "OwnServer", FALSE)?NAP_NET_OWN_SERVER:0;
    net->flags |=
        xml_get_bool_attr(xnode, "IgnoreGlobal", FALSE)?NAP_NET_IGN_GLOBAL:0;
    net->flags |=
        xml_get_bool_attr(xnode, "EmulateWinMX", FALSE)?NAP_NET_EMU_WINMX:0;
    attr = xmlGetProp(xnode, "LastSuccess");
    if (attr)
    {
        net->last_success = strtoul(attr, NULL, 10);
        xmlFree(attr);
    }
    attr = xmlGetProp(xnode, "Fails");
    if (attr)
    {
        net->number_fails = strtoul(attr, NULL, 10);
        xmlFree(attr);
    }
    
    xnode = xnode->xmlChildrenNode;
    while (xnode != NULL)
    {
        if (!strcmp(xnode->name, "Server"))
        {
            nap_server_t* server = xml_get_server(xdoc, xnode);
            if (server) nap_net_add_server(net, server);
        }
        else if (!strcmp(xnode->name, "AutoJoin"))
        {
            attr = xmlNodeGetContent(xnode->xmlChildrenNode);
            if (attr)
            {
                net->autojoin = g_slist_append(net->autojoin, g_strdup(attr));
                xmlFree(attr);
            }
        }
        xnode = xnode->next;
    }
    return net;
}

static nap_ng_t* xml_get_group(xmlDocPtr xdoc, xmlNodePtr xnode) {
  nap_ng_t* ng;
  xmlChar* attr;

  attr = xmlGetProp(xnode, "Name");
  ng = nap_ng_new(attr);
  xmlFree(attr);
  attr = xmlGetProp(xnode, "StayConnected");
  if (attr) {
    ng->keep = atoi(attr);
    xmlFree(attr);
  }
  attr = xmlGetProp(xnode, "AutoConnect");
  if (attr) {
    if (strcmp(attr, "no")) ng->flags |= NAP_NG_AUTO_CONNECT;
    xmlFree(attr);
  }
  
  xnode = xnode->xmlChildrenNode;
  while (xnode != NULL) {
    if (!strcmp(xnode->name, "Network")) {
      nap_net_t* net = xml_get_net(xdoc, xnode);
      if (net) nap_ng_add_net(ng, net);
    }
    xnode = xnode->next;
  }
  return ng;
}

static void xml_load_groups(xmlDocPtr xfile, xmlNodePtr xroot)
{
    xmlNodePtr xsub = xroot->xmlChildrenNode;
    while (xsub != NULL)
    {
        if (!strcmp(xsub->name, "Group"))
        {
            nap_ng_t* ng = xml_get_group(xfile, xsub);
            if (ng)
            {
                NapNetModule->groups = g_list_append(NapNetModule->groups, ng);
            }
        }
        else if (!strcmp(xsub->name, "Service"))
        {
            xmlChar* attr1;
            attr1 = xmlGetProp(xsub, "Url");
            if (attr1) 
            {
                nap_service_add(attr1);
                xmlFree(attr1);
            }
        }
        xsub = xsub->next;
    }
}

static void nap_network_load()
{
    NapNetModule->save = FALSE;
    
    if (!lopster_xml_load("nap:serverlist"))
    {
        NapNetModule->save = TRUE;
    }
    if (!NapNetModule->services)
        nap_service_add("http://www.gotnap.com/servers.php");

  NapNetModule->save = TRUE;
}


/* void nap_net_move_flags(int from, int to) { */
/*   GList* dlist1; */
/*   GList* dlist2; */
/*   GList* dlist3; */

/*   g_assert(NapNetModule); */
/*   for (dlist1 = NapNetModule->groups; dlist1; dlist1 = dlist1->next) { */
/*     nap_ng_t* ng = dlist1->data; */
/*     for (dlist2 = ng->nets; dlist2; dlist2 = dlist2->next) { */
/*       nap_net_t* net = dlist2->data; */
/*       for (dlist3 = net->servers; dlist3; dlist3 = dlist3->next) { */
/*         nap_server_t* server = dlist3->data; */
/*         if ((server->flags & from) == from) { */
/*           server->flags &= ~from; */
/*           server->flags |= to; */
/*         } else { */
/*           server->flags &= ~to; */
/*         } */
/*         object_set_dirty(OBJ(server)); */
/*       } */
/*     } */
/*   } */
/* } */

/* void nap_net_remove_not_napi() */
/* { */
    /*FIXME*/
/*     GList* dlist1; */
    
/*     g_assert(NapNetModule); */
/*     dlist1 = NapNetModule->groups; */
/*     while (dlist1) */
/*     { */
/*         nap_ng_t* ng = dlist1->data; */
/*         GList* dlist2 = ng->nets; */
/*         dlist1 = dlist1->next; */
/*         while (dlist2) */
/*         { */
/*             nap_net_t* net = dlist2->data; */
/*             dlist2 = dlist2->next; */
/*             if (!(net->flags & NAP_NET_PERSISTENT)) */
/*             { */
/*                 GList* dlist3 = net->servers; */
/*                 while (dlist3) { */
/*                     nap_server_t* server = dlist3->data; */
/*                     dlist3 = dlist3->next; */
/*                     if (!(server->flags & NAP_SERVER_NAPI)) object_kill(OBJ(server)); */
/*                 } */
/*                 if (!net->servers) object_kill(OBJ(net)); */
/*             } */
/*         } */
/*         if (!ng->nets) object_kill(OBJ(ng)); */
/*     } */
/* } */

int nap_net_success(nap_net_t* net)
{
    g_assert(NapNetModule);
    if (!net || !net->active) return 0;
    
    socket_set_timeout(SOCK(net), 600);
    net->number_fails = 0;
    net->last_success = CurrentTime;
    net->group->online++;
    NapNetModule->online_nets = g_list_append(NapNetModule->online_nets, net);
    nap_update_connect_status();
    nap_update_stats();
    
    nap_server_set_status(net->active, NAP_SERVER_ONLINE, _("Online"));
    
    return 1;
}

static gboolean nap_net_connect_next()
{
    GList* dlist;
    
    g_assert(NapNetModule);

    for (dlist = NapNetModule->groups; dlist; dlist = dlist->next)
    {
        nap_ng_t* ng = dlist->data;
        GList* dlist2;

        if (ng->keep >= 0 && ng->keep <= ng->online) continue;
        
        for (dlist2 = ng->nets_auto; dlist2; dlist2 = dlist2->next)
        {
            nap_net_t* net = dlist2->data;
            GList* dlist3;

            if (net->active) continue;

            for (dlist3 = net->servers; dlist3; dlist3 = dlist3->next)
            {
                nap_server_t* server = dlist3->data;
                if (server->status != NAP_SERVER_OFFLINE) continue;
                nap_server_connect(server);
                break;
            }
        }
    }
    return TRUE;
}



static void nap_send_listen() {
  if (NapNetModule) {
    nap_net_send(NULL, NAP_S_CHANGE_PORT, FALSE, "%d", nap_get_listen_port());
  }
}

static const char* nap_handle_con(tcp_socket_t* s) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(s));
  char* pos;

  if (buffer_avail(buffer) < 3) {
    // we need more bytes
    b_socket_set_limit(BSOCK(s), 8);
    return NULL;
  }
  
  pos = buffer_read_pos(buffer);
  if (!strncmp(pos, "GETLIST", 7)) {
    buffer_consume_virt(buffer, 7);
    return nap_browse_incoming_send(s);
  } else if (!strncmp(pos, "SENDLIST", 8)) {
    buffer_consume_virt(buffer, 8);
    return nap_browse_incoming_get(s);
  } else if (!strncmp(pos, "GET", 3)) {
    buffer_consume_virt(buffer, 3);
    return nap_upload_incoming(s);
  } else if (!strncmp(pos, "SEND", 4)) {
    buffer_consume_virt(buffer, 4);
    return nap_download_incoming(s);
  } else {
    print_raw("invalid", pos, buffer_avail(buffer));
    return "incoming data is invalid";
  }
}

static tcp_socket_t* nap_accept_con(int fd, ip_t ip, port_t port) {
  tcp_socket_t* socket = tcp_socket_new("NapTemp", 30);
  
  listen_init_accepted(socket, fd, ip, port);

  // set handler which detects the connection type
  b_socket_set_consume(BSOCK(socket), (BConsumeInputFunc)nap_handle_con);
  tcp_socket_send_str(socket, "1");
  b_socket_start_receiving(BSOCK(socket), 1024);
  return socket;
}

static void nap_change_port() {
  nap_send_listen();
}

static void nap_start_listen_service() {
  g_assert(NapNetModule);
  NapNetModule->listen =
    listen_port_new("/nap/listen/port", "NapNetModule->listen",
                    (ListenAcceptFunc)nap_accept_con,
                    (ListenChangeFunc)nap_change_port);
  object_set_ptr((object_t**)&NapNetModule->listen);
  listen_port_start_listening(NapNetModule->listen, 10);
  nap_send_listen();
}

void nap_usermode_send(nap_net_t* net, int mode) {
  GList* dlist;
  g_assert(NapNetModule);
  if (!net) {
    for (dlist = NapNetModule->online_nets; dlist; dlist = dlist->next) {
      net = dlist->data;
      nap_usermode_send(net, mode);
    }
  } else {
    int mask;
    if (net->type == NAP_TYPE_SLAVANAP)
      mask = NAP_MODE_MASK_SN;
    else if (net->type == NAP_TYPE_OPENNAP)
      mask = NAP_MODE_MASK_ON;
    else if (net->type == NAP_TYPE_OPENNAP_NG)
      mask = NAP_MODE_MASK_NG;
    else
      mask = NAP_MODE_MASK_ALL;
    
    nap_net_send(net, NAP_S_SET_USERMODE, FALSE, "%s", NapUserMode(mode, mask));
  }
}

static file_node_t* share_cb(file_tree_t* tree, file_node_t* node, nap_net_t* net) {
  (void)tree;
  if (node && node->file) nap_net_share_file(net, node->file);
  return NULL;
}

static void nap_net_share_lib(nap_net_t* net) {
  g_assert(NapNetModule);
  library_foreach_shared((FileNodeTraverse)share_cb, net);
}

void nap_net_go(nap_net_t* net) {
  int mode = pref_get_ref("/nap/usermode", int);

  g_assert(NapNetModule);
  if (net->type == NAP_TYPE_SLAVANAP) {
    net->max_searches = 1;
    net->max_searches_pm = 3;
  } else if (net->type == NAP_TYPE_OPENNAP) {
    net->max_searches = 3;
    net->max_searches_pm = 100;  // allow 100 searches per minute
  } else if (net->type == NAP_TYPE_OPENNAP_NG) {
    net->max_searches = 3;
    net->max_searches_pm = 100;  // allow 100 searches per minute
  } else {
    // ??
    net->max_searches = 3;
    net->max_searches_pm = 100;  // allow 100 searches per minute
  }
  net->searches_pm = 0;
  nap_usermode_send(net, mode);
  nap_hot_joined_net(net);
  nap_net_share_lib(net);
  nap_chat_joined_net(net);
  nap_search_joined_net(net);
}


void nap_net_finish_channel(nap_net_t* net) {
  if (net->channel_list) {
    net->channel_list->refreshing = FALSE;
    object_set_dirty(OBJ(net->channel_list));
  }
}

void nap_net_add_channel(nap_net_t* net, const char* name,
                         int online, int level, int limit,
                         const char* topic) {
  if (net->channel_list) {
    nap_ce_t* ce = g_malloc0(sizeof(*ce));
    g_assert(NapNetModule);
    ce->net = net;
    ce->name = g_strdup(name);
    ce->topic = topic?g_strdup(topic):NULL;
    ce->online = online;
    ce->limit = limit;
    ce->level = level;
    net->channel_list->list = g_list_append(net->channel_list->list, ce);
  }
}
			 
nap_net_t* nap_get_net(int index) {
  g_assert(NapNetModule);
  return g_list_nth_data(NapNetModule->online_nets, index-1);
}

static gboolean nap_net_search_next(nap_net_t* net)
{
    search_t* search = 0;
    GList* dlist;
    const char* request;
    
    if (!net->queued_searches) return FALSE;

    if (net->cur_searches >= net->max_searches)
    {
        printf("too many: %s %d %d\n", OBJ(net)->name, 
               net->cur_searches, net->max_searches);
        return FALSE;
    }
    if (net->searches_pm >= net->max_searches_pm)
    {
        printf("too mpm: %s %d %d\n", OBJ(net)->name, 
               net->searches_pm, net->max_searches_pm);
        return FALSE;
    }
  
    // normal searches have higher priority
    for (dlist = net->queued_searches; dlist; dlist = dlist->next)
    {
        search = dlist->data;
        if (!search->download) break;
    }
    if (!search) search = net->queued_searches->data;

    net->queued_searches = g_list_remove(net->queued_searches, search);
    net->active_searches = g_list_append(net->active_searches, search);
    net->searches_pm++;
    net->cur_searches++;
    
    request = nap_search_create_request(search);
    nap_net_send_unimportant(net, NAP_S_SEARCH, TRUE, "%s", request);
    
    return TRUE;
}

void nap_net_remove_search(nap_net_t* net, search_t* search) {
  GList* dlist;
  
  g_assert(NapNetModule);
  g_assert(search);
  if (!net) {
    for (dlist = NapNetModule->online_nets; dlist; dlist = dlist->next) {
      net = dlist->data;
      nap_net_remove_search(net, search);
    }
  } else {
    net->queued_searches =
      g_list_remove(net->queued_searches, search);
    net->active_searches =
      g_list_remove(net->active_searches, search);
  }
}

static void nap_net_do_search(nap_net_t* net, search_t* search) {
  g_assert(net);
  if (!g_list_find(net->active_searches, search) &&
      !g_list_find(net->queued_searches, search)) {
    net->queued_searches = g_list_append(net->queued_searches, search);
    nap_net_search_next(net);
  }
}

void nap_net_add_search(nap_net_t* net, search_t* search) {
  g_assert(NapNetModule);
  if (net) {
    nap_net_do_search(net, search);
  } else {
    GList* dlist;
    for (dlist = NapNetModule->online_nets; dlist; dlist = dlist->next) {
      net = dlist->data;
      nap_net_add_search(net, search);
    }
  }
}

static gboolean nap_net_reset_searches()
{
    GList* dlist;
    g_assert(NapNetModule);
    for (dlist = NapNetModule->online_nets; dlist; dlist = dlist->next)
    {
        nap_net_t* net = dlist->data;
        net->searches_pm = 0;
        while (nap_net_search_next(net)) ;
    }
    return TRUE;
}

void nap_net_new_result(file_t* file, nap_user_t* user)
{
    GList* dlist;
    
    g_assert(NapNetModule);
    g_assert(file && user && user->net);
    
    for (dlist = user->net->active_searches; dlist; dlist = dlist->next)
    {
        search_t* search = dlist->data;
        attribute_t* attr;
        
        if (search->download)
        {
            if (download_matches_source(search->download, file, USER(user)))
                download_add_source(search->download, file, USER(user));
        }
        else
        {
            if (!search_pattern_check_file(SEARCHP(search), file, TRUE)) continue;
            
            attr = attr_find(ATTRC(search), "pattern", "NapSpeed");
            if (attr)
            {
                int min = attr_get_refi(attr, int, 0);
                int max = attr_get_refi(attr, int, 1);
                if (user->speed < min || user->speed > max) continue;
            }
            
            search_add_result(search, file, USER(user), 1, FALSE);
        }
    }
    
    /*
      g_log("NAP_NET", G_LOG_LEVEL_DEBUG,
      "no search found for file %s", file->filename);
    */
}

void nap_net_search_finish_latest(nap_net_t* net)
{
    GList* dlist;
    search_t* search;
    
    g_assert(NapNetModule);
    g_assert(net);
    
    if (!net->active_searches) return;
    dlist = g_list_last(net->active_searches);
    search = dlist->data;
    net->active_searches =
        g_list_delete_link(net->active_searches, dlist);
    net->cur_searches--;
    
    if (search)
    {
        g_log("SEARCH", G_LOG_LEVEL_DEBUG, "(%s) was finished, requeuing\n",
              OBJ(search)->name?OBJ(search)->name:"null");
        net->queued_searches = g_list_append(net->queued_searches, search);
    }
    nap_net_search_next(net);
}

void nap_net_search_finish(nap_net_t* net) {
  g_assert(net);

  g_assert(NapNetModule);
  if (!net->active_searches) return;
  net->active_searches =
    g_list_delete_link(net->active_searches, net->active_searches);
  net->cur_searches--;

  nap_net_search_next(net);
}

void nap_net_global_pass_changed(const char* passwd) {
  const char* pass;

  g_assert(NapNetModule);
  pass = pref_get_ptr("/nap/password", char);
  if (pass && *pass != 0) return;  // we have our own pass, no action
  nap_net_local_pass_changed(passwd);
}

void nap_net_local_pass_changed(const char* passwd) {
  GList* dlist;
  const char* pass;

  g_assert(NapNetModule);
  for (dlist = NapNetModule->online_nets; dlist; dlist = dlist->next) {
    nap_net_t* net = dlist->data;
    pass = net->pass;
    if (pass && *pass != 0) continue; // net has its own pass, no action
    nap_net_send(net, NAP_S_PASSWORD, TRUE, "%s", passwd);
  }
}

void nap_net_set_password(nap_net_t* net, const char* passwd) {
  g_assert(NapNetModule);

  g_free(net->pass);
  net->pass = (passwd && *passwd)?g_strdup(passwd):NULL;
  
  if (g_list_find(NapNetModule->online_nets, net)) {
    passwd = nap_net_get_pass(net);
    if (passwd) nap_net_send(net, NAP_S_PASSWORD, TRUE, "%s", passwd);
  }
  object_set_dirty(OBJ(net));
}

void nap_net_set_nick(nap_net_t* net, const char* nick) {
  g_assert(NapNetModule);
  g_free(net->nick);
  net->nick = (nick && *nick)?g_strdup(nick):NULL;
  object_set_dirty(OBJ(net));
}

GList* nap_get_groups()
{
    g_assert(NapNetModule);
    return NapNetModule->groups;
}

static void nap_start_connect()
{
    GList* dlist;

    g_assert(NapNetModule);
    NapNetModule->auto_connect = TRUE;

    for (dlist = NapNetModule->groups; dlist; dlist = dlist->next)
    {
        nap_ng_t* ng = dlist->data;
        if (ng->flags & NAP_NG_AUTO_CONNECT) nap_ng_start_connect(ng);
    }
    nap_update_connect_status();
}

static void nap_stop_connect(void)
{
    GList* dlist;
    
    g_assert(NapNetModule);
    NapNetModule->auto_connect = FALSE;

    for (dlist = NapNetModule->groups; dlist; dlist = dlist->next)
    {
        nap_ng_t* ng = dlist->data;
        nap_ng_disconnect(ng);
    }
    nap_update_connect_status();
}

void nap_connect_change(unsigned status)
{
    switch (status) {
    case CONNECT_START:
        nap_start_connect();
        break;
    case CONNECT_STOP:
        nap_stop_connect();
        break;
    }
}




nap_net_t* nap_find_net(const char* name) {
  GList* dlist;
  nap_net_t* net;

  g_assert(NapNetModule);
  for (dlist = NapNetModule->groups; dlist; dlist = dlist->next) {
    nap_ng_t* ng = dlist->data;
    net = nap_ng_search_net(ng, name);
    if (net) return net;
  }
  return NULL;
}


static int NapBitrate(int bitrate) {
  int i1;
  const int BitRate[18] = {
    8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 192, 224, 256, 320
  };
  for (i1 = 0; i1 < 18; i1++) {
    if (bitrate <= BitRate[i1])
      return BitRate[i1];
  }
  return 320;
}

static char* NapShareMediaStr(int no) {
  switch (no) {
  case MEDIA_AUDIO:       return "audio";
  case MEDIA_VIDEO:       return "video";
  case MEDIA_APPLICATION: return "application";
  case MEDIA_IMAGE:       return "image";
  case MEDIA_DOCUMENT:    return "text";
  default:                return "application";
  };
}

static void nap_net_share_file(nap_net_t* net, file_t* file) {
  attribute_t* attr1 = attr_find(ATTRC(file), "file", "Bitrate");
  attribute_t* attr2 = attr_find(ATTRC(file), "file", "Frequency");
  attribute_t* attr3 = attr_find(ATTRC(file), "file", "Duration");
  int int1 = attr1?attr_get_ref(attr1, int):24;
  int int2 = attr2?attr_get_ref(attr2, int):16000;
  int int3 = attr3?attr_get_ref(attr3, int):600;

  g_assert(file);
  
  g_assert(NapNetModule);
  // we share all audio with the original command for mp3s
  if (attr1 || attr2 || attr3 || file->media == MEDIA_AUDIO) {
    nap_net_send_unimportant(net, NAP_S_SHARE_FILE, FALSE,
                             "\"%s\" 000000000000000000000000000000 "FF" %d %d %d",
                             file->netname, file->size, NapBitrate(int1), int2, int3);
  } else {
    nap_net_send_unimportant(net, NAP_S_SHARE_GENERIC, FALSE,
                             "\"%s\" "FF" 000000000000000000000000000000 %s",
                             file->netname, file->size, NapShareMediaStr(file->media));
  }
}

static void nap_net_unshare_file(nap_net_t* net, file_t* file) {
  g_assert(NapNetModule);
  if (!file) {
    nap_net_send(net, NAP_S_UNSHARE_ALL, FALSE, "");
  } else {
    nap_net_send(net, NAP_S_UNSHARE_FILE, FALSE, "%s", file->netname);
  }
}


static void _file_shared(file_t* file) {
  if (NapNetModule) nap_net_share_file(NULL, file);
}
static void _file_unshared(file_t* file) {
  if (NapNetModule) nap_net_unshare_file(NULL, file);
}
static void _file_changed(file_t* file) {
  if (NapNetModule) {
    _file_unshared(file);
    _file_shared(file);
  }
}


/* void priv_nap_net_show_searches() { */
/*   GList* dlist; */
/*   for (dlist = NapNetModule->online_nets; dlist; dlist = dlist->next) { */
/*     nap_net_t* net = dlist->data; */
    
/*     chat_print(NULL, "+\0client", */
/*                "+\1c_text", l_printf("q%u a%u %d/%d %d/%d (%u)", */
/*                                      g_list_length(net->queued_searches), */
/*                                      g_list_length(net->active_searches), */
/*                                      net->cur_searches, net->max_searches, */
/*                                      net->searches_pm, net->max_searches_pm, */
/*                                      net->skip_search_end */
/*                                      ), */
/*                "+\1c_client", OBJ(net)->name, */
/*                "+\0nl", NULL); */
/*   } */
/* } */

static void nap_net_shutdown(void)
{
    nap_stop_connect();
}

static void nap_nets_init() {
  lopster_xml_register_home("nap:serverlist", SDIR("nap") "servers.xml",
                            "ServerList", xml_load_groups, xml_save_groups);
  
  pref_register_string("/nap/username", TRUE, NULL);
  pref_add_constraint_replace("/nap/username", ' ', '_');

  pref_register_string("/nap/password", TRUE, NULL);
  pref_add_constraint_replace("/nap/password", ' ', '_');

  pref_register_int("/nap/listen/port", TRUE, 6699);
  pref_add_constraint_port("/nap/listen/port");

  library_register_actions(_file_shared, _file_unshared, _file_changed);
  lopster_call_at_exit(nap_net_shutdown);
}

static void nap_nets_run()
{
    nap_network_load();
    nap_start_listen_service();
    lopster_call_periodic(1000, (CallFunc)nap_net_connect_next, NULL);
    lopster_call_periodic(60000, (CallFunc)nap_net_reset_searches, NULL);
    pref_observe("/nap/charset", (ObjectFunc)nap_charset_observer, NULL);
    
}
static void nap_nets_exit()
{
    //FIXME: free mem
    //  nap_network_save();
    NapNetModule = NULL;
}

static void nap_net_module_init(nap_net_module_t* module) {
  module_init(MODULE(module), "Nap Network Module",
              (ModuleFunc)nap_nets_init,
              (ModuleFunc)nap_nets_run,
              (ModuleFunc)nap_nets_exit);

  module->groups = NULL;
  module->online_nets = NULL;
  module->online_cnt = 0;
  module->resolver = NULL;
  module->auto_connect = FALSE;
  module->listen = NULL;
  module->save = TRUE;
  module->services = NULL;

  module_require(base_module_instance);
  module_require(xml_module_instance);
  module_require(buddy_module_instance);
  module_require(tsocket_module_instance);
  module_require(library_module_instance);
  module_require(search_module_instance);
  module_require(resolve_module_instance);
  module_require(listen_module_instance);
  module_require(download_module_instance);
  //FIXME: add more deps
}


module_t* nap_net_module_instance() {
  if (!NapNetModule) {
    NapNetModule = g_malloc0(sizeof(*NapNetModule));
    nap_net_module_init(NapNetModule);
  }
  return MODULE(NapNetModule);
}

#ifdef WIN32
#define _NAP_NETWORK_C_
#include "napaliasdef.c"
#endif

