/* Copyright (C) 2005-2006 sgop@users.sourceforge.net.
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.7 $
 * $Date: 2006/07/22 17:14:03 $
 * $Author: sid4u $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>

#include "lcore/attribute.h"
#include "lcore/utils.h"
#include "lcore/search.h"
#include "lcore/l_i18n.h"
#include "nap_network.h"
#include "nap_download.h"
#include "nap.h"

typedef struct {
  search_t* search;
  int timer;
} nap_as_t;

typedef struct {
  module_t _module;
  GList* searches;
} nap_search_module_t;

static nap_search_module_t* NapSearchModule = NULL;

static void on_research(search_t* search, nap_net_t* net) {
  nap_net_add_search(net, search);
}

static void nap_search_remove(search_t* search) {
  NapSearchModule->searches = g_list_remove(NapSearchModule->searches, search);
  nap_net_remove_search(NULL, search);
}

static void nap_search_new(search_t* search)
{
    g_assert(NapSearchModule);
    g_assert(search);
    if (!g_list_find(NapSearchModule->searches, search))
    {
        object_notify_d(OBJ(search), (ObjectFunc)nap_search_remove, NULL);
        if (search->download) 
            search_set_continuous(search, 30, (ObjectFunc)on_research, NULL);
        NapSearchModule->searches = g_list_append(NapSearchModule->searches, search);
    }
    nap_net_add_search(NULL, search);
}

void nap_search_joined_net(nap_net_t* net) {
  search_foreach_continuous((ObjectFunc)on_research, net);
}

static char* NapMediaStr(int no)
{
    switch (no) {
    case MEDIA_FOLDER:      return "folder";
    case MEDIA_AUDIO:       return "audio";
    case MEDIA_VIDEO:       return "video";
    case MEDIA_APPLICATION: return "application";
    case MEDIA_IMAGE:       return "image";
    case MEDIA_DOCUMENT:    return "text";
    default:                return "any";
    };
}

#define NAP_ADD_SEARCH_TAG(pattr, tname, type, fmt)                     \
    {                                                                   \
        attribute_t* attr = attr_find(ATTRC(search), "pattern", pattr); \
        if (attr)                                                       \
        {                                                               \
            type* vec = attr_get_ptr(attr, type);                       \
            if (vec[0] == vec[1])                                       \
            {                                                           \
                res = l_printf("%s %s \"EQUAL TO\" "fmt, res, tname, vec[0]); \
            }                                                           \
            else                                                        \
            {                                                           \
                if (vec[0])                                             \
                    res = l_printf("%s %s \"AT LEAST\" "fmt, res, tname, vec[0]); \
                if (vec[1])                                             \
                    res = l_printf("%s %s \"AT BEST\" "fmt, res, tname, vec[1]); \
            }                                                           \
        }                                                               \
    }

const char* nap_search_create_request(search_t* search)
{
    search_pattern_t* pattern = SEARCHP(search);
    char* res;
    
    if (search->download)
    {
        char* str = make_string_from_list(search->download->search_keys, " ");
        res = l_printf("FILENAME CONTAINS \"%s\" MAX_RESULTS 200"
                       " SIZE \"EQUAL TO\" "FF" TYPE any",
                       str, FILE(search->download)->size);
        g_free(str);
    }
    else
    {
        char* str;
        
        if (!pattern->include) return NULL;
        
        str = make_string_from_list(pattern->include, " ");
        res = l_printf("FILENAME CONTAINS \"%s\"", str);
        g_free(str);
        
        if (pattern->exclude)
        {
            str = make_string_from_list(pattern->exclude, " ");
            res = l_printf("%s FILENAME EXCLUDES \"%s\"", res, str);
            g_free(str);
        }
        
        res = l_printf("%s MAX_RESULTS 200", res);
        
        NAP_ADD_SEARCH_TAG("NapSpeed", "LINESPEED", int, "%d");
        NAP_ADD_SEARCH_TAG("Bitrate", "BITRATE", int, "\"%d\"");
        NAP_ADD_SEARCH_TAG("Frequency", "FREQ", int, "\"%d\"");
        NAP_ADD_SEARCH_TAG("Duration", "DURATION", int, "\"%d\"");
        NAP_ADD_SEARCH_TAG("Size", "SIZE", file_size_t, "\""FF"\"");
        
        if (SEARCHP(search)->media != MEDIA_AUDIO)
        {
            res = l_printf("%s TYPE %s", res, NapMediaStr(SEARCHP(search)->media));
        }
    }
    
    return res;
}

#undef NAP_ADD_SEARCH_TAG


const char* print_nspeed_minmax_ui(int* value) {
  const char* str1 = "";
  const char* str2 = "";

  if (value[0] == 0 && value[1] == 0) return "";

  if (value[0]) str1 = _(NapSpeedStrShort(value[0]));
  if (value[1]) str2 = _(NapSpeedStrShort(value[1]));

  if (value[0] == value[1]) {
    return l_printf(_("exactly %s"), str1);
  } else if (*str1 && *str2) {
    return l_printf(_("%s to %s"), str1, str2);
  } else if (*str1) {
    return l_printf(_("at least %s"), str1);
  } else if (*str2) {
    return l_printf(_("at most %s"), str2);
  } else {
    g_assert(FALSE);
    return "";
  }
}

static void nap_searches_init() {
  attr_register("pattern", _("NapSpeed"), ATTR_USEFUL,
                (UIPrintFunc)print_nspeed_minmax_ui,
                (XMLPrintFunc)print_int_minmax,
                (CompareFunc)compare_int_minmax,
                (CopyFunc)copy_int_minmax,
                (FreeFunc)NULL,
                (StoreStrFunc)store_str_int_minmax,
                (StoreFunc)store_int_minmax);

  search_register_action(nap_search_new);
}

static void nap_search_module_init(nap_search_module_t* module) {
  module_init(MODULE(module), "Nap Search Module",
              (ModuleFunc)nap_searches_init, NULL, NULL);

  module->searches = NULL;

  module_require(search_module_instance);
  module_require(nap_net_module_instance);
}


module_t* nap_search_module_instance() {
  if (!NapSearchModule) {
    NapSearchModule = g_malloc0(sizeof(*NapSearchModule));
    nap_search_module_init(NapSearchModule);
  }
  return MODULE(NapSearchModule);
}
