/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.6 $
 * $Date: 2006/07/22 17:14:03 $
 * $Author: sid4u $
 */

#include <string.h>
#include <stdlib.h>

#include "lcore/file_segment.h"
#include "lcore/preferences.h"
#include "nap_upload.h"
#include "nap_proto.h"


static void nap_upload_update_info(nap_upload_t* upload) {
  nap_user_t* user = NAP_USER(UPLOAD(upload)->user);
  if (!user->net) return;
  nap_whois_user(user, user->net, NAP_WHOIS_QUIET, FALSE);
}


static void nap_upload_queue(nap_upload_t* upload, int pos, int slots) {
  nap_user_t* user = NAP_USER(UPLOAD(upload)->user);
  if (!user->net) return;
  nap_net_send(user->net, NAP_S_QUEUE_UPLOAD, FALSE, "%s \"%s\" %d",
               auto_text(nap_net_to_utf8(user->net, OBJ(user)->name)), 
               UPLOAD(upload)->file->netname, pos);
}

static gboolean nap_upload_start(nap_upload_t* upload) {
  nap_user_t* user = NAP_USER(UPLOAD(upload)->user);
  if (!user->net) return FALSE;
  
  nap_upload_update_info(upload);
  
  nap_net_send(user->net, NAP_S_ACCEPT_UPLOAD, FALSE, "%s \"%s\"",
               auto_text(nap_net_to_utf8(user->net, OBJ(user)->name)), 
               UPLOAD(upload)->file->netname);
  return TRUE;
}

static void nap_upload_stopped(nap_upload_t* upload) {
  if (upload->started) {
    upload->started = FALSE;
    nap_net_send(NULL, NAP_S_UPLOAD_STOPPED, FALSE, "");
  }
}

static gboolean nap_upload_end(nap_upload_t* upload, const char* message) {
  // keep the connection if segment finished (9)
  // and this is an advanced upload connection
  if (upload->advanced && 
      socket_get_status(SOCK(upload)) == ST_UL_SEG_FIN) {
    socket_set_status(SOCK(upload), ST_UNKNOWN);
    return FALSE;
  } else {
    nap_upload_stopped(upload);
    return upload_end(UPLOAD(upload), message);
  }
}

static void nap_upload_destroy(nap_upload_t* upload) {
  upload_destroy(UPLOAD(upload));
}

static void
nap_upload_init(nap_upload_t* upload, file_t* file, nap_user_t* user) {
  upload_init(UPLOAD(upload), file, USER(user));
  object_add_type(OBJ(upload), "NapUpload");
  OBJ(upload)->destroy = (ObjectFunc)nap_upload_destroy;
  upload->advanced = FALSE;
  upload->started = FALSE;
  socket_set_end(SOCK(upload), (SockEndFunc)nap_upload_end);

  UPLOAD(upload)->queue = (UploadQueueFunc)nap_upload_queue;
  UPLOAD(upload)->start = (UploadStartFunc)nap_upload_start;
  
  nap_whois_user(user, user->net, NAP_WHOIS_QUIET, FALSE);
  nap_user_stat_cnt(user, NS_UP);

}

nap_upload_t* nap_upload_new(nap_user_t* user, file_t* file) {
  nap_upload_t* upload = g_malloc0(sizeof(*upload));
  nap_upload_init(upload, file, user);
  return upload;
}

static const char* nap_upload_get_advanced(nap_upload_t* upload) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(upload));
  guint32 avail = buffer_avail(buffer);
  char* data;
  guint32 len;
  file_size_t val;
  char save;
  segment_t* segment = UPLOAD(upload)->cur_segment;

  data = buffer_read_pos(buffer);
  if (avail < 4) return NULL;
  len = unpack_u32(&data, &avail);

  if (avail < len) return NULL;

  save = data[len];
  data[len] = 0;
  g_log("NAP_UPLOAD", G_LOG_LEVEL_DEBUG, "received: %u :'%s'", len, data);

  if (!strncmp(data, "STOP", 4)) {
    upload->advanced = TRUE;
    val = strtoll(data+5, NULL, 10);
    if (!upload_change_segment_stop(UPLOAD(upload), val))
      return "failed to change segment stop";
  } else if (!strncmp(data, "START", 5)) {
    const char* res;
    if (segment) return "error: active segment";
    if (!upload->advanced) return "START received but no advanced connection";
    val = strtoll(data+5, NULL, 10);
    
    upload_add_segment
      (UPLOAD(upload), val, UPLOAD(upload)->file->size, FALSE);

    tcp_socket_send_str(TSOCK(upload), "CONTINUED"); //could be done with upload->header()
    res = upload_prepare_segment(UPLOAD(upload));
    if (res) return res;
  } else if (!strncmp(data, "FINISHED", 8)) {
    if (!upload->advanced) return "FINISHED received but no advanced connection";
    return "finished";
  } else {
    return "invalid token received";
  }

  data[len] = save;
  buffer_consume(buffer, 4+len);
  if (buffer_avail(buffer)) return "AGAIN";
  else return NULL;
}

static void nap_upload_started(nap_upload_t* upload) {
  b_socket_set_consume(BSOCK(upload), (BConsumeInputFunc)nap_upload_get_advanced);
  b_socket_start_sending(BSOCK(upload));
  g_assert(!upload->started);
  upload->started = TRUE;
  nap_net_send(NULL, NAP_S_UPLOAD_STARTED, FALSE, "");
}

static const char* nap_upload_get_info_in(tcp_socket_t* socket) {
  char* nick;
  char* winname;
  char* progress;
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(socket));
  char* temp;
  upload_t* upload;
  const char* res;
  nap_user_t* user;

  if (buffer_avail(buffer) == 0) return NULL;

  // this function (like many other napster consumer) assume,
  // that the buffer is 0 terminated. The tcp_socket appends
  // a 0 whenever it reads from the socket :)
  nick = arg(buffer_read_pos(buffer), 0);
  winname = arg(NULL, 0);
  progress = arg(NULL, 0);
  
  if (!progress) return NULL;
  
  g_log("NAP_UPLOAD", G_LOG_LEVEL_DEBUG,
        "'%s' requested file '%s' at pos '%s'",
        nick, winname, progress);

  user = nap_user_find(nick, NULL); //FIXME: should we convert the nickname?
  user_set_ip(USER(user), socket->ip);

  upload = upload_find(USER(user), auto_text(nap_net_to_utf8(user->net, winname)));

  if (!upload) {
    socket_set_status(SOCK(socket), ST_ERROR);
    return "upload not found";
  }

  if (upload->status != UPLOAD_STARTED) {
    return "upload not started";
    // FIXME
  }

  tcp_socket_move(TSOCK(upload), socket, TRUE);
  
  temp = l_printf("%llu", upload->file->size);
  tcp_socket_send_str(TSOCK(upload), temp);
  
  upload_add_segment(upload, strtoll(progress, NULL, 10),
                     upload->file->size, FALSE);
  
  res = upload_prepare_segment(upload);
  
  if (res) {
    socket_done(SOCK(upload), ST_ERROR, res);
    return "bsocket->nap_upload(failed)";
  } else {
    nap_upload_started(NAP_UPLOAD(upload));
    return "bsocket->nap_upload";
  }
}

const char* nap_upload_incoming(tcp_socket_t* s) {
 
  b_socket_set_consume(BSOCK(s), (BConsumeInputFunc)nap_upload_get_info_in);
  // allow handshake with max 1024 chars
  b_socket_set_limit(BSOCK(s), 1024);

  if (buffer_avail(b_socket_get_in_buffer(BSOCK(s))))
    return "AGAIN";
  else
    return NULL;
}

static const char* nap_upload_get_info_out(nap_upload_t* upload) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(upload));
  char* pos;
  const char* res;

  if (buffer_avail(buffer) < 1) return "invalid input";
  pos = buffer_read_pos(buffer);

  g_log("NAP_UPLOAD", G_LOG_LEVEL_DEBUG, "got progress '%s'", pos);

  buffer_clear(buffer);

  upload_add_segment(UPLOAD(upload), strtoll(pos, NULL, 10),
                     UPLOAD(upload)->file->size, FALSE);

  res = upload_prepare_segment(UPLOAD(upload));
  
  if (res) {
    socket_done(SOCK(upload), ST_ERROR, res);
    return "failed to start upload";
  } else {
    nap_upload_started(upload);
    return NULL;
  }
}

static const char* nap_upload_accept(nap_upload_t* upload) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(upload));
  char* temp;
  nap_user_t* user = NAP_USER(UPLOAD(upload)->user);

  if (buffer_avail(buffer) < 1) return "invalid input";
  if (*(buffer_read_pos(buffer)) != '1') return "input != '1'";
  buffer_consume_virt(buffer, 1);

  g_log("NAP_UPLOAD", G_LOG_LEVEL_DEBUG, "got incoming '1'");

  temp = l_printf("SEND%s \"%s\" "FF, nap_net_get_nick(user->net),
                  UPLOAD(upload)->file->netname, UPLOAD(upload)->file->size);
  
  g_log("NAP_UPLOAD", G_LOG_LEVEL_DEBUG, "sending [%s]", temp);
  tcp_socket_send_str(TSOCK(upload), temp);

  b_socket_set_consume
    (BSOCK(upload), (BConsumeInputFunc)nap_upload_get_info_out);

  if (buffer_avail(buffer)) return "AGAIN";
  else return NULL;
}

static void nap_upload_connect_and_start(nap_upload_t* upload) {
  b_socket_set_consume
    (BSOCK(upload), (BConsumeInputFunc)nap_upload_accept);
  if (socket_activate(SOCK(upload)))
    b_socket_start_receiving(BSOCK(upload), 1024);
}

void nap_upload_outgoing(nap_upload_t* upload, ip_t ip, port_t port) {
  user_set_ip_port(UPLOAD(upload)->user, ip, port);
  TSOCK(upload)->ip = ip;
  TSOCK(upload)->port = port;
  nap_upload_connect_and_start(upload);
}
