/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.8 $
 * $Date: 2006/07/22 17:14:03 $
 * $Author: sid4u $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>

#include "lcore/lopster.h"
#include "nap_users.h"
#include "nap.h"
#include "nap_proto.h"

#ifdef WIN32
#include "napalias.h"
#endif

typedef struct {
  module_t _module;
  GHashTable* users;
} nap_user_module_t;

static nap_user_module_t* NapUserModule = NULL;

static void on_nap_user_net_removed(nap_net_t* net, nap_user_t* user)
{
    (void)net;
    user->net = NULL;
}

static void nap_user_destroy(nap_user_t* user)
{
    // we could assert here if the user is not found in hash table.
    g_hash_table_remove(NapUserModule->users, OBJ(user)->name);
    
    g_list_free(user->nets);
    user_destroy(USER(user));
}

static void nap_user_monitor(nap_user_t* user)
{
    nap_net_send(NULL, NAP_S_HOT_ADD, TRUE, "%s", OBJ(user)->name);
}

static void nap_user_unmonitor(nap_user_t* user)
{
    nap_net_send(NULL, NAP_S_HOT_REMOVE, TRUE, "%s", OBJ(user)->name);
}

static void nap_user_init(nap_user_t* user, const char* nick)
{
    user_init(USER(user), nick, NAP_ID);
    object_add_type(OBJ(user), "NapUser");
    OBJ(user)->destroy = (ObjectFunc)nap_user_destroy;
    USER(user)->monitor = (UserMonitor)nap_user_monitor;
    USER(user)->unmonitor = (UserMonitor)nap_user_unmonitor;
    USER(user)->ip = 0;
    user->net = NULL;
    user->speed = 0;
    user->shared = 0;
    user->client_id = NAP_CLIENT_UNKNOWN;
    user->client = NULL;
    user->whois = NULL;
}

static nap_user_t* nap_user_new(const char* nick)
{
    nap_user_t* user = g_malloc0(sizeof(*user));
    nap_user_init(user, nick);
    return user;
}

void nap_user_set_net(nap_user_t* user, nap_net_t* net)
{
    if (user->net == net) return;
    if (user->net)
        object_unnotify_o(OBJ(user->net), (ObjectFunc)on_nap_user_net_removed, OBJ(user));
    user->net = net;
    if (user->net)
        object_notify_o(OBJ(user->net), (ObjectFunc)on_nap_user_net_removed, OBJ(user));
}

void nap_user_setup(nap_user_t* user, ip_t ip, int speed)
{
    g_assert(NapUserModule);
    user->speed = speed;
    if (ip) user_set_ip(user, ip);
}

nap_user_t* nap_user_find(const char* nick, nap_net_t* net)
{
    nap_user_t* user = g_hash_table_lookup(NapUserModule->users, nick);
    g_assert(NapUserModule);
    if (!user)
    {
        user = nap_user_new(nick);
        g_hash_table_insert(NapUserModule->users, OBJ(user)->name, user);
    }
    if (net) nap_user_set_net(user, net);
    return user;
}

void nap_user_set_client(nap_user_t* user, const char* client)
{
    g_free(user->client);
    user->client = g_strdup(client);
    if (!user->client)
    {
        user->client_id = NAP_CLIENT_UNKNOWN;
    }
    else if (!strncmp(user->client, "WinMX v2.6", 10))
    {
        user->client_id = NAP_CLIENT_WINMX_26;
    }
    else if (!strncmp(user->client, "Lopster", 7))
    {
        int v1=0, v2=0, v3=0, v4=0;
        if (sscanf(user->client+7, " %d.%d.%d-dev%d", 
                   &v1, &v2, &v3, &v4) < 2)
        {
            user->client_id = NAP_CLIENT_UNKNOWN;
        }
        else if ((v1 > 1) ||
                 (v1 == 1 && v2 > 0) ||
                 (v1 == 1 && v2 == 0 && v3 > 1) ||
                 (v1 == 1 && v2 == 0 && v3 == 1 && v4 >= 6)) {
            user->client_id = NAP_CLIENT_LOPSTER_ADV;
        }
        else
        {
            user->client_id = NAP_CLIENT_UNKNOWN;
        }
    }
    else
    {
        user->client_id = NAP_CLIENT_UNKNOWN;
    }
}

static void nap_user_action(nap_user_t* user, guint16 action, nap_net_t* net,
                            const char* reason, unsigned timeout)
{
  if (reason && *reason)
  {
      //    convert_quotes();
      if (timeout > 0) 
          nap_net_send(net, action, TRUE, "%s \"%s\" %u",
                       OBJ(user)->name, reason, timeout);
      else 
          nap_net_send(net, action, TRUE,
                       "%s \"%s\"", OBJ(user)->name, reason);
  }
  else
  {
      nap_net_send(net, action, TRUE, "%s", OBJ(user)->name);
  }
}

void nap_user_muzzle(nap_user_t* user, nap_net_t* net, const char* reason)
{
    nap_user_action(user, NAP_S_MUZZLE, net, reason, 0);
}

void nap_user_unmuzzle(nap_user_t* user, nap_net_t* net, const char* reason)
{
    nap_user_action(user, NAP_S_UNMUZZLE, net, reason, 0);
}

void nap_user_kill(nap_user_t* user, nap_net_t* net, const char* reason)
{
    nap_user_action(user, NAP_S_KILL, net, reason, 0);
}

void nap_user_ban(nap_user_t* user, nap_net_t* net, const char* reason, unsigned timeout)
{
    nap_user_action(user, NAP_S_BAN, net, reason, timeout);
}

void nap_user_unban(nap_user_t* user, nap_net_t* net, const char* reason)
{
    nap_user_action(user, NAP_S_UNBAN, net, reason, 0);
}

void nap_user_nuke(nap_user_t* user, nap_net_t* net)
{
    nap_user_action(user, NAP_S_NUKE, net, NULL, 0);
}

static void nap_users_init()
{
    NapUserModule->users =
        g_hash_table_new_full((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal, NULL, NULL);
}

static void nap_users_exit()
{
    g_log("NAP_USERS", G_LOG_LEVEL_DEBUG,
          "we have %u in cache consuming approx. %u bytes", 
          g_hash_table_size(NapUserModule->users),
          (int)(g_hash_table_size(NapUserModule->users)*sizeof(nap_user_t)));
    
    g_hash_table_destroy(NapUserModule->users);
    NapUserModule->users = NULL;
    NapUserModule = NULL;
}

static void nap_user_module_init(nap_user_module_t* module)
{
    module_init(MODULE(module), "Napster User Module",
                (ModuleFunc)nap_users_init, NULL, (ModuleFunc)nap_users_exit);
    
    module->users = NULL;
    
    module_require(nap_net_module_instance);
    module_require(user_module_instance);
}


module_t* nap_user_module_instance()
{
    if (!NapUserModule)
    {
        NapUserModule = g_malloc0(sizeof(*NapUserModule));
        nap_user_module_init(NapUserModule);
    }
    return MODULE(NapUserModule);
}

#ifdef WIN32
#define _NAP_USERS_C_
#include "napaliasdef.c"
#endif
