/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.2 $
 * $Date: 2007/05/23 20:42:29 $
 * $Author: sid4u $
 */

#include <string.h>

#include <glib.h>

#include "lcore/string_list.h"
#include "lcore/preferences.h"
#include "lcore/ui.h"
#include "lcore/action_item.h"
#include "lcore/completion.h"
#include "lcore/l_base.h"
#include "lcore/l_i18n.h"
#include "lcore/xml_handler.h"
#include "slsk.h"
#include "slsk_chat.h"
#include "slsk_proto.h"
#include "slsk_ui.h"
#include "slsk_til.h"
#include "slsk_hotlist.h"

#ifdef WIN32
#include "slskalias.h"
#endif

typedef struct {
  module_t _module;
  GList* pages;
  GList* channels;
} slsk_chat_module_t;

typedef struct {
  char* channel;
  char* charset;
} slsk_cp_enc_t;

static slsk_chat_module_t* SlskChatModule = NULL;


static void slsk_channels_save()
{
    if (module_running(MODULE(SlskChatModule)))
        lopster_xml_save("slsk:channel:charset", 10000);
}

static slsk_cp_enc_t* slsk_channel_find_enc(const char* channel) {
  GList* dlist;

  for (dlist = SlskChatModule->channels; dlist; dlist = dlist->next) {
    slsk_cp_enc_t* enc = dlist->data;
    if (!strcmp(enc->channel, channel)) return enc;
  }
  return NULL;
}


static void slsk_cp_set_charset(slsk_cp_t* page, const char* charset) {
  slsk_cp_enc_t* enc;

  g_free(page->charset);
  page->charset = (charset&&*charset)?g_strdup(charset):NULL;
  page->charset_warned = FALSE;
  
  if (page->type != SLSK_PAGE_PUBLIC) return;
  
  enc = slsk_channel_find_enc(OBJ(page)->name);
  if (enc) {
    if (page->charset) {
      g_free(enc->charset);
      enc->charset = g_strdup(page->charset);
    } else {
      SlskChatModule->channels = g_list_remove(SlskChatModule->channels, enc);
      g_free(enc->channel);
      g_free(enc->charset);
      g_free(enc);
    }
  } else {
    if (page->charset) {
      enc = g_malloc(sizeof(enc));
      enc->channel = g_strdup(OBJ(page)->name);
      enc->charset = g_strdup(page->charset);
      SlskChatModule->channels = g_list_prepend(SlskChatModule->channels, enc);
    } else {
      // do nothing
    }
  }
  slsk_channels_save();
}




slsk_cp_t* slsk_cp_search_net(slsk_net_t* net, const char* channel, int type) {
  GList* dlist;
  
  for (dlist = SlskChatModule->pages; dlist; dlist = dlist->next) {
    slsk_cp_t* scp = dlist->data;
    if (scp->type == type && scp->net == net &&
        !strcmp(OBJ(scp)->name, channel)) return scp;
  }
  return NULL;
}

slsk_cp_t* slsk_cp_search(const char* channel, int type) {
  GList* dlist;
  
  for (dlist = SlskChatModule->pages; dlist; dlist = dlist->next) {
    slsk_cp_t* scp = dlist->data;
    if (scp->type == type &&
	!strcmp(OBJ(scp)->name, channel)) return scp;
  }
  return NULL;
}

static gboolean slsk_cp_closing(slsk_cp_t* page) {
  if (page->type != SLSK_PAGE_PUBLIC) return TRUE;
  if (!PAGE(page)->active) return TRUE;

  page->leaving = TRUE;
  chat_page_set_active(PAGE(page), FALSE);
  slsk_net_send(page->net, SLSK_S_PART_CHANNEL, OBJ(page)->name);
  return FALSE;
}

static void slsk_clear_users(slsk_cp_t* scp)
{
    g_list_foreach(scp->users, (GFunc)user_unref, NULL);
    g_list_free(scp->users);
    scp->users = NULL;
}

static void slsk_clear_ticker(slsk_cp_t* scp) {
  g_list_foreach(scp->ticker, (GFunc)object_kill, NULL);
  g_list_free(scp->ticker);
  scp->ticker = NULL;
}

static void slsk_cp_destroy(slsk_cp_t* page)
{
    if (page->type == SLSK_PAGE_PRIVATE)
    {
        user_stop_monitor(USER(slsk_user_find(OBJ(page)->name)));
    }
    else if (page->type == SLSK_PAGE_PUBLIC)
    {
        slsk_clear_users(page);
        slsk_clear_ticker(page);
    }
    SlskChatModule->pages = g_list_remove(SlskChatModule->pages, page);
    chat_page_destroy(PAGE(page));
}


static void slsk_cp_init(slsk_cp_t* scp, slsk_net_t* net, const char* name,
                         const char* label, const char* title, int type)
{
    chat_page_init(PAGE(scp), name, label, title, SLSK_ID);
    object_add_type(OBJ(scp), "SlskChatPage");
    OBJ(scp)->destroy = (ObjectFunc)slsk_cp_destroy;
    PAGE(scp)->close = (ChatCloseFunc)slsk_cp_closing;
    scp->net = net;
    scp->type = type;
    SlskChatModule->pages = g_list_append(SlskChatModule->pages, scp);
}

static slsk_cp_t* slsk_cp_new(slsk_net_t* net, const char* name, const char* label,
                              const char* title, int type)
{
    slsk_cp_t* scp = g_malloc0(sizeof(*scp));
    slsk_cp_init(scp, net, name, label, title, type);
    return scp;
}
  
static void on_net_status_changed_priv(slsk_net_t* net, slsk_cp_t* scp)
{
    gboolean online = (net->status == SLSK_NET_ONLINE);

    if (PAGE(scp)->active == online) return;
    chat_page_set_active(PAGE(scp), online);

    if (online)
        user_start_monitor(USER(slsk_user_find(OBJ(scp)->name)));
    else
        user_stop_monitor(USER(slsk_user_find(OBJ(scp)->name)));
}

slsk_cp_t* slsk_cp_new_private(slsk_net_t* net, const char* nick)
{
    slsk_cp_t* scp = slsk_cp_new(net, nick, nick,
                                 l_printf(_("Private conversation with %s"), nick),
                                 SLSK_PAGE_PRIVATE);
    SLSKUI.create_private_page(scp);
    
    object_monitor_o(OBJ(net), (ObjectFunc)on_net_status_changed_priv, OBJ(scp));
    user_start_monitor(USER(slsk_user_find(nick)));
    
    return scp;
}

static char* slsk_cp_to_utf8(slsk_cp_t* page, const char* message) {
  GError* error = NULL;
  char* enc = NULL;

  if (!message) return NULL;

  enc = to_utf8(message, -1, NULL, &error, 5,
                page->charset, "utf-8", slsk_charset_get(), "latin1", "windows-1252");
  if (error) {
    if (!page->charset_warned) {
      page->charset_warned = TRUE;
      lopster_notify(LOPSTER_MESSAGE_WARNING, error->message,
                     g_markup_printf_escaped
                     (_("Message from channel <b>%s</b> seems not to have the "
                        "right encoding. Please check the character set encoding "
                        "for that channel."),
                      OBJ(page)->name)
                     );
    }
    g_error_free(error);
  }
  return enc;
}

void slsk_cp_set_encoding(slsk_cp_t* page, const char* charset) {
  GList* dlist;

  slsk_cp_set_charset(page, charset);

  for (dlist = page->ticker; dlist; dlist = dlist->next) {
    slsk_ct_t* ticker = dlist->data;
    g_free(ticker->message);
    ticker->message = slsk_cp_to_utf8(page, ticker->orig_message);
    object_set_dirty(OBJ(ticker));
  }
}

static char* slsk_user_name(slsk_user_t* cu) {
  return OBJ(cu)->name;
}

static char* slsk_pub_expand(slsk_cp_t* page, const char* text) {
  const char* pos;
  GList* results;
  char* result;
  int offset;
  completion_t* c;

  c = completion_new((CompletionFunc)slsk_user_name);
  completion_set_compare(c, (GCompletionStrncmpFunc)strncasecmp);
  completion_set_items(c, &page->users);
  
  if (!text) {
    results = completion_complete(c, "", &result);
    offset = 0;
    pos = NULL;
  } else {
    pos = strrchr(text, ' ');
    if (!pos) pos = text;
    else pos++;

    if (*pos == '@') {
      results = completion_complete(c, pos+1, &result);
      offset = (pos-text)+1;
    } else {
      results = completion_complete(c, pos, &result);
      offset = (pos-text);
    }
  }

  if (g_list_length(results) > 1) {
    GList* dlist;
    chat_print(PAGE(page), "+\0client",
	       "+\1c_client", l_printf("%d matching users found", 
				       g_list_length(results)),
	       "+\0nl", NULL);
    for (dlist = results; dlist; dlist = dlist->next) {
      slsk_user_t* cu = dlist->data;
      chat_print(PAGE(page), "+\0client",
		 "+\1c_text", "  ",
		 "+\1nick", OBJ(cu)->name, "\0\0",
		 "+\0nl", NULL);
    }
    if (result) {
      char* t = result;
      result = g_malloc(offset+strlen(t)+1);
      strncpy(result, text, offset);
      sprintf(result+offset, "%s", t);
      g_free(t);
    }
  } else if (g_list_length(results) == 1) {
    slsk_user_t* cu = results->data;
    g_free(result);
    result = g_malloc(offset+strlen(OBJ(cu)->name)+3);
    strncpy(result, text, offset);
    if (offset < 2) 
      sprintf(result+offset, "%s: ", OBJ(cu)->name);
    else
      sprintf(result+offset, "%s ", OBJ(cu)->name);
  } else {
    chat_print(PAGE(page), "+\0client",
	       "+\1c_client", "No matching user found",
	       "+\0nl", NULL);
  }
  completion_free(c);
  return result;
}

static void on_net_status_changed_pub(slsk_net_t* net, slsk_cp_t* scp)
{
    if (net->status == SLSK_NET_ONLINE)
    {
        if (PAGE(scp)->active) return;
        slsk_net_send(net, SLSK_S_JOIN_CHANNEL, OBJ(scp)->name);
        //    chat_page_set_active(PAGE(scp), TRUE);
    }
    else
    {
        if (!PAGE(scp)->active) return;
        chat_page_set_active(PAGE(scp), FALSE);
    }
}

slsk_cp_t* slsk_cp_new_public(slsk_net_t* net, const char* channel)
{
    slsk_cp_enc_t* enc;
    slsk_cp_t* scp = slsk_cp_new(net, channel, channel, 
                                 l_printf("channel %s", channel),
                                 SLSK_PAGE_PUBLIC);

    PAGE(scp)->active = FALSE;
    PAGE(scp)->expand = (ChatExpandFunc)slsk_pub_expand;
    
    enc = slsk_channel_find_enc(channel);
    if (enc) scp->charset = g_strdup(enc->charset);
    
    SLSKUI.create_public_page(scp);
    
    object_monitor_o(OBJ(net), (ObjectFunc)on_net_status_changed_pub, OBJ(scp));
    return scp;
}


void slsk_cp_update_net(slsk_cp_t* scp, slsk_net_t* net) {
  if (!scp || scp->net == net) return;
  scp->net = net;
}

static GList* slsk_cp_user_search(slsk_cp_t* scp, const char* user) {
  GList* dlist;

  g_assert(scp && user);
  for (dlist = scp->users; dlist; dlist = dlist->next) {
    slsk_user_t* cu = dlist->data;
    if (!strcmp(OBJ(cu)->name, user)) return dlist;
  }
  return NULL;
}

void slsk_public_message(slsk_net_t* net, char* channel, char* user, char* message) {
  slsk_cp_t* page = slsk_cp_search_net(net, channel, SLSK_PAGE_PUBLIC);

  if (page) {
    char* enc = slsk_cp_to_utf8(page, message);
    if (enc) user_message(PAGE(page), NULL, NULL, user, enc);
    g_free(enc);
  }
}

void slsk_private_message(slsk_net_t* net, char* nick, char* message, int timestamp)
{
    slsk_cp_t* page;
    char* enc;
    
    (void)timestamp;
    
    //FIXME: convert message to utf8!
    
    if ((page = slsk_cp_search(nick, SLSK_PAGE_PRIVATE)))
    {
        // search for an existing priv page and force pipe if found
        slsk_cp_update_net(page, net);
    }
    else if ((pref_get_ref("/core/chat/pipe", int) & PIPE_PRIVATE))
    {
        // create new priv page if we want to pipe
        page = slsk_cp_new_private(net, nick);
    }
    else
    {
        page = NULL;
    }

    if (page)
    {
        enc = slsk_cp_to_utf8(page, message);
        if (enc) user_message(PAGE(page), NULL, NULL, nick, enc);
        g_free(enc);
    }
    else
    {
        enc = to_utf8(message, -1, NULL, NULL, 4,
                      "utf-8", slsk_charset_get(), "latin1", "windows-1252");
        if (enc) private_message(NULL, "SLSK", NULL, nick, enc);
        g_free(enc);
    }
}

void slsk_emote_message(slsk_net_t* net, char* channel, char* user, char* message)
{
    slsk_cp_t* page = slsk_cp_search_net(net, channel, SLSK_PAGE_PUBLIC);
    
    if (page)
    {
        char* enc = slsk_cp_to_utf8(page, message);
        if (enc) emote_message(PAGE(page), NULL, NULL, user, enc);
        g_free(enc);
    }
}

void slsk_global_message(slsk_net_t* net, char* message)
{
    char* enc = to_utf8(message, -1, NULL, NULL, 4,
                        "utf-8", slsk_charset_get(), "latin1", "windows-1252");
    global_message("SoulSeek", NULL, enc);
    g_free(enc);
    (void)net;
}


void slsk_user_joined(slsk_cp_t* scp, slsk_user_t* cu, gboolean show)
{
    g_assert(scp && cu);
    if (g_list_find(scp->users, cu)) return;
    scp->users = g_list_append(scp->users, user_ref(USER(cu)));
    SLSKUI.user_show(scp, cu);
    if (show) SLSKUI.user_joined(scp, cu);
}

void slsk_user_parted(slsk_cp_t* scp, const char* user)
{
    GList* dlist;
    slsk_user_t* cu;
    
    g_assert(scp && user);
    dlist = slsk_cp_user_search(scp, user);
    if (dlist)
    {
        cu = dlist->data;
        SLSKUI.user_parted(scp, cu);
        scp->users = g_list_delete_link(scp->users, dlist);
        user_unref(USER(cu));
    }
}

static void slsk_ticker_destroy(slsk_ct_t* ct)
{
    g_free(ct->message);
    g_free(ct->orig_message);
    object_destroy(OBJ(ct));
}

static void slsk_ticker_init(slsk_ct_t* ticker, slsk_cp_t* scp, 
                             const char* nick, const char* message)
{
    object_init(OBJ(ticker), nick, TRUE);
    object_add_type(OBJ(ticker), "SlskTicker");
    OBJ(ticker)->destroy = (ObjectFunc)slsk_ticker_destroy;
    ticker->orig_message = g_strdup(message);
    ticker->message = slsk_cp_to_utf8(scp, message);
}

void slsk_ticker_new(slsk_cp_t* scp, const char* nick, const char* message)
{
    slsk_ct_t* ticker = g_malloc0(sizeof(*ticker));
    slsk_ticker_init(ticker, scp, nick, message);
    
    scp->ticker = g_list_append(scp->ticker, ticker);
    SLSKUI.ticker_show(scp, ticker);
}

static GList* slsk_ticker_search(slsk_cp_t* scp, const char* user) {
  GList* dlist;

  g_assert(scp && user);
  for (dlist = scp->ticker; dlist; dlist = dlist->next) {
    slsk_ct_t* ct = dlist->data;
    if (!strcmp(OBJ(ct)->name, user)) return dlist;
  }
  return NULL;
}

void slsk_ticker_remove(slsk_cp_t* scp, const char* user) {
  GList* dlist;
  
  g_assert(scp && user);
  dlist = slsk_ticker_search(scp, user);
  if (dlist) {
    slsk_ct_t* ticker = dlist->data;
    scp->ticker = g_list_delete_link(scp->ticker, dlist);
    object_kill(OBJ(ticker));
  }
}

void slsk_set_ticker(slsk_cp_t* scp, const char* message) {
  g_assert(scp);
  slsk_net_send(scp->net, SLSK_S_SET_TICKER,
		OBJ(scp)->name, message?message:"");
}

void slsk_join_channel(slsk_net_t* net, const char* name) {
  slsk_cp_t* scp;

  scp = slsk_cp_search_net(net, name, SLSK_PAGE_PUBLIC);
  if (scp && PAGE(scp)->active) return;

  if (!scp) scp = slsk_cp_new_public(net, name);
  slsk_net_send(net, SLSK_S_JOIN_CHANNEL, name);
}

void slsk_channel_joined(slsk_cp_t* scp) {
  chat_page_set_active(PAGE(scp), TRUE);
  scp->leaving = FALSE;
  slsk_clear_users(scp);
  SLSKUI.channel_joined(scp);
}

void slsk_channel_parted(slsk_cp_t* scp) {
  if (!scp) return;

  if (scp->leaving) {
    object_kill(OBJ(scp));
  } else {
    chat_page_set_active(PAGE(scp), FALSE);
    // we didnt request the part!
    if (pref_get_ref("/core/chat/auto_rejoin", gboolean))
      slsk_net_send(scp->net, SLSK_S_JOIN_CHANNEL, OBJ(scp)->name);
  }
}

void slsk_send_private_message(slsk_net_t* net, char* nick, char *message) {
  chat_page_t* page;
  char* pos;
  int private;

  if (!net) return;

  page = chat_page_get_current();
  
  if (page && page->ID == SLSK_ID && 
      SCP(page)->type == SLSK_PAGE_PRIVATE &&
      !strcasecmp(OBJ(page)->name, nick)) {
    private = 1;
    message = auto_text
      (from_utf8(message, -1, NULL, NULL, 3, SCP(page)->charset, slsk_charset_get(), "utf-8"));
  } else {
    message = auto_text
      (from_utf8(message, -1, NULL, NULL, 2, slsk_charset_get(), "utf-8"));
    private = 0;
  }
  
  
  line_insert_lf(message, LFP);
  do {
    pos = strchr(message, '\n');
    if (pos) *pos = 0;
    slsk_net_send(net, SLSK_S_PRIVMSG, nick, message);
    if (private) user_message(page, NULL, NULL, net->nick, message);
    else  private_message(page, "to", nick, net->nick, message);
    if (!pos) break;
    message = pos + 1;
  } while (*message);

}

void slsk_send_public_message(slsk_cp_t* scp, char* message) {
  char *pos;

  if (!scp || !scp->net) return;
  if (scp->type != SLSK_PAGE_PUBLIC) return;
  if (!PAGE(scp)->active) return;

  message = auto_text
    (from_utf8(message, -1, NULL, NULL, 3, scp->charset, slsk_charset_get(), "utf-8"));

  line_insert_lf(message, LFP);
  do {
    pos = strchr(message, '\n');
    if (pos) *pos = 0;
    slsk_net_send(scp->net, SLSK_S_PUBLIC, OBJ(scp)->name, message);
    if (!pos) break;
    message = pos + 1;
  } while (*message);

}

/*

void slsk_send_emote_message(slsk_cp_t* scp, char *text) {
char *text2;
  char *pos;

  if (!scp) return;
  if (scp->type != PAGE_PUBLIC) return;
  if (!scp->net) return;
  if (!PAGE(scp)->active) return;

  slsk_net_send(scp->net, SLSK_S_EMOTE, PAGE(scp)->name, text);
}

 */

static void xml_load_channels(xmlDocPtr xfile, xmlNodePtr xroot) {
  xmlNodePtr xsub = xroot->xmlChildrenNode;
  (void)xfile;
  while (xsub != NULL) {
    if (!strcmp(xsub->name, "Channel")) {
      xmlChar* attr1 = xmlGetProp(xsub, "Name");
      xmlChar* attr2 = xmlGetProp(xsub, "Charset");

      if (attr1 && attr2) {
        slsk_cp_enc_t* enc = g_malloc(sizeof(*enc));
        enc->channel = g_strdup(attr1);
        enc->charset = g_strdup(attr2);
        SlskChatModule->channels = g_list_prepend(SlskChatModule->channels, enc);
      }
      if (attr1) xmlFree(attr1);
      if (attr2) xmlFree(attr2);
    }
    xsub = xsub->next;
  }
}

static void xml_save_channel(slsk_cp_enc_t* enc, xmlNodePtr xroot) {
  xmlNodePtr xsub = xmlNewChild(xroot, NULL, "Channel", NULL);
  xmlNewProp(xsub, "Name", enc->channel);
  xmlNewProp(xsub, "Charset", enc->charset);
}

static void xml_save_channels(xmlDocPtr xfile, xmlNodePtr xroot) {
  (void)xfile;
  g_list_foreach(SlskChatModule->channels, (GFunc)xml_save_channel, xroot);
}



static void slsk_chat_init() {
  lopster_xml_register_home("slsk:channel:charset", SDIR("slsk") "channel.xml",
                            "ChannelCharset", xml_load_channels, xml_save_channels);
}

static void slsk_chat_run() {
  lopster_xml_load("slsk:channel:charset");
}

static void slsk_chat_done() {
  //FIXME: free data
  SlskChatModule = NULL;
}

static void slsk_chat_module_init(slsk_chat_module_t* module) {
  module_init(MODULE(module), "Slsk Netlist Module",
              (ModuleFunc)slsk_chat_init, (ModuleFunc)slsk_chat_run,
              (ModuleFunc)slsk_chat_done);

  module->pages = NULL;
  module->channels = NULL;

  module_require(xml_module_instance);
}


module_t* slsk_chat_module_instance() {
  if (!SlskChatModule) {
    SlskChatModule = g_malloc0(sizeof(*SlskChatModule));
    slsk_chat_module_init(SlskChatModule);
  }
  return MODULE(SlskChatModule);
}

#ifdef WIN32
#define _SLSK_CHAT_C_
#include "slskaliasdef.c"
#endif
