/* Copyright (C) 2005-2006 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.1 $
 * $Date: 2007/04/29 09:55:05 $
 * $Author: sgop $
 */

#include "lcore/utils.h"
#include "slsk.h"
#include "slsk_con.h"

#define STRING_MAX  40

void slsk_con_destroy(slsk_con_t* con)
{
    tcp_socket_destroy(TSOCK(con));
}

void slsk_con_init(slsk_con_t* con, const char* name, char type, int timeout)
{
    tcp_socket_init(TSOCK(con), name, timeout);
    object_add_type(OBJ(con), "SlskConnection");
    con->type = type;
}

static void buffer_print_raw(buffer_t* buffer, unsigned char* message, int length) {
  int i1;
  
  buffer_add_data(buffer, "\n[", 2);
  for (i1 = 0; i1 < length; i1++) {
    if (message[i1] >= 32 && message[i1] < 127)
      buffer_add_data(buffer, message+i1, 1);
    else if (message[i1] == '\r')
      buffer_add_data(buffer, "[\\r]", 4);
    else if (message[i1] == '\n')
      buffer_add_data(buffer, "[\\n]\n", 5);
    else
      buffer_add_string(buffer, l_printf("[%02x]", message[i1]));
  }
  buffer_add_data(buffer, "]", 1);
}

void slsk_message(slsk_message_t* message, char dir,
                  slsk_con_t* con, int level, ...) {
  va_list ap;
  guint32 len;
  char* pos = message->data;
  int ttype;
  unsigned dlen = message->length;
  static buffer_t* buffer = NULL;

  if (!buffer) buffer = buffer_new(1024);
  buffer_clear(buffer);

  va_start(ap, level);
  ttype = va_arg(ap, int);
  
  buffer_add_string
    (buffer, l_printf("%c: %3u | %s - ", dir,
		      message->type, con?OBJ(con)->name:"incoming"));
  
  while (ttype != VAL_NONE && ttype != VAL_INC) {
    buffer_add_data(buffer, " ", 1);
    switch (ttype) {
    case VAL_STRING:
      REQ_LEN(4);
      len = unpack_u32(&pos, &dlen);
      REQ_LEN(len);
      if (len > STRING_MAX) {
        buffer_add_data(buffer, "...'", 4);
        buffer_add_data(buffer, pos+len-STRING_MAX, STRING_MAX);
        buffer_add_data(buffer, "'", 1);
      } else {
        buffer_add_data(buffer, "'", 1);
        buffer_add_data(buffer, pos, len);
        buffer_add_data(buffer, "'", 1);
      }
      pos += len; dlen -= len;
      break;
    case VAL_BIN:
      REQ_LEN(4);
      len = unpack_u32(&pos, &dlen);
      REQ_LEN(len);
      buffer_add_string(buffer, l_printf("<%u byte binary>", len));
      pos += len; dlen -= len;
      break;
    case VAL_INT:
      REQ_LEN(4);
      len = unpack_u32(&pos, &dlen);
      buffer_add_string(buffer, l_printf("%u", len));
      break;
    case VAL_DINT: {
      guint64 size;
      REQ_LEN(8);
      size = unpack_u32d(&pos, &dlen);
      buffer_add_string(buffer, l_printf("%llu", size));
      break;
    }
    case VAL_IP:
      REQ_LEN(4);
      len = htonl(unpack_u32(&pos, &dlen));
      buffer_add_string(buffer, l_printf("%s", ntoa(len)));
      break;
    case VAL_BYTE:
      REQ_LEN(1);
      buffer_add_string(buffer, l_printf("0x%02x", unpack_byte(&pos, &dlen)));
      break;
    }
    ttype = va_arg(ap, int);
  }
  va_end(ap);

  len = message->length-(pos-message->data);
  if (len != 0) {
    buffer_add_string(buffer, l_printf(" [%d byte left]", len));
    if (ttype == VAL_NONE) buffer_print_raw(buffer, pos, len);
  }
  buffer_need_space(buffer, 1);
  buffer->data[buffer->size] = 0;
  g_log(l_printf("SLSK_%c", con?con->type:'X'), level, "%s", buffer->data);
}

slsk_message_t* slsk_consume_message(buffer_t* buffer) {
  static slsk_message_t result;
  guint32 type;
  guint32 len;
  char* data;
  unsigned avail;

  if (!buffer) return NULL;
  avail = buffer_avail(buffer);
  if (avail < 4) return NULL;

  data = buffer_read_pos(buffer);
  len = unpack_u32(&data, &avail);

  if (len < 4) return NULL;
  if (avail < len) return NULL;

  type = unpack_u32(&data, &avail);

  result.type = type;
  result.length = len-4;
  result.data = data;

  buffer_consume_virt(buffer, len + 4);
  
  return &result;
}

int slsk_pack_file(char* buffer, file_t* file, gboolean winname) {
  guint32 added = 0;
  attribute_t* attr1;
  attribute_t* attr2;
  attribute_t* attr3;
  char* temp;

  added += pack_byte(buffer+added, 1); // always 1?
  if (winname) {
    temp = file->netname;
  } else {
    temp = auto_text(from_utf8(file->filename, -1, NULL, NULL, 2, slsk_charset_get(), "utf-8"));
  }
  added += pack_string_u32(buffer+added, temp);
  added += pack_u32d(buffer+added, file->size);
  
  attr1 = attr_find(ATTRC(file), "file", "Bitrate");
  attr2 = attr_find(ATTRC(file), "file", "Duration");
  attr3 = attr_find(ATTRC(file), "file", "VBR");

  // nicotine thinks that the attribute number does not have any special
  // meaning. I believe that it's the type of the attribute.
  // Nicotine implicitely uses fixed types if the suffix is 'mp3' otherwise
  // the attributes are ignored. So send 'mp3' as suffix if we have a bitrate
  // Note: after some investigations, i think nicotine is right.
  // Stupid protocol this is..

  if (attr1 || attr2) {
    int bitrate = attr1?attr_get_ref(attr1, int):0;
    int duration = attr2?attr_get_ref(attr2, int):0;
    int vbr = attr3?attr_get_ref(attr3, gboolean):FALSE;

    added += pack_string_u32(buffer+added, "mp3");
    added += pack_u32(buffer+added, 3);

    added += pack_u32(buffer+added, 0);  // attr bitrate
    added += pack_u32(buffer+added, bitrate);
    added += pack_u32(buffer+added, 1);  // attr duration
    added += pack_u32(buffer+added, duration);
    added += pack_u32(buffer+added, 2);  // attr vbr
    added += pack_u32(buffer+added, vbr);
  } else {
    added += pack_string_u32(buffer+added, "");
    added += pack_u32(buffer+added, 0);
  }

  return added;
}


file_t* slsk_unpack_file(char** data, unsigned* dlen) {
  char* filename;
  file_size_t filesize;
  char* suffix;
  guint32 num_attr;
  file_t* file;

  if (*dlen < 5) return NULL;
  (*data)++; (*dlen)--; // skip the first byte, seems to be always 0x01
  filename = unpack_string_u32(data, dlen);
  if (!filename) return NULL;

  if (*dlen < 8) return NULL;
  filesize = unpack_u32d(data, dlen);
  suffix = unpack_string_u32(data, dlen);
  if (!suffix) return NULL;

  if (*dlen < 4) return NULL;

  file = file_new();
  file_set_origname(file, g_strdup(filename));
  file_set_locname(file, to_utf8(filename, -1, NULL, NULL, 3, 
                                 "utf-8", slsk_charset_get(), "latin1"));
  file->size = filesize;

  num_attr = unpack_u32(data, dlen);
  while (num_attr > 0) {
    guint32 type;
    guint32 value;
    if (*dlen < 8) {
      object_kill(OBJ(file));
      return NULL;
    }
    type = unpack_u32(data, dlen);
    value = unpack_u32(data, dlen);
    switch (type) {
    case 0: // bitrate
      if (value) attr_store(ATTRC(file), "file", "Bitrate", value);
      break;
    case 1: // duration
      if (value) attr_store(ATTRC(file), "file", "Duration", value);
      break;
    case 2: // vbr
      if (value) attr_store(ATTRC(file), "file", "VBR", TRUE);
      break;
    default:
      g_log("SLSK_P", G_LOG_LEVEL_WARNING,
            "got unknown attribute: %d - %d - %s", type, value, filename);
      break;
    }
    num_attr--;
  }

  return file;
}

