/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.1 $
 * $Date: 2007/04/29 09:55:05 $
 * $Author: sgop $
 */

#include <string.h>

#include <libxml/parser.h>

#include "lcore/xml_handler.h"
#include "slsk_download.h"
#include "slsk.h"

typedef struct {
    char* filename;
    file_size_t size;
    slsk_user_t* user;
    slsk_source_t* source;
    download_t* download;
} slsk_pot_t;

typedef struct {
    module_t _module;
    guint32 id;
    GList* sources;
    GList* pots;
} slsk_download_module_t;

static slsk_download_module_t* SlskDownloadModule = NULL;

/* static slsk_source_t* slsk_source_new(file_t* file, slsk_user_t* user); */
static void on_source_destroyed(source_t* source, slsk_pot_t* pot);
static void on_download_destroyed(download_t* download, slsk_pot_t* pot);


static void slsk_pots_save()
{
    if (module_running(MODULE(SlskDownloadModule)))
        lopster_xml_save("slsk:pots", 1000);
}

static void slsk_pots_load()
{
    lopster_xml_load("slsk:pots");
}

static void slsk_pot_set_download(slsk_pot_t* pot, download_t* download)
{
    g_assert(!pot->download);
    pot->download = download;
    object_notify_d(OBJ(download), (ObjectFunc)on_download_destroyed, pot);
}

static void slsk_pot_set_source(slsk_pot_t* pot, slsk_source_t* source)
{
    if (pot->source)
        object_unnotify_d(OBJ(pot->source), (ObjectFunc)on_source_destroyed, pot);

    pot->source = source;
    
    if (pot->source)
        object_notify_d(OBJ(pot->source), (ObjectFunc)on_source_destroyed, pot);
}

static void on_source_destroyed(source_t* source, slsk_pot_t* pot)
{
    g_assert(SOURCE(pot->source) == source);
    pot->source = NULL;
}

static void on_user_changed(slsk_user_t* user, slsk_pot_t* pot)
{
    if (user->status > SLSK_USER_OFFLINE && !pot->source)
    {
        file_t* file;
        source_t* source;
        
        g_assert(pot->download);
        file = file_clone(FILE(pot->download), "file");
        
        file_set_origname(file, pot->filename);
        file_set_locname(file, FILE(pot->download)->localname);
        source = download_add_source(pot->download, file, USER(pot->user));
        
        slsk_pot_set_source(pot, SLSK_SOURCE(source));
    }
    else if (user->status == SLSK_USER_OFFLINE && pot->source)
    {
        if (SOURCE(pot->source)->status != SOURCE_LOADING)
            socket_done(SOCK(pot->source), ST_DELETE, "user went offline");
    }
}

static void slsk_pot_destroy(slsk_pot_t* pot)
{
    g_assert(!pot->download);
    g_assert(!pot->source);

    // stop observing user
    object_unmonitor_d(OBJ(pot->user), (ObjectFunc)on_user_changed, pot);

    g_free(pot->filename);
    user_unref(USER(pot->user));
    g_free(pot);

    SlskDownloadModule->pots = g_list_remove(SlskDownloadModule->pots, pot);
    slsk_pots_save();
}

static void on_download_destroyed(download_t* download, slsk_pot_t* pot)
{
    g_assert(pot->download == download);
    slsk_pot_set_source(pot, NULL);
    pot->download = NULL;
    slsk_pot_destroy(pot);
}

static slsk_pot_t* slsk_pot_find(slsk_user_t* user, file_t* file)
{
    GList* dlist;
    for (dlist = SlskDownloadModule->pots; dlist; dlist = dlist->next)
    {
        slsk_pot_t* pot = dlist->data;
        if (pot->user == user && !strcmp(pot->filename, file->netname)) return pot;
    }
    return NULL;
}

static void slsk_pot_init(slsk_pot_t* pot, user_t* user,
                          const char* filename, file_size_t size)
{
    pot->filename = g_strdup(filename);
    pot->user = user_ref(user);
    pot->size = size;
    
    user_start_monitor(user);
    object_monitor_d(OBJ(pot->user), (ObjectFunc)on_user_changed, pot);

    SlskDownloadModule->pots = g_list_append(SlskDownloadModule->pots, pot);
    slsk_pots_save();
}

static slsk_pot_t* slsk_pot_new(user_t* user, const char* filename, file_size_t size)
{
    slsk_pot_t* pot = g_malloc0(sizeof(*pot));
    slsk_pot_init(pot, user, filename, size);
    return pot;
}

static void slsk_pot_new_from_source(source_t* source, download_t* download)
{
    slsk_pot_t* pot = slsk_pot_find(SLSK_USER(source->user), source->file);
    if (pot) return;

    pot = slsk_pot_new(source->user, source->file->netname, source->file->size);
    slsk_pot_set_source(pot, SLSK_SOURCE(source));
    slsk_pot_set_download(pot, download);
}

static void slsk_pot_new_from_file(slsk_user_t* user, const char* filename,
                                   const char* localname, file_size_t size)
{
    download_t* download = download_find(localname, size);
    if (download)
    {
        slsk_pot_t* pot = slsk_pot_new(USER(user), filename, size);
        slsk_pot_set_download(pot, download);
    }
    else
    {
        g_log("SLSK_DOWN", G_LOG_LEVEL_DEBUG,
              "discard potential source '%s'", OBJ(user)->name);
    }
}



static void xml_save_pot(slsk_pot_t* pot, xmlNodePtr xnode)
{
    xmlNodePtr xsub = xmlNewChild(xnode, NULL, "Source", NULL);
    xmlNewProp(xsub, "User", OBJ(pot->user)->name);
    xmlNewProp(xsub, "Filesize", l_printf("%llu", pot->size));
    //FIXME: filename is not utf8! need to encode
    xmlNewProp(xsub, "Filename", pot->filename);
    if (pot->download)
        xmlNewProp(xsub, "Download", FILE(pot->download)->localname);
    else
        xmlNewProp(xsub, "Download", FILE(SOURCE(pot->source)->download)->localname);
}

static void xml_save_pots(xmlDocPtr xfile, xmlNodePtr xroot)
{
    (void)xfile;
    g_list_foreach(SlskDownloadModule->pots, (GFunc)xml_save_pot, xroot);
}

static void xml_get_pot(xmlNodePtr xnode)
{
    xmlChar* attr1 = xmlGetProp(xnode, "User");
    xmlChar* attr2 = xmlGetProp(xnode, "Filename");
    xmlChar* attr3 = xmlGetProp(xnode, "Filesize");
    xmlChar* attr4 = xmlGetProp(xnode, "Download");

    if (attr1 && attr2 && attr3 && attr4) {
        slsk_pot_new_from_file(slsk_user_find(attr1), attr2, attr4, strtoll(attr3, NULL, 10));
    }
    if (attr1) xmlFree(attr1);
    if (attr2) xmlFree(attr2);
    if (attr3) xmlFree(attr3);
    if (attr4) xmlFree(attr4);
}

static void xml_load_pots(xmlDocPtr xfile, xmlNodePtr xroot)
{
    xmlNodePtr xsub = xroot->xmlChildrenNode;
    (void)xfile;
    while (xsub != NULL)
    {
        if (!strcmp(xsub->name, "Source"))
        {
            xml_get_pot(xsub);
        }
        xsub = xsub->next;
    }
}

slsk_source_t* slsk_source_find_by_id(guint32 token)
{
    GList* dlist;
    // FIXME: speedup
    g_assert(SlskDownloadModule);
    for (dlist = SlskDownloadModule->sources; dlist; dlist = dlist->next)
    {
        slsk_source_t* source = dlist->data;
        if (source->id == token) return source;
    }
    return NULL;
}

static gboolean slsk_source_end(slsk_source_t* source, const char* message)
{
    int status = socket_get_status(SOCK(source));
    switch (status)
    {
    case ST_SOURCE_REMOTE:
        source_set_remote(SOURCE(source), 60*5, 0);
        slsk_peer_queue_pos(source);  // request the queue position
        source_end(SOURCE(source), message);
        socket_set_status(SOCK(source), ST_UNKNOWN);
        return FALSE;
    default:
        return source_end(SOURCE(source), message);
    }
}

static void slsk_source_destroy(slsk_source_t* source)
{
    user_stop_monitor(SOURCE(source)->user);

    SlskDownloadModule->sources = g_list_remove(SlskDownloadModule->sources, source);
    source_destroy(SOURCE(source));
}

static gboolean slsk_source_start(slsk_source_t* source)
{
    //FIXME:
    slsk_peer_request(source->id, SLSK_USER(SOURCE(source)->user),
                      SOURCE(source)->file, 0);
    return TRUE;
}

static void slsk_source_init(slsk_source_t* source, file_t* file, slsk_user_t* user)
 {
    source_init(SOURCE(source), file, USER(user), TRUE);
    object_add_type(OBJ(source), "SlskSource");
    OBJ(source)->destroy = (ObjectFunc)slsk_source_destroy;
    socket_set_end(SOCK(source), (SockEndFunc)slsk_source_end);
    SOURCE(source)->overlap = TRUE;
    SOURCE(source)->start = (SourceStartFunc)slsk_source_start;
    source->id = SlskDownloadModule->id++;
    SlskDownloadModule->sources = g_list_append(SlskDownloadModule->sources, source);
  
    user_start_monitor(USER(user));
}

static slsk_source_t* slsk_source_new(file_t* file, slsk_user_t* user)
{
    slsk_source_t* source = g_malloc0(sizeof(*source));
    slsk_source_init(source, file, user);
    return source;
}

const char* slsk_source_start_downloading(slsk_peer_t* peer, guint32 token)
{
    segment_t* seg;
    char buffer[8+4];
    const char* res;
    slsk_source_t* source;

    g_assert(SlskDownloadModule);
    source = slsk_source_find_by_id(token);

    if (!source)
    {
        g_log("SLSK_DOWN", G_LOG_LEVEL_WARNING,
              "source id %u not found", token);
        return "source not found";
    }
    else
    {
        tcp_socket_move(TSOCK(source), TSOCK(peer), TRUE);
    }

    seg = source_attach_segment(SOURCE(source));
    if (!seg)
    {
        socket_done(SOCK(source), ST_ERROR, "no free segment");
        return "no free segment";
    }

    g_log("SLSK_DOWN", G_LOG_LEVEL_DEBUG,
          "sending ID %u and offset "FF, source->id, seg->start+seg->size);
    pack_u32(buffer, source->id);
    pack_u32d(buffer+4, seg->start+seg->size);
    tcp_socket_send_bin(TSOCK(source), buffer, 12);
  
    res = source_start_downloading(SOURCE(source));
    if (res) socket_done(SOCK(source), ST_ERROR, res);

    return "outgoing->source";
}

static slsk_source_t* 
slsk_file_to_source(download_t* download, file_t* file, slsk_user_t* user)
{
    slsk_source_t* source = slsk_source_new(file, user);
    slsk_pot_new_from_source(SOURCE(source), download);
    return source;
}



static void slsk_downloads_init()
{
    lopster_xml_register_home("slsk:pots", SDIR("slsk") "potentials.xml",
                              "PotentialDownloads", xml_load_pots, xml_save_pots);
    source_factory_register(SLSK_ID, (FileToSourceFunc)slsk_file_to_source);
}

static void slsk_downloads_run()
{
    slsk_pots_load();
}

static void slsk_downloads_exit()
{
}

static void slsk_download_module_init(slsk_download_module_t* module)
{
    module_init(MODULE(module), "Slsk Download Module",
                (ModuleFunc)slsk_downloads_init,
                (ModuleFunc)slsk_downloads_run,
                (ModuleFunc)slsk_downloads_exit);

    module->id = 1000;
    module->sources = NULL;
    module->pots = NULL;

    module_require(xml_module_instance);
    module_require(download_module_instance);

}


module_t* slsk_download_module_instance()
{
    if (!SlskDownloadModule)
    {
        SlskDownloadModule = g_malloc0(sizeof(*SlskDownloadModule));
        slsk_download_module_init(SlskDownloadModule);
    }
    return MODULE(SlskDownloadModule);
}
