/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.3 $
 * $Date: 2007/11/21 20:20:07 $
 * $Author: sgop $
 */

#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <netinet/in.h>
#endif
#include <string.h>

#include <glib.h>

#include "lcore/preferences.h"
#include "lcore/buffer.h"
#include "lcore/l_http.h"
#include "lcore/l_listen.h"
#include "lcore/file_tree.h"
#include "lcore/library.h"
#include "lcore/l_resolve.h"
#include "lcore/action_item.h"
#include "lcore/l_base.h"
#include "slsk_ui.h"
#include "slsk_network.h"
#include "slsk_proto.h"
#include "slsk_chat.h"
#include "slsk.h"
#include "slsk_hotlist.h"
#include "slsk_peer.h"

#ifdef WIN32
#include "slskalias.h"
#endif

typedef struct {
  module_t _module;
  slsk_net_t* net;
  int share_updater;
  int relogin_timer;
  gboolean auto_connect;
  listen_t* listen;
} slsk_net_module_t;

static slsk_net_module_t* SlskNetModule = NULL;


static void slsk_net_destroy(slsk_net_t* net) {
  if (net->address) g_free(net->address);
  if (net->nick) g_free(net->nick);
  if (net->pass) g_free(net->pass);
  slsk_con_destroy(SLSK_CON(net));
}

static void slsk_net_send_shares(slsk_net_t* net)
{
    file_tree_t* tree = FILE_TREE(library_get_current());
    
    slsk_net_send(net, SLSK_S_SHARE_LIB,
                  tree?tree->fci[MEDIA_FOLDER][3]:0, // folders:all
                  tree?tree->fci[MEDIA_ANY][1]:0);  // all_files:shared
}

static void slsk_net_send_listen(slsk_net_t* net)
{
    slsk_net_send(net, SLSK_S_SET_LISTEN, 
                  (int)listen_port_get_port(SlskNetModule->listen));
}

void slsk_net_set_privileges(slsk_net_t* net, int priv) {
  g_assert(SlskNetModule);
  net->privileges = priv;
  object_set_dirty(OBJ(net));
}

static void slsk_net_reset(slsk_net_t* net)
{
    slsk_net_set_privileges(net, -1);
    slsk_net_channels_clear(net);
    slsk_hot_left_net(net);
}

static void slsk_update_status()
{
    int status = ONLINE_STATUS_IDLE;
    if (SlskNetModule->net && SlskNetModule->net->status == SLSK_NET_ONLINE)
        status |= ONLINE_STATUS_ONLINE;
    plugin_set_online_status(SLSK_ID, status);
}

static void slsk_net_set_status(slsk_net_t* net, int status, const char* message) {

  net->status = status;
  g_free(net->status_message);
  net->status_message = g_strdup(message);

  slsk_update_status();

  if (status == SLSK_NET_OFFLINE)
    socket_set_timeout(SOCK(net), 0);
  else if (status == SLSK_NET_CONNECTING)
    socket_set_timeout(SOCK(net), 60);
  else
    socket_set_timeout(SOCK(net), 6000);

  object_set_dirty(OBJ(net));
}

static gboolean slsk_net_connect_again()
{
    if (!SlskNetModule) return FALSE;
    if (SlskNetModule->relogin_timer)
    {
        lopster_timeout_remove(SlskNetModule->relogin_timer);
        SlskNetModule->relogin_timer = 0;
        if (SlskNetModule->net) slsk_net_connect(SlskNetModule->net);
    }
    return FALSE;
}

static gboolean slsk_net_end(slsk_net_t* net, const char* message) {
  
  tcp_socket_end(TSOCK(net), message);
  slsk_net_set_status(net, SLSK_NET_OFFLINE, message);
  
  slsk_net_reset(net);

  if (!SlskNetModule->relogin_timer && SlskNetModule->auto_connect) {
    SlskNetModule->relogin_timer =
      lopster_timeout_add(30000, (GSourceFunc)slsk_net_connect_again, NULL);
  }
  return FALSE;
}

static const char* slsk_net_consume_input(b_socket_t* socket) {
  slsk_net_t* net = SLSK_NET(socket);
  buffer_t* buffer;
  slsk_message_t* result;
  const char* message;
  
  while (1) {
    // always get buffer pointer, it might have changed
    buffer = b_socket_get_in_buffer(socket);
    result = slsk_consume_message(buffer);
    if (!result) break;   // no command avail
    message = slsk_handle_message(net, result);
    if (message) return message;
  }

  buffer_consume(buffer, 0); // real consume the data
  
  return NULL;
}

static void slsk_net_init(slsk_net_t* net, const char* address,
			  int port) {
  slsk_con_init(SLSK_CON(net), address, 'S', 0);
  object_add_type(OBJ(net), "SlskNetwork");
  OBJ(net)->destroy = (ObjectFunc)slsk_net_destroy;
  socket_set_end(SOCK(net), (SockEndFunc)slsk_net_end);
  b_socket_set_consume
    (BSOCK(net), (BConsumeInputFunc)slsk_net_consume_input);
  net->address = address?g_strdup(address):NULL;
  net->ip = 0;
  net->port = port;
  net->status = SLSK_NET_OFFLINE;
  net->privileges = -1;

  socket_add_load(SOCK(net), "Server");
}

static slsk_net_t* slsk_net_new(const char* address, int port) {
  slsk_net_t* net;

  if (SlskNetModule->net) {
    g_log("SLSK_NET", G_LOG_LEVEL_WARNING,
          "the soulseek module allows one server only");
    g_assert(FALSE);
    return NULL;
  }

  SlskNetModule->net = net = g_malloc0(sizeof(*net));
  slsk_net_init(net, address, port);

  return net;
}

static void slsk_create_default_net() {
  const char* addy = pref_get_ptr("/slsk/server/address", char);
  int port = pref_get_ref("/slsk/server/port", int);
  slsk_net_t* net = slsk_net_new(addy, port);
  
  SLSKUI.net_show(net);
}

void slsk_net_send(slsk_net_t* net, guint32 type, ...) {
  va_list ap;
  
  g_assert(SlskNetModule);
  if (!net && SlskNetModule->net) net = SlskNetModule->net;

  if (net && net->status != SLSK_NET_OFFLINE) {
    va_start(ap, type);
    slsk_pack_command(net, type, ap);
    va_end(ap);
    b_socket_start_sending(BSOCK(net));
  }
}


static void slsk_net_connect_real(slsk_net_t* net)
{
    
    if (net->status != SLSK_NET_OFFLINE) return;
    
    slsk_net_set_status(net, SLSK_NET_CONNECTING, "Connecting...");
    
    if (!tcp_socket_connect(TSOCK(net), net->ip, htons(net->port)))
    {
        net->ip = 0;
        return;
    }
    
    b_socket_start_receiving(BSOCK(net), 1024000);
    
    slsk_net_send(net, SLSK_S_LOGIN, net->nick, net->pass,
                  pref_get_ref("/slsk/protocol", int));
}

void slsk_net_resolver(slsk_net_t* net, unsigned long ip)
{
    g_assert(SlskNetModule);
    // check whether net is still valid
    if (SlskNetModule->net != net) return;
    
    net->ip = ip;
    if (ip == 0)
    {
        slsk_net_set_status(net, SLSK_NET_OFFLINE,
                            "Could not resolve address");
    }
    else
    {
        // now ready to connect
        slsk_net_set_status(net, SLSK_NET_OFFLINE,
                            "Successfully resolved address");
        slsk_net_connect_real(net);
    }
}

void slsk_net_disconnect(slsk_net_t* net) {
  g_assert(SlskNetModule);
  if (!net) net = SlskNetModule->net;
  if (net && net->status != SLSK_NET_OFFLINE)
    socket_done(SOCK(net), 0, "User disconnected");
}

const char* slsk_net_get_username(slsk_net_t* net) {
  const char* nick = NULL;
  
  g_assert(SlskNetModule);
  if (net) nick = net->nick;
  if (!nick || *nick==0) nick = pref_get_ptr("/slsk/username", char);
  if (!nick || *nick==0) nick = pref_get_ptr("/core/user/nick", char);
  return nick;
}

void slsk_net_connect(slsk_net_t* net)
{
    const char* nick = NULL;
    char* pass = NULL;
    
    g_assert(SlskNetModule);
    if (!net) return;
    
    if (net->status != SLSK_NET_OFFLINE) return;
    
    nick = slsk_net_get_username(NULL);
    if (!nick || *nick==0)
    {
        g_log("SLSK_NET", G_LOG_LEVEL_WARNING,
              "Cannot connect to SoulSeek (no username given)");
        return;
    }
    if (!pass || *pass==0) pass = pref_get_ptr("/slsk/password", char);
    if (!pass || *pass==0) pass = pref_get_ptr("/core/user/pass", char);
    if (!pass || *pass==0)
    {
        g_log("SLSK_NET", G_LOG_LEVEL_WARNING,
              "Cannot connect to SoulSeek (no password given)");
        return;
    }
    g_free(net->nick);
    net->nick = g_strdup(nick);
    g_free(net->pass);
    net->pass = g_strdup(pass);
    
    if (net->ip == 0)
    {
        resolve_address(net->address, (call_func_t)slsk_net_resolver, net);
    }
    else
    {
        slsk_net_connect_real(net);
    }
}

static void slsk_net_send_away(slsk_net_t* net)
{
    char* away;
    away = pref_get_ptr("/core/user/away", char);
    if (away && *away) 
        slsk_net_send(net, SLSK_S_SET_STATUS, SLSK_USER_AWAY);
}

void slsk_net_success(slsk_net_t* net, char* message)
{
    g_assert(SlskNetModule);
    slsk_net_set_status(net, SLSK_NET_ONLINE, "Online");
    
    slsk_net_send_listen(net);
    slsk_net_send_shares(net);
    slsk_net_channels_clear(net);
    slsk_net_send_away(net);
    slsk_hot_joined_net(net);
    
    server_message("SoulSeek", message);
}

void slsk_net_channel_add(slsk_net_t* net, slsk_channel_t* channel, gboolean update)
{
    g_assert(SlskNetModule);
    if (net && channel)
        net->channels = g_list_prepend(net->channels, channel);

    if (update) SLSKUI.update_channels(net);
}

void slsk_net_channel_remove(slsk_net_t* net, const char* channel)
{
    GList* dlist;
    
    g_assert(SlskNetModule);
    if (!net || !channel) return;
    for (dlist = net->channels; dlist; dlist = dlist->next)
    {
        slsk_channel_t* ch = dlist->data;
        if (!strcmp(channel, ch->name))
        {
            net->channels = g_list_delete_link(net->channels, dlist);
            g_free(ch->name);
            g_free(ch);
            SLSKUI.update_channels(net);
            return;
        }
    }
}

void slsk_net_channels_clear(slsk_net_t* net)
{
    GList* dlist;
    
    g_assert(SlskNetModule);
    if (!net) return;
    for (dlist = net->channels; dlist; dlist = dlist->next)
    {
        slsk_channel_t* channel = dlist->data;
        g_free(channel->name);
        g_free(channel);
    }
    g_list_free(net->channels);
    net->channels = NULL;
    SLSKUI.update_channels(net);
}

static gboolean share_updater(slsk_net_t* net) {
  lopster_timeout_remove(SlskNetModule->share_updater);
  SlskNetModule->share_updater = 0;
  slsk_net_send_shares(net);
  return FALSE;
}

void slsk_net_update_shares(slsk_net_t* net) {
  g_assert(SlskNetModule);
  if (SlskNetModule->share_updater) lopster_timeout_remove(SlskNetModule->share_updater);
  SlskNetModule->share_updater =
    lopster_timeout_add(5000, (GSourceFunc)share_updater, net);
}
static void _slsk_net_update_shares() {
  slsk_net_update_shares(NULL);
}

slsk_net_t* slsk_net_get_online() {
  g_assert(SlskNetModule);
  if (SlskNetModule->net && SlskNetModule->net->status == SLSK_NET_ONLINE) 
    return SlskNetModule->net;
  return NULL;
}

void slsk_net_check_privileges(slsk_net_t* net) {
  g_assert(SlskNetModule);
  slsk_net_set_privileges(net, -1);
  slsk_net_send(net, SLSK_S_CHECK_PRIVILEGES);
}

static void slsk_start_connect()
{
    g_assert(SlskNetModule);
    if (SlskNetModule->auto_connect) return;
    SlskNetModule->auto_connect = TRUE;
    if (SlskNetModule->net) slsk_net_connect(SlskNetModule->net);
    slsk_update_status();
}

static void slsk_stop_connect(void)
{
    g_assert(SlskNetModule);
    SlskNetModule->auto_connect = FALSE;
    slsk_net_disconnect(NULL);
    slsk_update_status();
}

void slsk_connect_change(unsigned status)
{
    switch (status) {
    case CONNECT_START:
        slsk_start_connect();
        break;
    case CONNECT_STOP:
        slsk_stop_connect();
        break;
    }
}

// LISTEN STUFF
static const char* slsk_handle_con(tcp_socket_t* s) {
  guint32 len;
  guint32 token;
  char* pos;
  buffer_t* buffer;
  char type;
  const char* result;
  unsigned dlen;
  slsk_message_t m;

  buffer = b_socket_get_in_buffer(BSOCK(s));
  
  dlen = buffer_avail(buffer);
  if (dlen < 4) {
    // allow to read the message length
    if (!buffer_space(buffer)) b_socket_set_limit(BSOCK(s), 4);
    return NULL;
  }
  pos = buffer_read_pos(buffer);
  
  len = unpack_u32(&pos, &dlen);
  if (dlen < len) {
    // allow to read the rest of the message
    if (!buffer_space(buffer)) b_socket_set_limit(BSOCK(s), len+4);
    return NULL;
  }
  dlen = len;
  
  if (dlen < 5) return "invalid connection in";

  type = unpack_byte(&pos, &dlen);

  m.length = dlen;
  m.data = pos;
  m.type = type;
  if (type == 0) {

    slsk_message(&m, 'R', NULL, G_LOG_LEVEL_DEBUG, VAL_INT, VAL_NONE);
    
    token = unpack_u32(&pos, &dlen);
    buffer_consume_virt(buffer, len+4);
    return slsk_peer_replied(s, token);
  } else if (type == 1) {
    char* user;
    char* mode;
    
    slsk_message(&m, 'R', NULL, G_LOG_LEVEL_DEBUG,
                 VAL_STRING, VAL_STRING, VAL_INT, VAL_NONE);

    user = unpack_string_u32(&pos, &dlen);
    if (!user) return "invalid connection in";
    mode = unpack_string_u32(&pos, &dlen);
    if (!mode) return "invalid connection in";
    if (dlen < 1) return "invalid connection in";
    token = unpack_u32(&pos, &dlen);
    
    if (token) {
      // we never got anything other than 0, so warn about that
      g_log("SLSK_X", G_LOG_LEVEL_WARNING,
            "token == %d received", token);
    }
    buffer_consume_virt(buffer, len+4);
    result = slsk_peer_requested(user, s, *mode);
    return result;
  } else {
    return l_printf("*** unknown mode %d", type);
  }
  g_assert(FALSE);
}

static tcp_socket_t* slsk_accept_con(int fd, ip_t ip, port_t port) {
  tcp_socket_t* socket = tcp_socket_new("SlskTemp", 30);
  
  listen_init_accepted(socket, fd, ip, port);

  b_socket_set_consume(BSOCK(socket), (BConsumeInputFunc)slsk_handle_con);
  return socket;
}

static void slsk_change_port()
{
    slsk_net_send_listen(NULL);
}

static void slsk_start_listen_service()
{
    SlskNetModule->listen =
        listen_port_new("/slsk/listen/port", "SlskNetModule->listen",
                        (ListenAcceptFunc)slsk_accept_con,
                        (ListenChangeFunc)slsk_change_port);
    
    object_set_ptr((object_t**)&SlskNetModule->listen);
    listen_port_start_listening(SlskNetModule->listen, 1000);
    slsk_net_send_listen(NULL);
}

static gboolean slsk_net_ping_servers() {
  if (SlskNetModule->net && SlskNetModule->net->status == SLSK_NET_ONLINE)
    slsk_net_send(SlskNetModule->net, SLSK_S_PING);
  return TRUE;
}

static void slsk_nets_init() {
  pref_register_int("/slsk/listen/port", TRUE, 2234);
  pref_register_string("/slsk/username", TRUE, NULL);
  pref_register_string("/slsk/password", TRUE, NULL);
/*   pref_register_string("/slsk/server/address", TRUE, "38.115.131.131"); */
  pref_register_string("/slsk/server/address", TRUE, "server.slsknet.org");
  pref_register_int("/slsk/server/port", TRUE, 2240);

  library_register_actions((LibNodeFunc)_slsk_net_update_shares,
                           (LibNodeFunc)_slsk_net_update_shares, NULL);

  // FIXME: is 60 secs ok, or is the interval given by the server?
  lopster_call_periodic(60000, (CallFunc)slsk_net_ping_servers, NULL);
}

static void slsk_nets_run() {
  slsk_create_default_net();
  slsk_start_listen_service();
}

static void slsk_net_module_init(slsk_net_module_t* module) {
  module_init(MODULE(module), "Slsk Netlist Module",
	      (ModuleFunc)slsk_nets_init, (ModuleFunc)slsk_nets_run, NULL);

  module->net = NULL;
  module->share_updater = 0;
  module->relogin_timer = 0;
  module->auto_connect = FALSE;
  module->listen = NULL;

  module_require(base_module_instance);
  module_require(buddy_module_instance);
  module_require(tsocket_module_instance);
  module_require(library_module_instance);
  module_require(search_module_instance);
  module_require(resolve_module_instance);
  //FIXME: add more deps
}


module_t* slsk_net_module_instance() {
  if (!SlskNetModule) {
    SlskNetModule = g_malloc0(sizeof(*SlskNetModule));
    slsk_net_module_init(SlskNetModule);
  }
  return MODULE(SlskNetModule);
}

#ifdef WIN32
#define _SLSK_NETWORK_C_
#include "slskaliasdef.c"
#endif

