/* Copyright (C) 2005 sgop@users.sourceforge.net
 * This is free software distributed under the terms of the
 * GNU Public License.  See the file COPYING for details.
 */
/* $Revision: 1.2 $
 * $Date: 2007/05/23 20:42:29 $
 * $Author: sid4u $
 */

#define SLSK_P_CONNECT     0x01
#define SLSK_P_INCOMING    0x02

#include <string.h>

#include "lcore/utils.h"
#include "lcore/lopster.h"
#include "lcore/buffer.h"
#include "lcore/zbuffer.h"
#include "lcore/library.h"
#include "lcore/files.h"
#include "slsk.h"
#include "slsk_peer.h"
#include "slsk_proto.h"
#include "slsk_peer_proto.h"
#include "slsk_users.h"
#include "slsk_peer.h"
#include "slsk_search.h"
#include "slsk_upload.h"
#include "slsk_download.h"

#ifdef WIN32
#include "slskalias.h"
#endif

struct _slsk_action_t {
  int id;
  GList* results;
  char* search;
  guint32 token;
  char* filename;
  int dir;
  file_size_t size;
};

static guint32 TokenCnt = 0;
static GList* PCons = NULL;
static int NumPCons = 0;
static GList* DCons = NULL;
static slsk_peer_t* DCon = NULL;
static unsigned ParentTimeout = 120;
static unsigned ParentInterval = 100;
static unsigned MinParentsInCache = 3;
static gboolean HaveEnoughParents = FALSE;

static slsk_action_t* slsk_action_new(int id) {
  slsk_action_t* action = g_malloc(sizeof(*action));
  action->id = id;
  action->results = NULL;
  action->search = NULL;
  return action;
}

static void slsk_action_destroy(slsk_action_t* action) {
  g_list_free(action->results);
  g_free(action->search);
  g_free(action);
}

static void slsk_peer_resume_safe(slsk_peer_t* con);

static void 
slsk_peer_add_action(slsk_peer_t* peer, slsk_action_t* action) {
  peer->actions = g_list_append(peer->actions, action);
  if (socket_connected(SOCK(peer))) slsk_peer_resume_safe(peer);
}

static slsk_action_t* slsk_peer_pop_action(slsk_peer_t* peer) {
  slsk_action_t* action;
  if (!peer->actions) return NULL;
  action = peer->actions->data;
  peer->actions = g_list_remove(peer->actions, action);
  return action;
}

void slsk_peer_remove_save_action(slsk_peer_t* peer) {
  slsk_action_t* action;
  if (!peer->actions_save) return;
  action = peer->actions_save->data;
  peer->actions_save = g_list_remove(peer->actions_save, action);
  slsk_action_destroy(action);
}

static void slsk_peer_reactivate_actions(slsk_peer_t* peer) {
  slsk_action_t* action;

  while ((action = slsk_peer_pop_action(peer)) != NULL) {
    peer->actions_save = g_list_append(peer->actions_save, action);
  }
  peer->actions = peer->actions_save;
  peer->actions_save = NULL;
}

static gboolean slsk_peer_end(slsk_peer_t* con, const char* message) {
  gboolean ret;

  if (!con->user->net) {
    // normal error, abort, etc., dont retry
    ret = TRUE;
  } else if (con->direction & SLSK_P_INCOMING) {
    if (con->direction & SLSK_P_CONNECT) {
      // we could not connect either
      /*
      slsk_net_send(con->user->net, SLSK_S_CANT_CONNECT,
		    con->token, OBJ(con->user)->name);
      */
    }
    ret = TRUE;
  } else if (con->direction & SLSK_P_CONNECT) {
    if (con->actions || con->actions_save || 
        buffer_avail(b_socket_get_out_buffer(BSOCK(con), 0))) {
      
      // cant connect, tell peer to try
      user_set_firewalled(USER(con->user), TRUE);
      
      slsk_net_send(con->user->net, SLSK_S_REQUEST_PEER, con->token,
		    OBJ(con->user)->name, SLSK_CON(con)->type);
      socket_set_timeout(SOCK(con), 30);
      con->direction = 0x0; // outgoing + peer connects
      // and move actions back to process queue
      slsk_peer_reactivate_actions(con);
      ret = FALSE;
    } else {
      // we are done
      ret = TRUE;
    }
  } else {
    ret = TRUE;
  }

  if (con->browse) {
    browse_t* browse = BROWSE(con->browse);
    if (ret == FALSE) {
      browse_set_status(browse, "Fallback to remote connect");
    } else {
      browse_set_status(browse, "Failed");
    }
    browse_set_progress(BROWSE(browse), 0, NULL);
  }

  tcp_socket_end(TSOCK(con), message);

  if (con == DCon && ret == TRUE) {
    DCon = NULL;
    slsk_peer_make_d();
  }

  return ret;
}

static void slsk_peer_destroy(slsk_peer_t* con) {
  NumPCons--;
  PCons = g_list_remove(PCons, con);
  user_unref(USER(con->user));
  slsk_con_destroy(SLSK_CON(con));
}

static void slsk_peer_init(slsk_peer_t* con, slsk_user_t* user,
                           char type, guint32 token) {
  slsk_con_init(SLSK_CON(con), OBJ(user)->name, type, 30);
  object_add_type(OBJ(con), "SlskPeer");
  OBJ(con)->destroy = (ObjectFunc)slsk_peer_destroy;
  socket_set_end(SOCK(con), (SockEndFunc)slsk_peer_end);
  b_socket_set_consume(BSOCK(con), (BConsumeInputFunc)slsk_peer_consume_input);
  con->user = user_ref(USER(user));
  con->token = token;

  socket_add_load(SOCK(con), "Peer");
}

static slsk_peer_t* slsk_peer_find(slsk_user_t* user, char type) {
  GList* dlist;
  // FIXME: we need an implementation faster than this
  for (dlist = PCons; dlist; dlist = dlist->next) {
    slsk_peer_t* con = dlist->data;
    if (con->user == user && SLSK_CON(con)->type == type) return con;
  }
  return NULL;
}

slsk_peer_t* slsk_peer_new(slsk_user_t* user, char type, guint32 token) {
  slsk_peer_t* con;

  if (type != 'P' && type != 'D' && type != 'F') {
    g_assert(FALSE);
    return NULL;
  }
  con = g_malloc0(sizeof(*con));
  slsk_peer_init(con, user, type, token);
  PCons = g_list_append(PCons, con);
  NumPCons++;
  return con;
}

static const char* slsk_peer_get_transfer(slsk_peer_t* con) {
  buffer_t* buffer = b_socket_get_in_buffer(BSOCK(con));
  guint32 dlen;
  char* data;
  slsk_upload_t* upload;
  guint32 reqnum;

  dlen = buffer_avail(buffer);
  data = buffer_read_pos(buffer);
  
  if (dlen < 4) return NULL;
  reqnum = unpack_u32(&data, &dlen);

  g_log("SLSK_UPLOAD", G_LOG_LEVEL_DEBUG,
        "got upload request num %u", reqnum);
  upload = slsk_upload_find_user_req(con->user, reqnum);
  if (!upload)
    return "upload not found (downloads not searched)";

  buffer_consume_virt(buffer, 4);

  return slsk_upload_adapt_peer(upload, con);
}

/** from incoming '1' connection */
const char* slsk_peer_requested(const char* nick, tcp_socket_t* s, char type) {
  slsk_peer_t* con;
  slsk_user_t* user = slsk_user_find(nick);
  
  //  slsk_user_set_ip(user, s->ip);
  con = slsk_peer_new(user, type, TokenCnt++);
  con->direction |= SLSK_P_INCOMING; // peer initiated and connected
  // now copy the socket.
  tcp_socket_move(TSOCK(con), s, TRUE);

  if (type == 'F') {
    b_socket_set_consume(BSOCK(con), (BConsumeInputFunc)slsk_peer_get_transfer);
  }
  
  b_socket_start_receiving(BSOCK(con), 1024000);
  return "tsocket->slsk_peer";
}

static void slsk_peer_greet(slsk_peer_t* con) {
  buffer_clear(b_socket_get_out_buffer(BSOCK(con), 0));
  if (con->direction & SLSK_P_INCOMING) {
    slsk_peer_send(con, SLSK_X_PIERCE_FW, con->token);
  } else {
    slsk_peer_send(con, SLSK_X_PEER_INIT,
		   slsk_net_get_username(con->user->net),
		   SLSK_CON(con)->type/*, con->token*/);
    //NOTE: we always send 0 as last argument atm
  }
}

static int slsk_peer_connect(slsk_peer_t* peer, unsigned long ip, unsigned short port) {
  g_log("SLSK_PEER", G_LOG_LEVEL_DEBUG,
	"trying to connect to '%s'", OBJ(peer->user)->name);
  user_set_firewalled(USER(peer->user), FALSE);
  if (!tcp_socket_connect(TSOCK(peer), ip, port)) return 0;
  slsk_peer_greet(peer);
  return 1;
}

/** peer couldnt connect, now we try */
void slsk_peer_request_fw(slsk_user_t* user, char type, guint32 token) {
  slsk_peer_t* con;
  
  con = slsk_peer_new(user, type, token);
  con->direction |= SLSK_P_INCOMING;  // peer initiated
  con->direction |= SLSK_P_CONNECT;   // we connect
  
  if (type == 'F') {
    b_socket_set_consume
      (BSOCK(con), (BConsumeInputFunc)slsk_peer_get_transfer);
  }
  if (!slsk_peer_connect
      (con, user_ip(con->user), user_port(con->user))) return;
  b_socket_start_receiving(BSOCK(con), 204800);
  //  slsk_peer_resume(con);
}

static slsk_peer_t* slsk_peer_initiate(slsk_user_t* user, char type) {
  slsk_peer_t* con;
  
  //  printf("*** initiating peer connection ***\n");
  if (!user) return NULL;
  if (!user->net) {
    //    printf("*** getting any online net ***\n");
    slsk_user_set_net(user, slsk_net_get_online());
  }
  if (!user->net || user->net->status != SLSK_NET_ONLINE) {
    //    printf("*** could not find online net ***\n");
    return NULL;
  }
  con = slsk_peer_find(user, type);
  if (!con) con = slsk_peer_new(user, type, TokenCnt++);
  //  else printf("*** using existent con ***\n");

  if (!con) return NULL;
  
  if (socket_connected(SOCK(con))) return con;

  if (user_get_firewalled(USER(con->user))) {
    con->direction = 0; // outgoing + peer connects
    slsk_net_send(con->user->net, SLSK_S_REQUEST_PEER, con->token,
		  OBJ(con->user)->name, SLSK_CON(con)->type);
    socket_set_timeout(SOCK(con), 30);
  } else {
    con->direction = SLSK_P_CONNECT; // outgoing + we connect
    if (!slsk_user_ip_known(con->user)) {
      slsk_net_send(con->user->net, SLSK_S_GET_ADDRESS,
		    OBJ(user)->name);
    } else if (!slsk_peer_connect
	       (con, user_ip(con->user), user_port(con->user))) {
      return NULL;
    }
  }
  return con;
}

static const char* slsk_peer_resume(slsk_peer_t* con) {
  slsk_action_t* action;
  g_assert(socket_connected(SOCK(con)));
  
  user_set_ip(USER(con->user), TSOCK(con)->ip);
  while ((action = slsk_peer_pop_action(con)) != NULL) {
    int id = action->id;
    gboolean save = FALSE;
    switch (id) {
    case SLSK_ACTION_UP_FAILED:
      slsk_peer_send(con, SLSK_X_UPLOAD_FAILED, action->filename);
      b_socket_start_receiving(BSOCK(con), 2048);
      break;
    case SLSK_ACTION_QUEUE_POS:
      slsk_peer_send(con, SLSK_S_QUEUE_POS, action->filename);
      b_socket_start_receiving(BSOCK(con), 2048);
      break;
    case SLSK_ACTION_DOWNLOAD:
      return slsk_source_start_downloading(con, action->token);
      break;
    case SLSK_ACTION_UPLOAD:
      slsk_peer_send(con, SLSK_X_TRANS_RESPONSE,
		     action->token, 1, (guint32)(action->size));
      b_socket_start_receiving(BSOCK(con), 204800);
      break;
    case SLSK_ACTION_QUEUED:
      slsk_peer_send(con, SLSK_X_TRANS_RESPONSE,
		     action->token, 0, "Queued");
      b_socket_start_receiving(BSOCK(con), 2048);
      break;
    case SLSK_ACTION_REQUEST:
/*       if (action->dir == 0) */
/*         slsk_peer_send(con, SLSK_X_QUEUE_IN, action->filename); */
      slsk_peer_send(con, SLSK_X_TRANS_REQUEST, action->dir,
                     action->token, action->filename);
      b_socket_start_receiving(BSOCK(con), 2048);
      if (action->dir == 0) save = TRUE;
      break;
    case SLSK_ACTION_WHOIS:
      slsk_peer_send(con, SLSK_X_WHOIS_REQUEST);
      b_socket_start_receiving(BSOCK(con), 20480000);
      save = TRUE;
      break;
    case SLSK_ACTION_SEARCH:
      slsk_peer_send(con, SLSK_X_SEARCH_REQUEST, con->token,
		     action->search);
      b_socket_start_receiving(BSOCK(con), 204800);
      save = TRUE;
      break;
    case SLSK_ACTION_BROWSE:
      browse_set_status(BROWSE(con->browse), "Getting file list");
      slsk_peer_send(con, SLSK_X_BROWSE_REQUEST);
      b_socket_start_receiving(BSOCK(con), 20480000);
      save = TRUE;
      break;
    case SLSK_ACTION_DISTRIB:
      socket_set_timeout(SOCK(con), ParentTimeout);
      b_socket_start_receiving(BSOCK(con), 204800);
      break;
    case SLSK_ACTION_RESULTS:
      slsk_search_answer(con, action->token, action->results);
      //    b_socket_start_receiving(BSOCK(con), 204800);
      //    if (!(con->direction & SLSK_P_CONNECT))
      //    b_socket_close_when_sent(BSOCK(con), TRUE);
      break;
    case SLSK_ACTION_NONE:
      b_socket_start_receiving(BSOCK(con), 204800);
      break;
    default:
      g_log("SLSK_P", G_LOG_LEVEL_WARNING, "action is %d", id);
      g_assert(FALSE);
    }
    if (save) {
      con->actions_save = g_list_append(con->actions_save, action);
    } else {
      slsk_action_destroy(action);
    }
  }
  return NULL;
}

static void slsk_peer_resume_safe(slsk_peer_t* con) {
  const char* res = slsk_peer_resume(con);
  if (res) socket_done(SOCK(con), ST_ERROR, res);
}

void slsk_peer_whois(slsk_user_t* user) {
  slsk_peer_t* con = slsk_peer_initiate(user, 'P');
  if (con) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_WHOIS);
    slsk_whois_new(user);
    slsk_peer_add_action(con, action);
  }
}

void slsk_peer_got_address(slsk_net_t* net, const char* nick,
			   ip_t ip, port_t port) {
  slsk_user_t* user;
  GList* dlist;

  (void)net;

  user = slsk_user_find(nick);
  slsk_user_set_net(user, net);
  slsk_user_set_ip_port(user, ip, port);

  if (!ip) return;
  for (dlist = PCons; dlist; dlist = dlist->next) {
    slsk_peer_t* con = dlist->data;
    if (con->direction == SLSK_P_CONNECT && con->user == user) {
      if (!slsk_peer_connect(con, user_ip(user), user_port(user))) return;
      slsk_peer_resume_safe(con);
      return;
    }
  }
}

const char* slsk_peer_replied(tcp_socket_t* socket, guint32 token) {
  GList* dlist;

  for (dlist = PCons; dlist; dlist = dlist->next) {
    slsk_peer_t* con = dlist->data;
    if (con->direction != 0 || con->token != token) continue;
    // copy the socket
    tcp_socket_move(TSOCK(con), socket, TRUE);
    //    slsk_peer_greet(con);
    slsk_peer_resume_safe(con);
    return "tsocket->slsk_peer";
  }
  return "socket not matched";
}



void slsk_peer_browse(slsk_user_t* user, slsk_browse_t* browse) {
  slsk_peer_t* con = slsk_peer_initiate(user, 'P');
  if (con) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_BROWSE);
    con->browse = browse;
    browse_set_status(BROWSE(con->browse), "Connecting to peer");
    slsk_peer_add_action(con, action);
  }
}

void slsk_peer_browse_removed(slsk_browse_t* browse) {
  GList* dlist;

  for (dlist = PCons; dlist; dlist = dlist->next) {
    slsk_peer_t* con = dlist->data;
    if (con->browse == browse) {
      socket_done(SOCK(con), S_BROWSE_REMOVED, "browse deleted");
      break;
    }
  }
}

static void slsk_peer_check_parents(slsk_net_t* net) {
  if (g_list_length(DCons) >= MinParentsInCache && !HaveEnoughParents) {
    slsk_net_send(net, SLSK_S_HAVE_NO_PARENT, 0);
    HaveEnoughParents = TRUE;
  } else if (g_list_length(DCons) < MinParentsInCache && HaveEnoughParents) {
    slsk_net_send(net, SLSK_S_HAVE_NO_PARENT, 1);
    HaveEnoughParents = FALSE;
  }
}

void slsk_peer_add_host(slsk_net_t* net, const char* nick,
                        ip_t ip, port_t port) {
  slsk_user_t* user;

  user = slsk_user_find(nick);
  if (net) slsk_user_set_net(user, net);
  slsk_user_set_ip_port(user, ip, port);
  
  DCons = g_list_append(DCons, user);
  slsk_peer_check_parents(net);
}

void slsk_peer_abort_d() {
  if (!DCon) return;
  socket_done(SOCK(DCon), ST_ERROR, "user aborted");
}
void slsk_peer_make_d() {
  slsk_peer_t* con;
  slsk_user_t* user;

  //  return;  // FIXME
  if (DCon || !DCons) return;
  
  //FIXME
  user = DCons->data;
  DCons = g_list_remove(DCons, user);
  slsk_peer_check_parents(NULL);
  
  DCon = con = slsk_peer_initiate(user, 'D');
  if (con) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_DISTRIB);
    con->search_cnt = 1;
    slsk_peer_add_action(con, action);
  }
}

void slsk_peer_send_results(slsk_user_t* user, guint32 token, GList* results) {
  slsk_peer_t* peer;
  
  if (!results) return;

  peer = slsk_peer_initiate(user, 'P');
  if (peer) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_RESULTS);
    action->results = results;
    action->token = token;
    slsk_peer_add_action(peer, action);
  } else {
    g_list_free(results);
  }
}

void slsk_peer_upload(guint32 token, slsk_user_t* user, file_size_t size) {
  slsk_peer_t* con = slsk_peer_initiate(user, 'P');
  if (con) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_UPLOAD);
    action->token = token;
    action->size = size;
    slsk_peer_add_action(con, action);
  }
}

void slsk_peer_queued(guint32 token, slsk_user_t* user) {
  slsk_peer_t* con = slsk_peer_initiate(user, 'P');
  if (con) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_QUEUED);
    action->token = token;
    slsk_peer_add_action(con, action);
  }
}
				    


void slsk_peer_set_parent_timeout(unsigned timeout) {
  ParentTimeout = timeout;
}

void slsk_peer_set_parent_interval(unsigned interval) {
  ParentInterval = interval;
  // make sure the servers doesnt trick us and has a too low timeout
  if (ParentTimeout < ParentInterval*6/5)
    ParentTimeout = ParentInterval*6/5;
}

void slsk_peer_set_min_parents(unsigned number) {
  MinParentsInCache = number;
}

void slsk_peer_search(const char* nick, const char* term) {
  slsk_peer_t* con = slsk_peer_initiate(slsk_user_find(nick), 'P');
  if (con) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_SEARCH);
    action->search = g_strdup(term);
    slsk_peer_add_action(con, action);
  }
}

void slsk_peer_request(guint32 token, slsk_user_t* user, file_t* file, int dir) {
  slsk_peer_t* con = slsk_peer_initiate(user, 'P');
  if (con) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_REQUEST);
    action->token = token;
    action->filename = g_strdup(file->netname);
    action->dir = dir;
    slsk_peer_add_action(con, action);
  }
}

void slsk_peer_download(slsk_source_t* source) {
  slsk_peer_t* con = slsk_peer_initiate(SLSK_USER(SOURCE(source)->user), 'F');
  if (con) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_DOWNLOAD);
    action->token = source->id;
    slsk_peer_add_action(con, action);
  }
}

void slsk_peer_queue_pos(slsk_source_t* source) {
  slsk_peer_t* con = slsk_peer_initiate(SLSK_USER(SOURCE(source)->user), 'P');
  if (con) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_QUEUE_POS);
    action->filename = g_strdup(SOURCE(source)->file->netname);
    slsk_peer_add_action(con, action);
  }
}

void slsk_peer_upload_failed(slsk_upload_t* upload) {
  slsk_peer_t* con = slsk_peer_initiate(SLSK_USER(UPLOAD(upload)->user), 'P');
  if (con) {
    slsk_action_t* action = slsk_action_new(SLSK_ACTION_UP_FAILED);
    action->filename = g_strdup(UPLOAD(upload)->file->netname);
    slsk_peer_add_action(con, action);
  }
}

#ifdef WIN32
#define _SLSK_PEER_C_
#include "slskaliasdef.c"
#endif
