/* Copyright (C) 2005 sgop@users.sourceforge.net This is free software
 * distributed under the terms of the GNU Public License.  See the
 * file COPYING for details.
 */
/* $Revision: 1.3 $
 * $Date: 2007/11/16 11:49:27 $
 * $Author: sid4u $
 */

#include <glib.h>

#include "lcore/l_base.h"
#include "lcore/utils.h"
#include "lcore/l_socket.h"
#include "lcore/l_i18n.h"
#include "lcore/chat.h"
#include "slsk_users.h"
#include "slsk.h"
#include "slsk_network.h"
#include "slsk_proto.h"

#ifdef WIN32
#include "slskalias.h"
#endif

typedef struct {
  module_t _module;
  GHashTable* users;
  unsigned privileged;
  gboolean got_privs;
} slsk_user_module_t;

static slsk_user_module_t* SlskUserModule = NULL;


static void on_slsk_user_net_removed(slsk_net_t* net, slsk_user_t* user)
{
    (void)net;
    slsk_user_set_net(user, NULL);
}

static void on_slsk_user_net_changed(slsk_net_t* net, slsk_user_t* user)
{
    if (net->status != SLSK_NET_ONLINE)
        slsk_user_set_status(user, SLSK_USER_UNKNOWN);
    else
        object_set_dirty(OBJ(user));
}

static void slsk_user_destroy(slsk_user_t* user)
{
    g_hash_table_remove(SlskUserModule->users, OBJ(user)->name);
    g_free(user->description);
    object_destroy(OBJ(user));
}

static void slsk_user_monitor(slsk_user_t* user)
{
    slsk_net_send(NULL, SLSK_S_ADD_USER, OBJ(user)->name);
}

static void slsk_user_unmonitor(slsk_user_t* user)
{
    slsk_net_send(NULL, SLSK_S_REMOVE_USER, OBJ(user)->name);
}

static void slsk_user_init(slsk_user_t* user, const char* nick)
{
    user_init(USER(user), nick, SLSK_ID);
    object_add_type(OBJ(user), "SlskUser");
    OBJ(user)->destroy = (ObjectFunc)slsk_user_destroy;
    USER(user)->monitor = (UserMonitor)slsk_user_monitor;
    USER(user)->unmonitor = (UserMonitor)slsk_user_unmonitor;
    user->privileges = SlskUserModule->got_privs?SLSK_PRIV_OFF:SLSK_PRIV_UNKNOWN;
    user->status = SLSK_USER_UNKNOWN;
    user->description = NULL;
    user->all_slots = -1;
    user->queued = -1;
    user->has_slots = FALSE;
    user->shared = -1;
    user->shared_dirs = -1;
    user->mon_cnt = 0;
}

static slsk_user_t* slsk_user_new(const char* nick)
{
    slsk_user_t* user = g_malloc0(sizeof(*user));
    slsk_user_init(user, nick);
    return user;
}

slsk_user_t* slsk_user_find(const char* nick)
{
    slsk_user_t* user = g_hash_table_lookup(SlskUserModule->users, nick);
    if (!user)
    {
        user = slsk_user_new(nick);
        g_hash_table_insert(SlskUserModule->users, OBJ(user)->name, user);
    }
    return user;
}

void slsk_user_set_privileges(slsk_user_t* user, int priv)
{
    g_assert(SlskUserModule);
    if (user->privileges == priv) return;
    if (user->privileges == SLSK_PRIV_ON) SlskUserModule->privileged--;
    user->privileges = priv;
    if (user->privileges == SLSK_PRIV_ON) SlskUserModule->privileged++;
    object_set_dirty(OBJ(user));
}

int slsk_user_get_privileges(const char* nick)
{
    slsk_user_t* user = slsk_user_find(nick);
    g_assert(SlskUserModule);
    return user->privileges;
}

int slsk_user_get_status(const char* nick)
{
    slsk_user_t* user = slsk_user_find(nick);
    g_assert(SlskUserModule);
    return user->status;
}

gboolean slsk_user_ip_known(slsk_user_t* user)
{
    g_assert(SlskUserModule);
    return (USER(user)->ip != 0 &&  user->refreshed >= CurrentTime-60*60);
}

void slsk_user_set_ip_port(slsk_user_t* user, ip_t ip, port_t port)
{
    g_assert(SlskUserModule);
    user->refreshed = CurrentTime;
    user_set_ip_port(USER(user), ip, port);
}

void slsk_user_set_status(slsk_user_t* user, int status)
{
    g_assert(SlskUserModule);
    if (user->status == status) return;
    user->status = status;
    user_set_online(USER(user), status>SLSK_USER_OFFLINE);
}

void slsk_user_set_net(slsk_user_t* user, slsk_net_t* net)
{
    g_assert(SlskUserModule);
    if (user->net == net) return;

    if (user->net)
    {
        object_unnotify_o(OBJ(user->net), (ObjectFunc)on_slsk_user_net_removed, OBJ(user));
        object_unmonitor_o(OBJ(user->net), (ObjectFunc)on_slsk_user_net_changed, OBJ(user));
    }

    user->net = net;

    if (user->net)
    {
        object_notify_o(OBJ(user->net), (ObjectFunc)on_slsk_user_net_removed, OBJ(user));
        object_monitor_o(OBJ(user->net), (ObjectFunc)on_slsk_user_net_changed, OBJ(user));
    }

    object_set_dirty(OBJ(user));
}

void slsk_user_set_info(slsk_user_t* user, int speed,
                        int down, int shared, int dirs)
{
    g_assert(SlskUserModule);
    if (user->speed != speed) user->speed = speed;
    if (user->downloads != down) user->downloads = down;
    if (user->shared != shared) user->shared = shared;
    if (user->shared_dirs != dirs) user->shared_dirs = dirs;
    object_set_dirty(OBJ(user));
}

gboolean slsk_user_online(slsk_user_t* user)
{
    g_assert(SlskUserModule);
    return (user && user->net &&
            (user->status == SLSK_USER_ONLINE ||
             user->status == SLSK_USER_AWAY));
}

void slsk_user_set_whois(slsk_user_t* user, int slots,
                         int queued, gboolean has_slots)
{
    g_assert(SlskUserModule);
    user->all_slots = slots;
    user->queued = queued;
    user->has_slots = has_slots;
    object_set_dirty(OBJ(user));
}

static void set_privs(const char* nick, slsk_user_t* user)
{
    (void)nick;
    if (user->privileges == SLSK_PRIV_UNKNOWN)
        slsk_user_set_privileges(user, SLSK_PRIV_OFF);
}

void slsk_user_set_know_privileges()
{
    g_assert(SlskUserModule);
    g_hash_table_foreach(SlskUserModule->users, (GHFunc)set_privs, NULL);
    SlskUserModule->got_privs = TRUE;
    {
        unsigned num = g_hash_table_size(SlskUserModule->users);
        const char* fmt = 
            ngettext("There is %u privileged user", "There are %u privileged users", num);
        server_message("SoulSeek", l_printf(fmt, num));
    }
}



static void slsk_users_init()
{
    // the hash table does not hold any real references to the users,
    // so we don't have a item destroy method.
    SlskUserModule->users =
        g_hash_table_new_full((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal, NULL, NULL);
}

static void slsk_users_exit()
{
    g_log("SLSK_USERS", G_LOG_LEVEL_DEBUG,
          "we have %u in cache consuming approx. %u bytes", 
          g_hash_table_size(SlskUserModule->users),
          (int)(g_hash_table_size(SlskUserModule->users)*sizeof(slsk_user_t)));
    
    g_hash_table_destroy(SlskUserModule->users);
    SlskUserModule->users = NULL;
    SlskUserModule = NULL;
}

static void slsk_user_module_init(slsk_user_module_t* module)
{
    module_init(MODULE(module), "Soulseek User Module",
                (ModuleFunc)slsk_users_init, NULL, (ModuleFunc)slsk_users_exit);
    
    module->users = NULL;
    module->privileged = 0;
    module->got_privs = FALSE;
    
    module_require(base_module_instance);
    module_require(user_module_instance);
    module_require(slsk_net_module_instance);
}


module_t* slsk_user_module_instance()
{
    if (!SlskUserModule)
    {
        SlskUserModule = g_malloc0(sizeof(*SlskUserModule));
        slsk_user_module_init(SlskUserModule);
    }
    return MODULE(SlskUserModule);
}

#ifdef WIN32
#define _SLSK_USER_C_
#include "slskaliasdef.c"
#endif

